#!/usr/bin/python3

version = "4.80.2"

_change_log = """
    Changelog since 4.70.1 released to PyPI on 20-Jul-2024

    """

__version__ = version  # For PEP 396 and PEP 345

# The shortened version of version
ver = version


# INFO START
port = "PySimpleGUI"
__license__ = "MIT License"
__author__ = "yunluo"
__email__ = "sp91@qq.com"
# INFO END

# 8""""8        8""""8                             8""""8 8   8 8
# 8    8 e    e 8      e  eeeeeee eeeee e     eeee 8    " 8   8 8
# 8eeee8 8    8 8eeeee 8  8  8  8 8   8 8     8    8e     8e  8 8e
# 88     8eeee8     88 8e 8e 8  8 8eee8 8e    8eee 88  ee 88  8 88
# 88       88   e   88 88 88 8  8 88    88    88   88   8 88  8 88
# 88       88   8eee88 88 88 8  8 88    88eee 88ee 88eee8 88ee8 88


"""
    Copyright 2018, 2019, 2020, 2021, 2022, 2023, 2024 PySimpleGUI(tm)
    Copyright 2024 PySimpleSoft, Inc.
    Copyright 2024 MB

    From the inception these have been the project principals upon which it is all built
    1. Fun - it's a serious goal of the project. If we're not having FUN while making stuff, then something's not right
    2. Successful - you need to be successful or it's all for naught
    3. You are the important party - It's your success that determines the success of PySimpleGUI

    If these 3 things are kept at the forefront, then the rest tends to fall into place.

    Please consider sponsoring all open source developers that make software you or your business use. They need your help.
    

    This software is available for your use under a LGPL3+ license

    This notice, these first 150 lines of code shall remain unchanged



    888      .d8888b.  8888888b.  888      .d8888b.          
    888     d88P  Y88b 888   Y88b 888     d88P  Y88b         
    888     888    888 888    888 888          .d88P
    888     888        888   d88P 888         8888"    888   
    888     888  88888 8888888P"  888          "Y8b. 8888888 
    888     888    888 888        888     888    888   888   
    888     Y88b  d88P 888        888     Y88b  d88P         
    88888888 "Y8888P88 888        88888888 "Y8888P"          


    In addition to the normal publishing requirements of LGPL3+, these also apply:

    A final note from mike...
    
        “Don’t aim at success. The more you aim at it and make it a target, the more you are going to miss it. 
        For success, like happiness, cannot be pursued; it must ensue, and it only does so as the unintended side effect of one’s personal dedication to a cause greater.”
            — Viktor Frankl
    
        I first saw this quote in a truncated format:
            "Happiness, cannot be pursued; it must ensue, and it only does so as the unintended side effect of one’s personal dedication to a cause greater."    
    
        Everyone is different, but my experience with the PySimpleGUI project matches this theory.  It's taken a lifetime of trying and "failing" and trying
        to find happiness before I finally figured this truth-for-me out.  If I do a long list of things, and live life in a kind & loving way, then the
        result is happiness.  It's a biproduct, not a directly produced thing.  This should be taught in school.  Or maybe it can't.
        I hope you find happiness, but more importantly, or maybe first, I hope you find that bigger-than-you thing. For me it's always been programming.  It seems to be
        the giving back part, not just the calling, that makes the happiness fusion-reactor operate.

    "Thank you" has fueled this project. I'm incredibly grateful to have users that are in turn grateful. It's a feedback loop of gratitude. What a fantastic thing!
"""
# all of the tkinter involved imports
import tkinter as tk
# import tkinter.scrolledtext as tkst
import tkinter.font
from tkinter import filedialog
from tkinter import ttk
from tkinter.colorchooser import askcolor

# end of tkinter specific imports
# get the tkinter detailed version
tclversion_detailed = tkinter.Tcl().eval("info patchlevel")
framework_version = tclversion_detailed

import time
import pickle
import calendar
import datetime
import textwrap

import socket
from hashlib import sha256 as hh
import inspect
import traceback
import difflib
import copy
import pprint

try:  # Because Raspberry Pi is still on 3.4....it's not critical if this module isn't imported on the Pi
    from typing import (
        List,
        Any,
        Union,
        Tuple,
        Dict,
        SupportsAbs,
        Optional,
    )  # because this code has to run on 2.7 can't use real type hints.  Must do typing only in comments
except Exception:
    print(
        '*** Skipping import of Typing module. "pip3 install typing" to remove this warning ***'
    )
import random
import warnings
from math import floor
from math import fabs
from functools import wraps

try:  # Because Raspberry Pi is still on 3.4....
    # from subprocess import run, PIPE, Popen
    import subprocess
except Exception as e:
    print("** Import error {} **".format(e))

import threading
import itertools
import json
import configparser
import queue

try:
    import webbrowser

    webbrowser_available = True
except Exception:
    webbrowser_available = False
# used for github upgrades
import urllib.request
import urllib.error
import urllib.parse
import pydoc
import os
import sys
import ctypes
import platform

pil_import_attempted = pil_imported = False

warnings.simplefilter("always", UserWarning)

g_time_start = 0
g_time_end = 0
g_time_delta = 0


# These timer routines are to help you quickly time portions of code.  Place the timer_start call at the point
# you want to start timing and the timer_stop at the end point. The delta between the start and stop calls
# is returned from calling timer_stop


def timer_start():
    """
    Time your code easily.... starts the timer.
    Uses the time.time value, a technique known to not be terribly accurage, but tis' gclose enough for our purposes
    """
    global g_time_start

    g_time_start = time.time()


def timer_stop():
    """
    Time your code easily.... stop the timer and print the number of MILLISECONDS since the timer start

    :return: delta in MILLISECONDS from timer_start was called
    :rtype:  int
    """
    global g_time_delta, g_time_end

    g_time_end = time.time()
    g_time_delta = g_time_end - g_time_start
    return int(g_time_delta * 1000)


def timer_stop_usec():
    """
    Time your code easily.... stop the timer and print the number of MICROSECONDS since the timer start

    :return: delta in MICROSECONDS from timer_start was called
    :rtype:  int
    """
    global g_time_delta, g_time_end

    g_time_end = time.time()
    g_time_delta = g_time_end - g_time_start
    return int(g_time_delta * 1000000)


def _timeit(func):
    """
    Put @_timeit as a decorator to a function to get the time spent in that function printed out

    :param func: Decorated function
    :type func:
    :return:     Execution time for the decorated function
    :rtype:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print("{} executed in {:.4f} seconds".format(func.__name__, end - start))
        return result

    return wrapper


_timeit_counter = 0
MAX_TIMEIT_COUNT = 1000
_timeit_total = 0


def _timeit_summary(func):
    """
    Same as the timeit decorator except that the value is shown as an averave
    Put @_timeit_summary as a decorator to a function to get the time spent in that function printed out

    :param func: Decorated function
    :type func:
    :return:     Execution time for the decorated function
    :rtype:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        global _timeit_counter, _timeit_total

        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        _timeit_counter += 1
        _timeit_total += end - start
        if _timeit_counter > MAX_TIMEIT_COUNT:
            print(
                "{} executed in {:.4f} seconds".format(
                    func.__name__, _timeit_total / MAX_TIMEIT_COUNT
                )
            )
            _timeit_counter = 0
            _timeit_total = 0
        return result

    return wrapper


def formatted_datetime_now():
    """
    Returns a string with current date and time formatted YYYY-MM-DD HH:MM:SS for easy logging

    :return:    String with date and time formatted YYYY-MM-DD  HH:MM:SS
    :rtype:     (str)
    """
    now = datetime.datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    return current_time


def running_linux():
    """
    Determines the OS is Linux by using sys.platform

    Returns True if Linux

    :return: True if sys.platform indicates running Linux
    :rtype:  (bool)
    """
    return sys.platform.startswith("linux")


def running_mac():
    """
    Determines the OS is Mac by using sys.platform

    Returns True if Mac

    :return: True if sys.platform indicates running Mac
    :rtype:  (bool)
    """
    return sys.platform.startswith("darwin")


def running_windows():
    """
    Determines the OS is Windows by using sys.platform

    Returns True if Windows

    :return: True if sys.platform indicates running Windows
    :rtype:  (bool)
    """
    return sys.platform.startswith("win")


def running_trinket():
    """
    A special case for Trinket.  Checks both the OS and the number of environment variables
    Currently, Trinket only has ONE environment variable.  This fact is used to figure out if Trinket is being used.

    Returns True if "Trinket" (in theory)

    :return: True if sys.platform indicates Linux and the number of environment variables is 1
    :rtype:  (bool)
    """
    if sys.platform.startswith("linux") and socket.gethostname().startswith("pygame-"):
        return True
    return False


def running_replit():
    """
    A special case for REPLIT.  Checks both the OS and for the existance of the number of environment variable REPL_OWNER
    Currently, Trinket only has ONE environment variable.  This fact is used to figure out if Trinket is being used.

    Returns True if running on "replit"

    :return: True if sys.platform indicates Linux and setting REPL_OWNER is found in the environment variables
    :rtype:  (bool)
    """
    if "REPL_OWNER" in os.environ and sys.platform.startswith("linux"):
        return True
    return False


# Handy python statements to increment and decrement with wrapping that I don't want to forget
# count = (count + (MAX - 1)) % MAX           # Decrement - roll over to MAX from 0
# count = (count + 1) % MAX                   # Increment to MAX then roll over to 0

"""
    Welcome to the "core" PySimpleGUI code....

    It's a mess.... really... it's a mess internally... it's the external-facing interfaces that
    are not a mess.  The Elements and the methods for them are well-designed.
    PEP8 - this code is far far from PEP8 compliant. 
    It was written PRIOR to learning that PEP8 existed. 

    I'll be honest.... started learning Python in Nov 2017, started writing PySimpleGUI in Feb 2018.
    Released PySimpleGUI in July 2018.  I knew so little about Python that my parameters were all named
    using CamelCase.  DOH!  Someone on Reddit set me straight on that.  So overnight I renamed all of the
    parameters to lower case.  Unfortunately, the internal naming conventions have been set.  Mixing them
    with PEP8 at this moment would be even MORE confusing.

    Code I write now, outside PySimpleGUI, IS PEP8 compliant.  

    The variable and function naming in particular are not compliant.  There is
    liberal use of CamelVariableAndFunctionNames, but for anything externally facing, there are aliases
    available for all functions.  If you've got a serious enough problem with 100% PEP8 compliance
    that you'll pass on this package, then that's your right and I invite you to do so.  However, if
    perhaps you're a practical thinker where it's the results that matter, then you'll have no
    trouble with this code base.  There is consisency however.  

    I truly hope you get a lot of enjoyment out of using PySimpleGUI.  It came from good intentions.
"""

# ----====----====----==== Constants the user CAN safely change ====----====----====----#

# Base64 encoded GIF file
DEFAULT_BASE64_ICON = b"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"

DEFAULT_BASE64_ICON_16_BY_16 = b"iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAKCSURBVDhPVZNbSFRRFIb35YwXItBIGtDsiqENEUTRjJlZkJggPSUYBD0UhULElE6hBY6ID/ZSpD1IDxaCEPhUaFLRQyWRNxIJe8syMxCjMCbB07fOsaMt+GftvWf//7/2Whyt1sTei/fCpDqQBTrGOi9Myrk7URwhnQUfQLeOvErJuUQgADlK6gObvAOl5sHx0doHljwARFRiCpxG5J1sjPxALiYNgn9kiQ3gafdYUYzseCd+FICX7sShw7LR++q6cl3XHaXQHFdOJLxFsJtvKHnbUr1nqp01hhStpXAzo7TZZXOjJ+9orT9pY74aY3ZobZZYW8D/GpjM19Ob088fmJxW2tkC4AJt17Oeg2MLrHX6jXWes16w1sbBkrFWBTB2nTLpv5VJg7wGNhRDwCS0tR1cbECkidwMQohAdoScqiz8/FCZUKlPCgSWlQ71elOI1fcco9hCXp1kS7dX3u+qVOm2L4nW8qE4Neetvl8v83NOb++9703BcUI/cU3imuWV7JedKtv5LdFaMRzHLW+N+zJoVDZzRLj6SFNfPlMYwy5bDiRcCojmz15tKx+6hKPv7LvjrG/Q2RoOwjSyzNDlahyzA2dAJeNtFcMHA2cfLn24STNr6P4I728jJ7hvf/lEGuaXLnkRAp0PyFK+hlyLSJGyGWnKyeBi2oJU0IPIjNd15uuL2f2PJgueQBKhVRETCgNeYU+xaeEpnWaw8cQPRM7g/McT8eF0De9u7P+49TqXF7no98BDEEkdvvXem8LAtfJniFRB/A5XeiAiG2+/icgHVQUW5d5KyAhl3M2y+U+ysv1FDukyKGQW3Y+vHJWvU7mz8RJSPZgDd3H2RqiUUn8BSQuaBvGjGpsAAAAASUVORK5CYII="

DEFAULT_BASE64_LOADING_GIF = b" "

# Old debugger logo
# PSG_DEBUGGER_LOGO = b'R0lGODlhMgAtAPcAAAAAADD/2akK/4yz0pSxyZWyy5u3zZ24zpW30pG52J250J+60aC60KS90aDC3a3E163F2K3F2bPI2bvO3rzP3qvJ4LHN4rnR5P/zuf/zuv/0vP/0vsDS38XZ6cnb6f/xw//zwv/yxf/1w//zyP/1yf/2zP/3z//30wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAP8ALAAAAAAyAC0AAAj/AP8JHEiwoMGDCBMqXMiwoUOFAiJGXBigYoAPDxlK3CigwUGLIAOEyIiQI8cCBUOqJFnQpEkGA1XKZPlPgkuXBATK3JmRws2bB3TuXNmQw8+jQoeCbHj0qIGkSgNobNoUqlKIVJs++BfV4oiEWalaHVpyosCwJidw7Sr1YMQFBDn+y4qSbUW3AiDElXiWqoK1bPEKGLixr1jAXQ9GuGn4sN22Bl02roo4Kla+c8OOJbsQM9rNPJlORlr5asbPpTk/RP2YJGu7rjWnDm2RIQLZrSt3zgp6ZmqwmkHAng3ccWDEMe8Kpnw8JEHlkXnPdh6SxHPILaU/dp60LFUP07dfRq5aYntohAO0m+c+nvT6pVMPZ3jv8AJu8xktyNbw+ATJDtKFBx9NlA20gWU0DVQBYwZhsJMICRrkwEYJJGRCSBtEqGGCAQEAOw=='

PSG_DEBUGGER_LOGO = b" "

UDEMY_ICON = b" "

BLANK_BASE64 = b"iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAADsMAAA7DAcdvqGQAAAANSURBVBhXY2BgYGAAAAAFAAGKM+MAAAAAAElFTkSuQmCC"
BLANK_BASE64 = b"iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII="

DEFAULT_WINDOW_ICON = DEFAULT_BASE64_ICON

DEFAULT_ELEMENT_SIZE = (45, 1)  # In CHARACTERS
DEFAULT_BUTTON_ELEMENT_SIZE = (10, 1)  # In CHARACTERS
DEFAULT_MARGINS = (10, 5)  # Margins for each LEFT/RIGHT margin is first term
DEFAULT_ELEMENT_PADDING = (5, 3)  # Padding between elements (row, col) in pixels
DEFAULT_AUTOSIZE_TEXT = True
DEFAULT_AUTOSIZE_BUTTONS = True
DEFAULT_FONT = ("Helvetica", 10)
DEFAULT_TEXT_JUSTIFICATION = "left"
DEFAULT_BORDER_WIDTH = 1
DEFAULT_AUTOCLOSE_TIME = 3  # time in seconds to show an autoclose form
DEFAULT_DEBUG_WINDOW_SIZE = (80, 20)
DEFAULT_WINDOW_LOCATION = (None, None)
MAX_SCROLLED_TEXT_BOX_HEIGHT = 50
DEFAULT_TOOLTIP_TIME = 400
DEFAULT_TOOLTIP_OFFSET = (0, -20)
DEFAULT_KEEP_ON_TOP = None
DEFAULT_SCALING = None
DEFAULT_ALPHA_CHANNEL = 1.0
DEFAULT_HIDE_WINDOW_WHEN_CREATING = True
TOOLTIP_BACKGROUND_COLOR = "#ffffe0"
TOOLTIP_FONT = None
DEFAULT_USE_BUTTON_SHORTCUTS = False
#################### COLOR STUFF ####################
BLUES = ("#082567", "#0A37A3", "#00345B")
PURPLES = ("#480656", "#4F2398", "#380474")
GREENS = ("#01826B", "#40A860", "#96D2AB", "#00A949", "#003532")
YELLOWS = ("#F3FB62", "#F0F595")
TANS = ("#FFF9D5", "#F4EFCF", "#DDD8BA")
NICE_BUTTON_COLORS = (
    (GREENS[3], TANS[0]),
    ("#000000", "#FFFFFF"),
    ("#FFFFFF", "#000000"),
    (YELLOWS[0], PURPLES[1]),
    (YELLOWS[0], GREENS[3]),
    (YELLOWS[0], BLUES[2]),
)

COLOR_SYSTEM_DEFAULT = "1234567890"  # A Magic Number kind of signal to PySimpleGUI that the color should not be set at all
DEFAULT_BUTTON_COLOR = (
    "white",
    BLUES[0],
)  # Foreground, Background (None, None) == System Default
OFFICIAL_PYSIMPLEGUI_BUTTON_COLOR = ("white", BLUES[0])

# The "default PySimpleGUI theme"
OFFICIAL_PYSIMPLEGUI_THEME = CURRENT_LOOK_AND_FEEL = "Dark Blue 3"

DEFAULT_ERROR_BUTTON_COLOR = ("#FFFFFF", "#FF0000")
DEFAULT_BACKGROUND_COLOR = None
DEFAULT_ELEMENT_BACKGROUND_COLOR = None
DEFAULT_ELEMENT_TEXT_COLOR = COLOR_SYSTEM_DEFAULT
DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR = None
DEFAULT_TEXT_COLOR = COLOR_SYSTEM_DEFAULT
DEFAULT_INPUT_ELEMENTS_COLOR = COLOR_SYSTEM_DEFAULT
DEFAULT_INPUT_TEXT_COLOR = COLOR_SYSTEM_DEFAULT
DEFAULT_SCROLLBAR_COLOR = None
# DEFAULT_BUTTON_COLOR = (YELLOWS[0], PURPLES[0])    # (Text, Background) or (Color "on", Color) as a way to remember
# DEFAULT_BUTTON_COLOR = (GREENS[3], TANS[0])    # Foreground, Background (None, None) == System Default
# DEFAULT_BUTTON_COLOR = (YELLOWS[0], GREENS[4])    # Foreground, Background (None, None) == System Default
# DEFAULT_BUTTON_COLOR = ('white', 'black')    # Foreground, Background (None, None) == System Default
# DEFAULT_BUTTON_COLOR = (YELLOWS[0], PURPLES[2])    # Foreground, Background (None, None) == System Default
# DEFAULT_PROGRESS_BAR_COLOR = (GREENS[2], GREENS[0])     # a nice green progress bar
# DEFAULT_PROGRESS_BAR_COLOR = (BLUES[1], BLUES[1])     # a nice green progress bar
# DEFAULT_PROGRESS_BAR_COLOR = (BLUES[0], BLUES[0])     # a nice green progress bar
# DEFAULT_PROGRESS_BAR_COLOR = (PURPLES[1],PURPLES[0])    # a nice purple progress bar


# A transparent button is simply one that matches the background
# TRANSPARENT_BUTTON = 'This constant has been depricated. You must set your button background = background it is on for it to be transparent appearing'


# --------------------------------------------------------------------------------
# Progress Bar Relief Choices
RELIEF_RAISED = "raised"
RELIEF_SUNKEN = "sunken"
RELIEF_FLAT = "flat"
RELIEF_RIDGE = "ridge"
RELIEF_GROOVE = "groove"
RELIEF_SOLID = "solid"
RELIEF_LIST = (
    RELIEF_RAISED,
    RELIEF_FLAT,
    RELIEF_SUNKEN,
    RELIEF_RIDGE,
    RELIEF_SOLID,
    RELIEF_GROOVE,
)

# These are the spepific themes that tkinter offers
THEME_DEFAULT = "default"  # this is a TTK theme, not a PSG theme!!!
THEME_WINNATIVE = "winnative"
THEME_CLAM = "clam"
THEME_ALT = "alt"
THEME_CLASSIC = "classic"
THEME_VISTA = "vista"
THEME_XPNATIVE = "xpnative"

# The theme to use by default for all windows
DEFAULT_TTK_THEME = THEME_DEFAULT
ttk_theme_in_use = None
# TTK_THEME_LIST = ('default', 'winnative', 'clam', 'alt', 'classic', 'vista', 'xpnative')


USE_TTK_BUTTONS = None

DEFAULT_PROGRESS_BAR_COLOR = ("#01826B", "#D0D0D0")  # a nice green progress bar
DEFAULT_PROGRESS_BAR_COMPUTE = (
    "#000000",
    "#000000",
)  # Means that the progress bar colors should be computed from other colors
DEFAULT_PROGRESS_BAR_COLOR_OFFICIAL = (
    "#01826B",
    "#D0D0D0",
)  # a nice green progress bar
DEFAULT_PROGRESS_BAR_SIZE = (
    20,
    20,
)  # Size of Progress Bar (characters for length, pixels for width)
DEFAULT_PROGRESS_BAR_BORDER_WIDTH = 1
DEFAULT_PROGRESS_BAR_RELIEF = RELIEF_GROOVE
# PROGRESS_BAR_STYLES = ('default', 'winnative', 'clam', 'alt', 'classic', 'vista', 'xpnative')
DEFAULT_PROGRESS_BAR_STYLE = DEFAULT_TTK_THEME
DEFAULT_METER_ORIENTATION = "Horizontal"
DEFAULT_SLIDER_ORIENTATION = "vertical"
DEFAULT_SLIDER_BORDER_WIDTH = 1
DEFAULT_SLIDER_RELIEF = tk.FLAT
DEFAULT_FRAME_RELIEF = tk.GROOVE

DEFAULT_LISTBOX_SELECT_MODE = tk.SINGLE
SELECT_MODE_MULTIPLE = tk.MULTIPLE
LISTBOX_SELECT_MODE_MULTIPLE = "multiple"
SELECT_MODE_BROWSE = tk.BROWSE
LISTBOX_SELECT_MODE_BROWSE = "browse"
SELECT_MODE_EXTENDED = tk.EXTENDED
LISTBOX_SELECT_MODE_EXTENDED = "extended"
SELECT_MODE_SINGLE = tk.SINGLE
LISTBOX_SELECT_MODE_SINGLE = "single"

TABLE_SELECT_MODE_NONE = tk.NONE
TABLE_SELECT_MODE_BROWSE = tk.BROWSE
TABLE_SELECT_MODE_EXTENDED = tk.EXTENDED
DEFAULT_TABLE_SELECT_MODE = TABLE_SELECT_MODE_EXTENDED
TABLE_CLICKED_INDICATOR = "+CLICKED+"  # Part of the tuple returned as an event when a Table element has click events enabled
DEFAULT_MODAL_WINDOWS_ENABLED = True
DEFAULT_MODAL_WINDOWS_FORCED = False

TAB_LOCATION_TOP = "top"
TAB_LOCATION_TOP_LEFT = "topleft"
TAB_LOCATION_TOP_RIGHT = "topright"
TAB_LOCATION_LEFT = "left"
TAB_LOCATION_LEFT_TOP = "lefttop"
TAB_LOCATION_LEFT_BOTTOM = "leftbottom"
TAB_LOCATION_RIGHT = "right"
TAB_LOCATION_RIGHT_TOP = "righttop"
TAB_LOCATION_RIGHT_BOTTOM = "rightbottom"
TAB_LOCATION_BOTTOM = "bottom"
TAB_LOCATION_BOTTOM_LEFT = "bottomleft"
TAB_LOCATION_BOTTOM_RIGHT = "bottomright"

TITLE_LOCATION_TOP = tk.N
TITLE_LOCATION_BOTTOM = tk.S
TITLE_LOCATION_LEFT = tk.W
TITLE_LOCATION_RIGHT = tk.E
TITLE_LOCATION_TOP_LEFT = tk.NW
TITLE_LOCATION_TOP_RIGHT = tk.NE
TITLE_LOCATION_BOTTOM_LEFT = tk.SW
TITLE_LOCATION_BOTTOM_RIGHT = tk.SE

TEXT_LOCATION_TOP = tk.N
TEXT_LOCATION_BOTTOM = tk.S
TEXT_LOCATION_LEFT = tk.W
TEXT_LOCATION_RIGHT = tk.E
TEXT_LOCATION_TOP_LEFT = tk.NW
TEXT_LOCATION_TOP_RIGHT = tk.NE
TEXT_LOCATION_BOTTOM_LEFT = tk.SW
TEXT_LOCATION_BOTTOM_RIGHT = tk.SE
TEXT_LOCATION_CENTER = tk.CENTER

GRAB_ANYWHERE_IGNORE_THESE_WIDGETS = (
    ttk.Sizegrip,
    tk.Scale,
    ttk.Scrollbar,
    tk.Scrollbar,
    tk.Entry,
    tk.Text,
    tk.PanedWindow,
    tk.Listbox,
    tk.OptionMenu,
    ttk.Treeview,
)

# ----====----====----==== Constants the user should NOT f-with ====----====----====----#
ThisRow = 555666777  # magic number

# DEFAULT_WINDOW_ICON = ''
MESSAGE_BOX_LINE_WIDTH = 60

# "Special" Key Values.. reserved
# Key representing a Read timeout
EVENT_TIMEOUT = TIMEOUT_EVENT = TIMEOUT_KEY = "__TIMEOUT__"
EVENT_TIMER = TIMER_KEY = "__TIMER EVENT__"
WIN_CLOSED = WINDOW_CLOSED = None
WINDOW_CLOSE_ATTEMPTED_EVENT = WIN_X_EVENT = WIN_CLOSE_ATTEMPTED_EVENT = (
    "-WINDOW CLOSE ATTEMPTED-"
)
WINDOW_CONFIG_EVENT = "__WINDOW CONFIG__"
TITLEBAR_MINIMIZE_KEY = "__TITLEBAR MINIMIZE__"
TITLEBAR_MAXIMIZE_KEY = "__TITLEBAR MAXIMIZE__"
TITLEBAR_CLOSE_KEY = "__TITLEBAR CLOSE__"
TITLEBAR_IMAGE_KEY = "__TITLEBAR IMAGE__"
TITLEBAR_TEXT_KEY = "__TITLEBAR TEXT__"
TITLEBAR_DO_NOT_USE_AN_ICON = "__TITLEBAR_NO_ICON__"

# Key indicating should not create any return values for element
WRITE_ONLY_KEY = "__WRITE ONLY__"

MENU_DISABLED_CHARACTER = "!"
MENU_SHORTCUT_CHARACTER = "&"
MENU_KEY_SEPARATOR = "::"
MENU_SEPARATOR_LINE = "---"
MENU_RIGHT_CLICK_EDITME_EXIT = ["", ["Edit Me", "Exit"]]
MENU_RIGHT_CLICK_EDITME_VER_EXIT = ["", ["Edit Me", "Version", "Exit"]]
MENU_RIGHT_CLICK_EDITME_VER_LOC_EXIT = [
    "",
    ["Edit Me", "Version", "File Location", "Exit"],
]
MENU_RIGHT_CLICK_EDITME_VER_SETTINGS_EXIT = [
    "",
    ["Edit Me", "Settings", "Version", "Exit"],
]
MENU_RIGHT_CLICK_EXIT = ["", ["Exit"]]
MENU_RIGHT_CLICK_DISABLED = ["", []]
_MENU_RIGHT_CLICK_TABGROUP_DEFAULT = ["TABGROUP DEFAULT", []]
ENABLE_TK_WINDOWS = False

USE_CUSTOM_TITLEBAR = None
CUSTOM_TITLEBAR_BACKGROUND_COLOR = None
CUSTOM_TITLEBAR_TEXT_COLOR = None
CUSTOM_TITLEBAR_ICON = None
CUSTOM_TITLEBAR_FONT = None
TITLEBAR_METADATA_MARKER = "This window has a titlebar"

CUSTOM_MENUBAR_METADATA_MARKER = "This is a custom menubar"

SUPPRESS_ERROR_POPUPS = False
SUPPRESS_RAISE_KEY_ERRORS = True
SUPPRESS_KEY_GUESSING = False
SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS = False
ENABLE_TREEVIEW_869_PATCH = True

# These are now set based on the global settings file
ENABLE_MAC_NOTITLEBAR_PATCH = False
ENABLE_MAC_MODAL_DISABLE_PATCH = False
ENABLE_MAC_DISABLE_GRAB_ANYWHERE_WITH_TITLEBAR = True
ENABLE_MAC_ALPHA_99_PATCH = False

OLD_TABLE_TREE_SELECTED_ROW_COLORS = ("#FFFFFF", "#4A6984")
ALTERNATE_TABLE_AND_TREE_SELECTED_ROW_COLORS = (
    "SystemHighlightText",
    "SystemHighlight",
)

# Some handy unicode symbols
SYMBOL_SQUARE = "█"
SYMBOL_CIRCLE = "⚫"
SYMBOL_CIRCLE_OUTLINE = "◯"
SYMBOL_BULLET = "•"
SYMBOL_UP = "▲"
SYMBOL_RIGHT = "►"
SYMBOL_LEFT = "◄"
SYMBOL_DOWN = "▼"
SYMBOL_X = "❎"
SYMBOL_CHECK = "✅"
SYMBOL_CHECK_SMALL = "✓"
SYMBOL_X_SMALL = "✗"
SYMBOL_BALLOT_X = "☒"
SYMBOL_BALLOT_CHECK = "☑"
SYMBOL_LEFT_DOUBLE = "«"
SYMBOL_RIGHT_DOUBLE = "»"
SYMBOL_LEFT_ARROWHEAD = "⮜"
SYMBOL_RIGHT_ARROWHEAD = "⮞"
SYMBOL_UP_ARROWHEAD = "⮝"
SYMBOL_DOWN_ARROWHEAD = "⮟"

if sum([int(i) for i in tclversion_detailed.split(".")]) > 19:
    SYMBOL_TITLEBAR_MINIMIZE = "_"
    SYMBOL_TITLEBAR_MAXIMIZE = "◻"
    SYMBOL_TITLEBAR_CLOSE = "Ｘ"
else:
    SYMBOL_TITLEBAR_MINIMIZE = "_"
    SYMBOL_TITLEBAR_MAXIMIZE = "O"
    SYMBOL_TITLEBAR_CLOSE = "X"

#################### PATHS for user_settings APIs ####################
# These paths are passed to os.path.expanduser to get the default path for user_settings
# They can be changed using set_options

DEFAULT_USER_SETTINGS_WIN_PATH = r"~\AppData\Local\PySimpleGUI\settings"
DEFAULT_USER_SETTINGS_LINUX_PATH = r"~/.config/PySimpleGUI/settings"
DEFAULT_USER_SETTINGS_MAC_PATH = r"~/Library/Application Support/PySimpleGUI/settings"
DEFAULT_USER_SETTINGS_TRINKET_PATH = r"."
DEFAULT_USER_SETTINGS_REPLIT_PATH = r"."
DEFAULT_USER_SETTINGS_UNKNOWN_OS_PATH = (
    r"~/Library/Application Support/PySimpleGUI/settings"
)
DEFAULT_USER_SETTINGS_PATH = None  # value set by user to override all paths above
DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH = (
    None  # location of the global PySimpleGUI settings
)
DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME = (
    "_PySimpleGUI_settings_global_.json"  # location of the global PySimpleGUI settings
)


# ====================================================================== #
# One-liner functions that are handy as f_ck                             #
# ====================================================================== #
def rgb(red, green, blue):
    """
    Given integer values of Red, Green, Blue, return a color string "#RRGGBB"
    :param red:   Red portion from 0 to 255
    :type red:    (int)
    :param green: Green portion from 0 to 255
    :type green:  (int)
    :param blue:  Blue portion from 0 to 255
    :type  blue:  (int)
    :return:      A single RGB String in the format "#RRGGBB" where each pair is a hex number.
    :rtype:       (str)
    """
    red = min(int(red), 255) if red > 0 else 0
    blue = min(int(blue), 255) if blue > 0 else 0
    green = min(int(green), 255) if green > 0 else 0
    return "#%02x%02x%02x" % (red, green, blue)


# ====================================================================== #
# Enums for types                                                        #
# ====================================================================== #
# -------------------------  Button types  ------------------------- #
# uncomment this line and indent to go back to using Enums
BUTTON_TYPE_BROWSE_FOLDER = 1
BUTTON_TYPE_BROWSE_FILE = 2
BUTTON_TYPE_BROWSE_FILES = 21
BUTTON_TYPE_SAVEAS_FILE = 3
BUTTON_TYPE_CLOSES_WIN = 5
BUTTON_TYPE_CLOSES_WIN_ONLY = 6
BUTTON_TYPE_READ_FORM = 7
BUTTON_TYPE_REALTIME = 9
BUTTON_TYPE_CALENDAR_CHOOSER = 30
BUTTON_TYPE_COLOR_CHOOSER = 40
BUTTON_TYPE_SHOW_DEBUGGER = 50

BROWSE_FILES_DELIMITER = (
    ";"  # the delimiter to be used between each file in the returned string
)

FILE_TYPES_ALL_FILES = (("ALL Files", "*.* *"),)

BUTTON_DISABLED_MEANS_IGNORE = "ignore"

# -------------------------  Element types  ------------------------- #

ELEM_TYPE_TEXT = "text"
ELEM_TYPE_INPUT_TEXT = "input"
ELEM_TYPE_INPUT_COMBO = "combo"
ELEM_TYPE_INPUT_OPTION_MENU = "option menu"
ELEM_TYPE_INPUT_RADIO = "radio"
ELEM_TYPE_INPUT_MULTILINE = "multiline"
ELEM_TYPE_INPUT_CHECKBOX = "checkbox"
ELEM_TYPE_INPUT_SPIN = "spind"
ELEM_TYPE_BUTTON = "button"
ELEM_TYPE_IMAGE = "image"
ELEM_TYPE_CANVAS = "canvas"
ELEM_TYPE_FRAME = "frame"
ELEM_TYPE_GRAPH = "graph"
ELEM_TYPE_TAB = "tab"
ELEM_TYPE_TAB_GROUP = "tabgroup"
ELEM_TYPE_INPUT_SLIDER = "slider"
ELEM_TYPE_INPUT_LISTBOX = "listbox"
ELEM_TYPE_OUTPUT = "output"
ELEM_TYPE_COLUMN = "column"
ELEM_TYPE_MENUBAR = "menubar"
ELEM_TYPE_PROGRESS_BAR = "progressbar"
ELEM_TYPE_BLANK = "blank"
ELEM_TYPE_TABLE = "table"
ELEM_TYPE_TREE = "tree"
ELEM_TYPE_ERROR = "error"
ELEM_TYPE_SEPARATOR = "separator"
ELEM_TYPE_STATUSBAR = "statusbar"
ELEM_TYPE_PANE = "pane"
ELEM_TYPE_BUTTONMENU = "buttonmenu"
ELEM_TYPE_TITLEBAR = "titlebar"
ELEM_TYPE_SIZEGRIP = "sizegrip"

# STRETCH == ERROR ELEMENT as a filler

# -------------------------  Popup Buttons Types  ------------------------- #
POPUP_BUTTONS_YES_NO = 1
POPUP_BUTTONS_CANCELLED = 2
POPUP_BUTTONS_ERROR = 3
POPUP_BUTTONS_OK_CANCEL = 4
POPUP_BUTTONS_OK = 0
POPUP_BUTTONS_NO_BUTTONS = 5

PSG_THEME_PART_BUTTON_TEXT = "Button Text Color"
PSG_THEME_PART_BUTTON_BACKGROUND = "Button Background Color"
PSG_THEME_PART_BACKGROUND = "Background Color"
PSG_THEME_PART_INPUT_BACKGROUND = "Input Element Background Color"
PSG_THEME_PART_INPUT_TEXT = "Input Element Text Color"
PSG_THEME_PART_TEXT = "Text Color"
PSG_THEME_PART_SLIDER = "Slider Color"
PSG_THEME_PART_LIST = [
    PSG_THEME_PART_BACKGROUND,
    PSG_THEME_PART_BUTTON_BACKGROUND,
    PSG_THEME_PART_BUTTON_TEXT,
    PSG_THEME_PART_INPUT_BACKGROUND,
    PSG_THEME_PART_INPUT_TEXT,
    PSG_THEME_PART_TEXT,
    PSG_THEME_PART_SLIDER,
]

# theme_button

TTK_SCROLLBAR_PART_TROUGH_COLOR = "Trough Color"
TTK_SCROLLBAR_PART_BACKGROUND_COLOR = "Background Color"
TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR = "Arrow Button Arrow Color"
TTK_SCROLLBAR_PART_FRAME_COLOR = "Frame Color"
TTK_SCROLLBAR_PART_SCROLL_WIDTH = "Frame Width"
TTK_SCROLLBAR_PART_ARROW_WIDTH = "Arrow Width"
TTK_SCROLLBAR_PART_RELIEF = "Relief"
TTK_SCROLLBAR_PART_LIST = [
    TTK_SCROLLBAR_PART_TROUGH_COLOR,
    TTK_SCROLLBAR_PART_BACKGROUND_COLOR,
    TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR,
    TTK_SCROLLBAR_PART_FRAME_COLOR,
    TTK_SCROLLBAR_PART_SCROLL_WIDTH,
    TTK_SCROLLBAR_PART_ARROW_WIDTH,
    TTK_SCROLLBAR_PART_RELIEF,
]
TTK_SCROLLBAR_PART_THEME_BASED_LIST = [
    TTK_SCROLLBAR_PART_TROUGH_COLOR,
    TTK_SCROLLBAR_PART_BACKGROUND_COLOR,
    TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR,
    TTK_SCROLLBAR_PART_FRAME_COLOR,
]
DEFAULT_TTK_PART_MAPPING_DICT = {
    TTK_SCROLLBAR_PART_TROUGH_COLOR: PSG_THEME_PART_SLIDER,
    TTK_SCROLLBAR_PART_BACKGROUND_COLOR: PSG_THEME_PART_BUTTON_BACKGROUND,
    TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR: PSG_THEME_PART_BUTTON_TEXT,
    TTK_SCROLLBAR_PART_FRAME_COLOR: PSG_THEME_PART_BACKGROUND,
    TTK_SCROLLBAR_PART_SCROLL_WIDTH: 12,
    TTK_SCROLLBAR_PART_ARROW_WIDTH: 12,
    TTK_SCROLLBAR_PART_RELIEF: RELIEF_RAISED,
}

ttk_part_mapping_dict = copy.copy(DEFAULT_TTK_PART_MAPPING_DICT)


class TTKPartOverrides:
    """
    This class contains "overrides" to the defaults for ttk scrollbars that are defined in the global settings file.
    This class is used in every element, in the Window class and there's a global one that is used by set_options.
    """

    def __init__(
            self,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
    ):
        self.sbar_trough_color = sbar_trough_color
        self.sbar_background_color = sbar_background_color
        self.sbar_arrow_color = sbar_arrow_color
        self.sbar_width = sbar_width
        self.sbar_arrow_width = sbar_arrow_width
        self.sbar_frame_color = sbar_frame_color
        self.sbar_relief = sbar_relief


ttk_part_overrides_from_options = TTKPartOverrides()

# -------------------------  tkinter BASIC cursors... there are some OS dependent ones too  ------------------------- #
TKINTER_CURSORS = [
    "X_cursor",
    "arrow",
    "based_arrow_down",
    "based_arrow_up",
    "boat",
    "bogosity",
    "bottom_left_corner",
    "bottom_right_corner",
    "bottom_side",
    "bottom_tee",
    "box_spiral",
    "center_ptr",
    "circle",
    "clock",
    "coffee_mug",
    "cross",
    "cross_reverse",
    "crosshair",
    "diamond_cross",
    "dot",
    "dotbox",
    "double_arrow",
    "draft_large",
    "draft_small",
    "draped_box",
    "exchange",
    "fleur",
    "gobbler",
    "gumby",
    "hand1",
    "hand2",
    "heart",
    "icon",
    "iron_cross",
    "left_ptr",
    "left_side",
    "left_tee",
    "leftbutton",
    "ll_angle",
    "lr_angle",
    "man",
    "middlebutton",
    "mouse",
    "pencil",
    "pirate",
    "plus",
    "question_arrow",
    "right_ptr",
    "right_side",
    "right_tee",
    "rightbutton",
    "rtl_logo",
    "sailboat",
    "sb_down_arrow",
    "sb_h_double_arrow",
    "sb_left_arrow",
    "sb_right_arrow",
    "sb_up_arrow",
    "sb_v_double_arrow",
    "shuttle",
    "sizing",
    "spider",
    "spraycan",
    "star",
    "target",
    "tcross",
    "top_left_arrow",
    "top_left_corner",
    "top_right_corner",
    "top_side",
    "top_tee",
    "trek",
    "ul_angle",
    "umbrella",
    "ur_angle",
    "watch",
    "xterm",
]

TKINTER_CURSORS = [
    "X_cursor",
    "arrow",
    "based_arrow_down",
    "based_arrow_up",
    "boat",
    "bogosity",
    "bottom_left_corner",
    "bottom_right_corner",
    "bottom_side",
    "bottom_tee",
    "box_spiral",
    "center_ptr",
    "circle",
    "clock",
    "coffee_mug",
    "cross",
    "cross_reverse",
    "crosshair",
    "diamond_cross",
    "dot",
    "dotbox",
    "double_arrow",
    "draft_large",
    "draft_small",
    "draped_box",
    "exchange",
    "fleur",
    "gobbler",
    "gumby",
    "hand1",
    "hand2",
    "heart",
    "ibeam",
    "icon",
    "iron_cross",
    "left_ptr",
    "left_side",
    "left_tee",
    "leftbutton",
    "ll_angle",
    "lr_angle",
    "man",
    "middlebutton",
    "mouse",
    "no",
    "none",
    "pencil",
    "pirate",
    "plus",
    "question_arrow",
    "right_ptr",
    "right_side",
    "right_tee",
    "rightbutton",
    "rtl_logo",
    "sailboat",
    "sb_down_arrow",
    "sb_h_double_arrow",
    "sb_left_arrow",
    "sb_right_arrow",
    "sb_up_arrow",
    "sb_v_double_arrow",
    "shuttle",
    "size",
    "size_ne_sw",
    "size_ns",
    "size_nw_se",
    "size_we",
    "sizing",
    "spider",
    "spraycan",
    "star",
    "starting",
    "target",
    "tcross",
    "top_left_arrow",
    "top_left_corner",
    "top_right_corner",
    "top_side",
    "top_tee",
    "trek",
    "ul_angle",
    "umbrella",
    "uparrow",
    "ur_angle",
    "wait",
    "watch",
    "xterm",
]
# -------------------------  tkinter key codes for bindings  ------------------------- #

# The keycode that when pressed will take a snapshot of the current window
DEFAULT_WINDOW_SNAPSHOT_KEY_CODE = None
DEFAULT_WINDOW_SNAPSHOT_KEY = "--SCREENSHOT THIS WINDOW--"

tkinter_keysyms = (
    "space",
    "exclam",
    "quotedbl",
    "numbersign",
    "dollar",
    "percent",
    "ampersand",
    "quoteright",
    "parenleft",
    "parenright",
    "asterisk",
    "plus",
    "comma",
    "minus",
    "period",
    "slash",
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "colon",
    "semicolon",
    "less",
    "equal",
    "greater",
    "question",
    "at",
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
    "T",
    "U",
    "V",
    "W",
    "X",
    "Y",
    "Z",
    "bracketleft",
    "backslash",
    "bracketright",
    "asciicircum",
    "underscore",
    "quoteleft",
    "a",
    "b",
    "c",
    "d",
    "e",
    "f",
    "g",
    "h",
    "i",
    "j",
    "k",
    "l",
    "m",
    "n",
    "o",
    "p",
    "q",
    "r",
    "s",
    "t",
    "u",
    "v",
    "w",
    "x",
    "y",
    "z",
    "braceleft",
    "bar",
    "braceright",
    "asciitilde",
    "nobreakspace",
    "exclamdown",
    "cent",
    "sterling",
    "currency",
    "yen",
    "brokenbar",
    "section",
    "diaeresis",
    "copyright",
    "ordfeminine",
    "guillemotleft",
    "notsign",
    "hyphen",
    "registered",
    "macron",
    "degree",
    "plusminus",
    "twosuperior",
    "threesuperior",
    "acute",
    "mu",
    "paragraph",
    "periodcentered",
    "cedilla",
    "onesuperior",
    "masculine",
    "guillemotright",
    "onequarter",
    "onehalf",
    "threequarters",
    "questiondown",
    "Agrave",
    "Aacute",
    "Acircumflex",
    "Atilde",
    "Adiaeresis",
    "Aring",
    "AE",
    "Ccedilla",
    "Egrave",
    "Eacute",
    "Ecircumflex",
    "Ediaeresis",
    "Igrave",
    "Iacute",
    "Icircumflex",
    "Idiaeresis",
    "Eth",
    "Ntilde",
    "Ograve",
    "Oacute",
    "Ocircumflex",
    "Otilde",
    "Odiaeresis",
    "multiply",
    "Ooblique",
    "Ugrave",
    "Uacute",
    "Ucircumflex",
    "Udiaeresis",
    "Yacute",
    "Thorn",
    "ssharp",
    "agrave",
    "aacute",
    "acircumflex",
    "atilde",
    "adiaeresis",
    "aring",
    "ae",
    "ccedilla",
    "egrave",
    "eacute",
    "ecircumflex",
    "ediaeresis",
    "igrave",
    "iacute",
    "icircumflex",
    "idiaeresis",
    "eth",
    "ntilde",
    "ograve",
    "oacute",
    "ocircumflex",
    "otilde",
    "odiaeresis",
    "division",
    "oslash",
    "ugrave",
    "uacute",
    "ucircumflex",
    "udiaeresis",
    "yacute",
    "thorn",
    "ydiaeresis",
    "Aogonek",
    "breve",
    "Lstroke",
    "Lcaron",
    "Sacute",
    "Scaron",
    "Scedilla",
    "Tcaron",
    "Zacute",
    "Zcaron",
    "Zabovedot",
    "aogonek",
    "ogonek",
    "lstroke",
    "lcaron",
    "sacute",
    "caron",
    "scaron",
    "scedilla",
    "tcaron",
    "zacute",
    "doubleacute",
    "zcaron",
    "zabovedot",
    "Racute",
    "Abreve",
    "Cacute",
    "Ccaron",
    "Eogonek",
    "Ecaron",
    "Dcaron",
    "Nacute",
    "Ncaron",
    "Odoubleacute",
    "Rcaron",
    "Uring",
    "Udoubleacute",
    "Tcedilla",
    "racute",
    "abreve",
    "cacute",
    "ccaron",
    "eogonek",
    "ecaron",
    "dcaron",
    "nacute",
    "ncaron",
    "odoubleacute",
    "rcaron",
    "uring",
    "udoubleacute",
    "tcedilla",
    "abovedot",
    "Hstroke",
    "Hcircumflex",
    "Iabovedot",
    "Gbreve",
    "Jcircumflex",
    "hstroke",
    "hcircumflex",
    "idotless",
    "gbreve",
    "jcircumflex",
    "Cabovedot",
    "Ccircumflex",
    "Gabovedot",
    "Gcircumflex",
    "Ubreve",
    "Scircumflex",
    "cabovedot",
    "ccircumflex",
    "gabovedot",
    "gcircumflex",
    "ubreve",
    "scircumflex",
    "kappa",
    "Rcedilla",
    "Itilde",
    "Lcedilla",
    "Emacron",
    "Gcedilla",
    "Tslash",
    "rcedilla",
    "itilde",
    "lcedilla",
    "emacron",
    "gacute",
    "tslash",
    "ENG",
    "eng",
    "Amacron",
    "Iogonek",
    "Eabovedot",
    "Imacron",
    "Ncedilla",
    "Omacron",
    "Kcedilla",
    "Uogonek",
    "Utilde",
    "Umacron",
    "amacron",
    "iogonek",
    "eabovedot",
    "imacron",
    "ncedilla",
    "omacron",
    "kcedilla",
    "uogonek",
    "utilde",
    "umacron",
    "overline",
    "kana_fullstop",
    "kana_openingbracket",
    "kana_closingbracket",
    "kana_comma",
    "kana_middledot",
    "kana_WO",
    "kana_a",
    "kana_i",
    "kana_u",
    "kana_e",
    "kana_o",
    "kana_ya",
    "kana_yu",
    "kana_yo",
    "kana_tu",
    "prolongedsound",
    "kana_A",
    "kana_I",
    "kana_U",
    "kana_E",
    "kana_O",
    "kana_KA",
    "kana_KI",
    "kana_KU",
    "kana_KE",
    "kana_KO",
    "kana_SA",
    "kana_SHI",
    "kana_SU",
    "kana_SE",
    "kana_SO",
    "kana_TA",
    "kana_TI",
    "kana_TU",
    "kana_TE",
    "kana_TO",
    "kana_NA",
    "kana_NI",
    "kana_NU",
    "kana_NE",
    "kana_NO",
    "kana_HA",
    "kana_HI",
    "kana_HU",
    "kana_HE",
    "kana_HO",
    "kana_MA",
    "kana_MI",
    "kana_MU",
    "kana_ME",
    "kana_MO",
    "kana_YA",
    "kana_YU",
    "kana_YO",
    "kana_RA",
    "kana_RI",
    "kana_RU",
    "kana_RE",
    "kana_RO",
    "kana_WA",
    "kana_N",
    "voicedsound",
    "semivoicedsound",
    "Arabic_comma",
    "Arabic_semicolon",
    "Arabic_question_mark",
    "Arabic_hamza",
    "Arabic_maddaonalef",
    "Arabic_hamzaonalef",
    "Arabic_hamzaonwaw",
    "Arabic_hamzaunderalef",
    "Arabic_hamzaonyeh",
    "Arabic_alef",
    "Arabic_beh",
    "Arabic_tehmarbuta",
    "Arabic_teh",
    "Arabic_theh",
    "Arabic_jeem",
    "Arabic_hah",
    "Arabic_khah",
    "Arabic_dal",
    "Arabic_thal",
    "Arabic_ra",
    "Arabic_zain",
    "Arabic_seen",
    "Arabic_sheen",
    "Arabic_sad",
    "Arabic_dad",
    "Arabic_tah",
    "Arabic_zah",
    "Arabic_ain",
    "Arabic_ghain",
    "Arabic_tatweel",
    "Arabic_feh",
    "Arabic_qaf",
    "Arabic_kaf",
    "Arabic_lam",
    "Arabic_meem",
    "Arabic_noon",
    "Arabic_heh",
    "Arabic_waw",
    "Arabic_alefmaksura",
    "Arabic_yeh",
    "Arabic_fathatan",
    "Arabic_dammatan",
    "Arabic_kasratan",
    "Arabic_fatha",
    "Arabic_damma",
    "Arabic_kasra",
    "Arabic_shadda",
    "Arabic_sukun",
    "Serbian_dje",
    "Macedonia_gje",
    "Cyrillic_io",
    "Ukranian_je",
    "Macedonia_dse",
    "Ukranian_i",
    "Ukranian_yi",
    "Serbian_je",
    "Serbian_lje",
    "Serbian_nje",
    "Serbian_tshe",
    "Macedonia_kje",
    "Byelorussian_shortu",
    "Serbian_dze",
    "numerosign",
    "Serbian_DJE",
    "Macedonia_GJE",
    "Cyrillic_IO",
    "Ukranian_JE",
    "Macedonia_DSE",
    "Ukranian_I",
    "Ukranian_YI",
    "Serbian_JE",
    "Serbian_LJE",
    "Serbian_NJE",
    "Serbian_TSHE",
    "Macedonia_KJE",
    "Byelorussian_SHORTU",
    "Serbian_DZE",
    "Cyrillic_yu",
    "Cyrillic_a",
    "Cyrillic_be",
    "Cyrillic_tse",
    "Cyrillic_de",
    "Cyrillic_ie",
    "Cyrillic_ef",
    "Cyrillic_ghe",
    "Cyrillic_ha",
    "Cyrillic_i",
    "Cyrillic_shorti",
    "Cyrillic_ka",
    "Cyrillic_el",
    "Cyrillic_em",
    "Cyrillic_en",
    "Cyrillic_o",
    "Cyrillic_pe",
    "Cyrillic_ya",
    "Cyrillic_er",
    "Cyrillic_es",
    "Cyrillic_te",
    "Cyrillic_u",
    "Cyrillic_zhe",
    "Cyrillic_ve",
    "Cyrillic_softsign",
    "Cyrillic_yeru",
    "Cyrillic_ze",
    "Cyrillic_sha",
    "Cyrillic_e",
    "Cyrillic_shcha",
    "Cyrillic_che",
    "Cyrillic_hardsign",
    "Cyrillic_YU",
    "Cyrillic_A",
    "Cyrillic_BE",
    "Cyrillic_TSE",
    "Cyrillic_DE",
    "Cyrillic_IE",
    "Cyrillic_EF",
    "Cyrillic_GHE",
    "Cyrillic_HA",
    "Cyrillic_I",
    "Cyrillic_SHORTI",
    "Cyrillic_KA",
    "Cyrillic_EL",
    "Cyrillic_EM",
    "Cyrillic_EN",
    "Cyrillic_O",
    "Cyrillic_PE",
    "Cyrillic_YA",
    "Cyrillic_ER",
    "Cyrillic_ES",
    "Cyrillic_TE",
    "Cyrillic_U",
    "Cyrillic_ZHE",
    "Cyrillic_VE",
    "Cyrillic_SOFTSIGN",
    "Cyrillic_YERU",
    "Cyrillic_ZE",
    "Cyrillic_SHA",
    "Cyrillic_E",
    "Cyrillic_SHCHA",
    "Cyrillic_CHE",
    "Cyrillic_HARDSIGN",
    "Greek_ALPHAaccent",
    "Greek_EPSILONaccent",
    "Greek_ETAaccent",
    "Greek_IOTAaccent",
    "Greek_IOTAdiaeresis",
    "Greek_IOTAaccentdiaeresis",
    "Greek_OMICRONaccent",
    "Greek_UPSILONaccent",
    "Greek_UPSILONdieresis",
    "Greek_UPSILONaccentdieresis",
    "Greek_OMEGAaccent",
    "Greek_alphaaccent",
    "Greek_epsilonaccent",
    "Greek_etaaccent",
    "Greek_iotaaccent",
    "Greek_iotadieresis",
    "Greek_iotaaccentdieresis",
    "Greek_omicronaccent",
    "Greek_upsilonaccent",
    "Greek_upsilondieresis",
    "Greek_upsilonaccentdieresis",
    "Greek_omegaaccent",
    "Greek_ALPHA",
    "Greek_BETA",
    "Greek_GAMMA",
    "Greek_DELTA",
    "Greek_EPSILON",
    "Greek_ZETA",
    "Greek_ETA",
    "Greek_THETA",
    "Greek_IOTA",
    "Greek_KAPPA",
    "Greek_LAMBDA",
    "Greek_MU",
    "Greek_NU",
    "Greek_XI",
    "Greek_OMICRON",
    "Greek_PI",
    "Greek_RHO",
    "Greek_SIGMA",
    "Greek_TAU",
    "Greek_UPSILON",
    "Greek_PHI",
    "Greek_CHI",
    "Greek_PSI",
    "Greek_OMEGA",
    "Greek_alpha",
    "Greek_beta",
    "Greek_gamma",
    "Greek_delta",
    "Greek_epsilon",
    "Greek_zeta",
    "Greek_eta",
    "Greek_theta",
    "Greek_iota",
    "Greek_kappa",
    "Greek_lambda",
    "Greek_mu",
    "Greek_nu",
    "Greek_xi",
    "Greek_omicron",
    "Greek_pi",
    "Greek_rho",
    "Greek_sigma",
    "Greek_finalsmallsigma",
    "Greek_tau",
    "Greek_upsilon",
    "Greek_phi",
    "Greek_chi",
    "Greek_psi",
    "Greek_omega",
    "leftradical",
    "topleftradical",
    "horizconnector",
    "topintegral",
    "botintegral",
    "vertconnector",
    "topleftsqbracket",
    "botleftsqbracket",
    "toprightsqbracket",
    "botrightsqbracket",
    "topleftparens",
    "botleftparens",
    "toprightparens",
    "botrightparens",
    "leftmiddlecurlybrace",
    "rightmiddlecurlybrace",
    "topleftsummation",
    "botleftsummation",
    "topvertsummationconnector",
    "botvertsummationconnector",
    "toprightsummation",
    "botrightsummation",
    "rightmiddlesummation",
    "lessthanequal",
    "notequal",
    "greaterthanequal",
    "integral",
    "therefore",
    "variation",
    "infinity",
    "nabla",
    "approximate",
    "similarequal",
    "ifonlyif",
    "implies",
    "identical",
    "radical",
    "includedin",
    "includes",
    "intersection",
    "union",
    "logicaland",
    "logicalor",
    "partialderivative",
    "function",
    "leftarrow",
    "uparrow",
    "rightarrow",
    "downarrow",
    "blank",
    "soliddiamond",
    "checkerboard",
    "ht",
    "ff",
    "cr",
    "lf",
    "nl",
    "vt",
    "lowrightcorner",
    "uprightcorner",
    "upleftcorner",
    "lowleftcorner",
    "crossinglines",
    "horizlinescan1",
    "horizlinescan3",
    "horizlinescan5",
    "horizlinescan7",
    "horizlinescan9",
    "leftt",
    "rightt",
    "bott",
    "topt",
    "vertbar",
    "emspace",
    "enspace",
    "em3space",
    "em4space",
    "digitspace",
    "punctspace",
    "thinspace",
    "hairspace",
    "emdash",
    "endash",
    "signifblank",
    "ellipsis",
    "doubbaselinedot",
    "onethird",
    "twothirds",
    "onefifth",
    "twofifths",
    "threefifths",
    "fourfifths",
    "onesixth",
    "fivesixths",
    "careof",
    "figdash",
    "leftanglebracket",
    "decimalpoint",
    "rightanglebracket",
    "marker",
    "oneeighth",
    "threeeighths",
    "fiveeighths",
    "seveneighths",
    "trademark",
    "signaturemark",
    "trademarkincircle",
    "leftopentriangle",
    "rightopentriangle",
    "emopencircle",
    "emopenrectangle",
    "leftsinglequotemark",
    "rightsinglequotemark",
    "leftdoublequotemark",
    "rightdoublequotemark",
    "prescription",
    "minutes",
    "seconds",
    "latincross",
    "hexagram",
    "filledrectbullet",
    "filledlefttribullet",
    "filledrighttribullet",
    "emfilledcircle",
    "emfilledrect",
    "enopencircbullet",
    "enopensquarebullet",
    "openrectbullet",
    "opentribulletup",
    "opentribulletdown",
    "openstar",
    "enfilledcircbullet",
    "enfilledsqbullet",
    "filledtribulletup",
    "filledtribulletdown",
    "leftpointer",
    "rightpointer",
    "club",
    "diamond",
    "heart",
    "maltesecross",
    "dagger",
    "doubledagger",
    "checkmark",
    "ballotcross",
    "musicalsharp",
    "musicalflat",
    "malesymbol",
    "femalesymbol",
    "telephone",
    "telephonerecorder",
    "phonographcopyright",
    "caret",
    "singlelowquotemark",
    "doublelowquotemark",
    "cursor",
    "leftcaret",
    "rightcaret",
    "downcaret",
    "upcaret",
    "overbar",
    "downtack",
    "upshoe",
    "downstile",
    "underbar",
    "jot",
    "quad",
    "uptack",
    "circle",
    "upstile",
    "downshoe",
    "rightshoe",
    "leftshoe",
    "lefttack",
    "righttack",
    "hebrew_aleph",
    "hebrew_beth",
    "hebrew_gimmel",
    "hebrew_daleth",
    "hebrew_he",
    "hebrew_waw",
    "hebrew_zayin",
    "hebrew_het",
    "hebrew_teth",
    "hebrew_yod",
    "hebrew_finalkaph",
    "hebrew_kaph",
    "hebrew_lamed",
    "hebrew_finalmem",
    "hebrew_mem",
    "hebrew_finalnun",
    "hebrew_nun",
    "hebrew_samekh",
    "hebrew_ayin",
    "hebrew_finalpe",
    "hebrew_pe",
    "hebrew_finalzadi",
    "hebrew_zadi",
    "hebrew_kuf",
    "hebrew_resh",
    "hebrew_shin",
    "hebrew_taf",
    "BackSpace",
    "Tab",
    "Linefeed",
    "Clear",
    "Return",
    "Pause",
    "Scroll_Lock",
    "Sys_Req",
    "Escape",
    "Multi_key",
    "Kanji",
    "Home",
    "Left",
    "Up",
    "Right",
    "Down",
    "Prior",
    "Next",
    "End",
    "Begin",
    "Win_L",
    "Win_R",
    "App",
    "Select",
    "Print",
    "Execute",
    "Insert",
    "Undo",
    "Redo",
    "Menu",
    "Find",
    "Cancel",
    "Help",
    "Break",
    "Hebrew_switch",
    "Num_Lock",
    "KP_Space",
    "KP_Tab",
    "KP_Enter",
    "KP_F1",
    "KP_F2",
    "KP_F3",
    "KP_F4",
    "KP_Multiply",
    "KP_Add",
    "KP_Separator",
    "KP_Subtract",
    "KP_Decimal",
    "KP_Divide",
    "KP_0",
    "KP_1",
    "KP_2",
    "KP_3",
    "KP_4",
    "KP_5",
    "KP_6",
    "KP_7",
    "KP_8",
    "KP_9",
    "KP_Equal",
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "F11",
    "F12",
    "L1",
    "L2",
    "L3",
    "L4",
    "L5",
    "L6",
    "L7",
    "L8",
    "L9",
    "L10",
    "R1",
    "R2",
    "R3",
    "R4",
    "R5",
    "R6",
    "R7",
    "R8",
    "R9",
    "R10",
    "R11",
    "R12",
    "F33",
    "R14",
    "R15",
    "Shift_L",
    "Shift_R",
    "Control_L",
    "Control_R",
    "Caps_Lock",
    "Shift_Lock",
    "Meta_L",
    "Meta_R",
    "Alt_L",
    "Alt_R",
    "Super_L",
    "Super_R",
    "Hyper_L",
    "Hyper_R",
    "Delete",
)


# ------------------------------------------------------------------------- #
#                       ToolTip used by the Elements                        #
# ------------------------------------------------------------------------- #


class ToolTip:
    """
    Create a tooltip for a given widget
    (inspired by https://stackoverflow.com/a/36221216)
    This is an INTERNALLY USED only class.  Users should not refer to this class at all.
    """

    def __init__(self, widget, text, timeout=DEFAULT_TOOLTIP_TIME):
        """
        :param widget:  The tkinter widget
        :type widget:   widget type varies
        :param text:    text for the tooltip. It can inslude \n
        :type text:     (str)
        :param timeout: Time in milliseconds that mouse must remain still before tip is shown
        :type timeout:  (int)
        """
        self.widget = widget
        self.text = text
        self.timeout = timeout
        # self.wraplength = wraplength if wraplength else widget.winfo_screenwidth() // 2
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)

    def enter(self, event=None):
        """
        Called by tkinter when mouse enters a widget
        :param event: from tkinter.  Has x,y coordinates of mouse
        :type event:

        """
        self.x = event.x
        self.y = event.y
        self.schedule()

    def leave(self, event=None):
        """
        Called by tktiner when mouse exits a widget
        :param event: from tkinter.  Event info that's not used by function.
        :type event:

        """
        self.unschedule()
        self.hidetip()

    def schedule(self):
        """
        Schedule a timer to time how long mouse is hovering
        """
        self.unschedule()
        self.id = self.widget.after(self.timeout, self.showtip)

    def unschedule(self):
        """
        Cancel timer used to time mouse hover
        """
        if self.id:
            self.widget.after_cancel(self.id)
        self.id = None

    def showtip(self):
        """
        Creates a topoltip window with the tooltip text inside of it
        """
        if self.tipwindow:
            return
        x = self.widget.winfo_rootx() + self.x + DEFAULT_TOOLTIP_OFFSET[0]
        y = self.widget.winfo_rooty() + self.y + DEFAULT_TOOLTIP_OFFSET[1]
        self.tipwindow = tk.Toplevel(self.widget)
        # if not sys.platform.startswith('darwin'):
        try:
            self.tipwindow.wm_overrideredirect(True)
            # if running_mac() and ENABLE_MAC_NOTITLEBAR_PATCH:
            if _mac_should_apply_notitlebar_patch():
                self.tipwindow.wm_overrideredirect(False)
        except Exception as e:
            print("* Error performing wm_overrideredirect in showtip *", e)
        self.tipwindow.wm_geometry("+%d+%d" % (x, y))
        self.tipwindow.wm_attributes("-topmost", 1)

        label = ttk.Label(
            self.tipwindow,
            text=self.text,
            justify=tk.LEFT,
            background=TOOLTIP_BACKGROUND_COLOR,
            relief=tk.SOLID,
            borderwidth=1,
        )
        if TOOLTIP_FONT is not None:
            label.config(font=TOOLTIP_FONT)
        label.pack()

    def hidetip(self):
        """
        Destroy the tooltip window
        """
        if self.tipwindow:
            self.tipwindow.destroy()
        self.tipwindow = None


# ---------------------------------------------------------------------- #
# Cascading structure.... Objects get larger                             #
#   Button                                                               #
#       Element                                                          #
#           Row                                                          #
#               Form                                                     #
# ---------------------------------------------------------------------- #
# ------------------------------------------------------------------------- #
#                       Element CLASS                                       #
# ------------------------------------------------------------------------- #
class Element:
    """The base class for all Elements. Holds the basic description of an Element like size and colors"""

    def __init__(
            self,
            type_,
            size=(None, None),
            auto_size_text=None,
            font=None,
            background_color=None,
            text_color=None,
            key=None,
            pad=None,
            tooltip=None,
            visible=True,
            metadata=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
    ):
        """
        Element base class. Only used internally.  User will not create an Element object by itself

        :param type_:                        The type of element. These constants all start with "ELEM_TYPE_"
        :type type_:                         (int) (could be enum)
        :param size:                        w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                         (int, int) | (None, None) | int
        :param auto_size_text:              True if the Widget should be shrunk to exactly fit the number of chars to show
        :type auto_size_text:               bool
        :param font:                        specifies the font family, size. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                         (str or (str, int[, str]) or None)
        :param background_color:            color of background. Can be in #RRGGBB format or a color name "black"
        :type background_color:             (str)
        :param text_color:                  element's text color. Can be in #RRGGBB format or a color name "black"
        :type text_color:                   (str)
        :param key:                         Identifies an Element. Should be UNIQUE to this window.
        :type key:                          str | int | tuple | object
        :param pad:                         Amount of padding to put around element in pixels (left/right, top/bottom). If an int is given, then auto-converted to tuple (int, int)
        :type pad:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:                     text, that will appear when mouse hovers over the element
        :type tooltip:                      (str)
        :param visible:                     set visibility state of the element (Default = True)
        :type visible:                      (bool)
        :param metadata:                    User metadata that can be set to ANYTHING
        :type metadata:                     (Any)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        """

        if size is not None and size != (None, None):
            if isinstance(size, int):
                size = (size, 1)
            if isinstance(size, tuple) and len(size) == 1:
                size = (size[0], 1)

        if pad is not None and pad != (None, None):
            if isinstance(pad, int):
                pad = (pad, pad)

        self.Size = size
        self.Type = type_
        self.AutoSizeText = auto_size_text

        self.Pad = pad
        self.Font = font

        self.TKStringVar = None
        self.TKIntVar = None
        self.TKText = None
        self.TKEntry = None
        self.TKImage = None
        self.ttk_style_name = ""  # The ttk style name (if this is a ttk widget)
        self.ttk_style = None  # The ttk Style object (if this is a ttk widget)
        self._metadata = None  # type: Any

        self.ParentForm = None  # type: Window
        self.ParentContainer = None  # will be a Form, Column, or Frame element # UNBIND
        self.TextInputDefault = None
        self.Position = (0, 0)  # Default position Row 0, Col 0
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_ELEMENT_BACKGROUND_COLOR
        )
        self.TextColor = (
            text_color if text_color is not None else DEFAULT_ELEMENT_TEXT_COLOR
        )
        self.Key = key  # dictionary key for return values
        self.Tooltip = tooltip
        self.TooltipObject = None
        self._visible = visible
        self.TKRightClickMenu = None
        self.Widget = (
            None  # Set when creating window. Has the main tkinter widget for element
        )
        self.Tearoff = False  # needed because of right click menu code
        self.ParentRowFrame = None  # type tk.Frame
        self.metadata = metadata
        self.user_bind_dict = {}  # Used when user defines a tkinter binding using bind method - convert bind string to key modifier
        self.user_bind_event = None  # Used when user defines a tkinter binding using bind method - event data from tkinter
        # self.pad_used = (0, 0)  # the amount of pad used when was inserted into the layout
        self._popup_menu_location = (None, None)
        self.pack_settings = None
        self.vsb_style_name = None  # ttk style name used for the verical scrollbar if one is attached to element
        self.hsb_style_name = None  # ttk style name used for the horizontal scrollbar if one is attached to element
        self.vsb_style = None  # The ttk style used for the vertical scrollbar if one is attached to element
        self.hsb_style = None  # The ttk style used for the horizontal scrollbar if one is attached to element
        self.hsb = None  # The horizontal scrollbar if one is attached to element
        self.vsb = None  # The vertical scrollbar if one is attached to element
        ## TTK Scrollbar Settings
        self.ttk_part_overrides = TTKPartOverrides(
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )

        PSG_THEME_PART_FUNC_MAP = {
            PSG_THEME_PART_BACKGROUND: theme_background_color,
            PSG_THEME_PART_BUTTON_BACKGROUND: theme_button_color_background,
            PSG_THEME_PART_BUTTON_TEXT: theme_button_color_text,
            PSG_THEME_PART_INPUT_BACKGROUND: theme_input_background_color,
            PSG_THEME_PART_INPUT_TEXT: theme_input_text_color,
            PSG_THEME_PART_TEXT: theme_text_color,
            PSG_THEME_PART_SLIDER: theme_slider_color,
        }

        # class Theme_Parts():
        #     PSG_THEME_PART_FUNC_MAP = {PSG_THEME_PART_BACKGROUND: theme_background_color,
        if sbar_trough_color is not None:
            self.scroll_trough_color = sbar_trough_color
        else:
            self.scroll_trough_color = PSG_THEME_PART_FUNC_MAP.get(
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_TROUGH_COLOR],
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_TROUGH_COLOR],
            )
            if callable(self.scroll_trough_color):
                self.scroll_trough_color = self.scroll_trough_color()

        if sbar_background_color is not None:
            self.scroll_background_color = sbar_background_color
        else:
            self.scroll_background_color = PSG_THEME_PART_FUNC_MAP.get(
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_BACKGROUND_COLOR],
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_BACKGROUND_COLOR],
            )
            if callable(self.scroll_background_color):
                self.scroll_background_color = self.scroll_background_color()

        if sbar_arrow_color is not None:
            self.scroll_arrow_color = sbar_arrow_color
        else:
            self.scroll_arrow_color = PSG_THEME_PART_FUNC_MAP.get(
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR],
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_ARROW_BUTTON_ARROW_COLOR],
            )
            if callable(self.scroll_arrow_color):
                self.scroll_arrow_color = self.scroll_arrow_color()

        if sbar_frame_color is not None:
            self.scroll_frame_color = sbar_frame_color
        else:
            self.scroll_frame_color = PSG_THEME_PART_FUNC_MAP.get(
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_FRAME_COLOR],
                ttk_part_mapping_dict[TTK_SCROLLBAR_PART_FRAME_COLOR],
            )
            if callable(self.scroll_frame_color):
                self.scroll_frame_color = self.scroll_frame_color()

        if sbar_relief is not None:
            self.scroll_relief = sbar_relief
        else:
            self.scroll_relief = ttk_part_mapping_dict[TTK_SCROLLBAR_PART_RELIEF]

        if sbar_width is not None:
            self.scroll_width = sbar_width
        else:
            self.scroll_width = ttk_part_mapping_dict[TTK_SCROLLBAR_PART_SCROLL_WIDTH]

        if sbar_arrow_width is not None:
            self.scroll_arrow_width = sbar_arrow_width
        else:
            self.scroll_arrow_width = ttk_part_mapping_dict[
                TTK_SCROLLBAR_PART_ARROW_WIDTH
            ]

        if not hasattr(self, "DisabledTextColor"):
            self.DisabledTextColor = None
        if not hasattr(self, "ItemFont"):
            self.ItemFont = None
        if not hasattr(self, "RightClickMenu"):
            self.RightClickMenu = None
        if not hasattr(self, "Disabled"):
            self.Disabled = None  # in case the element hasn't defined this, add it here

    @property
    def visible(self):
        """
        Returns visibility state for the element.  This is a READONLY property
        :return: Visibility state for element
        :rtype:  (bool)
        """
        return self._visible

    @property
    def metadata(self):
        """
        Metadata is an Element property that you can use at any time to hold any value
        :return: the current metadata value
        :rtype:  (Any)
        """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """
         Metadata is an Element property that you can use at any time to hold any value
        :param value: Anything you want it to be
        :type value:  (Any)
        """
        self._metadata = value

    @property
    def key(self):
        """
        Returns key for the element.  This is a READONLY property.
        Keys can be any hashable object (basically anything except a list... tuples are ok, but not lists)
        :return: The window's Key
        :rtype:  (Any)
        """
        return self.Key

    @property
    def widget(self):
        """
        Returns tkinter widget for the element.  This is a READONLY property.
        The implementation is that the Widget member variable is returned. This is a backward compatible addition
        :return:    The element's underlying tkinter widget
        :rtype:     (tkinter.Widget)
        """
        return self.Widget

    def _RightClickMenuCallback(self, event):
        """
        Callback function that's called when a right click happens. Shows right click menu as result

        :param event: information provided by tkinter about the event including x,y location of click
        :type event:

        """
        if self.Type == ELEM_TYPE_TAB_GROUP:
            try:
                index = self.Widget.index("@{},{}".format(event.x, event.y))
                tab = self.Widget.tab(index, "text")
                key = self.find_key_from_tab_name(tab)
                tab_element = self.ParentForm.key_dict[key]
                if (
                        tab_element.RightClickMenu is None
                ):  # if this tab didn't explicitly have a menu, then don't show anything
                    return
                tab_element.TKRightClickMenu.tk_popup(event.x_root, event.y_root, 0)
                self.TKRightClickMenu.grab_release()
            except Exception:
                pass
            return
        self.TKRightClickMenu.tk_popup(event.x_root, event.y_root, 0)
        self.TKRightClickMenu.grab_release()
        if self.Type == ELEM_TYPE_GRAPH:
            self._update_position_for_returned_values(event)

    def _tearoff_menu_callback(self, parent, menu):
        """
        Callback function that's called when a right click menu is torn off.
        The reason for this function is to relocate the torn-off menu. It will default to 0,0 otherwise
        This callback moves the right click menu window to the location of the current window

        :param parent: information provided by tkinter - the parent of the Meny
        :type parent:
        :param menu:   information provided by tkinter - the menu window
        :type menu:

        """
        if self._popup_menu_location == (None, None):
            winx, winy = self.ParentForm.current_location()
        else:
            winx, winy = self._popup_menu_location
        # self.ParentForm.TKroot.update()
        self.ParentForm.TKroot.tk.call(
            "wm", "geometry", menu, "+{}+{}".format(winx, winy)
        )

    def _MenuItemChosenCallback(self, item_chosen):  # TEXT Menu item callback
        """
        Callback function called when user chooses a menu item from menubar, Button Menu or right click menu

        :param item_chosen: String holding the value chosen.
        :type item_chosen:  str

        """
        # print('IN MENU ITEM CALLBACK', item_chosen)
        self.MenuItemChosen = item_chosen
        self.ParentForm.LastButtonClicked = self.MenuItemChosen
        self.ParentForm.FormRemainedOpen = True
        _exit_mainloop(self.ParentForm)
        # Window._window_that_exited = self.ParentForm
        # self.ParentForm.TKroot.quit()  # kick the users out of the mainloop

    def _FindReturnKeyBoundButton(self, form):
        """
        Searches for which Button has the flag Button.BindReturnKey set.  It is called recursively when a
        "Container Element" is encountered. Func has to walk entire window including these "sub-forms"

        :param form: the Window object to search
        :type form:
        :return:     Button Object if a button is found, else None
        :rtype:      Button | None
        """
        for row in form.Rows:
            for element in row:
                if element.Type == ELEM_TYPE_BUTTON:
                    if element.BindReturnKey:
                        return element
                if element.Type == ELEM_TYPE_COLUMN:
                    rc = self._FindReturnKeyBoundButton(element)
                    if rc is not None:
                        return rc
                if element.Type == ELEM_TYPE_FRAME:
                    rc = self._FindReturnKeyBoundButton(element)
                    if rc is not None:
                        return rc
                if element.Type == ELEM_TYPE_TAB_GROUP:
                    rc = self._FindReturnKeyBoundButton(element)
                    if rc is not None:
                        return rc
                if element.Type == ELEM_TYPE_TAB:
                    rc = self._FindReturnKeyBoundButton(element)
                    if rc is not None:
                        return rc
                if element.Type == ELEM_TYPE_PANE:
                    rc = self._FindReturnKeyBoundButton(element)
                    if rc is not None:
                        return rc
        return None

    def _TextClickedHandler(self, event):
        """
        Callback that's called when a text element is clicked on with events enabled on the Text Element.
        Result is that control is returned back to user (quits mainloop).

        :param event:
        :type event:

        """
        # If this is a minimize button for a custom titlebar, then minimize the window
        if self.Key in (
                TITLEBAR_MINIMIZE_KEY,
                TITLEBAR_MAXIMIZE_KEY,
                TITLEBAR_CLOSE_KEY,
        ):
            self.ParentForm._custom_titlebar_callback(self.Key)
        self._generic_callback_handler(self.DisplayText)
        return

    def _ReturnKeyHandler(self, event):
        """
        Internal callback for the ENTER / RETURN key. Results in calling the ButtonCallBack for element that has the return key bound to it, just as if button was clicked.

        :param event:
        :type event:

        """
        # if the element is disabled, ignore the event
        if self.Disabled:
            return

        MyForm = self.ParentForm
        button_element = self._FindReturnKeyBoundButton(MyForm)
        if button_element is not None:
            # if the Button has been disabled, then don't perform the callback
            if button_element.Disabled:
                return
            button_element.ButtonCallBack()

    def _generic_callback_handler(self, alternative_to_key=None, force_key_to_be=None):
        """
        Peforms the actions that were in many of the callback functions previously.  Combined so that it's
        easier to modify and is in 1 place now

        :param alternate_to_key: If key is None, then use this value instead
        :type alternate_to_key:  Any
        """
        if force_key_to_be is not None:
            self.ParentForm.LastButtonClicked = force_key_to_be
        elif self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = alternative_to_key
        self.ParentForm.FormRemainedOpen = True

        _exit_mainloop(self.ParentForm)
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     Window._window_that_exited = self.ParentForm
        #     self.ParentForm.TKroot.quit()  # kick the users out of the mainloop

    def _ListboxSelectHandler(self, event):
        """
        Internal callback function for when a listbox item is selected

        :param event: Information from tkinter about the callback
        :type event:

        """
        self._generic_callback_handler("")

    def _ComboboxSelectHandler(self, event):
        """
        Internal callback function for when an entry is selected in a Combobox.
        :param event: Event data from tkinter (not used)
        :type event:

        """
        self._generic_callback_handler("")

    def _SpinboxSelectHandler(self, event=None):
        """
        Internal callback function for when an entry is selected in a Spinbox.
        Note that the parm is optional because it's not used if arrows are used to change the value
        but if the return key is pressed, it will include the event parm
        :param event: Event data passed in by tkinter (not used)
        :type event:
        """
        self._generic_callback_handler("")

    def _RadioHandler(self):
        """
        Internal callback for when a radio button is selected and enable events was set for radio
        """
        self._generic_callback_handler("")

    def _CheckboxHandler(self):
        """
        Internal callback for when a checkbnox is selected and enable events was set for checkbox
        """
        self._generic_callback_handler("")

    def _TabGroupSelectHandler(self, event):
        """
        Internal callback for when a Tab is selected and enable events was set for TabGroup

        :param event: Event data passed in by tkinter (not used)
        :type event:
        """
        self._generic_callback_handler("")

    def _KeyboardHandler(self, event):
        """
        Internal callback for when a key is pressed andd return keyboard events was set for window

        :param event: Event data passed in by tkinter (not used)
        :type event:
        """

        # if the element is disabled, ignore the event
        if self.Disabled:
            return
        self._generic_callback_handler("")

    def _ClickHandler(self, event):
        """
        Internal callback for when a mouse was clicked... I think.

        :param event: Event data passed in by tkinter (not used)
        :type event:
        """
        self._generic_callback_handler("")

    def _this_elements_window_closed(self, quick_check=True):
        if self.ParentForm is not None:
            return self.ParentForm.is_closed(quick_check=quick_check)

        return True

    def _user_bind_callback(self, bind_string, event, propagate=True):
        """
        Used when user binds a tkinter event directly to an element

        :param bind_string: The event that was bound so can lookup the key modifier
        :type bind_string:  (str)
        :param event:       Event data passed in by tkinter (not used)
        :type event:        (Any)
        :param propagate:   If True then tkinter will be told to propagate the event to the element
        :type propagate:    (bool)
        """
        key_suffix = self.user_bind_dict.get(bind_string, "")
        self.user_bind_event = event
        if self.Type == ELEM_TYPE_GRAPH:
            self._update_position_for_returned_values(event)
        if self.Key is not None:
            if isinstance(self.Key, str):
                key = self.Key + str(key_suffix)
            else:
                key = (
                    self.Key,
                    key_suffix,
                )  # old way (pre 2021) was to make a brand new tuple
                # key = self.Key + (key_suffix,)   # in 2021 tried this. It will break existing applications though - if key is a tuple, add one more item
        else:
            key = bind_string

        self._generic_callback_handler(force_key_to_be=key)

        return "break" if propagate is not True else None

    def bind(self, bind_string, key_modifier, propagate=True):
        """
        Used to add tkinter events to an Element.
        The tkinter specific data is in the Element's member variable user_bind_event
        :param bind_string:  The string tkinter expected in its bind function
        :type bind_string:   (str)
        :param key_modifier: Additional data to be added to the element's key when event is returned
        :type key_modifier:  (str)
        :param propagate:    If True then tkinter will be told to propagate the event to the element
        :type propagate:     (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        try:
            self.Widget.bind(
                bind_string,
                lambda evt: self._user_bind_callback(bind_string, evt, propagate),
            )
        except Exception as e:
            self.Widget.unbind_all(bind_string)
            return

        self.user_bind_dict[bind_string] = key_modifier

    def unbind(self, bind_string):
        """
        Removes a previously bound tkinter event from an Element.
        :param bind_string: The string tkinter expected in its bind function
        :type bind_string:  (str)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return
        self.Widget.unbind(bind_string)
        self.user_bind_dict.pop(bind_string, None)

    def set_tooltip(self, tooltip_text):
        """
        Called by application to change the tooltip text for an Element.  Normally invoked using the Element Object such as: window.Element('key').SetToolTip('New tip').

        :param tooltip_text: the text to show in tooltip.
        :type tooltip_text:  (str)
        """

        if self.TooltipObject:
            try:
                self.TooltipObject.leave()
            except Exception:
                pass

        self.TooltipObject = ToolTip(
            self.Widget, text=tooltip_text, timeout=DEFAULT_TOOLTIP_TIME
        )

    def set_focus(self, force=False):
        """
        Sets the current focus to be on this element

        :param force: if True will call focus_force otherwise calls focus_set
        :type force:  bool
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return
        try:
            if force:
                self.Widget.focus_force()
            else:
                self.Widget.focus_set()
        except Exception as e:
            _error_popup_with_traceback(
                "Exception blocking focus. Check your element's Widget", e
            )

    def block_focus(self, block=True):
        """
        Enable or disable the element from getting focus by using the keyboard.
        If the block parameter is True, then this element will not be given focus by using
        the keyboard to go from one element to another.
        You CAN click on the element and utilize it.

        :param block: if True the element will not get focus via the keyboard
        :type block:  bool
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return
        try:
            self.ParentForm.TKroot.focus_force()
            if block:
                self.Widget.configure(takefocus=0)
            else:
                self.Widget.configure(takefocus=1)
        except Exception as e:
            _error_popup_with_traceback(
                "Exception blocking focus. Check your element's Widget", e
            )

    def get_next_focus(self):
        """
        Gets the next element that should get focus after this element.

        :return:    Element that will get focus after this one
        :rtype:     (Element)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return None

        try:
            next_widget_focus = self.widget.tk_focusNext()
            return self.ParentForm.widget_to_element(next_widget_focus)
        except Exception as e:
            _error_popup_with_traceback(
                "Exception getting next focus. Check your element's Widget", e
            )

    def get_previous_focus(self):
        """
        Gets the element that should get focus previous to this element.

        :return:    Element that should get the focus before this one
        :rtype:     (Element)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return None
        try:
            next_widget_focus = self.widget.tk_focusPrev()  # tkinter.Widget
            return self.ParentForm.widget_to_element(next_widget_focus)
        except Exception as e:
            _error_popup_with_traceback(
                "Exception getting previous focus. Check your element's Widget", e
            )

    def set_size(self, size=(None, None)):
        """
        Changes the size of an element to a specific size.
        It's possible to specify None for one of sizes so that only 1 of the element's dimensions are changed.

        :param size: The size in characters, rows typically. In some cases they are pixels
        :type size:  (int, int)
        """
        try:
            if size[0] is not None:
                self.Widget.config(width=size[0])
        except Exception:
            print("Warning, error setting width on element with key=", self.Key)
        try:
            if size[1] is not None:
                self.Widget.config(height=size[1])
        except Exception:
            try:
                self.Widget.config(length=size[1])
            except Exception:
                print("Warning, error setting height on element with key=", self.Key)

        if self.Type == ELEM_TYPE_GRAPH:
            self.CanvasSize = size

    def get_size(self):
        """
        Return the size of an element in Pixels.  Care must be taken as some elements use characters to specify their size but will return pixels when calling this get_size method.
        :return: width and height of the element
        :rtype:  (int, int)
        """
        try:
            w = self.Widget.winfo_width()
            h = self.Widget.winfo_height()
        except Exception:
            print("Warning, error getting size of element", self.Key)
            w = h = None
        return w, h

    def hide_row(self):
        """
        Hide the entire row an Element is located on.
        Use this if you must have all space removed when you are hiding an element, including the row container
        """
        try:
            self.ParentRowFrame.pack_forget()
        except Exception:
            print("Warning, error hiding element row for key =", self.Key)

    def unhide_row(self):
        """
        Unhides (makes visible again) the row container that the Element is located on.
        Note that it will re-appear at the bottom of the window / container, most likely.
        """
        try:
            self.ParentRowFrame.pack()
        except Exception:
            print("Warning, error hiding element row for key =", self.Key)

    def expand(self, expand_x=False, expand_y=False, expand_row=True):
        """
        Causes the Element to expand to fill available space in the X and Y directions.  Can specify which or both directions

        :param expand_x:   If True Element will expand in the Horizontal directions
        :type expand_x:    (bool)
        :param expand_y:   If True Element will expand in the Vertical directions
        :type expand_y:    (bool)
        :param expand_row: If True the row containing the element will also expand. Without this your element is "trapped" within the row
        :type expand_row:  (bool)
        """
        if expand_x and expand_y:
            fill = tk.BOTH
        elif expand_x:
            fill = tk.X
        elif expand_y:
            fill = tk.Y
        else:
            return

        if not self._widget_was_created():
            return
        self.Widget.pack(expand=True, fill=fill)
        self.ParentRowFrame.pack(expand=expand_row, fill=fill)
        if self.element_frame is not None:
            self.element_frame.pack(expand=True, fill=fill)

    def set_cursor(self, cursor=None, cursor_color=None):
        """
        Sets the cursor for the current Element.
        "Cursor" is used in 2 different ways in this call.
        For the parameter "cursor" it's actually the mouse pointer.
        If you do not want any mouse pointer, then use the string "none"
        For the parameter "cursor_color" it's the color of the beam used when typing into an input element

        :param cursor:       The tkinter cursor name
        :type cursor:        (str)
        :param cursor_color: color to set the "cursor" to
        :type cursor_color:  (str)
        """
        if not self._widget_was_created():
            return
        if cursor is not None:
            try:
                self.Widget.config(cursor=cursor)
            except Exception as e:
                print("Warning bad cursor specified ", cursor)
                print(e)
        if cursor_color is not None:
            try:
                self.Widget.config(insertbackground=cursor_color)
            except Exception as e:
                print("Warning bad cursor color", cursor_color)
                print(e)

    def set_vscroll_position(self, percent_from_top):
        """
        Attempts to set the vertical scroll postition for an element's Widget
        :param percent_from_top: From 0 to 1.0, the percentage from the top to move scrollbar to
        :type percent_from_top:  (float)
        """
        if self.Type == ELEM_TYPE_COLUMN and self.Scrollable:
            widget = self.widget.canvas  # scrollable column is a special case
        else:
            widget = self.widget

        try:
            widget.yview_moveto(percent_from_top)
        except Exception as e:
            print("Warning setting the vertical scroll (yview_moveto failed)")
            print(e)

    def _widget_was_created(self):
        """
        Determines if a Widget was created for this element.

        :return: True if a Widget has been created previously (Widget is not None)
        :rtype:  (bool)
        """
        if self.Widget is not None:
            return True
        else:
            if SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS:
                return False

            warnings.warn(
                "You cannot Update element with key = {} until the window.read() is called or set finalize=True when creating window".format(
                    self.Key
                ),
                UserWarning,
            )
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "Unable to complete operation on element with key {}".format(
                        self.Key
                    ),
                    "You cannot perform operations (such as calling update) on an Element until:",
                    " window.read() is called or finalize=True when Window created.",
                    'Adding a "finalize=True" parameter to your Window creation will likely fix this.',
                    _create_error_message(),
                )
            return False

    def _grab_anywhere_on_using_control_key(self):
        """
        Turns on Grab Anywhere functionality AFTER a window has been created.  Don't try on a window that's not yet
        been Finalized or Read.
        """
        self.Widget.bind("<Control-Button-1>", self.ParentForm._StartMove)
        self.Widget.bind("<Control-ButtonRelease-1>", self.ParentForm._StopMove)
        self.Widget.bind("<Control-B1-Motion>", self.ParentForm._OnMotion)

    def _grab_anywhere_on(self):
        """
        Turns on Grab Anywhere functionality AFTER a window has been created.  Don't try on a window that's not yet
        been Finalized or Read.
        """
        self.Widget.bind("<ButtonPress-1>", self.ParentForm._StartMove)
        self.Widget.bind("<ButtonRelease-1>", self.ParentForm._StopMove)
        self.Widget.bind("<B1-Motion>", self.ParentForm._OnMotion)

    def _grab_anywhere_off(self):
        """
        Turns off Grab Anywhere functionality AFTER a window has been created.  Don't try on a window that's not yet
        been Finalized or Read.
        """
        self.Widget.unbind("<ButtonPress-1>")
        self.Widget.unbind("<ButtonRelease-1>")
        self.Widget.unbind("<B1-Motion>")

    def grab_anywhere_exclude(self):
        """
        Excludes this element from being used by the grab_anywhere feature
        Handy for elements like a Graph element when dragging is enabled. You want the Graph element to get the drag events instead of the window dragging.
        """
        self.ParentForm._grab_anywhere_ignore_these_list.append(self.Widget)

    def grab_anywhere_include(self):
        """
        Includes this element in the grab_anywhere feature
        This will allow you to make a Multline element drag a window for example
        """
        self.ParentForm._grab_anywhere_include_these_list.append(self.Widget)

    def set_right_click_menu(self, menu=None):
        """
        Sets a right click menu for an element.
        If a menu is already set for the element, it will call the tkinter destroy method to remove it
        :param menu:                   A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type menu:                    List[List[ List[str] | str ]]
        """
        if menu == MENU_RIGHT_CLICK_DISABLED:
            return
        if menu is None:
            menu = self.ParentForm.RightClickMenu
            if menu is None:
                return
        if menu:
            # If previously had a menu destroy it
            if self.TKRightClickMenu:
                try:
                    self.TKRightClickMenu.destroy()
                except Exception:
                    pass
            top_menu = tk.Menu(
                self.ParentForm.TKroot,
                tearoff=self.ParentForm.right_click_menu_tearoff,
                tearoffcommand=self._tearoff_menu_callback,
            )

            if self.ParentForm.right_click_menu_background_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(bg=self.ParentForm.right_click_menu_background_color)
            if self.ParentForm.right_click_menu_text_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(fg=self.ParentForm.right_click_menu_text_color)
            if self.ParentForm.right_click_menu_disabled_text_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    disabledforeground=self.ParentForm.right_click_menu_disabled_text_color
                )
            if self.ParentForm.right_click_menu_font is not None:
                top_menu.config(font=self.ParentForm.right_click_menu_font)

            if self.ParentForm.right_click_menu_selected_colors[0] not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    activeforeground=self.ParentForm.right_click_menu_selected_colors[0]
                )
            if self.ParentForm.right_click_menu_selected_colors[1] not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    activebackground=self.ParentForm.right_click_menu_selected_colors[1]
                )
            AddMenuItem(top_menu, menu[1], self, right_click_menu=True)
            self.TKRightClickMenu = top_menu
            if self.ParentForm.RightClickMenu:  # if the top level has a right click menu, then setup a callback for the Window itself
                if self.ParentForm.TKRightClickMenu is None:
                    self.ParentForm.TKRightClickMenu = top_menu
                    if running_mac():
                        self.ParentForm.TKroot.bind(
                            "<ButtonRelease-2>", self.ParentForm._RightClickMenuCallback
                        )
                    else:
                        self.ParentForm.TKroot.bind(
                            "<ButtonRelease-3>", self.ParentForm._RightClickMenuCallback
                        )
            if running_mac():
                self.Widget.bind("<ButtonRelease-2>", self._RightClickMenuCallback)
            else:
                self.Widget.bind("<ButtonRelease-3>", self._RightClickMenuCallback)

    def save_element_screenshot_to_disk(self, filename=None):
        """
        Saves an image of the PySimpleGUI window provided into the filename provided

        :param filename:        Optional filename to save screenshot to. If not included, the User Settinds are used to get the filename
        :return:                A PIL ImageGrab object that can be saved or manipulated
        :rtype:                 (PIL.ImageGrab | None)
        """
        global pil_import_attempted, pil_imported, PIL, ImageGrab, Image

        if not pil_import_attempted:
            try:
                import PIL as PIL
                from PIL import ImageGrab
                from PIL import Image

                pil_imported = True
                pil_import_attempted = True
            except Exception:
                pil_imported = False
                pil_import_attempted = True
                print("FAILED TO IMPORT PIL!")
                return None
        try:
            # Add a little to the X direction if window has a titlebar
            rect = (
                self.widget.winfo_rootx(),
                self.widget.winfo_rooty(),
                self.widget.winfo_rootx() + self.widget.winfo_width(),
                self.widget.winfo_rooty() + self.widget.winfo_height(),
            )

            grab = ImageGrab.grab(bbox=rect)
            # Save the grabbed image to disk
        except Exception as e:
            # print(e)
            popup_error_with_traceback(
                "Screen capture failure",
                "Error happened while trying to save screencapture of an element",
                e,
            )
            return None

        # return grab
        if filename is None:
            folder = pysimplegui_user_settings.get("-screenshots folder-", "")
            filename = pysimplegui_user_settings.get("-screenshots filename-", "")
            full_filename = os.path.join(folder, filename)
        else:
            full_filename = filename
        if full_filename:
            try:
                grab.save(full_filename)
            except Exception as e:
                popup_error_with_traceback(
                    "Screen capture failure",
                    "Error happened while trying to save screencapture",
                    e,
                )
        else:
            popup_error_with_traceback(
                "Screen capture failure",
                "You have attempted a screen capture but have not set up a good filename to save to",
            )
        return grab

    def _pack_forget_save_settings(self, alternate_widget=None):
        """
        Performs a pack_forget which will make a widget invisible.
        This method saves the pack settings so that they can be restored if the element is made visible again

        :param alternate_widget:   Widget to use that's different than the one defined in Element.Widget. These are usually Frame widgets
        :type alternate_widget:    (tk.Widget)
        """

        if alternate_widget is not None and self.Widget is None:
            return

        widget = alternate_widget if alternate_widget is not None else self.Widget
        # if the widget is already invisible (i.e. not packed) then will get an error
        try:
            pack_settings = widget.pack_info()
            self.pack_settings = pack_settings
            widget.pack_forget()
        except Exception:
            pass

    def _pack_restore_settings(self, alternate_widget=None):
        """
        Restores a previously packated widget which will make it visible again.
        If no settings were saved, then the widget is assumed to have not been unpacked and will not try to pack it again

        :param alternate_widget:   Widget to use that's different than the one defined in Element.Widget. These are usually Frame widgets
        :type alternate_widget:    (tk.Widget)
        """

        # if there are no saved pack settings, then assume it hasnb't been packaed before. The request will be ignored
        if self.pack_settings is None:
            return

        widget = alternate_widget if alternate_widget is not None else self.Widget
        if widget is not None:
            widget.pack(**self.pack_settings)

    def update(self, *args, **kwargs):
        """
        A dummy update call.  This will only be called if an element hasn't implemented an update method
        It is provided here for docstring purposes.  If you got here by browing code via PyCharm, know
        that this is not the function that will be called.  Your actual element's update method will be called.

        If you call update, you must call window.refresh if you want the change to happen prior to your next
        window.read() call. Normally uou don't do this as the window.read call is likely going to happen next.
        """
        print(
            "* Base Element Class update was called. Your element does not seem to have an update method"
        )

    def __call__(self, *args, **kwargs):
        """
        Makes it possible to "call" an already existing element.  When you do make the "call", it actually calls
        the Update method for the element.
        Example:    If this text element was in yoiur layout:
                    sg.Text('foo', key='T')
                    Then you can call the Update method for that element by writing:
                    window.find_element('T')('new text value')
        """
        return self.update(*args, **kwargs)

    SetTooltip = set_tooltip
    SetFocus = set_focus


# ---------------------------------------------------------------------- #
#                           Input Class                                  #
# ---------------------------------------------------------------------- #
class Input(Element):
    """
    Display a single text input field.  Based on the tkinter Widget `Entry`
    """

    def __init__(
            self,
            default_text="",
            size=(None, None),
            s=(None, None),
            disabled=False,
            password_char="",
            justification=None,
            background_color=None,
            text_color=None,
            font=None,
            tooltip=None,
            border_width=None,
            change_submits=False,
            enable_events=False,
            do_not_clear=True,
            key=None,
            k=None,
            focus=False,
            pad=None,
            p=None,
            use_readonly_for_disable=True,
            readonly=False,
            disabled_readonly_background_color=None,
            disabled_readonly_text_color=None,
            selected_text_color=None,
            selected_background_color=None,
            expand_x=False,
            expand_y=False,
            right_click_menu=None,
            visible=True,
            metadata=None,
    ):
        """
        :param default_text:                       Text initially shown in the input box as a default value(Default value = ''). Will automatically be converted to string
        :type default_text:                        (Any)
        :param size:                               w=characters-wide, h=rows-high. If an int is supplied rather than a tuple, then a tuple is created width=int supplied and heigh=1
        :type size:                                (int, int) |  (int, None) | int
        :param s:                                  Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                                   (int, int)  | (None, None) | int
        :param disabled:                           set disable state for element (Default = False)
        :type disabled:                            (bool)
        :param password_char:                      Password character if this is a password field (Default value = '')
        :type password_char:                       (char)
        :param justification:                      justification for data display. Valid choices - left, right, center
        :type justification:                       (str)
        :param background_color:                   color of background in one of the color formats
        :type background_color:                    (str)
        :param text_color:                         color of the text
        :type text_color:                          (str)
        :param font:                               specifies the font family, size. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                                (str or (str, int[, str]) or None)
        :param tooltip:                            text, that will appear when mouse hovers over the element
        :type tooltip:                             (str)
        :param border_width:                       width of border around element in pixels
        :type border_width:                        (int)
        :param change_submits:                     * DEPRICATED DO NOT USE. Use `enable_events` instead
        :type change_submits:                      (bool)
        :param enable_events:                      If True then changes to this element are immediately reported as an event. Use this instead of change_submits (Default = False)
        :type enable_events:                       (bool)
        :param do_not_clear:                       If False then the field will be set to blank after ANY event (button, any event) (Default = True)
        :type do_not_clear:                        (bool)
        :param key:                                Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                                 str | int | tuple | object
        :param k:                                  Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                                   str | int | tuple | object
        :param focus:                              Determines if initial focus should go to this element.
        :type focus:                               (bool)
        :param pad:                                Amount of padding to put around element. Normally (horizontal pixels, vertical pixels) but can be split apart further into ((horizontal left, horizontal right), (vertical above, vertical below)). If int is given, then converted to tuple (int, int) with the value provided duplicated
        :type pad:                                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                                  Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                                   (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param use_readonly_for_disable:           If True (the default) tkinter state set to 'readonly'. Otherwise state set to 'disabled'
        :type use_readonly_for_disable:            (bool)
        :param readonly:                           If True tkinter state set to 'readonly'.  Use this in place of use_readonly_for_disable as another way of achieving readonly.  Note cannot set BOTH readonly and disabled as tkinter only supplies a single flag
        :type readonly:                            (bool)
        :param disabled_readonly_background_color: If state is set to readonly or disabled, the color to use for the background
        :type disabled_readonly_background_color:  (str)
        :param disabled_readonly_text_color:       If state is set to readonly or disabled, the color to use for the text
        :type disabled_readonly_text_color:        (str)
        :param selected_text_color:                Color of text when it is selected (using mouse or control+A, etc)
        :type selected_text_color:                 (str)
        :param selected_background_color:          Color of background when it is selected (using mouse or control+A, etc)
        :type selected_background_color:           (str)
        :param expand_x:                           If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                            (bool)
        :param expand_y:                           If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                            (bool)
        :param right_click_menu:                   A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:                    List[List[ List[str] | str ]]
        :param visible:                            set visibility state of the element (Default = True)
        :type visible:                             (bool)
        :param metadata:                           User metadata that can be set to ANYTHING
        :type metadata:                            (Any)
        """

        self.DefaultText = default_text if default_text is not None else ""
        self.PasswordCharacter = password_char
        bg = (
            background_color
            if background_color is not None
            else DEFAULT_INPUT_ELEMENTS_COLOR
        )
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        self.selected_text_color = selected_text_color
        self.selected_background_color = selected_background_color
        self.Focus = focus
        self.do_not_clear = do_not_clear
        self.Justification = justification
        self.Disabled = disabled
        self.ChangeSubmits = change_submits or enable_events
        self.RightClickMenu = right_click_menu
        self.UseReadonlyForDisable = use_readonly_for_disable
        self.disabled_readonly_background_color = disabled_readonly_background_color
        self.disabled_readonly_text_color = disabled_readonly_text_color
        self.ReadOnly = readonly
        self.BorderWidth = (
            border_width if border_width is not None else DEFAULT_BORDER_WIDTH
        )
        self.TKEntry = self.Widget = None  # type: tk.Entry
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_TEXT,
            size=sz,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            font=font,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )

    def update(
            self,
            value=None,
            disabled=None,
            select=None,
            visible=None,
            text_color=None,
            background_color=None,
            font=None,
            move_cursor_to="end",
            password_char=None,
            paste=None,
            readonly=None,
    ):
        """
        Changes some of the settings for the Input Element. Must call `Window.Read` or `Window.Finalize` prior.
        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:            new text to display as default text in Input field
        :type value:             (str)
        :param disabled:         disable or enable state of the element (sets Entry Widget to readonly or normal)
        :type disabled:          (bool)
        :param select:           if True, then the text will be selected
        :type select:            (bool)
        :param visible:          change visibility of element
        :type visible:           (bool)
        :param text_color:       change color of text being typed
        :type text_color:        (str)
        :param background_color: change color of the background
        :type background_color:  (str)
        :param font:             specifies the font family, size. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param move_cursor_to:   Moves the cursor to a particular offset. Defaults to 'end'
        :type move_cursor_to:    int | str
        :param password_char:    Password character if this is a password field
        :type password_char:     str
        :param paste:            If True "Pastes" the value into the element rather than replacing the entire element. If anything is selected it is replaced. The text is inserted at the current cursor location.
        :type paste:             bool
        :param readonly:         if True make element readonly (user cannot change any choices). Enables the element if either choice are made.
        :type readonly:          (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Input.update - The window was closed")
            return

        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKEntry.configure(background=background_color)
            self.BackgroundColor = background_color
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKEntry.configure(fg=text_color)
            self.TextColor = text_color

        if disabled:
            if self.UseReadonlyForDisable:
                if self.disabled_readonly_text_color not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ):
                    self.TKEntry.configure(fg=self.disabled_readonly_text_color)
                self.TKEntry["state"] = "readonly"
            else:
                if self.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                    self.TKEntry.configure(fg=self.TextColor)
                self.TKEntry["state"] = "disabled"
            self.Disabled = True
        elif not disabled:
            self.TKEntry["state"] = "normal"
            if self.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                self.TKEntry.configure(fg=self.TextColor)
            self.Disabled = False

        if readonly:
            self.TKEntry["state"] = "readonly"
        elif not readonly:
            self.TKEntry["state"] = "normal"

        if value is not None:
            if paste:
                try:
                    self.TKStringVar.set(value)
                except Exception:
                    pass
            self.DefaultText = value
            if paste:
                try:
                    self.TKEntry.delete("sel.first", "sel.last")
                except Exception:
                    pass
                self.TKEntry.insert("insert", value)
            if move_cursor_to == "end":
                self.TKEntry.icursor(tk.END)
            elif move_cursor_to is not None:
                self.TKEntry.icursor(move_cursor_to)
        if select:
            self.TKEntry.select_range(0, "end")
        if not visible:
            self._pack_forget_save_settings()
            # self.TKEntry.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKEntry.pack(padx=self.pad_used[0], pady=self.pad_used[1])
            # self.TKEntry.pack(padx=self.pad_used[0], pady=self.pad_used[1], in_=self.ParentRowFrame)
        if visible is not None:
            self._visible = visible
        if password_char is not None:
            self.TKEntry.configure(show=password_char)
            self.PasswordCharacter = password_char
        if font is not None:
            self.TKEntry.configure(font=font)

    def set_ibeam_color(self, ibeam_color=None):
        """
        Sets the color of the I-Beam that is used to "insert" characters. This is oftens called a "Cursor" by
        many users.  To keep from being confused with tkinter's definition of cursor (the mouse pointer), the term
        ibeam is used in this case.
        :param ibeam_color: color to set the "I-Beam" used to indicate where characters will be inserted
        :type ibeam_color:  (str)
        """

        if not self._widget_was_created():
            return
        if ibeam_color is not None:
            try:
                self.Widget.config(insertbackground=ibeam_color)
            except Exception as e:
                _error_popup_with_traceback(
                    "Error setting I-Beam color in set_ibeam_color",
                    "The element has a key:",
                    self.Key,
                    "The color passed in was:",
                    ibeam_color,
                )

    def get(self):
        """
        Read and return the current value of the input element. Must call `Window.Read` or `Window.Finalize` prior

        :return: current value of Input field or '' if error encountered
        :rtype:  (str)
        """
        try:
            text = self.TKStringVar.get()
        except Exception:
            text = ""
        return text

    Get = get
    Update = update


# -------------------------  INPUT TEXT Element lazy functions  ------------------------- #
In = Input
InputText = Input
I = Input


# ---------------------------------------------------------------------- #
#                           Combo                                        #
# ---------------------------------------------------------------------- #
class Combo(Element):
    """
    ComboBox Element - A combination of a single-line input and a drop-down menu. User can type in their own value or choose from list.
    """

    def __init__(
            self,
            values,
            default_value=None,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            background_color=None,
            text_color=None,
            button_background_color=None,
            button_arrow_color=None,
            bind_return_key=False,
            change_submits=False,
            enable_events=False,
            enable_per_char_events=None,
            disabled=False,
            key=None,
            k=None,
            pad=None,
            p=None,
            expand_x=False,
            expand_y=False,
            tooltip=None,
            readonly=False,
            font=None,
            visible=True,
            metadata=None,
    ):
        """
        :param values:                  values to choose. While displayed as text, the items returned are what the caller supplied, not text
        :type values:                   List[Any] or Tuple[Any]
        :param default_value:           Choice to be displayed as initial value. Must match one of values variable contents
        :type default_value:            (Any)
        :param size:                    width, height. Width = characters-wide, height = NOTE it's the number of entries to show in the list. If an Int is passed rather than a tuple, then height is auto-set to 1 and width is value of the int
        :type size:                     (int, int)  | (None, None) | int
        :param s:                       Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                        (int, int)  | (None, None) | int
        :param auto_size_text:          True if element should be the same size as the contents
        :type auto_size_text:           (bool)
        :param background_color:        color of background
        :type background_color:         (str)
        :param text_color:              color of the text
        :type text_color:               (str)
        :param button_background_color: The color of the background of the button on the combo box
        :type button_background_color:  (str)
        :param button_arrow_color:      The color of the arrow on the button on the combo box
        :type button_arrow_color:       (str)
        :param bind_return_key:         If True, then the return key will cause a the Combo to generate an event when return key is pressed
        :type bind_return_key:          (bool)
        :param change_submits:          DEPRICATED DO NOT USE. Use `enable_events` instead
        :type change_submits:           (bool)
        :param enable_events:           Turns on the element specific events. Combo event is when a choice is made
        :type enable_events:            (bool)
        :param enable_per_char_events:  Enables generation of events for every character that's input. This is like the Input element's events
        :type enable_per_char_events:   (bool)
        :param disabled:                set disable state for element
        :type disabled:                 (bool)
        :param key:                     Used with window.find_element and with return values to uniquely identify this element
        :type key:                      str | int | tuple | object
        :param k:                       Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                        str | int | tuple | object
        :param pad:                     Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                       Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param expand_x:                If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                 (bool)
        :param expand_y:                If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                 (bool)
        :param tooltip:                 text that will appear when mouse hovers over this element
        :type tooltip:                  (str)
        :param readonly:                make element readonly (user can't change). True means user cannot change
        :type readonly:                 (bool)
        :param font:                    specifies the font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                     (str or (str, int[, str]) or None)
        :param visible:                 set visibility state of the element
        :type visible:                  (bool)
        :param metadata:                User metadata that can be set to ANYTHING
        :type metadata:                 (Any)
        """

        self.Values = values
        self.DefaultValue = default_value
        self.ChangeSubmits = change_submits or enable_events
        self.Widget = self.TKCombo = None  # type: ttk.Combobox
        self.Disabled = disabled
        self.Readonly = readonly
        self.BindReturnKey = bind_return_key
        bg = background_color if background_color else DEFAULT_INPUT_ELEMENTS_COLOR
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y
        if button_background_color is None:
            self.button_background_color = theme_button_color()[1]
        else:
            self.button_background_color = button_background_color
        if button_arrow_color is None:
            self.button_arrow_color = theme_button_color()[0]
        else:
            self.button_arrow_color = button_arrow_color
        self.enable_per_char_events = enable_per_char_events

        super().__init__(
            ELEM_TYPE_INPUT_COMBO,
            size=sz,
            auto_size_text=auto_size_text,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            tooltip=tooltip,
            font=font or DEFAULT_FONT,
            visible=visible,
            metadata=metadata,
        )

    def update(
            self,
            value=None,
            values=None,
            set_to_index=None,
            disabled=None,
            readonly=None,
            font=None,
            visible=None,
            size=(None, None),
            select=None,
            text_color=None,
            background_color=None,
    ):
        """
        Changes some of the settings for the Combo Element. Must call `Window.Read` or `Window.Finalize` prior.
        Note that the state can be in 3 states only.... enabled, disabled, readonly even
        though more combinations are available. The easy way to remember is that if you
        change the readonly parameter then you are enabling the element.

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:            change which value is current selected based on new list of previous list of choices
        :type value:             (Any)
        :param values:           change list of choices
        :type values:            List[Any]
        :param set_to_index:     change selection to a particular choice starting with index = 0
        :type set_to_index:      (int)
        :param disabled:         disable or enable state of the element
        :type disabled:          (bool)
        :param readonly:         if True make element readonly (user cannot change any choices). Enables the element if either choice are made.
        :type readonly:          (bool)
        :param font:             specifies the font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param visible:          control visibility of element
        :type visible:           (bool)
        :param size:             width, height. Width = characters-wide, height = NOTE it's the number of entries to show in the list
        :type size:              (int, int)
        :param select:           if True, then the text will be selected, if False then selection will be cleared
        :type select:            (bool)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        """

        if size != (None, None):
            if isinstance(size, int):
                size = (size, 1)
            if isinstance(size, tuple) and len(size) == 1:
                size = (size[0], 1)

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Combo.update - The window was closed")
            return

        if values is not None:
            try:
                self.TKCombo["values"] = values
                # self.TKCombo.current(0)       # don't set any value if a new set of values was made
            except Exception:
                pass
            self.Values = values
            if value is None:
                self.TKCombo.set("")
            if size == (None, None):
                max_line_len = (
                    max([len(str(l)) for l in self.Values]) if len(self.Values) else 0
                )
                if not self.AutoSizeText:
                    width = self.Size[0]
                else:
                    width = max_line_len + 1
                self.TKCombo.configure(width=width)
            else:
                self.TKCombo.configure(height=size[1])
                self.TKCombo.configure(width=size[0])
        if value is not None:
            if value not in self.Values:
                self.TKCombo.set(value)
            else:
                for index, v in enumerate(self.Values):
                    if v == value:
                        try:
                            self.TKCombo.current(index)
                        except Exception:
                            pass
                        self.DefaultValue = value
                        break
        if set_to_index is not None:
            try:
                self.TKCombo.current(set_to_index)
                self.DefaultValue = self.Values[set_to_index]
            except Exception:
                pass
        if readonly:
            self.Readonly = True
            self.TKCombo["state"] = "readonly"
        elif not readonly:
            self.Readonly = False
            self.TKCombo["state"] = "enable"
        if disabled:
            self.TKCombo["state"] = "disable"
        elif disabled is False and self.Readonly is True:
            self.TKCombo["state"] = "readonly"
        elif disabled is False and self.Readonly is False:
            self.TKCombo["state"] = "enable"
        self.Disabled = disabled if disabled is not None else self.Disabled

        combostyle = self.ttk_style
        style_name = self.ttk_style_name
        if text_color is not None:
            combostyle.configure(style_name, foreground=text_color)
            combostyle.configure(style_name, selectforeground=text_color)
            combostyle.configure(style_name, insertcolor=text_color)
            combostyle.map(style_name, fieldforeground=[("readonly", text_color)])
            self.TextColor = text_color
        if background_color is not None:
            combostyle.configure(style_name, selectbackground=background_color)
            combostyle.map(style_name, fieldbackground=[("readonly", background_color)])
            combostyle.configure(style_name, fieldbackground=background_color)
            self.BackgroundColor = background_color

        if self.Readonly:
            if text_color not in (None, COLOR_SYSTEM_DEFAULT):
                combostyle.configure(style_name, selectforeground=text_color)
            if background_color not in (None, COLOR_SYSTEM_DEFAULT):
                combostyle.configure(style_name, selectbackground=background_color)

        if font is not None:
            self.Font = font
            self.TKCombo.configure(font=font)
            self._dropdown_newfont = tkinter.font.Font(font=font)
            self.ParentRowFrame.option_add(
                "*TCombobox*Listbox*Font", self._dropdown_newfont
            )

        # make tcl call to deal with colors for the drop-down formatting
        try:
            if self.BackgroundColor not in (
                    None,
                    COLOR_SYSTEM_DEFAULT,
            ) and self.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                self.Widget.tk.eval(
                    "[ttk::combobox::PopdownWindow {}].f.l configure -foreground {} -background {} -selectforeground {} -selectbackground {} -font {}".format(
                        self.Widget,
                        self.TextColor,
                        self.BackgroundColor,
                        self.BackgroundColor,
                        self.TextColor,
                        self._dropdown_newfont,
                    )
                )
        except Exception as e:
            pass  # going to let this one slide

        if not visible:
            self._pack_forget_save_settings()
            # self.TKCombo.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKCombo.pack(padx=self.pad_used[0], pady=self.pad_used[1])
        if visible is not None:
            self._visible = visible
        if select:
            self.TKCombo.select_range(0, tk.END)
        elif not select:
            self.TKCombo.select_clear()

    def get(self):
        """
        Returns the current (right now) value of the Combo.  DO NOT USE THIS AS THE NORMAL WAY OF READING A COMBO!
        You should be using values from your call to window.read instead.  Know what you're doing if you use it.

        :return: Returns the value of what is currently chosen
        :rtype:  Any | None
        """
        try:
            if (
                    self.TKCombo.current() == -1
            ):  # if the current value was not in the original list
                value = self.TKCombo.get()  # then get the value typed in by user
            else:
                value = self.Values[
                    self.TKCombo.current()
                ]  # get value from original list given index
        except Exception:
            value = None  # only would happen if user closes window
        return value

    Get = get
    Update = update


# -------------------------  INPUT COMBO Element lazy functions  ------------------------- #
InputCombo = Combo
DropDown = InputCombo
Drop = InputCombo
DD = Combo


# ---------------------------------------------------------------------- #
#                           Option Menu                                  #
# ---------------------------------------------------------------------- #
class OptionMenu(Element):
    """
    Option Menu is an Element available ONLY on the tkinter port of PySimpleGUI.  It's is a widget that is unique
    to tkinter.  However, it looks much like a ComboBox.  Instead of an arrow to click to pull down the list of
    choices, another little graphic is shown on the widget to indicate where you click.  After clicking to activate,
    it looks like a Combo Box that you scroll to select a choice.
    """

    def __init__(
            self,
            values,
            default_value=None,
            size=(None, None),
            s=(None, None),
            disabled=False,
            auto_size_text=None,
            expand_x=False,
            expand_y=False,
            background_color=None,
            text_color=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            tooltip=None,
            visible=True,
            metadata=None,
    ):
        """
        :param values:           Values to be displayed
        :type values:            List[Any] or Tuple[Any]
        :param default_value:    the value to choose by default
        :type default_value:     (Any)
        :param size:             (width, height) size in characters (wide), height is ignored and present to be consistent with other elements
        :type size:              (int, int) (width, UNUSED)
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param disabled:         control enabled / disabled
        :type disabled:          (bool)
        :param auto_size_text:   True if size of Element should match the contents of the items
        :type auto_size_text:    (bool)
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param key:              Used with window.find_element and with return values to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:          text that will appear when mouse hovers over this element
        :type tooltip:           (str)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.Values = values
        self.DefaultValue = default_value
        self.Widget = self.TKOptionMenu = None  # type: tk.OptionMenu
        self.Disabled = disabled
        bg = background_color if background_color else DEFAULT_INPUT_ELEMENTS_COLOR
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_OPTION_MENU,
            size=sz,
            auto_size_text=auto_size_text,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )

    def update(
            self, value=None, values=None, disabled=None, visible=None, size=(None, None)
    ):
        """
        Changes some of the settings for the OptionMenu Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:    the value to choose by default
        :type value:     (Any)
        :param values:   Values to be displayed
        :type values:    List[Any]
        :param disabled: disable or enable state of the element
        :type disabled:  (bool)
        :param visible:  control visibility of element
        :type visible:   (bool)
        :param size:     (width, height) size in characters (wide), height is ignored and present to be consistent with other elements
        :type size:      (int, int) (width, UNUSED)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in OptionMenu.update - The window was closed"
            )
            return

        if values is not None:
            self.Values = values
            self.TKOptionMenu["menu"].delete(0, "end")

            # Insert list of new options (tk._setit hooks them up to var)
            # self.TKStringVar.set(self.Values[0])
            for new_value in self.Values:
                self.TKOptionMenu["menu"].add_command(
                    label=new_value, command=tk._setit(self.TKStringVar, new_value)
                )
            if value is None:
                self.TKStringVar.set("")

            if size == (None, None):
                max_line_len = (
                    max([len(str(l)) for l in self.Values]) if len(self.Values) else 0
                )
                if not self.AutoSizeText:
                    width = self.Size[0]
                else:
                    width = max_line_len + 1
                self.TKOptionMenu.configure(width=width)
            else:
                self.TKOptionMenu.configure(width=size[0])

        if value is not None:
            self.DefaultValue = value
            self.TKStringVar.set(value)

        if disabled:
            self.TKOptionMenu["state"] = "disabled"
        elif not disabled:
            self.TKOptionMenu["state"] = "normal"
        self.Disabled = disabled if disabled is not None else self.Disabled
        if not visible:
            self._pack_forget_save_settings()
            # self.TKOptionMenu.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKOptionMenu.pack(padx=self.pad_used[0], pady=self.pad_used[1])
        if visible is not None:
            self._visible = visible

    Update = update


# -------------------------  OPTION MENU Element lazy functions  ------------------------- #
InputOptionMenu = OptionMenu


# ---------------------------------------------------------------------- #
#                           Listbox                                      #
# ---------------------------------------------------------------------- #
class Listbox(Element):
    """
    A List Box.  Provide a list of values for the user to choose one or more of.   Returns a list of selected rows
    when a window.read() is executed.
    """

    def __init__(
            self,
            values,
            default_values=None,
            select_mode=None,
            change_submits=False,
            enable_events=False,
            bind_return_key=False,
            size=(None, None),
            s=(None, None),
            disabled=False,
            justification=None,
            auto_size_text=None,
            font=None,
            no_scrollbar=False,
            horizontal_scroll=False,
            background_color=None,
            text_color=None,
            highlight_background_color=None,
            highlight_text_color=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            tooltip=None,
            expand_x=False,
            expand_y=False,
            right_click_menu=None,
            visible=True,
            metadata=None,
    ):
        """
        :param values:                     list of values to display. Can be any type including mixed types as long as they have __str__ method
        :type values:                      List[Any] or Tuple[Any]
        :param default_values:             which values should be initially selected
        :type default_values:              List[Any]
        :param select_mode:                Select modes are used to determine if only 1 item can be selected or multiple and how they can be selected.   Valid choices begin with "LISTBOX_SELECT_MODE_" and include: LISTBOX_SELECT_MODE_SINGLE LISTBOX_SELECT_MODE_MULTIPLE LISTBOX_SELECT_MODE_BROWSE LISTBOX_SELECT_MODE_EXTENDED
        :type select_mode:                 [enum]
        :param change_submits:             DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:              (bool)
        :param enable_events:              Turns on the element specific events. Listbox generates events when an item is clicked
        :type enable_events:               (bool)
        :param bind_return_key:            If True, then the return key will cause a the Listbox to generate an event when return key is pressed
        :type bind_return_key:             (bool)
        :param size:                       w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                        (int, int) |  (int, None) | int
        :param s:                          Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                           (int, int)  | (None, None) | int
        :param disabled:                   set disable state for element
        :type disabled:                    (bool)
        :param justification:              justification for items in listbox. Valid choices - left, right, center.  Default is left. NOTE - on some older versions of tkinter, not available
        :type justification:               (str)
        :param auto_size_text:             True if element should be the same size as the contents
        :type auto_size_text:              (bool)
        :param font:                       specifies the font family, size, etc.  Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                        (str or (str, int[, str]) or None)
        :param no_scrollbar:               Controls if a scrollbar should be shown.  If True, no scrollbar will be shown
        :type no_scrollbar:                (bool)
        :param horizontal_scroll:          Controls if a horizontal scrollbar should be shown.  If True a horizontal scrollbar will be shown in addition to vertical
        :type horizontal_scroll:           (bool)
        :param background_color:           color of background
        :type background_color:            (str)
        :param text_color:                 color of the text
        :type text_color:                  (str)
        :param highlight_background_color: color of the background when an item is selected. Defaults to normal text color (a reverse look)
        :type highlight_background_color:  (str)
        :param highlight_text_color:       color of the text when an item is selected. Defaults to the normal background color (a rerverse look)
        :type highlight_text_color:        (str)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        :param key:                        Used with window.find_element and with return values to uniquely identify this element
        :type key:                         str | int | tuple | object
        :param k:                          Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                           str | int | tuple | object
        :param pad:                        Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                         (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                          Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                           (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:                    text, that will appear when mouse hovers over the element
        :type tooltip:                     (str)
        :param expand_x:                   If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                    (bool)
        :param expand_y:                   If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                    (bool)
        :param right_click_menu:           A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:            List[List[ List[str] | str ]]
        :param visible:                    set visibility state of the element
        :type visible:                     (bool)
        :param metadata:                   User metadata that can be set to ANYTHING
        :type metadata:                    (Any)
        """

        if values is None:
            _error_popup_with_traceback(
                "Error in your Listbox definition - The values parameter cannot be None",
                "Use an empty list if you want no values in your Listbox",
            )

        self.Values = values
        self.DefaultValues = default_values
        self.TKListbox = None
        self.ChangeSubmits = change_submits or enable_events
        self.BindReturnKey = bind_return_key
        self.Disabled = disabled
        if select_mode == LISTBOX_SELECT_MODE_BROWSE:
            self.SelectMode = SELECT_MODE_BROWSE
        elif select_mode == LISTBOX_SELECT_MODE_EXTENDED:
            self.SelectMode = SELECT_MODE_EXTENDED
        elif select_mode == LISTBOX_SELECT_MODE_MULTIPLE:
            self.SelectMode = SELECT_MODE_MULTIPLE
        elif select_mode == LISTBOX_SELECT_MODE_SINGLE:
            self.SelectMode = SELECT_MODE_SINGLE
        else:
            self.SelectMode = DEFAULT_LISTBOX_SELECT_MODE
        bg = (
            background_color
            if background_color is not None
            else theme_input_background_color()
        )
        fg = text_color if text_color is not None else theme_input_text_color()
        self.HighlightBackgroundColor = (
            highlight_background_color if highlight_background_color is not None else fg
        )
        self.HighlightTextColor = (
            highlight_text_color if highlight_text_color is not None else bg
        )
        self.RightClickMenu = right_click_menu
        self.vsb = None  # type: tk.Scrollbar or None
        self.hsb = None  # type: tk.Scrollbar | None
        self.TKListbox = self.Widget = None  # type: tk.Listbox
        self.element_frame = None  # type: tk.Frame
        self.NoScrollbar = no_scrollbar
        self.HorizontalScroll = horizontal_scroll
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y
        self.justification = justification

        super().__init__(
            ELEM_TYPE_INPUT_LISTBOX,
            size=sz,
            auto_size_text=auto_size_text,
            font=font,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )

    def update(
            self,
            values=None,
            disabled=None,
            set_to_index=None,
            scroll_to_index=None,
            select_mode=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Listbox Element. Must call `Window.Read` or `Window.Finalize` prior
        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param values:          new list of choices to be shown to user
        :type values:           List[Any]
        :param disabled:        disable or enable state of the element
        :type disabled:         (bool)
        :param set_to_index:    highlights the item(s) indicated. If parm is an int one entry will be set. If is a list, then each entry in list is highlighted
        :type set_to_index:     int | list | tuple
        :param scroll_to_index: scroll the listbox so that this index is the first shown
        :type scroll_to_index:  (int)
        :param select_mode:     changes the select mode according to tkinter's listbox widget
        :type select_mode:      (str)
        :param visible:         control visibility of element
        :type visible:          (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Listbox.update - The window was closed"
            )
            return

        if disabled:
            self.TKListbox.configure(state="disabled")
        elif not disabled:
            self.TKListbox.configure(state="normal")
        self.Disabled = disabled if disabled is not None else self.Disabled

        if values is not None:
            self.TKListbox.delete(0, "end")
            for item in list(values):
                self.TKListbox.insert(tk.END, item)
            # self.TKListbox.selection_set(0, 0)
            self.Values = list(values)
        if set_to_index is not None:
            self.TKListbox.selection_clear(
                0, len(self.Values)
            )  # clear all listbox selections
            if type(set_to_index) in (tuple, list):
                for i in set_to_index:
                    try:
                        self.TKListbox.selection_set(i, i)
                    except Exception:
                        warnings.warn(
                            "* Listbox Update selection_set failed with index {}*".format(
                                set_to_index
                            )
                        )
            else:
                try:
                    self.TKListbox.selection_set(set_to_index, set_to_index)
                except Exception:
                    warnings.warn(
                        "* Listbox Update selection_set failed with index {}*".format(
                            set_to_index
                        )
                    )
        if not visible:
            self._pack_forget_save_settings(self.element_frame)
        elif visible:
            self._pack_restore_settings(self.element_frame)
        if scroll_to_index is not None and len(self.Values):
            self.TKListbox.yview_moveto(scroll_to_index / len(self.Values))
        if select_mode is not None:
            try:
                self.TKListbox.config(selectmode=select_mode)
            except Exception:
                print("Listbox.update error trying to change mode to: ", select_mode)
        if visible is not None:
            self._visible = visible

    def set_value(self, values):
        """
        Set listbox highlighted choices

        :param values: new values to choose based on previously set values
        :type values:  List[Any] | Tuple[Any]

        """
        for index, item in enumerate(self.Values):
            try:
                if item in values:
                    self.TKListbox.selection_set(index)
                else:
                    self.TKListbox.selection_clear(index)
            except Exception:
                pass
        self.DefaultValues = values

    def get_list_values(self):
        # type: (Listbox) -> List[Any]
        """
        Returns list of Values provided by the user in the user's format

        :return: List of values. Can be any / mixed types -> []
        :rtype:  List[Any]
        """
        return self.Values

    def get_indexes(self):
        """
        Returns the items currently selected as a list of indexes

        :return: A list of offsets into values that is currently selected
        :rtype:  List[int]
        """
        return self.TKListbox.curselection()

    def get(self):
        """
        Returns the list of items currently selected in this listbox.  It should be identical
        to the value you would receive when performing a window.read() call.

        :return: The list of currently selected items. The actual items are returned, not the indexes
        :rtype:  List[Any]
        """
        try:
            items = self.TKListbox.curselection()
            value = [self.Values[int(item)] for item in items]
        except Exception:
            value = []
        return value

    def select_index(
            self, index, highlight_text_color=None, highlight_background_color=None
    ):
        """
        Selects an index while providing capability to setting the selected color for the index to specific text/background color

        :param index:                      specifies which item to change. index starts at 0 and goes to length of values list minus one
        :type  index:                      (int)
        :param highlight_text_color:       color of the text when this item is selected.
        :type  highlight_text_color:        (str)
        :param highlight_background_color: color of the background when this item is selected
        :type  highlight_background_color:  (str)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Listbox.select_item - The window was closed"
            )
            return

        if index >= len(self.Values):
            _error_popup_with_traceback(
                "Index {} is out of range for Listbox.select_index. Max allowed index is {}.".format(
                    index, len(self.Values) - 1
                )
            )
            return

        self.TKListbox.selection_set(index, index)

        if highlight_text_color is not None:
            self.widget.itemconfig(index, selectforeground=highlight_text_color)
        if highlight_background_color is not None:
            self.widget.itemconfig(index, selectbackground=highlight_background_color)

    def set_index_color(
            self,
            index,
            text_color=None,
            background_color=None,
            highlight_text_color=None,
            highlight_background_color=None,
    ):
        """
        Sets the color of a specific item without selecting it

        :param index:                      specifies which item to change. index starts at 0 and goes to length of values list minus one
        :type  index:                      (int)
        :param text_color:                 color of the text for this item
        :type  text_color:                 (str)
        :param background_color:           color of the background for this item
        :type  background_color:           (str)
        :param highlight_text_color:       color of the text when this item is selected.
        :type  highlight_text_color:       (str)
        :param highlight_background_color: color of the background when this item is selected
        :type  highlight_background_color: (str)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Listbox.set_item_color - The window was closed"
            )
            return

        if index >= len(self.Values):
            _error_popup_with_traceback(
                "Index {} is out of range for Listbox.set_index_color. Max allowed index is {}.".format(
                    index, len(self.Values) - 1
                )
            )
            return

        if text_color is not None:
            self.widget.itemconfig(index, fg=text_color)
        if background_color is not None:
            self.widget.itemconfig(index, bg=background_color)
        if highlight_text_color is not None:
            self.widget.itemconfig(index, selectforeground=highlight_text_color)
        if highlight_background_color is not None:
            self.widget.itemconfig(index, selectbackground=highlight_background_color)

    GetIndexes = get_indexes
    GetListValues = get_list_values
    SetValue = set_value
    Update = update


LBox = Listbox
LB = Listbox


# ---------------------------------------------------------------------- #
#                           Radio                                        #
# ---------------------------------------------------------------------- #
class Radio(Element):
    """
    Radio Button Element - Used in a group of other Radio Elements to provide user with ability to select only
    1 choice in a list of choices.
    """

    def __init__(
            self,
            text,
            group_id,
            default=False,
            disabled=False,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            background_color=None,
            text_color=None,
            circle_color=None,
            font=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            tooltip=None,
            change_submits=False,
            enable_events=False,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param text:             Text to display next to button
        :type text:              (str)
        :param group_id:         Groups together multiple Radio Buttons. Any type works
        :type group_id:          (Any)
        :param default:          Set to True for the one element of the group you want initially selected
        :type default:           (bool)
        :param disabled:         set disable state
        :type disabled:          (bool)
        :param size:             (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:              (int, int)  | (None, None) | int
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param auto_size_text:   if True will size the element to match the length of the text
        :type auto_size_text:    (bool)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param circle_color:     color of background of the circle that has the dot selection indicator in it
        :type circle_color:      (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param key:              Used with window.find_element and with return values to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param change_submits:   DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:    (bool)
        :param enable_events:    Turns on the element specific events. Radio Button events happen when an item is selected
        :type enable_events:     (bool)
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.InitialState = default
        self.Text = text
        self.Widget = self.TKRadio = None  # type: tk.Radiobutton
        self.GroupID = group_id
        self.Value = None
        self.Disabled = disabled
        self.TextColor = text_color if text_color else theme_text_color()
        self.RightClickMenu = right_click_menu

        if circle_color is None:
            # ---- compute color of circle background ---
            try:  # something in here will fail if a color is not specified in Hex
                text_hsl = _hex_to_hsl(self.TextColor)
                background_hsl = _hex_to_hsl(
                    background_color if background_color else theme_background_color()
                )
                l_delta = abs(text_hsl[2] - background_hsl[2]) / 10
                if (
                        text_hsl[2] > background_hsl[2]
                ):  # if the text is "lighter" than the background then make background darker
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] - l_delta,
                    )
                else:
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] + l_delta,
                    )
                self.CircleBackgroundColor = rgb(*bg_rbg)
            except Exception:
                self.CircleBackgroundColor = (
                    background_color if background_color else theme_background_color()
                )
        else:
            self.CircleBackgroundColor = circle_color
        self.ChangeSubmits = change_submits or enable_events
        self.EncodedRadioValue = None
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_RADIO,
            size=sz,
            auto_size_text=auto_size_text,
            font=font,
            background_color=background_color,
            text_color=self.TextColor,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )

    def update(
            self,
            value=None,
            text=None,
            background_color=None,
            text_color=None,
            circle_color=None,
            disabled=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Radio Button Element. Must call `Window.read` or `Window.finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:            if True change to selected and set others in group to unselected
        :type value:             (bool)
        :param text:             Text to display next to radio button
        :type text:              (str)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text. Note this also changes the color of the selection dot
        :type text_color:        (str)
        :param circle_color:     color of background of the circle that has the dot selection indicator in it
        :type circle_color:      (str)
        :param disabled:         disable or enable state of the element
        :type disabled:          (bool)
        :param visible:          control visibility of element
        :type visible:           (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Radio.update - The window was closed")
            return

        if value is not None:
            try:
                if value:
                    self.TKIntVar.set(self.EncodedRadioValue)
                elif not value:
                    if self.TKIntVar.get() == self.EncodedRadioValue:
                        self.TKIntVar.set(0)
            except Exception:
                print("Error updating Radio")
            self.InitialState = value
        if text is not None:
            self.Text = str(text)
            self.TKRadio.configure(text=self.Text)
        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKRadio.configure(background=background_color)
            self.BackgroundColor = background_color
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKRadio.configure(fg=text_color)
            self.TextColor = text_color

        if circle_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.CircleBackgroundColor = circle_color
            self.TKRadio.configure(
                selectcolor=self.CircleBackgroundColor
            )  # The background of the radio button
        elif text_color or background_color:
            if (
                    self.TextColor not in (None, COLOR_SYSTEM_DEFAULT)
                    and self.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT)
                    and self.TextColor.startswith("#")
                    and self.BackgroundColor.startswith("#")
            ):
                # ---- compute color of circle background ---
                text_hsl = _hex_to_hsl(self.TextColor)
                background_hsl = _hex_to_hsl(
                    self.BackgroundColor
                    if self.BackgroundColor
                    else theme_background_color()
                )
                l_delta = abs(text_hsl[2] - background_hsl[2]) / 10
                if (
                        text_hsl[2] > background_hsl[2]
                ):  # if the text is "lighter" than the background then make background darker
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] - l_delta,
                    )
                else:
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] + l_delta,
                    )
                self.CircleBackgroundColor = rgb(*bg_rbg)
                self.TKRadio.configure(
                    selectcolor=self.CircleBackgroundColor
                )  # The background of the checkbox

        if disabled:
            self.TKRadio["state"] = "disabled"
        elif not disabled:
            self.TKRadio["state"] = "normal"
        self.Disabled = disabled if disabled is not None else self.Disabled

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()
        if visible is not None:
            self._visible = visible

    def reset_group(self):
        """
        Sets all Radio Buttons in the group to not selected
        """
        self.TKIntVar.set(0)

    def get(self):
        # type: (Radio) -> bool
        """
        A snapshot of the value of Radio Button -> (bool)

        :return: True if this radio button is selected
        :rtype:  (bool)
        """
        return self.TKIntVar.get() == self.EncodedRadioValue

    Get = get
    ResetGroup = reset_group
    Update = update


R = Radio
Rad = Radio


# ---------------------------------------------------------------------- #
#                           Checkbox                                     #
# ---------------------------------------------------------------------- #
class Checkbox(Element):
    """
    Checkbox Element - Displays a checkbox and text next to it
    """

    def __init__(
            self,
            text,
            default=False,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            font=None,
            background_color=None,
            text_color=None,
            checkbox_color=None,
            highlight_thickness=1,
            change_submits=False,
            enable_events=False,
            disabled=False,
            key=None,
            k=None,
            pad=None,
            p=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param text:                Text to display next to checkbox
        :type text:                 (str)
        :param default:             Set to True if you want this checkbox initially checked
        :type default:              (bool)
        :param size:                (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                 (int, int)  | (None, None) | int
        :param s:                   Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                    (int, int)  | (None, None) | int
        :param auto_size_text:      if True will size the element to match the length of the text
        :type auto_size_text:       (bool)
        :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                 (str or (str, int[, str]) or None)
        :param background_color:    color of background
        :type background_color:     (str)
        :param text_color:          color of the text
        :type text_color:           (str)
        :param checkbox_color:      color of background of the box that has the check mark in it. The checkmark is the same color as the text
        :type checkbox_color:       (str)
        :param highlight_thickness: thickness of border around checkbox when gets focus
        :type highlight_thickness:  (int)
        :param change_submits:      DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:       (bool)
        :param enable_events:       Turns on the element specific events. Checkbox events happen when an item changes
        :type enable_events:        (bool)
        :param disabled:            set disable state
        :type disabled:             (bool)
        :param key:                 Used with window.find_element and with return values to uniquely identify this element
        :type key:                  str | int | tuple | object
        :param k:                   Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                    str | int | tuple | object
        :param pad:                 Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                  (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                   Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:             text, that will appear when mouse hovers over the element
        :type tooltip:              (str)
        :param right_click_menu:    A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:     List[List[ List[str] | str ]]
        :param expand_x:            If True the element will automatically expand in the X direction to fill available space
        :type expand_x:             (bool)
        :param expand_y:            If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:             (bool)
        :param visible:             set visibility state of the element
        :type visible:              (bool)
        :param metadata:            User metadata that can be set to ANYTHING
        :type metadata:             (Any)
        """

        self.Text = text
        self.InitialState = bool(default)
        self.Value = None
        self.TKCheckbutton = self.Widget = None  # type: tk.Checkbutton
        self.Disabled = disabled
        self.TextColor = text_color if text_color else theme_text_color()
        self.RightClickMenu = right_click_menu
        self.highlight_thickness = highlight_thickness

        # ---- compute color of circle background ---
        if checkbox_color is None:
            try:  # something in here will fail if a color is not specified in Hex
                text_hsl = _hex_to_hsl(self.TextColor)
                background_hsl = _hex_to_hsl(
                    background_color if background_color else theme_background_color()
                )
                l_delta = abs(text_hsl[2] - background_hsl[2]) / 10
                if (
                        text_hsl[2] > background_hsl[2]
                ):  # if the text is "lighter" than the background then make background darker
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] - l_delta,
                    )
                else:
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] + l_delta,
                    )
                self.CheckboxBackgroundColor = rgb(*bg_rbg)
            except Exception:
                self.CheckboxBackgroundColor = (
                    background_color if background_color else theme_background_color()
                )
        else:
            self.CheckboxBackgroundColor = checkbox_color
        self.ChangeSubmits = change_submits or enable_events
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_CHECKBOX,
            size=sz,
            auto_size_text=auto_size_text,
            font=font,
            background_color=background_color,
            text_color=self.TextColor,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )

    def get(self):
        # type: (Checkbox) -> bool
        """
        Return the current state of this checkbox

        :return: Current state of checkbox
        :rtype:  (bool)
        """
        return self.TKIntVar.get() != 0

    def update(
            self,
            value=None,
            text=None,
            background_color=None,
            text_color=None,
            checkbox_color=None,
            disabled=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Checkbox Element. Must call `Window.Read` or `Window.Finalize` prior.
        Note that changing visibility may cause element to change locations when made visible after invisible

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param checkbox_color:
        :param value:            if True checks the checkbox, False clears it
        :type value:             (bool)
        :param text:             Text to display next to checkbox
        :type text:              (str)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text. Note this also changes the color of the checkmark
        :type text_color:        (str)
        :param disabled:         disable or enable element
        :type disabled:          (bool)
        :param visible:          control visibility of element
        :type visible:           (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Checkbox.update - The window was closed"
            )
            return

        if value is not None:
            value = bool(value)
            try:
                self.TKIntVar.set(value)
                self.InitialState = value
            except Exception:
                print("Checkbox update failed")
        if disabled:
            self.TKCheckbutton.configure(state="disabled")
        elif not disabled:
            self.TKCheckbutton.configure(state="normal")
        self.Disabled = disabled if disabled is not None else self.Disabled

        if text is not None:
            self.Text = str(text)
            self.TKCheckbutton.configure(text=self.Text)
        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKCheckbutton.configure(background=background_color)
            self.BackgroundColor = background_color
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKCheckbutton.configure(fg=text_color)
            self.TextColor = text_color
        # Color the checkbox itself
        if checkbox_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.CheckboxBackgroundColor = checkbox_color
            self.TKCheckbutton.configure(
                selectcolor=self.CheckboxBackgroundColor
            )  # The background of the checkbox
        elif text_color or background_color:
            if (
                    self.CheckboxBackgroundColor is not None
                    and self.TextColor is not None
                    and self.BackgroundColor is not None
                    and self.TextColor.startswith("#")
                    and self.BackgroundColor.startswith("#")
            ):
                # ---- compute color of checkbox background ---
                text_hsl = _hex_to_hsl(self.TextColor)
                background_hsl = _hex_to_hsl(
                    self.BackgroundColor
                    if self.BackgroundColor
                    else theme_background_color()
                )
                l_delta = abs(text_hsl[2] - background_hsl[2]) / 10
                if (
                        text_hsl[2] > background_hsl[2]
                ):  # if the text is "lighter" than the background then make background darker
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] - l_delta,
                    )
                else:
                    bg_rbg = _hsl_to_rgb(
                        background_hsl[0],
                        background_hsl[1],
                        background_hsl[2] + l_delta,
                    )
                self.CheckboxBackgroundColor = rgb(*bg_rbg)
                self.TKCheckbutton.configure(
                    selectcolor=self.CheckboxBackgroundColor
                )  # The background of the checkbox

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible

    Get = get
    Update = update


# -------------------------  CHECKBOX Element lazy functions  ------------------------- #
CB = Checkbox
CBox = Checkbox
Check = Checkbox


# ---------------------------------------------------------------------- #
#                           Spin                                         #
# ---------------------------------------------------------------------- #


class Spin(Element):
    """
    A spinner with up/down buttons and a single line of text. Choose 1 values from list
    """

    def __init__(
            self,
            values,
            initial_value=None,
            disabled=False,
            change_submits=False,
            enable_events=False,
            readonly=False,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            bind_return_key=None,
            font=None,
            background_color=None,
            text_color=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            wrap=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param values:           List of valid values
        :type values:            Tuple[Any] or List[Any]
        :param initial_value:    Initial item to show in window. Choose from list of values supplied
        :type initial_value:     (Any)
        :param disabled:         set disable state
        :type disabled:          (bool)
        :param change_submits:   DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:    (bool)
        :param enable_events:    Turns on the element specific events. Spin events happen when an item changes
        :type enable_events:     (bool)
        :param readonly:         If True, then users cannot type in values. Only values from the values list are allowed.
        :type readonly:          (bool)
        :param size:             (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:              (int, int)  | (None, None) | int
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param auto_size_text:   if True will size the element to match the length of the text
        :type auto_size_text:    (bool)
        :param bind_return_key:  If True, then the return key will cause a the element to generate an event when return key is pressed
        :type bind_return_key:   (bool)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param key:              Used with window.find_element and with return values to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param wrap:             Determines if the values should "Wrap". Default is False. If True, when reaching last value, will continue back to the first value.
        :type wrap:              (bool)
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.Values = values
        self.DefaultValue = initial_value
        self.ChangeSubmits = change_submits or enable_events
        self.TKSpinBox = self.Widget = None  # type: tk.Spinbox
        self.Disabled = disabled
        self.Readonly = readonly
        self.RightClickMenu = right_click_menu
        self.BindReturnKey = bind_return_key
        self.wrap = wrap

        bg = background_color if background_color else DEFAULT_INPUT_ELEMENTS_COLOR
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_SPIN,
            size=sz,
            auto_size_text=auto_size_text,
            font=font,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(
            self, value=None, values=None, disabled=None, readonly=None, visible=None
    ):
        """
        Changes some of the settings for the Spin Element. Must call `Window.Read` or `Window.Finalize` prior
        Note that the state can be in 3 states only.... enabled, disabled, readonly even
        though more combinations are available. The easy way to remember is that if you
        change the readonly parameter then you are enabling the element.

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:    set the current value from list of choices
        :type value:     (Any)
        :param values:   set available choices
        :type values:    List[Any]
        :param disabled: disable. Note disabled and readonly cannot be mixed. It must be one OR the other
        :type disabled:  (bool)
        :param readonly: make element readonly.  Note disabled and readonly cannot be mixed. It must be one OR the other
        :type readonly:  (bool)
        :param visible:  control visibility of element
        :type visible:   (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Spin.update - The window was closed")
            return

        if values is not None:
            old_value = self.TKStringVar.get()
            self.Values = values
            self.TKSpinBox.configure(values=values)
            self.TKStringVar.set(old_value)
        if value is not None:
            try:
                self.TKStringVar.set(value)
                self.DefaultValue = value
            except Exception:
                pass

        if readonly:
            self.Readonly = True
            self.TKSpinBox["state"] = "readonly"
        elif not readonly:
            self.Readonly = False
            self.TKSpinBox["state"] = "normal"
        if disabled:
            self.TKSpinBox["state"] = "disable"
        elif not disabled:
            if self.Readonly:
                self.TKSpinBox["state"] = "readonly"
            else:
                self.TKSpinBox["state"] = "normal"
        self.Disabled = disabled if disabled is not None else self.Disabled

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()
        if visible is not None:
            self._visible = visible

    def _SpinChangedHandler(self, event):
        """
        Callback function. Used internally only. Called by tkinter when Spinbox Widget changes.  Results in Window.Read() call returning

        :param event: passed in from tkinter
        :type event:
        """
        # first, get the results table built
        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = ""
        self.ParentForm.FormRemainedOpen = True
        _exit_mainloop(self.ParentForm)
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     Window._window_that_exited = self.ParentForm
        #     self.ParentForm.TKroot.quit()  # kick the users out of the mainloop

    def set_ibeam_color(self, ibeam_color=None):
        """
        Sets the color of the I-Beam that is used to "insert" characters. This is oftens called a "Cursor" by
        many users.  To keep from being confused with tkinter's definition of cursor (the mouse pointer), the term
        ibeam is used in this case.
        :param ibeam_color: color to set the "I-Beam" used to indicate where characters will be inserted
        :type ibeam_color:  (str)
        """

        if not self._widget_was_created():
            return
        if ibeam_color is not None:
            try:
                self.Widget.config(insertbackground=ibeam_color)
            except Exception as e:
                _error_popup_with_traceback(
                    "Error setting I-Beam color in set_ibeam_color",
                    "The element has a key:",
                    self.Key,
                    "The color passed in was:",
                    ibeam_color,
                )

    def get(self):
        """
        Return the current chosen value showing in spinbox.
        This value will be the same as what was provided as list of choices.  If list items are ints, then the
        item returned will be an int (not a string)

        :return: The currently visible entry
        :rtype:  (Any)
        """
        value = self.TKStringVar.get()
        for v in self.Values:
            if str(v) == value:
                value = v
                break
        return value

    Get = get
    Update = update


Sp = Spin  # type: Spin


# ---------------------------------------------------------------------- #
#                           Multiline                                    #
# ---------------------------------------------------------------------- #
class Multiline(Element):
    """
    Multiline Element - Display and/or read multiple lines of text.  This is both an input and output element.
    Other PySimpleGUI ports have a separate MultilineInput and MultilineOutput elements.  May want to split this
    one up in the future too.
    """

    def __init__(
            self,
            default_text="",
            enter_submits=False,
            disabled=False,
            autoscroll=False,
            autoscroll_only_at_bottom=False,
            border_width=None,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            background_color=None,
            text_color=None,
            selected_text_color=None,
            selected_background_color=None,
            horizontal_scroll=False,
            change_submits=False,
            enable_events=False,
            do_not_clear=True,
            key=None,
            k=None,
            write_only=False,
            auto_refresh=False,
            reroute_stdout=False,
            reroute_stderr=False,
            reroute_cprint=False,
            echo_stdout_stderr=False,
            focus=False,
            font=None,
            pad=None,
            p=None,
            tooltip=None,
            justification=None,
            no_scrollbar=False,
            wrap_lines=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
            expand_x=False,
            expand_y=False,
            rstrip=True,
            right_click_menu=None,
            visible=True,
            metadata=None,
    ):
        """
        :param default_text:                 Initial text to show
        :type default_text:                  (Any)
        :param enter_submits:                if True, the Window.read call will return is enter key is pressed in this element
        :type enter_submits:                 (bool)
        :param disabled:                     set disable state
        :type disabled:                      (bool)
        :param autoscroll:                   If True the contents of the element will automatically scroll as more data added to the end
        :type autoscroll:                    (bool)
        :param autoscroll_only_at_bottom:    If True the contents of the element will automatically scroll only if the scrollbar is at the bottom of the multiline
        :type autoscroll_only_at_bottom:     (bool)
        :param border_width:                 width of border around element in pixels
        :type border_width:                  (int)
        :param size:                         (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                          (int, int)  | (None, None) | int
        :param s:                            Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                             (int, int)  | (None, None) | int
        :param auto_size_text:               if True will size the element to match the length of the text
        :type auto_size_text:                (bool)
        :param background_color:             color of background
        :type background_color:              (str)
        :param text_color:                   color of the text
        :type text_color:                    (str)
        :param selected_text_color:          Color of text when it is selected (using mouse or control+A, etc)
        :type selected_text_color:           (str)
        :param selected_background_color:    Color of background when it is selected (using mouse or control+A, etc)
        :type selected_background_color:     (str)
        :param horizontal_scroll:            Controls if a horizontal scrollbar should be shown.  If True a horizontal scrollbar will be shown in addition to vertical
        :type horizontal_scroll:             (bool)
        :param change_submits:               DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:                (bool)
        :param enable_events:                If True then any key press that happens when the element has focus will generate an event.
        :type enable_events:                 (bool)
        :param do_not_clear:                 if False the element will be cleared any time the Window.read call returns
        :type do_not_clear:                  (bool)
        :param key:                          Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                           str | int | tuple | object
        :param k:                            Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                             str | int | tuple | object
        :param write_only:                   If True then no entry will be added to the values dictionary when the window is read
        :type write_only:                    bool
        :param auto_refresh:                 If True then anytime the element is updated, the window will be refreshed so that the change is immediately displayed
        :type auto_refresh:                  (bool)
        :param reroute_stdout:               If True then all output to stdout will be output to this element
        :type reroute_stdout:                (bool)
        :param reroute_stderr:               If True then all output to stderr will be output to this element
        :type reroute_stderr:                (bool)
        :param reroute_cprint:               If True your cprint calls will output to this element. It's the same as you calling cprint_set_output_destination
        :type reroute_cprint:                (bool)
        :param echo_stdout_stderr:           If True then output to stdout and stderr will be output to this element AND also to the normal console location
        :type echo_stdout_stderr:            (bool)
        :param focus:                        if True initial focus will go to this element
        :type focus:                         (bool)
        :param font:                         specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                          (str or (str, int[, str]) or None)
        :param pad:                          Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                           (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                            Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                             (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param tooltip:                      text, that will appear when mouse hovers over the element
        :type tooltip:                       (str)
        :param justification:                text justification. left, right, center. Can use single characters l, r, c.
        :type justification:                 (str)
        :param no_scrollbar:                 If False then a vertical scrollbar will be shown (the default)
        :type no_scrollbar:                  (bool)
        :param wrap_lines:                   If True, the lines will be wrapped automatically. Other parms affect this setting, but this one will override them all. Default is it does nothing and uses previous settings for wrapping.
        :type wrap_lines:                    (bool)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        :param expand_x:                     If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                      (bool)
        :param expand_y:                     If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                      (bool)
        :param rstrip:                       If True the value returned in will have whitespace stripped from the right side
        :type rstrip:                        (bool)
        :param right_click_menu:             A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:              List[List[ List[str] | str ]]
        :param visible:                      set visibility state of the element
        :type visible:                       (bool)
        :param metadata:                     User metadata that can be set to ANYTHING
        :type metadata:                      (Any)
        """

        self.DefaultText = str(default_text)
        self.EnterSubmits = enter_submits
        bg = background_color if background_color else DEFAULT_INPUT_ELEMENTS_COLOR
        self.Focus = focus
        self.do_not_clear = do_not_clear
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        fg = text_color if text_color is not None else DEFAULT_INPUT_TEXT_COLOR
        self.selected_text_color = selected_text_color
        self.selected_background_color = selected_background_color
        self.Autoscroll = autoscroll
        self.Disabled = disabled
        self.ChangeSubmits = change_submits or enable_events
        self.RightClickMenu = right_click_menu
        self.BorderWidth = (
            border_width if border_width is not None else DEFAULT_BORDER_WIDTH
        )
        self.TagCounter = 0
        self.TKText = self.Widget = None  # type: tk.Text
        self.element_frame = None  # type: tk.Frame
        self.HorizontalScroll = horizontal_scroll
        self.tags = set()
        self.WriteOnly = write_only
        self.AutoRefresh = auto_refresh
        key = key if key is not None else k
        self.reroute_cprint = reroute_cprint
        self.echo_stdout_stderr = echo_stdout_stderr
        self.Justification = "left" if justification is None else justification
        self.justification_tag = self.just_center_tag = self.just_left_tag = (
            self.just_right_tag
        ) = None
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y
        self.rstrip = rstrip
        self.wrap_lines = wrap_lines
        self.reroute_stdout = reroute_stdout
        self.reroute_stderr = reroute_stderr
        self.no_scrollbar = no_scrollbar
        self.hscrollbar = None  # The horizontal scrollbar
        self.auto_scroll_only_at_bottom = autoscroll_only_at_bottom
        sz = size if size != (None, None) else s

        super().__init__(
            ELEM_TYPE_INPUT_MULTILINE,
            size=sz,
            auto_size_text=auto_size_text,
            background_color=bg,
            text_color=fg,
            key=key,
            pad=pad,
            tooltip=tooltip,
            font=font or DEFAULT_FONT,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )
        return

    def update(
            self,
            value=None,
            disabled=None,
            append=False,
            font=None,
            text_color=None,
            background_color=None,
            text_color_for_value=None,
            background_color_for_value=None,
            visible=None,
            autoscroll=None,
            justification=None,
            font_for_value=None,
    ):
        """
        Changes some of the settings for the Multiline Element. Must call `Window.read` or set finalize=True when creating window.

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:                      new text to display
        :type value:                       (Any)
        :param disabled:                   disable or enable state of the element
        :type disabled:                    (bool)
        :param append:                     if True then new value will be added onto the end of the current value. if False then contents will be replaced.
        :type append:                      (bool)
        :param font:                       specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the entire element
        :type font:                        (str or (str, int[, str]) or None)
        :param text_color:                 color of the text
        :type text_color:                  (str)
        :param background_color:           color of background
        :type background_color:            (str)
        :param text_color_for_value:       color of the new text being added (the value paramter)
        :type text_color_for_value:        (str)
        :param background_color_for_value: color of the new background of the text being added (the value paramter)
        :type background_color_for_value:  (str)
        :param visible:                    set visibility state of the element
        :type visible:                     (bool)
        :param autoscroll:                 if True then contents of element are scrolled down when new text is added to the end
        :type autoscroll:                  (bool)
        :param justification:              text justification. left, right, center. Can use single characters l, r, c. Sets only for this value, not entire element
        :type justification:               (str)
        :param font_for_value:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the value being updated
        :type font_for_value:              str | (str, int)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            # _error_popup_with_traceback('Error in Multiline.update - The window was closed')
            return

        if autoscroll is not None:
            self.Autoscroll = autoscroll
        current_scroll_position = self.TKText.yview()[1]

        if justification is not None:
            if justification.startswith("l"):
                just_tag = "left"
            if justification.startswith("r"):
                just_tag = "right"
            if justification.startswith("c"):
                just_tag = "center"
        else:
            just_tag = self.justification_tag

        starting_point = self.Widget.index(tk.INSERT)
        tag = None
        if value is not None:
            value = str(value)
            if (
                    background_color_for_value is not None
                    or text_color_for_value is not None
                    or font_for_value is not None
            ):
                try:
                    tag = (
                            "Multiline("
                            + str(text_color_for_value)
                            + ","
                            + str(background_color_for_value)
                            + ","
                            + str(font_for_value)
                            + ")"
                    )
                    if tag not in self.tags:
                        self.tags.add(tag)
                    if background_color_for_value is not None:
                        self.TKText.tag_configure(
                            tag, background=background_color_for_value
                        )
                    if text_color_for_value is not None:
                        self.TKText.tag_configure(tag, foreground=text_color_for_value)
                    if font_for_value is not None:
                        self.TKText.tag_configure(tag, font=font_for_value)
                except Exception as e:
                    print("* Multiline.update - bad color likely specified:", e)
            if self.Disabled:
                self.TKText.configure(state="normal")
            try:
                if not append:
                    self.TKText.delete("1.0", tk.END)
                if tag is not None or just_tag is not None:
                    self.TKText.insert(tk.END, value, (just_tag, tag))
                else:
                    self.TKText.insert(tk.END, value)

                # self.TKText.tag_add(just_tag, starting_point, starting_point)

            except Exception as e:
                print("* Error setting multiline *", e)
            if self.Disabled:
                self.TKText.configure(state="disabled")
            self.DefaultText = value

        # if self.Autoscroll:
        #     self.TKText.see(tk.END)
        if self.Autoscroll:
            if not self.auto_scroll_only_at_bottom or (
                    self.auto_scroll_only_at_bottom and current_scroll_position == 1.0
            ):
                self.TKText.see(tk.END)
        if disabled:
            self.TKText.configure(state="disabled")
        elif not disabled:
            self.TKText.configure(state="normal")
        self.Disabled = disabled if disabled is not None else self.Disabled

        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(background=background_color)
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(fg=text_color)
        if font is not None:
            self.TKText.configure(font=font)

        if not visible:
            self._pack_forget_save_settings(alternate_widget=self.element_frame)
            # self.element_frame.pack_forget()
        elif visible:
            self._pack_restore_settings(alternate_widget=self.element_frame)
            # self.element_frame.pack(padx=self.pad_used[0], pady=self.pad_used[1])

        if self.AutoRefresh and self.ParentForm:
            try:  # in case the window was destroyed
                self.ParentForm.refresh()
            except Exception:
                pass
        if visible is not None:
            self._visible = visible

    def get(self):
        """
        Return current contents of the Multiline Element

        :return: current contents of the Multiline Element (used as an input type of Multiline
        :rtype:  (str)
        """
        value = str(self.TKText.get(1.0, tk.END))
        if self.rstrip:
            return value.rstrip()
        return value

    def print(
            self,
            *args,
            end=None,
            sep=None,
            text_color=None,
            background_color=None,
            justification=None,
            font=None,
            colors=None,
            t=None,
            b=None,
            c=None,
            autoscroll=True,
    ):
        """
        Print like Python normally prints except route the output to a multiline element and also add colors if desired

        colors -(str, str) or str.  A combined text/background color definition in a single parameter

        There are also "aliases" for text_color, background_color and colors (t, b, c)
        t - An alias for color of the text (makes for shorter calls)
        b - An alias for the background_color parameter
        c - (str, str) - "shorthand" way of specifying color. (foreground, backgrouned)
        c - str - can also be a string of the format "foreground on background"  ("white on red")

        With the aliases it's possible to write the same print but in more compact ways:
        cprint('This will print white text on red background', c=('white', 'red'))
        cprint('This will print white text on red background', c='white on red')
        cprint('This will print white text on red background', text_color='white', background_color='red')
        cprint('This will print white text on red background', t='white', b='red')

        :param args:             The arguments to print
        :type args:              (Any)
        :param end:              The end char to use just like print uses
        :type end:               (str)
        :param sep:              The separation character like print uses
        :type sep:               (str)
        :param text_color:       The color of the text
        :type text_color:        (str)
        :param background_color: The background color of the line
        :type background_color:  (str)
        :param justification:    text justification. left, right, center. Can use single characters l, r, c. Sets only for this value, not entire element
        :type justification:     (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the args being printed
        :type font:              (str or (str, int[, str]) or None)
        :param colors:           Either a tuple or a string that has both the text and background colors. Or just the text color
        :type colors:            (str) or (str, str)
        :param t:                Color of the text
        :type t:                 (str)
        :param b:                The background color of the line
        :type b:                 (str)
        :param c:                Either a tuple or a string that has both the text and background colors or just tex color (same as the color parm)
        :type c:                 (str) or (str, str)
        :param autoscroll:       If True the contents of the element will automatically scroll as more data added to the end
        :type autoscroll:        (bool)
        """

        kw_text_color = text_color or t
        kw_background_color = background_color or b
        dual_color = colors or c
        try:
            if isinstance(dual_color, tuple):
                kw_text_color = dual_color[0]
                kw_background_color = dual_color[1]
            elif isinstance(dual_color, str):
                if (
                        " on " in dual_color
                ):  # if has "on" in the string, then have both text and background
                    kw_text_color = dual_color.split(" on ")[0]
                    kw_background_color = dual_color.split(" on ")[1]
                else:  # if no "on" then assume the color string is just the text color
                    kw_text_color = dual_color
        except Exception as e:
            print("* multiline print warning * you messed up with color formatting", e)

        _print_to_element(
            self,
            *args,
            end=end,
            sep=sep,
            text_color=kw_text_color,
            background_color=kw_background_color,
            justification=justification,
            autoscroll=autoscroll,
            font=font,
        )

    def reroute_stdout_to_here(self):
        """
        Sends stdout (prints) to this element
        """
        # if nothing on the stack, then need to save the very first stdout
        if len(Window._rerouted_stdout_stack) == 0:
            Window._original_stdout = sys.stdout
        Window._rerouted_stdout_stack.insert(0, (self.ParentForm, self))
        sys.stdout = self

    def reroute_stderr_to_here(self):
        """
        Sends stderr to this element
        """
        if len(Window._rerouted_stderr_stack) == 0:
            Window._original_stderr = sys.stderr
        Window._rerouted_stderr_stack.insert(0, (self.ParentForm, self))
        sys.stderr = self

    def restore_stdout(self):
        """
        Restore a previously re-reouted stdout back to the original destination
        """
        Window._restore_stdout()

    def restore_stderr(self):
        """
        Restore a previously re-reouted stderr back to the original destination
        """
        Window._restore_stderr()

    def write(self, txt):
        """
        Called by Python (not tkinter?) when stdout or stderr wants to write

        :param txt: text of output
        :type txt:  (str)
        """
        try:
            self.update(txt, append=True)
            # if need to echo, then send the same text to the destinatoin that isn't thesame as this one
            if self.echo_stdout_stderr:
                if sys.stdout != self:
                    sys.stdout.write(txt)
                elif sys.stderr != self:
                    sys.stderr.write(txt)
        except Exception:
            pass

    def flush(self):
        """
        Flush parameter was passed into a print statement.
        For now doing nothing.  Not sure what action should be taken to ensure a flush happens regardless.
        """
        # try:
        #     self.previous_stdout.flush()
        # except Exception:
        #     pass
        return

    def set_ibeam_color(self, ibeam_color=None):
        """
        Sets the color of the I-Beam that is used to "insert" characters. This is oftens called a "Cursor" by
        many users.  To keep from being confused with tkinter's definition of cursor (the mouse pointer), the term
        ibeam is used in this case.
        :param ibeam_color: color to set the "I-Beam" used to indicate where characters will be inserted
        :type ibeam_color:  (str)
        """

        if not self._widget_was_created():
            return
        if ibeam_color is not None:
            try:
                self.Widget.config(insertbackground=ibeam_color)
            except Exception as e:
                _error_popup_with_traceback(
                    "Error setting I-Beam color in set_ibeam_color",
                    "The element has a key:",
                    self.Key,
                    "The color passed in was:",
                    ibeam_color,
                )

    def __del__(self):
        """
        AT ONE TIME --- If this Widget is deleted, be sure and restore the old stdout, stderr
        Now the restore is done differently. Do not want to RELY on Python to call this method
        in order for stdout and stderr to be restored.  Instead explicit restores are called.

        """

        return

    Get = get
    Update = update


ML = Multiline
MLine = Multiline


# ---------------------------------------------------------------------- #
#                                       Text                             #
# ---------------------------------------------------------------------- #
class Text(Element):
    """
    Text - Display some text in the window.  Usually this means a single line of text.  However, the text can also be multiple lines.  If multi-lined there are no scroll bars.
    """

    def __init__(
            self,
            text="",
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            click_submits=False,
            enable_events=False,
            relief=None,
            font=None,
            text_color=None,
            background_color=None,
            border_width=None,
            justification=None,
            pad=None,
            p=None,
            key=None,
            k=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            grab=None,
            tooltip=None,
            visible=True,
            metadata=None,
    ):
        """
        :param text:             The text to display. Can include /n to achieve multiple lines.  Will convert (optional) parameter into a string
        :type text:              Any
        :param size:             (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:              (int, int) |  (int, None) | (None, None) | (int, ) | int
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int) |  (int, None) | (None, None) | (int, ) | int
        :param auto_size_text:   if True size of the Text Element will be sized to fit the string provided in 'text' parm
        :type auto_size_text:    (bool)
        :param click_submits:    DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type click_submits:     (bool)
        :param enable_events:    Turns on the element specific events. Text events happen when the text is clicked
        :type enable_events:     (bool)
        :param relief:           relief style around the text. Values are same as progress meter relief values. Should be a constant that is defined at starting with RELIEF - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type relief:            (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param background_color: color of background
        :type background_color:  (str)
        :param border_width:     number of pixels for the border (if using a relief)
        :type border_width:      (int)
        :param justification:    how string should be aligned within space provided by size. Valid choices = `left`, `right`, `center`
        :type justification:     (str)
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:              Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:               str or int or tuple or object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param grab:             If True can grab this element and move the window around. Default is False
        :type grab:              (bool)
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.DisplayText = str(text)
        self.TextColor = text_color if text_color else DEFAULT_TEXT_COLOR
        self.Justification = justification
        self.Relief = relief
        self.ClickSubmits = click_submits or enable_events
        if background_color is None:
            bg = DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR
        else:
            bg = background_color
        self.RightClickMenu = right_click_menu
        self.TKRightClickMenu = None
        self.BorderWidth = border_width
        self.Grab = grab
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_TEXT,
            auto_size_text=auto_size_text,
            size=sz,
            background_color=bg,
            font=font if font else DEFAULT_FONT,
            text_color=self.TextColor,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )

    def update(
            self,
            value=None,
            background_color=None,
            text_color=None,
            font=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Text Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:            new text to show
        :type value:             (Any)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Text.update - The window was closed")
            return

        if value is not None:
            self.DisplayText = str(value)
            self.TKStringVar.set(str(value))
        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(background=background_color)
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(fg=text_color)
        if font is not None:
            self.TKText.configure(font=font)
        if not visible:
            self._pack_forget_save_settings()
            # self.TKText.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKText.pack(padx=self.pad_used[0], pady=self.pad_used[1])
        if visible is not None:
            self._visible = visible

    def get(self):
        """
        Gets the current value of the displayed text

        :return: The current value
        :rtype:  (str)
        """
        try:
            text = self.TKStringVar.get()
        except Exception:
            text = ""
        return text

    @classmethod
    def fonts_installed_list(cls):
        """
        Returns a list of strings that tkinter reports as the installed fonts

        :return:          List of the installed font names
        :rtype:           List[str]
        """
        # A window must exist before can perform this operation. Create the hidden master root if it doesn't exist
        _get_hidden_master_root()

        fonts = list(tkinter.font.families())
        fonts.sort()

        return fonts

    @classmethod
    def char_width_in_pixels(cls, font, character="W"):
        """
        Get the with of the character "W" in pixels for the font being passed in or
        the character of your choosing if "W" is not a good representative character.
        Cannot be used until a window has been created.
        If an error occurs, 0 will be returned
        :param font:      specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike, to be measured
        :type font:       (str or (str, int[, str]) or None)
        :param character: specifies a SINGLE CHARACTER character to measure
        :type character:  (str)
        :return:          Width in pixels of "A"
        :rtype:           (int)
        """
        # A window must exist before can perform this operation. Create the hidden master root if it doesn't exist
        _get_hidden_master_root()

        size = 0
        try:
            size = tkinter.font.Font(font=font).measure(
                character
            )  # single character width
        except Exception as e:
            _error_popup_with_traceback("Exception retrieving char width in pixels", e)

        return size

    @classmethod
    def char_height_in_pixels(cls, font):
        """
        Get the height of a string if using the supplied font in pixels.
        Cannot be used until a window has been created.
        If an error occurs, 0 will be returned
        :param font: specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike, to be measured
        :type font:  (str or (str, int[, str]) or None)
        :return:     Height in pixels of "A"
        :rtype:      (int)
        """

        # A window must exist before can perform this operation. Create the hidden master root if it doesn't exist
        _get_hidden_master_root()

        size = 0
        try:
            size = tkinter.font.Font(font=font).metrics("linespace")
        except Exception as e:
            _error_popup_with_traceback("Exception retrieving char height in pixels", e)

        return size

    @classmethod
    def string_width_in_pixels(cls, font, string):
        """
        Get the with of the supplied string in pixels for the font being passed in.
        If an error occurs, 0 will be returned
        :param font:   specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike, to be measured
        :type font:    (str or (str, int[, str]) or None)
        :param string: the string to measure
        :type string:  str
        :return:       Width in pixels of string
        :rtype:        (int)
        """

        # A window must exist before can perform this operation. Create the hidden master root if it doesn't exist
        _get_hidden_master_root()

        size = 0
        try:
            size = tkinter.font.Font(font=font).measure(string)  # string's  width
        except Exception as e:
            _error_popup_with_traceback(
                "Exception retrieving string width in pixels", e
            )

        return size

    def _print_to_element(
            self,
            *args,
            end=None,
            sep=None,
            text_color=None,
            background_color=None,
            autoscroll=None,
            justification=None,
            font=None,
            append=None,
    ):
        """
        Print like Python normally prints except route the output to a multiline element and also add colors if desired

        :param multiline_element: The multiline element to be output to
        :type multiline_element:  (Multiline)
        :param args:              The arguments to print
        :type args:               List[Any]
        :param end:               The end char to use just like print uses
        :type end:                (str)
        :param sep:               The separation character like print uses
        :type sep:                (str)
        :param text_color:        color of the text
        :type text_color:         (str)
        :param background_color:  The background color of the line
        :type background_color:   (str)
        :param autoscroll:        If True (the default), the element will scroll to bottom after updating
        :type autoscroll:         (bool)
        :param font:              specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the value being updated
        :type font:               str | (str, int)
        """
        end_str = str(end) if end is not None else "\n"
        sep_str = str(sep) if sep is not None else " "

        outstring = ""
        num_args = len(args)
        for i, arg in enumerate(args):
            outstring += str(arg)
            if i != num_args - 1:
                outstring += sep_str
        outstring += end_str
        if append:
            outstring = self.get() + outstring

        self.update(
            outstring,
            text_color=text_color,
            background_color=background_color,
            font=font,
        )

        try:  # if the element is set to autorefresh, then refresh the parent window
            if self.AutoRefresh:
                self.ParentForm.refresh()
        except Exception:
            pass

    def print(
            self,
            *args,
            end=None,
            sep=None,
            text_color=None,
            background_color=None,
            justification=None,
            font=None,
            colors=None,
            t=None,
            b=None,
            c=None,
            autoscroll=True,
            append=True,
    ):
        """
        Print like Python normally prints except route the output to a multiline element and also add colors if desired

        colors -(str, str) or str.  A combined text/background color definition in a single parameter

        There are also "aliases" for text_color, background_color and colors (t, b, c)
        t - An alias for color of the text (makes for shorter calls)
        b - An alias for the background_color parameter
        c - (str, str) - "shorthand" way of specifying color. (foreground, backgrouned)
        c - str - can also be a string of the format "foreground on background"  ("white on red")

        With the aliases it's possible to write the same print but in more compact ways:
        cprint('This will print white text on red background', c=('white', 'red'))
        cprint('This will print white text on red background', c='white on red')
        cprint('This will print white text on red background', text_color='white', background_color='red')
        cprint('This will print white text on red background', t='white', b='red')

        :param append:
        :param args:             The arguments to print
        :type args:              (Any)
        :param end:              The end char to use just like print uses
        :type end:               (str)
        :param sep:              The separation character like print uses
        :type sep:               (str)
        :param text_color:       The color of the text
        :type text_color:        (str)
        :param background_color: The background color of the line
        :type background_color:  (str)
        :param justification:    text justification. left, right, center. Can use single characters l, r, c. Sets only for this value, not entire element
        :type justification:     (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the args being printed
        :type font:              (str or (str, int[, str]) or None)
        :param colors:           Either a tuple or a string that has both the text and background colors. Or just the text color
        :type colors:            (str) or (str, str)
        :param t:                Color of the text
        :type t:                 (str)
        :param b:                The background color of the line
        :type b:                 (str)
        :param c:                Either a tuple or a string that has both the text and background colors or just tex color (same as the color parm)
        :type c:                 (str) or (str, str)
        :param autoscroll:       If True the contents of the element will automatically scroll as more data added to the end
        :type autoscroll:        (bool)
        """

        kw_text_color = text_color or t
        kw_background_color = background_color or b
        dual_color = colors or c
        try:
            if isinstance(dual_color, tuple):
                kw_text_color = dual_color[0]
                kw_background_color = dual_color[1]
            elif isinstance(dual_color, str):
                if (
                        " on " in dual_color
                ):  # if has "on" in the string, then have both text and background
                    kw_text_color = dual_color.split(" on ")[0]
                    kw_background_color = dual_color.split(" on ")[1]
                else:  # if no "on" then assume the color string is just the text color
                    kw_text_color = dual_color
        except Exception as e:
            print("* multiline print warning * you messed up with color formatting", e)

        self._print_to_element(
            *args,
            end=end,
            sep=sep,
            text_color=kw_text_color,
            background_color=kw_background_color,
            justification=justification,
            autoscroll=autoscroll,
            font=font,
            append=append,
        )

    Get = get
    Update = update


# -------------------------  Text Element lazy functions  ------------------------- #

Txt = Text  # type: Text
T = Text  # type: Text


# ---------------------------------------------------------------------- #
#                                       StatusBar                        #
# ---------------------------------------------------------------------- #
class StatusBar(Element):
    """
    A StatusBar Element creates the sunken text-filled strip at the bottom. Many Windows programs have this line
    """

    def __init__(
            self,
            text,
            size=(None, None),
            s=(None, None),
            auto_size_text=None,
            click_submits=None,
            enable_events=False,
            relief=RELIEF_SUNKEN,
            font=None,
            text_color=None,
            background_color=None,
            justification=None,
            pad=None,
            p=None,
            key=None,
            k=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            tooltip=None,
            visible=True,
            metadata=None,
    ):
        """
        :param text:             Text that is to be displayed in the widget
        :type text:              (str)
        :param size:             (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:              (int, int) |  (int, None) | int
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param auto_size_text:   True if size should fit the text length
        :type auto_size_text:    (bool)
        :param click_submits:    DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type click_submits:     (bool)
        :param enable_events:    Turns on the element specific events. StatusBar events occur when the bar is clicked
        :type enable_events:     (bool)
        :param relief:           relief style. Values are same as progress meter relief values.  Can be a constant or a string: `RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID`
        :type relief:            (enum)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param background_color: color of background
        :type background_color:  (str)
        :param justification:    how string should be aligned within space provided by size. Valid choices = `left`, `right`, `center`
        :type justification:     (str)
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:              Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.DisplayText = text
        self.TextColor = text_color if text_color else DEFAULT_TEXT_COLOR
        self.Justification = justification
        self.Relief = relief
        self.ClickSubmits = click_submits or enable_events
        if background_color is None:
            bg = DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR
        else:
            bg = background_color
        self.TKText = self.Widget = None  # type: tk.Label
        key = key if key is not None else k
        self.RightClickMenu = right_click_menu
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_STATUSBAR,
            size=sz,
            auto_size_text=auto_size_text,
            background_color=bg,
            font=font or DEFAULT_FONT,
            text_color=self.TextColor,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(
            self,
            value=None,
            background_color=None,
            text_color=None,
            font=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Status Bar Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:            new text to show
        :type value:             (str)
        :param background_color: color of background
        :type background_color:  (str)
        :param text_color:       color of the text
        :type text_color:        (str)
        :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:              (str or (str, int[, str]) or None)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in StatusBar.update - The window was closed"
            )
            return

        if value is not None:
            self.DisplayText = value
            stringvar = self.TKStringVar
            stringvar.set(value)
        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(background=background_color)
        if text_color not in (None, COLOR_SYSTEM_DEFAULT):
            self.TKText.configure(fg=text_color)
        if font is not None:
            self.TKText.configure(font=font)
        if not visible:
            self._pack_forget_save_settings()
            # self.TKText.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKText.pack(padx=self.pad_used[0], pady=self.pad_used[1])
        if visible is not None:
            self._visible = visible

    Update = update


SBar = StatusBar


# ---------------------------------------------------------------------- #
#                       TKProgressBar                                    #
#  Emulate the TK ProgressBar using canvas and rectangles
# ---------------------------------------------------------------------- #


class TKProgressBar:
    uniqueness_counter = 0

    def __init__(
            self,
            root,
            max,
            length=400,
            width=DEFAULT_PROGRESS_BAR_SIZE[1],
            ttk_theme=DEFAULT_TTK_THEME,
            style_name="",
            relief=DEFAULT_PROGRESS_BAR_RELIEF,
            border_width=DEFAULT_PROGRESS_BAR_BORDER_WIDTH,
            orientation="horizontal",
            BarColor=(None, None),
            key=None,
    ):
        """
        :param root:         The root window bar is to be shown in
        :type root:          tk.Tk | tk.TopLevel
        :param max:          Maximum value the bar will be measuring
        :type max:           (int)
        :param length:       length in pixels of the bar
        :type length:        (int)
        :param width:        width in pixels of the bar
        :type width:         (int)
        :param style_name:   Progress bar style to use.  Set in the packer function
        :type style_name:    (str)
        :param ttk_theme:    Progress bar style defined as one of these 'default', 'winnative', 'clam', 'alt', 'classic', 'vista', 'xpnative'
        :type ttk_theme:     (str)
        :param relief:       relief style. Values are same as progress meter relief values.  Can be a constant or a string: `RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID` (Default value = DEFAULT_PROGRESS_BAR_RELIEF)
        :type relief:        (str)
        :param border_width: The amount of pixels that go around the outside of the bar
        :type border_width:  (int)
        :param orientation:  'horizontal' or 'vertical' ('h' or 'v' work) (Default value = 'vertical')
        :type orientation:   (str)
        :param BarColor:     The 2 colors that make up a progress bar. One is the background, the other is the bar
        :type BarColor:      (str, str)
        :param key:          Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:           str | int | tuple | object
        """

        self.Length = length
        self.Width = width
        self.Max = max
        self.Orientation = orientation
        self.Count = None
        self.PriorCount = 0
        self.style_name = style_name

        TKProgressBar.uniqueness_counter += 1

        if orientation.lower().startswith("h"):
            s = ttk.Style()
            _change_ttk_theme(s, ttk_theme)

            # self.style_name = str(key) + str(TKProgressBar.uniqueness_counter) + "my.Horizontal.TProgressbar"
            if BarColor != COLOR_SYSTEM_DEFAULT and BarColor[0] != COLOR_SYSTEM_DEFAULT:
                s.configure(
                    self.style_name,
                    background=BarColor[0],
                    troughcolor=BarColor[1],
                    troughrelief=relief,
                    borderwidth=border_width,
                    thickness=width,
                )
            else:
                s.configure(
                    self.style_name,
                    troughrelief=relief,
                    borderwidth=border_width,
                    thickness=width,
                )

            self.TKProgressBarForReal = ttk.Progressbar(
                root,
                maximum=self.Max,
                style=self.style_name,
                length=length,
                orient=tk.HORIZONTAL,
                mode="determinate",
            )
        else:
            s = ttk.Style()
            _change_ttk_theme(s, ttk_theme)
            # self.style_name = str(key) + str(TKProgressBar.uniqueness_counter) + "my.Vertical.TProgressbar"
            if BarColor != COLOR_SYSTEM_DEFAULT and BarColor[0] != COLOR_SYSTEM_DEFAULT:
                s.configure(
                    self.style_name,
                    background=BarColor[0],
                    troughcolor=BarColor[1],
                    troughrelief=relief,
                    borderwidth=border_width,
                    thickness=width,
                )
            else:
                s.configure(
                    self.style_name,
                    troughrelief=relief,
                    borderwidth=border_width,
                    thickness=width,
                )

            self.TKProgressBarForReal = ttk.Progressbar(
                root,
                maximum=self.Max,
                style=self.style_name,
                length=length,
                orient=tk.VERTICAL,
                mode="determinate",
            )

    def Update(self, count=None, max=None):
        """
        Update the current value of the bar and/or update the maximum value the bar can reach
        :param count: current value
        :type count:  (int)
        :param max:   the maximum value
        :type max:    (int)
        """
        if max is not None:
            self.Max = max
            try:
                self.TKProgressBarForReal.config(maximum=max)
            except Exception:
                return False
        if count is not None:
            try:
                self.TKProgressBarForReal["value"] = count
            except Exception:
                return False
        return True


# ---------------------------------------------------------------------- #
#                           Output                                       #
#  Routes stdout, stderr to a scrolled window                            #
# ---------------------------------------------------------------------- #
class Output(Multiline):
    """
    Output Element - a multi-lined text area to where stdout, stderr, cprint are rerouted.

    The Output Element is now based on the Multiline Element.  When you make an Output Element, you're
    creating a Multiline Element with some specific settings set:
        auto_refresh = True
        auto_scroll = True
        reroute_stdout = True
        reroute_stderr = True
        reroute_cprint = True
        write_only = True

    If you choose to use a Multiline element to replace an Output element, be sure an turn on the write_only paramter in the Multiline
    so that an item is not included in the values dictionary on every window.read call
    """

    def __init__(
            self,
            size=(None, None),
            s=(None, None),
            background_color=None,
            text_color=None,
            pad=None,
            p=None,
            autoscroll_only_at_bottom=False,
            echo_stdout_stderr=False,
            font=None,
            tooltip=None,
            key=None,
            k=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
            wrap_lines=None,
            horizontal_scroll=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
    ):
        """
        :param size:                        (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                         (int, int)  | (None, None) | int
        :param s:                           Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                            (int, int)  | (None, None) | int
        :param background_color:            color of background
        :type background_color:             (str)
        :param text_color:                  color of the text
        :type text_color:                   (str)
        :param pad:                         Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                           Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                            (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param autoscroll_only_at_bottom:   If True the contents of the element will automatically scroll only if the scrollbar is at the bottom of the multiline
        :type autoscroll_only_at_bottom:    (bool)
        :param echo_stdout_stderr:          If True then output to stdout will be output to this element AND also to the normal console location
        :type echo_stdout_stderr:           (bool)
        :param font:                        specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                         (str or (str, int[, str]) or None)
        :param tooltip:                     text, that will appear when mouse hovers over the element
        :type tooltip:                      (str)
        :param key:                         Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                          str | int | tuple | object
        :param k:                           Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                            str | int | tuple | object
        :param right_click_menu:            A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:             List[List[ List[str] | str ]]
        :param expand_x:                    If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                     (bool)
        :param expand_y:                    If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                     (bool)
        :param visible:                     set visibility state of the element
        :type visible:                      (bool)
        :param metadata:                    User metadata that can be set to ANYTHING
        :type metadata:                     (Any)
        :param wrap_lines:                  If True, the lines will be wrapped automatically. Other parms affect this setting, but this one will override them all. Default is it does nothing and uses previous settings for wrapping.
        :type wrap_lines:                   (bool)
        :param horizontal_scroll:           Controls if a horizontal scrollbar should be shown. If True, then line wrapping will be off by default
        :type horizontal_scroll:            (bool)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        """

        super().__init__(
            size=size,
            s=s,
            background_color=background_color,
            autoscroll_only_at_bottom=autoscroll_only_at_bottom,
            text_color=text_color,
            pad=pad,
            p=p,
            echo_stdout_stderr=echo_stdout_stderr,
            font=font,
            tooltip=tooltip,
            wrap_lines=wrap_lines,
            horizontal_scroll=horizontal_scroll,
            key=key,
            k=k,
            right_click_menu=right_click_menu,
            write_only=True,
            reroute_stdout=True,
            reroute_stderr=True,
            reroute_cprint=True,
            autoscroll=True,
            auto_refresh=True,
            expand_x=expand_x,
            expand_y=expand_y,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )


# ---------------------------------------------------------------------- #
#                           Button Class                                 #
# ---------------------------------------------------------------------- #
class Button(Element):
    """
    Button Element - Defines all possible buttons. The shortcuts such as Submit, FileBrowse, ... each create a Button
    """

    def __init__(
            self,
            button_text="",
            button_type=BUTTON_TYPE_READ_FORM,
            target=(None, None),
            tooltip=None,
            file_types=FILE_TYPES_ALL_FILES,
            initial_folder=None,
            default_extension="",
            disabled=False,
            change_submits=False,
            enable_events=False,
            image_filename=None,
            image_data=None,
            image_size=(None, None),
            image_subsample=None,
            image_zoom=None,
            image_source=None,
            border_width=None,
            size=(None, None),
            s=(None, None),
            auto_size_button=None,
            button_color=None,
            disabled_button_color=None,
            highlight_colors=None,
            mouseover_colors=(None, None),
            use_ttk_buttons=None,
            font=None,
            bind_return_key=False,
            focus=False,
            pad=None,
            p=None,
            key=None,
            k=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param button_text:           Text to be displayed on the button
        :type button_text:            (str)
        :param button_type:           You  should NOT be setting this directly. ONLY the shortcut functions set this
        :type button_type:            (int)
        :param target:                key or (row,col) target for the button. Note that -1 for column means 1 element to the left of this one. The constant ThisRow is used to indicate the current row. The Button itself is a valid target for some types of button
        :type target:                 str | (int, int)
        :param tooltip:               text, that will appear when mouse hovers over the element
        :type tooltip:                (str)
        :param file_types:            the filetypes that will be used to match files. To indicate all files: (("ALL Files", "*.* *"),).
        :type file_types:             Tuple[(str, str), ...]
        :param initial_folder:        starting path for folders and files
        :type initial_folder:         (str)
        :param default_extension:     If no extension entered by user, add this to filename (only used in saveas dialogs)
        :type default_extension:      (str)
        :param disabled:              If True button will be created disabled. If BUTTON_DISABLED_MEANS_IGNORE then the button will be ignored rather than disabled using tkinter
        :type disabled:               (bool | str)
        :param change_submits:        DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:         (bool)
        :param enable_events:         Turns on the element specific events. If this button is a target, should it generate an event when filled in
        :type enable_events:          (bool)
        :param image_source:          Image to place on button. Use INSTEAD of the image_filename and image_data. Unifies these into 1 easier to use parm
        :type image_source:           (str | bytes)
        :param image_filename:        image filename if there is a button image. GIFs and PNGs only.
        :type image_filename:         (str)
        :param image_data:            Raw or Base64 representation of the image to put on button. Choose either filename or data
        :type image_data:             bytes | str
        :param image_size:            Size of the image in pixels (width, height)
        :type image_size:             (int, int)
        :param image_subsample:       amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type image_subsample:        (int)
        :param image_zoom:            amount to increase the size of the image. 2=twice size, 3=3 times, etc
        :type image_zoom:             (int)
        :param border_width:          width of border around button in pixels
        :type border_width:           (int)
        :param size:                  (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                   (int | None, int | None)  | (None, None) | int
        :param s:                     Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                      (int | None, int | None)  | (None, None) | int
        :param auto_size_button:      if True the button size is sized to fit the text
        :type auto_size_button:       (bool)
        :param button_color:          Color of button. default is from theme or the window. Easy to remember which is which if you say "ON" between colors. "red" on "green". Normally a tuple, but can be a simplified-button-color-string "foreground on background". Can be a single color if want to set only the background.
        :type button_color:           (str, str) | str
        :param disabled_button_color: colors to use when button is disabled (text, background). Use None for a color if don't want to change. Only ttk buttons support both text and background colors. tk buttons only support changing text color
        :type disabled_button_color:  (str, str) | str
        :param highlight_colors:      colors to use when button has focus (has focus, does not have focus). None will use colors based on theme. Only used by Linux and only for non-TTK button
        :type highlight_colors:       (str, str)
        :param mouseover_colors:      Important difference between Linux & Windows! Linux - Colors when mouse moved over button.  Windows - colors when button is pressed. The default is to switch the text and background colors (an inverse effect)
        :type mouseover_colors:       (str, str) | str
        :param use_ttk_buttons:       True = use ttk buttons. False = do not use ttk buttons.  None (Default) = use ttk buttons only if on a Mac and not with button images
        :type use_ttk_buttons:        (bool)
        :param font:                  specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                   (str or (str, int[, str]) or None)
        :param bind_return_key:       If True then pressing the return key in an Input or Multiline Element will cause this button to appear to be clicked (generates event with this button's key
        :type bind_return_key:        (bool)
        :param focus:                 if True, initial focus will be put on this button
        :type focus:                  (bool)
        :param pad:                   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:                   Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                    str | int | tuple | object
        :param k:                     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                      str | int | tuple | object
        :param right_click_menu:      A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:       List[List[ List[str] | str ]]
        :param expand_x:              If True the element will automatically expand in the X direction to fill available space
        :type expand_x:               (bool)
        :param expand_y:              If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:               (bool)
        :param visible:               set visibility state of the element
        :type visible:                (bool)
        :param metadata:              User metadata that can be set to ANYTHING
        :type metadata:               (Any)
        """

        self.AutoSizeButton = auto_size_button
        self.BType = button_type
        if (
                file_types is not None
                and len(file_types) == 2
                and isinstance(file_types[0], str)
                and isinstance(file_types[1], str)
        ):
            warnings.warn(
                "file_types parameter not correctly specified. This parameter is a LIST of TUPLES. You have passed (str,str) rather than ((str, str),). Fixing it for you this time.\nchanging {} to {}\nPlease correct your code".format(
                    file_types, ((file_types[0], file_types[1]),)
                ),
                UserWarning,
            )
            file_types = ((file_types[0], file_types[1]),)
        self.FileTypes = file_types
        self.Widget = self.TKButton = None  # type: tk.Button
        self.Target = target
        self.ButtonText = str(button_text)
        self.RightClickMenu = right_click_menu
        # Button colors can be a tuple (text, background) or a string with format "text on background"
        # bc = button_color
        # if button_color is None:
        #     bc = DEFAULT_BUTTON_COLOR
        # else:
        #     try:
        #         if isinstance(button_color,str):
        #             bc = button_color.split(' on ')
        #     except Exception as e:
        #         print('* cprint warning * you messed up with color formatting', e)
        #     if bc[1] is None:
        #         bc = (bc[0], theme_button_color()[1])
        # self.ButtonColor = bc
        self.ButtonColor = button_color_to_tuple(button_color)

        # experimental code to compute disabled button text color
        # if disabled_button_color is None:
        #     try:
        #         disabled_button_color = (get_complimentary_hex(theme_button_color()[0]), theme_button_color()[1])
        #         # disabled_button_color = disabled_button_color
        #     except Exception:
        #         print('* Problem computing disabled button color *')
        self.DisabledButtonColor = (
            button_color_to_tuple(disabled_button_color)
            if disabled_button_color is not None
            else (None, None)
        )
        if image_source is not None:
            if isinstance(image_source, bytes):
                image_data = image_source
            elif isinstance(image_source, str):
                image_filename = image_source
        self.ImageFilename = image_filename
        self.ImageData = image_data
        self.ImageSize = image_size
        self.ImageSubsample = image_subsample
        self.zoom = int(image_zoom) if image_zoom is not None else None
        self.UserData = None
        self.BorderWidth = (
            border_width if border_width is not None else DEFAULT_BORDER_WIDTH
        )
        self.BindReturnKey = bind_return_key
        self.Focus = focus
        self.TKCal = None
        self.calendar_default_date_M_D_Y = (None, None, None)
        self.calendar_close_when_chosen = False
        self.calendar_locale = None
        self.calendar_format = None
        self.calendar_location = (None, None)
        self.calendar_no_titlebar = True
        self.calendar_begin_at_sunday_plus = 0
        self.calendar_month_names = None
        self.calendar_day_abbreviations = None
        self.calendar_title = ""
        self.calendar_selection = ""
        self.default_button = None
        self.InitialFolder = initial_folder
        self.DefaultExtension = default_extension
        self.Disabled = disabled
        self.ChangeSubmits = change_submits or enable_events
        self.UseTtkButtons = use_ttk_buttons
        self._files_delimiter = BROWSE_FILES_DELIMITER  # used by the file browse button. used when multiple files are selected by user
        if use_ttk_buttons is None and running_mac():
            self.UseTtkButtons = True
        # if image_filename or image_data:
        #     self.UseTtkButtons = False              # if an image is to be displayed, then force the button to not be a TTK Button
        if key is None and k is None:
            _key = self.ButtonText
            if DEFAULT_USE_BUTTON_SHORTCUTS:
                pos = _key.find(MENU_SHORTCUT_CHARACTER)
                if pos != -1:
                    if (
                            pos < len(MENU_SHORTCUT_CHARACTER)
                            or _key[pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                    ):
                        _key = _key[:pos] + _key[pos + len(MENU_SHORTCUT_CHARACTER):]
                    else:
                        _key = _key.replace(
                            "\\" + MENU_SHORTCUT_CHARACTER, MENU_SHORTCUT_CHARACTER
                        )
        else:
            _key = key if key is not None else k
        if highlight_colors is not None:
            self.HighlightColors = highlight_colors
        else:
            self.HighlightColors = self._compute_highlight_colors()

        if mouseover_colors != (None, None):
            self.MouseOverColors = button_color_to_tuple(mouseover_colors)
        elif button_color is not None:
            self.MouseOverColors = (self.ButtonColor[1], self.ButtonColor[0])
        else:
            self.MouseOverColors = (theme_button_color()[1], theme_button_color()[0])
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        sz = size if size != (None, None) else s
        super().__init__(
            ELEM_TYPE_BUTTON,
            size=sz,
            font=font,
            pad=pad,
            key=_key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def _compute_highlight_colors(self):
        """
        Determines the color to use to indicate the button has focus. This setting is only used by Linux.
        :return: Pair of colors. (Highlight, Highlight Background)
        :rtype:  (str, str)
        """
        highlight_color = highlight_background = COLOR_SYSTEM_DEFAULT
        if (
                self.ButtonColor != COLOR_SYSTEM_DEFAULT
                and theme_background_color() != COLOR_SYSTEM_DEFAULT
        ):
            highlight_background = theme_background_color()
        if (
                self.ButtonColor != COLOR_SYSTEM_DEFAULT
                and self.ButtonColor[0] != COLOR_SYSTEM_DEFAULT
        ):
            if self.ButtonColor[0] != theme_background_color():
                highlight_color = self.ButtonColor[0]
            else:
                highlight_color = "red"
        return highlight_color, highlight_background

        # Realtime button release callback

    def ButtonReleaseCallBack(self, parm):
        """
        Not a user callable function.  Called by tkinter when a "realtime" button is released

        :param parm: the event info from tkinter
        :type parm:

        """
        self.LastButtonClickedWasRealtime = False
        self.ParentForm.LastButtonClicked = None

    # Realtime button callback
    def ButtonPressCallBack(self, parm):
        """
        Not a user callable method. Callback called by tkinter when a "realtime" button is pressed

        :param parm: Event info passed in by tkinter
        :type parm:

        """
        self.ParentForm.LastButtonClickedWasRealtime = True
        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = self.ButtonText
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     Window._window_that_exited = self.ParentForm
        #     self.ParentForm.TKroot.quit()  # kick out of loop if read was called
        _exit_mainloop(self.ParentForm)

    def _find_target(self):
        target = self.Target
        target_element = None

        if target[0] == ThisRow:
            target = [self.Position[0], target[1]]
            if target[1] < 0:
                target[1] = self.Position[1] + target[1]
        strvar = None
        should_submit_window = False
        if target == (None, None):
            strvar = self.TKStringVar
        else:
            # Need a try-block because if the target is not hashable, the "in" test will raise exception
            try:
                if target in self.ParentForm.AllKeysDict:
                    target_element = self.ParentForm.AllKeysDict[target]
            except Exception:
                pass
            # if target not found or the above try got exception, then keep looking....
            if target_element is None:
                if not isinstance(target, str):
                    if target[0] < 0:
                        target = [self.Position[0] + target[0], target[1]]
                    target_element = self.ParentContainer._GetElementAtLocation(target)
                else:
                    target_element = self.ParentForm.find_element(target)
            try:
                strvar = target_element.TKStringVar
            except Exception:
                pass
            try:
                if target_element.ChangeSubmits:
                    should_submit_window = True
            except Exception:
                pass
        return target_element, strvar, should_submit_window

    # -------  Button Callback  ------- #
    def ButtonCallBack(self):
        """
        Not user callable! Called by tkinter when a button is clicked.  This is where all the fun begins!
        """

        if self.Disabled == BUTTON_DISABLED_MEANS_IGNORE:
            return
        target_element, strvar, should_submit_window = self._find_target()

        filetypes = FILE_TYPES_ALL_FILES if self.FileTypes is None else self.FileTypes

        if self.BType == BUTTON_TYPE_BROWSE_FOLDER:
            if running_mac():  # macs don't like seeing the parent window (go firgure)
                folder_name = tk.filedialog.askdirectory(
                    initialdir=self.InitialFolder
                )  # show the 'get folder' dialog box
            else:
                folder_name = tk.filedialog.askdirectory(
                    initialdir=self.InitialFolder, parent=self.ParentForm.TKroot
                )  # show the 'get folder' dialog box
            if folder_name:
                try:
                    strvar.set(folder_name)
                    self.TKStringVar.set(folder_name)
                except Exception:
                    pass
            else:  # if "cancel" button clicked, don't generate an event
                should_submit_window = False
        elif self.BType == BUTTON_TYPE_BROWSE_FILE:
            if running_mac():
                # Workaround for the "*.*" issue on Mac
                is_all = [
                    (x, y) for (x, y) in filetypes if all(ch in "* ." for ch in y)
                ]
                if not len(set(filetypes)) > 1 and (
                        len(is_all) != 0 or filetypes == FILE_TYPES_ALL_FILES
                ):
                    file_name = tk.filedialog.askopenfilename(
                        initialdir=self.InitialFolder
                    )
                else:
                    file_name = tk.filedialog.askopenfilename(
                        initialdir=self.InitialFolder, filetypes=filetypes
                    )  # show the 'get file' dialog box
                # elif _mac_allow_filetypes():
                # file_name = tk.filedialog.askopenfilename(initialdir=self.InitialFolder, filetypes=filetypes)  # show the 'get file' dialog box
                # else:
                #     file_name = tk.filedialog.askopenfilename(initialdir=self.InitialFolder)  # show the 'get file' dialog box
            else:
                file_name = tk.filedialog.askopenfilename(
                    filetypes=filetypes,
                    initialdir=self.InitialFolder,
                    parent=self.ParentForm.TKroot,
                )  # show the 'get file' dialog box

            if file_name:
                strvar.set(file_name)
                self.TKStringVar.set(file_name)
            else:  # if "cancel" button clicked, don't generate an event
                should_submit_window = False
        elif self.BType == BUTTON_TYPE_COLOR_CHOOSER:
            color = tk.colorchooser.askcolor(
                parent=self.ParentForm.TKroot, color=self.default_color
            )  # show the 'get file' dialog box
            color = color[1]  # save only the #RRGGBB portion
            if color is not None:
                strvar.set(color)
                self.TKStringVar.set(color)
        elif self.BType == BUTTON_TYPE_BROWSE_FILES:
            if running_mac():
                # Workaround for the "*.*" issue on Mac
                is_all = [
                    (x, y) for (x, y) in filetypes if all(ch in "* ." for ch in y)
                ]
                if not len(set(filetypes)) > 1 and (
                        len(is_all) != 0 or filetypes == FILE_TYPES_ALL_FILES
                ):
                    file_name = tk.filedialog.askopenfilenames(
                        initialdir=self.InitialFolder
                    )
                else:
                    file_name = tk.filedialog.askopenfilenames(
                        filetypes=filetypes, initialdir=self.InitialFolder
                    )
                # elif _mac_allow_filetypes():
                #     file_name = tk.filedialog.askopenfilenames(filetypes=filetypes, initialdir=self.InitialFolder)
                # else:
                #     file_name = tk.filedialog.askopenfilenames(initialdir=self.InitialFolder)
            else:
                file_name = tk.filedialog.askopenfilenames(
                    filetypes=filetypes,
                    initialdir=self.InitialFolder,
                    parent=self.ParentForm.TKroot,
                )

            if file_name:
                file_name = self._files_delimiter.join(file_name)  # normally a ';'
                strvar.set(file_name)
                self.TKStringVar.set(file_name)
            else:  # if "cancel" button clicked, don't generate an event
                should_submit_window = False
        elif self.BType == BUTTON_TYPE_SAVEAS_FILE:
            # show the 'get file' dialog box
            if running_mac():
                # Workaround for the "*.*" issue on Mac
                is_all = [
                    (x, y) for (x, y) in filetypes if all(ch in "* ." for ch in y)
                ]
                if not len(set(filetypes)) > 1 and (
                        len(is_all) != 0 or filetypes == FILE_TYPES_ALL_FILES
                ):
                    file_name = tk.filedialog.asksaveasfilename(
                        defaultextension=self.DefaultExtension,
                        initialdir=self.InitialFolder,
                    )
                else:
                    file_name = tk.filedialog.asksaveasfilename(
                        filetypes=filetypes,
                        defaultextension=self.DefaultExtension,
                        initialdir=self.InitialFolder,
                    )
                # elif _mac_allow_filetypes():
                #     file_name = tk.filedialog.asksaveasfilename(filetypes=filetypes, defaultextension=self.DefaultExtension, initialdir=self.InitialFolder)
                # else:
                #     file_name = tk.filedialog.asksaveasfilename(defaultextension=self.DefaultExtension, initialdir=self.InitialFolder)
            else:
                file_name = tk.filedialog.asksaveasfilename(
                    filetypes=filetypes,
                    defaultextension=self.DefaultExtension,
                    initialdir=self.InitialFolder,
                    parent=self.ParentForm.TKroot,
                )

            if file_name:
                strvar.set(file_name)
                self.TKStringVar.set(file_name)
            else:  # if "cancel" button clicked, don't generate an event
                should_submit_window = False
        elif (
                self.BType == BUTTON_TYPE_CLOSES_WIN
        ):  # this is a return type button so GET RESULTS and destroy window
            # first, get the results table built
            # modify the Results table in the parent FlexForm object
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = self.Key
            else:
                self.ParentForm.LastButtonClicked = self.ButtonText
            self.ParentForm.FormRemainedOpen = False
            self.ParentForm._Close()
            _exit_mainloop(self.ParentForm)

            if self.ParentForm.NonBlocking:
                self.ParentForm.TKroot.destroy()
                Window._DecrementOpenCount()
        elif (
                self.BType == BUTTON_TYPE_READ_FORM
        ):  # LEAVE THE WINDOW OPEN!! DO NOT CLOSE
            # This is a PLAIN BUTTON
            # first, get the results table built
            # modify the Results table in the parent FlexForm object
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = self.Key
            else:
                self.ParentForm.LastButtonClicked = self.ButtonText
            self.ParentForm.FormRemainedOpen = True
            _exit_mainloop(self.ParentForm)
        elif (
                self.BType == BUTTON_TYPE_CLOSES_WIN_ONLY
        ):  # special kind of button that does not exit main loop
            self.ParentForm._Close(without_event=True)
            self.ParentForm.TKroot.destroy()  # close the window with tkinter
            Window._DecrementOpenCount()
        elif (
                self.BType == BUTTON_TYPE_CALENDAR_CHOOSER
        ):  # this is a return type button so GET RESULTS and destroy window
            # ------------ new chooser code -------------
            self.ParentForm.LastButtonClicked = (
                self.Key
            )  # key should have been generated already if not set by user
            self.ParentForm.FormRemainedOpen = True
            should_submit_window = False
            _exit_mainloop(self.ParentForm)
        # elif self.BType == BUTTON_TYPE_SHOW_DEBUGGER:
        # **** DEPRICATED *****
        # if self.ParentForm.DebuggerEnabled:
        # show_debugger_popout_window()

        if should_submit_window:
            self.ParentForm.LastButtonClicked = target_element.Key
            self.ParentForm.FormRemainedOpen = True
            _exit_mainloop(self.ParentForm)

        return

    def update(
            self,
            text=None,
            button_color=(None, None),
            disabled=None,
            image_source=None,
            image_data=None,
            image_filename=None,
            visible=None,
            image_subsample=None,
            image_zoom=None,
            disabled_button_color=(None, None),
            image_size=None,
    ):
        """
        Changes some of the settings for the Button Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param text:                  sets button text
        :type text:                   (str)
        :param button_color:          Color of button. default is from theme or the window. Easy to remember which is which if you say "ON" between colors. "red" on "green". Normally a tuple, but can be a simplified-button-color-string "foreground on background". Can be a single color if want to set only the background.
        :type button_color:           (str, str) | str
        :param disabled:              True/False to enable/disable at the GUI level. Use BUTTON_DISABLED_MEANS_IGNORE to ignore clicks (won't change colors)
        :type disabled:               (bool | str)
        :param image_source:          Image to place on button. Use INSTEAD of the image_filename and image_data. Unifies these into 1 easier to use parm
        :type image_source:           (str | bytes)
        :param image_data:            Raw or Base64 representation of the image to put on button. Choose either filename or data
        :type image_data:             bytes | str
        :param image_filename:        image filename if there is a button image. GIFs and PNGs only.
        :type image_filename:         (str)
        :param disabled_button_color: colors to use when button is disabled (text, background). Use None for a color if don't want to change. Only ttk buttons support both text and background colors. tk buttons only support changing text color
        :type disabled_button_color:  (str, str)
        :param visible:               control visibility of element
        :type visible:                (bool)
        :param image_subsample:       amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type image_subsample:        (int)
        :param image_zoom:            amount to increase the size of the image. 2=twice size, 3=3 times, etc
        :type image_zoom:             (int)
        :param image_size:            Size of the image in pixels (width, height)
        :type image_size:             (int, int)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Button.update - The window was closed"
            )
            return

        if image_source is not None:
            if isinstance(image_source, bytes):
                image_data = image_source
            elif isinstance(image_source, str):
                image_filename = image_source

        if self.UseTtkButtons:
            style_name = (
                self.ttk_style_name
            )  # created when made initial window (in the pack)
            # style_name = str(self.Key) + 'custombutton.TButton'
            button_style = ttk.Style()
        if text is not None:
            btext = text
            if DEFAULT_USE_BUTTON_SHORTCUTS:
                pos = btext.find(MENU_SHORTCUT_CHARACTER)
                if pos != -1:
                    if (
                            pos < len(MENU_SHORTCUT_CHARACTER)
                            or btext[pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                    ):
                        btext = (
                                btext[:pos] + btext[pos + len(MENU_SHORTCUT_CHARACTER):]
                        )
                    else:
                        btext = btext.replace(
                            "\\" + MENU_SHORTCUT_CHARACTER, MENU_SHORTCUT_CHARACTER
                        )
                        pos = -1
                if pos != -1:
                    self.TKButton.config(underline=pos)
            self.TKButton.configure(text=btext)
            self.ButtonText = text
        if button_color != (None, None) and button_color != COLOR_SYSTEM_DEFAULT:
            bc = button_color_to_tuple(button_color, self.ButtonColor)
            # if isinstance(button_color, str):
            #     try:
            #         button_color = button_color.split(' on ')
            #     except Exception as e:
            #         print('** Error in formatting your button color **', button_color, e)
            if self.UseTtkButtons:
                if bc[0] not in (None, COLOR_SYSTEM_DEFAULT):
                    button_style.configure(style_name, foreground=bc[0])
                if bc[1] not in (None, COLOR_SYSTEM_DEFAULT):
                    button_style.configure(style_name, background=bc[1])
            else:
                if bc[0] not in (None, COLOR_SYSTEM_DEFAULT):
                    self.TKButton.config(foreground=bc[0], activebackground=bc[0])
                if bc[1] not in (None, COLOR_SYSTEM_DEFAULT):
                    self.TKButton.config(background=bc[1], activeforeground=bc[1])
            self.ButtonColor = bc
        if disabled is True:
            self.TKButton["state"] = "disabled"
        elif disabled is False:
            self.TKButton["state"] = "normal"
        elif disabled == BUTTON_DISABLED_MEANS_IGNORE:
            self.TKButton["state"] = "normal"
        self.Disabled = disabled if disabled is not None else self.Disabled

        if image_data is not None:
            image = tk.PhotoImage(data=image_data)
            if image_subsample:
                image = image.subsample(image_subsample)
            if image_zoom is not None:
                image = image.zoom(int(image_zoom))
            if image_size is not None:
                width, height = image_size
            else:
                width, height = image.width(), image.height()
            if self.UseTtkButtons:
                button_style.configure(
                    style_name, image=image, width=width, height=height
                )
            else:
                self.TKButton.config(image=image, width=width, height=height)
            self.TKButton.image = image
        if image_filename is not None:
            image = tk.PhotoImage(file=image_filename)
            if image_subsample:
                image = image.subsample(image_subsample)
            if image_zoom is not None:
                image = image.zoom(int(image_zoom))
            if image_size is not None:
                width, height = image_size
            else:
                width, height = image.width(), image.height()
            if self.UseTtkButtons:
                button_style.configure(
                    style_name, image=image, width=width, height=height
                )
            else:
                self.TKButton.config(
                    highlightthickness=0, image=image, width=width, height=height
                )
            self.TKButton.image = image
        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()
        if (
                disabled_button_color != (None, None)
                and disabled_button_color != COLOR_SYSTEM_DEFAULT
        ):
            if not self.UseTtkButtons:
                self.TKButton["disabledforeground"] = disabled_button_color[0]
            else:
                if disabled_button_color[0] is not None:
                    button_style.map(
                        style_name, foreground=[("disabled", disabled_button_color[0])]
                    )
                if disabled_button_color[1] is not None:
                    button_style.map(
                        style_name, background=[("disabled", disabled_button_color[1])]
                    )
            self.DisabledButtonColor = (
                disabled_button_color[0]
                if disabled_button_color[0] is not None
                else self.DisabledButtonColor[0],
                disabled_button_color[1]
                if disabled_button_color[1] is not None
                else self.DisabledButtonColor[1],
            )

        if visible is not None:
            self._visible = visible

    def get_text(self):
        """
        Returns the current text shown on a button

        :return: The text currently displayed on the button
        :rtype:  (str)
        """
        return self.ButtonText

    def click(self):
        """
        Generates a click of the button as if the user clicked the button
        Calls the tkinter invoke method for the button
        """
        try:
            self.TKButton.invoke()
        except Exception:
            print("Exception clicking button")

    Click = click
    GetText = get_text
    Update = update


# -------------------------  Button lazy functions  ------------------------- #
B = Button
Btn = Button


# ---------------------------------------------------------------------- #
#                           ButtonMenu Class                             #
# ---------------------------------------------------------------------- #
class ButtonMenu(Element):
    """
    The Button Menu Element.  Creates a button that when clicked will show a menu similar to right click menu
    """

    def __init__(
            self,
            button_text,
            menu_def,
            tooltip=None,
            disabled=False,
            image_source=None,
            image_filename=None,
            image_data=None,
            image_size=(None, None),
            image_subsample=None,
            image_zoom=None,
            border_width=None,
            size=(None, None),
            s=(None, None),
            auto_size_button=None,
            button_color=None,
            text_color=None,
            background_color=None,
            disabled_text_color=None,
            font=None,
            item_font=None,
            pad=None,
            p=None,
            expand_x=False,
            expand_y=False,
            key=None,
            k=None,
            tearoff=False,
            visible=True,
            metadata=None,
    ):
        """
        :param button_text:               Text to be displayed on the button
        :type button_text:                (str)
        :param menu_def:                  A list of lists of Menu items to show when this element is clicked. See docs for format as they are the same for all menu types
        :type menu_def:                   List[List[str]]
        :param tooltip:                   text, that will appear when mouse hovers over the element
        :type tooltip:                    (str)
        :param disabled:                  If True button will be created disabled
        :type disabled:                   (bool)
        :param image_source:              Image to place on button. Use INSTEAD of the image_filename and image_data. Unifies these into 1 easier to use parm
        :type image_source:               (str | bytes)
        :param image_filename:            image filename if there is a button image. GIFs and PNGs only.
        :type image_filename:             (str)
        :param image_data:                Raw or Base64 representation of the image to put on button. Choose either filename or data
        :type image_data:                 bytes | str
        :param image_size:                Size of the image in pixels (width, height)
        :type image_size:                 (int, int)
        :param image_subsample:           amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type image_subsample:            (int)
        :param image_zoom:                amount to increase the size of the image. 2=twice size, 3=3 times, etc
        :type image_zoom:                 (int)
        :param border_width:              width of border around button in pixels
        :type border_width:               (int)
        :param size:                      (w, h) w=characters-wide, h=rows-high. If an int instead of a tuple is supplied, then height is auto-set to 1
        :type size:                       (int, int)  | (None, None) | int
        :param s:                         Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                          (int, int)  | (None, None) | int
        :param auto_size_button:          if True the button size is sized to fit the text
        :type auto_size_button:           (bool)
        :param button_color:              of button. Easy to remember which is which if you say "ON" between colors. "red" on "green"
        :type button_color:               (str, str) | str
        :param background_color:          color of the background
        :type background_color:           (str)
        :param text_color:                element's text color. Can be in #RRGGBB format or a color name "black"
        :type text_color:                 (str)
        :param disabled_text_color:       color to use for text when item is disabled. Can be in #RRGGBB format or a color name "black"
        :type disabled_text_color:        (str)
        :param font:                      specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                       (str or (str, int[, str]) or None)
        :param item_font:                 specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike, for the menu items
        :type item_font:                  (str or (str, int[, str]) or None)
        :param pad:                       Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                         Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param expand_x:                  If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                   (bool)
        :param expand_y:                  If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                   (bool)
        :param key:                       Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                        str | int | tuple | object
        :param k:                         Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                          str | int | tuple | object
        :param tearoff:                   Determines if menus should allow them to be torn off
        :type tearoff:                    (bool)
        :param visible:                   set visibility state of the element
        :type visible:                    (bool)
        :param metadata:                  User metadata that can be set to ANYTHING
        :type metadata:                   (Any)
        """

        self.MenuDefinition = copy.deepcopy(menu_def)

        self.AutoSizeButton = auto_size_button
        self.ButtonText = button_text
        self.ButtonColor = button_color_to_tuple(button_color)
        # self.TextColor = self.ButtonColor[0]
        # self.BackgroundColor = self.ButtonColor[1]
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else theme_input_background_color()
        )
        self.TextColor = (
            text_color if text_color is not None else theme_input_text_color()
        )
        self.DisabledTextColor = (
            disabled_text_color
            if disabled_text_color is not None
            else COLOR_SYSTEM_DEFAULT
        )
        self.ItemFont = item_font
        self.BorderWidth = (
            border_width if border_width is not None else DEFAULT_BORDER_WIDTH
        )
        if image_source is not None:
            if isinstance(image_source, str):
                image_filename = image_source
            elif isinstance(image_source, bytes):
                image_data = image_source
            else:
                warnings.warn(
                    "ButtonMenu element - image_source is not a valid type: {}".format(
                        type(image_source)
                    ),
                    UserWarning,
                )

        self.ImageFilename = image_filename
        self.ImageData = image_data
        self.ImageSize = image_size
        self.ImageSubsample = image_subsample
        self.zoom = int(image_zoom) if image_zoom is not None else None
        self.Disabled = disabled
        self.IsButtonMenu = True
        self.MenuItemChosen = None
        self.Widget = self.TKButtonMenu = None  # type: tk.Menubutton
        self.TKMenu = None  # type: tk.Menu
        self.part_of_custom_menubar = False
        self.custom_menubar_key = None
        # self.temp_size = size if size != (NONE, NONE) else
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_BUTTONMENU,
            size=sz,
            font=font,
            pad=pad,
            key=key,
            tooltip=tooltip,
            text_color=self.TextColor,
            background_color=self.BackgroundColor,
            visible=visible,
            metadata=metadata,
        )
        self.Tearoff = tearoff

    def _MenuItemChosenCallback(
            self, item_chosen
    ):  # ButtonMenu Menu Item Chosen Callback
        """
        Not a user callable function.  Called by tkinter when an item is chosen from the menu.

        :param item_chosen: The menu item chosen.
        :type item_chosen:  (str)
        """
        # print('IN MENU ITEM CALLBACK', item_chosen)
        self.MenuItemChosen = item_chosen
        self.ParentForm.LastButtonClicked = self.Key
        self.ParentForm.FormRemainedOpen = True
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     self.ParentForm.TKroot.quit()  # kick the users out of the mainloop
        _exit_mainloop(self.ParentForm)

    def update(
            self,
            menu_definition=None,
            visible=None,
            image_source=None,
            image_size=(None, None),
            image_subsample=None,
            image_zoom=None,
            button_text=None,
            button_color=None,
    ):
        """
        Changes some of the settings for the ButtonMenu Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param menu_definition: (New menu definition (in menu definition format)
        :type menu_definition:  List[List]
        :param visible:         control visibility of element
        :type visible:          (bool)
        :param image_source:    new image if image is to be changed. Can be a filename or a base64 encoded byte-string
        :type image_source:     (str | bytes)
        :param image_size:      Size of the image in pixels (width, height)
        :type image_size:       (int, int)
        :param image_subsample: amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type image_subsample:  (int)
        :param image_zoom:      amount to increase the size of the image. 2=twice size, 3=3 times, etc
        :type image_zoom:       (int)
        :param button_text:     Text to be shown on the button
        :type button_text:      (str)
        :param button_color:    Normally a tuple, but can be a simplified-button-color-string "foreground on background". Can be a single color if want to set only the background.
        :type button_color:     (str, str) | str
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in ButtonMenu.update - The window was closed"
            )
            return

        if menu_definition is not None:
            self.MenuDefinition = copy.deepcopy(menu_definition)
            top_menu = self.TKMenu = tk.Menu(
                self.TKButtonMenu,
                tearoff=self.Tearoff,
                font=self.ItemFont,
                tearoffcommand=self._tearoff_menu_callback,
            )

            if self.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                top_menu.config(bg=self.BackgroundColor)
            if self.TextColor not in (COLOR_SYSTEM_DEFAULT, None):
                top_menu.config(fg=self.TextColor)
            if self.DisabledTextColor not in (COLOR_SYSTEM_DEFAULT, None):
                top_menu.config(disabledforeground=self.DisabledTextColor)
            if self.ItemFont is not None:
                top_menu.config(font=self.ItemFont)
            AddMenuItem(self.TKMenu, self.MenuDefinition[1], self)
            self.TKButtonMenu.configure(menu=self.TKMenu)
        if image_source is not None:
            filename = data = None
            if image_source is not None:
                if isinstance(image_source, bytes):
                    data = image_source
                elif isinstance(image_source, str):
                    filename = image_source
                else:
                    warnings.warn(
                        "ButtonMenu element - image_source is not a valid type: {}".format(
                            type(image_source)
                        ),
                        UserWarning,
                    )
            image = None
            if filename is not None:
                image = tk.PhotoImage(file=filename)
                if image_subsample is not None:
                    image = image.subsample(image_subsample)
                if image_zoom is not None:
                    image = image.zoom(int(image_zoom))
            elif data is not None:
                # if type(data) is bytes:
                try:
                    image = tk.PhotoImage(data=data)
                    if image_subsample is not None:
                        image = image.subsample(image_subsample)
                    if image_zoom is not None:
                        image = image.zoom(int(image_zoom))
                except Exception as e:
                    image = data

            if image is not None:
                if type(image) is not bytes:
                    width, height = (
                        image_size[0] if image_size[0] is not None else image.width(),
                        image_size[1] if image_size[1] is not None else image.height(),
                    )
                else:
                    width, height = image_size

                self.TKButtonMenu.config(
                    image=image, compound=tk.CENTER, width=width, height=height
                )
                self.TKButtonMenu.image = image
        if button_text is not None:
            self.TKButtonMenu.configure(text=button_text)
            self.ButtonText = button_text
        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()
        if visible is not None:
            self._visible = visible
        if button_color != (None, None) and button_color != COLOR_SYSTEM_DEFAULT:
            bc = button_color_to_tuple(button_color, self.ButtonColor)
            if bc[0] not in (None, COLOR_SYSTEM_DEFAULT):
                self.TKButtonMenu.config(foreground=bc[0], activeforeground=bc[0])
            if bc[1] not in (None, COLOR_SYSTEM_DEFAULT):
                self.TKButtonMenu.config(background=bc[1], activebackground=bc[1])
            self.ButtonColor = bc

    def click(self):
        """
        Generates a click of the button as if the user clicked the button
        Calls the tkinter invoke method for the button
        """
        try:
            self.TKMenu.invoke(1)
        except Exception:
            print("Exception clicking button")

    Update = update
    Click = click


BMenu = ButtonMenu
BM = ButtonMenu


# ---------------------------------------------------------------------- #
#                           ProgreessBar                                 #
# ---------------------------------------------------------------------- #
class ProgressBar(Element):
    """
    Progress Bar Element - Displays a colored bar that is shaded as progress of some operation is made
    """

    def __init__(
            self,
            max_value,
            orientation=None,
            size=(None, None),
            s=(None, None),
            size_px=(None, None),
            auto_size_text=None,
            bar_color=None,
            style=None,
            border_width=None,
            relief=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param max_value:        max value of progressbar
        :type max_value:         (int)
        :param orientation:      'horizontal' or 'vertical'
        :type orientation:       (str)
        :param size:             Size of the bar.  If horizontal (chars long, pixels wide), vert (chars high, pixels wide). Vert height measured using horizontal chars units.
        :type size:              (int, int) |  (int, None)
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None)
        :param size_px:          Size in pixels (length, width). Will be used in place of size parm if specified
        :type size_px:           (int, int) | (None, None)
        :param auto_size_text:   Not sure why this is here
        :type auto_size_text:    (bool)
        :param bar_color:        The 2 colors that make up a progress bar. Either a tuple of 2 strings or a string. Tuple - (bar, background). A string with 1 color changes the background of the bar only. A string with 2 colors separated by "on" like "red on blue" specifies a red bar on a blue background.
        :type bar_color:         (str, str) or str
        :param style:            Progress bar style defined as one of these 'default', 'winnative', 'clam', 'alt', 'classic', 'vista', 'xpnative'
        :type style:             (str)
        :param border_width:     The amount of pixels that go around the outside of the bar
        :type border_width:      (int)
        :param relief:           relief style. Values are same as progress meter relief values.  Can be a constant or a string: `RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID` (Default value = DEFAULT_PROGRESS_BAR_RELIEF)
        :type relief:            (str)
        :param key:              Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.MaxValue = max_value
        self.TKProgressBar = None  # type: TKProgressBar
        self.Cancelled = False
        self.NotRunning = True
        self.Orientation = orientation if orientation else DEFAULT_METER_ORIENTATION
        self.RightClickMenu = right_click_menu
        # Progress Bar colors can be a tuple (text, background) or a string with format "bar on background" - examples "red on white" or ("red", "white")
        if bar_color is None:
            bar_color = DEFAULT_PROGRESS_BAR_COLOR
        else:
            bar_color = _simplified_dual_color_to_tuple(
                bar_color, default=DEFAULT_PROGRESS_BAR_COLOR
            )

        self.BarColor = bar_color  # should be a tuple at this point
        self.BarStyle = style if style else DEFAULT_TTK_THEME
        self.BorderWidth = (
            border_width if border_width else DEFAULT_PROGRESS_BAR_BORDER_WIDTH
        )
        self.Relief = relief if relief else DEFAULT_PROGRESS_BAR_RELIEF
        self.BarExpired = False
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y
        self.size_px = size_px

        super().__init__(
            ELEM_TYPE_PROGRESS_BAR,
            size=sz,
            auto_size_text=auto_size_text,
            key=key,
            pad=pad,
            visible=visible,
            metadata=metadata,
        )

    # returns False if update failed
    def update_bar(self, current_count, max=None):
        """
        DEPRECATED BUT STILL USABLE - has been combined with the normal ProgressBar.update method.
        Change what the bar shows by changing the current count and optionally the max count

        :param current_count: sets the current value
        :type current_count:  (int)
        :param max:           changes the max value
        :type max:            (int)
        """

        if self.ParentForm.TKrootDestroyed:
            return False
        self.TKProgressBar.Update(current_count, max=max)
        try:
            self.ParentForm.TKroot.update()
        except Exception:
            Window._DecrementOpenCount()
            # _my_windows.Decrement()
            return False
        return True

    def update(self, current_count=None, max=None, bar_color=None, visible=None):
        """
        Changes some of the settings for the ProgressBar Element. Must call `Window.Read` or `Window.Finalize` prior
        Now has the ability to modify the count so that the update_bar method is not longer needed separately

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param current_count: sets the current value
        :type current_count:  (int)
        :param max:           changes the max value
        :type max:            (int)
        :param bar_color:     The 2 colors that make up a progress bar. Easy to remember which is which if you say "ON" between colors. "red" on "green".
        :type bar_color:      (str, str) or str
        :param visible:       control visibility of element
        :type visible:        (bool)
        :return:              Returns True if update was OK.  False means something wrong with window or it was closed
        :rtype:               (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return False

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in ProgressBar.update - The window was closed"
            )
            return

        if self.ParentForm.TKrootDestroyed:
            return False

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible
        if bar_color is not None:
            bar_color = _simplified_dual_color_to_tuple(
                bar_color, default=DEFAULT_PROGRESS_BAR_COLOR
            )
            self.BarColor = bar_color
            style = ttk.Style()
            style.configure(
                self.ttk_style_name, background=bar_color[0], troughcolor=bar_color[1]
            )
        if current_count is not None:
            self.TKProgressBar.Update(current_count, max=max)

        try:
            self.ParentForm.TKroot.update()
        except Exception:
            # Window._DecrementOpenCount()
            # _my_windows.Decrement()
            return False
        return True

    Update = update
    UpdateBar = update_bar


PBar = ProgressBar
Prog = ProgressBar
Progress = ProgressBar


# ---------------------------------------------------------------------- #
#                           Image                                        #
# ---------------------------------------------------------------------- #
class Image(Element):
    """
    Image Element - show an image in the window. Should be a GIF or a PNG only
    """

    def __init__(
            self,
            source=None,
            filename=None,
            data=None,
            background_color=None,
            size=(None, None),
            s=(None, None),
            pad=None,
            p=None,
            key=None,
            k=None,
            tooltip=None,
            subsample=None,
            zoom=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            enable_events=False,
            metadata=None,
    ):
        """
        :param source:           A filename or a base64 bytes. Will automatically detect the type and fill in filename or data for you.
        :type source:            str | bytes | None
        :param filename:         image filename if there is a button image. GIFs and PNGs only.
        :type filename:          str | None
        :param data:             Raw or Base64 representation of the image to put on button. Choose either filename or data
        :type data:              bytes | str | None
        :param background_color: color of background
        :type background_color:
        :param size:             (width, height) size of image in pixels
        :type size:              (int, int)
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:              Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param subsample:        amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type subsample:         (int)
        :param zoom:             amount to increase the size of the image.
        :type zoom:              (int)
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param enable_events:    Turns on the element specific events. For an Image element, the event is "image clicked"
        :type enable_events:     (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        if source is not None:
            if isinstance(source, bytes):
                data = source
            elif isinstance(source, str):
                filename = source
            else:
                warnings.warn(
                    "Image element - source is not a valid type: {}".format(
                        type(source)
                    ),
                    UserWarning,
                )

        self.Filename = filename
        self.Data = data
        self.Widget = self.tktext_label = None  # type: tk.Label
        self.BackgroundColor = background_color
        if data is None and filename is None:
            self.Filename = ""
        self.EnableEvents = enable_events
        self.RightClickMenu = right_click_menu
        self.AnimatedFrames = None
        self.CurrentFrameNumber = 0
        self.TotalAnimatedFrames = 0
        self.LastFrameTime = 0
        self.ImageSubsample = subsample
        self.zoom = int(zoom) if zoom is not None else None

        self.Source = filename if filename is not None else data
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_IMAGE,
            size=sz,
            background_color=background_color,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(
            self,
            source=None,
            filename=None,
            data=None,
            size=(None, None),
            subsample=None,
            zoom=None,
            visible=None,
    ):
        """
        Changes some of the settings for the Image Element. Must call `Window.Read` or `Window.Finalize` prior.
        To clear an image that's been displayed, call with NONE of the options set.  A blank update call will
        delete the previously shown image.

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param source:   A filename or a base64 bytes. Will automatically detect the type and fill in filename or data for you.
        :type source:    str | bytes | None
        :param filename: filename to the new image to display.
        :type filename:  (str)
        :param data:     Base64 encoded string OR a tk.PhotoImage object
        :type data:      str | tkPhotoImage
        :param size:     (width, height) size of image in pixels
        :type size:      Tuple[int,int]
        :param subsample: amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type subsample: (int)
        :param zoom:     amount to increase the size of the image
        :type zoom:      (int)
        :param visible:  control visibility of element
        :type visible:   (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Image.update - The window was closed")
            return

        if source is not None:
            if isinstance(source, bytes):
                data = source
            elif isinstance(source, str):
                filename = source
            else:
                warnings.warn(
                    "Image element - source is not a valid type: {}".format(
                        type(source)
                    ),
                    UserWarning,
                )

        image = None
        if filename is not None:
            try:
                image = tk.PhotoImage(file=filename)
                if subsample is not None:
                    image = image.subsample(subsample)
                if zoom is not None:
                    image = image.zoom(int(zoom))
            except Exception as e:
                _error_popup_with_traceback("Exception updating Image element", e)

        elif data is not None:
            # if type(data) is bytes:
            try:
                image = tk.PhotoImage(data=data)
                if subsample is not None:
                    image = image.subsample(subsample)
                if zoom is not None:
                    image = image.zoom(int(zoom))
            except Exception as e:
                image = data
                # return  # an error likely means the window has closed so exit

        if image is not None:
            self.tktext_label.configure(image="")  # clear previous image
            if self.tktext_label.image is not None:
                del self.tktext_label.image
            if type(image) is not bytes:
                width, height = (
                    size[0] if size[0] is not None else image.width(),
                    size[1] if size[1] is not None else image.height(),
                )
            else:
                width, height = size
            try:  # sometimes crashes if user closed with X
                self.tktext_label.configure(image=image, width=width, height=height)
            except Exception as e:
                _error_popup_with_traceback("Exception updating Image element", e)
            self.tktext_label.image = image
        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        # if everything is set to None, then delete the image
        if (
                filename is None
                and image is None
                and visible is None
                and size == (None, None)
        ):
            # Using a try because the image may have been previously deleted and don't want an error if that's happened
            try:
                self.tktext_label.configure(image="", width=1, height=1, bd=0)
                self.tktext_label.image = None
            except Exception:
                pass

        if visible is not None:
            self._visible = visible

    def update_animation(self, source, time_between_frames=0):
        """
        Show an Animated GIF. Call the function as often as you like. The function will determine when to show the next frame and will automatically advance to the next frame at the right time.
        NOTE - does NOT perform a sleep call to delay
        :param source:              Filename or Base64 encoded string containing Animated GIF
        :type source:               str | bytes | None
        :param time_between_frames: Number of milliseconds to wait between showing frames
        :type time_between_frames:  (int)
        """

        if self.Source != source:
            self.AnimatedFrames = None
            self.Source = source

        if self.AnimatedFrames is None:
            self.TotalAnimatedFrames = 0
            self.AnimatedFrames = []
            # Load up to 1000 frames of animation.  stops when a bad frame is returns by tkinter
            for i in range(1000):
                if type(source) is not bytes:
                    try:
                        self.AnimatedFrames.append(
                            tk.PhotoImage(file=source, format="gif -index %i" % i)
                        )
                    except Exception as e:
                        break
                else:
                    try:
                        self.AnimatedFrames.append(
                            tk.PhotoImage(data=source, format="gif -index %i" % i)
                        )
                    except Exception as e:
                        break
            self.TotalAnimatedFrames = len(self.AnimatedFrames)
            self.LastFrameTime = time.time()
            self.CurrentFrameNumber = (
                -1
            )  # start at -1 because it is incremented before every frame is shown
        # show the frame

        now = time.time()

        if time_between_frames:
            if (now - self.LastFrameTime) * 1000 > time_between_frames:
                self.LastFrameTime = now
                self.CurrentFrameNumber = (
                                                  self.CurrentFrameNumber + 1
                                          ) % self.TotalAnimatedFrames
            else:  # don't reshow the frame again if not time for new frame
                return
        else:
            self.CurrentFrameNumber = (
                                              self.CurrentFrameNumber + 1
                                      ) % self.TotalAnimatedFrames
        image = self.AnimatedFrames[self.CurrentFrameNumber]
        try:  # needed in case the window was closed with an "X"
            self.tktext_label.configure(
                image=image, width=image.width(), heigh=image.height()
            )
        except Exception as e:
            print("Exception in update_animation", e)

    def update_animation_no_buffering(self, source, time_between_frames=0):
        """
        Show an Animated GIF. Call the function as often as you like. The function will determine when to show the next frame and will automatically advance to the next frame at the right time.
        NOTE - does NOT perform a sleep call to delay

        :param source:              Filename or Base64 encoded string containing Animated GIF
        :type source:               str | bytes
        :param time_between_frames: Number of milliseconds to wait between showing frames
        :type time_between_frames:  (int)
        """

        if self.Source != source:
            self.AnimatedFrames = None
            self.Source = source
            self.frame_num = 0

        now = time.time()

        if time_between_frames:
            if (now - self.LastFrameTime) * 1000 > time_between_frames:
                self.LastFrameTime = now
            else:  # don't reshow the frame again if not time for new frame
                return

        # read a frame
        while True:
            if type(source) is not bytes:
                try:
                    self.image = tk.PhotoImage(
                        file=source, format="gif -index %i" % self.frame_num
                    )
                    self.frame_num += 1
                except Exception:
                    self.frame_num = 0
            else:
                try:
                    self.image = tk.PhotoImage(
                        data=source, format="gif -index %i" % self.frame_num
                    )
                    self.frame_num += 1
                except Exception:
                    self.frame_num = 0
            if self.frame_num:
                break

        try:  # needed in case the window was closed with an "X"
            self.tktext_label.configure(
                image=self.image, width=self.image.width(), heigh=self.image.height()
            )

        except Exception:
            pass

    Update = update
    UpdateAnimation = update_animation


Im = Image


# ---------------------------------------------------------------------- #
#                           Canvas                                       #
# ---------------------------------------------------------------------- #
class Canvas(Element):
    def __init__(
            self,
            canvas=None,
            background_color=None,
            size=(None, None),
            s=(None, None),
            pad=None,
            p=None,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            border_width=0,
            metadata=None,
    ):
        """
        :param canvas:           Your own tk.Canvas if you already created it. Leave blank to create a Canvas
        :type canvas:            (tk.Canvas)
        :param background_color: color of background
        :type background_color:  (str)
        :param size:             (width in char, height in rows) size in pixels to make canvas
        :type size:              (int,int) | (None, None)
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None) | int
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:              Used with window.find_element and with return values to uniquely identify this element
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param tooltip:          text, that will appear when mouse hovers over the element
        :type tooltip:           (str)
        :param right_click_menu: A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:  List[List[ List[str] | str ]]
        :param expand_x:         If True the element will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param border_width:     width of border around element in pixels. Not normally used with Canvas element
        :type border_width:      (int)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self._TKCanvas = self.Widget = canvas
        self.RightClickMenu = right_click_menu
        self.BorderWidth = border_width
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_CANVAS,
            background_color=background_color,
            size=sz,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(self, background_color=None, visible=None):
        """

        :param background_color: color of background
        :type background_color:  (str)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        """

        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Canvas.update - The window was closed"
            )
            return

        if background_color not in (None, COLOR_SYSTEM_DEFAULT):
            self._TKCanvas.configure(background=background_color)
        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()
        if visible is not None:
            self._visible = visible

    @property
    def tk_canvas(self):
        """
        Returns the underlying tkiner Canvas widget

        :return: The tkinter canvas widget
        :rtype:  (tk.Canvas)
        """
        if self._TKCanvas is None:
            print(
                "*** Did you forget to call Finalize()? Your code should look something like: ***"
            )
            print('*** window = sg.Window("My Form", layout, finalize=True) ***')
        return self._TKCanvas

    TKCanvas = tk_canvas


# ---------------------------------------------------------------------- #
#                           Graph                                        #
# ---------------------------------------------------------------------- #
class Graph(Element):
    """
    Creates an area for you to draw on.  The MAGICAL property this Element has is that you interact
    with the element using your own coordinate system.  This is an important point!!  YOU define where the location
    is for (0,0).  Want (0,0) to be in the middle of the graph like a math 4-quadrant graph?  No problem!  Set your
    lower left corner to be (-100,-100) and your upper right to be (100,100) and you've got yourself a graph with
    (0,0) at the center.
    One of THE coolest of the Elements.
    You can also use float values. To do so, be sure and set the float_values parameter.
    Mouse click and drag events are possible and return the (x,y) coordinates of the mouse
    Drawing primitives return an "id" that is referenced when you want to operation on that item (e.g. to erase it)
    """

    def __init__(
            self,
            canvas_size,
            graph_bottom_left,
            graph_top_right,
            background_color=None,
            pad=None,
            p=None,
            change_submits=False,
            drag_submits=False,
            enable_events=False,
            motion_events=False,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            float_values=False,
            border_width=0,
            metadata=None,
    ):
        """
        :param canvas_size:       size of the canvas area in pixels
        :type canvas_size:        (int, int)
        :param graph_bottom_left: (x,y) The bottoms left corner of your coordinate system
        :type graph_bottom_left:  (int, int)
        :param graph_top_right:   (x,y) The top right corner of  your coordinate system
        :type graph_top_right:    (int, int)
        :param background_color:  background color of the drawing area
        :type background_color:   (str)
        :param pad:               Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                 Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                  (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param change_submits:    * DEPRICATED DO NOT USE. Use `enable_events` instead
        :type change_submits:     (bool)
        :param drag_submits:      if True and Events are enabled for the Graph, will report Events any time the mouse moves while button down.  When the mouse button is released, you'll get an event = graph key + '+UP' (if key is a string.. if not a string, it'll be made into a tuple)
        :type drag_submits:       (bool)
        :param enable_events:     If True then clicks on the Graph are immediately reported as an event. Use this instead of change_submits
        :type enable_events:      (bool)
        :param motion_events:     If True then if no button is down and the mouse is moved, an event is generated with key = graph key + '+MOVE' (if key is a string, it not a string then a tuple is returned)
        :type motion_events:      (bool)
        :param key:               Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                str | int | tuple | object
        :param k:                 Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                  str | int | tuple | object
        :param tooltip:           text, that will appear when mouse hovers over the element
        :type tooltip:            (str)
        :param right_click_menu:  A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:   List[List[ List[str] | str ]]
        :param expand_x:          If True the element will automatically expand in the X direction to fill available space
        :type expand_x:           (bool)
        :param expand_y:          If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:           (bool)
        :param visible:           set visibility state of the element (Default = True)
        :type visible:            (bool)
        :param float_values:      If True x,y coordinates are returned as floats, not ints
        :type float_values:       (bool)
        :param border_width:      width of border around element in pixels. Not normally used for Graph Elements
        :type border_width:       (int)
        :param metadata:          User metadata that can be set to ANYTHING
        :type metadata:           (Any)
        """

        self.CanvasSize = canvas_size
        self.BottomLeft = graph_bottom_left
        self.TopRight = graph_top_right
        # self._TKCanvas = None               # type: tk.Canvas
        self._TKCanvas2 = self.Widget = None  # type: tk.Canvas
        self.ChangeSubmits = change_submits or enable_events
        self.DragSubmits = drag_submits
        self.ClickPosition = (None, None)
        self.MouseButtonDown = False
        self.Images = {}
        self.RightClickMenu = right_click_menu
        self.FloatValues = float_values
        self.BorderWidth = border_width
        key = key if key is not None else k
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y
        self.motion_events = motion_events

        super().__init__(
            ELEM_TYPE_GRAPH,
            background_color=background_color,
            size=canvas_size,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def _convert_xy_to_canvas_xy(self, x_in, y_in):
        """
        Not user callable.  Used to convert user's coordinates into the ones used by tkinter
        :param x_in: The x coordinate to convert
        :type x_in:  int | float
        :param y_in: The y coordinate to convert
        :type y_in:  int | float
        :return:     (int, int) The converted canvas coordinates
        :rtype:      (int, int)
        """
        if None in (x_in, y_in):
            return None, None
        try:
            scale_x = (self.CanvasSize[0] - 0) / (self.TopRight[0] - self.BottomLeft[0])
            scale_y = (0 - self.CanvasSize[1]) / (self.TopRight[1] - self.BottomLeft[1])
        except Exception:
            scale_x = scale_y = 0

        new_x = 0 + scale_x * (x_in - self.BottomLeft[0])
        new_y = self.CanvasSize[1] + scale_y * (y_in - self.BottomLeft[1])
        return new_x, new_y

    def _convert_canvas_xy_to_xy(self, x_in, y_in):
        """
        Not user callable.  Used to convert tkinter Canvas coords into user's coordinates

        :param x_in: The x coordinate in canvas coordinates
        :type x_in:  (int)
        :param y_in: (int) The y coordinate in canvas coordinates
        :type y_in:
        :return:     The converted USER coordinates
        :rtype:      (int, int) | Tuple[float, float]
        """
        if None in (x_in, y_in):
            return None, None
        scale_x = (self.CanvasSize[0] - 0) / (self.TopRight[0] - self.BottomLeft[0])
        scale_y = (0 - self.CanvasSize[1]) / (self.TopRight[1] - self.BottomLeft[1])

        new_x = x_in / scale_x + self.BottomLeft[0]
        new_y = (y_in - self.CanvasSize[1]) / scale_y + self.BottomLeft[1]
        if self.FloatValues:
            return new_x, new_y
        else:
            return floor(new_x), floor(new_y)

    def draw_line(self, point_from, point_to, color="black", width=1):
        """
        Draws a line from one point to another point using USER'S coordinates. Can set the color and width of line
        :param point_from: Starting point for line
        :type point_from:  (int, int) | Tuple[float, float]
        :param point_to:   Ending point for line
        :type point_to:    (int, int) | Tuple[float, float]
        :param color:      Color of the line
        :type color:       (str)
        :param width:      width of line in pixels
        :type width:       (int)
        :return:           id returned from tktiner or None if user closed the window. id is used when you
        :rtype:            int | None
        """
        if point_from == (None, None):
            return
        converted_point_from = self._convert_xy_to_canvas_xy(
            point_from[0], point_from[1]
        )
        converted_point_to = self._convert_xy_to_canvas_xy(point_to[0], point_to[1])
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case window was closed with an X
            id_ = self._TKCanvas2.create_line(
                converted_point_from, converted_point_to, width=width, fill=color
            )
        except Exception:
            id_ = None
        return id_

    def draw_lines(self, points, color="black", width=1):
        """
        Draw a series of lines given list of points

        :param points: list of points that define the polygon
        :type points:  List[(int, int) | Tuple[float, float]]
        :param color:  Color of the line
        :type color:   (str)
        :param width:  width of line in pixels
        :type width:   (int)
        :return:       id returned from tktiner or None if user closed the window. id is used when you
        :rtype:        int | None
        """
        converted_points = [
            self._convert_xy_to_canvas_xy(point[0], point[1]) for point in points
        ]

        try:  # in case window was closed with an X
            id = self._TKCanvas2.create_line(*converted_points, width=width, fill=color)
        except Exception:
            if self._TKCanvas2 is None:
                print(
                    "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
                )
                print("Call Window.Finalize() prior to this operation")
            id = None
        return id

    def draw_point(self, point, size=2, color="black"):
        """
        Draws a "dot" at the point you specify using the USER'S coordinate system
        :param point: Center location using USER'S coordinate system
        :type point:  (int, int) | Tuple[float, float]
        :param size:  Radius? (Or is it the diameter?) in user's coordinate values.
        :type size:   int | float
        :param color: color of the point to draw
        :type color:  (str)
        :return:      id returned from tkinter that you'll need if you want to manipulate the point
        :rtype:       int | None
        """
        if point == (None, None):
            return
        converted_point = self._convert_xy_to_canvas_xy(point[0], point[1])
        size_converted = self._convert_xy_to_canvas_xy(point[0] + size, point[1])
        size = size_converted[0] - converted_point[0]
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # needed in case window was closed with an X
            point1 = converted_point[0] - size // 2, converted_point[1] - size // 2
            point2 = converted_point[0] + size // 2, converted_point[1] + size // 2
            id = self._TKCanvas2.create_oval(
                point1[0],
                point1[1],
                point2[0],
                point2[1],
                width=0,
                fill=color,
                outline=color,
            )
        except Exception:
            id = None
        return id

    def draw_circle(
            self, center_location, radius, fill_color=None, line_color="black", line_width=1
    ):
        """
        Draws a circle, cenetered at the location provided.  Can set the fill and outline colors
        :param center_location: Center location using USER'S coordinate system
        :type center_location:  (int, int) | Tuple[float, float]
        :param radius:          Radius in user's coordinate values.
        :type radius:           int | float
        :param fill_color:      color of the point to draw
        :type fill_color:       (str)
        :param line_color:      color of the outer line that goes around the circle (sorry, can't set thickness)
        :type line_color:       (str)
        :param line_width:      width of the line around the circle, the outline, in pixels
        :type line_width:       (int)
        :return:                id returned from tkinter that you'll need if you want to manipulate the circle
        :rtype:                 int | None
        """
        if center_location == (None, None):
            return
        converted_point = self._convert_xy_to_canvas_xy(
            center_location[0], center_location[1]
        )
        radius_converted = self._convert_xy_to_canvas_xy(
            center_location[0] + radius, center_location[1]
        )
        radius = radius_converted[0] - converted_point[0]
        # radius = radius_converted[1]-5
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        # print('Oval parms', int(converted_point[0]) - int(radius), int(converted_point[1]) - int(radius),
        #                                      int(converted_point[0]) + int(radius), int(converted_point[1]) + int(radius))
        try:  # needed in case the window was closed with an X
            id = self._TKCanvas2.create_oval(
                int(converted_point[0]) - int(radius),
                int(converted_point[1]) - int(radius),
                int(converted_point[0]) + int(radius),
                int(converted_point[1]) + int(radius),
                fill=fill_color,
                outline=line_color,
                width=line_width,
            )
        except Exception:
            id = None
        return id

    def draw_oval(
            self, top_left, bottom_right, fill_color=None, line_color=None, line_width=1
    ):
        """
        Draws an oval based on coordinates in user coordinate system. Provide the location of a "bounding rectangle"
        :param top_left:     the top left point of bounding rectangle
        :type top_left:      (int, int) | Tuple[float, float]
        :param bottom_right: the bottom right point of bounding rectangle
        :type bottom_right:  (int, int) | Tuple[float, float]
        :param fill_color:   color of the interrior
        :type fill_color:    (str)
        :param line_color:   color of outline of oval
        :type line_color:    (str)
        :param line_width:   width of the line around the oval, the outline, in pixels
        :type line_width:    (int)
        :return:             id returned from tkinter that you'll need if you want to manipulate the oval
        :rtype:              int | None
        """
        converted_top_left = self._convert_xy_to_canvas_xy(top_left[0], top_left[1])
        converted_bottom_right = self._convert_xy_to_canvas_xy(
            bottom_right[0], bottom_right[1]
        )
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case windows close with X
            id = self._TKCanvas2.create_oval(
                converted_top_left[0],
                converted_top_left[1],
                converted_bottom_right[0],
                converted_bottom_right[1],
                fill=fill_color,
                outline=line_color,
                width=line_width,
            )
        except Exception:
            id = None

        return id

    def draw_arc(
            self,
            top_left,
            bottom_right,
            extent,
            start_angle,
            style=None,
            arc_color="black",
            line_width=1,
            fill_color=None,
    ):
        """
        Draws different types of arcs.  Uses a "bounding box" to define location
        :param line_width:
        :param top_left:     the top left point of bounding rectangle
        :type top_left:      (int, int) | Tuple[float, float]
        :param bottom_right: the bottom right point of bounding rectangle
        :type bottom_right:  (int, int) | Tuple[float, float]
        :param extent:       Andle to end drawing. Used in conjunction with start_angle
        :type extent:        (float)
        :param start_angle:  Angle to begin drawing. Used in conjunction with extent
        :type start_angle:   (float)
        :param style:        Valid choices are One of these Style strings- 'pieslice', 'chord', 'arc', 'first', 'last', 'butt', 'projecting', 'round', 'bevel', 'miter'
        :type style:         (str)
        :param arc_color:    color to draw arc with
        :type arc_color:     (str)
        :param fill_color:   color to fill the area
        :type fill_color:    (str)
        :return:             id returned from tkinter that you'll need if you want to manipulate the arc
        :rtype:              int | None
        """
        converted_top_left = self._convert_xy_to_canvas_xy(top_left[0], top_left[1])
        converted_bottom_right = self._convert_xy_to_canvas_xy(
            bottom_right[0], bottom_right[1]
        )
        tkstyle = tk.PIESLICE if style is None else style
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case closed with X
            id = self._TKCanvas2.create_arc(
                converted_top_left[0],
                converted_top_left[1],
                converted_bottom_right[0],
                converted_bottom_right[1],
                extent=extent,
                start=start_angle,
                style=tkstyle,
                outline=arc_color,
                width=line_width,
                fill=fill_color,
            )
        except Exception as e:
            print("Error encountered drawing arc.", e)
            id = None
        return id

    def draw_rectangle(
            self, top_left, bottom_right, fill_color=None, line_color=None, line_width=None
    ):
        """
        Draw a rectangle given 2 points. Can control the line and fill colors

        :param top_left:     the top left point of rectangle
        :type top_left:      (int, int) | Tuple[float, float]
        :param bottom_right: the bottom right point of rectangle
        :type bottom_right:  (int, int) | Tuple[float, float]
        :param fill_color:   color of the interior
        :type fill_color:    (str)
        :param line_color:   color of outline
        :type line_color:    (str)
        :param line_width:   width of the line in pixels
        :type line_width:    (int)
        :return:             int | None id returned from tkinter that you'll need if you want to manipulate the rectangle
        :rtype:              int | None
        """

        converted_top_left = self._convert_xy_to_canvas_xy(top_left[0], top_left[1])
        converted_bottom_right = self._convert_xy_to_canvas_xy(
            bottom_right[0], bottom_right[1]
        )
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        if line_width is None:
            line_width = 1
        try:  # in case closed with X
            id = self._TKCanvas2.create_rectangle(
                converted_top_left[0],
                converted_top_left[1],
                converted_bottom_right[0],
                converted_bottom_right[1],
                fill=fill_color,
                outline=line_color,
                width=line_width,
            )
        except Exception:
            id = None
        return id

    def draw_polygon(self, points, fill_color=None, line_color=None, line_width=None):
        """
        Draw a polygon given list of points

        :param points:     list of points that define the polygon
        :type points:      List[(int, int) | Tuple[float, float]]
        :param fill_color: color of the interior
        :type fill_color:  (str)
        :param line_color: color of outline
        :type line_color:  (str)
        :param line_width: width of the line in pixels
        :type line_width:  (int)
        :return:           id returned from tkinter that you'll need if you want to manipulate the rectangle
        :rtype:            int | None
        """

        converted_points = [
            self._convert_xy_to_canvas_xy(point[0], point[1]) for point in points
        ]
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case closed with X
            id = self._TKCanvas2.create_polygon(
                converted_points, fill=fill_color, outline=line_color, width=line_width
            )
        except Exception:
            id = None
        return id

    def draw_text(
            self,
            text,
            location,
            color="black",
            font=None,
            angle=0,
            text_location=TEXT_LOCATION_CENTER,
    ):
        """
        Draw some text on your graph.  This is how you label graph number lines for example

        :param text:          text to display
        :type text:           (Any)
        :param location:      location to place first letter
        :type location:       (int, int) | Tuple[float, float]
        :param color:         text color
        :type color:          (str)
        :param font:          specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:           (str or (str, int[, str]) or None)
        :param angle:         Angle 0 to 360 to draw the text.  Zero represents horizontal text
        :type angle:          (float)
        :param text_location: "anchor" location for the text. Values start with TEXT_LOCATION_
        :type text_location:  (enum)
        :return:              id returned from tkinter that you'll need if you want to manipulate the text
        :rtype:               int | None
        """
        text = str(text)
        if location == (None, None):
            return
        converted_point = self._convert_xy_to_canvas_xy(location[0], location[1])
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case closed with X
            id = self._TKCanvas2.create_text(
                converted_point[0],
                converted_point[1],
                text=text,
                font=font,
                fill=color,
                angle=angle,
                anchor=text_location,
            )
        except Exception:
            id = None
        return id

    def draw_image(self, filename=None, data=None, location=(None, None)):
        """
        Places an image onto your canvas.  It's a really important method for this element as it enables so much

        :param filename: if image is in a file, path and filename for the image. (GIF and PNG only!)
        :type filename:  (str)
        :param data:     if image is in Base64 format or raw? format then use instead of filename
        :type data:      str | bytes
        :param location: the (x,y) location to place image's top left corner
        :type location:  (int, int) | Tuple[float, float]
        :return:         id returned from tkinter that you'll need if you want to manipulate the image
        :rtype:          int | None
        """
        if location == (None, None):
            return
        if filename is not None:
            image = tk.PhotoImage(file=filename)
        elif data is not None:
            # if type(data) is bytes:
            try:
                image = tk.PhotoImage(data=data)
            except Exception:
                return None  # an error likely means the window has closed so exit
        converted_point = self._convert_xy_to_canvas_xy(location[0], location[1])
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        try:  # in case closed with X
            id = self._TKCanvas2.create_image(
                converted_point, image=image, anchor=tk.NW
            )
            self.Images[id] = image
        except Exception:
            id = None
        return id

    def erase(self):
        """
        Erase the Graph - Removes all figures previously "drawn" using the Graph methods (e.g. DrawText)
        """
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        self.Images = {}
        try:  # in case window was closed with X
            self._TKCanvas2.delete("all")
        except Exception:
            pass

    def delete_figure(self, id_):
        """
        Remove from the Graph the figure represented by id. The id is given to you anytime you call a drawing primitive

        :param id_: the id returned to you when calling one of the drawing methods
        :type id_:  (int)
        """
        try:
            self._TKCanvas2.delete(id_)
        except Exception:
            print("DeleteFigure - bad ID {}".format(id_))
        try:
            del self.Images[
                id_
            ]  # in case was an image. If wasn't an image, then will get exception
        except Exception:
            pass

    def update(self, background_color=None, visible=None):
        """
        Changes some of the settings for the Graph Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param background_color: color of background
        :type background_color:  ???
        :param visible:          control visibility of element
        :type visible:           (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Graph.update - The window was closed")
            return

        if background_color is not None and background_color != COLOR_SYSTEM_DEFAULT:
            self._TKCanvas2.configure(background=background_color)

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible

    def move(self, x_direction, y_direction):
        """
        Moves the entire drawing area (the canvas) by some delta from the current position.  Units are indicated in your coordinate system indicated number of ticks in your coordinate system

        :param x_direction: how far to move in the "X" direction in your coordinates
        :type x_direction:  int | float
        :param y_direction: how far to move in the "Y" direction in your coordinates
        :type y_direction:  int | float
        """
        zero_converted = self._convert_xy_to_canvas_xy(0, 0)
        shift_converted = self._convert_xy_to_canvas_xy(x_direction, y_direction)
        shift_amount = (
            shift_converted[0] - zero_converted[0],
            shift_converted[1] - zero_converted[1],
        )
        if self._TKCanvas2 is None:
            print(
                "*** WARNING - The Graph element has not been finalized and cannot be drawn upon ***"
            )
            print("Call Window.Finalize() prior to this operation")
            return None
        self._TKCanvas2.move("all", shift_amount[0], shift_amount[1])

    def move_figure(self, figure, x_direction, y_direction):
        """
        Moves a previously drawn figure using a "delta" from current position

        :param figure:      Previously obtained figure-id. These are returned from all Draw methods
        :type figure:       (id)
        :param x_direction: delta to apply to position in the X direction
        :type x_direction:  int | float
        :param y_direction: delta to apply to position in the Y direction
        :type y_direction:  int | float
        """
        zero_converted = self._convert_xy_to_canvas_xy(0, 0)
        shift_converted = self._convert_xy_to_canvas_xy(x_direction, y_direction)
        shift_amount = (
            shift_converted[0] - zero_converted[0],
            shift_converted[1] - zero_converted[1],
        )
        if figure is None:
            print("* move_figure warning - your figure is None *")
            return None
        self._TKCanvas2.move(figure, shift_amount[0], shift_amount[1])

    def relocate_figure(self, figure, x, y):
        """
        Move a previously made figure to an arbitrary (x,y) location. This differs from the Move methods because it
        uses absolute coordinates versus relative for Move

        :param figure: Previously obtained figure-id. These are returned from all Draw methods
        :type figure:  (id)
        :param x:      location on X axis (in user coords) to move the upper left corner of the figure
        :type x:       int | float
        :param y:      location on Y axis (in user coords) to move the upper left corner of the figure
        :type y:       int | float
        """

        zero_converted = self._convert_xy_to_canvas_xy(0, 0)
        shift_converted = self._convert_xy_to_canvas_xy(x, y)
        shift_amount = (
            shift_converted[0] - zero_converted[0],
            shift_converted[1] - zero_converted[1],
        )
        if figure is None:
            print(
                "*** WARNING - Your figure is None. It most likely means your did not Finalize your Window ***"
            )
            print("Call Window.Finalize() prior to all graph operations")
            return None
        xy = self._TKCanvas2.coords(figure)
        self._TKCanvas2.move(
            figure, shift_converted[0] - xy[0], shift_converted[1] - xy[1]
        )

    def send_figure_to_back(self, figure):
        """
        Changes Z-order of figures on the Graph.  Sends the indicated figure to the back of all other drawn figures

        :param figure: value returned by tkinter when creating the figure / drawing
        :type figure:  (int)
        """
        self.TKCanvas.tag_lower(
            figure
        )  # move figure to the "bottom" of all other figure

    def bring_figure_to_front(self, figure):
        """
        Changes Z-order of figures on the Graph.  Brings the indicated figure to the front of all other drawn figures

        :param figure: value returned by tkinter when creating the figure / drawing
        :type figure:  (int)
        """
        self.TKCanvas.tag_raise(figure)  # move figure to the "top" of all other figures

    def get_figures_at_location(self, location):
        """
        Returns a list of figures located at a particular x,y location within the Graph

        :param location: point to check
        :type location:  (int, int) | Tuple[float, float]
        :return:         a list of previously drawn "Figures" (returned from the drawing primitives)
        :rtype:          List[int]
        """
        x, y = self._convert_xy_to_canvas_xy(location[0], location[1])
        ids = self.TKCanvas.find_overlapping(x, y, x, y)
        return ids

    def get_bounding_box(self, figure):
        """
        Given a figure, returns the upper left and lower right bounding box coordinates

        :param figure: a previously drawing figure
        :type figure:  object
        :return:       upper left x, upper left y, lower right x, lower right y
        :rtype:        Tuple[int, int, int, int] | Tuple[float, float, float, float]
        """
        box = self.TKCanvas.bbox(figure)
        top_left = self._convert_canvas_xy_to_xy(box[0], box[1])
        bottom_right = self._convert_canvas_xy_to_xy(box[2], box[3])
        return top_left, bottom_right

    def change_coordinates(self, graph_bottom_left, graph_top_right):
        """
        Changes the corrdinate system to a new one.  The same 2 points in space are used to define the coorinate
        system - the bottom left and the top right values of your graph.

        :param graph_bottom_left: The bottoms left corner of your coordinate system
        :type graph_bottom_left:  (int, int) (x,y)
        :param graph_top_right:   The top right corner of  your coordinate system
        :type graph_top_right:    (int, int)  (x,y)
        """
        self.BottomLeft = graph_bottom_left
        self.TopRight = graph_top_right

    @property
    def tk_canvas(self):
        """
        Returns the underlying tkiner Canvas widget

        :return: The tkinter canvas widget
        :rtype:  (tk.Canvas)
        """
        if self._TKCanvas2 is None:
            print(
                "*** Did you forget to call Finalize()? Your code should look something like: ***"
            )
            print('*** form = sg.Window("My Form").Layout(layout).Finalize() ***')
        return self._TKCanvas2

    # button release callback
    def button_release_call_back(self, event):
        """
        Not a user callable method.  Used to get Graph click events. Called by tkinter when button is released

        :param event: (event) event info from tkinter. Note not used in this method
        :type event:
        """
        if not self.DragSubmits:
            return  # only report mouse up for drag operations
        self.ClickPosition = self._convert_canvas_xy_to_xy(event.x, event.y)
        self.ParentForm.LastButtonClickedWasRealtime = False
        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = (
                "__GRAPH__"  # need to put something rather than None
            )
        _exit_mainloop(self.ParentForm)
        if isinstance(self.ParentForm.LastButtonClicked, str):
            self.ParentForm.LastButtonClicked = (
                    self.ParentForm.LastButtonClicked + "+UP"
            )
        else:
            self.ParentForm.LastButtonClicked = (
                self.ParentForm.LastButtonClicked,
                "+UP",
            )
        self.MouseButtonDown = False

    # button callback
    def button_press_call_back(self, event):
        """
        Not a user callable method.  Used to get Graph click events. Called by tkinter when button is released

        :param event: (event) event info from tkinter. Contains the x and y coordinates of a click
        :type event:
        """

        self.ClickPosition = self._convert_canvas_xy_to_xy(event.x, event.y)
        self.ParentForm.LastButtonClickedWasRealtime = self.DragSubmits
        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = (
                "__GRAPH__"  # need to put something rather than None
            )
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     self.ParentForm.TKroot.quit()  # kick out of loop if read was called
        _exit_mainloop(self.ParentForm)
        self.MouseButtonDown = True

    def _update_position_for_returned_values(self, event):
        """
        Updates the variable that's used when the values dictionary is returned from a window read.

        Not called by the user.  It's called from another method/function that tkinter calledback

        :param event: (event) event info from tkinter. Contains the x and y coordinates of a click
        :type event:
        """
        """
        Updates the variable that's used when the values dictionary is returned from a window read.

        Not called by the user.  It's called from another method/function that tkinter calledback

        :param event: (event) event info from tkinter. Contains the x and y coordinates of a click
        :type event:  
        """

        self.ClickPosition = self._convert_canvas_xy_to_xy(event.x, event.y)

    # button callback
    def motion_call_back(self, event):
        """
        Not a user callable method.  Used to get Graph mouse motion events. Called by tkinter when mouse moved

        :param event: (event) event info from tkinter. Contains the x and y coordinates of a mouse
        :type event:
        """

        if not self.MouseButtonDown and not self.motion_events:
            return
        self.ClickPosition = self._convert_canvas_xy_to_xy(event.x, event.y)
        self.ParentForm.LastButtonClickedWasRealtime = self.DragSubmits
        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = (
                "__GRAPH__"  # need to put something rather than None
            )
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     self.ParentForm.TKroot.quit()  # kick out of loop if read was called
        if self.motion_events and not self.MouseButtonDown:
            if isinstance(self.ParentForm.LastButtonClicked, str):
                self.ParentForm.LastButtonClicked = (
                        self.ParentForm.LastButtonClicked + "+MOVE"
                )
            else:
                self.ParentForm.LastButtonClicked = (
                    self.ParentForm.LastButtonClicked,
                    "+MOVE",
                )
        _exit_mainloop(self.ParentForm)

    BringFigureToFront = bring_figure_to_front
    ButtonPressCallBack = button_press_call_back
    ButtonReleaseCallBack = button_release_call_back
    DeleteFigure = delete_figure
    DrawArc = draw_arc
    DrawCircle = draw_circle
    DrawImage = draw_image
    DrawLine = draw_line
    DrawOval = draw_oval
    DrawPoint = draw_point
    DrawPolygon = draw_polygon
    DrawLines = draw_lines
    DrawRectangle = draw_rectangle
    DrawText = draw_text
    GetFiguresAtLocation = get_figures_at_location
    GetBoundingBox = get_bounding_box
    Erase = erase
    MotionCallBack = motion_call_back
    Move = move
    MoveFigure = move_figure
    RelocateFigure = relocate_figure
    SendFigureToBack = send_figure_to_back
    TKCanvas = tk_canvas
    Update = update


G = Graph


# ---------------------------------------------------------------------- #
#                           Frame                                        #
# ---------------------------------------------------------------------- #
class Frame(Element):
    """
    A Frame Element that contains other Elements. Encloses with a line around elements and a text label.
    """

    def __init__(
            self,
            title,
            layout,
            title_color=None,
            background_color=None,
            title_location=None,
            relief=DEFAULT_FRAME_RELIEF,
            size=(None, None),
            s=(None, None),
            font=None,
            pad=None,
            p=None,
            border_width=None,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            grab=None,
            visible=True,
            element_justification="left",
            vertical_alignment=None,
            metadata=None,
    ):
        """
        :param title:                 text that is displayed as the Frame's "label" or title
        :type title:                  (str)
        :param layout:                The layout to put inside the Frame
        :type layout:                 List[List[Elements]]
        :param title_color:           color of the title text
        :type title_color:            (str)
        :param background_color:      background color of the Frame
        :type background_color:       (str)
        :param title_location:        location to place the text title.  Choices include: TITLE_LOCATION_TOP TITLE_LOCATION_BOTTOM TITLE_LOCATION_LEFT TITLE_LOCATION_RIGHT TITLE_LOCATION_TOP_LEFT TITLE_LOCATION_TOP_RIGHT TITLE_LOCATION_BOTTOM_LEFT TITLE_LOCATION_BOTTOM_RIGHT
        :type title_location:         (enum)
        :param relief:                relief style. Values are same as other elements with reliefs. Choices include RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID
        :type relief:                 (enum)
        :param size:                  (width, height) Sets an initial hard-coded size for the Frame. This used to be a problem, but was fixed in 4.53.0 and works better than Columns when using the size paramter
        :type size:                   (int, int)
        :param s:                     Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                      (int, int)  | (None, None) | int
        :param font:                  specifies the  font family, size, etc. for the TITLE. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                   (str or (str, int[, str]) or None)
        :param pad:                   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param border_width:          width of border around element in pixels
        :type border_width:           (int)
        :param key:                   Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                    str | int | tuple | object
        :param k:                     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                      str | int | tuple | object
        :param tooltip:               text, that will appear when mouse hovers over the element
        :type tooltip:                (str)
        :param right_click_menu:      A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:       List[List[ List[str] | str ]]
        :param expand_x:              If True the element will automatically expand in the X direction to fill available space
        :type expand_x:               (bool)
        :param expand_y:              If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:               (bool)
        :param grab:                  If True can grab this element and move the window around. Default is False
        :type grab:                   (bool)
        :param visible:               set visibility state of the element
        :type visible:                (bool)
        :param element_justification: All elements inside the Frame will have this justification 'left', 'right', 'center' are valid values
        :type element_justification:  (str)
        :param vertical_alignment:    Place the Frame at the 'top', 'center', 'bottom' of the row (can also use t,c,r). Defaults to no setting (tkinter decides)
        :type vertical_alignment:     (str)
        :param metadata:              User metadata that can be set to ANYTHING
        :type metadata:               (Any)
        """

        self.UseDictionary = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.ParentWindow = None
        self.Rows = []
        # self.ParentForm = None
        self.TKFrame = None
        self.Title = title
        self.Relief = relief
        self.TitleLocation = title_location
        self.BorderWidth = border_width
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.RightClickMenu = right_click_menu
        self.ContainerElemementNumber = Window._GetAContainerNumber()
        self.ElementJustification = element_justification
        self.VerticalAlignment = vertical_alignment
        self.Widget = None  # type: tk.LabelFrame
        self.Grab = grab
        self.Layout(layout)
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_FRAME,
            background_color=background_color,
            text_color=title_color,
            size=sz,
            font=font,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def add_row(self, *args):
        """
        Not recommended user call.  Used to add rows of Elements to the Frame Element.

        """
        NumRows = len(self.Rows)  # number of existing rows is our row number
        CurrentRowNumber = NumRows  # this row's number
        CurrentRow = []  # start with a blank row and build up
        # -------------------------  Add the elements to a row  ------------------------- #
        for i, element in enumerate(
                args
        ):  # Loop through list of elements and add them to the row
            if type(element) == list:
                PopupError(
                    "Error creating Frame layout",
                    "Layout has a LIST instead of an ELEMENT",
                    "This sometimes means you have a badly placed ]",
                    "The offensive list is:",
                    element,
                    "This list will be stripped from your layout",
                    keep_on_top=True,
                )
                continue
            elif callable(element) and not isinstance(element, Element):
                PopupError(
                    "Error creating Frame layout",
                    "Layout has a FUNCTION instead of an ELEMENT",
                    "This likely means you are missing () from your layout",
                    "The offensive list is:",
                    element,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                )
                continue
            if element.ParentContainer is not None:
                warnings.warn(
                    "*** YOU ARE ATTEMPTING TO REUSE AN ELEMENT IN YOUR LAYOUT! Once placed in a layout, an element cannot be used in another layout. ***",
                    UserWarning,
                )
                _error_popup_with_traceback(
                    "Error creating Frame layout",
                    "The layout specified has already been used",
                    'You MUST start witha "clean", unused layout every time you create a window',
                    "The offensive Element = ",
                    element,
                    "and has a key = ",
                    element.Key,
                    "This item will be stripped from your layout",
                    'Hint - try printing your layout and matching the IDs "print(layout)"',
                )
                continue
            element.Position = (CurrentRowNumber, i)
            element.ParentContainer = self
            CurrentRow.append(element)
            if element.Key is not None:
                self.UseDictionary = True
        # -------------------------  Append the row to list of Rows  ------------------------- #
        self.Rows.append(CurrentRow)

    def layout(self, rows):
        """
        Can use like the Window.Layout method, but it's better to use the layout parameter when creating

        :param rows: The rows of Elements
        :type rows:  List[List[Element]]
        :return:     Used for chaining
        :rtype:      (Frame)
        """

        for row in rows:
            try:
                iter(row)
            except TypeError:
                PopupError(
                    "Error creating Frame layout",
                    "Your row is not an iterable (e.g. a list)",
                    "Instead of a list, the type found was {}".format(type(row)),
                    "The offensive row = ",
                    row,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            self.AddRow(*row)
        return self

    def _GetElementAtLocation(self, location):
        """
        Not user callable. Used to find the Element at a row, col position within the layout

        :param location: (row, column) position of the element to find in layout
        :type location:  (int, int)
        :return:         (Element) The element found at the location
        :rtype:          (Element)
        """

        (row_num, col_num) = location
        row = self.Rows[row_num]
        element = row[col_num]
        return element

    def update(self, value=None, visible=None):
        """
        Changes some of the settings for the Frame Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:   New text value to show on frame
        :type value:    (Any)
        :param visible: control visibility of element
        :type visible:  (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Frame.update - The window was closed")
            return

        if not visible:
            self._pack_forget_save_settings()
            # self.TKFrame.pack_forget()
        elif visible:
            self._pack_restore_settings()
            # self.TKFrame.pack(padx=self.pad_used[0], pady=self.pad_used[1])
        if value is not None:
            self.TKFrame.config(text=str(value))
        if visible is not None:
            self._visible = visible

    AddRow = add_row
    Layout = layout
    Update = update


Fr = Frame


# ---------------------------------------------------------------------- #
#                           Vertical Separator                           #
# ---------------------------------------------------------------------- #
class VerticalSeparator(Element):
    """
    Vertical Separator Element draws a vertical line at the given location. It will span 1 "row". Usually paired with
    Column Element if extra height is needed
    """

    def __init__(self, color=None, pad=None, p=None, key=None, k=None):
        """
        :param color: Color of the line. Defaults to theme's text color. Can be name or #RRGGBB format
        :type color:  (str)
        :param pad:   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:   Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:    str | int | tuple | object
        :param k:     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:      str | int | tuple | object
        """
        key = key if key is not None else k
        pad = pad if pad is not None else p
        self.expand_x = None
        self.expand_y = None
        self.Orientation = "vertical"  # for now only vertical works
        self.color = color if color is not None else theme_text_color()
        super().__init__(ELEM_TYPE_SEPARATOR, pad=pad, key=key)


VSeperator = VerticalSeparator
VSeparator = VerticalSeparator
VSep = VerticalSeparator


# ---------------------------------------------------------------------- #
#                           Horizontal Separator                           #
# ---------------------------------------------------------------------- #
class HorizontalSeparator(Element):
    """
    Horizontal Separator Element draws a Horizontal line at the given location.
    """

    def __init__(self, color=None, pad=None, p=None, key=None, k=None):
        """
        :param color: Color of the line. Defaults to theme's text color. Can be name or #RRGGBB format
        :type color:  (str)
        :param pad:   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:   Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:    str | int | tuple | object
        :param k:     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:      str | int | tuple | object
        """

        self.Orientation = "horizontal"  # for now only vertical works
        self.color = color if color is not None else theme_text_color()
        self.expand_x = True
        self.expand_y = None
        key = key if key is not None else k
        pad = pad if pad is not None else p

        super().__init__(ELEM_TYPE_SEPARATOR, pad=pad, key=key)


HSeparator = HorizontalSeparator
HSep = HorizontalSeparator


# ---------------------------------------------------------------------- #
#                           Sizegrip                                     #
# ---------------------------------------------------------------------- #
class Sizegrip(Element):
    """
    Sizegrip element will be added to the bottom right corner of your window.
    It should be placed on the last row of your window along with any other elements on that row.
    The color will match the theme's background color.
    """

    def __init__(self, background_color=None, pad=None, p=(0, 0), key=None, k=None):
        """
        Sizegrip Element
        :param background_color: color to use for the background of the grip
        :type background_color:  str
        :param pad:   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:   Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:    str | int | tuple | object
        :param k:     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:      str | int | tuple | object
        """

        bg = (
            background_color
            if background_color is not None
            else theme_background_color()
        )
        pad = pad if pad is not None else p
        key = key if key is not None else k

        super().__init__(ELEM_TYPE_SIZEGRIP, background_color=bg, key=key, pad=pad)


SGrip = Sizegrip


# ---------------------------------------------------------------------- #
#                           Tab                                          #
# ---------------------------------------------------------------------- #
class Tab(Element):
    """
    Tab Element is another "Container" element that holds a layout and displays a tab with text. Used with TabGroup only
    Tabs are never placed directly into a layout.  They are always "Contained" in a TabGroup layout
    """

    def __init__(
            self,
            title,
            layout,
            title_color=None,
            background_color=None,
            font=None,
            pad=None,
            p=None,
            disabled=False,
            border_width=None,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            element_justification="left",
            image_source=None,
            image_subsample=None,
            image_zoom=None,
            metadata=None,
    ):
        """
        :param title:                 text to show on the tab
        :type title:                  (str)
        :param layout:                The element layout that will be shown in the tab
        :type layout:                 List[List[Element]]
        :param title_color:           color of the tab text (note not currently working on tkinter)
        :type title_color:            (str)
        :param background_color:      color of background of the entire layout
        :type background_color:       (str)
        :param font:                  NOT USED in the tkinter port
        :type font:                   (str or (str, int[, str]) or None)
        :param pad:                   Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                    (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                     Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param disabled:              If True button will be created disabled
        :type disabled:               (bool)
        :param border_width:          NOT USED in tkinter port
        :type border_width:           (int)
        :param key:                   Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                    str | int | tuple | object
        :param k:                     Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                      str | int | tuple | object
        :param tooltip:               text, that will appear when mouse hovers over the element
        :type tooltip:                (str)
        :param right_click_menu:      A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:       List[List[ List[str] | str ]]
        :param expand_x:              If True the element will automatically expand in the X direction to fill available space
        :type expand_x:               (bool)
        :param expand_y:              If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:               (bool)
        :param visible:               set visibility state of the element
        :type visible:                (bool)
        :param element_justification: All elements inside the Tab will have this justification 'left', 'right', 'center' are valid values
        :type element_justification:  (str)
        :param image_source:          A filename or a base64 bytes of an image to place on the Tab
        :type image_source:            str | bytes | None
        :param image_subsample:       amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
        :type image_subsample:        (int)
        :param image_zoom:            amount to increase the size of the image. 2=twice size, 3=3 times, etc
        :type image_zoom:             (int)
        :param metadata:              User metadata that can be set to ANYTHING
        :type metadata:               (Any)
        """

        filename = data = None
        if image_source is not None:
            if isinstance(image_source, bytes):
                data = image_source
            elif isinstance(image_source, str):
                filename = image_source
            else:
                warnings.warn(
                    "Image element - source is not a valid type: {}".format(
                        type(image_source)
                    ),
                    UserWarning,
                )

        self.Filename = filename
        self.Data = data
        self.ImageSubsample = image_subsample
        self.zoom = int(image_zoom) if image_zoom is not None else None
        self.UseDictionary = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.ParentWindow = None
        self.Rows = []
        self.TKFrame = None
        self.Widget = None  # type: tk.Frame
        self.Title = title
        self.BorderWidth = border_width
        self.Disabled = disabled
        self.ParentNotebook = None
        self.TabID = None
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.RightClickMenu = right_click_menu
        self.ContainerElemementNumber = Window._GetAContainerNumber()
        self.ElementJustification = element_justification
        key = key if key is not None else k
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        self.Layout(layout)

        super().__init__(
            ELEM_TYPE_TAB,
            background_color=background_color,
            text_color=title_color,
            font=font,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def add_row(self, *args):
        """
        Not recommended use call.  Used to add rows of Elements to the Frame Element.

        """
        NumRows = len(self.Rows)  # number of existing rows is our row number
        CurrentRowNumber = NumRows  # this row's number
        CurrentRow = []  # start with a blank row and build up
        # -------------------------  Add the elements to a row  ------------------------- #
        for i, element in enumerate(
                args
        ):  # Loop through list of elements and add them to the row
            if type(element) == list:
                popup_error_with_traceback(
                    "Error creating Tab layout",
                    "Layout has a LIST instead of an ELEMENT",
                    "This sometimes means you have a badly placed ]",
                    "The offensive list is:",
                    element,
                    "This list will be stripped from your layout",
                )
                continue
            elif callable(element) and not isinstance(element, Element):
                popup_error_with_traceback(
                    "Error creating Tab layout",
                    "Layout has a FUNCTION instead of an ELEMENT",
                    "This likely means you are missing () from your layout",
                    "The offensive list is:",
                    element,
                    "This item will be stripped from your layout",
                )
                continue
            if element.ParentContainer is not None:
                warnings.warn(
                    "*** YOU ARE ATTEMPTING TO REUSE AN ELEMENT IN YOUR LAYOUT! Once placed in a layout, an element cannot be used in another layout. ***",
                    UserWarning,
                )
                popup_error_with_traceback(
                    "Error creating Tab layout",
                    "The layout specified has already been used",
                    'You MUST start witha "clean", unused layout every time you create a window',
                    "The offensive Element = ",
                    element,
                    "and has a key = ",
                    element.Key,
                    "This item will be stripped from your layout",
                    'Hint - try printing your layout and matching the IDs "print(layout)"',
                )
                continue
            element.Position = (CurrentRowNumber, i)
            element.ParentContainer = self
            CurrentRow.append(element)
            if element.Key is not None:
                self.UseDictionary = True
        # -------------------------  Append the row to list of Rows  ------------------------- #
        self.Rows.append(CurrentRow)

    def layout(self, rows):
        """
        Not user callable.  Use layout parameter instead. Creates the layout using the supplied rows of Elements

        :param rows: List[List[Element]] The list of rows
        :type rows:  List[List[Element]]
        :return:     (Tab) used for chaining
        :rtype:
        """

        for row in rows:
            try:
                iter(row)
            except TypeError:
                PopupError(
                    "Error creating Tab layout",
                    "Your row is not an iterable (e.g. a list)",
                    "Instead of a list, the type found was {}".format(type(row)),
                    "The offensive row = ",
                    row,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            self.AddRow(*row)
        return self

    def update(self, title=None, disabled=None, visible=None):
        """
        Changes some of the settings for the Tab Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param title:    tab title
        :type title:     (str)
        :param disabled: disable or enable state of the element
        :type disabled:  (bool)
        :param visible:  control visibility of element
        :type visible:   (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Tab.update - The window was closed")
            return

        state = "normal"
        if disabled is not None:
            self.Disabled = disabled
            if disabled:
                state = "disabled"
        if not visible:
            state = "hidden"
        if visible is not None:
            self._visible = visible

        self.ParentNotebook.tab(self.TabID, state=state)

        if title is not None:
            self.Title = str(title)
            self.ParentNotebook.tab(self.TabID, text=self.Title)
            # self.ParentNotebook.tab(self.ContainerElemementNumber-1, text=self.Title)

        # if visible is False:
        #     self.ParentNotebook.pack_forget()
        # elif visible is True:
        #     self.ParentNotebook.pack()
        return self

    def _GetElementAtLocation(self, location):
        """
        Not user callable. Used to find the Element at a row, col position within the layout

        :param location: (row, column) position of the element to find in layout
        :type location:  (int, int)
        :return:         The element found at the location
        :rtype:          (Element)
        """

        (row_num, col_num) = location
        row = self.Rows[row_num]
        element = row[col_num]
        return element

    def select(self):
        """
        Create a tkinter event that mimics user clicking on a tab. Must have called window.Finalize / Read first!

        """
        # Use a try in case the window has been destoyed
        try:
            self.ParentNotebook.select(self.TabID)
        except Exception as e:
            print("Exception Selecting Tab {}".format(e))

    AddRow = add_row
    Layout = layout
    Select = select
    Update = update


# ---------------------------------------------------------------------- #
#                           TabGroup                                     #
# ---------------------------------------------------------------------- #
class TabGroup(Element):
    """
    TabGroup Element groups together your tabs into the group of tabs you see displayed in your window
    """

    def __init__(
            self,
            layout,
            tab_location=None,
            title_color=None,
            tab_background_color=None,
            selected_title_color=None,
            selected_background_color=None,
            background_color=None,
            focus_color=None,
            font=None,
            change_submits=False,
            enable_events=False,
            pad=None,
            p=None,
            border_width=None,
            tab_border_width=None,
            theme=None,
            key=None,
            k=None,
            size=(None, None),
            s=(None, None),
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param layout:                    Layout of Tabs. Different than normal layouts. ALL Tabs should be on first row
        :type layout:                     List[List[Tab]]
        :param tab_location:              location that tabs will be displayed. Choices are left, right, top, bottom, lefttop, leftbottom, righttop, rightbottom, bottomleft, bottomright, topleft, topright
        :type tab_location:               (str)
        :param title_color:               color of text on tabs
        :type title_color:                (str)
        :param tab_background_color:      color of all tabs that are not selected
        :type tab_background_color:       (str)
        :param selected_title_color:      color of tab text when it is selected
        :type selected_title_color:       (str)
        :param selected_background_color: color of tab when it is selected
        :type selected_background_color:  (str)
        :param background_color:          color of background area that tabs are located on
        :type background_color:           (str)
        :param focus_color:               color of focus indicator on the tabs
        :type focus_color:                (str)
        :param font:                      specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                       (str or (str, int[, str]) or None)
        :param change_submits:            * DEPRICATED DO NOT USE. Use `enable_events` instead
        :type change_submits:             (bool)
        :param enable_events:             If True then switching tabs will generate an Event
        :type enable_events:              (bool)
        :param pad:                       Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                         Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param border_width:              width of border around element in pixels
        :type border_width:               (int)
        :param tab_border_width:          width of border around the tabs
        :type tab_border_width:           (int)
        :param theme:                     DEPRICATED - You can only specify themes using set options or when window is created. It's not possible to do it on an element basis
        :type theme:                      (enum)
        :param key:                       Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                        str | int | tuple | object
        :param k:                         Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                          str | int | tuple | object
        :param size:                      (width, height) w=pixels-wide, h=pixels-high. Either item in tuple can be None to indicate use the computed value and set only 1 direction
        :type size:                       (int|None, int|None)
        :param s:                         Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                          (int|None, int|None)
        :param tooltip:                   text, that will appear when mouse hovers over the element
        :type tooltip:                    (str)
        :param right_click_menu:          A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:           List[List[ List[str] | str ]]
        :param expand_x:                  If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                   (bool)
        :param expand_y:                  If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                   (bool)
        :param visible:                   DEPRECATED  - Should you need to control visiblity for the TabGroup as a whole, place it into a Column element
        :type visible:                    (bool)
        :param metadata:                  User metadata that can be set to ANYTHING
        :type metadata:                   (Any)
        """

        self.UseDictionary = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.ParentWindow = None
        self.SelectedTitleColor = (
            selected_title_color
            if selected_title_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["TEXT"]
        )
        self.SelectedBackgroundColor = (
            selected_background_color
            if selected_background_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["BACKGROUND"]
        )
        title_color = (
            title_color
            if title_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["TEXT_INPUT"]
        )
        self.TabBackgroundColor = (
            tab_background_color
            if tab_background_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["INPUT"]
        )
        self.Rows = []
        self.TKNotebook = None  # type: ttk.Notebook
        self.Widget = None  # type: ttk.Notebook
        self.tab_index_to_key = {}  # has a list of the tabs in the notebook and their associated key
        self.TabCount = 0
        self.BorderWidth = border_width
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.ChangeSubmits = change_submits or enable_events
        self.TabLocation = tab_location
        self.ElementJustification = "left"
        self.RightClickMenu = right_click_menu
        self.TabBorderWidth = tab_border_width
        self.FocusColor = focus_color

        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        self.Layout(layout)

        super().__init__(
            ELEM_TYPE_TAB_GROUP,
            size=sz,
            background_color=background_color,
            text_color=title_color,
            font=font,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def add_row(self, *args):
        """
        Not recommended user call.  Used to add rows of Elements to the Frame Element.

        """

        NumRows = len(self.Rows)  # number of existing rows is our row number
        CurrentRowNumber = NumRows  # this row's number
        CurrentRow = []  # start with a blank row and build up
        # -------------------------  Add the elements to a row  ------------------------- #
        for i, element in enumerate(
                args
        ):  # Loop through list of elements and add them to the row
            if type(element) == list:
                PopupError(
                    "Error creating Tab layout",
                    "Layout has a LIST instead of an ELEMENT",
                    "This sometimes means you have a badly placed ]",
                    "The offensive list is:",
                    element,
                    "This list will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            elif callable(element) and not isinstance(element, Element):
                PopupError(
                    "Error creating Tab layout",
                    "Layout has a FUNCTION instead of an ELEMENT",
                    "This likely means you are missing () from your layout",
                    "The offensive list is:",
                    element,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            if element.ParentContainer is not None:
                warnings.warn(
                    "*** YOU ARE ATTEMPTING TO REUSE AN ELEMENT IN YOUR LAYOUT! Once placed in a layout, an element cannot be used in another layout. ***",
                    UserWarning,
                )
                PopupError(
                    "Error creating Tab layout",
                    "The layout specified has already been used",
                    'You MUST start witha "clean", unused layout every time you create a window',
                    "The offensive Element = ",
                    element,
                    "and has a key = ",
                    element.Key,
                    "This item will be stripped from your layout",
                    'Hint - try printing your layout and matching the IDs "print(layout)"',
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            element.Position = (CurrentRowNumber, i)
            element.ParentContainer = self
            CurrentRow.append(element)
            if element.Key is not None:
                self.UseDictionary = True
        # -------------------------  Append the row to list of Rows  ------------------------- #
        self.Rows.append(CurrentRow)

    def layout(self, rows):
        """
        Can use like the Window.Layout method, but it's better to use the layout parameter when creating

        :param rows: The rows of Elements
        :type rows:  List[List[Element]]
        :return:     Used for chaining
        :rtype:      (Frame)
        """
        for row in rows:
            try:
                iter(row)
            except TypeError:
                PopupError(
                    "Error creating Tab layout",
                    "Your row is not an iterable (e.g. a list)",
                    "Instead of a list, the type found was {}".format(type(row)),
                    "The offensive row = ",
                    row,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            self.AddRow(*row)
        return self

    def _GetElementAtLocation(self, location):
        """
        Not user callable. Used to find the Element at a row, col position within the layout

        :param location: (row, column) position of the element to find in layout
        :type location:  (int, int)
        :return:         The element found at the location
        :rtype:          (Element)
        """

        (row_num, col_num) = location
        row = self.Rows[row_num]
        element = row[col_num]
        return element

    def find_key_from_tab_name(self, tab_name):
        """
        Searches through the layout to find the key that matches the text on the tab. Implies names should be unique

        :param tab_name: name of a tab
        :type tab_name:  str
        :return:         Returns the key or None if no key found
        :rtype:          key | None
        """
        for row in self.Rows:
            for element in row:
                if element.Title == tab_name:
                    return element.Key
        return None

    def find_currently_active_tab_key(self):
        """
        Returns the key for the currently active tab in this TabGroup
        :return:    Returns the key or None of no key found
        :rtype:     key | None
        """
        try:
            current_index = self.TKNotebook.index("current")
            key = self.tab_index_to_key.get(current_index, None)
        except Exception:
            key = None

        return key

    def get(self):
        """
        Returns the current value for the Tab Group, which will be the currently selected tab's KEY or the text on
        the tab if no key is defined.  Returns None if an error occurs.
        Note that this is exactly the same data that would be returned from a call to Window.read. Are you sure you
        are using this method correctly?

        :return: The key of the currently selected tab or None if there is an error
        :rtype:  Any | None
        """

        try:
            current_index = self.TKNotebook.index("current")
            key = self.tab_index_to_key.get(current_index, None)
        except Exception:
            key = None

        return key

    def add_tab(self, tab_element):
        """
        Add a new tab to an existing TabGroup
        This call was written so that tabs can be added at runtime as your user performs operations.
        Your Window should already be created and finalized.

        :param tab_element: A Tab Element that has a layout in it
        :type tab_element:  Tab
        """

        self.add_row(tab_element)
        tab_element.TKFrame = tab_element.Widget = tk.Frame(self.TKNotebook)
        form = self.ParentForm
        form._BuildKeyDictForWindow(form, tab_element, form.AllKeysDict)
        form.AllKeysDict[tab_element.Key] = tab_element
        # Pack the tab's layout into the tab. NOTE - This does NOT pack the Tab itself... for that see below...
        PackFormIntoFrame(tab_element, tab_element.TKFrame, self.ParentForm)

        # - This is below -    Perform the same operation that is performed when a Tab is packed into the window.
        # If there's an image in the tab, then do the imagey-stuff
        # ------------------- start of imagey-stuff -------------------
        try:
            if tab_element.Filename is not None:
                photo = tk.PhotoImage(file=tab_element.Filename)
            elif tab_element.Data is not None:
                photo = tk.PhotoImage(data=tab_element.Data)
            else:
                photo = None

            if tab_element.ImageSubsample and photo is not None:
                photo = photo.subsample(tab_element.ImageSubsample)
                # print('*ERROR laying out form.... Image Element has no image specified*')
        except Exception as e:
            photo = None
            _error_popup_with_traceback(
                "Your Window has an Tab Element with an IMAGE problem",
                "The traceback will show you the Window with the problem layout",
                "Look in this Window's layout for an Image tab_element that has a key of {}".format(
                    tab_element.Key
                ),
                "The error occuring is:",
                e,
            )

        tab_element.photo = photo
        # add the label
        if photo is not None:
            width, height = photo.width(), photo.height()
            tab_element.tktext_label = tk.Label(
                tab_element.ParentRowFrame,
                image=photo,
                width=width,
                height=height,
                bd=0,
            )
        else:
            tab_element.tktext_label = tk.Label(tab_element.ParentRowFrame, bd=0)
        # ------------------- end of imagey-stuff -------------------

        state = "normal"
        if tab_element.Disabled:
            state = "disabled"
        if not tab_element.visible:
            state = "hidden"
        if photo is not None:
            self.TKNotebook.add(
                tab_element.TKFrame,
                text=tab_element.Title,
                compound=tk.LEFT,
                state=state,
                image=photo,
            )
        else:
            self.TKNotebook.add(
                tab_element.TKFrame, text=tab_element.Title, state=state
            )
        tab_element.ParentNotebook = self.TKNotebook
        tab_element.TabID = self.TabCount
        tab_element.ParentForm = self.ParentForm
        self.TabCount += 1
        if (
                tab_element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                and tab_element.BackgroundColor is not None
        ):
            tab_element.TKFrame.configure(
                background=tab_element.BackgroundColor,
                highlightbackground=tab_element.BackgroundColor,
                highlightcolor=tab_element.BackgroundColor,
            )
        if tab_element.BorderWidth is not None:
            tab_element.TKFrame.configure(borderwidth=tab_element.BorderWidth)
        if tab_element.Tooltip is not None:
            tab_element.TooltipObject = ToolTip(
                tab_element.TKFrame,
                text=tab_element.Tooltip,
                timeout=DEFAULT_TOOLTIP_TIME,
            )
        _add_right_click_menu(tab_element, form)

    def update(self, visible=None):
        """
        Enables changing the visibility

        :param visible:  control visibility of element
        :type visible:   (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in TabGroup.update - The window was closed"
            )
            return

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible

    AddRow = add_row
    FindKeyFromTabName = find_key_from_tab_name
    Get = get
    Layout = layout


# ---------------------------------------------------------------------- #
#                           Slider                                       #
# ---------------------------------------------------------------------- #
class Slider(Element):
    """
    A slider, horizontal or vertical
    """

    def __init__(
            self,
            range=(None, None),
            default_value=None,
            resolution=None,
            tick_interval=None,
            orientation=None,
            disable_number_display=False,
            border_width=None,
            relief=None,
            change_submits=False,
            enable_events=False,
            disabled=False,
            size=(None, None),
            s=(None, None),
            font=None,
            background_color=None,
            text_color=None,
            trough_color=None,
            key=None,
            k=None,
            pad=None,
            p=None,
            expand_x=False,
            expand_y=False,
            tooltip=None,
            visible=True,
            metadata=None,
    ):
        """
        :param range:                  slider's range (min value, max value)
        :type range:                   (int, int) | Tuple[float, float]
        :param default_value:          starting value for the slider
        :type default_value:           int | float
        :param resolution:             the smallest amount the slider can be moved
        :type resolution:              int | float
        :param tick_interval:          how often a visible tick should be shown next to slider
        :type tick_interval:           int | float
        :param orientation:            'horizontal' or 'vertical' ('h' or 'v' also work)
        :type orientation:             (str)
        :param disable_number_display: if True no number will be displayed by the Slider Element
        :type disable_number_display:  (bool)
        :param border_width:           width of border around element in pixels
        :type border_width:            (int)
        :param relief:                 relief style. Use constants - RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID
        :type relief:                  str | None
        :param change_submits:         * DEPRICATED DO NOT USE. Use `enable_events` instead
        :type change_submits:          (bool)
        :param enable_events:          If True then moving the slider will generate an Event
        :type enable_events:           (bool)
        :param disabled:               set disable state for element
        :type disabled:                (bool)
        :param size:                   (l=length chars/rows, w=width pixels)
        :type size:                    (int, int)
        :param s:                      Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                       (int, int)  | (None, None)
        :param font:                   specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                    (str or (str, int[, str]) or None)
        :param background_color:       color of slider's background
        :type background_color:        (str)
        :param text_color:             color of the slider's text
        :type text_color:              (str)
        :param trough_color:           color of the slider's trough
        :type trough_color:            (str)
        :param key:                    Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                     str | int | tuple | object
        :param k:                      Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                       str | int | tuple | object
        :param pad:                    Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                     (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                      Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                       (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param expand_x:               If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                (bool)
        :param expand_y:               If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                (bool)
        :param tooltip:                text, that will appear when mouse hovers over the element
        :type tooltip:                 (str)
        :param visible:                set visibility state of the element
        :type visible:                 (bool)
        :param metadata:               User metadata that can be set to ANYTHING
        :type metadata:                (Any)
        """

        self.TKScale = self.Widget = None  # type: tk.Scale
        self.Range = (1, 10) if range == (None, None) else range
        self.DefaultValue = self.Range[0] if default_value is None else default_value
        self.Orientation = orientation if orientation else DEFAULT_SLIDER_ORIENTATION
        self.BorderWidth = border_width if border_width else DEFAULT_SLIDER_BORDER_WIDTH
        self.Relief = relief if relief else DEFAULT_SLIDER_RELIEF
        self.Resolution = 1 if resolution is None else resolution
        self.ChangeSubmits = change_submits or enable_events
        self.Disabled = disabled
        self.TickInterval = tick_interval
        self.DisableNumericDisplay = disable_number_display
        self.TroughColor = trough_color or DEFAULT_SCROLLBAR_COLOR
        sz = size if size != (None, None) else s
        temp_size = sz
        if temp_size == (None, None):
            temp_size = (20, 20) if self.Orientation.startswith("h") else (8, 20)
        key = key if key is not None else k
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_INPUT_SLIDER,
            size=temp_size,
            font=font,
            background_color=background_color,
            text_color=text_color,
            key=key,
            pad=pad,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(self, value=None, range=(None, None), disabled=None, visible=None):
        """
        Changes some of the settings for the Slider Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param value:    sets current slider value
        :type value:     int | float
        :param range:    Sets a new range for slider
        :type range:     (int, int) | Tuple[float, float
        :param disabled: disable or enable state of the element
        :type disabled:  (bool)
        :param visible:  control visibility of element
        :type visible:   (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Slider.update - The window was closed"
            )
            return

        if range != (None, None):
            self.TKScale.config(from_=range[0], to_=range[1])
        if value is not None:
            try:
                self.TKIntVar.set(value)
            except Exception:
                pass
            self.DefaultValue = value
        if disabled:
            self.TKScale["state"] = "disabled"
        elif not disabled:
            self.TKScale["state"] = "normal"
        self.Disabled = disabled if disabled is not None else self.Disabled

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible

    def _SliderChangedHandler(self, event):
        """
        Not user callable.  Callback function for when slider is moved.

        :param event: (event) the event data provided by tkinter. Unknown format. Not used.
        :type event:
        """

        if self.Key is not None:
            self.ParentForm.LastButtonClicked = self.Key
        else:
            self.ParentForm.LastButtonClicked = ""
        self.ParentForm.FormRemainedOpen = True
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     self.ParentForm.TKroot.quit()  # kick the users out of the mainloop
        _exit_mainloop(self.ParentForm)

    Update = update


Sl = Slider


# ---------------------------------------------------------------------- #
#                          TkFixedFrame (Used by Column)                 #
# ---------------------------------------------------------------------- #
class TkFixedFrame(tk.Frame):
    """
    A tkinter frame that is used with Column Elements that do not have a scrollbar
    """

    def __init__(self, master, **kwargs):
        """
        :param master:   The parent widget
        :type master:    (tk.Widget)
        :param **kwargs: The keyword args
        :type **kwargs:
        """
        tk.Frame.__init__(self, master, **kwargs)

        self.canvas = tk.Canvas(self)

        self.canvas.pack(side="left", fill="both", expand=True)

        # reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.TKFrame = tk.Frame(self.canvas, **kwargs)
        self.frame_id = self.canvas.create_window(
            0, 0, window=self.TKFrame, anchor="nw"
        )
        self.canvas.config(borderwidth=0, highlightthickness=0)
        self.TKFrame.config(borderwidth=0, highlightthickness=0)
        self.config(borderwidth=0, highlightthickness=0)


# ---------------------------------------------------------------------- #
#                          TkScrollableFrame (Used by Column)            #
# ---------------------------------------------------------------------- #
class TkScrollableFrame(tk.Frame):
    """
    A frame with one or two scrollbars.  Used to make Columns with scrollbars
    """

    def __init__(self, master, vertical_only, element, window, **kwargs):
        """
        :param master:        The parent widget
        :type master:         (tk.Widget)
        :param vertical_only: if True the only a vertical scrollbar will be shown
        :type vertical_only:  (bool)
        :param element:       The element containing this object
        :type element:        (Column)
        """
        tk.Frame.__init__(self, master, **kwargs)
        # create a canvas object and a vertical scrollbar for scrolling it

        self.canvas = tk.Canvas(self)
        element.Widget = self.canvas
        # Okay, we're gonna make a list. Containing the y-min, x-min, y-max, and x-max of the frame
        element.element_frame = self
        _make_ttk_scrollbar(element, "v", window)
        # element.vsb = tk.Scrollbar(self, orient=tk.VERTICAL)
        element.vsb.pack(side="right", fill="y", expand="false")

        if not vertical_only:
            _make_ttk_scrollbar(element, "h", window)
            # self.hscrollbar = tk.Scrollbar(self, orient=tk.HORIZONTAL)
            element.hsb.pack(side="bottom", fill="x", expand="false")
            self.canvas.config(xscrollcommand=element.hsb.set)
            # self.canvas = tk.Canvas(self, )
        # else:
        #     self.canvas = tk.Canvas(self)

        self.canvas.config(yscrollcommand=element.vsb.set)
        self.canvas.pack(side="left", fill="both", expand=True)
        element.vsb.config(command=self.canvas.yview)
        if not vertical_only:
            element.hsb.config(command=self.canvas.xview)

        # reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.TKFrame = tk.Frame(self.canvas, **kwargs)
        self.frame_id = self.canvas.create_window(
            0, 0, window=self.TKFrame, anchor="nw"
        )
        self.canvas.config(borderwidth=0, highlightthickness=0)
        self.TKFrame.config(borderwidth=0, highlightthickness=0)
        self.config(borderwidth=0, highlightthickness=0)

        # Canvas can be: master, canvas, TKFrame

        # Chr0nic

        # self.unhookMouseWheel(None)
        # self.TKFrame.bind("<Enter>", self.hookMouseWheel)
        # self.TKFrame.bind("<Leave>", self.unhookMouseWheel)
        # self.bind('<Configure>', self.set_scrollregion)

        self.unhookMouseWheel(None)
        self.canvas.bind("<Enter>", self.hookMouseWheel)
        self.canvas.bind("<Leave>", self.unhookMouseWheel)
        self.bind("<Configure>", self.set_scrollregion)

    # Chr0nic
    def hookMouseWheel(self, e):
        # print("enter")
        VarHolder.canvas_holder = self.canvas
        self.canvas.bind_all("<4>", self.yscroll, add="+")
        self.canvas.bind_all("<5>", self.yscroll, add="+")
        self.canvas.bind_all("<MouseWheel>", self.yscroll, add="+")
        self.canvas.bind_all("<Shift-MouseWheel>", self.xscroll, add="+")

    # Chr0nic
    def unhookMouseWheel(self, e):
        # print("leave")
        VarHolder.canvas_holder = None
        self.canvas.unbind_all("<4>")
        self.canvas.unbind_all("<5>")
        self.canvas.unbind_all("<MouseWheel>")
        self.canvas.unbind_all("<Shift-MouseWheel>")

    def resize_frame(self, e):
        self.canvas.itemconfig(self.frame_id, height=e.height, width=e.width)

    def yscroll(self, event):
        if self.canvas.yview() == (0.0, 1.0):
            return
        if event.num == 5 or event.delta < 0:
            self.canvas.yview_scroll(1, "unit")
        elif event.num == 4 or event.delta > 0:
            self.canvas.yview_scroll(-1, "unit")

    def xscroll(self, event):
        if event.num == 5 or event.delta < 0:
            self.canvas.xview_scroll(1, "unit")
        elif event.num == 4 or event.delta > 0:
            self.canvas.xview_scroll(-1, "unit")

    def bind_mouse_scroll(self, parent, mode):
        # ~~ Windows only
        parent.bind("<MouseWheel>", mode)
        # ~~ Unix only
        parent.bind("<Button-4>", mode)
        parent.bind("<Button-5>", mode)

    def set_scrollregion(self, event=None):
        """Set the scroll region on the canvas"""
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))


# ---------------------------------------------------------------------- #
#                           Column                                       #
# ---------------------------------------------------------------------- #
class Column(Element):
    """
    A container element that is used to create a layout within your window's layout
    """

    def __init__(
            self,
            layout,
            background_color=None,
            size=(None, None),
            s=(None, None),
            size_subsample_width=1,
            size_subsample_height=2,
            pad=None,
            p=None,
            scrollable=False,
            vertical_scroll_only=False,
            right_click_menu=None,
            key=None,
            k=None,
            visible=True,
            justification=None,
            element_justification=None,
            vertical_alignment=None,
            grab=None,
            expand_x=None,
            expand_y=None,
            metadata=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
    ):
        """
        :param layout:                      Layout that will be shown in the Column container
        :type layout:                       List[List[Element]]
        :param background_color:            color of background of entire Column
        :type background_color:             (str)
        :param size:                        (width, height) size in pixels (doesn't work quite right, sometimes only 1 dimension is set by tkinter. Use a Sizer Element to help set sizes
        :type size:                         (int | None, int | None)
        :param s:                           Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                            (int | None, int | None)
        :param size_subsample_width:        Determines the size of a scrollable column width based on 1/size_subsample * required size. 1 = match the contents exactly, 2 = 1/2 contents size, 3 = 1/3. Can be a fraction to make larger than required.
        :type size_subsample_width:         (float)
        :param size_subsample_height:       Determines the size of a scrollable height based on 1/size_subsample * required size. 1 = match the contents exactly, 2 = 1/2 contents size, 3 = 1/3. Can be a fraction to make larger than required..
        :type size_subsample_height:        (float)
        :param pad:                         Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                           Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                            (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param scrollable:                  if True then scrollbars will be added to the column. If you update the contents of a scrollable column, be sure and call Column.contents_changed also
        :type scrollable:                   (bool)
        :param vertical_scroll_only:        if True then no horizontal scrollbar will be shown if a scrollable column
        :type vertical_scroll_only:         (bool)
        :param right_click_menu:            A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:             List[List[ List[str] | str ]]
        :param key:                         Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                          str | int | tuple | object
        :param k:                           Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                            str | int | tuple | object
        :param visible:                     set visibility state of the element
        :type visible:                      (bool)
        :param justification:               set justification for the Column itself. Note entire row containing the Column will be affected
        :type justification:                (str)
        :param element_justification:       All elements inside the Column will have this justification 'left', 'right', 'center' are valid values
        :type element_justification:        (str)
        :param vertical_alignment:          Place the column at the 'top', 'center', 'bottom' of the row (can also use t,c,r). Defaults to no setting (tkinter decides)
        :type vertical_alignment:           (str)
        :param grab:                        If True can grab this element and move the window around. Default is False
        :type grab:                         (bool)
        :param expand_x:                    If True the column will automatically expand in the X direction to fill available space
        :type expand_x:                     (bool)
        :param expand_y:                    If True the column will automatically expand in the Y direction to fill available space
        :type expand_y:                     (bool)
        :param metadata:                    User metadata that can be set to ANYTHING
        :type metadata:                     (Any)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        """

        self.UseDictionary = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.ParentWindow = None
        self.ParentPanedWindow = None
        self.Rows = []
        self.TKFrame = None
        self.TKColFrame = None  # type: tk.Frame
        self.Scrollable = scrollable
        self.VerticalScrollOnly = vertical_scroll_only

        self.RightClickMenu = right_click_menu
        bg = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.ContainerElemementNumber = Window._GetAContainerNumber()
        self.ElementJustification = element_justification
        self.Justification = justification
        self.VerticalAlignment = vertical_alignment
        key = key if key is not None else k
        self.Grab = grab
        self.expand_x = expand_x
        self.expand_y = expand_y
        self.Layout(layout)
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.size_subsample_width = size_subsample_width
        self.size_subsample_height = size_subsample_height

        super().__init__(
            ELEM_TYPE_COLUMN,
            background_color=bg,
            size=sz,
            pad=pad,
            key=key,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )
        return

    def add_row(self, *args):
        """
        Not recommended user call.  Used to add rows of Elements to the Column Element.

        """

        NumRows = len(self.Rows)  # number of existing rows is our row number
        CurrentRowNumber = NumRows  # this row's number
        CurrentRow = []  # start with a blank row and build up
        # -------------------------  Add the elements to a row  ------------------------- #
        for i, element in enumerate(
                args
        ):  # Loop through list of elements and add them to the row
            if type(element) == list:
                PopupError(
                    "Error creating Column layout",
                    "Layout has a LIST instead of an ELEMENT",
                    "This sometimes means you have a badly placed ]",
                    "The offensive list is:",
                    element,
                    "This list will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            elif callable(element) and not isinstance(element, Element):
                PopupError(
                    "Error creating Column layout",
                    "Layout has a FUNCTION instead of an ELEMENT",
                    "This likely means you are missing () from your layout",
                    "The offensive list is:",
                    element,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            if element.ParentContainer is not None:
                warnings.warn(
                    "*** YOU ARE ATTEMPTING TO REUSE AN ELEMENT IN YOUR LAYOUT! Once placed in a layout, an element cannot be used in another layout. ***",
                    UserWarning,
                )
                PopupError(
                    "Error creating Column layout",
                    "The layout specified has already been used",
                    'You MUST start witha "clean", unused layout every time you create a window',
                    "The offensive Element = ",
                    element,
                    "and has a key = ",
                    element.Key,
                    "This item will be stripped from your layout",
                    'Hint - try printing your layout and matching the IDs "print(layout)"',
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            element.Position = (CurrentRowNumber, i)
            element.ParentContainer = self
            CurrentRow.append(element)
            if element.Key is not None:
                self.UseDictionary = True
        # -------------------------  Append the row to list of Rows  ------------------------- #
        self.Rows.append(CurrentRow)

    def layout(self, rows):
        """
        Can use like the Window.Layout method, but it's better to use the layout parameter when creating

        :param rows: The rows of Elements
        :type rows:  List[List[Element]]
        :return:     Used for chaining
        :rtype:      (Column)
        """

        for row in rows:
            try:
                iter(row)
            except TypeError:
                PopupError(
                    "Error creating Column layout",
                    "Your row is not an iterable (e.g. a list)",
                    "Instead of a list, the type found was {}".format(type(row)),
                    "The offensive row = ",
                    row,
                    "This item will be stripped from your layout",
                    keep_on_top=True,
                    image=_random_error_emoji(),
                )
                continue
            self.AddRow(*row)
        return self

    def _GetElementAtLocation(self, location):
        """
        Not user callable. Used to find the Element at a row, col position within the layout

        :param location:     (row, column) position of the element to find in layout
        :type  location:     (int, int)
        :return:             The element found at the location
        :rtype:              (Element)
        """

        (row_num, col_num) = location
        row = self.Rows[row_num]
        element = row[col_num]
        return element

    def update(self, visible=None):
        """
        Changes some of the settings for the Column Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param visible: control visibility of element
        :type visible:  (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback(
                "Error in Column.update - The window was closed"
            )
            return

        if self.expand_x and self.expand_y:
            expand = tk.BOTH
        elif self.expand_x:
            expand = tk.X
        elif self.expand_y:
            expand = tk.Y
        else:
            expand = None

        if not visible:
            if self.TKColFrame:
                self._pack_forget_save_settings()
                # self.TKColFrame.pack_forget()
            if self.ParentPanedWindow:
                self.ParentPanedWindow.remove(self.TKColFrame)
        elif visible:
            if self.TKColFrame:
                self._pack_restore_settings()
                # self.TKColFrame.pack(padx=self.pad_used[0], pady=self.pad_used[1], fill=expand)
            if self.ParentPanedWindow:
                self.ParentPanedWindow.add(self.TKColFrame)
        if visible is not None:
            self._visible = visible

    def contents_changed(self):
        """
        When a scrollable column has part of its layout changed by making elements visible or invisible or the
        layout is extended for the Column, then this method needs to be called so that the new scroll area
        is computed to match the new contents.
        """
        self.TKColFrame.canvas.config(scrollregion=self.TKColFrame.canvas.bbox("all"))

    AddRow = add_row
    Layout = layout
    Update = update


Col = Column


# ---------------------------------------------------------------------- #
#                           Pane                                         #
# ---------------------------------------------------------------------- #
class Pane(Element):
    """
    A sliding Pane that is unique to tkinter.  Uses Columns to create individual panes
    """

    def __init__(
            self,
            pane_list,
            background_color=None,
            size=(None, None),
            s=(None, None),
            pad=None,
            p=None,
            orientation="vertical",
            show_handle=True,
            relief=RELIEF_RAISED,
            handle_size=None,
            border_width=None,
            key=None,
            k=None,
            expand_x=None,
            expand_y=None,
            visible=True,
            metadata=None,
    ):
        """
        :param pane_list:        Must be a list of Column Elements. Each Column supplied becomes one pane that's shown
        :type pane_list:         List[Column] | Tuple[Column]
        :param background_color: color of background
        :type background_color:  (str)
        :param size:             (width, height) w=characters-wide, h=rows-high How much room to reserve for the Pane
        :type size:              (int, int)
        :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                 (int, int)  | (None, None)
        :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param orientation:      'horizontal' or 'vertical' or ('h' or 'v'). Direction the Pane should slide
        :type orientation:       (str)
        :param show_handle:      if True, the handle is drawn that makes it easier to grab and slide
        :type show_handle:       (bool)
        :param relief:           relief style. Values are same as other elements that use relief values. RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID
        :type relief:            (enum)
        :param handle_size:      Size of the handle in pixels
        :type handle_size:       (int)
        :param border_width:     width of border around element in pixels
        :type border_width:      (int)
        :param key:              Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:               str | int | tuple | object
        :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                 str | int | tuple | object
        :param expand_x:         If True the column will automatically expand in the X direction to fill available space
        :type expand_x:          (bool)
        :param expand_y:         If True the column will automatically expand in the Y direction to fill available space
        :type expand_y:          (bool)
        :param visible:          set visibility state of the element
        :type visible:           (bool)
        :param metadata:         User metadata that can be set to ANYTHING
        :type metadata:          (Any)
        """

        self.UseDictionary = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.ParentWindow = None
        self.Rows = []
        self.TKFrame = None
        self.PanedWindow = None
        self.Orientation = orientation
        self.PaneList = pane_list
        self.ShowHandle = show_handle
        self.Relief = relief
        self.HandleSize = handle_size or 8
        self.BorderDepth = border_width
        bg = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )

        self.Rows = [pane_list]
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_PANE,
            background_color=bg,
            size=sz,
            pad=pad,
            key=key,
            visible=visible,
            metadata=metadata,
        )
        return

    def update(self, visible=None):
        """
        Changes some of the settings for the Pane Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param visible: control visibility of element
        :type visible:  (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Pane.update - The window was closed")
            return

        if not visible:
            self._pack_forget_save_settings()
        elif visible:
            self._pack_restore_settings()

        if visible is not None:
            self._visible = visible

    Update = update


# ---------------------------------------------------------------------- #
#                           Calendar                                     #
# ---------------------------------------------------------------------- #
class TKCalendar(ttk.Frame):
    """
    This code was shamelessly lifted from moshekaplan's repository - moshekaplan/tkinter_components
    NONE of this code is user callable.  Stay away!
    """

    # XXX ToDo: cget and configure

    datetime = calendar.datetime.datetime
    timedelta = calendar.datetime.timedelta

    def __init__(
            self,
            master=None,
            target_element=None,
            close_when_chosen=True,
            default_date=(None, None, None),
            **kw,
    ):
        """WIDGET-SPECIFIC OPTIONS: locale, firstweekday, year, month, selectbackground, selectforeground"""
        self._TargetElement = target_element
        default_mon, default_day, default_year = default_date
        # remove custom options from kw before initializating ttk.Frame
        fwday = kw.pop("firstweekday", calendar.MONDAY)
        year = kw.pop("year", default_year or self.datetime.now().year)
        month = kw.pop("month", default_mon or self.datetime.now().month)
        locale = kw.pop("locale", None)
        sel_bg = kw.pop("selectbackground", "#ecffc4")
        sel_fg = kw.pop("selectforeground", "#05640e")
        self.format = kw.pop("format")
        if self.format is None:
            self.format = "%Y-%m-%d %H:%M:%S"

        self._date = self.datetime(year, month, default_day or 1)
        self._selection = None  # no date selected
        self._master = master
        self.close_when_chosen = close_when_chosen
        ttk.Frame.__init__(self, master, **kw)

        # instantiate proper calendar class
        if locale is None:
            self._cal = calendar.TextCalendar(fwday)
        else:
            self._cal = calendar.LocaleTextCalendar(fwday, locale)

        self.__setup_styles()  # creates custom styles
        self.__place_widgets()  # pack/grid used widgets
        self.__config_calendar()  # adjust calendar columns and setup tags
        # configure a canvas, and proper bindings, for selecting dates
        self.__setup_selection(sel_bg, sel_fg)

        # store items ids, used for insertion later
        self._items = [self._calendar.insert("", "end", values="") for _ in range(6)]
        # insert dates in the currently empty calendar
        self._build_calendar()

    def __setitem__(self, item, value):
        if item in ("year", "month"):
            raise AttributeError("attribute '%s' is not writeable" % item)
        elif item == "selectbackground":
            self._canvas["background"] = value
        elif item == "selectforeground":
            self._canvas.itemconfigure(self._canvas.text, item=value)
        else:
            ttk.Frame.__setitem__(self, item, value)

    def __getitem__(self, item):
        if item in ("year", "month"):
            return getattr(self._date, item)
        elif item == "selectbackground":
            return self._canvas["background"]
        elif item == "selectforeground":
            return self._canvas.itemcget(self._canvas.text, "fill")
        else:
            r = ttk.tclobjs_to_py({item: ttk.Frame.__getitem__(self, item)})
            return r[item]

    def __setup_styles(self):
        # custom ttk styles
        style = ttk.Style(self.master)
        arrow_layout = lambda dir: (
            [("Button.focus", {"children": [("Button.%sarrow" % dir, None)]})]
        )
        style.layout("L.TButton", arrow_layout("left"))
        style.layout("R.TButton", arrow_layout("right"))

    def __place_widgets(self):
        # header frame and its widgets
        hframe = ttk.Frame(self)
        lbtn = ttk.Button(hframe, style="L.TButton", command=self._prev_month)
        rbtn = ttk.Button(hframe, style="R.TButton", command=self._next_month)
        self._header = ttk.Label(hframe, width=15, anchor="center")
        # the calendar
        self._calendar = ttk.Treeview(self, show="", selectmode="none", height=7)

        # pack the widgets
        hframe.pack(in_=self, side="top", pady=4, anchor="center")
        lbtn.grid(in_=hframe)
        self._header.grid(in_=hframe, column=1, row=0, padx=12)
        rbtn.grid(in_=hframe, column=2, row=0)
        self._calendar.pack(in_=self, expand=1, fill="both", side="bottom")

    def __config_calendar(self):
        cols = self._cal.formatweekheader(3).split()
        self._calendar["columns"] = cols
        self._calendar.tag_configure("header", background="grey90")
        self._calendar.insert("", "end", values=cols, tag="header")
        # adjust its columns width
        font = tkinter.font.Font()
        maxwidth = max(font.measure(col) for col in cols)
        for col in cols:
            self._calendar.column(col, width=maxwidth, minwidth=maxwidth, anchor="e")

    def __setup_selection(self, sel_bg, sel_fg):
        self._font = tkinter.font.Font()
        self._canvas = canvas = tk.Canvas(
            self._calendar, background=sel_bg, borderwidth=0, highlightthickness=0
        )
        canvas.text = canvas.create_text(0, 0, fill=sel_fg, anchor="w")

        canvas.bind("<ButtonPress-1>", lambda evt: canvas.place_forget())
        self._calendar.bind("<Configure>", lambda evt: canvas.place_forget())
        self._calendar.bind("<ButtonPress-1>", self._pressed)

    def __minsize(self, evt):
        width, height = self._calendar.master.geometry().split("x")
        height = height[: height.index("+")]
        self._calendar.master.minsize(width, height)

    def _build_calendar(self):
        year, month = self._date.year, self._date.month

        # update header text (Month, YEAR)
        header = self._cal.formatmonthname(year, month, 0)
        self._header["text"] = header.title()

        # update calendar shown dates
        cal = self._cal.monthdayscalendar(year, month)
        for indx, item in enumerate(self._items):
            week = cal[indx] if indx < len(cal) else []
            fmt_week = [("%02d" % day) if day else "" for day in week]
            self._calendar.item(item, values=fmt_week)

    def _show_selection(self, text, bbox):
        """Configure canvas for a new selection."""
        x, y, width, height = bbox

        textw = self._font.measure(text)

        canvas = self._canvas
        canvas.configure(width=width, height=height)
        canvas.coords(canvas.text, width - textw, height / 2 - 1)
        canvas.itemconfigure(canvas.text, text=text)
        canvas.place(in_=self._calendar, x=x, y=y)

    # Callbacks

    def _pressed(self, evt):
        """Clicked somewhere in the calendar."""
        x, y, widget = evt.x, evt.y, evt.widget
        item = widget.identify_row(y)
        column = widget.identify_column(x)

        if not column or not item in self._items:
            # clicked in the weekdays row or just outside the columns
            return

        item_values = widget.item(item)["values"]
        if not len(item_values):  # row is empty for this month
            return

        text = item_values[int(column[1]) - 1]
        if not text:  # date is empty
            return

        bbox = widget.bbox(item, column)
        if not bbox:  # calendar not visible yet
            return

        # update and then show selection
        text = "%02d" % text
        self._selection = (text, item, column)
        self._show_selection(text, bbox)
        year, month = self._date.year, self._date.month
        now = self.datetime.now()
        try:
            self._TargetElement.Update(
                self.datetime(
                    year,
                    month,
                    int(self._selection[0]),
                    now.hour,
                    now.minute,
                    now.second,
                ).strftime(self.format)
            )
            if self._TargetElement.ChangeSubmits:
                self._TargetElement.ParentForm.LastButtonClicked = (
                    self._TargetElement.Key
                )
                self._TargetElement.ParentForm.FormRemainedOpen = True
                self._TargetElement.ParentForm.TKroot.quit()  # kick the users out of the mainloop
        except Exception:
            pass
        if self.close_when_chosen:
            self._master.destroy()

    def _prev_month(self):
        """Updated calendar to show the previous month."""
        self._canvas.place_forget()

        self._date = self._date - self.timedelta(days=1)
        self._date = self.datetime(self._date.year, self._date.month, 1)
        self._build_calendar()  # reconstuct calendar

    def _next_month(self):
        """Update calendar to show the next month."""
        self._canvas.place_forget()

        year, month = self._date.year, self._date.month
        self._date = self._date + self.timedelta(
            days=calendar.monthrange(year, month)[1] + 1
        )
        self._date = self.datetime(self._date.year, self._date.month, 1)
        self._build_calendar()  # reconstruct calendar

    # Properties

    @property
    def selection(self):
        if not self._selection:
            return None

        year, month = self._date.year, self._date.month
        return self.datetime(year, month, int(self._selection[0]))


# ---------------------------------------------------------------------- #
#                           Menu                                         #
# ---------------------------------------------------------------------- #
class Menu(Element):
    """
    Menu Element is the Element that provides a Menu Bar that goes across the top of the window, just below titlebar.
    Here is an example layout.  The "&" are shortcut keys ALT+key.
    Is a List of -  "Item String" + List
    Where Item String is what will be displayed on the Menubar itself.
    The List that follows the item represents the items that are shown then Menu item is clicked
    Notice how an "entry" in a mennu can be a list which means it branches out and shows another menu, etc. (resursive)
    menu_def = [['&File', ['!&Open', '&Save::savekey', '---', '&Properties', 'E&xit']],
                ['!&Edit', ['!&Paste', ['Special', 'Normal', ], 'Undo'], ],
                ['&Debugger', ['Popout', 'Launch Debugger']],
                ['&Toolbar', ['Command &1', 'Command &2', 'Command &3', 'Command &4']],
                ['&Help', '&About...'], ]
    Important Note!  The colors, font, look of the Menubar itself cannot be changed, only the menus shown AFTER clicking the menubar
    can be changed.  If you want to change the style/colors the Menubar, then you will have to use the MenubarCustom element.
    Finally, "keys" can be added to entries so make them unique.  The "Save" entry has a key associated with it. You
    can see it has a "::" which signifies the beginning of a key.  The user will not see the key portion when the
    menu is shown.  The key portion is returned as part of the event.
    """

    def __init__(
            self,
            menu_definition,
            background_color=None,
            text_color=None,
            disabled_text_color=None,
            size=(None, None),
            s=(None, None),
            tearoff=False,
            font=None,
            pad=None,
            p=None,
            key=None,
            k=None,
            visible=True,
            metadata=None,
    ):
        """
        :param menu_definition:           The Menu definition specified using lists (docs explain the format)
        :type menu_definition:            List[List[Tuple[str, List[str]]]
        :param background_color:          color of the background of menus, NOT the Menubar
        :type background_color:           (str)
        :param text_color:                text color for menus, NOT the Menubar. Can be in #RRGGBB format or a color name "black".
        :type text_color:                 (str)
        :param disabled_text_color:       color to use for text when item in submenu, not the menubar itself, is disabled. Can be in #RRGGBB format or a color name "black"
        :type disabled_text_color:        (str)
        :param size:                      Not used in the tkinter port
        :type size:                       (int, int)
        :param s:                         Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
        :type s:                          (int, int)  | (None, None)
        :param tearoff:                   if True, then can tear the menu off from the window ans use as a floating window. Very cool effect
        :type tearoff:                    (bool)
        :param pad:                       Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                         Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                          (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param font:                      specifies the  font family, size, etc. of submenus. Does NOT apply to the Menubar itself. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                       (str or (str, int[, str]) or None)
        :param key:                       Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
        :type key:                        str | int | tuple | object
        :param k:                         Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                          str | int | tuple | object
        :param visible:                   set visibility state of the element
        :type visible:                    (bool)
        :param metadata:                  User metadata that can be set to ANYTHING
        :type metadata:                   (Any)
        """

        self.BackgroundColor = (
            background_color
            if background_color is not None
            else theme_input_background_color()
        )
        self.TextColor = (
            text_color if text_color is not None else theme_input_text_color()
        )

        self.DisabledTextColor = (
            disabled_text_color
            if disabled_text_color is not None
            else COLOR_SYSTEM_DEFAULT
        )
        self.MenuDefinition = copy.deepcopy(menu_definition)
        self.Widget = self.TKMenu = None  # type: tk.Menu
        self.MenuItemChosen = None
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p

        super().__init__(
            ELEM_TYPE_MENUBAR,
            background_color=self.BackgroundColor,
            text_color=self.TextColor,
            size=sz,
            pad=pad,
            key=key,
            visible=visible,
            font=font,
            metadata=metadata,
        )
        # super().__init__(ELEM_TYPE_MENUBAR, background_color=COLOR_SYSTEM_DEFAULT, text_color=COLOR_SYSTEM_DEFAULT, size=sz, pad=pad, key=key, visible=visible, font=None, metadata=metadata)

        self.Tearoff = tearoff

        return

    def _MenuItemChosenCallback(self, item_chosen):  # Menu Menu Item Chosen Callback
        """
        Not user callable.  Called when some end-point on the menu (an item) has been clicked.  Send the information back to the application as an event.  Before event can be sent

        :param item_chosen: the text that was clicked on / chosen from the menu
        :type item_chosen:  (str)
        """
        # print('IN MENU ITEM CALLBACK', item_chosen)
        self.MenuItemChosen = item_chosen
        self.ParentForm.LastButtonClicked = item_chosen
        self.ParentForm.FormRemainedOpen = True
        # if self.ParentForm.CurrentlyRunningMainloop:
        #     self.ParentForm.TKroot.quit()  # kick the users out of the mainloop
        _exit_mainloop(self.ParentForm)

    def update(self, menu_definition=None, visible=None):
        """
        Update a menubar - can change the menu definition and visibility.  The entire menu has to be specified

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param menu_definition: The menu definition list
        :type menu_definition:  List[List[Tuple[str, List[str]]]
        :param visible:         control visibility of element
        :type visible:          (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Menu.update - The window was closed")
            return

        if menu_definition is not None:
            self.MenuDefinition = copy.deepcopy(menu_definition)
            if self.TKMenu is None:  # if no menu exists, make one
                self.TKMenu = tk.Menu(
                    self.ParentForm.TKroot,
                    tearoff=self.Tearoff,
                    tearoffcommand=self._tearoff_menu_callback,
                )  # create the menubar
            menubar = self.TKMenu
            # Delete all the menu items (assuming 10000 should be a high enough number to cover them all)
            menubar.delete(0, 10000)
            self.Widget = (
                self.TKMenu
            )  # same the new menu so user can access to extend PySimpleGUI
            for menu_entry in self.MenuDefinition:
                baritem = tk.Menu(
                    menubar,
                    tearoff=self.Tearoff,
                    tearoffcommand=self._tearoff_menu_callback,
                )

                if self.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                    baritem.config(bg=self.BackgroundColor)
                if self.TextColor not in (COLOR_SYSTEM_DEFAULT, None):
                    baritem.config(fg=self.TextColor)
                if self.DisabledTextColor not in (COLOR_SYSTEM_DEFAULT, None):
                    baritem.config(disabledforeground=self.DisabledTextColor)
                if self.Font is not None:
                    baritem.config(font=self.Font)

                if self.Font is not None:
                    baritem.config(font=self.Font)
                pos = menu_entry[0].find(MENU_SHORTCUT_CHARACTER)
                # print(pos)
                if pos != -1:
                    if (
                            pos == 0
                            or menu_entry[0][pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                    ):
                        menu_entry[0] = (
                                menu_entry[0][:pos]
                                + menu_entry[0][pos + len(MENU_SHORTCUT_CHARACTER):]
                        )
                if menu_entry[0][0] == MENU_DISABLED_CHARACTER:
                    menubar.add_cascade(
                        label=menu_entry[0][len(MENU_DISABLED_CHARACTER):],
                        menu=baritem,
                        underline=pos,
                    )
                    menubar.entryconfig(
                        menu_entry[0][len(MENU_DISABLED_CHARACTER):], state="disabled"
                    )
                else:
                    menubar.add_cascade(
                        label=menu_entry[0], menu=baritem, underline=pos
                    )

                if len(menu_entry) > 1:
                    AddMenuItem(baritem, menu_entry[1], self)

        if not visible:
            self.ParentForm.TKroot.configure(
                menu=[]
            )  # this will cause the menubar to disappear
        elif self.TKMenu is not None:
            self.ParentForm.TKroot.configure(menu=self.TKMenu)
        if visible is not None:
            self._visible = visible

    Update = update


MenuBar = Menu  # another name for Menu to make it clear it's the Menu Bar
Menubar = Menu  # another name for Menu to make it clear it's the Menu Bar


# ---------------------------------------------------------------------- #
#                           Table                                        #
# ---------------------------------------------------------------------- #
class Table(Element):
    def __init__(
            self,
            values,
            headings=None,
            visible_column_map=None,
            col_widths=None,
            cols_justification=None,
            def_col_width=10,
            auto_size_columns=True,
            max_col_width=20,
            select_mode=None,
            display_row_numbers=False,
            starting_row_number=0,
            num_rows=None,
            row_height=None,
            font=None,
            justification="right",
            text_color=None,
            background_color=None,
            alternating_row_color=None,
            selected_row_colors=(None, None),
            header_text_color=None,
            header_background_color=None,
            header_font=None,
            header_border_width=None,
            header_relief=None,
            row_colors=None,
            vertical_scroll_only=True,
            hide_vertical_scroll=False,
            border_width=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
            size=(None, None),
            s=(None, None),
            change_submits=False,
            enable_events=False,
            enable_click_events=False,
            right_click_selects=False,
            bind_return_key=False,
            pad=None,
            p=None,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param values:                  Your table data represented as a 2-dimensions table... a list of rows, with each row representing a row in your table.
        :type values:                   List[List[str | int | float]]
        :param headings:                The headings to show on the top line
        :type headings:                 List[str]
        :param visible_column_map:      One entry for each column. False indicates the column is not shown
        :type visible_column_map:       List[bool]
        :param col_widths:              Number of characters that each column will occupy
        :type col_widths:               List[int]
        :param cols_justification:      Justification for EACH column. Is a list of strings with the value 'l', 'r', 'c' that indicates how the column will be justified. Either no columns should be set, or have to have one for every colun
        :type cols_justification:       List[str] or Tuple[str] or None
        :param def_col_width:           Default column width in characters
        :type def_col_width:            (int)
        :param auto_size_columns:       if True columns will be sized automatically
        :type auto_size_columns:        (bool)
        :param max_col_width:           Maximum width for all columns in characters
        :type max_col_width:            (int)
        :param select_mode:             Select Mode. Valid values start with "TABLE_SELECT_MODE_".  Valid values are: TABLE_SELECT_MODE_NONE TABLE_SELECT_MODE_BROWSE TABLE_SELECT_MODE_EXTENDED
        :type select_mode:              (enum)
        :param display_row_numbers:     if True, the first column of the table will be the row #
        :type display_row_numbers:      (bool)
        :param starting_row_number:     The row number to use for the first row. All following rows will be based on this starting value. Default is 0.
        :type starting_row_number:      (int)
        :param num_rows:                The number of rows of the table to display at a time
        :type num_rows:                 (int)
        :param row_height:              height of a single row in pixels
        :type row_height:               (int)
        :param font:                    specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                     (str or (str, int[, str]) or None)
        :param justification:           'left', 'right', 'center' are valid choices
        :type justification:            (str)
        :param text_color:              color of the text
        :type text_color:               (str)
        :param background_color:        color of background
        :type background_color:         (str)
        :param alternating_row_color:   if set then every other row will have this color in the background.
        :type alternating_row_color:    (str)
        :param selected_row_colors:     Sets the text color and background color for a selected row. Same format as button colors - tuple ('red', 'yellow') or string 'red on yellow'. Defaults to theme's button color
        :type selected_row_colors:      str or (str, str)
        :param header_text_color:       sets the text color for the header
        :type header_text_color:        (str)
        :param header_background_color: sets the background color for the header
        :type header_background_color:  (str)
        :param header_font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type header_font:              (str or (str, int[, str]) or None)
        :param header_border_width:     Border width for the header portion
        :type header_border_width:      (int | None)
        :param header_relief:           Relief style for the header. Values are same as other elements that use relief. RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID
        :type header_relief:            (str | None)
        :param row_colors:              list of tuples of (row, background color) OR (row, foreground color, background color). Sets the colors of listed rows to the color(s) provided (note the optional foreground color)
        :type row_colors:               List[Tuple[int, str] | Tuple[Int, str, str]]
        :param vertical_scroll_only:    if True only the vertical scrollbar will be visible
        :type vertical_scroll_only:     (bool)
        :param hide_vertical_scroll:    if True vertical scrollbar will be hidden
        :type hide_vertical_scroll:     (bool)
        :param border_width:            Border width/depth in pixels
        :type border_width:             (int)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        :param size:                    DO NOT USE! Use num_rows instead
        :type size:                     (int, int)
        :param change_submits:          DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:           (bool)
        :param enable_events:           Turns on the element specific events. Table events happen when row is clicked
        :type enable_events:            (bool)
        :param enable_click_events:     Turns on the element click events that will give you (row, col) click data when the table is clicked
        :type enable_click_events:      (bool)
        :param right_click_selects:     If True, then right clicking a row will select that row if multiple rows are not currently selected
        :type right_click_selects:      (bool)
        :param bind_return_key:         if True, pressing return key will cause event coming from Table, ALSO a left button double click will generate an event if this parameter is True
        :type bind_return_key:          (bool)
        :param pad:                     Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                       Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:                     Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                      str | int | tuple | object
        :param k:                       Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                        str | int | tuple | object
        :param tooltip:                 text, that will appear when mouse hovers over the element
        :type tooltip:                  (str)
        :param right_click_menu:        A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:         List[List[ List[str] | str ]]
        :param expand_x:                If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                 (bool)
        :param expand_y:                If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                 (bool)
        :param visible:                 set visibility state of the element
        :type visible:                  (bool)
        :param metadata:                User metadata that can be set to ANYTHING
        :type metadata:                 (Any)
        """

        self.Values = values
        self.ColumnHeadings = headings
        self.ColumnsToDisplay = visible_column_map
        self.ColumnWidths = col_widths
        self.cols_justification = cols_justification
        self.MaxColumnWidth = max_col_width
        self.DefaultColumnWidth = def_col_width
        self.AutoSizeColumns = auto_size_columns
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.TextColor = text_color
        self.HeaderTextColor = (
            header_text_color
            if header_text_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["TEXT_INPUT"]
        )
        self.HeaderBackgroundColor = (
            header_background_color
            if header_background_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["INPUT"]
        )
        self.HeaderFont = header_font
        self.Justification = justification
        self.InitialState = None
        self.SelectMode = select_mode
        self.DisplayRowNumbers = display_row_numbers
        self.NumRows = num_rows if num_rows is not None else size[1]
        self.RowHeight = row_height
        self.Widget = self.TKTreeview = None  # type: ttk.Treeview
        self.AlternatingRowColor = alternating_row_color
        self.VerticalScrollOnly = vertical_scroll_only
        self.HideVerticalScroll = hide_vertical_scroll
        self.SelectedRows = []
        self.ChangeSubmits = change_submits or enable_events
        self.BindReturnKey = bind_return_key
        self.StartingRowNumber = (
            starting_row_number  # When displaying row numbers, where to start
        )
        self.RowHeaderText = "Row"
        self.enable_click_events = enable_click_events
        self.right_click_selects = right_click_selects
        self.last_clicked_position = (None, None)
        self.HeaderBorderWidth = header_border_width
        self.BorderWidth = border_width
        self.HeaderRelief = header_relief
        self.table_ttk_style_name = None  # the ttk style name for the Table itself
        if selected_row_colors == (None, None):
            # selected_row_colors = DEFAULT_TABLE_AND_TREE_SELECTED_ROW_COLORS
            selected_row_colors = theme_button_color()
        else:
            try:
                if isinstance(selected_row_colors, str):
                    selected_row_colors = selected_row_colors.split(" on ")
            except Exception as e:
                print(
                    "* Table Element Warning * you messed up with color formatting of Selected Row Color",
                    e,
                )
        self.SelectedRowColors = selected_row_colors

        self.RightClickMenu = right_click_menu
        self.RowColors = row_colors
        self.tree_ids = []  # ids returned when inserting items into table - will use to delete colors
        key = key if key is not None else k
        sz = size if size != (None, None) else s
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_TABLE,
            text_color=text_color,
            background_color=background_color,
            font=font,
            size=sz,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )
        return

    def update(
            self,
            values=None,
            num_rows=None,
            visible=None,
            select_rows=None,
            alternating_row_color=None,
            row_colors=None,
    ):
        """
        Changes some of the settings for the Table Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param values:                A new 2-dimensional table to show
        :type values:                 List[List[str | int | float]]
        :param num_rows:              How many rows to display at a time
        :type num_rows:               (int)
        :param visible:               if True then will be visible
        :type visible:                (bool)
        :param select_rows:           List of rows to select as if user did
        :type select_rows:            List[int]
        :param alternating_row_color: the color to make every other row
        :type alternating_row_color:  (str)
        :param row_colors:            list of tuples of (row, background color) OR (row, foreground color, background color). Changes the colors of listed rows to the color(s) provided (note the optional foreground color)
        :type row_colors:             List[Tuple[int, str] | Tuple[Int, str, str]]
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Table.update - The window was closed")
            return

        if values is not None:
            for id in self.tree_ids:
                self.TKTreeview.item(id, tags=())
                if (
                        self.BackgroundColor is not None
                        and self.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    self.TKTreeview.tag_configure(id, background=self.BackgroundColor)
                else:
                    self.TKTreeview.tag_configure(
                        id, background="#FFFFFF", foreground="#000000"
                    )
                if (
                        self.TextColor is not None
                        and self.TextColor != COLOR_SYSTEM_DEFAULT
                ):
                    self.TKTreeview.tag_configure(id, foreground=self.TextColor)
                else:
                    self.TKTreeview.tag_configure(id, foreground="#000000")

            children = self.TKTreeview.get_children()
            for i in children:
                self.TKTreeview.detach(i)
                self.TKTreeview.delete(i)
            children = self.TKTreeview.get_children()

            self.tree_ids = []
            for i, value in enumerate(values):
                if self.DisplayRowNumbers:
                    value = [i + self.StartingRowNumber] + value
                id = self.TKTreeview.insert(
                    "", "end", text=value, iid=i + 1, values=value, tag=i
                )
                if (
                        self.BackgroundColor is not None
                        and self.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    self.TKTreeview.tag_configure(id, background=self.BackgroundColor)
                else:
                    self.TKTreeview.tag_configure(id, background="#FFFFFF")
                self.tree_ids.append(id)
            self.Values = values
            self.SelectedRows = []
        if not visible:
            self._pack_forget_save_settings(self.element_frame)
        elif visible:
            self._pack_restore_settings(self.element_frame)

        if num_rows is not None:
            self.TKTreeview.config(height=num_rows)
        if select_rows is not None:
            rows_to_select = [i + 1 for i in select_rows]
            self.TKTreeview.selection_set(rows_to_select)

        if alternating_row_color is not None:  # alternating colors
            self.AlternatingRowColor = alternating_row_color

        if self.AlternatingRowColor is not None:
            for row in range(0, len(self.Values), 2):
                self.TKTreeview.tag_configure(row, background=self.AlternatingRowColor)
        if row_colors is not None:  # individual row colors
            self.RowColors = row_colors
            for row_def in self.RowColors:
                if len(row_def) == 2:  # only background is specified
                    self.TKTreeview.tag_configure(row_def[0], background=row_def[1])
                else:
                    self.TKTreeview.tag_configure(
                        row_def[0], background=row_def[2], foreground=row_def[1]
                    )
        if visible is not None:
            self._visible = visible

    def _treeview_selected(self, event):
        """
        Not user callable.  Callback function that is called when something is selected from Table.
        Stores the selected rows in Element as they are being selected. If events enabled, then returns from Read

        :param event: event information from tkinter
        :type event:  (unknown)
        """
        # print('**-- in treeview selected --**')
        selections = self.TKTreeview.selection()
        self.SelectedRows = [int(x) - 1 for x in selections]
        if self.ChangeSubmits:
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = self.Key
            else:
                self.ParentForm.LastButtonClicked = ""
            self.ParentForm.FormRemainedOpen = True
            # if self.ParentForm.CurrentlyRunningMainloop:
            #     self.ParentForm.TKroot.quit()
            _exit_mainloop(self.ParentForm)

    def _treeview_double_click(self, event):
        """
        Not user callable.  Callback function that is called when something is selected from Table.
        Stores the selected rows in Element as they are being selected. If events enabled, then returns from Read

        :param event: event information from tkinter
        :type event:  (unknown)
        """
        selections = self.TKTreeview.selection()
        self.SelectedRows = [int(x) - 1 for x in selections]
        if self.BindReturnKey:  # Signifies BOTH a return key AND a double click
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = self.Key
            else:
                self.ParentForm.LastButtonClicked = ""
            self.ParentForm.FormRemainedOpen = True
            # if self.ParentForm.CurrentlyRunningMainloop:
            #     self.ParentForm.TKroot.quit()
            _exit_mainloop(self.ParentForm)

    def _table_clicked(self, event):
        """
        Not user callable.  Callback function that is called a click happens on a table.
        Stores the selected rows in Element as they are being selected. If events enabled, then returns from Read

        :param event: event information from tkinter
        :type event:  (unknown)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return
        # popup(obj_to_string_single_obj(event))
        try:
            region = self.Widget.identify("region", event.x, event.y)
            if region == "heading":
                row = -1
            elif region == "cell":
                row = int(self.Widget.identify_row(event.y)) - 1
            elif region == "separator":
                row = None
            else:
                row = None
            col_identified = self.Widget.identify_column(event.x)
            if col_identified:  # Sometimes tkinter returns a value of '' which would cause an error if cast to an int
                column = (
                        int(self.Widget.identify_column(event.x)[1:])
                        - 1
                        - int(self.DisplayRowNumbers is True)
                )
            else:
                column = None
        except Exception as e:
            warnings.warn(
                "Error getting table click data for table with key= {}\nError: {}".format(
                    self.Key, e
                ),
                UserWarning,
            )
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "Unable to complete operation getting the clicked event for table with key {}".format(
                        self.Key
                    ),
                    _create_error_message(),
                    e,
                    "Event data:",
                    obj_to_string_single_obj(event),
                )
            row = column = None

        self.last_clicked_position = (row, column)

        # update the rows being selected if appropriate
        self.ParentForm.TKroot.update()
        # self.TKTreeview.()
        selections = self.TKTreeview.selection()
        if self.right_click_selects and len(selections) <= 1:
            if (event.num == 3 and not running_mac()) or (
                    event.num == 2 and running_mac()
            ):
                if row != -1 and row is not None:
                    selections = [row + 1]
                    self.TKTreeview.selection_set(selections)
        # print(selections)
        self.SelectedRows = [int(x) - 1 for x in selections]
        # print('The new selected rows = ', self.SelectedRows, 'selections =', selections)
        if self.enable_click_events:
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = (
                    self.Key,
                    TABLE_CLICKED_INDICATOR,
                    (row, column),
                )
            else:
                self.ParentForm.LastButtonClicked = ""
            self.ParentForm.FormRemainedOpen = True
            _exit_mainloop(self.ParentForm)

    def get(self):
        """
        Get the selected rows using tktiner's selection method.  Returns a list of the selected rows.

        :return: a list of the index of the selected rows (a list of ints)
        :rtype:  List[int]
        """

        selections = self.TKTreeview.selection()
        selected_rows = [int(x) - 1 for x in selections]
        return selected_rows

    def get_last_clicked_position(self):
        """
        Returns a tuple with the row and column of the cell that was last clicked.
        Headers will have a row == -1 and the Row Number Column (if present) will have a column == -1
        :return: The (row,col) position of the last cell clicked in the table
        :rtype:  (int | None, int | None)
        """
        return self.last_clicked_position

    Update = update
    Get = get


# ---------------------------------------------------------------------- #
#                           Tree                                         #
# ---------------------------------------------------------------------- #
class Tree(Element):
    """
    Tree Element - Presents data in a tree-like manner, much like a file/folder browser.  Uses the TreeData class
    to hold the user's data and pass to the element for display.
    """

    def __init__(
            self,
            data=None,
            headings=None,
            visible_column_map=None,
            col_widths=None,
            col0_width=10,
            col0_heading="",
            def_col_width=10,
            auto_size_columns=True,
            max_col_width=20,
            select_mode=None,
            show_expanded=False,
            change_submits=False,
            enable_events=False,
            click_toggles_select=None,
            font=None,
            justification="right",
            text_color=None,
            border_width=None,
            background_color=None,
            selected_row_colors=(None, None),
            header_text_color=None,
            header_background_color=None,
            header_font=None,
            header_border_width=None,
            header_relief=None,
            num_rows=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
            row_height=None,
            vertical_scroll_only=True,
            hide_vertical_scroll=False,
            pad=None,
            p=None,
            key=None,
            k=None,
            tooltip=None,
            right_click_menu=None,
            expand_x=False,
            expand_y=False,
            visible=True,
            metadata=None,
    ):
        """
        :param data:                    The data represented using a PySimpleGUI provided TreeData class
        :type data:                     (TreeData)
        :param headings:                List of individual headings for each column
        :type headings:                 List[str]
        :param visible_column_map:      Determines if a column should be visible. If left empty, all columns will be shown
        :type visible_column_map:       List[bool]
        :param col_widths:              List of column widths so that individual column widths can be controlled
        :type col_widths:               List[int]
        :param col0_width:              Size of Column 0 which is where the row numbers will be optionally shown
        :type col0_width:               (int)
        :param col0_heading:            Text to be shown in the header for the left-most column
        :type col0_heading:             (str)
        :param def_col_width:           default column width
        :type def_col_width:            (int)
        :param auto_size_columns:       if True, the size of a column is determined  using the contents of the column
        :type auto_size_columns:        (bool)
        :param max_col_width:           the maximum size a column can be
        :type max_col_width:            (int)
        :param select_mode:             Use same values as found on Table Element.  Valid values include: TABLE_SELECT_MODE_NONE TABLE_SELECT_MODE_BROWSE TABLE_SELECT_MODE_EXTENDED
        :type select_mode:              (enum)
        :param show_expanded:           if True then the tree will be initially shown with all nodes completely expanded
        :type show_expanded:            (bool)
        :param change_submits:          DO NOT USE. Only listed for backwards compat - Use enable_events instead
        :type change_submits:           (bool)
        :param enable_events:           Turns on the element specific events. Tree events happen when row is clicked
        :type enable_events:            (bool)
        :param click_toggles_select:    If True then clicking a row will cause the selection for that row to toggle between selected and deselected
        :type click_toggles_select:     (bool)
        :param font:                    specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                     (str or (str, int[, str]) or None)
        :param justification:           'left', 'right', 'center' are valid choices
        :type justification:            (str)
        :param text_color:              color of the text
        :type text_color:               (str)
        :param border_width:            Border width/depth in pixels
        :type border_width:             (int)
        :param background_color:        color of background
        :type background_color:         (str)
        :param selected_row_colors:     Sets the text color and background color for a selected row. Same format as button colors - tuple ('red', 'yellow') or string 'red on yellow'. Defaults to theme's button color
        :type selected_row_colors:      str or (str, str)
        :param header_text_color:       sets the text color for the header
        :type header_text_color:        (str)
        :param header_background_color: sets the background color for the header
        :type header_background_color:  (str)
        :param header_font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type header_font:              (str or (str, int[, str]) or None)
        :param header_border_width:     Border width for the header portion
        :type header_border_width:      (int | None)
        :param header_relief:           Relief style for the header. Values are same as other elements that use relief. RELIEF_RAISED RELIEF_SUNKEN RELIEF_FLAT RELIEF_RIDGE RELIEF_GROOVE RELIEF_SOLID
        :type header_relief:            (str | None)
        :param num_rows:                The number of rows of the table to display at a time
        :type num_rows:                 (int)
        :param row_height:              height of a single row in pixels
        :type row_height:               (int)
        :param vertical_scroll_only:    if True only the vertical scrollbar will be visible
        :type vertical_scroll_only:     (bool)
        :param hide_vertical_scroll:    if True vertical scrollbar will be hidden
        :type hide_vertical_scroll:     (bool)
        :param sbar_trough_color:           Scrollbar color of the trough
        :type sbar_trough_color:            (str)
        :param sbar_background_color:       Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:        (str)
        :param sbar_arrow_color:            Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:             (str)
        :param sbar_width:                  Scrollbar width in pixels
        :type sbar_width:                   (int)
        :param sbar_arrow_width:            Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:             (int)
        :param sbar_frame_color:            Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:             (str)
        :param sbar_relief:                 Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                  (str)
        :param pad:                     Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
        :type pad:                      (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param p:                       Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
        :type p:                        (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
        :param key:                     Used with window.find_element and with return values to uniquely identify this element to uniquely identify this element
        :type key:                      str | int | tuple | object
        :param k:                       Same as the Key. You can use either k or key. Which ever is set will be used.
        :type k:                        str | int | tuple | object
        :param tooltip:                 text, that will appear when mouse hovers over the element
        :type tooltip:                  (str)
        :param right_click_menu:        A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:         List[List[str] | str]]
        :param expand_x:                If True the element will automatically expand in the X direction to fill available space
        :type expand_x:                 (bool)
        :param expand_y:                If True the element will automatically expand in the Y direction to fill available space
        :type expand_y:                 (bool)
        :param visible:                 set visibility state of the element
        :type visible:                  (bool)
        :param metadata:                User metadata that can be set to ANYTHING
        :type metadata:                 (Any)
        """

        self.image_dict = {}

        self.TreeData = data
        self.ColumnHeadings = headings
        self.ColumnsToDisplay = visible_column_map
        self.ColumnWidths = col_widths
        self.MaxColumnWidth = max_col_width
        self.DefaultColumnWidth = def_col_width
        self.AutoSizeColumns = auto_size_columns
        self.BackgroundColor = (
            background_color
            if background_color is not None
            else DEFAULT_BACKGROUND_COLOR
        )
        self.TextColor = text_color
        self.HeaderTextColor = (
            header_text_color
            if header_text_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["TEXT_INPUT"]
        )
        self.HeaderBackgroundColor = (
            header_background_color
            if header_background_color is not None
            else LOOK_AND_FEEL_TABLE[CURRENT_LOOK_AND_FEEL]["INPUT"]
        )
        self.HeaderBorderWidth = header_border_width
        self.BorderWidth = border_width
        self.HeaderRelief = header_relief
        self.click_toggles_select = click_toggles_select
        if selected_row_colors == (None, None):
            # selected_row_colors = DEFAULT_TABLE_AND_TREE_SELECTED_ROW_COLORS
            selected_row_colors = theme_button_color()
        else:
            try:
                if isinstance(selected_row_colors, str):
                    selected_row_colors = selected_row_colors.split(" on ")
            except Exception as e:
                print(
                    "* Table Element Warning * you messed up with color formatting of Selected Row Color",
                    e,
                )
        self.SelectedRowColors = selected_row_colors

        self.HeaderFont = header_font
        self.Justification = justification
        self.InitialState = None
        self.SelectMode = select_mode
        self.ShowExpanded = show_expanded
        self.NumRows = num_rows
        self.Col0Width = col0_width
        self.col0_heading = col0_heading
        self.TKTreeview = None  # type: ttk.Treeview
        self.element_frame = None  # type: tk.Frame
        self.VerticalScrollOnly = vertical_scroll_only
        self.HideVerticalScroll = hide_vertical_scroll
        self.SelectedRows = []
        self.ChangeSubmits = change_submits or enable_events
        self.RightClickMenu = right_click_menu
        self.RowHeight = row_height
        self.IconList = {}
        self.IdToKey = {"": ""}
        self.KeyToID = {"": ""}
        key = key if key is not None else k
        pad = pad if pad is not None else p
        self.expand_x = expand_x
        self.expand_y = expand_y

        super().__init__(
            ELEM_TYPE_TREE,
            text_color=text_color,
            background_color=background_color,
            font=font,
            pad=pad,
            key=key,
            tooltip=tooltip,
            visible=visible,
            metadata=metadata,
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )
        return

    def _treeview_selected(self, event):
        """
        Not a user function.  Callback function that happens when an item is selected from the tree.  In this
        method, it saves away the reported selections so they can be properly returned.

        :param event: An event parameter passed in by tkinter.  Not used
        :type event:  (Any)
        """

        selections = self.TKTreeview.selection()
        selected_rows = [self.IdToKey[x] for x in selections]
        if self.click_toggles_select:
            if set(self.SelectedRows) == set(selected_rows):
                for item in selections:
                    self.TKTreeview.selection_remove(item)
                selections = []
        self.SelectedRows = [self.IdToKey[x] for x in selections]

        if self.ChangeSubmits:
            MyForm = self.ParentForm
            if self.Key is not None:
                self.ParentForm.LastButtonClicked = self.Key
            else:
                self.ParentForm.LastButtonClicked = ""
            self.ParentForm.FormRemainedOpen = True
            # if self.ParentForm.CurrentlyRunningMainloop:
            #     self.ParentForm.TKroot.quit()
            _exit_mainloop(self.ParentForm)

    def add_treeview_data(self, node):
        """
        Not a user function.  Recursive method that inserts tree data into the tkinter treeview widget.

        :param node: The node to insert.  Will insert all nodes from starting point downward, recursively
        :type node:  (TreeData)
        """
        if node.key != "":
            if node.icon:
                try:
                    if node.icon not in self.image_dict:
                        if type(node.icon) is bytes:
                            photo = tk.PhotoImage(data=node.icon)
                        else:
                            photo = tk.PhotoImage(file=node.icon)
                        self.image_dict[node.icon] = photo
                    else:
                        photo = self.image_dict.get(node.icon)

                    node.photo = photo
                    id = self.TKTreeview.insert(
                        self.KeyToID[node.parent],
                        "end",
                        iid=None,
                        text=node.text,
                        values=node.values,
                        open=self.ShowExpanded,
                        image=node.photo,
                    )
                    self.IdToKey[id] = node.key
                    self.KeyToID[node.key] = id
                except Exception:
                    self.photo = None
            else:
                id = self.TKTreeview.insert(
                    self.KeyToID[node.parent],
                    "end",
                    iid=None,
                    text=node.text,
                    values=node.values,
                    open=self.ShowExpanded,
                )
                self.IdToKey[id] = node.key
                self.KeyToID[node.key] = id

        for node in node.children:
            self.add_treeview_data(node)

    def update(
            self, values=None, key=None, value=None, text=None, icon=None, visible=None
    ):
        """
        Changes some of the settings for the Tree Element. Must call `Window.Read` or `Window.Finalize` prior

        Changes will not be visible in your window until you call window.read or window.refresh.

        If you change visibility, your element may MOVE. If you want it to remain stationary, use the "layout helper"
        function "pin" to ensure your element is "pinned" to that location in your layout so that it returns there
        when made visible.

        :param values:  Representation of the tree
        :type values:   (TreeData)
        :param key:     identifies a particular item in tree to update
        :type key:      str | int | tuple | object
        :param value:   sets the node identified by key to a particular value
        :type value:    (Any)
        :param text:    sets the node identified by key to this string
        :type text:     (str)
        :param icon:    can be either a base64 icon or a filename for the icon
        :type icon:     bytes | str
        :param visible: control visibility of element
        :type visible:  (bool)
        """
        if (
                not self._widget_was_created()
        ):  # if widget hasn't been created yet, then don't allow
            return

        if self._this_elements_window_closed():
            _error_popup_with_traceback("Error in Tree.update - The window was closed")
            return

        if values is not None:
            children = self.TKTreeview.get_children()
            for i in children:
                self.TKTreeview.detach(i)
                self.TKTreeview.delete(i)
            children = self.TKTreeview.get_children()
            self.TreeData = values
            self.IdToKey = {"": ""}
            self.KeyToID = {"": ""}
            self.add_treeview_data(self.TreeData.root_node)
            self.SelectedRows = []
        if key is not None:
            for id in self.IdToKey.keys():
                if key == self.IdToKey[id]:
                    break
            else:
                id = None
                print("** Key not found **")
        else:
            id = None
        if id:
            # item = self.TKTreeview.item(id)
            if value is not None:
                self.TKTreeview.item(id, values=value)
            if text is not None:
                self.TKTreeview.item(id, text=text)
            if icon is not None:
                try:
                    if type(icon) is bytes:
                        photo = tk.PhotoImage(data=icon)
                    else:
                        photo = tk.PhotoImage(file=icon)
                    self.TKTreeview.item(id, image=photo)
                    self.IconList[key] = (
                        photo  # save so that it's not deleted (save reference)
                    )
                except Exception:
                    pass
            # item = self.TKTreeview.item(id)
        if not visible:
            self._pack_forget_save_settings(self.element_frame)
        elif visible:
            self._pack_restore_settings(self.element_frame)

        if visible is not None:
            self._visible = visible

        return self

    Update = update


class TreeData(object):
    """
    Class that user fills in to represent their tree data. It's a very simple tree representation with a root "Node"
    with possibly one or more children "Nodes".  Each Node contains a key, text to display, list of values to display
    and an icon.  The entire tree is built using a single method, Insert.  Nothing else is required to make the tree.
    """

    class Node(object):
        """
        Contains information about the individual node in the tree
        """

        def __init__(self, parent, key, text, values, icon=None):
            """
            Represents a node within the TreeData class

            :param parent: The parent Node
            :type parent:  (TreeData.Node)
            :param key:    Used to uniquely identify this node
            :type key:     str | int | tuple | object
            :param text:   The text that is displayed at this node's location
            :type text:    (str)
            :param values: The list of values that are displayed at this node
            :type values:  List[Any]
            :param icon:   just a icon
            :type icon:    str | bytes
            """

            self.parent = parent  # type: TreeData.Node
            self.children = []  # type: List[TreeData.Node]
            self.key = key  # type: str
            self.text = text  # type: str
            self.values = values  # type: List[Any]
            self.icon = icon  # type: str | bytes

        def _Add(self, node):
            self.children.append(node)

    def __init__(self):
        """
        Instantiate the object, initializes the Tree Data, creates a root node for you
        """
        self.tree_dict = {}  # type: Dict[str, TreeData.Node]
        self.root_node = self.Node("", "", "root", [], None)  # The root node
        self.tree_dict[""] = self.root_node  # Start the tree out with the root node

    def _AddNode(self, key, node):
        """
        Adds a node to tree dictionary (not user callable)

        :param key:  Uniquely identifies this Node
        :type key:   (str)
        :param node: Node being added
        :type node:  (TreeData.Node)
        """
        self.tree_dict[key] = node

    def insert(self, parent, key, text, values, icon=None):
        """
        Inserts a node into the tree. This is how user builds their tree, by Inserting Nodes
        This is the ONLY user callable method in the TreeData class

        :param parent: the parent Node
        :type parent:  (Node)
        :param key:    Used to uniquely identify this node
        :type key:     str | int | tuple | object
        :param text:   The text that is displayed at this node's location
        :type text:    (str)
        :param values: The list of values that are displayed at this node
        :type values:  List[Any]
        :param icon:   icon
        :type icon:    str | bytes
        """

        node = self.Node(parent, key, text, values, icon)
        self.tree_dict[key] = node
        parent_node = self.tree_dict[parent]
        parent_node._Add(node)

    def __repr__(self):
        """
        Converts the TreeData into a printable version, nicely formatted

        :return: (str) A formatted, text version of the TreeData
        :rtype:
        """
        return self._NodeStr(self.root_node, 1)

    def _NodeStr(self, node, level):
        """
        Does the magic of converting the TreeData into a nicely formatted string version

        :param node:  The node to begin printing the tree
        :type node:   (TreeData.Node)
        :param level: The indentation level for string formatting
        :type level:  (int)
        """
        return "\n".join(
            [
                str(node.key)
                + " : "
                + str(node.text)
                + " [ "
                + ", ".join([str(v) for v in node.values])
                + " ]"
            ]
            + [
                " " * 4 * level + self._NodeStr(child, level + 1)
                for child in node.children
            ]
        )

    Insert = insert


# ---------------------------------------------------------------------- #
#                           Error Element                                #
# ---------------------------------------------------------------------- #
class ErrorElement(Element):
    """
    A "dummy Element" that is returned when there are error conditions, like trying to find an element that's invalid
    """

    def __init__(self, key=None, metadata=None):
        """
        :param key: Used with window.find_element and with return values to uniquely identify this element
        :type key:
        """
        self.Key = key

        super().__init__(ELEM_TYPE_ERROR, key=key, metadata=metadata)

    def update(self, silent_on_error=True, *args, **kwargs):
        """
        Update method for the Error Element, an element that should not be directly used by developer

        :param silent_on_error: if False, then a Popup window will be shown
        :type silent_on_error:  (bool)
        :return:                returns 'self' so call can be chained
        :rtype:                 (ErrorElement)
        """
        print(
            "** Your update is being ignored because you supplied a bad key earlier **"
        )
        return self

    def get(self):
        """
        One of the method names found in other Elements. Used here to return an error string in case it's called

        :return: A warning text string.
        :rtype:  (str)
        """
        return "This is NOT a valid Element!\nSTOP trying to do things with it or I will have to crash at some point!"

    Get = get
    Update = update


# ---------------------------------------------------------------------- #
#                           Stretch Element                              #
# ---------------------------------------------------------------------- #
# This is for source code compatibility with tkinter version. No tkinter equivalent but you can fake it using a Text element that expands in the X direction
def Push(background_color=None):
    """
    Acts like a Stretch element found in the Qt port.
    Used in a Horizontal fashion.  Placing one on each side of an element will enter the element.
    Place one to the left and the element to the right will be right justified.  See VStretch for vertical type
    :param background_color: color of background may be needed because of how this is implemented
    :type background_color:  (str)
    :return:                 (Text)
    """
    return Text(
        font="_ 1", background_color=background_color, pad=(0, 0), expand_x=True
    )


P = Push
Stretch = Push


def VPush(background_color=None):
    """
    Acts like a Stretch element found in the Qt port.
    Used in a Vertical fashion.
    :param background_color: color of background may be needed because of how this is implemented
    :type background_color:  (str)
    :return:                 (Text)
    """
    return Text(
        font="_ 1", background_color=background_color, pad=(0, 0), expand_y=True
    )


VStretch = VPush
VP = VPush


# ------------------------------------------------------------------------- #
#                       _TimerPeriodic CLASS                                #
# ------------------------------------------------------------------------- #


class _TimerPeriodic:
    id_counter = 1
    # Dictionary containing the active timers.  Format is {id : _TimerPeriodic object}
    active_timers = {}  # type: dict[int:_TimerPeriodic]

    def __init__(self, window, frequency_ms, key=EVENT_TIMER, repeating=True):
        """
        :param window:          The window to send events to
        :type window:           Window
        :param frequency_ms:    How often to send events in milliseconds
        :type frequency_ms:     int
        :param repeating:       If True then the timer will run, repeatedly sending events, until stopped
        :type repeating:        bool
        """
        self.window = window
        self.frequency_ms = frequency_ms
        self.repeating = repeating
        self.key = key
        self.id = _TimerPeriodic.id_counter
        _TimerPeriodic.id_counter += 1
        self.start()

    @classmethod
    def stop_timer_with_id(cls, timer_id):
        """
        Not user callable!
        :return: A simple counter that makes each container element unique
        :rtype:
        """
        timer = cls.active_timers.get(timer_id, None)
        if timer is not None:
            timer.stop()

    @classmethod
    def stop_all_timers_for_window(cls, window):
        """
        Stops all timers for a given window
        :param window:      The window to stop timers for
        :type window:       Window
        """
        for timer in _TimerPeriodic.active_timers.values():
            if timer.window == window:
                timer.running = False

    @classmethod
    def get_all_timers_for_window(cls, window):
        """
        Returns a list of timer IDs for a given window
        :param window:      The window to find timers for
        :type window:       Window
        :return:            List of timer IDs for the window
        :rtype:             List[int]
        """
        timers = []
        for timer in _TimerPeriodic.active_timers.values():
            if timer.window == window:
                timers.append(timer.id)

        return timers

    def timer_thread(self):
        """
        The thread that sends events to the window.  Runs either once or in a loop until timer is stopped
        """

        if not self.running:  # if timer has been cancelled, abort
            del _TimerPeriodic.active_timers[self.id]
            return
        while True:
            time.sleep(self.frequency_ms / 1000)
            if not self.running:  # if timer has been cancelled, abort
                del _TimerPeriodic.active_timers[self.id]
                return
            self.window.write_event_value(self.key, self.id)

            if not self.repeating:  # if timer does not repeat, then exit thread
                del _TimerPeriodic.active_timers[self.id]
                return

    def start(self):
        """
        Starts a timer by starting a timer thread
        Adds timer to the list of active timers
        """
        self.running = True
        self.thread = threading.Thread(target=self.timer_thread, daemon=True)
        self.thread.start()
        _TimerPeriodic.active_timers[self.id] = self

    def stop(self):
        """
        Stops a timer
        """
        self.running = False


# ------------------------------------------------------------------------- #
#                       Window CLASS                                        #
# ------------------------------------------------------------------------- #
class Window:
    """
    Represents a single Window
    """

    NumOpenWindows = 0
    _user_defined_icon = None
    hidden_master_root = None  # type: tk.Tk
    _animated_popup_dict = {}  # type: Dict
    _active_windows = {}  # type: Dict[Window, tk.Tk()]
    _move_all_windows = False  # if one window moved, they will move
    _window_that_exited = None  # type: Window
    _root_running_mainloop = None  # type: tk.Tk()    # (may be the hidden root or a window's root)
    _timeout_key = None
    _TKAfterID = None  # timer that is used to run reads with timeouts
    _window_running_mainloop = None  # The window that is running the mainloop
    _container_element_counter = (
        0  # used to get a number of Container Elements (Frame, Column, Tab)
    )
    _read_call_from_debugger = False
    _timeout_0_counter = 0  # when timeout=0 then go through each window one at a time
    _counter_for_ttk_widgets = 0
    _floating_debug_window_build_needed = False
    _main_debug_window_build_needed = False
    # rereouted stdout info. List of tuples (window, element, previous destination)
    _rerouted_stdout_stack = []  # type: List[Tuple[Window, Element]]
    _rerouted_stderr_stack = []  # type: List[Tuple[Window, Element]]
    _original_stdout = None
    _original_stderr = None
    _watermark = None
    _watermark_temp_forced = False
    _watermark_user_text = ""

    def __init__(
            self,
            title,
            layout=None,
            default_element_size=None,
            default_button_element_size=(None, None),
            auto_size_text=None,
            auto_size_buttons=None,
            location=(None, None),
            relative_location=(None, None),
            size=(None, None),
            element_padding=None,
            margins=(None, None),
            button_color=None,
            font=None,
            progress_bar_color=(None, None),
            background_color=None,
            border_depth=None,
            auto_close=False,
            auto_close_duration=DEFAULT_AUTOCLOSE_TIME,
            icon=None,
            force_toplevel=False,
            alpha_channel=None,
            return_keyboard_events=False,
            use_default_focus=True,
            text_justification=None,
            no_titlebar=False,
            grab_anywhere=False,
            grab_anywhere_using_control=True,
            keep_on_top=None,
            resizable=False,
            disable_close=False,
            disable_minimize=False,
            right_click_menu=None,
            transparent_color=None,
            debugger_enabled=True,
            right_click_menu_background_color=None,
            right_click_menu_text_color=None,
            right_click_menu_disabled_text_color=None,
            right_click_menu_selected_colors=(None, None),
            right_click_menu_font=None,
            right_click_menu_tearoff=False,
            finalize=False,
            element_justification="left",
            ttk_theme=None,
            use_ttk_buttons=None,
            modal=False,
            enable_close_attempted_event=False,
            enable_window_config_events=False,
            titlebar_background_color=None,
            titlebar_text_color=None,
            titlebar_font=None,
            titlebar_icon=None,
            use_custom_titlebar=None,
            scaling=None,
            sbar_trough_color=None,
            sbar_background_color=None,
            sbar_arrow_color=None,
            sbar_width=None,
            sbar_arrow_width=None,
            sbar_frame_color=None,
            sbar_relief=None,
            watermark=None,
            metadata=None,
    ):
        """
        :param title:                                The title that will be displayed in the Titlebar and on the Taskbar
        :type title:                                 (str)
        :param layout:                               The layout for the window. Can also be specified in the Layout method
        :type layout:                                List[List[Element]] | Tuple[Tuple[Element]]
        :param default_element_size:                 size in characters (wide) and rows (high) for all elements in this window
        :type default_element_size:                  (int, int) - (width, height)
        :param default_button_element_size:          (width, height) size in characters (wide) and rows (high) for all Button elements in this window
        :type default_button_element_size:           (int, int)
        :param auto_size_text:                       True if Elements in Window should be sized to exactly fir the length of text
        :type auto_size_text:                        (bool)
        :param auto_size_buttons:                    True if Buttons in this Window should be sized to exactly fit the text on this.
        :type auto_size_buttons:                     (bool)
        :param relative_location:                    (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
        :type relative_location:                     (int, int)
        :param location:                             (x,y) location, in pixels, to locate the upper left corner of the window on the screen. Default is to center on screen. None will not set any location meaning the OS will decide
        :type location:                              (int, int) or (None, None) or None
        :param size:                                 (width, height) size in pixels for this window. Normally the window is autosized to fit contents, not set to an absolute size by the user. Try not to set this value. You risk, the contents being cut off, etc. Let the layout determine the window size instead
        :type size:                                  (int, int)
        :param element_padding:                      Default amount of padding to put around elements in window (left/right, top/bottom) or ((left, right), (top, bottom)), or an int. If an int, then it's converted into a tuple (int, int)
        :type element_padding:                       (int, int) or ((int, int),(int,int)) or int
        :param margins:                              (left/right, top/bottom) Amount of pixels to leave inside the window's frame around the edges before your elements are shown.
        :type margins:                               (int, int)
        :param button_color:                         Default button colors for all buttons in the window
        :type button_color:                          (str, str) | str
        :param font:                                 specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                                  (str or (str, int[, str]) or None)
        :param progress_bar_color:                   (bar color, background color) Sets the default colors for all progress bars in the window
        :type progress_bar_color:                    (str, str)
        :param background_color:                     color of background
        :type background_color:                      (str)
        :param border_depth:                         Default border depth (width) for all elements in the window
        :type border_depth:                          (int)
        :param auto_close:                           If True, the window will automatically close itself
        :type auto_close:                            (bool)
        :param auto_close_duration:                  Number of seconds to wait before closing the window
        :type auto_close_duration:                   (int)
        :param icon:                                 Can be either a filename or Base64 value. For Windows if filename, it MUST be ICO format. For Linux, must NOT be ICO. Most portable is to use a Base64 of a PNG file. This works universally across all OS's
        :type icon:                                  (str | bytes)
        :param force_toplevel:                       If True will cause this window to skip the normal use of a hidden master window
        :type force_toplevel:                        (bool)
        :param alpha_channel:                        Sets the opacity of the window. 0 = invisible 1 = completely visible. Values bewteen 0 & 1 will produce semi-transparent windows in SOME environments (The Raspberry Pi always has this value at 1 and cannot change.
        :type alpha_channel:                         (float)
        :param return_keyboard_events:               if True key presses on the keyboard will be returned as Events from Read calls
        :type return_keyboard_events:                (bool)
        :param use_default_focus:                    If True will use the default focus algorithm to set the focus to the "Correct" element
        :type use_default_focus:                     (bool)
        :param text_justification:                   Default text justification for all Text Elements in window
        :type text_justification:                    'left' | 'right' | 'center'
        :param no_titlebar:                          If true, no titlebar nor frame will be shown on window. This means you cannot minimize the window and it will not show up on the taskbar
        :type no_titlebar:                           (bool)
        :param grab_anywhere:                        If True can use mouse to click and drag to move the window. Almost every location of the window will work except input fields on some systems
        :type grab_anywhere:                         (bool)
        :param grab_anywhere_using_control:          If True can use CONTROL key + left mouse mouse to click and drag to move the window. DEFAULT is TRUE. Unlike normal grab anywhere, it works on all elements.
        :type grab_anywhere_using_control:           (bool)
        :param keep_on_top:                          If True, window will be created on top of all other windows on screen. It can be bumped down if another window created with this parm
        :type keep_on_top:                           (bool)
        :param resizable:                            If True, allows the user to resize the window. Note the not all Elements will change size or location when resizing.
        :type resizable:                             (bool)
        :param disable_close:                        If True, the X button in the top right corner of the window will no work.  Use with caution and always give a way out toyour users
        :type disable_close:                         (bool)
        :param disable_minimize:                     if True the user won't be able to minimize window.  Good for taking over entire screen and staying that way.
        :type disable_minimize:                      (bool)
        :param right_click_menu:                     A list of lists of Menu items to show when this element is right clicked. See user docs for exact format.
        :type right_click_menu:                      List[List[ List[str] | str ]]
        :param transparent_color:                    Any portion of the window that has this color will be completely transparent. You can even click through these spots to the window under this window.
        :type transparent_color:                     (str)
        :param debugger_enabled:                     If True then the internal debugger will be enabled
        :type debugger_enabled:                      (bool)
        :param right_click_menu_background_color:    Background color for right click menus
        :type right_click_menu_background_color:     (str)
        :param right_click_menu_text_color:          Text color for right click menus
        :type right_click_menu_text_color:           (str)
        :param right_click_menu_disabled_text_color: Text color for disabled right click menu items
        :type right_click_menu_disabled_text_color:  (str)
        :param right_click_menu_selected_colors:     Text AND background colors for a selected item. Can be a Tuple OR a color string. simplified-button-color-string "foreground on background". Can be a single color if want to set only the background. Normally a tuple, but can be a simplified-dual-color-string "foreground on background". Can be a single color if want to set only the background.
        :type right_click_menu_selected_colors:      (str, str) | str | Tuple
        :param right_click_menu_font:                Font for right click menus
        :type right_click_menu_font:                 (str or (str, int[, str]) or None)
        :param right_click_menu_tearoff:             If True then all right click menus can be torn off
        :type right_click_menu_tearoff:              bool
        :param finalize:                             If True then the Finalize method will be called. Use this rather than chaining .Finalize for cleaner code
        :type finalize:                              (bool)
        :param element_justification:                All elements in the Window itself will have this justification 'left', 'right', 'center' are valid values
        :type element_justification:                 (str)
        :param ttk_theme:                            Set the tkinter ttk "theme" of the window.  Default = DEFAULT_TTK_THEME.  Sets all ttk widgets to this theme as their default
        :type ttk_theme:                             (str)
        :param use_ttk_buttons:                      Affects all buttons in window. True = use ttk buttons. False = do not use ttk buttons.  None = use ttk buttons only if on a Mac
        :type use_ttk_buttons:                       (bool)
        :param modal:                                If True then this window will be the only window a user can interact with until it is closed
        :type modal:                                 (bool)
        :param enable_close_attempted_event:         If True then the window will not close when "X" clicked. Instead an event WINDOW_CLOSE_ATTEMPTED_EVENT if returned from window.read
        :type enable_close_attempted_event:          (bool)
        :param enable_window_config_events:          If True then window configuration events (resizing or moving the window) will return WINDOW_CONFIG_EVENT from window.read. Note you will get several when Window is created.
        :type enable_window_config_events:           (bool)
        :param titlebar_background_color:            If custom titlebar indicated by use_custom_titlebar, then use this as background color
        :type titlebar_background_color:             (str | None)
        :param titlebar_text_color:                  If custom titlebar indicated by use_custom_titlebar, then use this as text color
        :type titlebar_text_color:                   (str | None)
        :param titlebar_font:                        If custom titlebar indicated by use_custom_titlebar, then use this as title font
        :type titlebar_font:                         (str or (str, int[, str]) or None)
        :param titlebar_icon:                        If custom titlebar indicated by use_custom_titlebar, then use this as the icon (file or base64 bytes)
        :type titlebar_icon:                         (bytes | str)
        :param use_custom_titlebar:                  If True, then a custom titlebar will be used instead of the normal titlebar
        :type use_custom_titlebar:                   bool
        :param scaling:                              Apply scaling to the elements in the window. Can be set on a global basis using set_options
        :type scaling:                               float
        :param sbar_trough_color:                    Scrollbar color of the trough
        :type sbar_trough_color:                     (str)
        :param sbar_background_color:                Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
        :type sbar_background_color:                 (str)
        :param sbar_arrow_color:                     Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
        :type sbar_arrow_color:                      (str)
        :param sbar_width:                           Scrollbar width in pixels
        :type sbar_width:                            (int)
        :param sbar_arrow_width:                     Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
        :type sbar_arrow_width:                      (int)
        :param sbar_frame_color:                     Scrollbar Color of frame around scrollbar (available only on some ttk themes)
        :type sbar_frame_color:                      (str)
        :param sbar_relief:                          Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
        :type sbar_relief:                           (str)
        :param watermark:                            If True, then turns on watermarking temporarily for ALL windows created from this point forward. See global settings doc for more info
        :type watermark:                             bool
        :param metadata:                             User metadata that can be set to ANYTHING
        :type metadata:                              (Any)
        """

        self._metadata = None  # type: Any
        self.AutoSizeText = (
            auto_size_text if auto_size_text is not None else DEFAULT_AUTOSIZE_TEXT
        )
        self.AutoSizeButtons = (
            auto_size_buttons
            if auto_size_buttons is not None
            else DEFAULT_AUTOSIZE_BUTTONS
        )
        self.Title = str(title)
        self.Rows = []  # a list of ELEMENTS for this row
        self.DefaultElementSize = (
            default_element_size
            if default_element_size is not None
            else DEFAULT_ELEMENT_SIZE
        )
        self.DefaultButtonElementSize = (
            default_button_element_size
            if default_button_element_size != (None, None)
            else DEFAULT_BUTTON_ELEMENT_SIZE
        )
        if DEFAULT_WINDOW_LOCATION != (None, None) and location == (None, None):
            self.Location = DEFAULT_WINDOW_LOCATION
        else:
            self.Location = location
        self.RelativeLoction = relative_location
        self.ButtonColor = button_color_to_tuple(button_color)
        self.BackgroundColor = (
            background_color if background_color else DEFAULT_BACKGROUND_COLOR
        )
        self.ParentWindow = None
        self.Font = font if font else DEFAULT_FONT
        self.RadioDict = {}
        self.BorderDepth = border_depth
        if icon:
            self.WindowIcon = icon
        elif Window._user_defined_icon is not None:
            self.WindowIcon = Window._user_defined_icon
        else:
            self.WindowIcon = DEFAULT_WINDOW_ICON
        self.AutoClose = auto_close
        self.NonBlocking = False
        self.TKroot = None  # type: tk.Tk
        self.TKrootDestroyed = False
        self.CurrentlyRunningMainloop = False
        self.FormRemainedOpen = False
        self.TKAfterID = None
        self.ProgressBarColor = progress_bar_color
        self.AutoCloseDuration = auto_close_duration
        self.RootNeedsDestroying = False
        self.Shown = False
        self.ReturnValues = None
        self.ReturnValuesList = []
        self.ReturnValuesDictionary = {}
        self.DictionaryKeyCounter = 0
        self.LastButtonClicked = None
        self.LastButtonClickedWasRealtime = False
        self.UseDictionary = False
        self.UseDefaultFocus = use_default_focus
        self.ReturnKeyboardEvents = return_keyboard_events
        self.LastKeyboardEvent = None
        self.TextJustification = text_justification
        self.NoTitleBar = no_titlebar
        self.Grab = grab_anywhere
        self.GrabAnywhere = grab_anywhere
        self.GrabAnywhereUsingControlKey = grab_anywhere_using_control
        if keep_on_top is None and DEFAULT_KEEP_ON_TOP is not None:
            keep_on_top = DEFAULT_KEEP_ON_TOP
        elif keep_on_top is None:
            keep_on_top = False
        self.KeepOnTop = keep_on_top
        self.ForceTopLevel = force_toplevel
        self.Resizable = resizable
        self._AlphaChannel = (
            alpha_channel if alpha_channel is not None else DEFAULT_ALPHA_CHANNEL
        )
        self.Timeout = None
        self.TimeoutKey = TIMEOUT_KEY
        self.TimerCancelled = False
        self.DisableClose = disable_close
        self.DisableMinimize = disable_minimize
        self._Hidden = False
        self._Size = size
        self.XFound = False
        if element_padding is not None:
            if isinstance(element_padding, int):
                element_padding = (element_padding, element_padding)

        if element_padding is None:
            self.ElementPadding = DEFAULT_ELEMENT_PADDING
        else:
            self.ElementPadding = element_padding
        self.RightClickMenu = right_click_menu
        self.Margins = margins if margins != (None, None) else DEFAULT_MARGINS
        self.ContainerElemementNumber = Window._GetAContainerNumber()
        # The dictionary containing all elements and keys for the window
        # The keys are the keys for the elements and the values are the elements themselves.
        self.AllKeysDict = {}
        self.TransparentColor = transparent_color
        self.UniqueKeyCounter = 0
        self.DebuggerEnabled = debugger_enabled
        self.WasClosed = False
        self.ElementJustification = element_justification
        self.FocusSet = False
        self.metadata = metadata
        self.TtkTheme = ttk_theme or DEFAULT_TTK_THEME
        self.UseTtkButtons = (
            use_ttk_buttons if use_ttk_buttons is not None else USE_TTK_BUTTONS
        )
        self.user_bind_dict = {}  # Used when user defines a tkinter binding using bind method - convert bind string to key modifier
        self.user_bind_event = None  # Used when user defines a tkinter binding using bind method - event data from tkinter
        self.modal = modal
        self.thread_queue = None  # type: queue.Queue
        self.thread_lock = None  # type: threading.Lock
        self.thread_timer = None  # type: tk.Misc
        self.thread_strvar = None  # type: tk.StringVar
        self.read_closed_window_count = 0
        self.config_last_size = (None, None)
        self.config_last_location = (None, None)
        self.starting_window_position = (None, None)
        self.not_completed_initial_movement = True
        self.config_count = 0
        self.saw_00 = False
        self.maximized = False
        self.right_click_menu_background_color = (
            right_click_menu_background_color
            if right_click_menu_background_color is not None
            else theme_input_background_color()
        )
        self.right_click_menu_text_color = (
            right_click_menu_text_color
            if right_click_menu_text_color is not None
            else theme_input_text_color()
        )
        self.right_click_menu_disabled_text_color = (
            right_click_menu_disabled_text_color
            if right_click_menu_disabled_text_color is not None
            else COLOR_SYSTEM_DEFAULT
        )
        self.right_click_menu_font = (
            right_click_menu_font if right_click_menu_font is not None else self.Font
        )
        self.right_click_menu_tearoff = right_click_menu_tearoff
        self.auto_close_timer_needs_starting = False
        self.finalize_in_progress = False
        self.close_destroys_window = (
            not enable_close_attempted_event
            if enable_close_attempted_event is not None
            else None
        )
        self.enable_window_config_events = enable_window_config_events
        self.override_custom_titlebar = False
        self.use_custom_titlebar = use_custom_titlebar or theme_use_custom_titlebar()
        self.titlebar_background_color = titlebar_background_color
        self.titlebar_text_color = titlebar_text_color
        self.titlebar_font = titlebar_font
        self.titlebar_icon = titlebar_icon
        self.right_click_menu_selected_colors = _simplified_dual_color_to_tuple(
            right_click_menu_selected_colors,
            (self.right_click_menu_background_color, self.right_click_menu_text_color),
        )
        self.TKRightClickMenu = None
        self._grab_anywhere_ignore_these_list = []
        self._grab_anywhere_include_these_list = []
        self._has_custom_titlebar = use_custom_titlebar
        self._mousex = self._mousey = 0
        self._startx = self._starty = 0
        self.scaling = scaling if scaling is not None else DEFAULT_SCALING
        if self.use_custom_titlebar:
            self.Margins = (0, 0)
            self.NoTitleBar = True
        self._mouse_offset_x = self._mouse_offset_y = 0

        if watermark:
            Window._watermark_temp_forced = True
            _global_settings_get_watermark_info()
        elif not watermark:
            Window._watermark = None
            Window._watermark_temp_forced = False

        self.ttk_part_overrides = TTKPartOverrides(
            sbar_trough_color=sbar_trough_color,
            sbar_background_color=sbar_background_color,
            sbar_arrow_color=sbar_arrow_color,
            sbar_width=sbar_width,
            sbar_arrow_width=sbar_arrow_width,
            sbar_frame_color=sbar_frame_color,
            sbar_relief=sbar_relief,
        )

        if no_titlebar:
            self.override_custom_titlebar = True

        if layout is not None and type(layout) not in (list, tuple):
            warnings.warn("Your layout is not a list or tuple... this is not good!")

        if layout is not None:
            self.Layout(layout)
            if finalize:
                self.Finalize()

        if CURRENT_LOOK_AND_FEEL == "Default":
            print(
                "Window will be a boring gray. Try removing the theme call entirely\n",
                "You will get the default theme or the one set in global settings\n"
                "If you seriously want this gray window and no more nagging, add  theme('DefaultNoMoreNagging')  or theme('Gray Gray Gray') for completely gray/System Defaults",
            )

    @classmethod
    def _GetAContainerNumber(cls):
        """
        Not user callable!
        :return: A simple counter that makes each container element unique
        :rtype:
        """
        cls._container_element_counter += 1
        return cls._container_element_counter

    @classmethod
    def _IncrementOpenCount(self):
        """
        Not user callable!  Increments the number of open windows
        Note - there is a bug where this count easily gets out of sync. Issue has been opened already. No ill effects
        """
        self.NumOpenWindows += 1
        # print('+++++ INCREMENTING Num Open Windows = {} ---'.format(Window.NumOpenWindows))

    @classmethod
    def _DecrementOpenCount(self):
        """
        Not user callable!  Decrements the number of open windows
        """
        self.NumOpenWindows -= 1 * (self.NumOpenWindows != 0)  # decrement if not 0
        # print('----- DECREMENTING Num Open Windows = {} ---'.format(Window.NumOpenWindows))

    @classmethod
    def get_screen_size(self):
        """
        This is a "Class Method" meaning you call it by writing: width, height = Window.get_screen_size()
        Returns the size of the "screen" as determined by tkinter.  This can vary depending on your operating system and the number of monitors installed on your system.  For Windows, the primary monitor's size is returns. On some multi-monitored Linux systems, the monitors are combined and the total size is reported as if one screen.

        :return: Size of the screen in pixels as determined by tkinter
        :rtype:  (int, int)
        """
        root = _get_hidden_master_root()
        screen_width = root.winfo_screenwidth()
        screen_height = root.winfo_screenheight()
        return screen_width, screen_height

    @property
    def metadata(self):
        """
        Metadata is available for all windows. You can set to any value.
        :return: the current metadata value
        :rtype:  (Any)
        """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """
        Metadata is available for all windows. You can set to any value.
        :param value: Anything you want it to be
        :type value:  (Any)
        """
        self._metadata = value

    # ------------------------- Add ONE Row to Form ------------------------- #
    def add_row(self, *args):
        """
        Adds a single row of elements to a window's self.Rows variables.
        Generally speaking this is NOT how users should be building Window layouts.
        Users, create a single layout (a list of lists) and pass as a parameter to Window object, or call Window.Layout(layout)

        """
        NumRows = len(self.Rows)  # number of existing rows is our row number
        CurrentRowNumber = NumRows  # this row's number
        CurrentRow = []  # start with a blank row and build up
        # -------------------------  Add the elements to a row  ------------------------- #
        for i, element in enumerate(
                args
        ):  # Loop through list of elements and add them to the row
            if isinstance(element, tuple) or isinstance(element, list):
                self.add_row(*element)
                continue
                _error_popup_with_traceback(
                    "Error creating Window layout",
                    "Layout has a LIST instead of an ELEMENT",
                    "This sometimes means you have a badly placed ]",
                    "The offensive list is:",
                    element,
                    "This list will be stripped from your layout",
                )
                continue
            elif callable(element) and not isinstance(element, Element):
                _error_popup_with_traceback(
                    "Error creating Window layout",
                    "Layout has a FUNCTION instead of an ELEMENT",
                    "This likely means you are missing () from your layout",
                    "The offensive list is:",
                    element,
                    "This item will be stripped from your layout",
                )
                continue
            if element.ParentContainer is not None:
                warnings.warn(
                    "*** YOU ARE ATTEMPTING TO REUSE AN ELEMENT IN YOUR LAYOUT! Once placed in a layout, an element cannot be used in another layout. ***",
                    UserWarning,
                )
                _error_popup_with_traceback(
                    "Error detected in layout - Contains an element that has already been used.",
                    "You have attempted to reuse an element in your layout.",
                    "The layout specified has an element that's already been used.",
                    'You MUST start with a "clean", unused layout every time you create a window',
                    "The offensive Element = ",
                    element,
                    "and has a key = ",
                    element.Key,
                    "This item will be stripped from your layout",
                    'Hint - try printing your layout and matching the IDs "print(layout)"',
                )
                continue
            element.Position = (CurrentRowNumber, i)
            element.ParentContainer = self
            CurrentRow.append(element)
            # if this element is a titlebar, then automatically set the window margins to (0,0) and turn off normal titlebar
            if element.metadata == TITLEBAR_METADATA_MARKER:
                self.Margins = (0, 0)
                self.NoTitleBar = True
        # -------------------------  Append the row to list of Rows  ------------------------- #
        self.Rows.append(CurrentRow)

    # ------------------------- Add Multiple Rows to Form ------------------------- #
    def add_rows(self, rows):
        """
        Loops through a list of lists of elements and adds each row, list, to the layout.
        This is NOT the best way to go about creating a window.  Sending the entire layout at one time and passing
        it as a parameter to the Window call is better.

        :param rows: A list of a list of elements
        :type rows:  List[List[Elements]]
        """
        for row in rows:
            try:
                iter(row)
            except TypeError:
                _error_popup_with_traceback(
                    "Error Creating Window Layout",
                    "Error creating Window layout",
                    "Your row is not an iterable (e.g. a list)",
                    "Instead of a list, the type found was {}".format(type(row)),
                    "The offensive row = ",
                    row,
                    "This item will be stripped from your layout",
                )
                continue
            self.add_row(*row)
        # if _optional_window_data(self) is not None:
        #     self.add_row(_optional_window_data(self))
        if Window._watermark is not None:
            self.add_row(Window._watermark(self))

    def layout(self, rows):
        """
        Second of two preferred ways of telling a Window what its layout is. The other way is to pass the layout as
        a parameter to Window object.  The parameter method is the currently preferred method. This call to Layout
        has been removed from examples contained in documents and in the Demo Programs. Trying to remove this call
        from history and replace with sending as a parameter to Window.

        :param rows: Your entire layout
        :type rows:  List[List[Elements]]
        :return:     self so that you can chain method calls
        :rtype:      (Window)
        """
        if self.use_custom_titlebar and not self.override_custom_titlebar:
            if self.titlebar_icon is not None:
                icon = self.titlebar_icon
            elif CUSTOM_TITLEBAR_ICON is not None:
                icon = CUSTOM_TITLEBAR_ICON
            elif self.titlebar_icon is not None:
                icon = self.titlebar_icon
            elif self.WindowIcon == DEFAULT_WINDOW_ICON:
                icon = DEFAULT_BASE64_ICON_16_BY_16
            else:
                icon = None

            new_rows = [
                           [
                               Titlebar(
                                   title=self.Title,
                                   icon=icon,
                                   text_color=self.titlebar_text_color,
                                   background_color=self.titlebar_background_color,
                                   font=self.titlebar_font,
                               )
                           ]
                       ] + rows
        else:
            new_rows = rows
        self.add_rows(new_rows)
        self._BuildKeyDict()

        if self._has_custom_titlebar_element():
            self.Margins = (0, 0)
            self.NoTitleBar = True
            self._has_custom_titlebar = True
        return self

    def extend_layout(self, container, rows):
        """
        Adds new rows to an existing container element inside of this window
        If the container is a scrollable Column, you need to also call the contents_changed() method

        :param container: The container Element the layout will be placed inside of
        :type container:  Frame | Column | Tab
        :param rows:      The layout to be added
        :type rows:       (List[List[Element]])
        :return:          (Window) self so could be chained
        :rtype:           (Window)
        """
        column = Column(rows, pad=(0, 0), background_color=container.BackgroundColor)
        if self == container:
            frame = self.TKroot
        elif isinstance(container.Widget, TkScrollableFrame):
            frame = container.Widget.TKFrame
        else:
            frame = container.Widget
        PackFormIntoFrame(column, frame, self)
        # sg.PackFormIntoFrame(col, window.TKroot, window)
        self.AddRow(column)
        self.AllKeysDict = self._BuildKeyDictForWindow(self, column, self.AllKeysDict)
        return self

    def LayoutAndRead(self, rows, non_blocking=False):
        """
        Deprecated!!  Now your layout your window's rows (layout) and then separately call Read.

        :param rows:         The layout of the window
        :type rows:          List[List[Element]]
        :param non_blocking: if True the Read call will not block
        :type non_blocking:  (bool)
        """
        _error_popup_with_traceback(
            "LayoutAndRead Depricated",
            "Wow!  You have been using PySimpleGUI for a very long time.",
            "The Window.LayoutAndRead call is no longer supported",
        )

        raise DeprecationWarning(
            "LayoutAndRead is no longer supported... change your call window.Layout(layout).Read()\nor window(title, layout).Read()"
        )
        # self.AddRows(rows)
        # self._Show(non_blocking=non_blocking)
        # return self.ReturnValues

    def LayoutAndShow(self, rows):
        """
        Deprecated - do not use any longer.  Layout your window and then call Read.  Or can add a Finalize call before the Read
        """
        raise DeprecationWarning("LayoutAndShow is no longer supported... ")

    def _Show(self, non_blocking=False):
        """
        NOT TO BE CALLED BY USERS.  INTERNAL ONLY!
        It's this method that first shows the window to the user, collects results

        :param non_blocking: if True, this is a non-blocking call
        :type non_blocking:  (bool)
        :return:             Tuple[Any, Dict] The event, values turple that is returned from Read calls
        :rtype:
        """
        self.Shown = True
        # Compute num rows & num cols (it'll come in handy debugging)
        self.NumRows = len(self.Rows)
        self.NumCols = max(len(row) for row in self.Rows)
        self.NonBlocking = non_blocking

        # Search through entire form to see if any elements set the focus
        # if not, then will set the focus to the first input element
        found_focus = False
        for row in self.Rows:
            for element in row:
                try:
                    if element.Focus:
                        found_focus = True
                except Exception:
                    pass
                try:
                    if element.Key is not None:
                        self.UseDictionary = True
                except Exception:
                    pass

        if not found_focus and self.UseDefaultFocus:
            self.UseDefaultFocus = True
        else:
            self.UseDefaultFocus = False
        # -=-=-=-=-=-=-=-=- RUN the GUI -=-=-=-=-=-=-=-=- ##
        StartupTK(self)
        # If a button or keyboard event happened but no results have been built, build the results
        if self.LastKeyboardEvent is not None or self.LastButtonClicked is not None:
            return _BuildResults(self, False, self)
        return self.ReturnValues

    # ------------------------- SetIcon - set the window's fav icon ------------------------- #
    def set_icon(self, icon=None, pngbase64=None):
        """
        Changes the icon that is shown on the title bar and on the task bar.
        NOTE - The file type is IMPORTANT and depends on the OS!
        Can pass in:
        * filename which must be a .ICO icon file for windows, PNG file for Linux
        * bytes object
        * BASE64 encoded file held in a variable

        :param icon:      Filename or bytes object
        :type icon:       (str)
        :param pngbase64: Base64 encoded image
        :type pngbase64:  (bytes)
        """
        if type(icon) is bytes or pngbase64 is not None:
            wicon = tkinter.PhotoImage(data=icon if icon is not None else pngbase64)
            try:
                self.TKroot.tk.call("wm", "iconphoto", self.TKroot._w, wicon)
            except Exception:
                wicon = tkinter.PhotoImage(data=DEFAULT_BASE64_ICON)
                try:
                    self.TKroot.tk.call("wm", "iconphoto", self.TKroot._w, wicon)
                except Exception:
                    pass
            self.WindowIcon = wicon
            return

        wicon = icon
        try:
            self.TKroot.iconbitmap(icon)
        except Exception:
            try:
                wicon = tkinter.PhotoImage(file=icon)
                self.TKroot.tk.call("wm", "iconphoto", self.TKroot._w, wicon)
            except Exception:
                try:
                    wicon = tkinter.PhotoImage(data=DEFAULT_BASE64_ICON)
                    try:
                        self.TKroot.tk.call("wm", "iconphoto", self.TKroot._w, wicon)
                    except Exception:
                        pass
                except Exception:
                    pass
        self.WindowIcon = wicon

    def _GetElementAtLocation(self, location):
        """
        Given a (row, col) location in a layout, return the element located at that position

        :param location: (int, int) Return the element located at (row, col) in layout
        :type location:
        :return:         (Element) The Element located at that position in this window
        :rtype:
        """

        (row_num, col_num) = location
        row = self.Rows[row_num]
        element = row[col_num]
        return element

    def _GetDefaultElementSize(self):
        """
        Returns the default elementSize

        :return: (width, height) of the default element size
        :rtype:  (int, int)
        """

        return self.DefaultElementSize

    def _AutoCloseAlarmCallback(self):
        """
        Function that's called by tkinter when autoclode timer expires.  Closes the window

        """
        try:
            window = self
            if window:
                if window.NonBlocking:
                    self.Close()
                else:
                    window._Close()
                    self.TKroot.quit()
                    self.RootNeedsDestroying = True
        except Exception:
            pass

    def _TimeoutAlarmCallback(self):
        """
        Read Timeout Alarm callback. Will kick a mainloop call out of the tkinter event loop and cause it to return
        """
        # first, get the results table built
        # modify the Results table in the parent FlexForm object
        # print('TIMEOUT CALLBACK')
        if self.TimerCancelled:
            # print('** timer was cancelled **')
            return
        self.LastButtonClicked = self.TimeoutKey
        self.FormRemainedOpen = True
        self.TKroot.quit()  # kick the users out of the mainloop

    def _calendar_chooser_button_clicked(self, elem):
        """

        :param elem:
        :type elem:
        :return:
        :rtype:
        """
        target_element, strvar, should_submit_window = elem._find_target()

        if elem.calendar_default_date_M_D_Y == (None, None, None):
            now = datetime.datetime.now()
            cur_month, cur_day, cur_year = now.month, now.day, now.year
        else:
            cur_month, cur_day, cur_year = elem.calendar_default_date_M_D_Y

        date_chosen = popup_get_date(
            start_mon=cur_month,
            start_day=cur_day,
            start_year=cur_year,
            close_when_chosen=elem.calendar_close_when_chosen,
            no_titlebar=elem.calendar_no_titlebar,
            begin_at_sunday_plus=elem.calendar_begin_at_sunday_plus,
            locale=elem.calendar_locale,
            location=elem.calendar_location,
            month_names=elem.calendar_month_names,
            day_abbreviations=elem.calendar_day_abbreviations,
            title=elem.calendar_title,
        )
        if date_chosen is not None:
            month, day, year = date_chosen
            now = datetime.datetime.now()
            hour, minute, second = now.hour, now.minute, now.second
            try:
                date_string = calendar.datetime.datetime(
                    year, month, day, hour, minute, second
                ).strftime(elem.calendar_format)
            except Exception as e:
                print("Bad format string in calendar chooser button", e)
                date_string = "Bad format string"

            if target_element is not None and target_element != elem:
                target_element.update(date_string)
            elif target_element == elem:
                elem.calendar_selection = date_string

            strvar.set(date_string)
            elem.TKStringVar.set(date_string)
            if should_submit_window:
                self.LastButtonClicked = target_element.Key
                results = _BuildResults(self, False, self)
        else:
            should_submit_window = False
        return should_submit_window

    # @_timeit_summary
    def read(self, timeout=None, timeout_key=TIMEOUT_KEY, close=False):
        """
        THE biggest deal method in the Window class! This is how you get all of your data from your Window.
            Pass in a timeout (in milliseconds) to wait for a maximum of timeout milliseconds. Will return timeout_key
            if no other GUI events happen first.

        :param timeout:     Milliseconds to wait until the Read will return IF no other GUI events happen first
        :type timeout:      (int)
        :param timeout_key: The value that will be returned from the call if the timer expired
        :type timeout_key:  (Any)
        :param close:       if True the window will be closed prior to returning
        :type close:        (bool)
        :return:            (event, values)
        :rtype:             Tuple[(Any), Dict[Any, Any], List[Any], None]
        """

        if Window._floating_debug_window_build_needed:
            Window._floating_debug_window_build_needed = False
            _Debugger.debugger._build_floating_window()

        if Window._main_debug_window_build_needed:
            Window._main_debug_window_build_needed = False
            _Debugger.debugger._build_main_debugger_window()

        # ensure called only 1 time through a single read cycle
        if not Window._read_call_from_debugger:
            _refresh_debugger()

        # if the user has not added timeout and a debug window is open, then set a timeout for them so the debugger continuously refreshes
        if _debugger_window_is_open() and not Window._read_call_from_debugger:
            if timeout is None or timeout > 3000:
                timeout = 200

        while True:
            Window._root_running_mainloop = self.TKroot
            results = self._read(timeout=timeout, timeout_key=timeout_key)
            if results is not None:
                if results[0] == DEFAULT_WINDOW_SNAPSHOT_KEY:
                    self.save_window_screenshot_to_disk()
                    popup_quick_message(
                        "Saved window screenshot to disk",
                        background_color="#1c1e23",
                        text_color="white",
                        keep_on_top=True,
                        font="_ 30",
                    )
                    continue
            # Post processing for Calendar Chooser Button
            try:
                if (
                        results[0] == timeout_key
                ):  # if a timeout, then not a calendar button
                    break
                elem = self.find_element(
                    results[0], silent_on_error=True
                )  # get the element that caused the event
                if elem.Type == ELEM_TYPE_BUTTON:
                    if elem.BType == BUTTON_TYPE_CALENDAR_CHOOSER:
                        if self._calendar_chooser_button_clicked(
                                elem
                        ):  # returns True if should break out
                            # results[0] = self.LastButtonClicked
                            results = self.ReturnValues
                            break
                        else:
                            continue
                break
            except Exception:
                break  # wasn't a calendar button for sure

        if close:
            self.close()

        return results

    # @_timeit
    def _read(self, timeout=None, timeout_key=TIMEOUT_KEY):
        """
        THE biggest deal method in the Window class! This is how you get all of your data from your Window.
            Pass in a timeout (in milliseconds) to wait for a maximum of timeout milliseconds. Will return timeout_key
            if no other GUI events happen first.

        :param timeout:     Milliseconds to wait until the Read will return IF no other GUI events happen first
        :type timeout:      (int)
        :param timeout_key: The value that will be returned from the call if the timer expired
        :type timeout_key:  (Any)
        :return:            (event, values) (event or timeout_key or None, Dictionary of values or List of values from all elements in the Window)
        :rtype:             Tuple[(Any), Dict[Any, Any], List[Any], None]
        """

        # if there are events in the thread event queue, then return those events before doing anything else.
        if self._queued_thread_event_available():
            self.ReturnValues = results = _BuildResults(self, False, self)
            return results

        if self.finalize_in_progress and self.auto_close_timer_needs_starting:
            self._start_autoclose_timer()
            self.auto_close_timer_needs_starting = False

        timeout = int(timeout) if timeout is not None else None
        if timeout == 0:  # timeout of zero runs the old readnonblocking
            event, values = self._ReadNonBlocking()
            if event is None:
                event = timeout_key
            if values is None:
                event = None
            return event, values  # make event None if values was None and return
        # Read with a timeout
        self.Timeout = timeout
        self.TimeoutKey = timeout_key
        self.NonBlocking = False
        if self.TKrootDestroyed:
            self.read_closed_window_count += 1
            if self.read_closed_window_count > 100:
                popup_error_with_traceback(
                    "Trying to read a closed window",
                    "You have tried 100 times to read a closed window.",
                    "You need to add a check for event == WIN_CLOSED",
                )
            return None, None
        if not self.Shown:
            self._Show()
        else:
            # if already have a button waiting, the return previously built results
            if (
                    self.LastButtonClicked is not None
                    and not self.LastButtonClickedWasRealtime
            ):
                results = _BuildResults(self, False, self)
                self.LastButtonClicked = None
                return results
            InitializeResults(self)

            if self._queued_thread_event_available():
                self.ReturnValues = results = _BuildResults(self, False, self)
                return results

            # if the last button clicked was realtime, emulate a read non-blocking
            # the idea is to quickly return realtime buttons without any blocks until released
            if self.LastButtonClickedWasRealtime:
                # clear the realtime flag if the element is not a button element (for example a graph element that is dragging)
                if self.AllKeysDict.get(self.LastButtonClicked, None):
                    if (
                            self.AllKeysDict.get(self.LastButtonClicked).Type
                            != ELEM_TYPE_BUTTON
                    ):
                        self.LastButtonClickedWasRealtime = False  # stops from generating events until something changes
                else:  # it is possible for the key to not be in the dicitonary because it has a modifier. If so, then clear the realtime button flag
                    self.LastButtonClickedWasRealtime = (
                        False  # stops from generating events until something changes
                    )

                try:
                    rc = self.TKroot.update()
                except Exception:
                    self.TKrootDestroyed = True
                    Window._DecrementOpenCount()
                    # _my_windows.Decrement()
                    # print('ROOT Destroyed')
                results = _BuildResults(self, False, self)
                if results[0] is not None and results[0] != timeout_key:
                    return results
                else:
                    pass

                # else:
                #     print("** REALTIME PROBLEM FOUND **", results)

            if self.RootNeedsDestroying:
                # print('*** DESTROYING really late***')
                try:
                    self.TKroot.destroy()
                except Exception:
                    pass
                # _my_windows.Decrement()
                self.LastButtonClicked = None
                return None, None

            # normal read blocking code....
            if timeout is not None:
                self.TimerCancelled = False
                self.TKAfterID = self.TKroot.after(timeout, self._TimeoutAlarmCallback)
            self.CurrentlyRunningMainloop = True
            # self.TKroot.protocol("WM_DESTROY_WINDOW", self._OnClosingCallback)
            # self.TKroot.protocol("WM_DELETE_WINDOW", self._OnClosingCallback)
            Window._window_running_mainloop = self
            try:
                Window._root_running_mainloop.mainloop()
            except Exception:
                print("**** EXITING ****")
                exit(-1)
            # print('Out main')
            self.CurrentlyRunningMainloop = False
            # if self.LastButtonClicked != TIMEOUT_KEY:
            try:
                self.TKroot.after_cancel(self.TKAfterID)
                del self.TKAfterID
            except Exception:
                pass
                # print('** tkafter cancel failed **')
            self.TimerCancelled = True
            if self.RootNeedsDestroying:
                # print('*** DESTROYING LATE ***')
                try:
                    self.TKroot.destroy()
                except Exception:
                    pass
                Window._DecrementOpenCount()
                # _my_windows.Decrement()
                self.LastButtonClicked = None
                return None, None
            # if form was closed with X
            if (
                    self.LastButtonClicked is None
                    and self.LastKeyboardEvent is None
                    and self.ReturnValues[0] is None
            ):
                Window._DecrementOpenCount()
                # _my_windows.Decrement()
        # Determine return values
        if self.LastKeyboardEvent is not None or self.LastButtonClicked is not None:
            results = _BuildResults(self, False, self)
            if not self.LastButtonClickedWasRealtime:
                self.LastButtonClicked = None
            return results
        else:
            if self._queued_thread_event_available():
                self.ReturnValues = results = _BuildResults(self, False, self)
                return results
            if (
                    not self.XFound
                    and self.Timeout != 0
                    and self.Timeout is not None
                    and self.ReturnValues[0] is None
            ):  # Special Qt case because returning for no reason so fake timeout
                self.ReturnValues = (
                    self.TimeoutKey,
                    self.ReturnValues[1],
                )  # fake a timeout
            elif (
                    not self.XFound and self.ReturnValues[0] is None
            ):  # Return a timeout event... can happen when autoclose used on another window
                # print("*** Faking timeout ***")
                self.ReturnValues = (
                    self.TimeoutKey,
                    self.ReturnValues[1],
                )  # fake a timeout
            return self.ReturnValues

    def _ReadNonBlocking(self):
        """
        Should be NEVER called directly by the user.  The user can call Window.read(timeout=0) to get same effect

        :return: (event, values). (event or timeout_key or None, Dictionary of values or List of values from all elements in the Window)
        :rtype:  Tuple[(Any), Dict[Any, Any] | List[Any] | None]
        """
        if self.TKrootDestroyed:
            try:
                self.TKroot.quit()
                self.TKroot.destroy()
            except Exception:
                pass
                # print('DESTROY FAILED')
            return None, None
        if not self.Shown:
            self._Show(non_blocking=True)
        try:
            rc = self.TKroot.update()
        except Exception:
            self.TKrootDestroyed = True
            Window._DecrementOpenCount()
            # _my_windows.Decrement()
            # print("read failed")
            # return None, None
        if self.RootNeedsDestroying:
            # print('*** DESTROYING LATE ***', self.ReturnValues)
            self.TKroot.destroy()
            Window._DecrementOpenCount()
            # _my_windows.Decrement()
            self.Values = None
            self.LastButtonClicked = None
            return None, None
        return _BuildResults(self, False, self)

    def _start_autoclose_timer(self):
        duration = (
            DEFAULT_AUTOCLOSE_TIME
            if self.AutoCloseDuration is None
            else self.AutoCloseDuration
        )
        self.TKAfterID = self.TKroot.after(
            int(duration * 1000), self._AutoCloseAlarmCallback
        )

    def finalize(self):
        """
        Use this method to cause your layout to built into a real tkinter window.  In reality this method is like
        Read(timeout=0).  It doesn't block and uses your layout to create tkinter widgets to represent the elements.
        Lots of action!

        :return: Returns 'self' so that method "Chaining" can happen (read up about it as it's very cool!)
        :rtype:  (Window)
        """

        if self.TKrootDestroyed:
            return self
        self.finalize_in_progress = True

        self.Read(timeout=1)

        if self.AutoClose:
            self.auto_close_timer_needs_starting = True
        # add the window to the list of active windows
        Window._active_windows[self] = Window.hidden_master_root
        return self
        # OLD CODE FOLLOWS
        if not self.Shown:
            self._Show(non_blocking=True)
        try:
            rc = self.TKroot.update()
        except Exception:
            self.TKrootDestroyed = True
            Window._DecrementOpenCount()
            print("** Finalize failed **")
            # _my_windows.Decrement()
            # return None, None
        return self

    def refresh(self):
        """
        Refreshes the window by calling tkroot.update().  Can sometimes get away with a refresh instead of a Read.
        Use this call when you want something to appear in your Window immediately (as soon as this function is called).
        If you change an element in a window, your change will not be visible until the next call to Window.read
        or a call to Window.refresh()

        :return: `self` so that method calls can be easily "chained"
        :rtype:  (Window)
        """

        if self.TKrootDestroyed:
            return self
        try:
            rc = self.TKroot.update()
        except Exception:
            pass
        return self

    def fill(self, values_dict):
        """
        Fill in elements that are input fields with data based on a 'values dictionary'

        :param values_dict: pairs
        :type values_dict:  (Dict[Any, Any]) - {Element_key : value}
        :return:            returns self so can be chained with other methods
        :rtype:             (Window)
        """

        FillFormWithValues(self, values_dict)
        return self

    def _find_closest_key(self, search_key):
        if not isinstance(search_key, str):
            search_key = str(search_key)
        matches = difflib.get_close_matches(
            search_key, [str(k) for k in self.AllKeysDict.keys()]
        )
        if not len(matches):
            return None
        for k in self.AllKeysDict.keys():
            if matches[0] == str(k):
                return k
        return matches[0] if len(matches) else None

    def FindElement(self, key, silent_on_error=False):
        """
        ** Warning ** This call will eventually be depricated. **

        It is suggested that you modify your code to use the recommended window[key] lookup or the PEP8 compliant window.find_element(key)

        For now, you'll only see a message printed and the call will continue to funcation as before.

        :param key:             Used with window.find_element and with return values to uniquely identify this element
        :type key:              str | int | tuple | object
        :param silent_on_error: If True do not display popup nor print warning of key errors
        :type silent_on_error:  (bool)
        :return:                Return value can be: the Element that matches the supplied key if found; an Error Element if silent_on_error is False; None if silent_on_error True;
        :rtype:                 Element | Error Element | None
        """

        warnings.warn(
            "Use of FindElement is not recommended.\nEither switch to the recommended window[key] format\nor the PEP8 compliant find_element",
            UserWarning,
        )
        print(
            "** Warning - FindElement should not be used to look up elements. window[key] or window.find_element are recommended. **"
        )

        return self.find_element(key, silent_on_error=silent_on_error)

    def find_element(
            self, key, silent_on_error=False, supress_guessing=None, supress_raise=None
    ):
        """
        Find element object associated with the provided key.
        THIS METHOD IS NO LONGER NEEDED to be called by the user

        You can perform the same operation by writing this statement:
        element = window[key]

        You can drop the entire "find_element" function name and use [ ] instead.

        However, if you wish to perform a lookup without error checking, and don't have error popups turned
        off globally, you'll need to make this call so that you can disable error checks on this call.

        find_element is typically used in combination with a call to element's update method (or any other element method!):
        window[key].update(new_value)

        Versus the "old way"
        window.FindElement(key).Update(new_value)

        This call can be abbreviated to any of these:
        find_element = FindElement == Element == Find
        With find_element being the PEP8 compliant call that should be used.

        Rememeber that this call will return None if no match is found which may cause your code to crash if not
        checked for.

        :param key:              Used with window.find_element and with return values to uniquely identify this element
        :type key:               str | int | tuple | object
        :param silent_on_error:  If True do not display popup nor print warning of key errors
        :type silent_on_error:   (bool)
        :param supress_guessing: Override for the global key guessing setting.
        :type supress_guessing:  (bool | None)
        :param supress_raise:    Override for the global setting that determines if a key error should raise an exception
        :type supress_raise:     (bool | None)
        :return:                 Return value can be: the Element that matches the supplied key if found; an Error Element if silent_on_error is False; None if silent_on_error True
        :rtype:                  Element | ErrorElement | None
        """

        key_error = False
        closest_key = None
        supress_guessing = (
            supress_guessing if supress_guessing is not None else SUPPRESS_KEY_GUESSING
        )
        supress_raise = (
            supress_raise if supress_raise is not None else SUPPRESS_RAISE_KEY_ERRORS
        )
        try:
            element = self.AllKeysDict[key]
        except KeyError:
            key_error = True
            closest_key = self._find_closest_key(key)
            if not silent_on_error:
                print(
                    "** Error looking up your element using the key: ",
                    key,
                    "The closest matching key: ",
                    closest_key,
                )
                _error_popup_with_traceback(
                    "Key Error",
                    "Problem finding your key " + str(key),
                    "Closest match = " + str(closest_key),
                    emoji=EMOJI_BASE64_KEY,
                )
                element = ErrorElement(key=key)
            else:
                element = None
            if not supress_raise:
                raise KeyError(key)

        if key_error:
            if not supress_guessing and closest_key is not None:
                element = self.AllKeysDict[closest_key]

        return element

    Element = find_element  # Shortcut function
    Find = find_element  # Shortcut function, most likely not used by many people.
    Elem = find_element  # NEW for 2019!  More laziness... Another shortcut

    def find_element_with_focus(self):
        """
        Returns the Element that currently has focus as reported by tkinter. If no element is found None is returned!
        :return: An Element if one has been found with focus or None if no element found
        :rtype:  Element | None
        """
        element = _FindElementWithFocusInSubForm(self)
        return element

    def widget_to_element(self, widget):
        """
        Returns the element that matches a supplied tkinter widget.
        If no matching element is found, then None is returned.


        :return:    Element that uses the specified widget
        :rtype:     Element | None
        """
        if self.AllKeysDict is None or len(self.AllKeysDict) == 0:
            return None
        for key, element in self.AllKeysDict.items():
            if element.Widget == widget:
                return element
        return None

    def _BuildKeyDict(self):
        """
        Used internally only! Not user callable
        Builds a dictionary containing all elements with keys for this window.
        """
        dict = {}
        self.AllKeysDict = self._BuildKeyDictForWindow(self, self, dict)

    def _BuildKeyDictForWindow(self, top_window, window, key_dict):
        """
        Loop through all Rows and all Container Elements for this window and create the keys for all of them.
        Note that the calls are recursive as all pathes must be walked

        :param top_window: The highest level of the window
        :type top_window:  (Window)
        :param window:     The "sub-window" (container element) to be searched
        :type window:      Column | Frame | TabGroup | Pane | Tab
        :param key_dict:   The dictionary as it currently stands.... used as part of recursive call
        :type key_dict:
        :return:           (dict) Dictionary filled with all keys in the window
        :rtype:
        """
        for row_num, row in enumerate(window.Rows):
            for col_num, element in enumerate(row):
                if element.Type == ELEM_TYPE_COLUMN:
                    key_dict = self._BuildKeyDictForWindow(
                        top_window, element, key_dict
                    )
                if element.Type == ELEM_TYPE_FRAME:
                    key_dict = self._BuildKeyDictForWindow(
                        top_window, element, key_dict
                    )
                if element.Type == ELEM_TYPE_TAB_GROUP:
                    key_dict = self._BuildKeyDictForWindow(
                        top_window, element, key_dict
                    )
                if element.Type == ELEM_TYPE_PANE:
                    key_dict = self._BuildKeyDictForWindow(
                        top_window, element, key_dict
                    )
                if element.Type == ELEM_TYPE_TAB:
                    key_dict = self._BuildKeyDictForWindow(
                        top_window, element, key_dict
                    )
                if (
                        element.Key is None
                ):  # if no key has been assigned.... create one for input elements
                    if element.Type == ELEM_TYPE_BUTTON:
                        element.Key = element.ButtonText
                    elif element.Type == ELEM_TYPE_TAB:
                        element.Key = element.Title
                    if element.Type in (
                            ELEM_TYPE_MENUBAR,
                            ELEM_TYPE_BUTTONMENU,
                            ELEM_TYPE_INPUT_SLIDER,
                            ELEM_TYPE_GRAPH,
                            ELEM_TYPE_IMAGE,
                            ELEM_TYPE_INPUT_CHECKBOX,
                            ELEM_TYPE_INPUT_LISTBOX,
                            ELEM_TYPE_INPUT_COMBO,
                            ELEM_TYPE_INPUT_MULTILINE,
                            ELEM_TYPE_INPUT_OPTION_MENU,
                            ELEM_TYPE_INPUT_SPIN,
                            ELEM_TYPE_INPUT_RADIO,
                            ELEM_TYPE_INPUT_TEXT,
                            ELEM_TYPE_PROGRESS_BAR,
                            ELEM_TYPE_TABLE,
                            ELEM_TYPE_TREE,
                            ELEM_TYPE_TAB_GROUP,
                            ELEM_TYPE_SEPARATOR,
                    ):
                        element.Key = top_window.DictionaryKeyCounter
                        top_window.DictionaryKeyCounter += 1
                if element.Key is not None:
                    if element.Key in key_dict.keys():
                        if (
                                element.Type == ELEM_TYPE_BUTTON
                                and WARN_DUPLICATE_BUTTON_KEY_ERRORS
                        ):  # for Buttons see if should complain
                            warnings.warn(
                                "*** Duplicate key found in your layout {} ***".format(
                                    element.Key
                                ),
                                UserWarning,
                            )
                            warnings.warn(
                                "*** Replaced new key with {} ***".format(
                                    str(element.Key) + str(self.UniqueKeyCounter)
                                )
                            )
                            if not SUPPRESS_ERROR_POPUPS:
                                _error_popup_with_traceback(
                                    "Duplicate key found in your layout",
                                    "Dupliate key: {}".format(element.Key),
                                    "Is being replaced with: {}".format(
                                        str(element.Key) + str(self.UniqueKeyCounter)
                                    ),
                                    "The line of code above shows you which layout, but does not tell you exactly where the element was defined",
                                    "The element type is {}".format(element.Type),
                                )
                        element.Key = str(element.Key) + str(self.UniqueKeyCounter)
                        self.UniqueKeyCounter += 1
                    key_dict[element.Key] = element
        return key_dict

    def element_list(self):
        """
        Returns a list of all elements in the window

        :return: List of all elements in the window and container elements in the window
        :rtype:  List[Element]
        """
        return self._build_element_list()

    def _build_element_list(self):
        """
        Used internally only! Not user callable
        Builds a dictionary containing all elements with keys for this window.
        """
        elem_list = []
        elem_list = self._build_element_list_for_form(self, self, elem_list)
        return elem_list

    def _build_element_list_for_form(self, top_window, window, elem_list):
        """
        Loop through all Rows and all Container Elements for this window and create a list
        Note that the calls are recursive as all pathes must be walked

        :param top_window: The highest level of the window
        :type top_window:  (Window)
        :param window:     The "sub-window" (container element) to be searched
        :type window:      Column | Frame | TabGroup | Pane | Tab
        :param elem_list:  The element list as it currently stands.... used as part of recursive call
        :type elem_list:   ???
        :return:           List of all elements in this sub-window
        :rtype:            List[Element]
        """
        for row_num, row in enumerate(window.Rows):
            for col_num, element in enumerate(row):
                elem_list.append(element)
                if element.Type in (
                        ELEM_TYPE_COLUMN,
                        ELEM_TYPE_FRAME,
                        ELEM_TYPE_TAB_GROUP,
                        ELEM_TYPE_PANE,
                        ELEM_TYPE_TAB,
                ):
                    elem_list = self._build_element_list_for_form(
                        top_window, element, elem_list
                    )
        return elem_list

    def save_to_disk(self, filename):
        """
        Saves the values contained in each of the input areas of the form. Basically saves what would be returned from a call to Read.  It takes these results and saves them to disk using pickle.
         Note that every element in your layout that is to be saved must have a key assigned to it.

        :param filename: Filename to save the values to in pickled form
        :type filename:  str
        """
        try:
            event, values = _BuildResults(self, False, self)
            remove_these = []
            for key in values:
                if self.Element(key).Type == ELEM_TYPE_BUTTON:
                    remove_these.append(key)
            for key in remove_these:
                del values[key]
            with open(filename, "wb") as sf:
                pickle.dump(values, sf)
        except Exception:
            print("*** Error saving Window contents to disk ***")

    def load_from_disk(self, filename):
        """
        Restore values from a previous call to SaveToDisk which saves the returned values dictionary in Pickle format

        :param filename: Pickle Filename to load
        :type filename:  (str)
        """
        try:
            with open(filename, "rb") as df:
                self.Fill(pickle.load(df))
        except Exception:
            print("*** Error loading form to disk ***")

    def get_screen_dimensions(self):
        """
        Get the screen dimensions.  NOTE - you must have a window already open for this to work (blame tkinter not me)

        :return: Tuple containing width and height of screen in pixels
        :rtype:  Tuple[None, None] | Tuple[width, height]
        """

        if self.TKrootDestroyed or self.TKroot is None:
            return Window.get_screen_size()
        screen_width = (
            self.TKroot.winfo_screenwidth()
        )  # get window info to move to middle of screen
        screen_height = self.TKroot.winfo_screenheight()
        return screen_width, screen_height

    def move(self, x, y):
        """
        Move the upper left corner of this window to the x,y coordinates provided
        :param x: x coordinate in pixels
        :type x:  (int)
        :param y: y coordinate in pixels
        :type y:  (int)
        """
        try:
            self.TKroot.geometry("+%s+%s" % (x, y))
            self.config_last_location = (int(x), (int(y)))

        except Exception:
            pass

    def move_to_center(self):
        """
        Recenter your window after it's been moved or the size changed.

        This is a conveinence method. There are no tkinter calls involved, only pure PySimpleGUI API calls.
        """
        if not self._is_window_created("tried Window.move_to_center"):
            return
        screen_width, screen_height = self.get_screen_dimensions()
        win_width, win_height = self.size
        x, y = (screen_width - win_width) // 2, (screen_height - win_height) // 2
        self.move(x, y)

    def minimize(self):
        """
        Minimize this window to the task bar
        """
        if not self._is_window_created("tried Window.minimize"):
            return
        if self.use_custom_titlebar:
            self._custom_titlebar_minimize()
        else:
            self.TKroot.iconify()
        self.maximized = False

    def maximize(self):
        """
        Maximize the window. This is done differently on a windows system versus a linux or mac one.  For non-Windows
        the root attribute '-fullscreen' is set to True.  For Windows the "root" state is changed to "zoomed"
        The reason for the difference is the title bar is removed in some cases when using fullscreen option
        """

        if not self._is_window_created("tried Window.maximize"):
            return
        if not running_linux():
            self.TKroot.state("zoomed")
        else:
            self.TKroot.attributes("-fullscreen", True)
        # this method removes the titlebar too
        # self.TKroot.attributes('-fullscreen', True)
        self.maximized = True

    def normal(self):
        """
        Restore a window to a non-maximized state.  Does different things depending on platform.  See Maximize for more.
        """
        if not self._is_window_created("tried Window.normal"):
            return
        if self.use_custom_titlebar:
            self._custom_titlebar_restore()
        else:
            if self.TKroot.state() == "iconic":
                self.TKroot.deiconify()
            else:
                if not running_linux():
                    self.TKroot.state("normal")
                else:
                    self.TKroot.attributes("-fullscreen", False)
            self.maximized = False

    def _StartMoveUsingControlKey(self, event):
        """
        Used by "Grab Anywhere" style windows. This function is bound to mouse-down. It marks the beginning of a drag.
        :param event: event information passed in by tkinter. Contains x,y position of mouse
        :type event:  (event)
        """
        self._start_move_save_offset(event)
        return

    def _StartMoveGrabAnywhere(self, event):
        """
        Used by "Grab Anywhere" style windows. This function is bound to mouse-down. It marks the beginning of a drag.
        :param event: event information passed in by tkinter. Contains x,y position of mouse
        :type event:  (event)
        """
        if (
                isinstance(event.widget, GRAB_ANYWHERE_IGNORE_THESE_WIDGETS)
                or event.widget in self._grab_anywhere_ignore_these_list
        ) and event.widget not in self._grab_anywhere_include_these_list:
            # print('Found widget to ignore in grab anywhere...')
            return
        self._start_move_save_offset(event)

    def _StartMove(self, event):
        self._start_move_save_offset(event)
        return

    def _StopMove(self, event):
        """
        Used by "Grab Anywhere" style windows. This function is bound to mouse-up. It marks the ending of a drag.
        Sets the position of the window to this final x,y coordinates
        :param event: event information passed in by tkinter. Contains x,y position of mouse
        :type event:  (event)
        """
        return

    def _start_move_save_offset(self, event):
        self._mousex = event.x + event.widget.winfo_rootx()
        self._mousey = event.y + event.widget.winfo_rooty()
        geometry = self.TKroot.geometry()
        location = geometry[geometry.find("+") + 1:].split("+")
        self._startx = int(location[0])
        self._starty = int(location[1])
        self._mouse_offset_x = self._mousex - self._startx
        self._mouse_offset_y = self._mousey - self._starty
        # ------ Move All Windows code ------
        if Window._move_all_windows:
            # print('Moving all')
            for win in Window._active_windows:
                if win == self:
                    continue
                geometry = win.TKroot.geometry()
                location = geometry[geometry.find("+") + 1:].split("+")
                _startx = int(location[0])
                _starty = int(location[1])
                win._mouse_offset_x = event.x_root - _startx
                win._mouse_offset_y = event.y_root - _starty

    def _OnMotionUsingControlKey(self, event):
        self._OnMotion(event)

    def _OnMotionGrabAnywhere(self, event):
        """
        Used by "Grab Anywhere" style windows. This function is bound to mouse motion. It actually moves the window
        :param event: event information passed in by tkinter. Contains x,y position of mouse
        :type event:  (event)
        """
        if (
                isinstance(event.widget, GRAB_ANYWHERE_IGNORE_THESE_WIDGETS)
                or event.widget in self._grab_anywhere_ignore_these_list
        ) and event.widget not in self._grab_anywhere_include_these_list:
            # print('Found widget to ignore in grab anywhere...')
            return

        self._OnMotion(event)

    def _OnMotion(self, event):
        self.TKroot.geometry(
            f"+{event.x_root - self._mouse_offset_x}+{event.y_root - self._mouse_offset_y}"
        )
        # print(f"+{event.x_root}+{event.y_root}")
        # ------ Move All Windows code ------
        try:
            if Window._move_all_windows:
                for win in Window._active_windows:
                    if win == self:
                        continue
                    win.TKroot.geometry(
                        f"+{event.x_root - win._mouse_offset_x}+{event.y_root - win._mouse_offset_y}"
                    )
        except Exception as e:
            print("on motion error", e)

    def _focus_callback(self, event):
        print("Focus event = {} window = {}".format(event, self.Title))

    def _config_callback(self, event):
        """
        Called when a config event happens for the window

        :param event:            From tkinter and is not used
        :type event:             Any
        """
        self.LastButtonClicked = WINDOW_CONFIG_EVENT
        self.FormRemainedOpen = True
        self.user_bind_event = event
        _exit_mainloop(self)

    def _move_callback(self, event):
        """
        Called when a control + arrow key is pressed.
        This is a built-in window positioning key sequence

        :param event:            From tkinter and is not used
        :type event:             Any
        """
        if not self._is_window_created("Tried to move window using arrow keys"):
            return
        x, y = self.current_location()
        if event.keysym == "Up":
            self.move(x, y - 1)
        elif event.keysym == "Down":
            self.move(x, y + 1)
        elif event.keysym == "Left":
            self.move(x - 1, y)
        elif event.keysym == "Right":
            self.move(x + 1, y)

    """
    def _config_callback(self, event):
        new_x = event.x
        new_y = event.y


        if self.not_completed_initial_movement:
            if self.starting_window_position != (new_x, new_y):
                return
            self.not_completed_initial_movement = False
            return

        if not self.saw_00:
            if new_x == 0 and new_y == 0:
                self.saw_00 = True

        # self.config_count += 1
        # if self.config_count < 40:
        #     return

        print('Move LOGIC')

        if self.config_last_size != (event.width, event.height):
            self.config_last_size = (event.width, event.height)

        if self.config_last_location[0] != new_x or self.config_last_location[1] != new_y:
            if self.config_last_location == (None, None):
                self.config_last_location = (new_x, new_y)
                return

        deltax = self.config_last_location[0] - event.x
        deltay = self.config_last_location[1] - event.y
        if deltax == 0 and deltay == 0:
            print('not moving so returning')
            return
        if Window._move_all_windows:
            print('checking all windows')
            for window in Window._active_windows:
                if window == self:
                    continue
                x = window.TKroot.winfo_x() + deltax
                y = window.TKroot.winfo_y() + deltay
                # window.TKroot.geometry("+%s+%s" % (x, y))  # this is what really moves the window
                # window.config_last_location = (x,y)
    """

    def _KeyboardCallback(self, event):
        """
        Window keyboard callback. Called by tkinter.  Will kick user out of the tkinter event loop. Should only be
        called if user has requested window level keyboard events

        :param event: object provided by tkinter that contains the key information
        :type event:  (event)
        """
        self.LastButtonClicked = None
        self.FormRemainedOpen = True
        if event.char != "":
            self.LastKeyboardEvent = event.char
        else:
            self.LastKeyboardEvent = str(event.keysym) + ":" + str(event.keycode)
        # if not self.NonBlocking:
        #     _BuildResults(self, False, self)
        _exit_mainloop(self)

    def _MouseWheelCallback(self, event):
        """
        Called by tkinter when a mouse wheel event has happened. Only called if keyboard events for the window
        have been enabled

        :param event: object sent in by tkinter that has the wheel direction
        :type event:  (event)
        """
        self.LastButtonClicked = None
        self.FormRemainedOpen = True
        self.LastKeyboardEvent = (
            "MouseWheel:Down" if event.delta < 0 or event.num == 5 else "MouseWheel:Up"
        )
        # if not self.NonBlocking:
        #     _BuildResults(self, False, self)
        _exit_mainloop(self)

    def _Close(self, without_event=False):
        """
        The internal close call that does the real work of building. This method basically sets up for closing
        but doesn't destroy the window like the User's version of Close does

        :parm without_event: if True, then do not cause an event to be generated, "silently" close the window
        :type without_event: (bool)
        """

        try:
            self.TKroot.update()
        except Exception:
            pass

        if not self.NonBlocking or not without_event:
            _BuildResults(self, False, self)
        if self.TKrootDestroyed:
            return
        self.TKrootDestroyed = True
        self.RootNeedsDestroying = True
        return

    def close(self):
        """
        Closes window.  Users can safely call even if window has been destroyed.   Should always call when done with
        a window so that resources are properly freed up within your thread.
        """

        try:
            del Window._active_windows[
                self
            ]  # will only be in the list if window was explicitly finalized
        except Exception:
            pass

        try:
            self.TKroot.update()  # On Linux must call update if the user closed with X or else won't actually close the window
        except Exception:
            pass

        self._restore_stdout()
        self._restore_stderr()

        _TimerPeriodic.stop_all_timers_for_window(self)

        if self.TKrootDestroyed:
            return
        try:
            self.TKroot.destroy()
            self.TKroot.update()
            Window._DecrementOpenCount()
        except Exception:
            pass
        # if down to 1 window, try and destroy the hidden window, if there is one
        # if Window.NumOpenWindows == 1:
        #     try:
        #         Window.hidden_master_root.destroy()
        #         Window.NumOpenWindows = 0  # if no hidden window, then this won't execute
        #     except Exception:
        #         pass
        self.TKrootDestroyed = True

        # Free up anything that was held in the layout and the root variables
        self.Rows = None
        self.TKroot = None

    def is_closed(self, quick_check=None):
        """
        Returns True is the window is maybe closed.  Can be difficult to tell sometimes
        NOTE - the call to TKroot.update was taking over 500 ms sometimes so added a flag to bypass the lengthy call.
        :type quick_check:  bool
        :return:            True if the window was closed or destroyed
        :rtype:             (bool)
        """

        if self.TKrootDestroyed or self.TKroot is None:
            return True

        # if performing a quick check only, then skip calling tkinter for performance reasons
        if quick_check:
            return False

        # see if can do an update... if not, then it's been destroyed
        try:
            rc = self.TKroot.update()
        except Exception:
            return True
        return False

    # IT FINALLY WORKED! 29-Oct-2018 was the first time this damned thing got called
    def _OnClosingCallback(self):
        """
        Internally used method ONLY. Not sure callable.  tkinter calls this when the window is closed by clicking X
        """
        # global _my_windows
        # print('Got closing callback', self.DisableClose)
        if self.DisableClose:
            return
        if (
                self.CurrentlyRunningMainloop
        ):  # quit if this is the current mainloop, otherwise don't quit!
            _exit_mainloop(self)
            if self.close_destroys_window:
                self.TKroot.destroy()  # destroy this window
                self.TKrootDestroyed = True
                self.XFound = True
            else:
                self.LastButtonClicked = WINDOW_CLOSE_ATTEMPTED_EVENT
        elif Window._root_running_mainloop == Window.hidden_master_root:
            _exit_mainloop(self)
        else:
            if self.close_destroys_window:
                self.TKroot.destroy()  # destroy this window
                self.XFound = True
            else:
                self.LastButtonClicked = WINDOW_CLOSE_ATTEMPTED_EVENT
        if self.close_destroys_window:
            self.RootNeedsDestroying = True
        self._restore_stdout()
        self._restore_stderr()

    def disable(self):
        """
        Disables window from taking any input from the user
        """
        if not self._is_window_created("tried Window.disable"):
            return
        self.TKroot.attributes("-disabled", 1)
        # self.TKroot.grab_set_global()

    def enable(self):
        """
        Re-enables window to take user input after having it be Disabled previously
        """
        if not self._is_window_created("tried Window.enable"):
            return
        self.TKroot.attributes("-disabled", 0)
        # self.TKroot.grab_release()

    def hide(self):
        """
        Hides the window from the screen and the task bar
        """
        if not self._is_window_created("tried Window.hide"):
            return
        self._Hidden = True
        self.TKroot.withdraw()

    def un_hide(self):
        """
        Used to bring back a window that was previously hidden using the Hide method
        """
        if not self._is_window_created("tried Window.un_hide"):
            return
        if self._Hidden:
            self.TKroot.deiconify()
            self._Hidden = False

    def is_hidden(self):
        """
            Returns True if the window is currently hidden
        :return:    Returns True if the window is currently hidden
        :rtype:     bool
        """
        return self._Hidden

    def disappear(self):
        """
        Causes a window to "disappear" from the screen, but remain on the taskbar. It does this by turning the alpha
        channel to 0.  NOTE that on some platforms alpha is not supported. The window will remain showing on these
        platforms.  The Raspberry Pi for example does not have an alpha setting
        """
        if not self._is_window_created("tried Window.disappear"):
            return
        self.TKroot.attributes("-alpha", 0)

    def reappear(self):
        """
        Causes a window previously made to "Disappear" (using that method). Does this by restoring the alpha channel
        """
        if not self._is_window_created("tried Window.reappear"):
            return
        self.TKroot.attributes("-alpha", 255)

    def set_alpha(self, alpha):
        """
        Sets the Alpha Channel for a window.  Values are between 0 and 1 where 0 is completely transparent

        :param alpha: 0 to 1. 0 is completely transparent.  1 is completely visible and solid (can't see through)
        :type alpha:  (float)
        """
        if not self._is_window_created("tried Window.set_alpha"):
            return
        self._AlphaChannel = alpha
        self.TKroot.attributes("-alpha", alpha)

    @property
    def alpha_channel(self):
        """
        A property that changes the current alpha channel value (internal value)
        :return: the current alpha channel setting according to self, not read directly from tkinter
        :rtype:  (float)
        """
        return self._AlphaChannel

    @alpha_channel.setter
    def alpha_channel(self, alpha):
        """
        The setter method for this "property".
        Planning on depricating so that a Set call is always used by users. This is more in line with the SDK
        :param alpha: 0 to 1. 0 is completely transparent.  1 is completely visible and solid (can't see through)
        :type alpha:  (float)
        """
        if not self._is_window_created("tried Window.alpha_channel"):
            return
        self._AlphaChannel = alpha
        self.TKroot.attributes("-alpha", alpha)

    def bring_to_front(self):
        """
        Brings this window to the top of all other windows (perhaps may not be brought before a window made to "stay
        on top")
        """
        if not self._is_window_created("tried Window.bring_to_front"):
            return
        if running_windows():
            try:
                self.TKroot.wm_attributes("-topmost", 0)
                self.TKroot.wm_attributes("-topmost", 1)
                if not self.KeepOnTop:
                    self.TKroot.wm_attributes("-topmost", 0)
            except Exception as e:
                warnings.warn("Problem in Window.bring_to_front" + str(e), UserWarning)
        else:
            try:
                self.TKroot.lift()
            except Exception:
                pass

    def send_to_back(self):
        """
        Pushes this window to the bottom of the stack of windows. It is the opposite of BringToFront
        """
        if not self._is_window_created("tried Window.send_to_back"):
            return
        try:
            self.TKroot.lower()
        except Exception:
            pass

    def keep_on_top_set(self):
        """
        Sets keep_on_top after a window has been created.  Effect is the same
        as if the window was created with this set.  The Window is also brought
        to the front
        """
        if not self._is_window_created("tried Window.keep_on_top_set"):
            return
        self.KeepOnTop = True
        self.bring_to_front()
        try:
            self.TKroot.wm_attributes("-topmost", 1)
        except Exception as e:
            warnings.warn(
                "Problem in Window.keep_on_top_set trying to set wm_attributes topmost"
                + str(e),
                UserWarning,
            )

    def keep_on_top_clear(self):
        """
        Clears keep_on_top after a window has been created.  Effect is the same
        as if the window was created with this set.
        """
        if not self._is_window_created("tried Window.keep_on_top_clear"):
            return
        self.KeepOnTop = False
        try:
            self.TKroot.wm_attributes("-topmost", 0)
        except Exception as e:
            warnings.warn(
                "Problem in Window.keep_on_top_clear trying to clear wm_attributes topmost"
                + str(e),
                UserWarning,
            )

    def current_location(self, more_accurate=False, without_titlebar=False):
        """
        Get the current location of the window's top left corner.
        Sometimes, depending on the environment, the value returned does not include the titlebar,etc
        A new option, more_accurate, can be used to get the theoretical upper leftmost corner of the window.
        The titlebar and menubar are crated by the OS. It gets really confusing when running in a webpage (repl, trinket)
        Thus, the values can appear top be "off" due to the sometimes unpredictable way the location is calculated.
        If without_titlebar is set then the location of the root x,y is used which should not include the titlebar but
            may be OS dependent.

        :param more_accurate:    If True, will use the window's geometry to get the topmost location with titlebar, menubar taken into account
        :type more_accurate:     (bool)
        :param without_titlebar: If True, return location of top left of main window area without the titlebar (may be OS dependent?)
        :type without_titlebar:  (bool)
        :return:                 The x and y location in tuple form (x,y)
        :rtype:                  Tuple[(int | None), (int | None)]
        """

        if not self._is_window_created("tried Window.current_location"):
            return None, None
        try:
            if without_titlebar:
                x, y = self.TKroot.winfo_rootx(), self.TKroot.winfo_rooty()
            elif more_accurate:
                geometry = self.TKroot.geometry()
                location = geometry[geometry.find("+") + 1:].split("+")
                x, y = int(location[0]), int(location[1])
            else:
                x, y = int(self.TKroot.winfo_x()), int(self.TKroot.winfo_y())
        except Exception as e:
            warnings.warn(
                "Error in Window.current_location. Trouble getting x,y location\n"
                + str(e),
                UserWarning,
            )
            x, y = (None, None)
        return x, y

    def current_size_accurate(self):
        """
        Get the current location of the window based on tkinter's geometry setting

        :return:              The x and y size in tuple form (x,y)
        :rtype:               Tuple[(int | None), (int | None)]
        """

        if not self._is_window_created("tried Window.current_location"):
            return None, None
        try:
            geometry = self.TKroot.geometry()
            geometry_tuple = geometry.split("+")
            window_size = geometry_tuple[0].split("x")
            x, y = int(window_size[0]), int(window_size[1])
        except Exception as e:
            warnings.warn(
                "Error in Window.current_size_accurate. Trouble getting x,y size\n{} {}".format(
                    geometry, geometry_tuple
                )
                + str(e),
                UserWarning,
            )
            x, y = (None, None)
        return x, y

    @property
    def size(self):
        """
        Return the current size of the window in pixels

        :return: (width, height) of the window
        :rtype:  Tuple[(int), (int)] or Tuple[None, None]
        """
        if not self._is_window_created("Tried to use Window.size property"):
            return None, None
        win_width = self.TKroot.winfo_width()
        win_height = self.TKroot.winfo_height()
        return win_width, win_height

    @size.setter
    def size(self, size):
        """
        Changes the size of the window, if possible

        :param size: (width, height) of the desired window size
        :type size:  (int, int)
        """
        try:
            self.TKroot.geometry("%sx%s" % (size[0], size[1]))
            self.TKroot.update_idletasks()
        except Exception:
            pass

    def set_size(self, size):
        """
        Changes the size of the window, if possible. You can also use the Window.size prooerty
        to set/get the size.

        :param size: (width, height) of the desired window size
        :type size:  (int, int)
        """
        if not self._is_window_created(
                "Tried to change the size of the window prior to creation."
        ):
            return
        try:
            self.TKroot.geometry("%sx%s" % (size[0], size[1]))
            self.TKroot.update_idletasks()
        except Exception:
            pass

    def set_min_size(self, size):
        """
        Changes the minimum size of the window. Note Window must be read or finalized first.

        :param size: (width, height) tuple (int, int) of the desired window size in pixels
        :type size:  (int, int)
        """
        if not self._is_window_created("tried Window.set_min_size"):
            return
        self.TKroot.minsize(size[0], size[1])
        self.TKroot.update_idletasks()

    def set_resizable(self, x_axis_enable, y_axis_enable):
        """
        Changes if a window can be resized in either the X or the Y direction.
        Note Window must be read or finalized first.

        :param x_axis_enable: If True, the window can be changed in the X-axis direction. If False, it cannot
        :type x_axis_enable: (bool)
        :param y_axis_enable: If True, the window can be changed in the Y-axis direction. If False, it cannot
        :type y_axis_enable: (bool)
        """

        if not self._is_window_created("tried Window.set_resixable"):
            return
        try:
            self.TKroot.resizable(x_axis_enable, y_axis_enable)
        except Exception as e:
            _error_popup_with_traceback(
                "Window.set_resizable - tkinter reported error", e
            )

    def visibility_changed(self):
        """
        When making an element in a column or someplace that has a scrollbar, then you'll want to call this function
        prior to the column's contents_changed() method.
        """
        self.refresh()

    def set_transparent_color(self, color):
        """
        Set the color that will be transparent in your window. Areas with this color will be SEE THROUGH.

        :param color: Color string that defines the transparent color
        :type color:  (str)
        """
        if not self._is_window_created("tried Window.set_transparent_color"):
            return
        try:
            self.TKroot.attributes("-transparentcolor", color)
            self.TransparentColor = color
        except Exception:
            print("Transparent color not supported on this platform (windows only)")

    def mouse_location(self):
        """
        Return the (x,y) location of the mouse relative to the entire screen.  It's the same location that
        you would use to create a window, popup, etc.

        :return:    The location of the mouse pointer
        :rtype:     (int, int)
        """
        if not self._is_window_created("tried Window.mouse_location"):
            return 0, 0

        return self.TKroot.winfo_pointerx(), self.TKroot.winfo_pointery()

    def grab_any_where_on(self):
        """
        Turns on Grab Anywhere functionality AFTER a window has been created.  Don't try on a window that's not yet
        been Finalized or Read.
        """
        if not self._is_window_created("tried Window.grab_any_where_on"):
            return
        self.TKroot.bind("<ButtonPress-1>", self._StartMoveGrabAnywhere)
        self.TKroot.bind("<ButtonRelease-1>", self._StopMove)
        self.TKroot.bind("<B1-Motion>", self._OnMotionGrabAnywhere)

    def grab_any_where_off(self):
        """
        Turns off Grab Anywhere functionality AFTER a window has been created.  Don't try on a window that's not yet
        been Finalized or Read.
        """
        if not self._is_window_created("tried Window.grab_any_where_off"):
            return
        self.TKroot.unbind("<ButtonPress-1>")
        self.TKroot.unbind("<ButtonRelease-1>")
        self.TKroot.unbind("<B1-Motion>")

    def _user_bind_callback(self, bind_string, event, propagate=True):
        """
        Used when user binds a tkinter event directly to an element

        :param bind_string: The event that was bound so can lookup the key modifier
        :type bind_string:  (str)
        :param event:       Event data passed in by tkinter (not used)
        :type event:
        :param propagate:   If True then tkinter will be told to propagate the event
        :type propagate:    (bool)
        """
        # print('bind callback', bind_string, event)
        key = self.user_bind_dict.get(bind_string, "")
        self.user_bind_event = event
        if key is not None:
            self.LastButtonClicked = key
        else:
            self.LastButtonClicked = bind_string
        self.FormRemainedOpen = True
        # if self.CurrentlyRunningMainloop:
        #     self.TKroot.quit()
        _exit_mainloop(self)
        return "break" if propagate is not True else None

    def bind(self, bind_string, key, propagate=True):
        """
        Used to add tkinter events to a Window.
        The tkinter specific data is in the Window's member variable user_bind_event
        :param bind_string: The string tkinter expected in its bind function
        :type bind_string:  (str)
        :param key:         The event that will be generated when the tkinter event occurs
        :type key:          str | int | tuple | object
        :param propagate:   If True then tkinter will be told to propagate the event
        :type propagate:    (bool)
        """
        if not self._is_window_created("tried Window.bind"):
            return
        try:
            self.TKroot.bind(
                bind_string,
                lambda evt: self._user_bind_callback(bind_string, evt, propagate),
            )
        except Exception as e:
            self.TKroot.unbind_all(bind_string)
            return
            # _error_popup_with_traceback('Window.bind error', e)
        self.user_bind_dict[bind_string] = key

    def unbind(self, bind_string):
        """
        Used to remove tkinter events to a Window.
        This implementation removes ALL of the binds of the bind_string from the Window.  If there
        are multiple binds for the Window itself, they will all be removed.  This can be extended later if there
        is a need.
        :param bind_string: The string tkinter expected in its bind function
        :type bind_string:  (str)
        """
        if not self._is_window_created("tried Window.unbind"):
            return
        self.TKroot.unbind(bind_string)

    def _callback_main_debugger_window_create_keystroke(self, event):
        """
        Called when user presses the key that creates the main debugger window
        March 2022 - now causes the user reads to return timeout events automatically
        :param event: (event) not used. Passed in event info
        :type event:
        """
        Window._main_debug_window_build_needed = True
        # exit the event loop in a way that resembles a timeout occurring
        self.LastButtonClicked = self.TimeoutKey
        self.FormRemainedOpen = True
        self.TKroot.quit()  # kick the users out of the mainloop

    def _callback_popout_window_create_keystroke(self, event):
        """
        Called when user presses the key that creates the floating debugger window
        March 2022 - now causes the user reads to return timeout events automatically
        :param event: (event) not used. Passed in event info
        :type event:
        """
        Window._floating_debug_window_build_needed = True
        # exit the event loop in a way that resembles a timeout occurring
        self.LastButtonClicked = self.TimeoutKey
        self.FormRemainedOpen = True
        self.TKroot.quit()  # kick the users out of the mainloop

    def enable_debugger(self):
        """
        Enables the internal debugger. By default, the debugger IS enabled
        """
        if not self._is_window_created("tried Window.enable_debugger"):
            return
        self.TKroot.bind(
            "<Cancel>", self._callback_main_debugger_window_create_keystroke
        )
        self.TKroot.bind("<Pause>", self._callback_popout_window_create_keystroke)
        self.DebuggerEnabled = True

    def disable_debugger(self):
        """
        Disable the internal debugger. By default the debugger is ENABLED
        """
        if not self._is_window_created("tried Window.disable_debugger"):
            return
        self.TKroot.unbind("<Cancel>")
        self.TKroot.unbind("<Pause>")
        self.DebuggerEnabled = False

    def set_title(self, title):
        """
        Change the title of the window

        :param title: The string to set the title to
        :type title:  (str)
        """
        if not self._is_window_created("tried Window.set_title"):
            return
        if self._has_custom_titlebar:
            try:  # just in case something goes badly, don't crash
                self.find_element(TITLEBAR_TEXT_KEY).update(title)
            except Exception:
                pass
        # even with custom titlebar, set the main window's title too so it'll match when minimized
        self.TKroot.wm_title(str(title))

    def make_modal(self):
        """
        Makes a window into a "Modal Window"
        This means user will not be able to interact with other windows until this one is closed

        NOTE - Sorry Mac users - you can't have modal windows.... lobby your tkinter Mac devs
        """
        if not self._is_window_created("tried Window.make_modal"):
            return

        if running_mac() and ENABLE_MAC_MODAL_DISABLE_PATCH:
            return

        # if modal windows have been disabled globally
        if not DEFAULT_MODAL_WINDOWS_ENABLED and not DEFAULT_MODAL_WINDOWS_FORCED:
            # if not DEFAULT_MODAL_WINDOWS_ENABLED:
            return

        try:
            self.TKroot.transient()
            self.TKroot.grab_set()
            self.TKroot.focus_force()
        except Exception as e:
            print("Exception trying to make modal", e)

    def force_focus(self):
        """
        Forces this window to take focus
        """
        if not self._is_window_created("tried Window.force_focus"):
            return
        self.TKroot.focus_force()

    def was_closed(self):
        """
        Returns True if the window was closed

        :return: True if the window is closed
        :rtype:  bool
        """
        return self.TKrootDestroyed

    def set_cursor(self, cursor):
        """
        Sets the cursor for the window.
        If you do not want any mouse pointer, then use the string "none"

        :param cursor: The tkinter cursor name
        :type cursor:  (str)
        """

        if not self._is_window_created("tried Window.set_cursor"):
            return
        try:
            self.TKroot.config(cursor=cursor)
        except Exception as e:
            print("Warning bad cursor specified ", cursor)
            print(e)

    def ding(self, display_number=0):
        """
        Make a "bell" sound. A capability provided by tkinter.  Your window needs to be finalized prior to calling.
        Ring a display's bell is the tkinter description of the call.
        :param display_number: Passed to tkinter's bell method as parameter "displayof".
        :type display_number:  int
        """
        if not self._is_window_created("tried Window.ding"):
            return
        try:
            self.TKroot.bell(display_number)
        except Exception as e:
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "Window.ding() - tkinter reported error from bell() call", e
                )

    def _window_tkvar_changed_callback(self, *args):
        """
        Internal callback function for when the thread

        :param event: Information from tkinter about the callback
        :type event:

        """
        # print('Thread callback info', threading.current_thread())
        # print(event)
        # trace_details = traceback.format_stack()
        # print(''.join(trace_details))
        # self.thread_lock.acquire()
        # if self.thread_timer:
        # self.TKroot.after_cancel(id=self.thread_timer)
        # self.thread_timer = None
        # self.thread_lock.release()

        if self._queued_thread_event_available():
            self.FormRemainedOpen = True
            _exit_mainloop(self)

    def _create_thread_queue(self):
        """
        Creates the queue used by threads to communicate with this window
        """

        if self.thread_queue is None:
            self.thread_queue = queue.Queue()

        if self.thread_lock is None:
            self.thread_lock = threading.Lock()

        if self.thread_strvar is None:
            self.thread_strvar = tk.StringVar()
            self.thread_strvar.trace("w", self._window_tkvar_changed_callback)

    def write_event_value(self, key, value):
        """
        Adds a key & value tuple to the queue that is used by threads to communicate with the window

        :param key:   The key that will be returned as the event when reading the window
        :type key:    Any
        :param value: The value that will be in the values dictionary
        :type value:  Any
        """

        if self.thread_queue is None:
            print("*** Warning Window.write_event_value - no thread queue found ***")
            return
        # self.thread_lock.acquire()  # first lock the critical section
        self.thread_queue.put(item=(key, value))
        self.TKroot.tk.willdispatch()  # brilliant bit of code provided by Giuliano who I owe a million thank yous!
        self.thread_strvar.set("new item")

        # self.thread_queue.put(item=(key, value))
        # self.thread_strvar.set('new item')
        # March 28 2021 - finally found a solution!  It needs a little more work and a lock
        # if no timer is running, then one should be started
        # if self.thread_timer is None:
        #     print('Starting a timer')
        #     self.thread_timer = self.TKroot.after(1, self._window_tkvar_changed_callback)
        # self.thread_lock.release()

    def _queued_thread_event_read(self):
        if self.thread_queue is None:
            return None

        try:  # see if something has been posted to Queue
            message = self.thread_queue.get_nowait()
        except queue.Empty:  # get_nowait() will get exception when Queue is empty
            return None

        return message

    def _queued_thread_event_available(self):
        if self.thread_queue is None:
            return False
        # self.thread_lock.acquire()
        qsize = self.thread_queue.qsize()
        if qsize == 0:
            self.thread_timer = None
        # self.thread_lock.release()
        return qsize != 0

    def _RightClickMenuCallback(self, event):
        """
        When a right click menu is specified for an entire window, then this callback catches right clicks
        that happen to the window itself, when there are no elements that are in that area.

        The only portion that is not currently covered correctly is the row frame itself.  There will still
        be parts of the window, at the moment, that don't respond to a right click.  It's getting there, bit
        by bit.

        Callback function that's called when a right click happens. Shows right click menu as result.

        :param event: information provided by tkinter about the event including x,y location of click
        :type event:
        """
        # if there are widgets under the mouse, then see if it's the root only.  If not, then let the widget (element) show their menu instead
        x, y = self.TKroot.winfo_pointerxy()
        widget = self.TKroot.winfo_containing(x, y)
        if widget != self.TKroot:
            return
        self.TKRightClickMenu.tk_popup(event.x_root, event.y_root, 0)
        self.TKRightClickMenu.grab_release()

    def save_window_screenshot_to_disk(self, filename=None):
        """
        Saves an image of the PySimpleGUI window provided into the filename provided

        :param filename:        Optional filename to save screenshot to. If not included, the User Settinds are used to get the filename
        :return:                A PIL ImageGrab object that can be saved or manipulated
        :rtype:                 (PIL.ImageGrab | None)
        """
        global pil_import_attempted, pil_imported, PIL, ImageGrab, Image

        if not pil_import_attempted:
            try:
                import PIL as PIL
                from PIL import ImageGrab
                from PIL import Image

                pil_imported = True
                pil_import_attempted = True
            except Exception:
                pil_imported = False
                pil_import_attempted = True
                print("FAILED TO IMPORT PIL!")
                return None
        try:
            # Get location of window to save
            pos = self.current_location()
            # Add a little to the X direction if window has a titlebar
            if not self.NoTitleBar:
                pos = (pos[0] + 7, pos[1])
            # Get size of wiondow
            size = self.current_size_accurate()
            # Get size of the titlebar
            titlebar_height = self.TKroot.winfo_rooty() - self.TKroot.winfo_y()
            # Add titlebar to size of window so that titlebar and window will be saved
            size = (size[0], size[1] + titlebar_height)
            if not self.NoTitleBar:
                size_adjustment = (2, 1)
            else:
                size_adjustment = (0, 0)
            # Make the "Bounding rectangle" used by PLK to do the screen grap "operation
            rect = (
                pos[0],
                pos[1],
                pos[0] + size[0] + size_adjustment[0],
                pos[1] + size[1] + size_adjustment[1],
            )
            # Grab the image
            grab = ImageGrab.grab(bbox=rect)
            # Save the grabbed image to disk
        except Exception as e:
            # print(e)
            popup_error_with_traceback(
                "Screen capture failure",
                "Error happened while trying to save screencapture",
                e,
            )

            return None
        # return grab
        if filename is None:
            folder = pysimplegui_user_settings.get("-screenshots folder-", "")
            filename = pysimplegui_user_settings.get("-screenshots filename-", "")
            full_filename = os.path.join(folder, filename)
        else:
            full_filename = filename
        if full_filename:
            try:
                grab.save(full_filename)
            except Exception as e:
                popup_error_with_traceback(
                    "Screen capture failure",
                    "Error happened while trying to save screencapture",
                    e,
                )
        else:
            popup_error_with_traceback(
                "Screen capture failure",
                "You have attempted a screen capture but have not set up a good filename to save to",
            )
        return grab

    def perform_long_operation(self, func, end_key=None):
        """
        Call your function that will take a long time to execute.  When it's complete, send an event
        specified by the end_key.

        Starts a thread on your behalf.

        This is a way for you to "ease into" threading without learning the details of threading.
        Your function will run, and when it returns 2 things will happen:
        1. The value you provide for end_key will be returned to you when you call window.read()
        2. If your function returns a value, then the value returned will also be included in your windows.read call in the values dictionary

        IMPORTANT - This method uses THREADS... this means you CANNOT make any PySimpleGUI calls from
        the function you provide with the exception of one function, Window.write_event_value.

        :param func:    A lambda or a function name with no parms
        :type func:     Any
        :param end_key: Optional key that will be generated when the function returns
        :type end_key:  (Any | None)
        :return:        The id of the thread
        :rtype:         threading.Thread
        """

        thread = threading.Thread(
            target=_long_func_thread, args=(self, end_key, func), daemon=True
        )
        thread.start()
        return thread

    @property
    def key_dict(self):
        """
        Returns a dictionary with all keys and their corresponding elements
        { key : Element }
        :return: Dictionary of keys and elements
        :rtype:  Dict[Any, Element]
        """
        return self.AllKeysDict

    def key_is_good(self, key):
        """
        Checks to see if this is a good key for this window
        If there's an element with the key provided, then True is returned
        :param key:     The key to check
        :type key:      str | int | tuple | object
        :return:        True if key is an element in this window
        :rtype:         bool
        """
        if key in self.key_dict:
            return True
        return False

    def get_scaling(self):
        """
        Returns the current scaling value set for this window

        :return:    Scaling according to tkinter. Returns DEFAULT_SCALING if error
        :rtype:     float
        """

        if not self._is_window_created("Tried Window.set_scaling"):
            return DEFAULT_SCALING
        try:
            scaling = self.TKroot.tk.call("tk", "scaling")
        except Exception as e:
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "Window.get_scaling() - tkinter reported error", e
                )
            scaling = DEFAULT_SCALING

        return scaling

    def _custom_titlebar_restore_callback(self, event):
        self._custom_titlebar_restore()

    def _custom_titlebar_restore(self):
        if running_linux():
            # if self._skip_first_restore_callback:
            #     self._skip_first_restore_callback = False
            #     return
            self.TKroot.unbind("<Button-1>")
            self.TKroot.deiconify()

            # self.ParentForm.TKroot.wm_overrideredirect(True)
            self.TKroot.wm_attributes("-type", "dock")

        else:
            self.TKroot.unbind("<Expose>")
            self.TKroot.wm_overrideredirect(True)
        if self.TKroot.state() == "iconic":
            self.TKroot.deiconify()
        else:
            if not running_linux():
                self.TKroot.state("normal")
            else:
                self.TKroot.attributes("-fullscreen", False)
        self.maximized = False

    def _custom_titlebar_minimize(self):
        if running_linux():
            self.TKroot.wm_attributes("-type", "normal")
            # self.ParentForm.TKroot.state('icon')
            # return
            # self.ParentForm.maximize()
            self.TKroot.wm_overrideredirect(False)
            # self.ParentForm.minimize()
            # self.ParentForm.TKroot.wm_overrideredirect(False)
            self.TKroot.iconify()
            # self._skip_first_restore_callback = True
            self.TKroot.bind("<Button-1>", self._custom_titlebar_restore_callback)
        else:
            self.TKroot.wm_overrideredirect(False)
            self.TKroot.iconify()
            self.TKroot.bind("<Expose>", self._custom_titlebar_restore_callback)

    def _custom_titlebar_callback(self, key):
        """
        One of the Custom Titlbar buttons was clicked
        :param key:
        :return:
        """
        if key == TITLEBAR_MINIMIZE_KEY:
            if not self.DisableMinimize:
                self._custom_titlebar_minimize()
        elif key == TITLEBAR_MAXIMIZE_KEY:
            if self.Resizable:
                if self.maximized:
                    self.normal()
                else:
                    self.maximize()
        elif key == TITLEBAR_CLOSE_KEY:
            if not self.DisableClose:
                self._OnClosingCallback()

    def timer_start(self, frequency_ms, key=EVENT_TIMER, repeating=True):
        """
        Starts a timer that gnerates Timer Events.  The default is to repeat the timer events until timer is stopped.
        You can provide your own key or a default key will be used.  The default key is defined
        with the constants EVENT_TIMER or TIMER_KEY.  They both equal the same value.
        The values dictionary will contain the timer ID that is returned from this function.

        :param frequency_ms:    How often to generate timer events in milliseconds
        :type frequency_ms:     int
        :param key:             Key to be returned as the timer event
        :type key:              str | int | tuple | object
        :param repeating:       If True then repeat timer events until timer is explicitly stopped
        :type repeating:        bool
        :return:                Timer ID for the timer
        :rtype:                 int
        """
        timer = _TimerPeriodic(
            self, frequency_ms=frequency_ms, key=key, repeating=repeating
        )
        return timer.id

    def timer_stop(self, timer_id):
        """
        Stops a timer with a given ID

        :param timer_id:        Timer ID of timer to stop
        :type timer_id:         int
        :return:
        """
        _TimerPeriodic.stop_timer_with_id(timer_id)

    def timer_stop_all(self):
        """
        Stops all timers for THIS window
        """
        _TimerPeriodic.stop_all_timers_for_window(self)

    def timer_get_active_timers(self):
        """
        Returns a list of currently active timers for a window
        :return:    List of timers for the window
        :rtype:     List[int]
        """
        return _TimerPeriodic.get_all_timers_for_window(self)

    @classmethod
    def _restore_stdout(cls):
        for item in cls._rerouted_stdout_stack:
            (window, element) = item  # type: (Window, Element)
            if not window.is_closed():
                sys.stdout = element
                break
        cls._rerouted_stdout_stack = [
            item for item in cls._rerouted_stdout_stack if not item[0].is_closed()
        ]
        if len(cls._rerouted_stdout_stack) == 0 and cls._original_stdout is not None:
            sys.stdout = cls._original_stdout
        # print('Restored stdout... new stack:',  [item[0].Title for item in cls._rerouted_stdout_stack ])

    @classmethod
    def _restore_stderr(cls):
        for item in cls._rerouted_stderr_stack:
            (window, element) = item  # type: (Window, Element)
            if not window.is_closed():
                sys.stderr = element
                break
        cls._rerouted_stderr_stack = [
            item for item in cls._rerouted_stderr_stack if not item[0].is_closed()
        ]
        if len(cls._rerouted_stderr_stack) == 0 and cls._original_stderr is not None:
            sys.stderr = cls._original_stderr
        # print('Restored stderr... new stack:',  [item[0].Title for item in cls._rerouted_stderr_stack ])

    # def __enter__(self):
    #     """
    #     WAS used with context managers which are no longer needed nor advised.  It is here for legacy support and
    #     am afraid of removing right now
    #     :return: (window)
    #      :rtype:
    #     """
    #     return self

    def __getitem__(self, key):
        """
        Returns Element that matches the passed in key.
        This is "called" by writing code as thus:
        window['element key'].update

        :param key: The key to find
        :type key:  str | int | tuple | object
        :return:    The element found
        :rtype:     Element | Input | Combo | OptionMenu | Listbox | Radio | Checkbox | Spin | Multiline | Text | StatusBar | Output | Button | ButtonMenu | ProgressBar | Image | Canvas | Graph | Frame | VerticalSeparator | HorizontalSeparator | Tab | TabGroup | Slider | Column | Pane | Menu | Table | Tree | ErrorElement | None
        """

        return self.find_element(key)

    def __call__(self, *args, **kwargs):
        """
        Call window.read but without having to type it out.
        window() == window.read()
        window(timeout=50) == window.read(timeout=50)

        :return: The famous event, values that read returns.
        :rtype:  Tuple[Any, Dict[Any, Any]]
        """
        return self.read(*args, **kwargs)

    def _is_window_created(self, additional_message=""):
        msg = str(additional_message)
        if self.TKroot is None:
            warnings.warn(
                'You cannot perform operations on a Window until it is read or finalized. Adding a "finalize=True" parameter to your Window creation will fix this. '
                + msg,
                UserWarning,
            )
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "You cannot perform operations on a Window until it is read or finalized.",
                    'Adding a "finalize=True" parameter to your Window creation will likely fix this',
                    msg,
                )
            return False
        return True

    def _has_custom_titlebar_element(self):
        for elem in self.AllKeysDict.values():
            if elem.Key in (
                    TITLEBAR_MAXIMIZE_KEY,
                    TITLEBAR_CLOSE_KEY,
                    TITLEBAR_IMAGE_KEY,
            ):
                return True
            if elem.metadata == TITLEBAR_METADATA_MARKER:
                return True
        return False

    AddRow = add_row
    AddRows = add_rows
    AlphaChannel = alpha_channel
    BringToFront = bring_to_front
    Close = close
    CurrentLocation = current_location
    Disable = disable
    DisableDebugger = disable_debugger
    Disappear = disappear
    Enable = enable
    EnableDebugger = enable_debugger
    Fill = fill
    Finalize = finalize
    # FindElement = find_element
    FindElementWithFocus = find_element_with_focus
    GetScreenDimensions = get_screen_dimensions
    GrabAnyWhereOff = grab_any_where_off
    GrabAnyWhereOn = grab_any_where_on
    Hide = hide
    Layout = layout
    LoadFromDisk = load_from_disk
    Maximize = maximize
    Minimize = minimize
    Move = move
    Normal = normal
    Read = read
    Reappear = reappear
    Refresh = refresh
    SaveToDisk = save_to_disk
    SendToBack = send_to_back
    SetAlpha = set_alpha
    SetIcon = set_icon
    SetTransparentColor = set_transparent_color
    Size = size
    UnHide = un_hide
    VisibilityChanged = visibility_changed
    CloseNonBlocking = close
    CloseNonBlockingForm = close
    start_thread = perform_long_operation
    #
    # def __exit__(self, *a):
    #     """
    #     WAS used with context managers which are no longer needed nor advised.  It is here for legacy support and
    #     am afraid of removing right now
    #     :param *a: (?) Not sure what's passed in.
    #      :type *a:
    #     :return:   Always returns False which was needed for context manager to work
    #      :rtype:
    #     """
    #     self.__del__()
    #     return False
    #
    # def __del__(self):
    #     # print('DELETING WINDOW')
    #     for row in self.Rows:
    #         for element in row:
    #             element.__del__()


# -------------------------------- PEP8-ify the Window Class USER Interfaces -------------------------------- #


FlexForm = Window


def _long_func_thread(window, end_key, original_func):
    """
    Used to run long operations on the user's behalf. Called by the window object

    :param window:        The window that will get the event
    :type window:         (Window)
    :param end_key:       The event that will be sent when function returns. If None then no event will be sent when exiting thread
    :type end_key:        (Any|None)
    :param original_func: The user's function that is called. Can be a function with no arguments or a lambda experession
    :type original_func:  (Any)
    """

    return_value = original_func()
    if end_key is not None:
        window.write_event_value(end_key, return_value)


def _exit_mainloop(exiting_window):
    if (
            exiting_window == Window._window_running_mainloop
            or Window._root_running_mainloop == Window.hidden_master_root
    ):
        Window._window_that_exited = exiting_window
        if Window._root_running_mainloop is not None:
            Window._root_running_mainloop.quit()
        # print('** Exited window mainloop **')


def _timeout_alarm_callback_hidden():
    """
    Read Timeout Alarm callback. Will kick a mainloop call out of the tkinter event loop and cause it to return
    """

    del Window._TKAfterID

    # first, get the results table built
    # modify the Results table in the parent FlexForm object
    # print('TIMEOUT CALLBACK')
    Window._root_running_mainloop.quit()  # kick the users out of the mainloop

    # Get window that caused return
    Window._window_that_exited = None


def read_all_windows(timeout=None, timeout_key=TIMEOUT_KEY):
    """
    Reads all windows that are "active" when the call is made. "Active" means that it's been finalized or read.
    If a window has not been finalized then it will not be considered an "active window"

    If any of the active windows returns a value then the window and its event and values
    are returned.

    If no windows are open, then the value (None, WIN_CLOSED, None) will be returned
        Since WIN_CLOSED is None, it means (None, None, None) is what's returned when no windows remain opened

    :param timeout:     Time in milliseconds to delay before a returning a timeout event
    :type timeout:      (int)
    :param timeout_key: Key to return when a timeout happens. Defaults to the standard TIMEOUT_KEY
    :type timeout_key:  (Any)
    :return:            A tuple with the  (Window, event, values dictionary/list)
    :rtype:             (Window, Any, Dict | List)
    """

    if len(Window._active_windows) == 0:
        return None, WIN_CLOSED, None

    # first see if any queued events are waiting for any of the windows
    for window in Window._active_windows.keys():
        if window._queued_thread_event_available():
            _BuildResults(window, False, window)
            event, values = window.ReturnValues
            return window, event, values

    Window._root_running_mainloop = Window.hidden_master_root
    Window._timeout_key = timeout_key

    if timeout == 0:
        window = list(Window._active_windows.keys())[Window._timeout_0_counter]
        event, values = window._ReadNonBlocking()
        if event is None:
            event = timeout_key
        if values is None:
            event = None
        Window._timeout_0_counter = (Window._timeout_0_counter + 1) % len(
            Window._active_windows
        )
        return window, event, values

    Window._timeout_0_counter = (
        0  # reset value if not reading with timeout 0 so ready next time needed
    )

    # setup timeout timer
    if timeout is not None:
        try:
            Window.hidden_master_root.after_cancel(Window._TKAfterID)
            del Window._TKAfterID
        except Exception:
            pass

        Window._TKAfterID = Window.hidden_master_root.after(
            timeout, _timeout_alarm_callback_hidden
        )

    # ------------ Call Mainloop ------------
    Window._root_running_mainloop.mainloop()

    try:
        Window.hidden_master_root.after_cancel(Window._TKAfterID)
        del Window._TKAfterID
    except Exception:
        pass
        # print('** tkafter cancel failed **')

    # Get window that caused return

    window = Window._window_that_exited

    if window is None:
        return None, timeout_key, None

    if window.XFound:
        event, values = None, None
        window.close()
        try:
            del Window._active_windows[window]
        except Exception:
            pass
            # print('Error deleting window, but OK')
    else:
        _BuildResults(window, False, window)
        event, values = window.ReturnValues

    return window, event, values


# MP""""""`MM                     dP
# M  mmmmm..M                     88
# M.      `YM dP    dP .d8888b. d8888P .d8888b. 88d8b.d8b.
# MMMMMMM.  M 88    88 Y8ooooo.   88   88ooood8 88'`88'`88
# M. .MMM'  M 88.  .88       88   88   88.  ... 88  88  88
# Mb.     .dM `8888P88 `88888P'   dP   `88888P' dP  dP  dP
# MMMMMMMMMMM      .88
#              d8888P
# M""""""""M
# Mmmm  mmmM
# MMMM  MMMM 88d888b. .d8888b. dP    dP
# MMMM  MMMM 88'  `88 88'  `88 88    88
# MMMM  MMMM 88       88.  .88 88.  .88
# MMMM  MMMM dP       `88888P8 `8888P88
# MMMMMMMMMM                        .88
#                               d8888P

# ------------------------------------------------------------------------- #
#                       SystemTray - class for implementing a psyeudo tray  #
# ------------------------------------------------------------------------- #

# -------------------------------- System Tray Begins Here -------------------------------- #
# Feb 2020 - Just starting on this so code commented out for now. Basing on PySimpleGUIQt's implementation / call format


# -------------------------------------------------------------------
# fade in/out info and default window alpha
SYSTEM_TRAY_WIN_MARGINS = 160, 60  # from right edge of screen, from bottom of screen
SYSTEM_TRAY_MESSAGE_MAX_LINE_LENGTH = 50
# colors
SYSTEM_TRAY_MESSAGE_WIN_COLOR = "#282828"
SYSTEM_TRAY_MESSAGE_TEXT_COLOR = "#ffffff"

SYSTEM_TRAY_MESSAGE_DISPLAY_DURATION_IN_MILLISECONDS = (
    3000  # how long to display the window
)
SYSTEM_TRAY_MESSAGE_FADE_IN_DURATION = 1000  # how long to fade in / fade out the window

EVENT_SYSTEM_TRAY_ICON_DOUBLE_CLICKED = "__DOUBLE_CLICKED__"
EVENT_SYSTEM_TRAY_ICON_ACTIVATED = "__ACTIVATED__"
EVENT_SYSTEM_TRAY_MESSAGE_CLICKED = "__MESSAGE_CLICKED__"

# Base64 Images to use as icons in the window
_tray_icon_error = b"iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAAA3NCSVQICAjb4U/gAAAACXBIWXMAAADlAAAA5QGP5Zs8AAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAIpQTFRF////20lt30Bg30pg4FJc409g4FBe4E9f4U9f4U9g4U9f4E9g31Bf4E9f4E9f4E9f4E9f4E9f4FFh4Vdm4lhn42Bv5GNx5W575nJ/6HqH6HyI6YCM6YGM6YGN6oaR8Kev9MPI9cbM9snO9s3R+Nfb+dzg+d/i++vt/O7v/fb3/vj5//z8//7+////KofnuQAAABF0Uk5TAAcIGBktSYSXmMHI2uPy8/XVqDFbAAAA8UlEQVQ4y4VT15LCMBBTQkgPYem9d9D//x4P2I7vILN68kj2WtsAhyDO8rKuyzyLA3wjSnvi0Eujf3KY9OUP+kno651CvlB0Gr1byQ9UXff+py5SmRhhIS0oPj4SaUUCAJHxP9+tLb/ezU0uEYDUsCc+l5/T8smTIVMgsPXZkvepiMj0Tm5txQLENu7gSF7HIuMreRxYNkbmHI0u5Hk4PJOXkSMz5I3nyY08HMjbpOFylF5WswdJPmYeVaL28968yNfGZ2r9gvqFalJNUy2UWmq1Wa7di/3Kxl3tF1671YHRR04dWn3s9cXRV09f3vb1fwPD7z9j1WgeRgAAAABJRU5ErkJggg=="
_tray_icon_success = b"iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAAA3NCSVQICAjb4U/gAAAACXBIWXMAAAEKAAABCgEWpLzLAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAHJQTFRF////ZsxmbbZJYL9gZrtVar9VZsJcbMRYaMZVasFYaL9XbMFbasRZaMFZacRXa8NYasFaasJaasFZasJaasNZasNYasJYasJZasJZasJZasJZasJZasJYasJZasJZasJZasJZasJaasJZasJZasJZasJZ2IAizQAAACV0Uk5TAAUHCA8YGRobHSwtPEJJUVtghJeYrbDByNjZ2tvj6vLz9fb3/CyrN0oAAADnSURBVDjLjZPbWoUgFIQnbNPBIgNKiwwo5v1fsQvMvUXI5oqPf4DFOgCrhLKjC8GNVgnsJY3nKm9kgTsduVHU3SU/TdxpOp15P7OiuV/PVzk5L3d0ExuachyaTWkAkLFtiBKAqZHPh/yuAYSv8R7XE0l6AVXnwBNJUsE2+GMOzWL8k3OEW7a/q5wOIS9e7t5qnGExvF5Bvlc4w/LEM4Abt+d0S5BpAHD7seMcf7+ZHfclp10TlYZc2y2nOqc6OwruxUWx0rDjNJtyp6HkUW4bJn0VWdf/a7nDpj1u++PBOR694+Ftj/8PKNdnDLn/V8YAAAAASUVORK5CYII="
_tray_icon_halt = b"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"
_tray_icon_notallowed = b"iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAMAUExURQAAAPcICPcLC/cMDPcQEPcSEvcXF/cYGPcaGvcbG/ccHPgxMfgyMvg0NPg5Ofg6Ovg7O/hBQfhCQvlFRflGRvljY/pkZPplZfpnZ/p2dgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMgEwNYAAAEAdFJOU////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////wBT9wclAAAACXBIWXMAAA7DAAAOwwHHb6hkAAABE0lEQVQ4T4WT65bDIAiExWbbtN0m3Uua+P4P6g4jGtN4NvNL4DuCCC5WWobe++uwmEmtwNxJUTebcwWCt5jJBwsYcKf3NE4hTOOJxj1FEnBTz4NH6qH2jUcCGr/QLLpkQgHe/6VWJXVqFgBB4yI/KVCkBCoFgPrPHw0CWbwCL8RibBFwzQDQH62/QeAtHQBeADUIDbkF/UnmnkB1ixtERrN3xCgyuF5kMntHTCJXh2vyv+wIdMhvgTeCQJ0C2hBMgSKfZlM1wSLXZ5oqgs8sjSpaCQ2VVlfKhLU6fdZGSvyWz9JMb+NE4jt/Nwfm0yJZSkBpYDg7TcJGrjm0Z7jK0B6P/fHiHK8e9Pp/eSmuf1+vf4x/ralnCN9IrncAAAAASUVORK5CYII="
_tray_icon_stop = b"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"
_tray_icon_exclamation = b"iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAMAUExURQAAAN0zM900NN01Nd02Nt03N944ON45Od46Ot47O98/P99BQd9CQt9DQ+FPT+JSUuJTU+JUVOJVVeJWVuNbW+ReXuVjY+Zra+dxceh4eOl7e+l8fOl+ful/f+qBgeqCguqDg+qFheuJieuLi+yPj+yQkO2Wlu+cnO+hofGqqvGtrfre3vrf3/ri4vvn5/75+f76+v/+/v///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMQ8SQkAAAEAdFJOU////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////wBT9wclAAAACXBIWXMAAA7DAAAOwwHHb6hkAAABJElEQVQ4T4WS63KCMBBGsyBai62X0otY0aq90ZZa3v/dtpvsJwTijOfXt7tnILOJYY9tNonjNCtQOlqhuKKG0RrNVjgkmIHBHgMId+h7zHSiwg2a9FNVVYScupETmjkd67o+CWpYwft+R6CpCgeUlq5AOyf45+8JsRUKFI6eQLkI3n5CIREBUekLxGaLpATCymRISiAszARJCYSxiZGUQKDLQoqgnPnFhUPOTWeRoZD3FvVZlmVHkE2OEM9iV71GVoZDBGUpAg9QWN5/jx+Ilsi9hz0q4VHOWD+hEF70yc1QEr1a4Q0F0S3eJDfLuv8T4QEFXduZE1rj+et7g6hzCDxF08N+X4DAu+6lUSTnc5wE5tx73ckSTV8QVoux3N88Rykw/wP3i+vwPKk17AAAAABJRU5ErkJggg=="
_tray_icon_none = None

SYSTEM_TRAY_MESSAGE_ICON_INFORMATION = _tray_icon_success
SYSTEM_TRAY_MESSAGE_ICON_WARNING = _tray_icon_exclamation
SYSTEM_TRAY_MESSAGE_ICON_CRITICAL = _tray_icon_stop
SYSTEM_TRAY_MESSAGE_ICON_NOICON = _tray_icon_none


# ------------------------------------------------------------------------- #
#                       Tray CLASS                                      #
# ------------------------------------------------------------------------- #
class SystemTray:
    """
    A "Simulated System Tray" that duplicates the API calls available to PySimpleGUIWx and PySimpleGUIQt users.

    All of the functionality works. The icon is displayed ABOVE the system tray rather than inside of it.
    """

    def __init__(
            self,
            menu=None,
            filename=None,
            data=None,
            data_base64=None,
            tooltip=None,
            metadata=None,
    ):
        """
        SystemTray - create an icon in the system tray
        :param menu:        Menu definition. Example - ['UNUSED', ['My', 'Simple', '---', 'Menu', 'Exit']]
        :type menu:         List[List[List[str] or str]]
        :param filename:    filename for icon
        :type filename:     (str)
        :param data:        in-ram image for icon (same as data_base64 parm)
        :type data:         (bytes)
        :param data_base64: base-64 data for icon
        :type data_base64:  (bytes)
        :param tooltip:     tooltip string
        :type tooltip:      (str)
        :param metadata:    User metadata that can be set to ANYTHING
        :type metadata:     (Any)
        """
        self._metadata = None
        self.Menu = menu
        self.TrayIcon = None
        self.Shown = False
        self.MenuItemChosen = TIMEOUT_KEY
        self.metadata = metadata
        self.last_message_event = None

        screen_size = Window.get_screen_size()

        if filename:
            image_elem = Image(
                filename=filename,
                background_color="red",
                enable_events=True,
                tooltip=tooltip,
                key="-IMAGE-",
            )
        elif data_base64:
            image_elem = Image(
                data=data_base64,
                background_color="red",
                enable_events=True,
                tooltip=tooltip,
                key="-IMAGE-",
            )
        elif data:
            image_elem = Image(
                data=data,
                background_color="red",
                enable_events=True,
                tooltip=tooltip,
                key="-IMAGE-",
            )
        else:
            image_elem = Image(
                background_color="red",
                enable_events=True,
                tooltip=tooltip,
                key="-IMAGE-",
            )
        layout = [
            [image_elem],
        ]
        self.window = Window(
            "Window Title",
            layout,
            element_padding=(0, 0),
            margins=(0, 0),
            grab_anywhere=True,
            no_titlebar=True,
            transparent_color="red",
            keep_on_top=True,
            right_click_menu=menu,
            location=(screen_size[0] - 100, screen_size[1] - 100),
            finalize=True,
        )

        self.window["-IMAGE-"].bind("<Double-Button-1>", "+DOUBLE_CLICK")

    @property
    def metadata(self):
        """
        Metadata is an SystemTray property that you can use at any time to hold any value
        :return: the current metadata value
        :rtype:  (Any)
        """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """
        Metadata is an SystemTray property that you can use at any time to hold any value
        :param value: Anything you want it to be
        :type value:  (Any)
        """
        self._metadata = value

    def read(self, timeout=None):
        """
        Reads the context menu
        :param timeout: Optional.  Any value other than None indicates a non-blocking read
        :type timeout:
        :return:
        :rtype:
        """
        if (
                self.last_message_event != TIMEOUT_KEY
                and self.last_message_event is not None
        ):
            event = self.last_message_event
            self.last_message_event = None
            return event
        event, values = self.window.read(timeout=timeout)
        if event.endswith("DOUBLE_CLICK"):
            return EVENT_SYSTEM_TRAY_ICON_DOUBLE_CLICKED
        elif event == "-IMAGE-":
            return EVENT_SYSTEM_TRAY_ICON_ACTIVATED

        return event

    def hide(self):
        """
        Hides the icon
        """
        self.window.hide()

    def un_hide(self):
        """
        Restores a previously hidden icon
        """
        self.window.un_hide()

    def show_message(
            self,
            title,
            message,
            filename=None,
            data=None,
            data_base64=None,
            messageicon=None,
            time=(
                    SYSTEM_TRAY_MESSAGE_FADE_IN_DURATION,
                    SYSTEM_TRAY_MESSAGE_DISPLAY_DURATION_IN_MILLISECONDS,
            ),
    ):
        """
        Shows a balloon above icon in system tray
        :param messageicon:
        :param title:       Title shown in balloon
        :type title:        str
        :param message:     Message to be displayed
        :type message:      str
        :param filename:    Optional icon filename
        :type filename:     str
        :param data:        Optional in-ram icon
        :type data:         b''
        :param data_base64: Optional base64 icon
        :type data_base64:  b''
        :param time:        Amount of time to display message in milliseconds. If tuple, first item is fade in/out duration
        :type time:         int | (int, int)
        :return:            The event that happened during the display such as user clicked on message
        :rtype:             Any
        """

        if isinstance(time, tuple):
            fade_duration, display_duration = time
        else:
            fade_duration = SYSTEM_TRAY_MESSAGE_FADE_IN_DURATION
            display_duration = time

        user_icon = data_base64 or filename or data or messageicon

        event = self.notify(
            title,
            message,
            icon=user_icon,
            fade_in_duration=fade_duration,
            display_duration_in_ms=display_duration,
        )
        self.last_message_event = event
        return event

    def close(self):
        """
        Close the system tray window
        """
        self.window.close()

    def update(
            self,
            menu=None,
            tooltip=None,
            filename=None,
            data=None,
            data_base64=None,
    ):
        """
        Updates the menu, tooltip or icon
        :param menu:        menu defintion
        :type menu:         ???
        :param tooltip:     string representing tooltip
        :type tooltip:      ???
        :param filename:    icon filename
        :type filename:     ???
        :param data:        icon raw image
        :type data:         ???
        :param data_base64: icon base 64 image
        :type data_base64:  ???
        """
        # Menu
        if menu is not None:
            top_menu = tk.Menu(self.window.TKroot, tearoff=False)
            AddMenuItem(top_menu, menu[1], self.window["-IMAGE-"])
            self.window["-IMAGE-"].TKRightClickMenu = top_menu

        if filename:
            self.window["-IMAGE-"].update(filename=filename)
        elif data_base64:
            self.window["-IMAGE-"].update(data=data_base64)
        elif data:
            self.window["-IMAGE-"].update(data=data)

        if tooltip:
            self.window["-IMAGE-"].set_tooltip(tooltip)

    @classmethod
    def notify(
            cls,
            title,
            message,
            icon=_tray_icon_success,
            display_duration_in_ms=SYSTEM_TRAY_MESSAGE_DISPLAY_DURATION_IN_MILLISECONDS,
            fade_in_duration=SYSTEM_TRAY_MESSAGE_FADE_IN_DURATION,
            alpha=0.9,
            location=None,
    ):
        """
        Displays a "notification window", usually in the bottom right corner of your display.  Has an icon, a title, and a message
        The window will slowly fade in and out if desired.  Clicking on the window will cause it to move through the end the current "phase". For example, if the window was fading in and it was clicked, then it would immediately stop fading in and instead be fully visible.  It's a way for the user to quickly dismiss the window.
        :param title:                  Text to be shown at the top of the window in a larger font
        :type title:                   (str)
        :param message:                Text message that makes up the majority of the window
        :type message:                 (str)
        :param icon:                   A base64 encoded PNG/GIF image or PNG/GIF filename that will be displayed in the window
        :type icon:                    bytes | str
        :param display_duration_in_ms: Number of milliseconds to show the window
        :type display_duration_in_ms:  (int)
        :param fade_in_duration:       Number of milliseconds to fade window in and out
        :type fade_in_duration:        (int)
        :param alpha:                  Alpha channel. 0 - invisible 1 - fully visible
        :type alpha:                   (float)
        :param location:               Location on the screen to display the window
        :type location:                (int, int)
        :return:                       (int) reason for returning
        :rtype:                        (int)
        """

        messages = message.split("\n")
        full_msg = ""
        for m in messages:
            m_wrap = textwrap.fill(m, SYSTEM_TRAY_MESSAGE_MAX_LINE_LENGTH)
            full_msg += m_wrap + "\n"
        message = full_msg[:-1]

        win_msg_lines = message.count("\n") + 1
        max_line = max(message.split("\n"))

        screen_res_x, screen_res_y = Window.get_screen_size()
        win_margin = SYSTEM_TRAY_WIN_MARGINS  # distance from screen edges
        win_width, win_height = 364, 66 + (14.8 * win_msg_lines)

        layout = [
            [
                Graph(
                    canvas_size=(win_width, win_height),
                    graph_bottom_left=(0, win_height),
                    graph_top_right=(win_width, 0),
                    key="-GRAPH-",
                    background_color=SYSTEM_TRAY_MESSAGE_WIN_COLOR,
                    enable_events=True,
                )
            ]
        ]

        win_location = (
            location
            if location is not None
            else (
                screen_res_x - win_width - win_margin[0],
                screen_res_y - win_height - win_margin[1],
            )
        )
        window = Window(
            title,
            layout,
            background_color=SYSTEM_TRAY_MESSAGE_WIN_COLOR,
            no_titlebar=True,
            location=win_location,
            keep_on_top=True,
            alpha_channel=0,
            margins=(0, 0),
            element_padding=(0, 0),
            grab_anywhere=True,
            finalize=True,
        )

        window["-GRAPH-"].draw_rectangle(
            (win_width, win_height),
            (-win_width, -win_height),
            fill_color=SYSTEM_TRAY_MESSAGE_WIN_COLOR,
            line_color=SYSTEM_TRAY_MESSAGE_WIN_COLOR,
        )
        if type(icon) is bytes:
            window["-GRAPH-"].draw_image(data=icon, location=(20, 20))
        elif icon is not None:
            window["-GRAPH-"].draw_image(filename=icon, location=(20, 20))
        window["-GRAPH-"].draw_text(
            title,
            location=(64, 20),
            color=SYSTEM_TRAY_MESSAGE_TEXT_COLOR,
            font=("Helvetica", 12, "bold"),
            text_location=TEXT_LOCATION_TOP_LEFT,
        )
        window["-GRAPH-"].draw_text(
            message,
            location=(64, 44),
            color=SYSTEM_TRAY_MESSAGE_TEXT_COLOR,
            font=("Helvetica", 9),
            text_location=TEXT_LOCATION_TOP_LEFT,
        )
        window["-GRAPH-"].set_cursor("hand2")

        if fade_in_duration:
            for i in range(1, int(alpha * 100)):  # fade in
                window.set_alpha(i / 100)
                event, values = window.read(timeout=fade_in_duration // 100)
                if event != TIMEOUT_KEY:
                    window.set_alpha(1)
                    break
            if event != TIMEOUT_KEY:
                window.close()
                return (
                    EVENT_SYSTEM_TRAY_MESSAGE_CLICKED if event == "-GRAPH-" else event
                )
            event, values = window(timeout=display_duration_in_ms)
            if event == TIMEOUT_KEY:
                for i in range(int(alpha * 100), 1, -1):  # fade out
                    window.set_alpha(i / 100)
                    event, values = window.read(timeout=fade_in_duration // 100)
                    if event != TIMEOUT_KEY:
                        break
        else:
            window.set_alpha(alpha)
            event, values = window(timeout=display_duration_in_ms)
        window.close()

        return EVENT_SYSTEM_TRAY_MESSAGE_CLICKED if event == "-GRAPH-" else event

    Close = close
    Hide = hide
    Read = read
    ShowMessage = show_message
    UnHide = un_hide
    Update = update


# ################################################################################
# ################################################################################
#  END OF ELEMENT DEFINITIONS
# ################################################################################
# ################################################################################


# =========================================================================== #
# Button Lazy Functions so the caller doesn't have to define a bunch of stuff #
# =========================================================================== #


# ------------------------- A fake Element... the Pad Element ------------------------- #
def Sizer(h_pixels=0, v_pixels=0):
    """
    "Pushes" out the size of whatever it is placed inside of.  This includes Columns, Frames, Tabs and Windows

    :param h_pixels: number of horizontal pixels
    :type h_pixels:  (int)
    :param v_pixels: number of vertical pixels
    :type v_pixels:  (int)
    :return:         (Canvas) A canvas element that has a pad setting set according to parameters
    :rtype:          (Canvas)
    """

    return Canvas(size=(0, 0), pad=((h_pixels, 0), (v_pixels, 0)))


def pin(elem, vertical_alignment=None, shrink=True, expand_x=None, expand_y=None):
    """
    Pin's an element provided into a layout so that when it's made invisible and visible again, it will
     be in the correct place.  Otherwise it will be placed at the end of its containing window/column.

     The element you want to pin is the element that you'll be making visibile/invisible.

    The pin helper function also causes containers to shrink to fit the contents correct after something inside
     has changed visiblity.  Note that setting a hardcoded size on your window can impact this ability to shrink.

    :param elem:               the element to put into the layout
    :type elem:                Element
    :param vertical_alignment: Aligns elements vertically. 'top', 'center', 'bottom'. Can be shortened to 't', 'c', 'b'
    :type vertical_alignment:  str | None
    :param shrink:             If True, then the space will shrink down to a single pixel when hidden. False leaves the area large and blank
    :type shrink:              bool
    :param expand_x:           If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_x:            (bool)
    :param expand_y:           If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_y:            (bool)
    :return:                   A column element containing the provided element
    :rtype:                    Column
    """
    if shrink:
        # return Column([[elem, Canvas(size=(0, 0),background_color=elem.BackgroundColor, pad=(0, 0))]], pad=(0, 0), vertical_alignment=vertical_alignment, expand_x=expand_x, expand_y=expand_y)
        return Column(
            [[elem, Column([[]], pad=(0, 0))]],
            pad=(0, 0),
            vertical_alignment=vertical_alignment,
            expand_x=expand_x,
            expand_y=expand_y,
        )
    else:
        return Column(
            [[elem]],
            pad=(0, 0),
            vertical_alignment=vertical_alignment,
            expand_x=expand_x,
            expand_y=expand_y,
        )


def vtop(elem_or_row, expand_x=None, expand_y=None, background_color=None):
    """
    Align an element or a row of elements to the top of the row that contains it

    :param elem_or_row:      the element or row of elements
    :type elem_or_row:       Element | List[Element] | Tuple[Element]
    :param expand_x:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_x:          (bool)
    :param expand_y:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_y:          (bool)
    :param background_color: Background color for container that is used by vtop to do the alignment
    :type background_color:  str | None
    :return:                 A column element containing the provided element aligned to the top or list of elements (a row)
    :rtype:                  Column | List[Column]
    """

    if isinstance(elem_or_row, list) or isinstance(elem_or_row, tuple):
        return [
            Column(
                [[e]],
                pad=(0, 0),
                vertical_alignment="top",
                expand_x=expand_x,
                expand_y=expand_y,
                background_color=background_color,
            )
            for e in elem_or_row
        ]

    return Column(
        [[elem_or_row]],
        pad=(0, 0),
        vertical_alignment="top",
        expand_x=expand_x,
        expand_y=expand_y,
        background_color=background_color,
    )


def vcenter(elem_or_row, expand_x=None, expand_y=None, background_color=None):
    """
    Align an element or a row of elements to the center of the row that contains it

    :param elem_or_row:      the element or row of elements
    :type elem_or_row:       Element | List[Element] | Tuple[Element]
    :param expand_x:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_x:          (bool)
    :param expand_y:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_y:          (bool)
    :param background_color: Background color for container that is used by vcenter to do the alignment
    :type background_color:  str | None
    :return:                 A column element containing the provided element aligned to the center or list of elements (a row)
    :rtype:                  Column | List[Column]
    """

    if isinstance(elem_or_row, list) or isinstance(elem_or_row, tuple):
        return [
            Column(
                [[e]],
                pad=(0, 0),
                vertical_alignment="center",
                expand_x=expand_x,
                expand_y=expand_y,
                background_color=background_color,
            )
            for e in elem_or_row
        ]

    return Column(
        [[elem_or_row]],
        pad=(0, 0),
        vertical_alignment="center",
        expand_x=expand_x,
        expand_y=expand_y,
        background_color=background_color,
    )


def vbottom(elem_or_row, expand_x=None, expand_y=None, background_color=None):
    """
    Align an element or a row of elements to the bottom of the row that contains it

    :param elem_or_row:      the element or row of elements
    :type elem_or_row:       Element | List[Element] | Tuple[Element]
    :param expand_x:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_x:          (bool)
    :param expand_y:         If True/False the value will be passed to the Column Elements used to make this feature
    :type expand_y:          (bool)
    :param background_color: Background color for container that is used by vcenter to do the alignment
    :type background_color:  str | None
    :return:                 A column element containing the provided element aligned to the bottom or list of elements (a row)
    :rtype:                  Column | List[Column]
    """

    if isinstance(elem_or_row, list) or isinstance(elem_or_row, tuple):
        return [
            Column(
                [[e]],
                pad=(0, 0),
                vertical_alignment="bottom",
                expand_x=expand_x,
                expand_y=expand_y,
                background_color=background_color,
            )
            for e in elem_or_row
        ]

    return Column(
        [[elem_or_row]],
        pad=(0, 0),
        vertical_alignment="bottom",
        expand_x=expand_x,
        expand_y=expand_y,
        background_color=background_color,
    )


def Titlebar(
        title="",
        icon=None,
        text_color=None,
        background_color=None,
        font=None,
        key=None,
        k=None,
):
    """
    A custom titlebar that replaces the OS provided titlebar, thus giving you control
    the is not possible using the OS provided titlebar such as the color.

    NOTE LINUX USERS - at the moment the minimize function is not yet working.  Windows users
    should have no problem and it should function as a normal window would.

    This titlebar is created from a row of elements that is then encapsulated into a
    one Column element which is what this Titlebar function returns to you.

    A custom titlebar removes the margins from your window.  If you want the  remainder
    of your Window to have margins, place the layout after the Titlebar into a Column and
    set the pad of that Column to the dimensions you would like your margins to have.

    The Titlebar is a COLUMN element.  You can thus call the update method for the column and
    perform operations such as making the column visible/invisible

    :param icon:             Can be either a filename or Base64 byte string of a PNG or GIF. This is used in an Image element to create the titlebar
    :type icon:              str or bytes or None
    :param title:            The "title" to show in the titlebar
    :type title:             str
    :param text_color:       Text color for titlebar
    :type text_color:        str | None
    :param background_color: Background color for titlebar
    :type background_color:  str | None
    :param font:             Font to be used for the text and the symbols
    :type font:              (str or (str, int[, str]) or None)
    :param key:              Identifies an Element. Should be UNIQUE to this window.
    :type key:               str | int | tuple | object | None
    :param k:                Exactly the same as key.  Choose one of them to use
    :type k:                 str | int | tuple | object | None
    :return:                 A single Column element that has eveything in 1 element
    :rtype:                  Column
    """
    bc = background_color or CUSTOM_TITLEBAR_BACKGROUND_COLOR or theme_button_color()[1]
    tc = text_color or CUSTOM_TITLEBAR_TEXT_COLOR or theme_button_color()[0]
    font = font or CUSTOM_TITLEBAR_FONT or ("Helvetica", 12)
    key = k or key

    if isinstance(icon, bytes):
        icon_and_text_portion = [
            Image(data=icon, background_color=bc, key=TITLEBAR_IMAGE_KEY)
        ]
    elif icon == TITLEBAR_DO_NOT_USE_AN_ICON:
        icon_and_text_portion = []
    elif icon is not None:
        icon_and_text_portion = [
            Image(filename=icon, background_color=bc, key=TITLEBAR_IMAGE_KEY)
        ]
    elif CUSTOM_TITLEBAR_ICON is not None:
        if isinstance(CUSTOM_TITLEBAR_ICON, bytes):
            icon_and_text_portion = [
                Image(
                    data=CUSTOM_TITLEBAR_ICON,
                    background_color=bc,
                    key=TITLEBAR_IMAGE_KEY,
                )
            ]
        else:
            icon_and_text_portion = [
                Image(
                    filename=CUSTOM_TITLEBAR_ICON,
                    background_color=bc,
                    key=TITLEBAR_IMAGE_KEY,
                )
            ]
    else:
        icon_and_text_portion = [
            Image(
                data=DEFAULT_BASE64_ICON_16_BY_16,
                background_color=bc,
                key=TITLEBAR_IMAGE_KEY,
            )
        ]

    icon_and_text_portion += [
        T(
            title,
            text_color=tc,
            background_color=bc,
            font=font,
            grab=True,
            key=TITLEBAR_TEXT_KEY,
        )
    ]

    return Column(
        [
            [
                Column([icon_and_text_portion], pad=(0, 0), background_color=bc),
                Column(
                    [
                        [
                            T(
                                SYMBOL_TITLEBAR_MINIMIZE,
                                text_color=tc,
                                background_color=bc,
                                enable_events=True,
                                font=font,
                                key=TITLEBAR_MINIMIZE_KEY,
                            ),
                            Text(
                                SYMBOL_TITLEBAR_MAXIMIZE,
                                text_color=tc,
                                background_color=bc,
                                enable_events=True,
                                font=font,
                                key=TITLEBAR_MAXIMIZE_KEY,
                            ),
                            Text(
                                SYMBOL_TITLEBAR_CLOSE,
                                text_color=tc,
                                background_color=bc,
                                font=font,
                                enable_events=True,
                                key=TITLEBAR_CLOSE_KEY,
                            ),
                        ]
                    ],
                    element_justification="r",
                    expand_x=True,
                    grab=True,
                    pad=(0, 0),
                    background_color=bc,
                ),
            ]
        ],
        expand_x=True,
        grab=True,
        background_color=bc,
        pad=(0, 0),
        metadata=TITLEBAR_METADATA_MARKER,
        key=key,
    )


def MenubarCustom(
        menu_definition,
        disabled_text_color=None,
        bar_font=None,
        font=None,
        tearoff=False,
        pad=0,
        p=None,
        background_color=None,
        text_color=None,
        bar_background_color=None,
        bar_text_color=None,
        key=None,
        k=None,
):
    """
    A custom Menubar that replaces the OS provided Menubar

    Why?
    Two reasons - 1. they look great (see custom titlebar) 2. if you have a custom titlebar, then you have to use a custom menubar if you want a menubar

    :param menu_definition:      The Menu definition specified using lists (docs explain the format)
    :type menu_definition:       List[List[Tuple[str, List[str]]]
    :param disabled_text_color:  color to use for text when item is disabled. Can be in #RRGGBB format or a color name "black"
    :type disabled_text_color:   (str)
    :param bar_font:             specifies the font family, size to be used for the chars in the bar itself
    :type bar_font:              (str or (str, int[, str]) or None)
    :param font:                 specifies the font family, size to be used for the menu items
    :type font:                  (str or (str, int[, str]) or None)
    :param tearoff:              if True, then can tear the menu off from the window ans use as a floating window. Very cool effect
    :type tearoff:               (bool)
    :param pad:                  Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int).  TIP - 0 will make flush with titlebar
    :type pad:                   (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                    Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                     (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param background_color:     color to use for background of the menus that are displayed after making a section. Can be in #RRGGBB format or a color name "black". Defaults to the color of the bar text
    :type background_color:      (str)
    :param text_color:           color to use for the text of the many items in the displayed menus. Can be in #RRGGBB format or a color name "black". Defaults to the bar background
    :type text_color:            (str)
    :param bar_background_color: color to use for the menubar. Can be in #RRGGBB format or a color name "black". Defaults to theme's button text color
    :type bar_background_color:  (str)
    :param bar_text_color:       color to use for the menu items text when item is disabled. Can be in #RRGGBB format or a color name "black". Defaults to theme's button background color
    :type bar_text_color:        (str)
    :param key:                  Value that uniquely identifies this element from all other elements. Used when Finding an element or in return values. Must be unique to the window
    :type key:                   str | int | tuple | object
    :param k:                    Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                     str | int | tuple | object
    :returns:                    A Column element that has a series of ButtonMenu elements
    :rtype:                      Column
    """

    bar_bg = (
        bar_background_color
        if bar_background_color is not None
        else theme_button_color()[0]
    )
    bar_text = bar_text_color if bar_text_color is not None else theme_button_color()[1]
    menu_bg = background_color if background_color is not None else bar_text
    menu_text = text_color if text_color is not None else bar_bg
    pad = pad if pad is not None else p

    row = []
    for menu in menu_definition:
        text = menu[0]
        if MENU_SHORTCUT_CHARACTER in text:
            text = text.replace(MENU_SHORTCUT_CHARACTER, "")
        if text.startswith(MENU_DISABLED_CHARACTER):
            disabled = True
            text = text[len(MENU_DISABLED_CHARACTER):]
        else:
            disabled = False

        button_menu = ButtonMenu(
            text,
            menu,
            border_width=0,
            button_color=(bar_text, bar_bg),
            key=text,
            pad=(0, 0),
            disabled=disabled,
            font=bar_font,
            item_font=font,
            disabled_text_color=disabled_text_color,
            text_color=menu_text,
            background_color=menu_bg,
            tearoff=tearoff,
        )
        button_menu.part_of_custom_menubar = True
        button_menu.custom_menubar_key = key if key is not None else k
        row += [button_menu]
    return Column(
        [row],
        pad=pad,
        background_color=bar_bg,
        expand_x=True,
        key=key if key is not None else k,
    )


# -------------------------  FOLDER BROWSE Element lazy function  ------------------------- #
def FolderBrowse(
        button_text="Browse",
        target=(ThisRow, -1),
        initial_folder=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        change_submits=False,
        enable_events=False,
        font=None,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    :param button_text:      text in the button (Default value = 'Browse')
    :type button_text:       (str)
    :param target:           target for the button (Default value = (ThisRow, -1))
    :type target:            str | (int, int)
    :param initial_folder:   starting path for folders and files
    :type initial_folder:    (str)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param change_submits:   If True, pressing Enter key submits window (Default = False)
    :type enable_events:     (bool)
    :param enable_events:    Turns on the element specific events.(Default = False)
    :type enable_events:     (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              Used with window.find_element and with return values to uniquely identify this element
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 The Button created
    :rtype:                  (Button)
    """

    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_BROWSE_FOLDER,
        target=target,
        initial_folder=initial_folder,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        disabled=disabled,
        button_color=button_color,
        change_submits=change_submits,
        enable_events=enable_events,
        font=font,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  FILE BROWSE Element lazy function  ------------------------- #
def FileBrowse(
        button_text="Browse",
        target=(ThisRow, -1),
        file_types=FILE_TYPES_ALL_FILES,
        initial_folder=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        change_submits=False,
        enable_events=False,
        font=None,
        disabled=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Browse')
    :type button_text:       (str)
    :param target:           key or (row,col) target for the button (Default value = (ThisRow, -1))
    :type target:            str | (int, int)
    :param file_types:       filter file types Default value = (("ALL Files", "*.* *"),).
    :type file_types:        Tuple[(str, str), ...]
    :param initial_folder:   starting path for folders and files
    :type initial_folder:
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param change_submits:   If True, pressing Enter key submits window (Default = False)
    :type change_submits:    (bool)
    :param enable_events:    Turns on the element specific events.(Default = False)
    :type enable_events:     (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_BROWSE_FILE,
        target=target,
        file_types=file_types,
        initial_folder=initial_folder,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        change_submits=change_submits,
        enable_events=enable_events,
        disabled=disabled,
        button_color=button_color,
        font=font,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  FILES BROWSE Element (Multiple file selection) lazy function  ------------------------- #
def FilesBrowse(
        button_text="Browse",
        target=(ThisRow, -1),
        file_types=FILE_TYPES_ALL_FILES,
        disabled=False,
        initial_folder=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        change_submits=False,
        enable_events=False,
        font=None,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        files_delimiter=BROWSE_FILES_DELIMITER,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    Allows browsing of multiple files. File list is returned as a single list with the delimiter defined using the files_delimiter parameter.

    :param button_text:      text in the button (Default value = 'Browse')
    :type button_text:       (str)
    :param target:           key or (row,col) target for the button (Default value = (ThisRow, -1))
    :type target:            str | (int, int)
    :param file_types:       Default value = (("ALL Files", "*.* *"),).
    :type file_types:        Tuple[(str, str), ...]
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param initial_folder:   starting path for folders and files
    :type initial_folder:    (str)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param change_submits:   If True, pressing Enter key submits window (Default = False)
    :type change_submits:    (bool)
    :param enable_events:    Turns on the element specific events.(Default = False)
    :type enable_events:     (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param files_delimiter:  String to place between files when multiple files are selected. Normally a ;
    :type files_delimiter:   str
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    button = Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_BROWSE_FILES,
        target=target,
        file_types=file_types,
        initial_folder=initial_folder,
        change_submits=change_submits,
        enable_events=enable_events,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        disabled=disabled,
        button_color=button_color,
        font=font,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )
    button._files_delimiter = files_delimiter
    return button


# -------------------------  FILE BROWSE Element lazy function  ------------------------- #
def FileSaveAs(
        button_text="Save As...",
        target=(ThisRow, -1),
        file_types=FILE_TYPES_ALL_FILES,
        initial_folder=None,
        default_extension="",
        disabled=False,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        change_submits=False,
        enable_events=False,
        font=None,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:       text in the button (Default value = 'Save As...')
    :type button_text:        (str)
    :param target:            key or (row,col) target for the button (Default value = (ThisRow, -1))
    :type target:             str | (int, int)
    :param file_types:        Default value = (("ALL Files", "*.* *"),).
    :type file_types:         Tuple[(str, str), ...]
    :param default_extension: If no extension entered by user, add this to filename (only used in saveas dialogs)
    :type default_extension:  (str)
    :param initial_folder:    starting path for folders and files
    :type initial_folder:     (str)
    :param disabled:          set disable state for element (Default = False)
    :type disabled:           (bool)
    :param tooltip:           text, that will appear when mouse hovers over the element
    :type tooltip:            (str)
    :param size:              (w,h) w=characters-wide, h=rows-high
    :type size:               (int, int)
    :param s:                 Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                  (int, int)  | (None, None) | int
    :param auto_size_button:  True if button size is determined by button text
    :type auto_size_button:   (bool)
    :param button_color:      button color (foreground, background)
    :type button_color:       (str, str) | str
    :param change_submits:    If True, pressing Enter key submits window (Default = False)
    :type change_submits:     (bool)
    :param enable_events:     Turns on the element specific events.(Default = False)
    :type enable_events:      (bool)
    :param font:              specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:               (str or (str, int[, str]) or None)
    :param pad:               Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:                (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                 Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                  (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:               key for uniquely identify this element (for window.find_element)
    :type key:                str | int | tuple | object
    :param k:                 Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                  str | int | tuple | object
    :param visible:           set initial visibility state of the Button
    :type visible:            (bool)
    :param metadata:          Anything you want to store along with this button
    :type metadata:           (Any)
    :param expand_x:          If True Element will expand in the Horizontal directions
    :type expand_x:           (bool)
    :param expand_y:          If True Element will expand in the Vertical directions
    :type expand_y:           (bool)        :return:                  returns a button
    :rtype:                   (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_SAVEAS_FILE,
        target=target,
        file_types=file_types,
        initial_folder=initial_folder,
        default_extension=default_extension,
        tooltip=tooltip,
        size=size,
        s=s,
        disabled=disabled,
        auto_size_button=auto_size_button,
        button_color=button_color,
        change_submits=change_submits,
        enable_events=enable_events,
        font=font,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  SAVE AS Element lazy function  ------------------------- #
def SaveAs(
        button_text="Save As...",
        target=(ThisRow, -1),
        file_types=FILE_TYPES_ALL_FILES,
        initial_folder=None,
        default_extension="",
        disabled=False,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        change_submits=False,
        enable_events=False,
        font=None,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:       text in the button (Default value = 'Save As...')
    :type button_text:        (str)
    :param target:            key or (row,col) target for the button (Default value = (ThisRow, -1))
    :type target:             str | (int, int)
    :param file_types:        Default value = (("ALL Files", "*.* *"),).
    :type file_types:         Tuple[(str, str), ...]
    :param default_extension: If no extension entered by user, add this to filename (only used in saveas dialogs)
    :type default_extension:  (str)
    :param initial_folder:    starting path for folders and files
    :type initial_folder:     (str)
    :param disabled:          set disable state for element (Default = False)
    :type disabled:           (bool)
    :param tooltip:           text, that will appear when mouse hovers over the element
    :type tooltip:            (str)
    :param size:              (w,h) w=characters-wide, h=rows-high
    :type size:               (int, int)
    :param s:                 Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                  (int, int)  | (None, None) | int
    :param auto_size_button:  True if button size is determined by button text
    :type auto_size_button:   (bool)
    :param button_color:      button color (foreground, background)
    :type button_color:       (str, str) or str
    :param change_submits:    If True, pressing Enter key submits window (Default = False)
    :type change_submits:     (bool)
    :param enable_events:     Turns on the element specific events.(Default = False)
    :type enable_events:      (bool)
    :param font:              specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:               (str or (str, int[, str]) or None)
    :param pad:               Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:                (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                 Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                  (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int    :param key:               key for uniquely identify this element (for window.find_element)
    :type key:                str | int | tuple | object
    :param k:                 Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                  str | int | tuple | object
    :param visible:           set initial visibility state of the Button
    :type visible:            (bool)
    :param metadata:          Anything you want to store along with this button
    :type metadata:           (Any)
    :param expand_x:          If True Element will expand in the Horizontal directions
    :type expand_x:           (bool)
    :param expand_y:          If True Element will expand in the Vertical directions
    :type expand_y:           (bool)
    :return:                  returns a button
    :rtype:                   (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_SAVEAS_FILE,
        target=target,
        file_types=file_types,
        initial_folder=initial_folder,
        default_extension=default_extension,
        tooltip=tooltip,
        size=size,
        s=s,
        disabled=disabled,
        auto_size_button=auto_size_button,
        button_color=button_color,
        change_submits=change_submits,
        enable_events=enable_events,
        font=font,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  SAVE BUTTON Element lazy function  ------------------------- #
def Save(
        button_text="Save",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        bind_return_key=True,
        disabled=False,
        tooltip=None,
        font=None,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Save')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  SUBMIT BUTTON Element lazy function  ------------------------- #
def Submit(
        button_text="Submit",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        bind_return_key=True,
        tooltip=None,
        font=None,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Submit')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  OPEN BUTTON Element lazy function  ------------------------- #
# -------------------------  OPEN BUTTON Element lazy function  ------------------------- #
def Open(
        button_text="Open",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        bind_return_key=True,
        tooltip=None,
        font=None,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Open')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  OK BUTTON Element lazy function  ------------------------- #
def OK(
        button_text="OK",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        bind_return_key=True,
        tooltip=None,
        font=None,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'OK')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  YES BUTTON Element lazy function  ------------------------- #
def Ok(
        button_text="Ok",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        bind_return_key=True,
        tooltip=None,
        font=None,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Ok')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  CANCEL BUTTON Element lazy function  ------------------------- #
def Cancel(
        button_text="Cancel",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        tooltip=None,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Cancel')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  QUIT BUTTON Element lazy function  ------------------------- #
def Quit(
        button_text="Quit",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        tooltip=None,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Quit')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:             (bool)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  Exit BUTTON Element lazy function  ------------------------- #
def Exit(
        button_text="Exit",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        tooltip=None,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Exit')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  YES BUTTON Element lazy function  ------------------------- #
def Yes(
        button_text="Yes",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        tooltip=None,
        font=None,
        bind_return_key=True,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Yes')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = True) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  NO BUTTON Element lazy function  ------------------------- #
def No(
        button_text="No",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        tooltip=None,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'No')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, then the return key will cause a the Listbox to generate an event
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  NO BUTTON Element lazy function  ------------------------- #
def Help(
        button_text="Help",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        font=None,
        tooltip=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button (Default value = 'Help')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  NO BUTTON Element lazy function  ------------------------- #
def Debug(
        button_text="",
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        font=None,
        tooltip=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    This Button has been changed in how it works!!
    Your button has been replaced with a normal button that has the PySimpleGUI Debugger buggon logo on it.
    In your event loop, you will need to check for the event of this button and then call:
            show_debugger_popout_window()
    :param button_text:      text in the button (Default value = '')
    :type button_text:       (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """

    user_key = key if key is not None else k if k is not None else button_text

    return Button(
        button_text="",
        button_type=BUTTON_TYPE_READ_FORM,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=theme_button_color(),
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=user_key,
        k=k,
        visible=visible,
        image_data=PSG_DEBUGGER_LOGO,
        image_subsample=2,
        border_width=0,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  GENERIC BUTTON Element lazy function  ------------------------- #
def SimpleButton(
        button_text,
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        border_width=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        font=None,
        bind_return_key=False,
        disabled=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    DEPIRCATED

    This Button should not be used.

    :param border_width:
    :param button_text:      text in the button
    :type button_text:       (str)
    :param image_filename:   image filename if there is a button image
    :type image_filename:    image filename if there is a button image
    :param image_data:       in-RAM image to be displayed on button
    :type image_data:        in-RAM image to be displayed on button
    :param image_size:       image size (O.K.)
    :type image_size:        (Default = (None))
    :param image_subsample:  amount to reduce the size of the image
    :type image_subsample:   amount to reduce the size of the image
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_CLOSES_WIN,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        disabled=disabled,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  CLOSE BUTTON Element lazy function  ------------------------- #
def CloseButton(
        button_text,
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        border_width=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        font=None,
        bind_return_key=False,
        disabled=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    DEPRICATED

    This button should not be used.  Instead explicitly close your windows by calling window.close() or by using
    the close parameter in window.read

    :param border_width:
    :param button_text:      text in the button
    :type button_text:       (str)
    :param image_filename:   image filename if there is a button image
    :type image_filename:    image filename if there is a button image
    :param image_data:       in-RAM image to be displayed on button
    :type image_data:        in-RAM image to be displayed on button
    :param image_size:       image size (O.K.)
    :type image_size:        (Default = (None))
    :param image_subsample:  amount to reduce the size of the image
    :type image_subsample:   amount to reduce the size of the image
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_CLOSES_WIN,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        disabled=disabled,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


CButton = CloseButton


# -------------------------  GENERIC BUTTON Element lazy function  ------------------------- #
def ReadButton(
        button_text,
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        border_width=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        font=None,
        bind_return_key=False,
        disabled=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    :param button_text:      text in the button
    :type button_text:       (str)
    :param image_filename:   image filename if there is a button image
    :type image_filename:    image filename if there is a button image
    :param image_data:       in-RAM image to be displayed on button
    :type image_data:        in-RAM image to be displayed on button
    :param image_size:       image size (O.K.)
    :type image_size:        (Default = (None))
    :param image_subsample:  amount to reduce the size of the image
    :type image_subsample:   amount to reduce the size of the image
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param focus:            if focus should be set to this
    :type focus:             idk_yetReally
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param border_width:     width of border around element
    :type border_width:      (int)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 Button created
    :rtype:                  (Button)
    """

    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_READ_FORM,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        size=size,
        s=s,
        disabled=disabled,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


ReadFormButton = ReadButton
RButton = ReadFormButton


# -------------------------  Realtime BUTTON Element lazy function  ------------------------- #
def RealtimeButton(
        button_text,
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        border_width=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        font=None,
        disabled=False,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button
    :type button_text:       (str)
    :param image_filename:   image filename if there is a button image
    :type image_filename:    image filename if there is a button image
    :param image_data:       in-RAM image to be displayed on button
    :type image_data:        in-RAM image to be displayed on button
    :param image_size:       image size (O.K.)
    :type image_size:        (Default = (None))
    :param image_subsample:  amount to reduce the size of the image
    :type image_subsample:   amount to reduce the size of the image
    :param border_width:     width of border around element
    :type border_width:      (int)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:             (bool)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 Button created
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_REALTIME,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        disabled=disabled,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  Dummy BUTTON Element lazy function  ------------------------- #
def DummyButton(
        button_text,
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        border_width=None,
        tooltip=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        font=None,
        disabled=False,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    This is a special type of Button.

    It will close the window but NOT send an event that the window has been closed.

    It's used in conjunction with non-blocking windows to silently close them.  They are used to
    implement the non-blocking popup windows. They're also found in some Demo Programs, so look there for proper use.

    :param button_text:      text in the button
    :type button_text:       (str)
    :param image_filename:   image filename if there is a button image
    :type image_filename:    image filename if there is a button image
    :param image_data:       in-RAM image to be displayed on button
    :type image_data:        in-RAM image to be displayed on button
    :param image_size:       image size (O.K.)
    :type image_size:        (Default = (None))
    :param image_subsample:  amount to reduce the size of the image
    :type image_subsample:   amount to reduce the size of the image
    :param border_width:     width of border around element
    :type border_width:      (int)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            if focus should be set to this
    :type focus:             (bool)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         Anything you want to store along with this button
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    return Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_CLOSES_WIN_ONLY,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )


# -------------------------  Calendar Chooser Button lazy function  ------------------------- #
def CalendarButton(
        button_text,
        target=(ThisRow, -1),
        close_when_date_chosen=True,
        default_date_m_d_y=(None, None, None),
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        tooltip=None,
        border_width=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        enable_events=None,
        key=None,
        k=None,
        visible=True,
        locale=None,
        format="%Y-%m-%d %H:%M:%S",
        begin_at_sunday_plus=0,
        month_names=None,
        day_abbreviations=None,
        title="Choose Date",
        no_titlebar=True,
        location=(None, None),
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """
    Button that will show a calendar chooser window.  Fills in the target element with result

    :param enable_events:
    :param button_text:            text in the button
    :type button_text:             (str)
    :param target:                 Key or "coordinate" (see docs) of target element
    :type target:                  (int, int) | Any
    :param close_when_date_chosen: (Default = True)
    :type close_when_date_chosen:  bool
    :param default_date_m_d_y:     Beginning date to show
    :type default_date_m_d_y:      (int, int or None, int)
    :param image_filename:         image filename if there is a button image
    :type image_filename:          image filename if there is a button image
    :param image_data:             in-RAM image to be displayed on button
    :type image_data:              in-RAM image to be displayed on button
    :param image_size:             image size (O.K.)
    :type image_size:              (Default = (None))
    :param image_subsample:        amount to reduce the size of the image
    :type image_subsample:         amount to reduce the size of the image
    :param tooltip:                text, that will appear when mouse hovers over the element
    :type tooltip:                 (str)
    :param border_width:           width of border around element
    :type border_width:            width of border around element
    :param size:                   (w,h) w=characters-wide, h=rows-high
    :type size:                    (int, int)
    :param s:                      Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                       (int, int)  | (None, None) | int
    :param auto_size_button:       True if button size is determined by button text
    :type auto_size_button:        (bool)
    :param button_color:           button color (foreground, background)
    :type button_color:            (str, str) | str
    :param disabled:               set disable state for element (Default = False)
    :type disabled:                (bool)
    :param font:                   specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                    (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:         bool
    :param focus:                  if focus should be set to this
    :type focus:                   bool
    :param pad:                    Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:                     (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                      Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                       (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:                    key for uniquely identify this element (for window.find_element)
    :type key:                     str | int | tuple | object
    :param k:                      Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                       str | int | tuple | object
    :param locale:                 defines the locale used to get day names
    :type locale:                  str
    :param format:                 formats result using this strftime format
    :type format:                  str
    :param begin_at_sunday_plus:   Determines the left-most day in the display. 0=sunday, 1=monday, etc
    :type begin_at_sunday_plus:    (int)
    :param month_names:            optional list of month names to use (should be 12 items)
    :type month_names:             List[str]
    :param day_abbreviations:      optional list of abbreviations to display as the day of week
    :type day_abbreviations:       List[str]
    :param title:                  Title shown on the date chooser window
    :type title:                   (str)
    :param no_titlebar:            if True no titlebar will be shown on the date chooser window
    :type no_titlebar:             bool
    :param location:               Location on the screen (x,y) to show the calendar popup window
    :type location:                (int, int)
    :param visible:                set initial visibility state of the Button
    :type visible:                 (bool)
    :param metadata:               Anything you want to store along with this button
    :type metadata:                (Any)
    :param expand_x:               If True Element will expand in the Horizontal directions
    :type expand_x:                (bool)
    :param expand_y:               If True Element will expand in the Vertical directions
    :type expand_y:                (bool)
    :return:                       returns a button
    :rtype:                        (Button)
    """
    button = Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_CALENDAR_CHOOSER,
        target=target,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        enable_events=enable_events,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )
    button.calendar_close_when_chosen = close_when_date_chosen
    button.calendar_default_date_M_D_Y = default_date_m_d_y
    button.calendar_locale = locale
    button.calendar_format = format
    button.calendar_no_titlebar = no_titlebar
    button.calendar_location = location
    button.calendar_begin_at_sunday_plus = begin_at_sunday_plus
    button.calendar_month_names = month_names
    button.calendar_day_abbreviations = day_abbreviations
    button.calendar_title = title

    return button


# -------------------------  Calendar Chooser Button lazy function  ------------------------- #
def ColorChooserButton(
        button_text,
        target=(ThisRow, -1),
        image_filename=None,
        image_data=None,
        image_size=(None, None),
        image_subsample=None,
        tooltip=None,
        border_width=None,
        size=(None, None),
        s=(None, None),
        auto_size_button=None,
        button_color=None,
        disabled=False,
        font=None,
        bind_return_key=False,
        focus=False,
        pad=None,
        p=None,
        key=None,
        k=None,
        default_color=None,
        visible=True,
        metadata=None,
        expand_x=False,
        expand_y=False,
):
    """

    :param button_text:      text in the button
    :type button_text:       (str)
    :param target:           key or (row,col) target for the button. Note that -1 for column means 1 element to the left of this one. The constant ThisRow is used to indicate the current row. The Button itself is a valid target for some types of button
    :type target:            str | (int, int)
    :type image_filename:    (str)
    :param image_filename:   image filename if there is a button image. GIFs and PNGs only.
    :type image_filename:    (str)
    :param image_data:       Raw or Base64 representation of the image to put on button. Choose either filename or data
    :type image_data:        bytes | str
    :param image_size:       Size of the image in pixels (width, height)
    :type image_size:        (int, int)
    :param image_subsample:  amount to reduce the size of the image. Divides the size by this number. 2=1/2, 3=1/3, 4=1/4, etc
    :type image_subsample:   (int)
    :param tooltip:          text, that will appear when mouse hovers over the element
    :type tooltip:           (str)
    :param border_width:     width of border around element
    :type border_width:      (int)
    :param size:             (w,h) w=characters-wide, h=rows-high
    :type size:              (int, int)
    :param s:                Same as size parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, size will be used
    :type s:                 (int, int)  | (None, None) | int
    :param auto_size_button: True if button size is determined by button text
    :type auto_size_button:  (bool)
    :param button_color:     button color (foreground, background)
    :type button_color:      (str, str) | str
    :param disabled:         set disable state for element (Default = False)
    :type disabled:          (bool)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:              (str or (str, int[, str]) or None)
    :param bind_return_key:  (Default = False) If True, this button will appear to be clicked when return key is pressed in other elements such as Input and elements with return key options
    :type bind_return_key:   (bool)
    :param focus:            Determines if initial focus should go to this element.
    :type focus:             (bool)
    :param pad:              Amount of padding to put around element in pixels (left/right, top/bottom) or ((left, right), (top, bottom)) or an int. If an int, then it's converted into a tuple (int, int)
    :type pad:               (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param p:                Same as pad parameter.  It's an alias. If EITHER of them are set, then the one that's set will be used. If BOTH are set, pad will be used
    :type p:                 (int, int) or ((int, int),(int,int)) or (int,(int,int)) or  ((int, int),int) | int
    :param key:              key for uniquely identify this element (for window.find_element)
    :type key:               str | int | tuple | object
    :param k:                Same as the Key. You can use either k or key. Which ever is set will be used.
    :type k:                 str | int | tuple | object
    :param default_color:    Color to be sent to tkinter to use as the default color
    :type default_color:     str
    :param visible:          set initial visibility state of the Button
    :type visible:           (bool)
    :param metadata:         User metadata that can be set to ANYTHING
    :type metadata:          (Any)
    :param expand_x:         If True Element will expand in the Horizontal directions
    :type expand_x:          (bool)
    :param expand_y:         If True Element will expand in the Vertical directions
    :type expand_y:          (bool)
    :return:                 returns a button
    :rtype:                  (Button)
    """
    button = Button(
        button_text=button_text,
        button_type=BUTTON_TYPE_COLOR_CHOOSER,
        target=target,
        image_filename=image_filename,
        image_data=image_data,
        image_size=image_size,
        image_subsample=image_subsample,
        border_width=border_width,
        tooltip=tooltip,
        size=size,
        s=s,
        auto_size_button=auto_size_button,
        button_color=button_color,
        font=font,
        disabled=disabled,
        bind_return_key=bind_return_key,
        focus=focus,
        pad=pad,
        p=p,
        key=key,
        k=k,
        visible=visible,
        metadata=metadata,
        expand_x=expand_x,
        expand_y=expand_y,
    )
    button.default_color = default_color
    return button


#####################################  -----  BUTTON Functions   ------ ##################################################


def button_color_to_tuple(color_tuple_or_string, default=(None, None)):
    """
    Convert a color tuple or color string into 2 components and returns them as a tuple
    (Text Color, Button Background Color)
    If None is passed in as the first parameter, then the theme's button color is
    returned

    :param color_tuple_or_string: Button color - tuple or a simplied color string with word "on" between color
    :type  color_tuple_or_string: str | (str, str)
    :param default:               The 2 colors to use if there is a problem. Otherwise defaults to the theme's button color
    :type  default:               (str, str)
    :return:                      (str | (str, str)
    :rtype:                       str | (str, str)
    """
    if default == (None, None):
        color_tuple = _simplified_dual_color_to_tuple(
            color_tuple_or_string, default=theme_button_color()
        )
    elif color_tuple_or_string == COLOR_SYSTEM_DEFAULT:
        color_tuple = (COLOR_SYSTEM_DEFAULT, COLOR_SYSTEM_DEFAULT)
    else:
        color_tuple = _simplified_dual_color_to_tuple(
            color_tuple_or_string, default=default
        )

    return color_tuple


def _simplified_dual_color_to_tuple(color_tuple_or_string, default=(None, None)):
    """
    Convert a color tuple or color string into 2 components and returns them as a tuple
    (Text Color, Button Background Color)
    If None is passed in as the first parameter, theme_

    :param color_tuple_or_string: Button color - tuple or a simplied color string with word "on" between color
    :type  color_tuple_or_string: str | (str, str} | (None, None)
    :param default:               The 2 colors to use if there is a problem. Otherwise defaults to the theme's button color
    :type  default:               (str, str)
    :return:                      (str | (str, str)
    :rtype:                       str | (str, str)
    """
    if color_tuple_or_string is None or color_tuple_or_string == (None, None):
        color_tuple_or_string = default
    if color_tuple_or_string == COLOR_SYSTEM_DEFAULT:
        return COLOR_SYSTEM_DEFAULT, COLOR_SYSTEM_DEFAULT
    text_color = background_color = COLOR_SYSTEM_DEFAULT
    try:
        if isinstance(color_tuple_or_string, (tuple, list)):
            if len(color_tuple_or_string) >= 2:
                text_color = color_tuple_or_string[0] or default[0]
                background_color = color_tuple_or_string[1] or default[1]
            elif len(color_tuple_or_string) == 1:
                background_color = color_tuple_or_string[0] or default[1]
        elif isinstance(color_tuple_or_string, str):
            color_tuple_or_string = color_tuple_or_string.lower()
            split_colors = color_tuple_or_string.split(" on ")
            if len(split_colors) >= 2:
                text_color = split_colors[0].strip() or default[0]
                background_color = split_colors[1].strip() or default[1]
            elif len(split_colors) == 1:
                split_colors = color_tuple_or_string.split("on")
                if len(split_colors) == 1:
                    text_color, background_color = default[0], split_colors[0].strip()
                else:
                    split_colors = split_colors[0].strip(), split_colors[1].strip()
                    text_color = split_colors[0] or default[0]
                    background_color = split_colors[1] or default[1]
                    # text_color, background_color = color_tuple_or_string, default[1]
            else:
                text_color, background_color = default
        else:
            if not SUPPRESS_ERROR_POPUPS:
                _error_popup_with_traceback(
                    "** Badly formatted dual-color... not a tuple nor string **",
                    color_tuple_or_string,
                )
            else:
                print(
                    "** Badly formatted dual-color... not a tuple nor string **",
                    color_tuple_or_string,
                )
            text_color, background_color = default
    except Exception as e:
        if not SUPPRESS_ERROR_POPUPS:
            _error_popup_with_traceback(
                "** Badly formatted button color **", color_tuple_or_string, e
            )
        else:
            print(
                "** Badly formatted button color... not a tuple nor string **",
                color_tuple_or_string,
                e,
            )
        text_color, background_color = default
    if isinstance(text_color, int):
        text_color = "#%06X" % text_color
    if isinstance(background_color, int):
        background_color = "#%06X" % background_color
    # print('converted button color', color_tuple_or_string, 'to', (text_color, background_color))

    return text_color, background_color


#####################################  -----  RESULTS   ------ ##################################################


def AddToReturnDictionary(form, element, value):
    form.ReturnValuesDictionary[element.Key] = value
    # if element.Key is None:
    #     form.ReturnValuesDictionary[form.DictionaryKeyCounter] = value
    #     element.Key = form.DictionaryKeyCounter
    #     form.DictionaryKeyCounter += 1
    # else:
    #     form.ReturnValuesDictionary[element.Key] = value


def AddToReturnList(form, value):
    form.ReturnValuesList.append(value)


# ----------------------------------------------------------------------------#
# -------  FUNCTION InitializeResults.  Sets up form results matrix  --------#
def InitializeResults(form):
    _BuildResults(form, True, form)
    return


# =====  Radio Button RadVar encoding and decoding =====#
# =====  The value is simply the row * 1000 + col  =====#
def DecodeRadioRowCol(RadValue):
    container = RadValue // 100000
    row = RadValue // 1000
    col = RadValue % 1000
    return container, row, col


def EncodeRadioRowCol(container, row, col):
    RadValue = container * 100000 + row * 1000 + col
    return RadValue


# -------  FUNCTION BuildResults.  Form exiting so build the results to pass back  ------- #
# format of return values is
# (Button Pressed, input_values)
def _BuildResults(form, initialize_only, top_level_form):
    # Results for elements are:
    #   TEXT - Nothing
    #   INPUT - Read value from TK
    #   Button - Button Text and position as a Tuple

    # Get the initialized results so we don't have to rebuild
    # form.DictionaryKeyCounter = 0
    form.ReturnValuesDictionary = {}
    form.ReturnValuesList = []
    _BuildResultsForSubform(form, initialize_only, top_level_form)
    if not top_level_form.LastButtonClickedWasRealtime:
        top_level_form.LastButtonClicked = None
    return form.ReturnValues


def _BuildResultsForSubform(form, initialize_only, top_level_form):
    event = top_level_form.LastButtonClicked
    for row_num, row in enumerate(form.Rows):
        for col_num, element in enumerate(row):
            if element.Key is not None and WRITE_ONLY_KEY in str(element.Key):
                continue
            value = None
            if element.Type == ELEM_TYPE_COLUMN:
                element.DictionaryKeyCounter = top_level_form.DictionaryKeyCounter
                element.ReturnValuesList = []
                element.ReturnValuesDictionary = {}
                _BuildResultsForSubform(element, initialize_only, top_level_form)
                for item in element.ReturnValuesList:
                    AddToReturnList(top_level_form, item)
                if element.UseDictionary:
                    top_level_form.UseDictionary = True
                if element.ReturnValues[0] is not None:  # if a button was clicked
                    event = element.ReturnValues[0]

            if element.Type == ELEM_TYPE_FRAME:
                element.DictionaryKeyCounter = top_level_form.DictionaryKeyCounter
                element.ReturnValuesList = []
                element.ReturnValuesDictionary = {}
                _BuildResultsForSubform(element, initialize_only, top_level_form)
                for item in element.ReturnValuesList:
                    AddToReturnList(top_level_form, item)
                if element.UseDictionary:
                    top_level_form.UseDictionary = True
                if element.ReturnValues[0] is not None:  # if a button was clicked
                    event = element.ReturnValues[0]

            if element.Type == ELEM_TYPE_PANE:
                element.DictionaryKeyCounter = top_level_form.DictionaryKeyCounter
                element.ReturnValuesList = []
                element.ReturnValuesDictionary = {}
                _BuildResultsForSubform(element, initialize_only, top_level_form)
                for item in element.ReturnValuesList:
                    AddToReturnList(top_level_form, item)
                if element.UseDictionary:
                    top_level_form.UseDictionary = True
                if element.ReturnValues[0] is not None:  # if a button was clicked
                    event = element.ReturnValues[0]

            if element.Type == ELEM_TYPE_TAB_GROUP:
                element.DictionaryKeyCounter = top_level_form.DictionaryKeyCounter
                element.ReturnValuesList = []
                element.ReturnValuesDictionary = {}
                _BuildResultsForSubform(element, initialize_only, top_level_form)
                for item in element.ReturnValuesList:
                    AddToReturnList(top_level_form, item)
                if element.UseDictionary:
                    top_level_form.UseDictionary = True
                if element.ReturnValues[0] is not None:  # if a button was clicked
                    event = element.ReturnValues[0]

            if element.Type == ELEM_TYPE_TAB:
                element.DictionaryKeyCounter = top_level_form.DictionaryKeyCounter
                element.ReturnValuesList = []
                element.ReturnValuesDictionary = {}
                _BuildResultsForSubform(element, initialize_only, top_level_form)
                for item in element.ReturnValuesList:
                    AddToReturnList(top_level_form, item)
                if element.UseDictionary:
                    top_level_form.UseDictionary = True
                if element.ReturnValues[0] is not None:  # if a button was clicked
                    event = element.ReturnValues[0]

            if not initialize_only:
                if element.Type == ELEM_TYPE_INPUT_TEXT:
                    try:
                        value = element.TKStringVar.get()
                    except Exception:
                        value = ""
                    if (
                            not top_level_form.NonBlocking
                            and not element.do_not_clear
                            and not top_level_form.ReturnKeyboardEvents
                    ):
                        element.TKStringVar.set("")
                elif element.Type == ELEM_TYPE_INPUT_CHECKBOX:
                    value = element.TKIntVar.get()
                    value = value != 0
                elif element.Type == ELEM_TYPE_INPUT_RADIO:
                    RadVar = element.TKIntVar.get()
                    this_rowcol = EncodeRadioRowCol(
                        form.ContainerElemementNumber, row_num, col_num
                    )
                    # this_rowcol = element.EncodedRadioValue       # could use the saved one
                    value = RadVar == this_rowcol
                elif element.Type == ELEM_TYPE_BUTTON:
                    if top_level_form.LastButtonClicked == element.Key:
                        event = top_level_form.LastButtonClicked
                        if (
                                element.BType != BUTTON_TYPE_REALTIME
                        ):  # Do not clear realtime buttons
                            top_level_form.LastButtonClicked = None
                    if element.BType == BUTTON_TYPE_CALENDAR_CHOOSER:
                        # value = None
                        value = element.calendar_selection
                    else:
                        try:
                            value = element.TKStringVar.get()
                        except Exception:
                            value = None
                elif element.Type == ELEM_TYPE_INPUT_COMBO:
                    element = element  # type: Combo
                    # value = element.TKStringVar.get()
                    try:
                        if (
                                element.TKCombo.current() == -1
                        ):  # if the current value was not in the original list
                            value = element.TKCombo.get()
                        else:
                            value = element.Values[
                                element.TKCombo.current()
                            ]  # get value from original list given index
                    except Exception:
                        value = "*Exception occurred*"
                elif element.Type == ELEM_TYPE_INPUT_OPTION_MENU:
                    value = element.TKStringVar.get()
                elif element.Type == ELEM_TYPE_INPUT_LISTBOX:
                    try:
                        items = element.TKListbox.curselection()
                        value = [element.Values[int(item)] for item in items]
                    except Exception as e:
                        value = ""
                elif element.Type == ELEM_TYPE_INPUT_SPIN:
                    try:
                        value = element.TKStringVar.get()
                        for v in element.Values:
                            if str(v) == value:
                                value = v
                                break
                    except Exception:
                        value = 0
                elif element.Type == ELEM_TYPE_INPUT_SLIDER:
                    try:
                        value = float(element.TKScale.get())
                    except Exception:
                        value = 0
                elif element.Type == ELEM_TYPE_INPUT_MULTILINE:
                    if element.WriteOnly:  # if marked as "write only" when created, then don't include with the values being returned
                        continue
                    try:
                        value = element.TKText.get(1.0, tk.END)
                        if element.rstrip:
                            value = value.rstrip()
                        if (
                                not top_level_form.NonBlocking
                                and not element.do_not_clear
                                and not top_level_form.ReturnKeyboardEvents
                        ):
                            element.TKText.delete("1.0", tk.END)
                    except Exception:
                        value = None
                elif element.Type == ELEM_TYPE_TAB_GROUP:
                    try:
                        value = element.TKNotebook.tab(
                            element.TKNotebook.index("current")
                        )["text"]
                        tab_key = element.find_currently_active_tab_key()
                        # tab_key = element.FindKeyFromTabName(value)
                        if tab_key is not None:
                            value = tab_key
                    except Exception:
                        value = None
                elif element.Type == ELEM_TYPE_TABLE:
                    value = element.SelectedRows
                elif element.Type == ELEM_TYPE_TREE:
                    value = element.SelectedRows
                elif element.Type == ELEM_TYPE_GRAPH:
                    value = element.ClickPosition
                elif element.Type == ELEM_TYPE_MENUBAR:
                    if element.MenuItemChosen is not None:
                        event = top_level_form.LastButtonClicked = (
                            element.MenuItemChosen
                        )
                    value = element.MenuItemChosen
                    element.MenuItemChosen = None
                elif element.Type == ELEM_TYPE_BUTTONMENU:
                    element = element  # type: ButtonMenu
                    value = element.MenuItemChosen
                    if element.part_of_custom_menubar:
                        if element.MenuItemChosen is not None:
                            value = event = element.MenuItemChosen
                            top_level_form.LastButtonClicked = element.MenuItemChosen
                            if element.custom_menubar_key is not None:
                                top_level_form.ReturnValuesDictionary[
                                    element.custom_menubar_key
                                ] = value
                            element.MenuItemChosen = None
                        else:
                            if (
                                    element.custom_menubar_key
                                    not in top_level_form.ReturnValuesDictionary
                            ):
                                top_level_form.ReturnValuesDictionary[
                                    element.custom_menubar_key
                                ] = None
                            value = None

                    # if element.MenuItemChosen is not None:
                    #     button_pressed_text = top_level_form.LastButtonClicked = element.MenuItemChosen
                    # value = element.MenuItemChosen
                    # element.MenuItemChosen = None
            else:
                value = None

            # if an input type element, update the results
            if element.Type not in (
                    ELEM_TYPE_BUTTON,
                    ELEM_TYPE_TEXT,
                    ELEM_TYPE_IMAGE,
                    ELEM_TYPE_OUTPUT,
                    ELEM_TYPE_PROGRESS_BAR,
                    ELEM_TYPE_COLUMN,
                    ELEM_TYPE_FRAME,
                    ELEM_TYPE_SEPARATOR,
                    ELEM_TYPE_TAB,
            ):
                if not (
                        element.Type == ELEM_TYPE_BUTTONMENU
                        and element.part_of_custom_menubar
                ):
                    AddToReturnList(form, value)
                    AddToReturnDictionary(top_level_form, element, value)
            elif (
                    element.Type == ELEM_TYPE_BUTTON
                    and element.BType == BUTTON_TYPE_COLOR_CHOOSER
                    and element.Target == (None, None)
            ) or (
                    element.Type == ELEM_TYPE_BUTTON
                    and element.Key is not None
                    and (
                            element.BType
                            in (
                                    BUTTON_TYPE_SAVEAS_FILE,
                                    BUTTON_TYPE_BROWSE_FILE,
                                    BUTTON_TYPE_BROWSE_FILES,
                                    BUTTON_TYPE_BROWSE_FOLDER,
                                    BUTTON_TYPE_CALENDAR_CHOOSER,
                            )
                    )
            ):
                AddToReturnList(form, value)
                AddToReturnDictionary(top_level_form, element, value)

    # if this is a column, then will fail so need to wrap with try
    try:
        if form.ReturnKeyboardEvents and form.LastKeyboardEvent is not None:
            event = form.LastKeyboardEvent
            form.LastKeyboardEvent = None
    except Exception:
        pass

    try:
        form.ReturnValuesDictionary.pop(
            None, None
        )  # clean up dictionary include None was included
    except Exception:
        pass

    # if no event was found
    if not initialize_only and event is None and form == top_level_form:
        queued_event_value = form._queued_thread_event_read()
        if queued_event_value is not None:
            event, value = queued_event_value
            AddToReturnList(form, value)
            form.ReturnValuesDictionary[event] = value

    if not form.UseDictionary:
        form.ReturnValues = event, form.ReturnValuesList
    else:
        form.ReturnValues = event, form.ReturnValuesDictionary

    return form.ReturnValues


def fill_form_with_values(window, values_dict):
    """
    Fills a window with values provided in a values dictionary { element_key : new_value }

    :param window:      The window object to fill
    :type window:       (Window)
    :param values_dict: A dictionary with element keys as key and value is values parm for Update call
    :type values_dict:  (Dict[Any, Any])
    :return:            None
    :rtype:             None
    """

    for element_key in values_dict:
        try:
            window.AllKeysDict[element_key].Update(values_dict[element_key])
        except Exception as e:
            print(
                "Problem filling form. Perhaps bad key?  This is a suspected bad key: {}".format(
                    element_key
                )
            )


def _FindElementWithFocusInSubForm(form):
    """
    Searches through a "sub-form" (can be a window or container) for the current element with focus

    :param form: a Window, Column, Frame, or TabGroup (container elements)
    :type form:  container elements
    :return:     Element
    :rtype:      Element | None
    """
    for row_num, row in enumerate(form.Rows):
        for col_num, element in enumerate(row):
            if element.Type == ELEM_TYPE_COLUMN:
                matching_elem = _FindElementWithFocusInSubForm(element)
                if matching_elem is not None:
                    return matching_elem
            elif element.Type == ELEM_TYPE_FRAME:
                matching_elem = _FindElementWithFocusInSubForm(element)
                if matching_elem is not None:
                    return matching_elem
            elif element.Type == ELEM_TYPE_TAB_GROUP:
                matching_elem = _FindElementWithFocusInSubForm(element)
                if matching_elem is not None:
                    return matching_elem
            elif element.Type == ELEM_TYPE_TAB:
                matching_elem = _FindElementWithFocusInSubForm(element)
                if matching_elem is not None:
                    return matching_elem
            elif element.Type == ELEM_TYPE_PANE:
                matching_elem = _FindElementWithFocusInSubForm(element)
                if matching_elem is not None:
                    return matching_elem
            elif element.Type == ELEM_TYPE_INPUT_TEXT:
                if element.TKEntry is not None:
                    if element.TKEntry is element.TKEntry.focus_get():
                        return element
            elif element.Type == ELEM_TYPE_INPUT_MULTILINE:
                if element.TKText is not None:
                    if element.TKText is element.TKText.focus_get():
                        return element
            elif element.Type == ELEM_TYPE_BUTTON:
                if element.TKButton is not None:
                    if element.TKButton is element.TKButton.focus_get():
                        return element
            else:  # The "Catch All" - if type isn't one of the above, try generic element.Widget
                try:
                    if element.Widget is not None:
                        if element.Widget is element.Widget.focus_get():
                            return element
                except Exception:
                    return None

    return None


def AddMenuItem(
        top_menu,
        sub_menu_info,
        element,
        is_sub_menu=False,
        skip=False,
        right_click_menu=False,
):
    """
    Only to be used internally. Not user callable
    :param right_click_menu:
    :param top_menu:      ???
    :type top_menu:       ???
    :param sub_menu_info: ???
    :type sub_menu_info:
    :param element:       ???
    :type element:        idk_yetReally
    :param is_sub_menu:   (Default = False)
    :type is_sub_menu:    (bool)
    :param skip:          (Default = False)
    :type skip:           (bool)

    """
    return_val = None
    if type(sub_menu_info) is str:
        if not is_sub_menu and not skip:
            pos = sub_menu_info.find(MENU_SHORTCUT_CHARACTER)
            if pos != -1:
                if (
                        pos < len(MENU_SHORTCUT_CHARACTER)
                        or sub_menu_info[pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                ):
                    sub_menu_info = (
                            sub_menu_info[:pos]
                            + sub_menu_info[pos + len(MENU_SHORTCUT_CHARACTER):]
                    )
            if sub_menu_info == "---":
                top_menu.add("separator")
            else:
                try:
                    item_without_key = sub_menu_info[
                        : sub_menu_info.index(MENU_KEY_SEPARATOR)
                    ]
                except Exception:
                    item_without_key = sub_menu_info

                if item_without_key[0] == MENU_DISABLED_CHARACTER:
                    top_menu.add_command(
                        label=item_without_key[len(MENU_DISABLED_CHARACTER):],
                        underline=pos - 1,
                        command=lambda: element._MenuItemChosenCallback(sub_menu_info),
                    )
                    top_menu.entryconfig(
                        item_without_key[len(MENU_DISABLED_CHARACTER):],
                        state="disabled",
                    )
                else:
                    top_menu.add_command(
                        label=item_without_key,
                        underline=pos,
                        command=lambda: element._MenuItemChosenCallback(sub_menu_info),
                    )
    else:
        i = 0
        while i < (len(sub_menu_info)):
            item = sub_menu_info[i]
            if i != len(sub_menu_info) - 1:
                if type(sub_menu_info[i + 1]) == list:
                    new_menu = tk.Menu(top_menu, tearoff=element.Tearoff)
                    # if a right click menu, then get styling from the top-level window
                    if right_click_menu:
                        window = element.ParentForm
                        if window.right_click_menu_background_color not in (
                                COLOR_SYSTEM_DEFAULT,
                                None,
                        ):
                            new_menu.config(bg=window.right_click_menu_background_color)
                            new_menu.config(
                                activeforeground=window.right_click_menu_background_color
                            )
                        if window.right_click_menu_text_color not in (
                                COLOR_SYSTEM_DEFAULT,
                                None,
                        ):
                            new_menu.config(fg=window.right_click_menu_text_color)
                            new_menu.config(
                                activebackground=window.right_click_menu_text_color
                            )
                        if window.right_click_menu_disabled_text_color not in (
                                COLOR_SYSTEM_DEFAULT,
                                None,
                        ):
                            new_menu.config(
                                disabledforeground=window.right_click_menu_disabled_text_color
                            )
                        if window.right_click_menu_font is not None:
                            new_menu.config(font=window.right_click_menu_font)
                    else:
                        if element.Font is not None:
                            new_menu.config(font=element.Font)
                        if element.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                            new_menu.config(bg=element.BackgroundColor)
                            new_menu.config(activeforeground=element.BackgroundColor)
                        if element.TextColor not in (COLOR_SYSTEM_DEFAULT, None):
                            new_menu.config(fg=element.TextColor)
                            new_menu.config(activebackground=element.TextColor)
                        if element.DisabledTextColor not in (
                                COLOR_SYSTEM_DEFAULT,
                                None,
                        ):
                            new_menu.config(
                                disabledforeground=element.DisabledTextColor
                            )
                        if element.ItemFont is not None:
                            new_menu.config(font=element.ItemFont)
                    return_val = new_menu
                    pos = sub_menu_info[i].find(MENU_SHORTCUT_CHARACTER)
                    if pos != -1:
                        if (
                                pos < len(MENU_SHORTCUT_CHARACTER)
                                or sub_menu_info[i][pos - len(MENU_SHORTCUT_CHARACTER)]
                                != "\\"
                        ):
                            sub_menu_info[i] = (
                                    sub_menu_info[i][:pos]
                                    + sub_menu_info[i][pos + len(MENU_SHORTCUT_CHARACTER):]
                            )
                    if sub_menu_info[i][0] == MENU_DISABLED_CHARACTER:
                        top_menu.add_cascade(
                            label=sub_menu_info[i][len(MENU_DISABLED_CHARACTER):],
                            menu=new_menu,
                            underline=pos,
                            state="disabled",
                        )
                    else:
                        top_menu.add_cascade(
                            label=sub_menu_info[i], menu=new_menu, underline=pos
                        )
                    AddMenuItem(
                        new_menu,
                        sub_menu_info[i + 1],
                        element,
                        is_sub_menu=True,
                        right_click_menu=right_click_menu,
                    )
                    i += 1  # skip the next one
                else:
                    AddMenuItem(
                        top_menu, item, element, right_click_menu=right_click_menu
                    )
            else:
                AddMenuItem(top_menu, item, element, right_click_menu=right_click_menu)
            i += 1
    return return_val


# 888    888      d8b          888
# 888    888      Y8P          888
# 888    888                   888
# 888888 888  888 888 88888b.  888888  .d88b.  888d888
# 888    888 .88P 888 888 "88b 888    d8P  Y8b 888P"
# 888    888888K  888 888  888 888    88888888 888
# Y88b.  888 "88b 888 888  888 Y88b.  Y8b.     888
#  "Y888 888  888 888 888  888  "Y888  "Y8888  888

# My crappy tkinter code starts here.  (search for "crappy" to get here quickly... that's the purpose if you hadn't caught on

"""
         )
        (
          ,
       ___)\
      (_____)
      (_______)

"""


# Chr0nic || This is probably *very* bad practice. But it works. Simple, but it works...
class VarHolder(object):
    canvas_holder = None

    def __init__(self):
        self.canvas_holder = None


# Also, to get to the point in the code where each element's widget is created, look for element + "p lacement" (without the space)


# ========================   TK CODE STARTS HERE ========================================= #
def _fixed_map(style, style_name, option, highlight_colors=(None, None)):
    # Fix for setting text colour for Tkinter 8.6.9
    # From: https://core.tcl.tk/tk/info/509cafafae

    # default_map = [elm for elm in style.map("Treeview", query_opt=option) if '!' not in elm[0]]
    # custom_map = [elm for elm in style.map(style_name, query_opt=option) if '!' not in elm[0]]
    default_map = [
        elm
        for elm in style.map("Treeview", query_opt=option)
        if "!" not in elm[0] and "selected" not in elm[0]
    ]
    custom_map = [
        elm
        for elm in style.map(style_name, query_opt=option)
        if "!" not in elm[0] and "selected" not in elm[0]
    ]
    if option == "background":
        custom_map.append(
            (
                "selected",
                highlight_colors[1]
                if highlight_colors[1] is not None
                else ALTERNATE_TABLE_AND_TREE_SELECTED_ROW_COLORS[1],
            )
        )
    elif option == "foreground":
        custom_map.append(
            (
                "selected",
                highlight_colors[0]
                if highlight_colors[0] is not None
                else ALTERNATE_TABLE_AND_TREE_SELECTED_ROW_COLORS[0],
            )
        )

    new_map = custom_map + default_map
    return new_map
    #
    # new_map = [elm for elm in style.map(style_name, query_opt=option) if elm[:2] != ('!disabled', '!selected')]
    #
    # if option == 'background':
    #     new_map.append(('selected', highlight_colors[1] if highlight_colors[1] is not None else ALTERNATE_TABLE_AND_TREE_SELECTED_ROW_COLORS[1]))
    # elif option == 'foreground':
    #     new_map.append(('selected', highlight_colors[0] if highlight_colors[0] is not None else ALTERNATE_TABLE_AND_TREE_SELECTED_ROW_COLORS[0]))
    # return new_map
    #


def _add_right_click_menu(element, toplevel_form):
    if element.RightClickMenu == MENU_RIGHT_CLICK_DISABLED:
        return
    if element.RightClickMenu or toplevel_form.RightClickMenu:
        menu = element.RightClickMenu or toplevel_form.RightClickMenu
        top_menu = tk.Menu(
            toplevel_form.TKroot,
            tearoff=toplevel_form.right_click_menu_tearoff,
            tearoffcommand=element._tearoff_menu_callback,
        )

        if toplevel_form.right_click_menu_background_color not in (
                COLOR_SYSTEM_DEFAULT,
                None,
        ):
            top_menu.config(bg=toplevel_form.right_click_menu_background_color)
        if toplevel_form.right_click_menu_text_color not in (
                COLOR_SYSTEM_DEFAULT,
                None,
        ):
            top_menu.config(fg=toplevel_form.right_click_menu_text_color)
        if toplevel_form.right_click_menu_disabled_text_color not in (
                COLOR_SYSTEM_DEFAULT,
                None,
        ):
            top_menu.config(
                disabledforeground=toplevel_form.right_click_menu_disabled_text_color
            )
        if toplevel_form.right_click_menu_font is not None:
            top_menu.config(font=toplevel_form.right_click_menu_font)

        if toplevel_form.right_click_menu_selected_colors[0] not in (
                COLOR_SYSTEM_DEFAULT,
                None,
        ):
            top_menu.config(
                activeforeground=toplevel_form.right_click_menu_selected_colors[0]
            )
        if toplevel_form.right_click_menu_selected_colors[1] not in (
                COLOR_SYSTEM_DEFAULT,
                None,
        ):
            top_menu.config(
                activebackground=toplevel_form.right_click_menu_selected_colors[1]
            )
        AddMenuItem(top_menu, menu[1], element, right_click_menu=True)
        element.TKRightClickMenu = top_menu
        if running_mac():
            element.Widget.bind("<ButtonRelease-2>", element._RightClickMenuCallback)
        else:
            element.Widget.bind("<ButtonRelease-3>", element._RightClickMenuCallback)


def _change_ttk_theme(style, theme_name):
    global ttk_theme_in_use
    if theme_name not in style.theme_names():
        _error_popup_with_traceback(
            'You are trying to use TTK theme "{}"'.format(theme_name),
            "This is not legal for your system",
            "The valid themes to choose from are: {}".format(
                ", ".join(style.theme_names())
            ),
        )
        return False

    style.theme_use(theme_name)
    ttk_theme_in_use = theme_name
    return True


# class Stylist:
#     """
#     A class to help get information about ttk styles
#     """
#     @staticmethod
#     def get_elements(layout):
#         """Return a list of elements contained in the style"""
#         elements = []
#         element = layout[0][0]
#         elements.append(element)
#         sublayout = layout[0][1]
#
#         if 'children' in sublayout:
#             child_elements = Stylist.get_elements(sublayout['children'])
#             elements.extend(child_elements)
#         return elements
#
#     @staticmethod
#     def get_options(ttkstyle, theme=None):
#         style = ttk.Style()
#         if theme is not None:
#             style.theme_use(theme)
#         layout = style.layout(ttkstyle)
#         elements = Stylist.get_elements(layout)
#         options = []
#         for e in elements:
#             _opts = style.element_options(e)
#             if _opts:
#                 options.extend(list(_opts))
#         return list(set(options))
#
#     @staticmethod
#     def create_style(base_style: str, theme=None, **kwargs):
#         style = ttk.Style()
#         if theme is not None:
#             style.theme_use(theme)
#         style_id = uuid4()
#         ttkstyle = '{}.{}'.format(style_id, base_style)
#         style.configure(ttkstyle, **kwargs)
#         return ttkstyle


def _make_ttk_style_name(base_style, element, primary_style=False):
    Window._counter_for_ttk_widgets += 1
    style_name = (
            str(Window._counter_for_ttk_widgets) + "___" + str(element.Key) + base_style
    )
    if primary_style:
        element.ttk_style_name = style_name
    return style_name


def _make_ttk_scrollbar(element, orientation, window):
    """
    Creates a ttk scrollbar for elements as they are being added to the layout

    :param element:     The element
    :type element:      (Element)
    :param orientation: The orientation vertical ('v') or horizontal ('h')
    :type orientation:  (str)
    :param window:      The window containing the scrollbar
    :type window:       (Window)
    """

    style = ttk.Style()
    _change_ttk_theme(style, window.TtkTheme)
    if orientation[0].lower() == "v":
        orient = "vertical"
        style_name = _make_ttk_style_name(".Vertical.TScrollbar", element)
        # style_name_thumb = _make_ttk_style_name('.Vertical.TScrollbar.thumb', element)
        element.vsb_style = style
        element.vsb = ttk.Scrollbar(
            element.element_frame,
            orient=orient,
            command=element.Widget.yview,
            style=style_name,
        )
        element.vsb_style_name = style_name
    else:
        orient = "horizontal"
        style_name = _make_ttk_style_name(".Horizontal.TScrollbar", element)
        element.hsb_style = style
        element.hsb = ttk.Scrollbar(
            element.element_frame,
            orient=orient,
            command=element.Widget.xview,
            style=style_name,
        )
        element.hsb_style_name = style_name

    # ------------------ Get the colors using heirarchy of element, window, options, settings ------------------
    # Trough Color
    if element.ttk_part_overrides.sbar_trough_color is not None:
        trough_color = element.ttk_part_overrides.sbar_trough_color
    elif window.ttk_part_overrides.sbar_trough_color is not None:
        trough_color = window.ttk_part_overrides.sbar_trough_color
    elif ttk_part_overrides_from_options.sbar_trough_color is not None:
        trough_color = ttk_part_overrides_from_options.sbar_trough_color
    else:
        trough_color = element.scroll_trough_color
    # Relief
    if element.ttk_part_overrides.sbar_relief is not None:
        scroll_relief = element.ttk_part_overrides.sbar_relief
    elif window.ttk_part_overrides.sbar_relief is not None:
        scroll_relief = window.ttk_part_overrides.sbar_relief
    elif ttk_part_overrides_from_options.sbar_relief is not None:
        scroll_relief = ttk_part_overrides_from_options.sbar_relief
    else:
        scroll_relief = element.scroll_relief
    # Frame Color
    if element.ttk_part_overrides.sbar_frame_color is not None:
        frame_color = element.ttk_part_overrides.sbar_frame_color
    elif window.ttk_part_overrides.sbar_frame_color is not None:
        frame_color = window.ttk_part_overrides.sbar_frame_color
    elif ttk_part_overrides_from_options.sbar_frame_color is not None:
        frame_color = ttk_part_overrides_from_options.sbar_frame_color
    else:
        frame_color = element.scroll_frame_color
    # Background Color
    if element.ttk_part_overrides.sbar_background_color is not None:
        background_color = element.ttk_part_overrides.sbar_background_color
    elif window.ttk_part_overrides.sbar_background_color is not None:
        background_color = window.ttk_part_overrides.sbar_background_color
    elif ttk_part_overrides_from_options.sbar_background_color is not None:
        background_color = ttk_part_overrides_from_options.sbar_background_color
    else:
        background_color = element.scroll_background_color
    # Arrow Color
    if element.ttk_part_overrides.sbar_arrow_color is not None:
        arrow_color = element.ttk_part_overrides.sbar_arrow_color
    elif window.ttk_part_overrides.sbar_arrow_color is not None:
        arrow_color = window.ttk_part_overrides.sbar_arrow_color
    elif ttk_part_overrides_from_options.sbar_arrow_color is not None:
        arrow_color = ttk_part_overrides_from_options.sbar_arrow_color
    else:
        arrow_color = element.scroll_arrow_color
    # Arrow Width
    if element.ttk_part_overrides.sbar_arrow_width is not None:
        arrow_width = element.ttk_part_overrides.sbar_arrow_width
    elif window.ttk_part_overrides.sbar_arrow_width is not None:
        arrow_width = window.ttk_part_overrides.sbar_arrow_width
    elif ttk_part_overrides_from_options.sbar_arrow_width is not None:
        arrow_width = ttk_part_overrides_from_options.sbar_arrow_width
    else:
        arrow_width = element.scroll_arrow_width
    # Scroll Width
    if element.ttk_part_overrides.sbar_width is not None:
        scroll_width = element.ttk_part_overrides.sbar_width
    elif window.ttk_part_overrides.sbar_width is not None:
        scroll_width = window.ttk_part_overrides.sbar_width
    elif ttk_part_overrides_from_options.sbar_width is not None:
        scroll_width = ttk_part_overrides_from_options.sbar_width
    else:
        scroll_width = element.scroll_width

    if trough_color not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, troughcolor=trough_color)

    if frame_color not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, framecolor=frame_color)
    if frame_color not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, bordercolor=frame_color)

    if (background_color not in (None, COLOR_SYSTEM_DEFAULT)) and (
            arrow_color not in (None, COLOR_SYSTEM_DEFAULT)
    ):
        style.map(
            style_name,
            background=[
                ("selected", background_color),
                ("active", arrow_color),
                ("background", background_color),
                ("!focus", background_color),
            ],
        )
    if (background_color not in (None, COLOR_SYSTEM_DEFAULT)) and (
            arrow_color not in (None, COLOR_SYSTEM_DEFAULT)
    ):
        style.map(
            style_name,
            arrowcolor=[
                ("selected", arrow_color),
                ("active", background_color),
                ("background", background_color),
                ("!focus", arrow_color),
            ],
        )

    if scroll_width not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, width=scroll_width)
    if arrow_width not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, arrowsize=arrow_width)

    if scroll_relief not in (None, COLOR_SYSTEM_DEFAULT):
        style.configure(style_name, relief=scroll_relief)


# if __name__ == '__main__':
#     root = tk.Tk()
#
#     # find out what options are available for the theme and widget style
#     options = Stylist.get_options('TFrame', 'default')
#     print('The options for this style and theme are', options)
#
#     # create a new style
#     frame_style = Stylist.create_style('TFrame', 'alt', relief=tk.RAISED, borderwidth=1)
#
#     # apply the new style
#     ttk.Frame(style=frame_style, width=100, height=100).pack(padx=10, pady=10)
#
#     root.mainloop()


# @_timeit
def PackFormIntoFrame(form, containing_frame, toplevel_form):
    """

    :param form:             a window class
    :type form:              (Window)
    :param containing_frame: ???
    :type containing_frame:  ???
    :param toplevel_form:    ???
    :type toplevel_form:     (Window)

    """

    # Old bindings
    def yscroll_old(event):
        try:
            if event.num == 5 or event.delta < 0:
                VarHolder.canvas_holder.yview_scroll(1, "unit")
            elif event.num == 4 or event.delta > 0:
                VarHolder.canvas_holder.yview_scroll(-1, "unit")
        except Exception:
            pass

    def xscroll_old(event):
        try:
            if event.num == 5 or event.delta < 0:
                VarHolder.canvas_holder.xview_scroll(1, "unit")
            elif event.num == 4 or event.delta > 0:
                VarHolder.canvas_holder.xview_scroll(-1, "unit")
        except Exception:
            pass

    # Chr0nic
    def testMouseHook2(em):
        combo = em.TKCombo
        combo.unbind_class("TCombobox", "<MouseWheel>")
        combo.unbind_class("TCombobox", "<ButtonPress-4>")
        combo.unbind_class("TCombobox", "<ButtonPress-5>")
        containing_frame.unbind_all("<4>")
        containing_frame.unbind_all("<5>")
        containing_frame.unbind_all("<MouseWheel>")
        containing_frame.unbind_all("<Shift-MouseWheel>")

    # Chr0nic
    def testMouseUnhook2(em):
        containing_frame.bind_all("<4>", yscroll_old, add="+")
        containing_frame.bind_all("<5>", yscroll_old, add="+")
        containing_frame.bind_all("<MouseWheel>", yscroll_old, add="+")
        containing_frame.bind_all("<Shift-MouseWheel>", xscroll_old, add="+")

    # Chr0nic
    def testMouseHook(em):
        containing_frame.unbind_all("<4>")
        containing_frame.unbind_all("<5>")
        containing_frame.unbind_all("<MouseWheel>")
        containing_frame.unbind_all("<Shift-MouseWheel>")

    # Chr0nic
    def testMouseUnhook(em):
        containing_frame.bind_all("<4>", yscroll_old, add="+")
        containing_frame.bind_all("<5>", yscroll_old, add="+")
        containing_frame.bind_all("<MouseWheel>", yscroll_old, add="+")
        containing_frame.bind_all("<Shift-MouseWheel>", xscroll_old, add="+")

    def _char_width_in_pixels(font):
        return tkinter.font.Font(font=font).measure("A")  # single character width

    def _char_height_in_pixels(font):
        return tkinter.font.Font(font=font).metrics("linespace")

    def _string_width_in_pixels(font, string):
        return tkinter.font.Font(font=font).measure(string)  # single character width

    # def _valid_theme(style, theme_name):
    #     if theme_name in style.theme_names():
    #         return True
    #     _error_popup_with_traceback('Your Window has an invalid ttk theme specified',
    #                                 'The traceback will show you the Window with the problem layout',
    #                                 '** Invalid ttk theme specified {} **'.format(theme_name),
    #                                 '\nValid choices include: {}'.format(style.theme_names()))
    #
    #     # print('** Invalid ttk theme specified {} **'.format(theme_name),
    #     #       '\nValid choices include: {}'.format(style.theme_names()))
    #     return False

    def _add_grab(element):
        try:
            if form.Grab is True or element.Grab is True:
                # if something already about to the button, then don't do the grab stuff
                if "<Button-1>" not in element.Widget.bind():
                    element.Widget.bind(
                        "<ButtonPress-1>", toplevel_form._StartMoveGrabAnywhere
                    )
                    element.Widget.bind("<ButtonRelease-1>", toplevel_form._StopMove)
                    element.Widget.bind(
                        "<B1-Motion>", toplevel_form._OnMotionGrabAnywhere
                    )
                element.ParentRowFrame.bind(
                    "<ButtonPress-1>", toplevel_form._StartMoveGrabAnywhere
                )
                element.ParentRowFrame.bind(
                    "<ButtonRelease-1>", toplevel_form._StopMove
                )
                element.ParentRowFrame.bind(
                    "<B1-Motion>", toplevel_form._OnMotionGrabAnywhere
                )
                if element.Type == ELEM_TYPE_COLUMN:
                    element.TKColFrame.canvas.bind(
                        "<ButtonPress-1>", toplevel_form._StartMoveGrabAnywhere
                    )
                    element.TKColFrame.canvas.bind(
                        "<ButtonRelease-1>", toplevel_form._StopMove
                    )
                    element.TKColFrame.canvas.bind(
                        "<B1-Motion>", toplevel_form._OnMotionGrabAnywhere
                    )
        except Exception as e:
            pass
            # print(e)

    def _add_right_click_menu_and_grab(element):
        if element.RightClickMenu == MENU_RIGHT_CLICK_DISABLED:
            return
        if (
                element.Type == ELEM_TYPE_TAB_GROUP
        ):  # unless everything disabled, then need to always set a right click menu for tabgroups
            if toplevel_form.RightClickMenu == MENU_RIGHT_CLICK_DISABLED:
                return
            menu = _MENU_RIGHT_CLICK_TABGROUP_DEFAULT
        else:
            menu = (
                    element.RightClickMenu
                    or form.RightClickMenu
                    or toplevel_form.RightClickMenu
            )

        if menu:
            top_menu = tk.Menu(
                toplevel_form.TKroot,
                tearoff=toplevel_form.right_click_menu_tearoff,
                tearoffcommand=element._tearoff_menu_callback,
            )

            if toplevel_form.right_click_menu_background_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(bg=toplevel_form.right_click_menu_background_color)
            if toplevel_form.right_click_menu_text_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(fg=toplevel_form.right_click_menu_text_color)
            if toplevel_form.right_click_menu_disabled_text_color not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    disabledforeground=toplevel_form.right_click_menu_disabled_text_color
                )
            if toplevel_form.right_click_menu_font is not None:
                top_menu.config(font=toplevel_form.right_click_menu_font)

            if toplevel_form.right_click_menu_selected_colors[0] not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    activeforeground=toplevel_form.right_click_menu_selected_colors[0]
                )
            if toplevel_form.right_click_menu_selected_colors[1] not in (
                    COLOR_SYSTEM_DEFAULT,
                    None,
            ):
                top_menu.config(
                    activebackground=toplevel_form.right_click_menu_selected_colors[1]
                )
            AddMenuItem(top_menu, menu[1], element, right_click_menu=True)
            element.TKRightClickMenu = top_menu
            if toplevel_form.RightClickMenu:  # if the top level has a right click menu, then setup a callback for the Window itself
                if toplevel_form.TKRightClickMenu is None:
                    toplevel_form.TKRightClickMenu = top_menu
                    if running_mac():
                        toplevel_form.TKroot.bind(
                            "<ButtonRelease-2>", toplevel_form._RightClickMenuCallback
                        )
                    else:
                        toplevel_form.TKroot.bind(
                            "<ButtonRelease-3>", toplevel_form._RightClickMenuCallback
                        )
            if running_mac():
                element.Widget.bind(
                    "<ButtonRelease-2>", element._RightClickMenuCallback
                )
            else:
                element.Widget.bind(
                    "<ButtonRelease-3>", element._RightClickMenuCallback
                )
                try:
                    if element.Type == ELEM_TYPE_COLUMN:
                        element.TKColFrame.canvas.bind(
                            "<ButtonRelease-3>", element._RightClickMenuCallback
                        )
                except Exception:
                    pass
        _add_grab(element)

    def _add_expansion(element, row_should_expand, row_fill_direction):
        expand = True
        if element.expand_x and element.expand_y:
            fill = tk.BOTH
            row_fill_direction = tk.BOTH
            row_should_expand = True
        elif element.expand_x:
            fill = tk.X
            row_fill_direction = (
                tk.X
                if row_fill_direction == tk.NONE
                else tk.BOTH
                if row_fill_direction == tk.Y
                else tk.X
            )
        elif element.expand_y:
            fill = tk.Y
            row_fill_direction = (
                tk.Y
                if row_fill_direction == tk.NONE
                else tk.BOTH
                if row_fill_direction == tk.X
                else tk.Y
            )
            row_should_expand = True
        else:
            fill = tk.NONE
            expand = False
        return expand, fill, row_should_expand, row_fill_direction

    tclversion_detailed = tkinter.Tcl().eval("info patchlevel")

    # --------------------------------------------------------------------------- #
    # ****************  Use FlexForm to build the tkinter window ********** ----- #
    # Building is done row by row.                                                #
    # WARNING - You can't use print in this function. If the user has rerouted   #
    # stdout then there will be an error saying the window isn't finalized        #
    # --------------------------------------------------------------------------- #
    ######################### LOOP THROUGH ROWS #########################
    # *********** -------  Loop through ROWS  ------- ***********#
    for row_num, flex_row in enumerate(form.Rows):
        ######################### LOOP THROUGH ELEMENTS ON ROW #########################
        # *********** -------  Loop through ELEMENTS  ------- ***********#
        # *********** Make TK Row                             ***********#
        tk_row_frame = tk.Frame(containing_frame)
        row_should_expand = False
        row_fill_direction = tk.NONE

        if form.ElementJustification is not None:
            row_justify = form.ElementJustification
        else:
            row_justify = "l"

        for col_num, element in enumerate(flex_row):
            element.ParentRowFrame = tk_row_frame
            element.element_frame = None  # for elements that have a scrollbar too
            element.ParentForm = toplevel_form  # save the button's parent form object
            if toplevel_form.Font and (
                    element.Font == DEFAULT_FONT or element.Font is None
            ):
                font = toplevel_form.Font
            elif element.Font is not None:
                font = element.Font
            else:
                font = DEFAULT_FONT
            # -------  Determine Auto-Size setting on a cascading basis ------- #
            if element.AutoSizeText is not None:  # if element overide
                auto_size_text = element.AutoSizeText
            elif toplevel_form.AutoSizeText is not None:  # if form override
                auto_size_text = toplevel_form.AutoSizeText
            else:
                auto_size_text = DEFAULT_AUTOSIZE_TEXT
            element_type = element.Type
            # Set foreground color
            text_color = element.TextColor
            elementpad = (
                element.Pad if element.Pad is not None else toplevel_form.ElementPadding
            )
            # element.pad_used = elementpad  # store the value used back into the element
            # Determine Element size
            element_size = element.Size
            if element_size == (None, None) and element_type not in (
                    ELEM_TYPE_BUTTON,
                    ELEM_TYPE_BUTTONMENU,
            ):  # user did not specify a size
                element_size = toplevel_form.DefaultElementSize
            elif element_size == (None, None) and element_type in (
                    ELEM_TYPE_BUTTON,
                    ELEM_TYPE_BUTTONMENU,
            ):
                element_size = toplevel_form.DefaultButtonElementSize
            else:
                auto_size_text = (
                    False  # if user has specified a size then it shouldn't autosize
                )

            border_depth = (
                toplevel_form.BorderDepth
                if toplevel_form.BorderDepth is not None
                else DEFAULT_BORDER_WIDTH
            )
            try:
                if element.BorderWidth is not None:
                    border_depth = element.BorderWidth
            except Exception:
                pass

            # -------------------------  COLUMN placement element  ------------------------- #
            if element_type == ELEM_TYPE_COLUMN:
                element = element  # type: Column
                # ----------------------- SCROLLABLE Column ----------------------
                if element.Scrollable:
                    element.Widget = element.TKColFrame = TkScrollableFrame(
                        tk_row_frame, element.VerticalScrollOnly, element, toplevel_form
                    )  # do not use yet!  not working
                    PackFormIntoFrame(
                        element, element.TKColFrame.TKFrame, toplevel_form
                    )
                    element.TKColFrame.TKFrame.update()
                    if element.Size == (
                            None,
                            None,
                    ):  # if no size specified, use column width x column height/2
                        element.TKColFrame.canvas.config(
                            width=element.TKColFrame.TKFrame.winfo_reqwidth()
                                  // element.size_subsample_width,
                            height=element.TKColFrame.TKFrame.winfo_reqheight()
                                   // element.size_subsample_height,
                        )
                    else:
                        element.TKColFrame.canvas.config(
                            width=element.TKColFrame.TKFrame.winfo_reqwidth()
                                  // element.size_subsample_width,
                            height=element.TKColFrame.TKFrame.winfo_reqheight()
                                   // element.size_subsample_height,
                        )
                        if None not in (element.Size[0], element.Size[1]):
                            element.TKColFrame.canvas.config(
                                width=element.Size[0], height=element.Size[1]
                            )
                        elif element.Size[1] is not None:
                            element.TKColFrame.canvas.config(height=element.Size[1])
                        elif element.Size[0] is not None:
                            element.TKColFrame.canvas.config(width=element.Size[0])
                    if not element.BackgroundColor in (None, COLOR_SYSTEM_DEFAULT):
                        element.TKColFrame.canvas.config(
                            background=element.BackgroundColor
                        )
                        element.TKColFrame.TKFrame.config(
                            background=element.BackgroundColor,
                            borderwidth=0,
                            highlightthickness=0,
                        )
                        element.TKColFrame.config(
                            background=element.BackgroundColor,
                            borderwidth=0,
                            highlightthickness=0,
                        )
                # ----------------------- PLAIN Column ----------------------
                else:
                    if element.Size != (None, None):
                        element.Widget = element.TKColFrame = TkFixedFrame(tk_row_frame)
                        PackFormIntoFrame(
                            element, element.TKColFrame.TKFrame, toplevel_form
                        )
                        element.TKColFrame.TKFrame.update()
                        if None not in (element.Size[0], element.Size[1]):
                            element.TKColFrame.canvas.config(
                                width=element.Size[0], height=element.Size[1]
                            )
                        elif element.Size[1] is not None:
                            element.TKColFrame.canvas.config(height=element.Size[1])
                        elif element.Size[0] is not None:
                            element.TKColFrame.canvas.config(width=element.Size[0])
                        if not element.BackgroundColor in (None, COLOR_SYSTEM_DEFAULT):
                            element.TKColFrame.canvas.config(
                                background=element.BackgroundColor
                            )
                            element.TKColFrame.TKFrame.config(
                                background=element.BackgroundColor,
                                borderwidth=0,
                                highlightthickness=0,
                            )
                    else:
                        element.Widget = element.TKColFrame = tk.Frame(tk_row_frame)
                        PackFormIntoFrame(element, element.TKColFrame, toplevel_form)
                        if element.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT):
                            element.TKColFrame.config(
                                background=element.BackgroundColor,
                                borderwidth=0,
                                highlightthickness=0,
                            )

                if element.Justification is None:
                    pass
                elif element.Justification.lower().startswith("l"):
                    row_justify = "l"
                elif element.Justification.lower().startswith("c"):
                    row_justify = "c"
                elif element.Justification.lower().startswith("r"):
                    row_justify = "r"

                # anchor=tk.NW
                # side = tk.LEFT
                # row_justify = element.Justification

                # element.Widget = element.TKColFrame

                expand = True
                if element.expand_x and element.expand_y:
                    fill = tk.BOTH
                    row_fill_direction = tk.BOTH
                    row_should_expand = True
                elif element.expand_x:
                    fill = tk.X
                    row_fill_direction = tk.X
                elif element.expand_y:
                    fill = tk.Y
                    row_fill_direction = tk.Y
                    row_should_expand = True
                else:
                    fill = tk.NONE
                    expand = False

                if element.VerticalAlignment is not None:
                    anchor = tk.CENTER  # Default to center if a bad choice is made

                    if element.VerticalAlignment.lower().startswith("t"):
                        anchor = tk.N
                    if element.VerticalAlignment.lower().startswith("c"):
                        anchor = tk.CENTER
                    if element.VerticalAlignment.lower().startswith("b"):
                        anchor = tk.S
                    element.TKColFrame.pack(
                        side=tk.LEFT,
                        anchor=anchor,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        expand=expand,
                        fill=fill,
                    )
                else:
                    element.TKColFrame.pack(
                        side=tk.LEFT,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        expand=expand,
                        fill=fill,
                    )

                # element.TKColFrame.pack(side=side, padx=elementpad[0], pady=elementpad[1], expand=True, fill='both')
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKColFrame.pack_forget()

                _add_right_click_menu_and_grab(element)
                # if element.Grab:
                #     element._grab_anywhere_on()
                # row_should_expand = True
            # -------------------------  Pane placement element  ------------------------- #
            if element_type == ELEM_TYPE_PANE:
                bd = (
                    element.BorderDepth
                    if element.BorderDepth is not None
                    else border_depth
                )
                element.PanedWindow = element.Widget = tk.PanedWindow(
                    tk_row_frame,
                    orient=tk.VERTICAL
                    if element.Orientation.startswith("v")
                    else tk.HORIZONTAL,
                    borderwidth=bd,
                    bd=bd,
                )
                if element.Relief is not None:
                    element.PanedWindow.configure(relief=element.Relief)
                element.PanedWindow.configure(handlesize=element.HandleSize)
                if element.ShowHandle:
                    element.PanedWindow.config(showhandle=True)
                if element.Size != (None, None):
                    element.PanedWindow.config(
                        width=element.Size[0], height=element.Size[1]
                    )
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.PanedWindow.configure(background=element.BackgroundColor)
                for pane in element.PaneList:
                    pane.Widget = pane.TKColFrame = tk.Frame(element.PanedWindow)
                    pane.ParentPanedWindow = element.PanedWindow
                    PackFormIntoFrame(pane, pane.TKColFrame, toplevel_form)
                    if pane.visible:
                        element.PanedWindow.add(pane.TKColFrame)
                    if (
                            pane.BackgroundColor != COLOR_SYSTEM_DEFAULT
                            and pane.BackgroundColor is not None
                    ):
                        pane.TKColFrame.configure(
                            background=pane.BackgroundColor,
                            highlightbackground=pane.BackgroundColor,
                            highlightcolor=pane.BackgroundColor,
                        )
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.PanedWindow.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                # element.PanedWindow.pack(side=tk.LEFT, padx=elementpad[0], pady=elementpad[1], expand=True, fill='both')
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.PanedWindow.pack_forget()
            # -------------------------  TEXT placement element  ------------------------- #
            elif element_type == ELEM_TYPE_TEXT:
                # auto_size_text = element.AutoSizeText
                element = element  # type: Text
                display_text = element.DisplayText  # text to display
                if not auto_size_text:
                    width, height = element_size
                else:
                    width, height = None, None
                    # lines = display_text.split('\n')
                    # max_line_len = max([len(l) for l in lines])
                    # num_lines = len(lines)
                    # if max_line_len > element_size[0]:  # if text exceeds element size, the will have to wrap
                    #     width = element_size[0]
                    # else:
                    #     width = max_line_len
                    # height = num_lines
                # ---===--- LABEL widget create and place --- #
                element = element  # type: Text
                bd = (
                    element.BorderWidth
                    if element.BorderWidth is not None
                    else border_depth
                )
                stringvar = tk.StringVar()
                element.TKStringVar = stringvar
                stringvar.set(str(display_text))
                if auto_size_text:
                    width = 0
                if element.Justification is not None:
                    justification = element.Justification
                elif toplevel_form.TextJustification is not None:
                    justification = toplevel_form.TextJustification
                else:
                    justification = DEFAULT_TEXT_JUSTIFICATION
                justify = (
                    tk.LEFT
                    if justification.startswith("l")
                    else tk.CENTER
                    if justification.startswith("c")
                    else tk.RIGHT
                )
                anchor = (
                    tk.NW
                    if justification.startswith("l")
                    else tk.N
                    if justification.startswith("c")
                    else tk.NE
                )
                tktext_label = element.Widget = tk.Label(
                    tk_row_frame,
                    textvariable=stringvar,
                    width=width,
                    height=height,
                    justify=justify,
                    bd=bd,
                    font=font,
                )
                # Set wrap-length for text (in PIXELS) == PAIN IN THE ASS
                wraplen = tktext_label.winfo_reqwidth()  # width of widget in Pixels
                if auto_size_text or (
                        not auto_size_text and height == 1
                ):  # if just 1 line high, ensure no wrap happens
                    wraplen = 0
                tktext_label.configure(
                    anchor=anchor, wraplen=wraplen
                )  # set wrap to width of widget
                if element.Relief is not None:
                    tktext_label.configure(relief=element.Relief)
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    tktext_label.configure(background=element.BackgroundColor)
                if (
                        element.TextColor != COLOR_SYSTEM_DEFAULT
                        and element.TextColor is not None
                ):
                    tktext_label.configure(fg=element.TextColor)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                tktext_label.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tktext_label.pack_forget()
                element.TKText = tktext_label
                if element.ClickSubmits:
                    tktext_label.bind("<Button-1>", element._TextClickedHandler)
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKText,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)
                if element.Grab:
                    element._grab_anywhere_on()
            # -------------------------  BUTTON placement element non-ttk version  ------------------------- #
            elif (
                    element_type == ELEM_TYPE_BUTTON and element.UseTtkButtons is False
            ) or (
                    element_type == ELEM_TYPE_BUTTON
                    and element.UseTtkButtons is not True
                    and toplevel_form.UseTtkButtons is not True
            ):
                element = element  # type: Button
                element.UseTtkButtons = False  # indicate that ttk button was not used
                stringvar = tk.StringVar()
                element.TKStringVar = stringvar
                element.Location = (row_num, col_num)
                btext = element.ButtonText
                btype = element.BType
                if element.AutoSizeButton is not None:
                    auto_size = element.AutoSizeButton
                else:
                    auto_size = toplevel_form.AutoSizeButtons
                if auto_size is False or element.Size[0] is not None:
                    width, height = element_size
                else:
                    width = 0
                    height = toplevel_form.DefaultButtonElementSize[1]
                if (
                        element.ButtonColor != (None, None)
                        and element.ButtonColor != DEFAULT_BUTTON_COLOR
                ):
                    bc = element.ButtonColor
                elif (
                        toplevel_form.ButtonColor != (None, None)
                        and toplevel_form.ButtonColor != DEFAULT_BUTTON_COLOR
                ):
                    bc = toplevel_form.ButtonColor
                else:
                    bc = DEFAULT_BUTTON_COLOR

                bd = element.BorderWidth
                pos = -1
                if DEFAULT_USE_BUTTON_SHORTCUTS:
                    pos = btext.find(MENU_SHORTCUT_CHARACTER)
                    if pos != -1:
                        if (
                                pos < len(MENU_SHORTCUT_CHARACTER)
                                or btext[pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                        ):
                            btext = (
                                    btext[:pos]
                                    + btext[pos + len(MENU_SHORTCUT_CHARACTER):]
                            )
                        else:
                            btext = btext.replace(
                                "\\" + MENU_SHORTCUT_CHARACTER, MENU_SHORTCUT_CHARACTER
                            )
                            pos = -1
                tkbutton = element.Widget = tk.Button(
                    tk_row_frame,
                    text=btext,
                    width=width,
                    height=height,
                    justify=tk.CENTER,
                    bd=bd,
                    font=font,
                )
                if pos != -1:
                    tkbutton.config(underline=pos)
                try:
                    if btype != BUTTON_TYPE_REALTIME:
                        tkbutton.config(command=element.ButtonCallBack)

                    else:
                        tkbutton.bind(
                            "<ButtonRelease-1>", element.ButtonReleaseCallBack
                        )
                        tkbutton.bind("<ButtonPress-1>", element.ButtonPressCallBack)
                    if bc != (None, None) and COLOR_SYSTEM_DEFAULT not in bc:
                        tkbutton.config(foreground=bc[0], background=bc[1])
                    else:
                        if bc[0] != COLOR_SYSTEM_DEFAULT:
                            tkbutton.config(foreground=bc[0])
                        if bc[1] != COLOR_SYSTEM_DEFAULT:
                            tkbutton.config(background=bc[1])
                except Exception as e:
                    _error_popup_with_traceback(
                        "Button has a problem....",
                        "The traceback information will not show the line in your layout with the problem, but it does tell you which window.",
                        "Error {}".format(e),
                        # 'Button Text: {}'.format(btext),
                        # 'Button key: {}'.format(element.Key),
                        # 'Color string: {}'.format(bc),
                        "Parent Window's Title: {}".format(toplevel_form.Title),
                    )

                if bd == 0 and not running_mac():
                    tkbutton.config(relief=tk.FLAT)

                element.TKButton = (
                    tkbutton  # not used yet but save the TK button in case
                )
                if elementpad[0] == 0 or elementpad[1] == 0:
                    tkbutton.config(highlightthickness=0)

                ## -------------- TK Button With Image -------------- ##
                if element.ImageFilename:  # if button has an image on it
                    tkbutton.config(highlightthickness=0)
                    try:
                        photo = tk.PhotoImage(file=element.ImageFilename)
                        if element.ImageSubsample:
                            photo = photo.subsample(element.ImageSubsample)
                        if element.zoom:
                            photo = photo.zoom(element.zoom)
                        if element.ImageSize != (None, None):
                            width, height = element.ImageSize
                        else:
                            width, height = photo.width(), photo.height()
                    except Exception as e:
                        _error_popup_with_traceback(
                            "Button Element error {}".format(e),
                            "Image filename: {}".format(element.ImageFilename),
                            "NOTE - file format must be PNG or GIF!",
                            "Button element key: {}".format(element.Key),
                            "Parent Window's Title: {}".format(toplevel_form.Title),
                        )
                    tkbutton.config(
                        image=photo, compound=tk.CENTER, width=width, height=height
                    )
                    tkbutton.image = photo
                if element.ImageData:  # if button has an image on it
                    tkbutton.config(highlightthickness=0)
                    try:
                        photo = tk.PhotoImage(data=element.ImageData)
                        if element.ImageSubsample:
                            photo = photo.subsample(element.ImageSubsample)
                        if element.zoom:
                            photo = photo.zoom(element.zoom)
                        if element.ImageSize != (None, None):
                            width, height = element.ImageSize
                        else:
                            width, height = photo.width(), photo.height()
                        tkbutton.config(
                            image=photo, compound=tk.CENTER, width=width, height=height
                        )
                        tkbutton.image = photo
                    except Exception as e:
                        _error_popup_with_traceback(
                            "Button Element error {}".format(e),
                            "Problem using BASE64 Image data Image Susample",
                            "Buton element key: {}".format(element.Key),
                            "Parent Window's Title: {}".format(toplevel_form.Title),
                        )

                if width != 0:
                    wraplen = width * _char_width_in_pixels(font)
                    tkbutton.configure(
                        wraplength=wraplen
                    )  # set wrap to width of widget
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )

                tkbutton.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tkbutton.pack_forget()
                if element.BindReturnKey:
                    element.TKButton.bind("<Return>", element._ReturnKeyHandler)
                if element.Focus is True or (
                        toplevel_form.UseDefaultFocus and not toplevel_form.FocusSet
                ):
                    toplevel_form.FocusSet = True
                    element.TKButton.bind("<Return>", element._ReturnKeyHandler)
                    element.TKButton.focus_set()
                    toplevel_form.TKroot.focus_force()
                if element.Disabled is True:
                    element.TKButton["state"] = "disabled"
                if element.DisabledButtonColor != (
                        None,
                        None,
                ) and element.DisabledButtonColor != (
                        COLOR_SYSTEM_DEFAULT,
                        COLOR_SYSTEM_DEFAULT,
                ):
                    if element.DisabledButtonColor[0] not in (
                            None,
                            COLOR_SYSTEM_DEFAULT,
                    ):
                        element.TKButton["disabledforeground"] = (
                            element.DisabledButtonColor[0]
                        )
                if element.MouseOverColors[1] not in (COLOR_SYSTEM_DEFAULT, None):
                    tkbutton.config(activebackground=element.MouseOverColors[1])
                if element.MouseOverColors[0] not in (COLOR_SYSTEM_DEFAULT, None):
                    tkbutton.config(activeforeground=element.MouseOverColors[0])

                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKButton,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                try:
                    if element.HighlightColors[1] != COLOR_SYSTEM_DEFAULT:
                        tkbutton.config(highlightbackground=element.HighlightColors[1])
                    if element.HighlightColors[0] != COLOR_SYSTEM_DEFAULT:
                        tkbutton.config(highlightcolor=element.HighlightColors[0])
                except Exception as e:
                    _error_popup_with_traceback(
                        "Button Element error {}".format(e),
                        "Button element key: {}".format(element.Key),
                        "Button text: {}".format(btext),
                        "Has a bad highlight color {}".format(element.HighlightColors),
                        "Parent Window's Title: {}".format(toplevel_form.Title),
                    )
                    # print('Button with text: ', btext, 'has a bad highlight color', element.HighlightColors)
                _add_right_click_menu_and_grab(element)

            # -------------------------  BUTTON placement element ttk version ------------------------- #
            elif element_type == ELEM_TYPE_BUTTON:
                element = element  # type: Button
                element.UseTtkButtons = True  # indicate that ttk button was used
                stringvar = tk.StringVar()
                element.TKStringVar = stringvar
                element.Location = (row_num, col_num)
                btext = element.ButtonText
                pos = -1
                if DEFAULT_USE_BUTTON_SHORTCUTS:
                    pos = btext.find(MENU_SHORTCUT_CHARACTER)
                    if pos != -1:
                        if (
                                pos < len(MENU_SHORTCUT_CHARACTER)
                                or btext[pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                        ):
                            btext = (
                                    btext[:pos]
                                    + btext[pos + len(MENU_SHORTCUT_CHARACTER):]
                            )
                        else:
                            btext = btext.replace(
                                "\\" + MENU_SHORTCUT_CHARACTER, MENU_SHORTCUT_CHARACTER
                            )
                            pos = -1
                btype = element.BType
                if element.AutoSizeButton is not None:
                    auto_size = element.AutoSizeButton
                else:
                    auto_size = toplevel_form.AutoSizeButtons
                if auto_size is False or element.Size[0] is not None:
                    width, height = element_size
                else:
                    width = 0
                    height = toplevel_form.DefaultButtonElementSize[1]
                if (
                        element.ButtonColor != (None, None)
                        and element.ButtonColor != COLOR_SYSTEM_DEFAULT
                ):
                    bc = element.ButtonColor
                elif (
                        toplevel_form.ButtonColor != (None, None)
                        and toplevel_form.ButtonColor != COLOR_SYSTEM_DEFAULT
                ):
                    bc = toplevel_form.ButtonColor
                else:
                    bc = DEFAULT_BUTTON_COLOR
                bd = element.BorderWidth
                tkbutton = element.Widget = ttk.Button(
                    tk_row_frame, text=btext, width=width
                )
                if pos != -1:
                    tkbutton.config(underline=pos)
                if btype != BUTTON_TYPE_REALTIME:
                    tkbutton.config(command=element.ButtonCallBack)
                else:
                    tkbutton.bind("<ButtonRelease-1>", element.ButtonReleaseCallBack)
                    tkbutton.bind("<ButtonPress-1>", element.ButtonPressCallBack)
                style_name = _make_ttk_style_name(
                    ".TButton", element, primary_style=True
                )
                button_style = ttk.Style()
                element.ttk_style = button_style
                _change_ttk_theme(button_style, toplevel_form.TtkTheme)
                button_style.configure(style_name, font=font)

                if bc != (None, None) and COLOR_SYSTEM_DEFAULT not in bc:
                    button_style.configure(
                        style_name, foreground=bc[0], background=bc[1]
                    )
                elif bc[0] != COLOR_SYSTEM_DEFAULT:
                    button_style.configure(style_name, foreground=bc[0])
                elif bc[1] != COLOR_SYSTEM_DEFAULT:
                    button_style.configure(style_name, background=bc[1])

                if bd == 0 and not running_mac():
                    button_style.configure(style_name, relief=tk.FLAT)
                    button_style.configure(style_name, borderwidth=0)
                else:
                    button_style.configure(style_name, borderwidth=bd)
                button_style.configure(style_name, justify=tk.CENTER)

                if element.MouseOverColors[1] not in (COLOR_SYSTEM_DEFAULT, None):
                    button_style.map(
                        style_name, background=[("active", element.MouseOverColors[1])]
                    )
                if element.MouseOverColors[0] not in (COLOR_SYSTEM_DEFAULT, None):
                    button_style.map(
                        style_name, foreground=[("active", element.MouseOverColors[0])]
                    )

                if element.DisabledButtonColor[0] not in (COLOR_SYSTEM_DEFAULT, None):
                    button_style.map(
                        style_name,
                        foreground=[("disabled", element.DisabledButtonColor[0])],
                    )
                if element.DisabledButtonColor[1] not in (COLOR_SYSTEM_DEFAULT, None):
                    button_style.map(
                        style_name,
                        background=[("disabled", element.DisabledButtonColor[1])],
                    )

                if height > 1:
                    button_style.configure(
                        style_name, padding=height * _char_height_in_pixels(font)
                    )  # should this be height instead?
                if width != 0:
                    wraplen = width * _char_width_in_pixels(
                        font
                    )  # width of widget in Pixels
                    button_style.configure(
                        style_name, wraplength=wraplen
                    )  # set wrap to width of widget

                ## -------------- TTK Button With Image -------------- ##
                if element.ImageFilename:  # if button has an image on it
                    button_style.configure(style_name, borderwidth=0)
                    # tkbutton.configure(highlightthickness=0)
                    photo = tk.PhotoImage(file=element.ImageFilename)
                    if element.ImageSubsample:
                        photo = photo.subsample(element.ImageSubsample)
                    if element.zoom:
                        photo = photo.zoom(element.zoom)
                    if element.ImageSize != (None, None):
                        width, height = element.ImageSize
                    else:
                        width, height = photo.width(), photo.height()
                    button_style.configure(
                        style_name,
                        image=photo,
                        compound=tk.CENTER,
                        width=width,
                        height=height,
                    )
                    tkbutton.image = photo
                if element.ImageData:  # if button has an image on it
                    # tkbutton.configure(highlightthickness=0)
                    button_style.configure(style_name, borderwidth=0)

                    photo = tk.PhotoImage(data=element.ImageData)
                    if element.ImageSubsample:
                        photo = photo.subsample(element.ImageSubsample)
                    if element.zoom:
                        photo = photo.zoom(element.zoom)
                    if element.ImageSize != (None, None):
                        width, height = element.ImageSize
                    else:
                        width, height = photo.width(), photo.height()
                    button_style.configure(
                        style_name,
                        image=photo,
                        compound=tk.CENTER,
                        width=width,
                        height=height,
                    )
                    # tkbutton.configure(image=photo, compound=tk.CENTER, width=width, height=height)
                    tkbutton.image = photo

                element.TKButton = (
                    tkbutton  # not used yet but save the TK button in case
                )
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                tkbutton.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tkbutton.pack_forget()
                if element.BindReturnKey:
                    element.TKButton.bind("<Return>", element._ReturnKeyHandler)
                if element.Focus is True or (
                        toplevel_form.UseDefaultFocus and not toplevel_form.FocusSet
                ):
                    toplevel_form.FocusSet = True
                    element.TKButton.bind("<Return>", element._ReturnKeyHandler)
                    element.TKButton.focus_set()
                    toplevel_form.TKroot.focus_force()
                if element.Disabled is True:
                    element.TKButton["state"] = "disabled"

                tkbutton.configure(
                    style=style_name
                )  # IMPORTANT!  Apply the style to the button!
                _add_right_click_menu_and_grab(element)

                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKButton,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
            # -------------------------  BUTTONMENU placement element  ------------------------- #
            elif element_type == ELEM_TYPE_BUTTONMENU:
                element = element  # type: ButtonMenu
                element.Location = (row_num, col_num)
                btext = element.ButtonText
                if element.AutoSizeButton is not None:
                    auto_size = element.AutoSizeButton
                else:
                    auto_size = toplevel_form.AutoSizeButtons
                if auto_size is False or element.Size[0] is not None:
                    width, height = element_size
                else:
                    width = 0
                    height = toplevel_form.DefaultButtonElementSize[1]
                if (
                        element.ButtonColor != (None, None)
                        and element.ButtonColor != DEFAULT_BUTTON_COLOR
                ):
                    bc = element.ButtonColor
                elif (
                        toplevel_form.ButtonColor != (None, None)
                        and toplevel_form.ButtonColor != DEFAULT_BUTTON_COLOR
                ):
                    bc = toplevel_form.ButtonColor
                else:
                    bc = DEFAULT_BUTTON_COLOR
                bd = element.BorderWidth
                if element.ItemFont is None:
                    element.ItemFont = font
                tkbutton = element.Widget = tk.Menubutton(
                    tk_row_frame,
                    text=btext,
                    width=width,
                    height=height,
                    justify=tk.LEFT,
                    bd=bd,
                    font=font,
                )
                element.TKButtonMenu = tkbutton
                if (
                        bc != (None, None)
                        and bc != COLOR_SYSTEM_DEFAULT
                        and bc[1] != COLOR_SYSTEM_DEFAULT
                ):
                    tkbutton.config(foreground=bc[0], background=bc[1])
                    tkbutton.config(activebackground=bc[0])
                    tkbutton.config(activeforeground=bc[1])
                elif bc[0] != COLOR_SYSTEM_DEFAULT:
                    tkbutton.config(foreground=bc[0])
                    tkbutton.config(activebackground=bc[0])
                if bd == 0 and not running_mac():
                    tkbutton.config(relief=RELIEF_FLAT)
                elif bd != 0:
                    tkbutton.config(relief=RELIEF_RAISED)

                element.TKButton = (
                    tkbutton  # not used yet but save the TK button in case
                )
                wraplen = tkbutton.winfo_reqwidth()  # width of widget in Pixels
                if element.ImageFilename:  # if button has an image on it
                    photo = tk.PhotoImage(file=element.ImageFilename)
                    if element.ImageSubsample:
                        photo = photo.subsample(element.ImageSubsample)
                    if element.zoom:
                        photo = photo.zoom(element.zoom)
                    if element.ImageSize != (None, None):
                        width, height = element.ImageSize
                    else:
                        width, height = photo.width(), photo.height()
                    tkbutton.config(
                        image=photo, compound=tk.CENTER, width=width, height=height
                    )
                    tkbutton.image = photo
                if element.ImageData:  # if button has an image on it
                    photo = tk.PhotoImage(data=element.ImageData)
                    if element.ImageSubsample:
                        photo = photo.subsample(element.ImageSubsample)
                    if element.zoom:
                        photo = photo.zoom(element.zoom)
                    if element.ImageSize != (None, None):
                        width, height = element.ImageSize
                    else:
                        width, height = photo.width(), photo.height()
                    tkbutton.config(
                        image=photo, compound=tk.CENTER, width=width, height=height
                    )
                    tkbutton.image = photo
                if width != 0:
                    tkbutton.configure(
                        wraplength=wraplen + 10
                    )  # set wrap to width of widget
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                tkbutton.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )

                menu_def = element.MenuDefinition

                element.TKMenu = top_menu = tk.Menu(
                    tkbutton,
                    tearoff=element.Tearoff,
                    font=element.ItemFont,
                    tearoffcommand=element._tearoff_menu_callback,
                )

                if element.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                    top_menu.config(bg=element.BackgroundColor)
                    top_menu.config(activeforeground=element.BackgroundColor)
                if element.TextColor not in (COLOR_SYSTEM_DEFAULT, None):
                    top_menu.config(fg=element.TextColor)
                    top_menu.config(activebackground=element.TextColor)
                if element.DisabledTextColor not in (COLOR_SYSTEM_DEFAULT, None):
                    top_menu.config(disabledforeground=element.DisabledTextColor)
                if element.ItemFont is not None:
                    top_menu.config(font=element.ItemFont)

                AddMenuItem(top_menu, menu_def[1], element)
                if elementpad[0] == 0 or elementpad[1] == 0:
                    tkbutton.config(highlightthickness=0)
                tkbutton.configure(menu=top_menu)
                element.TKMenu = top_menu
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tkbutton.pack_forget()
                if element.Disabled:
                    element.TKButton["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKButton,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )

            # -------------------------  INPUT placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_TEXT:
                element = element  # type: InputText
                default_text = element.DefaultText
                element.TKStringVar = tk.StringVar()
                element.TKStringVar.set(default_text)
                show = element.PasswordCharacter if element.PasswordCharacter else ""
                bd = border_depth
                if element.Justification is not None:
                    justification = element.Justification
                else:
                    justification = DEFAULT_TEXT_JUSTIFICATION
                justify = (
                    tk.LEFT
                    if justification.startswith("l")
                    else tk.CENTER
                    if justification.startswith("c")
                    else tk.RIGHT
                )
                # anchor = tk.NW if justification == 'left' else tk.N if justification == 'center' else tk.NE
                element.TKEntry = element.Widget = tk.Entry(
                    tk_row_frame,
                    width=element_size[0],
                    textvariable=element.TKStringVar,
                    bd=bd,
                    font=font,
                    show=show,
                    justify=justify,
                )
                if element.ChangeSubmits:
                    element.TKEntry.bind("<Key>", element._KeyboardHandler)
                element.TKEntry.bind("<Return>", element._ReturnKeyHandler)

                if element.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKEntry.configure(
                        background=element.BackgroundColor,
                        selectforeground=element.BackgroundColor,
                    )

                if text_color not in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKEntry.configure(
                        fg=text_color, selectbackground=text_color
                    )
                    element.TKEntry.config(insertbackground=text_color)
                if element.selected_background_color not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ):
                    element.TKEntry.configure(
                        selectbackground=element.selected_background_color
                    )
                if element.selected_text_color not in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKEntry.configure(
                        selectforeground=element.selected_text_color
                    )
                if element.disabled_readonly_background_color not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ):
                    element.TKEntry.config(
                        readonlybackground=element.disabled_readonly_background_color
                    )
                if (
                        element.disabled_readonly_text_color
                        not in (None, COLOR_SYSTEM_DEFAULT)
                        and element.Disabled
                ):
                    element.TKEntry.config(fg=element.disabled_readonly_text_color)

                element.Widget.config(highlightthickness=0)
                # element.pack_keywords = {'side':tk.LEFT, 'padx':elementpad[0], 'pady':elementpad[1], 'expand':False, 'fill':tk.NONE }
                # element.TKEntry.pack(**element.pack_keywords)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKEntry.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKEntry.pack_forget()
                if element.Focus is True or (
                        toplevel_form.UseDefaultFocus and not toplevel_form.FocusSet
                ):
                    toplevel_form.FocusSet = True
                    element.TKEntry.focus_set()
                if element.Disabled:
                    element.TKEntry["state"] = (
                        "readonly" if element.UseReadonlyForDisable else "disabled"
                    )
                if element.ReadOnly:
                    element.TKEntry["state"] = "readonly"

                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKEntry,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

                # row_should_expand = True

            # -------------------------  COMBO placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_COMBO:
                element = element  # type: Combo
                max_line_len = (
                    max([len(str(l)) for l in element.Values])
                    if len(element.Values)
                    else 0
                )
                if not auto_size_text:
                    width = element_size[0]
                else:
                    width = max_line_len + 1
                element.TKStringVar = tk.StringVar()
                style_name = _make_ttk_style_name(
                    ".TCombobox", element, primary_style=True
                )
                combostyle = ttk.Style()
                element.ttk_style = combostyle
                _change_ttk_theme(combostyle, toplevel_form.TtkTheme)

                # Creates a unique name for each field element(Sure there is a better way to do this)
                # unique_field = _make_ttk_style_name('.TCombobox.field', element)

                # Set individual widget options
                try:
                    if element.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                        combostyle.configure(style_name, foreground=element.TextColor)
                        combostyle.configure(
                            style_name, selectbackground=element.TextColor
                        )
                        combostyle.configure(style_name, insertcolor=element.TextColor)
                        combostyle.map(
                            style_name,
                            fieldforeground=[("readonly", element.TextColor)],
                        )
                    if element.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT):
                        combostyle.configure(
                            style_name, selectforeground=element.BackgroundColor
                        )
                        combostyle.map(
                            style_name,
                            fieldbackground=[("readonly", element.BackgroundColor)],
                        )
                        combostyle.configure(
                            style_name, fieldbackground=element.BackgroundColor
                        )

                    if element.button_arrow_color not in (None, COLOR_SYSTEM_DEFAULT):
                        combostyle.configure(
                            style_name, arrowcolor=element.button_arrow_color
                        )
                    if element.button_background_color not in (
                            None,
                            COLOR_SYSTEM_DEFAULT,
                    ):
                        combostyle.configure(
                            style_name, background=element.button_background_color
                        )
                    if element.Readonly:
                        if element.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                            combostyle.configure(
                                style_name, selectforeground=element.TextColor
                            )
                        if element.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT):
                            combostyle.configure(
                                style_name, selectbackground=element.BackgroundColor
                            )

                except Exception as e:
                    _error_popup_with_traceback(
                        "Combo Element error {}".format(e),
                        "Combo element key: {}".format(element.Key),
                        "One of your colors is bad. Check the text, background, button background and button arrow colors",
                        "Parent Window's Title: {}".format(toplevel_form.Title),
                    )

                # Strange code that is needed to set the font for the drop-down list
                element._dropdown_newfont = tkinter.font.Font(font=font)
                tk_row_frame.option_add(
                    "*TCombobox*Listbox*Font", element._dropdown_newfont
                )

                element.TKCombo = element.Widget = ttk.Combobox(
                    tk_row_frame,
                    width=width,
                    textvariable=element.TKStringVar,
                    font=font,
                    style=style_name,
                )

                # make tcl call to deal with colors for the drop-down formatting
                try:
                    if element.BackgroundColor not in (
                            None,
                            COLOR_SYSTEM_DEFAULT,
                    ) and element.TextColor not in (None, COLOR_SYSTEM_DEFAULT):
                        element.Widget.tk.eval(
                            "[ttk::combobox::PopdownWindow {}].f.l configure -foreground {} -background {} -selectforeground {} -selectbackground {}".format(
                                element.Widget,
                                element.TextColor,
                                element.BackgroundColor,
                                element.BackgroundColor,
                                element.TextColor,
                            )
                        )
                except Exception as e:
                    pass  # going to let this one slide

                # Chr0nic
                element.TKCombo.bind(
                    "<Enter>", lambda event, em=element: testMouseHook2(em)
                )
                element.TKCombo.bind(
                    "<Leave>", lambda event, em=element: testMouseUnhook2(em)
                )

                if toplevel_form.UseDefaultFocus and not toplevel_form.FocusSet:
                    toplevel_form.FocusSet = True
                    element.TKCombo.focus_set()

                if element.Size[1] != 1 and element.Size[1] is not None:
                    element.TKCombo.configure(height=element.Size[1])
                element.TKCombo["values"] = element.Values
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKCombo.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKCombo.pack_forget()
                if element.DefaultValue is not None:
                    element.TKCombo.set(element.DefaultValue)
                    # for i, v in enumerate(element.Values):
                    #     if v == element.DefaultValue:
                    #         element.TKCombo.current(i)
                    #         break
                # elif element.Values:
                #     element.TKCombo.current(0)
                if element.ChangeSubmits:
                    element.TKCombo.bind(
                        "<<ComboboxSelected>>", element._ComboboxSelectHandler
                    )
                if element.BindReturnKey:
                    element.TKCombo.bind("<Return>", element._ComboboxSelectHandler)
                if element.enable_per_char_events:
                    element.TKCombo.bind("<Key>", element._KeyboardHandler)
                if element.Readonly:
                    element.TKCombo["state"] = "readonly"
                if element.Disabled:  # note overrides readonly if disabled
                    element.TKCombo["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKCombo,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

            # -------------------------  OPTIONMENU placement Element (Like ComboBox but different) element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_OPTION_MENU:
                max_line_len = max([len(str(l)) for l in element.Values])
                if not auto_size_text:
                    width = element_size[0]
                else:
                    width = max_line_len
                element.TKStringVar = tk.StringVar()
                if element.DefaultValue:
                    element.TKStringVar.set(element.DefaultValue)
                element.TKOptionMenu = element.Widget = tk.OptionMenu(
                    tk_row_frame, element.TKStringVar, *element.Values
                )
                element.TKOptionMenu.config(
                    highlightthickness=0, font=font, width=width
                )
                element.TKOptionMenu["menu"].config(font=font)
                element.TKOptionMenu.config(borderwidth=border_depth)
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKOptionMenu.configure(background=element.BackgroundColor)
                    element.TKOptionMenu["menu"].config(
                        background=element.BackgroundColor
                    )
                if (
                        element.TextColor != COLOR_SYSTEM_DEFAULT
                        and element.TextColor is not None
                ):
                    element.TKOptionMenu.configure(fg=element.TextColor)
                    element.TKOptionMenu["menu"].config(fg=element.TextColor)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKOptionMenu.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKOptionMenu.pack_forget()
                if element.Disabled:
                    element.TKOptionMenu["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKOptionMenu,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
            # -------------------------  LISTBOX placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_LISTBOX:
                element = element  # type: Listbox
                max_line_len = (
                    max([len(str(l)) for l in element.Values])
                    if len(element.Values)
                    else 0
                )
                if not auto_size_text:
                    width = element_size[0]
                else:
                    width = max_line_len
                element_frame = tk.Frame(tk_row_frame)
                element.element_frame = element_frame

                justification = tk.LEFT
                if element.justification is not None:
                    if element.justification.startswith("l"):
                        justification = tk.LEFT
                    elif element.justification.startswith("r"):
                        justification = tk.RIGHT
                    elif element.justification.startswith("c"):
                        justification = tk.CENTER

                element.TKStringVar = tk.StringVar()
                element.TKListbox = element.Widget = tk.Listbox(
                    element_frame,
                    height=element_size[1],
                    width=width,
                    selectmode=element.SelectMode,
                    font=font,
                    exportselection=False,
                )
                # On OLD versions of tkinter the justify option isn't available
                try:
                    element.Widget.config(justify=justification)
                except Exception:
                    pass

                element.Widget.config(highlightthickness=0)
                for index, item in enumerate(element.Values):
                    element.TKListbox.insert(tk.END, item)
                    if (
                            element.DefaultValues is not None
                            and item in element.DefaultValues
                    ):
                        element.TKListbox.selection_set(index)
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKListbox.configure(background=element.BackgroundColor)
                if (
                        element.HighlightBackgroundColor is not None
                        and element.HighlightBackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKListbox.config(
                        selectbackground=element.HighlightBackgroundColor
                    )
                if text_color is not None and text_color != COLOR_SYSTEM_DEFAULT:
                    element.TKListbox.configure(fg=text_color)
                if (
                        element.HighlightTextColor is not None
                        and element.HighlightTextColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKListbox.config(
                        selectforeground=element.HighlightTextColor
                    )
                if element.ChangeSubmits:
                    element.TKListbox.bind(
                        "<<ListboxSelect>>", element._ListboxSelectHandler
                    )

                if not element.NoScrollbar:
                    _make_ttk_scrollbar(element, "v", toplevel_form)
                    element.Widget.configure(yscrollcommand=element.vsb.set)
                    element.vsb.pack(side=tk.RIGHT, fill="y")

                # Horizontal scrollbar
                if element.HorizontalScroll:
                    _make_ttk_scrollbar(element, "h", toplevel_form)
                    element.hsb.pack(side=tk.BOTTOM, fill="x")
                    element.Widget.configure(xscrollcommand=element.hsb.set)

                if not element.NoScrollbar or element.HorizontalScroll:
                    # Chr0nic
                    element.Widget.bind(
                        "<Enter>", lambda event, em=element: testMouseHook(em)
                    )
                    element.Widget.bind(
                        "<Leave>", lambda event, em=element: testMouseUnhook(em)
                    )

                # else:
                #     element.TKText.config(wrap='word')

                # if not element.NoScrollbar:
                #     # Vertical scrollbar
                #     element.vsb = tk.Scrollbar(element_frame, orient="vertical", command=element.TKListbox.yview)
                #     element.TKListbox.configure(yscrollcommand=element.vsb.set)
                #     element.vsb.pack(side=tk.RIGHT, fill='y')

                # Horizontal scrollbar
                # if element.HorizontalScroll:
                #     hscrollbar = tk.Scrollbar(element_frame, orient=tk.HORIZONTAL)
                #     hscrollbar.pack(side=tk.BOTTOM, fill='x')
                #     hscrollbar.config(command=element.Widget.xview)
                #     element.Widget.configure(xscrollcommand=hscrollbar.set)
                #     element.hsb = hscrollbar
                #
                #     # Chr0nic
                #     element.TKListbox.bind("<Enter>", lambda event, em=element: testMouseHook(em))
                #     element.TKListbox.bind("<Leave>", lambda event, em=element: testMouseUnhook(em))
                #
                #

                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element_frame.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    fill=fill,
                    expand=expand,
                )
                element.TKListbox.pack(side=tk.LEFT, fill=fill, expand=expand)
                if not element.visible:
                    element._pack_forget_save_settings(alternate_widget=element_frame)
                    # element_frame.pack_forget()
                if element.BindReturnKey:
                    element.TKListbox.bind("<Return>", element._ListboxSelectHandler)
                    element.TKListbox.bind(
                        "<Double-Button-1>", element._ListboxSelectHandler
                    )
                if element.Disabled:
                    element.TKListbox["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKListbox,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)
            # -------------------------  MULTILINE placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_MULTILINE:
                element = element  # type: Multiline
                width, height = element_size
                bd = element.BorderWidth
                element.element_frame = element_frame = tk.Frame(tk_row_frame)

                # if element.no_scrollbar:
                element.TKText = element.Widget = tk.Text(
                    element_frame,
                    width=width,
                    height=height,
                    bd=bd,
                    font=font,
                    relief=RELIEF_SUNKEN,
                )
                # else:
                #     element.TKText = element.Widget = tk.scrolledtext.ScrolledText(element_frame, width=width, height=height, bd=bd, font=font, relief=RELIEF_SUNKEN)

                if not element.no_scrollbar:
                    _make_ttk_scrollbar(element, "v", toplevel_form)

                    element.Widget.configure(yscrollcommand=element.vsb.set)
                    element.vsb.pack(side=tk.RIGHT, fill="y")

                # Horizontal scrollbar
                if element.HorizontalScroll:
                    element.TKText.config(wrap="none")
                    _make_ttk_scrollbar(element, "h", toplevel_form)
                    element.hsb.pack(side=tk.BOTTOM, fill="x")
                    element.Widget.configure(xscrollcommand=element.hsb.set)
                else:
                    element.TKText.config(wrap="word")

                if element.wrap_lines:
                    element.TKText.config(wrap="word")
                elif not element.wrap_lines:
                    element.TKText.config(wrap="none")

                if not element.no_scrollbar or element.HorizontalScroll:
                    # Chr0nic
                    element.TKText.bind(
                        "<Enter>", lambda event, em=element: testMouseHook(em)
                    )
                    element.TKText.bind(
                        "<Leave>", lambda event, em=element: testMouseUnhook(em)
                    )

                if element.DefaultText:
                    element.TKText.insert(
                        1.0, element.DefaultText
                    )  # set the default text
                element.TKText.config(highlightthickness=0)
                if text_color is not None and text_color != COLOR_SYSTEM_DEFAULT:
                    element.TKText.configure(fg=text_color, selectbackground=text_color)
                    element.TKText.config(insertbackground=text_color)
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKText.configure(
                        background=element.BackgroundColor,
                        selectforeground=element.BackgroundColor,
                    )
                if element.selected_background_color not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ):
                    element.TKText.configure(
                        selectbackground=element.selected_background_color
                    )
                if element.selected_text_color not in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKText.configure(
                        selectforeground=element.selected_text_color
                    )
                element.TKText.tag_configure("center", justify="center")
                element.TKText.tag_configure("left", justify="left")
                element.TKText.tag_configure("right", justify="right")

                if element.Justification.startswith("l"):
                    element.TKText.tag_add("left", 1.0, "end")
                    element.justification_tag = "left"
                elif element.Justification.startswith("r"):
                    element.TKText.tag_add("right", 1.0, "end")
                    element.justification_tag = "right"
                elif element.Justification.startswith("c"):
                    element.TKText.tag_add("center", 1.0, "end")
                    element.justification_tag = "center"
                # if DEFAULT_SCROLLBAR_COLOR not in (None, COLOR_SYSTEM_DEFAULT):               # only works on Linux so not including it
                #     element.TKText.vbar.config(troughcolor=DEFAULT_SCROLLBAR_COLOR)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )

                element.element_frame.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    fill=fill,
                    expand=expand,
                )
                element.Widget.pack(side=tk.LEFT, fill=fill, expand=expand)

                if not element.visible:
                    element._pack_forget_save_settings(alternate_widget=element_frame)
                    # element.element_frame.pack_forget()
                else:
                    # Chr0nic
                    element.TKText.bind(
                        "<Enter>", lambda event, em=element: testMouseHook(em)
                    )
                    element.TKText.bind(
                        "<Leave>", lambda event, em=element: testMouseUnhook(em)
                    )
                if element.ChangeSubmits:
                    element.TKText.bind("<Key>", element._KeyboardHandler)
                if element.EnterSubmits:
                    element.TKText.bind("<Return>", element._ReturnKeyHandler)
                if element.Focus is True or (
                        toplevel_form.UseDefaultFocus and not toplevel_form.FocusSet
                ):
                    toplevel_form.FocusSet = True
                    element.TKText.focus_set()

                if element.Disabled:
                    element.TKText["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKText,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )

                if element.reroute_cprint:
                    cprint_set_output_destination(toplevel_form, element.Key)

                _add_right_click_menu_and_grab(element)

                if element.reroute_stdout:
                    element.reroute_stdout_to_here()
                if element.reroute_stderr:
                    element.reroute_stderr_to_here()

                # row_should_expand = True
            # -------------------------  CHECKBOX pleacement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_CHECKBOX:
                element = element  # type: Checkbox
                width = 0 if auto_size_text else element_size[0]
                default_value = element.InitialState
                element.TKIntVar = tk.IntVar()
                element.TKIntVar.set(default_value if default_value is not None else 0)

                element.TKCheckbutton = element.Widget = tk.Checkbutton(
                    tk_row_frame,
                    anchor=tk.NW,
                    text=element.Text,
                    width=width,
                    variable=element.TKIntVar,
                    bd=border_depth,
                    font=font,
                )
                if element.ChangeSubmits:
                    element.TKCheckbutton.configure(command=element._CheckboxHandler)
                if element.Disabled:
                    element.TKCheckbutton.configure(state="disable")
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKCheckbutton.configure(background=element.BackgroundColor)
                    element.TKCheckbutton.configure(
                        selectcolor=element.CheckboxBackgroundColor
                    )  # The background of the checkbox
                    element.TKCheckbutton.configure(
                        activebackground=element.BackgroundColor
                    )
                if text_color is not None and text_color != COLOR_SYSTEM_DEFAULT:
                    element.TKCheckbutton.configure(fg=text_color)
                    element.TKCheckbutton.configure(activeforeground=element.TextColor)

                element.Widget.configure(highlightthickness=element.highlight_thickness)
                if element.BackgroundColor != COLOR_SYSTEM_DEFAULT:
                    element.TKCheckbutton.config(
                        highlightbackground=element.BackgroundColor
                    )
                if element.TextColor != COLOR_SYSTEM_DEFAULT:
                    element.TKCheckbutton.config(highlightcolor=element.TextColor)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKCheckbutton.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKCheckbutton.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKCheckbutton,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

            # -------------------------  PROGRESS placement element  ------------------------- #
            elif element_type == ELEM_TYPE_PROGRESS_BAR:
                element = element  # type: ProgressBar
                if element.size_px != (None, None):
                    progress_length, progress_width = element.size_px
                else:
                    width = element_size[0]
                    fnt = tkinter.font.Font()
                    char_width = fnt.measure("A")  # single character width
                    progress_length = width * char_width
                    progress_width = element_size[1]
                direction = element.Orientation
                if element.BarColor != (
                        None,
                        None,
                ):  # if element has a bar color, use it
                    bar_color = element.BarColor
                else:
                    bar_color = DEFAULT_PROGRESS_BAR_COLOR
                if element.Orientation.lower().startswith("h"):
                    base_style_name = ".Horizontal.TProgressbar"
                else:
                    base_style_name = ".Vertical.TProgressbar"
                style_name = _make_ttk_style_name(
                    base_style_name, element, primary_style=True
                )
                element.TKProgressBar = TKProgressBar(
                    tk_row_frame,
                    element.MaxValue,
                    progress_length,
                    progress_width,
                    orientation=direction,
                    BarColor=bar_color,
                    border_width=element.BorderWidth,
                    relief=element.Relief,
                    ttk_theme=toplevel_form.TtkTheme,
                    key=element.Key,
                    style_name=style_name,
                )
                element.Widget = element.TKProgressBar.TKProgressBarForReal
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKProgressBar.TKProgressBarForReal.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings(
                        alternate_widget=element.TKProgressBar.TKProgressBarForReal
                    )
                    # element.TKProgressBar.TKProgressBarForReal.pack_forget()
                _add_right_click_menu_and_grab(element)

                # -------------------------  RADIO placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_RADIO:
                element = element  # type: Radio
                width = 0 if auto_size_text else element_size[0]
                default_value = element.InitialState
                ID = element.GroupID
                # see if ID has already been placed
                value = EncodeRadioRowCol(
                    form.ContainerElemementNumber, row_num, col_num
                )  # value to set intvar to if this radio is selected
                element.EncodedRadioValue = value
                if ID in toplevel_form.RadioDict:
                    RadVar = toplevel_form.RadioDict[ID]
                else:
                    RadVar = tk.IntVar()
                    toplevel_form.RadioDict[ID] = RadVar
                element.TKIntVar = RadVar  # store the RadVar in Radio object
                if (
                        default_value
                ):  # if this radio is the one selected, set RadVar to match
                    element.TKIntVar.set(value)
                element.TKRadio = element.Widget = tk.Radiobutton(
                    tk_row_frame,
                    anchor=tk.NW,
                    text=element.Text,
                    width=width,
                    variable=element.TKIntVar,
                    value=value,
                    bd=border_depth,
                    font=font,
                )
                if element.ChangeSubmits:
                    element.TKRadio.configure(command=element._RadioHandler)
                if not element.BackgroundColor in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKRadio.configure(background=element.BackgroundColor)
                    element.TKRadio.configure(selectcolor=element.CircleBackgroundColor)
                    element.TKRadio.configure(activebackground=element.BackgroundColor)
                if text_color is not None and text_color != COLOR_SYSTEM_DEFAULT:
                    element.TKRadio.configure(fg=text_color)
                    element.TKRadio.configure(activeforeground=text_color)

                element.Widget.configure(highlightthickness=1)
                if element.BackgroundColor != COLOR_SYSTEM_DEFAULT:
                    element.TKRadio.config(highlightbackground=element.BackgroundColor)
                if element.TextColor != COLOR_SYSTEM_DEFAULT:
                    element.TKRadio.config(highlightcolor=element.TextColor)

                if element.Disabled:
                    element.TKRadio["state"] = "disabled"
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKRadio.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKRadio.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKRadio,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

                # -------------------------  SPIN placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_SPIN:
                element = element  # type: Spin
                width, height = element_size
                width = 0 if auto_size_text else element_size[0]
                element.TKStringVar = tk.StringVar()
                element.TKSpinBox = element.Widget = tk.Spinbox(
                    tk_row_frame,
                    values=element.Values,
                    textvariable=element.TKStringVar,
                    width=width,
                    bd=border_depth,
                )
                if element.DefaultValue is not None:
                    element.TKStringVar.set(element.DefaultValue)
                element.TKSpinBox.configure(font=font)  # set wrap to width of widget
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element.TKSpinBox.configure(background=element.BackgroundColor)
                    element.TKSpinBox.configure(
                        buttonbackground=element.BackgroundColor
                    )
                if text_color not in (None, COLOR_SYSTEM_DEFAULT):
                    element.TKSpinBox.configure(fg=text_color)
                    element.TKSpinBox.config(insertbackground=text_color)
                element.Widget.config(highlightthickness=0)
                if element.wrap:
                    element.Widget.configure(wrap=True)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKSpinBox.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.TKSpinBox.pack_forget()
                if element.ChangeSubmits:
                    element.TKSpinBox.configure(command=element._SpinboxSelectHandler)
                    # element.TKSpinBox.bind('<ButtonRelease-1>', element._SpinChangedHandler)
                    # element.TKSpinBox.bind('<Up>', element._SpinChangedHandler)
                    # element.TKSpinBox.bind('<Down>', element._SpinChangedHandler)
                if element.Readonly:
                    element.TKSpinBox["state"] = "readonly"
                if element.Disabled:  # note overrides readonly if disabled
                    element.TKSpinBox["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKSpinBox,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                if element.BindReturnKey:
                    element.TKSpinBox.bind("<Return>", element._SpinboxSelectHandler)
                _add_right_click_menu_and_grab(element)
                # -------------------------  IMAGE placement element  ------------------------- #
            elif element_type == ELEM_TYPE_IMAGE:
                element = element  # type: Image
                try:
                    if element.Filename is not None:
                        photo = tk.PhotoImage(file=element.Filename)
                    elif element.Data is not None:
                        photo = tk.PhotoImage(data=element.Data)
                    else:
                        photo = None

                    if photo is not None:
                        if element.ImageSubsample:
                            photo = photo.subsample(element.ImageSubsample)
                        if element.zoom:
                            photo = photo.zoom(element.zoom)
                        # print('*ERROR laying out form.... Image Element has no image specified*')
                except Exception as e:
                    photo = None
                    _error_popup_with_traceback(
                        "Your Window has an Image Element with a problem",
                        "The traceback will show you the Window with the problem layout",
                        "Look in this Window's layout for an Image element that has a key of {}".format(
                            element.Key
                        ),
                        "The error occuring is:",
                        e,
                    )

                element.tktext_label = element.Widget = tk.Label(tk_row_frame, bd=0)

                if photo is not None:
                    if (
                            element_size == (None, None)
                            or element_size is None
                            or element_size == toplevel_form.DefaultElementSize
                    ):
                        width, height = photo.width(), photo.height()
                    else:
                        width, height = element_size
                    element.tktext_label.config(image=photo, width=width, height=height)

                if not element.BackgroundColor in (None, COLOR_SYSTEM_DEFAULT):
                    element.tktext_label.config(background=element.BackgroundColor)

                element.tktext_label.image = photo
                # tktext_label.configure(anchor=tk.NW, image=photo)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.tktext_label.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )

                if not element.visible:
                    element._pack_forget_save_settings()
                    # element.tktext_label.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.tktext_label,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                if element.EnableEvents and element.tktext_label is not None:
                    element.tktext_label.bind("<ButtonPress-1>", element._ClickHandler)

                _add_right_click_menu_and_grab(element)

                # -------------------------  Canvas placement element  ------------------------- #
            elif element_type == ELEM_TYPE_CANVAS:
                element = element  # type: Canvas
                width, height = element_size
                if element._TKCanvas is None:
                    element._TKCanvas = tk.Canvas(
                        tk_row_frame, width=width, height=height, bd=border_depth
                    )
                else:
                    element._TKCanvas.master = tk_row_frame
                element.Widget = element._TKCanvas

                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element._TKCanvas.configure(
                        background=element.BackgroundColor, highlightthickness=0
                    )
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element._TKCanvas.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element._TKCanvas.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element._TKCanvas,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

                # -------------------------  Graph placement element  ------------------------- #
            elif element_type == ELEM_TYPE_GRAPH:
                element = element  # type: Graph
                width, height = element_size
                # I don't know why TWO canvases were being defined, on inside the other.  Was it so entire canvas can move?
                # if element._TKCanvas is None:
                #     element._TKCanvas = tk.Canvas(tk_row_frame, width=width, height=height, bd=border_depth)
                # else:
                #     element._TKCanvas.master = tk_row_frame
                element._TKCanvas2 = element.Widget = tk.Canvas(
                    tk_row_frame, width=width, height=height, bd=border_depth
                )
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element._TKCanvas2.pack(side=tk.LEFT, expand=expand, fill=fill)
                element._TKCanvas2.addtag_all("mytag")
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    element._TKCanvas2.configure(
                        background=element.BackgroundColor, highlightthickness=0
                    )
                    # element._TKCanvas.configure(background=element.BackgroundColor, highlightthickness=0)
                element._TKCanvas2.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # element._TKCanvas2.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element._TKCanvas2,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                if element.ChangeSubmits:
                    element._TKCanvas2.bind(
                        "<ButtonRelease-1>", element.ButtonReleaseCallBack
                    )
                    element._TKCanvas2.bind(
                        "<ButtonPress-1>", element.ButtonPressCallBack
                    )
                if element.DragSubmits:
                    element._TKCanvas2.bind("<Motion>", element.MotionCallBack)
                _add_right_click_menu_and_grab(element)
            # -------------------------  MENU placement element  ------------------------- #
            elif element_type == ELEM_TYPE_MENUBAR:
                element = element  # type: MenuBar
                menu_def = element.MenuDefinition
                element.TKMenu = element.Widget = tk.Menu(
                    toplevel_form.TKroot,
                    tearoff=element.Tearoff,
                    tearoffcommand=element._tearoff_menu_callback,
                )  # create the menubar
                menubar = element.TKMenu
                if (
                        font is not None
                ):  # if a font is used, make sure it's saved in the element
                    element.Font = font
                for menu_entry in menu_def:
                    baritem = tk.Menu(
                        menubar,
                        tearoff=element.Tearoff,
                        tearoffcommand=element._tearoff_menu_callback,
                    )
                    if element.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                        baritem.config(bg=element.BackgroundColor)
                        baritem.config(activeforeground=element.BackgroundColor)
                    if element.TextColor not in (COLOR_SYSTEM_DEFAULT, None):
                        baritem.config(fg=element.TextColor)
                        baritem.config(activebackground=element.TextColor)
                    if element.DisabledTextColor not in (COLOR_SYSTEM_DEFAULT, None):
                        baritem.config(disabledforeground=element.DisabledTextColor)
                    if font is not None:
                        baritem.config(font=font)
                    pos = menu_entry[0].find(MENU_SHORTCUT_CHARACTER)
                    # print(pos)
                    if pos != -1:
                        if (
                                pos == 0
                                or menu_entry[0][pos - len(MENU_SHORTCUT_CHARACTER)] != "\\"
                        ):
                            menu_entry[0] = (
                                    menu_entry[0][:pos] + menu_entry[0][pos + 1:]
                            )
                    if menu_entry[0][0] == MENU_DISABLED_CHARACTER:
                        menubar.add_cascade(
                            label=menu_entry[0][len(MENU_DISABLED_CHARACTER):],
                            menu=baritem,
                            underline=pos - 1,
                        )
                        menubar.entryconfig(
                            menu_entry[0][len(MENU_DISABLED_CHARACTER):],
                            state="disabled",
                        )
                    else:
                        menubar.add_cascade(
                            label=menu_entry[0], menu=baritem, underline=pos
                        )

                    if len(menu_entry) > 1:
                        AddMenuItem(baritem, menu_entry[1], element)
                toplevel_form.TKroot.configure(menu=element.TKMenu)
            # -------------------------  Frame placement element  ------------------------- #
            elif element_type == ELEM_TYPE_FRAME:
                element = element  # type: Frame
                labeled_frame = element.Widget = tk.LabelFrame(
                    tk_row_frame, text=element.Title, relief=element.Relief
                )
                element.TKFrame = labeled_frame
                PackFormIntoFrame(element, labeled_frame, toplevel_form)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                if element.VerticalAlignment is not None:
                    anchor = tk.CENTER  # Default to center if a bad choice is made
                    if element.VerticalAlignment.lower().startswith("t"):
                        anchor = tk.N
                    if element.VerticalAlignment.lower().startswith("c"):
                        anchor = tk.CENTER
                    if element.VerticalAlignment.lower().startswith("b"):
                        anchor = tk.S
                    labeled_frame.pack(
                        side=tk.LEFT,
                        anchor=anchor,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        expand=expand,
                        fill=fill,
                    )
                else:
                    labeled_frame.pack(
                        side=tk.LEFT,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        expand=expand,
                        fill=fill,
                    )

                if element.Size != (None, None):
                    labeled_frame.config(width=element.Size[0], height=element.Size[1])
                    labeled_frame.pack_propagate(0)
                if not element.visible:
                    element._pack_forget_save_settings()
                    # labeled_frame.pack_forget()
                if (
                        element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                        and element.BackgroundColor is not None
                ):
                    labeled_frame.configure(
                        background=element.BackgroundColor,
                        highlightbackground=element.BackgroundColor,
                        highlightcolor=element.BackgroundColor,
                    )
                if (
                        element.TextColor != COLOR_SYSTEM_DEFAULT
                        and element.TextColor is not None
                ):
                    labeled_frame.configure(foreground=element.TextColor)
                if font is not None:
                    labeled_frame.configure(font=font)
                if element.TitleLocation is not None:
                    labeled_frame.configure(labelanchor=element.TitleLocation)
                if element.BorderWidth is not None:
                    labeled_frame.configure(borderwidth=element.BorderWidth)
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        labeled_frame,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)
                # row_should_expand=True
            # -------------------------  Tab placement element  ------------------------- #
            elif element_type == ELEM_TYPE_TAB:
                element = element  # type: Tab
                form = form  # type: TabGroup
                element.TKFrame = element.Widget = tk.Frame(form.TKNotebook)
                PackFormIntoFrame(element, element.TKFrame, toplevel_form)
                state = "normal"
                if element.Disabled:
                    state = "disabled"
                if not element.visible:
                    state = "hidden"
                # this code will add an image to the tab. Use it when adding the image on a tab enhancement
                try:
                    if element.Filename is not None:
                        photo = tk.PhotoImage(file=element.Filename)
                    elif element.Data is not None:
                        photo = tk.PhotoImage(data=element.Data)
                    else:
                        photo = None

                    if element.ImageSubsample and photo is not None:
                        photo = photo.subsample(element.ImageSubsample)
                    if element.zoom and photo is not None:
                        photo = photo.zoom(element.zoom)
                        # print('*ERROR laying out form.... Image Element has no image specified*')
                except Exception as e:
                    photo = None
                    _error_popup_with_traceback(
                        "Your Window has an Tab Element with an IMAGE problem",
                        "The traceback will show you the Window with the problem layout",
                        "Look in this Window's layout for an Image element that has a key of {}".format(
                            element.Key
                        ),
                        "The error occuring is:",
                        e,
                    )

                element.photo = photo
                if photo is not None:
                    if (
                            element_size == (None, None)
                            or element_size is None
                            or element_size == toplevel_form.DefaultElementSize
                    ):
                        width, height = photo.width(), photo.height()
                    else:
                        width, height = element_size
                    element.tktext_label = tk.Label(
                        tk_row_frame, image=photo, width=width, height=height, bd=0
                    )
                else:
                    element.tktext_label = tk.Label(tk_row_frame, bd=0)
                if photo is not None:
                    form.TKNotebook.add(
                        element.TKFrame,
                        text=element.Title,
                        compound=tk.LEFT,
                        state=state,
                        image=photo,
                    )

                # element.photo_image = tk.PhotoImage(data=DEFAULT_BASE64_ICON)
                # form.TKNotebook.add(element.TKFrame, text=element.Title, compound=tk.LEFT, state=state,image = element.photo_image)

                form.TKNotebook.add(element.TKFrame, text=element.Title, state=state)
                # July 28 2022 removing the expansion and pack as a test
                # expand, fill, row_should_expand, row_fill_direction = _add_expansion(element, row_should_expand, row_fill_direction)
                # form.TKNotebook.pack(side=tk.LEFT, padx=elementpad[0], pady=elementpad[1], fill=fill, expand=expand)

                element.ParentNotebook = form.TKNotebook
                element.TabID = form.TabCount
                form.tab_index_to_key[element.TabID] = (
                    element.key
                )  # has a list of the tabs in the notebook and their associated key
                form.TabCount += 1
                if element.BackgroundColor not in (COLOR_SYSTEM_DEFAULT, None):
                    element.TKFrame.configure(
                        background=element.BackgroundColor,
                        highlightbackground=element.BackgroundColor,
                        highlightcolor=element.BackgroundColor,
                    )

                # if element.BorderWidth is not None:
                #     element.TKFrame.configure(borderwidth=element.BorderWidth)
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKFrame,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)
                # row_should_expand = True
            # -------------------------  TabGroup placement element  ------------------------- #
            elif element_type == ELEM_TYPE_TAB_GROUP:
                element = element  # type: TabGroup
                # custom_style = str(element.Key) + 'customtab.TNotebook'
                custom_style = _make_ttk_style_name(
                    ".TNotebook", element, primary_style=True
                )
                style = ttk.Style()
                _change_ttk_theme(style, toplevel_form.TtkTheme)

                if element.TabLocation is not None:
                    position_dict = {
                        "left": "w",
                        "right": "e",
                        "top": "n",
                        "bottom": "s",
                        "lefttop": "wn",
                        "leftbottom": "ws",
                        "righttop": "en",
                        "rightbottom": "es",
                        "bottomleft": "sw",
                        "bottomright": "se",
                        "topleft": "nw",
                        "topright": "ne",
                    }
                    try:
                        tab_position = position_dict[element.TabLocation]
                    except Exception:
                        tab_position = position_dict["top"]
                    style.configure(custom_style, tabposition=tab_position)

                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    style.configure(custom_style, background=element.BackgroundColor)

                # FINALLY the proper styling to get tab colors!
                if (
                        element.SelectedTitleColor is not None
                        and element.SelectedTitleColor != COLOR_SYSTEM_DEFAULT
                ):
                    style.map(
                        custom_style + ".Tab",
                        foreground=[("selected", element.SelectedTitleColor)],
                    )
                if (
                        element.SelectedBackgroundColor is not None
                        and element.SelectedBackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    style.map(
                        custom_style + ".Tab",
                        background=[("selected", element.SelectedBackgroundColor)],
                    )
                if (
                        element.TabBackgroundColor is not None
                        and element.TabBackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    style.configure(
                        custom_style + ".Tab", background=element.TabBackgroundColor
                    )
                if (
                        element.TextColor is not None
                        and element.TextColor != COLOR_SYSTEM_DEFAULT
                ):
                    style.configure(custom_style + ".Tab", foreground=element.TextColor)
                if element.BorderWidth is not None:
                    style.configure(custom_style, borderwidth=element.BorderWidth)
                if element.TabBorderWidth is not None:
                    style.configure(
                        custom_style + ".Tab", borderwidth=element.TabBorderWidth
                    )  # if ever want to get rid of border around the TABS themselves
                if element.FocusColor not in (None, COLOR_SYSTEM_DEFAULT):
                    style.configure(
                        custom_style + ".Tab", focuscolor=element.FocusColor
                    )

                style.configure(custom_style + ".Tab", font=font)
                element.Style = style
                element.StyleName = custom_style
                element.TKNotebook = element.Widget = ttk.Notebook(
                    tk_row_frame, style=custom_style
                )

                PackFormIntoFrame(element, toplevel_form.TKroot, toplevel_form)

                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKNotebook.pack(
                    anchor=tk.SW,
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    fill=fill,
                    expand=expand,
                )

                if element.ChangeSubmits:
                    element.TKNotebook.bind(
                        "<<NotebookTabChanged>>", element._TabGroupSelectHandler
                    )
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKNotebook,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                if element.Size != (None, None):
                    element.TKNotebook.configure(
                        width=element.Size[0], height=element.Size[1]
                    )
                _add_right_click_menu_and_grab(element)
                if not element.visible:
                    element._pack_forget_save_settings()
                # row_should_expand = True
                # -------------------  SLIDER placement element  ------------------------- #
            elif element_type == ELEM_TYPE_INPUT_SLIDER:
                element = element  # type: Slider
                slider_length = element_size[0] * _char_width_in_pixels(font)
                slider_width = element_size[1]
                element.TKIntVar = tk.IntVar()
                element.TKIntVar.set(element.DefaultValue)
                if element.Orientation.startswith("v"):
                    range_from = element.Range[1]
                    range_to = element.Range[0]
                    slider_length += DEFAULT_MARGINS[1] * (
                            element_size[0] * 2
                    )  # add in the padding
                else:
                    range_from = element.Range[0]
                    range_to = element.Range[1]
                tkscale = element.Widget = tk.Scale(
                    tk_row_frame,
                    orient=element.Orientation,
                    variable=element.TKIntVar,
                    from_=range_from,
                    to_=range_to,
                    resolution=element.Resolution,
                    length=slider_length,
                    width=slider_width,
                    bd=element.BorderWidth,
                    relief=element.Relief,
                    font=font,
                    tickinterval=element.TickInterval,
                )
                tkscale.config(highlightthickness=0)
                if element.ChangeSubmits:
                    tkscale.config(command=element._SliderChangedHandler)
                if element.BackgroundColor not in (None, COLOR_SYSTEM_DEFAULT):
                    tkscale.configure(background=element.BackgroundColor)
                if element.TroughColor != COLOR_SYSTEM_DEFAULT:
                    tkscale.config(troughcolor=element.TroughColor)
                if element.DisableNumericDisplay:
                    tkscale.config(showvalue=0)
                if text_color not in (None, COLOR_SYSTEM_DEFAULT):
                    tkscale.configure(fg=text_color)
                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                tkscale.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tkscale.pack_forget()
                element.TKScale = tkscale
                if element.Disabled:
                    element.TKScale["state"] = "disabled"
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKScale,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

            # -------------------------  TABLE placement element  ------------------------- #
            elif element_type == ELEM_TYPE_TABLE:
                element = element  # type: Table
                element.element_frame = frame = tk.Frame(tk_row_frame)
                element.table_frame = frame
                height = element.NumRows
                if element.Justification.startswith("l"):
                    anchor = tk.W
                elif element.Justification.startswith("r"):
                    anchor = tk.E
                else:
                    anchor = tk.CENTER
                column_widths = {}
                # create column width list
                for row in element.Values:
                    for i, col in enumerate(row):
                        col_width = min(len(str(col)), element.MaxColumnWidth)
                        try:
                            if col_width > column_widths[i]:
                                column_widths[i] = col_width
                        except Exception:
                            column_widths[i] = col_width

                if element.ColumnsToDisplay is None:
                    displaycolumns = (
                        element.ColumnHeadings
                        if element.ColumnHeadings is not None
                        else element.Values[0]
                    )
                else:
                    displaycolumns = []
                    for i, should_display in enumerate(element.ColumnsToDisplay):
                        if should_display:
                            if element.ColumnHeadings is not None:
                                displaycolumns.append(element.ColumnHeadings[i])
                            else:
                                displaycolumns.append(str(i))

                column_headings = (
                    element.ColumnHeadings
                    if element.ColumnHeadings is not None
                    else displaycolumns
                )
                if (
                        element.DisplayRowNumbers
                ):  # if display row number, tack on the numbers to front of columns
                    displaycolumns = [
                                         element.RowHeaderText,
                                     ] + displaycolumns
                    if column_headings is not None:
                        column_headings = [
                                              element.RowHeaderText,
                                          ] + element.ColumnHeadings
                    else:
                        column_headings = [
                                              element.RowHeaderText,
                                          ] + displaycolumns
                element.TKTreeview = element.Widget = ttk.Treeview(
                    frame,
                    columns=column_headings,
                    displaycolumns=displaycolumns,
                    show="headings",
                    height=height,
                    selectmode=element.SelectMode,
                )
                treeview = element.TKTreeview
                if element.DisplayRowNumbers:
                    treeview.heading(
                        element.RowHeaderText, text=element.RowHeaderText
                    )  # make a dummy heading
                    row_number_header_width = (
                            _string_width_in_pixels(
                                element.HeaderFont, element.RowHeaderText
                            )
                            + 10
                    )
                    row_number_width = (
                            _string_width_in_pixels(font, str(len(element.Values))) + 10
                    )
                    row_number_width = max(row_number_header_width, row_number_width)
                    treeview.column(
                        element.RowHeaderText,
                        width=row_number_width,
                        minwidth=10,
                        anchor=anchor,
                        stretch=0,
                    )

                headings = (
                    element.ColumnHeadings
                    if element.ColumnHeadings is not None
                    else element.Values[0]
                )
                for i, heading in enumerate(headings):
                    # heading = str(heading)
                    treeview.heading(heading, text=heading)
                    if element.AutoSizeColumns:
                        col_width = column_widths.get(
                            i, len(heading)
                        )  # in case more headings than there are columns of data
                        width = max(
                            col_width * _char_width_in_pixels(font),
                            len(heading) * _char_width_in_pixels(element.HeaderFont),
                        )
                    else:
                        try:
                            width = element.ColumnWidths[i] * _char_width_in_pixels(
                                font
                            )
                        except Exception:
                            width = element.DefaultColumnWidth * _char_width_in_pixels(
                                font
                            )
                    if element.cols_justification is not None:
                        try:
                            if element.cols_justification[i].startswith("l"):
                                col_anchor = tk.W
                            elif element.cols_justification[i].startswith("r"):
                                col_anchor = tk.E
                            elif element.cols_justification[i].startswith("c"):
                                col_anchor = tk.CENTER
                            else:
                                col_anchor = anchor

                        except Exception:  # likely didn't specify enough entries (must be one per col)
                            col_anchor = anchor
                    else:
                        col_anchor = anchor
                    treeview.column(
                        heading,
                        width=width,
                        minwidth=10,
                        anchor=col_anchor,
                        stretch=element.expand_x,
                    )
                # Insert values into the tree
                for i, value in enumerate(element.Values):
                    if element.DisplayRowNumbers:
                        value = [i + element.StartingRowNumber] + value
                    id = treeview.insert(
                        "", "end", text=value, iid=i + 1, values=value, tag=i
                    )
                    element.tree_ids.append(id)
                if element.AlternatingRowColor not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ):  # alternating colors
                    for row in range(0, len(element.Values), 2):
                        treeview.tag_configure(
                            row, background=element.AlternatingRowColor
                        )
                if element.RowColors is not None:  # individual row colors
                    for row_def in element.RowColors:
                        if len(row_def) == 2:  # only background is specified
                            treeview.tag_configure(row_def[0], background=row_def[1])
                        else:
                            treeview.tag_configure(
                                row_def[0], background=row_def[2], foreground=row_def[1]
                            )
                # ------ Do Styling of Colors -----
                # style_name = str(element.Key) + 'customtable.Treeview'
                style_name = _make_ttk_style_name(
                    ".Treeview", element, primary_style=True
                )
                element.table_ttk_style_name = style_name
                table_style = ttk.Style()
                element.ttk_style = table_style

                _change_ttk_theme(table_style, toplevel_form.TtkTheme)

                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    table_style.configure(
                        style_name,
                        background=element.BackgroundColor,
                        fieldbackground=element.BackgroundColor,
                    )
                    if element.SelectedRowColors[1] is not None:
                        table_style.map(
                            style_name,
                            background=_fixed_map(
                                table_style,
                                style_name,
                                "background",
                                element.SelectedRowColors,
                            ),
                        )
                if (
                        element.TextColor is not None
                        and element.TextColor != COLOR_SYSTEM_DEFAULT
                ):
                    table_style.configure(style_name, foreground=element.TextColor)
                    if element.SelectedRowColors[0] is not None:
                        table_style.map(
                            style_name,
                            foreground=_fixed_map(
                                table_style,
                                style_name,
                                "foreground",
                                element.SelectedRowColors,
                            ),
                        )
                if element.RowHeight is not None:
                    table_style.configure(style_name, rowheight=element.RowHeight)
                else:
                    table_style.configure(
                        style_name, rowheight=_char_height_in_pixels(font)
                    )
                if (
                        element.HeaderTextColor is not None
                        and element.HeaderTextColor != COLOR_SYSTEM_DEFAULT
                ):
                    table_style.configure(
                        style_name + ".Heading", foreground=element.HeaderTextColor
                    )
                if (
                        element.HeaderBackgroundColor is not None
                        and element.HeaderBackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    table_style.configure(
                        style_name + ".Heading",
                        background=element.HeaderBackgroundColor,
                    )
                if element.HeaderFont is not None:
                    table_style.configure(
                        style_name + ".Heading", font=element.HeaderFont
                    )
                else:
                    table_style.configure(style_name + ".Heading", font=font)
                if element.HeaderBorderWidth is not None:
                    table_style.configure(
                        style_name + ".Heading", borderwidth=element.HeaderBorderWidth
                    )
                if element.HeaderRelief is not None:
                    table_style.configure(
                        style_name + ".Heading", relief=element.HeaderRelief
                    )
                table_style.configure(style_name, font=font)
                if element.BorderWidth is not None:
                    table_style.configure(style_name, borderwidth=element.BorderWidth)

                if element.HeaderBackgroundColor not in (
                        None,
                        COLOR_SYSTEM_DEFAULT,
                ) and element.HeaderTextColor not in (None, COLOR_SYSTEM_DEFAULT):
                    table_style.map(
                        style_name + ".Heading",
                        background=[
                            ("pressed", "!focus", element.HeaderBackgroundColor),
                            ("active", element.HeaderTextColor),
                        ],
                    )
                    table_style.map(
                        style_name + ".Heading",
                        foreground=[
                            ("pressed", "!focus", element.HeaderTextColor),
                            ("active", element.HeaderBackgroundColor),
                        ],
                    )

                treeview.configure(style=style_name)
                # scrollable_frame.pack(side=tk.LEFT,  padx=elementpad[0], pady=elementpad[1], expand=True, fill='both')
                if element.enable_click_events:
                    treeview.bind("<ButtonRelease-1>", element._table_clicked)
                if element.right_click_selects:
                    if running_mac():
                        treeview.bind("<Button-2>", element._table_clicked)
                    else:
                        treeview.bind("<Button-3>", element._table_clicked)
                treeview.bind("<<TreeviewSelect>>", element._treeview_selected)
                if element.BindReturnKey:
                    treeview.bind("<Return>", element._treeview_double_click)
                    treeview.bind("<Double-Button-1>", element._treeview_double_click)

                if not element.HideVerticalScroll:
                    _make_ttk_scrollbar(element, "v", toplevel_form)

                    element.Widget.configure(yscrollcommand=element.vsb.set)
                    element.vsb.pack(side=tk.RIGHT, fill="y")

                # Horizontal scrollbar
                if not element.VerticalScrollOnly:
                    # element.Widget.config(wrap='none')
                    _make_ttk_scrollbar(element, "h", toplevel_form)
                    element.hsb.pack(side=tk.BOTTOM, fill="x")
                    element.Widget.configure(xscrollcommand=element.hsb.set)

                if not element.HideVerticalScroll or not element.VerticalScrollOnly:
                    # Chr0nic
                    element.Widget.bind(
                        "<Enter>", lambda event, em=element: testMouseHook(em)
                    )
                    element.Widget.bind(
                        "<Leave>", lambda event, em=element: testMouseUnhook(em)
                    )

                # if not element.HideVerticalScroll:
                #     scrollbar = tk.Scrollbar(frame)
                #     scrollbar.pack(side=tk.RIGHT, fill='y')
                #     scrollbar.config(command=treeview.yview)
                #     treeview.configure(yscrollcommand=scrollbar.set)

                # if not element.VerticalScrollOnly:
                #     hscrollbar = tk.Scrollbar(frame, orient=tk.HORIZONTAL)
                #     hscrollbar.pack(side=tk.BOTTOM, fill='x')
                #     hscrollbar.config(command=treeview.xview)
                #     treeview.configure(xscrollcommand=hscrollbar.set)

                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKTreeview.pack(
                    side=tk.LEFT, padx=0, pady=0, expand=expand, fill=fill
                )
                frame.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings(
                        alternate_widget=element.element_frame
                    )  # seems like it should be the frame if following other elements conventions
                    # element.TKTreeview.pack_forget()
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKTreeview,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

                if tclversion_detailed == "8.6.9" and ENABLE_TREEVIEW_869_PATCH:
                    # print('*** tk version 8.6.9 detected.... patching ttk treeview code ***')
                    table_style.map(
                        style_name,
                        foreground=_fixed_map(
                            table_style,
                            style_name,
                            "foreground",
                            element.SelectedRowColors,
                        ),
                        background=_fixed_map(
                            table_style,
                            style_name,
                            "background",
                            element.SelectedRowColors,
                        ),
                    )
            # -------------------------  Tree placement element  ------------------------- #
            elif element_type == ELEM_TYPE_TREE:
                element = element  # type: Tree
                element.element_frame = element_frame = tk.Frame(tk_row_frame)

                height = element.NumRows
                if element.Justification.startswith("l"):  # justification
                    anchor = tk.W
                elif element.Justification.startswith("r"):
                    anchor = tk.E
                else:
                    anchor = tk.CENTER

                if element.ColumnsToDisplay is None:  # Which cols to display
                    displaycolumns = element.ColumnHeadings
                else:
                    displaycolumns = []
                    for i, should_display in enumerate(element.ColumnsToDisplay):
                        if should_display:
                            displaycolumns.append(element.ColumnHeadings[i])
                column_headings = element.ColumnHeadings
                # ------------- GET THE TREEVIEW WIDGET -------------
                element.TKTreeview = element.Widget = ttk.Treeview(
                    element_frame,
                    columns=column_headings,
                    displaycolumns=displaycolumns,
                    show="tree headings" if column_headings is not None else "tree",
                    height=height,
                    selectmode=element.SelectMode,
                )
                treeview = element.TKTreeview
                max_widths = {}
                for key, node in element.TreeData.tree_dict.items():
                    for i, value in enumerate(node.values):
                        max_width = max_widths.get(i, 0)
                        if len(str(value)) > max_width:
                            max_widths[i] = len(str(value))

                if element.ColumnHeadings is not None:
                    for i, heading in enumerate(
                            element.ColumnHeadings
                    ):  # Configure cols + headings
                        treeview.heading(heading, text=heading)
                        if element.AutoSizeColumns:
                            max_width = max_widths.get(i, 0)
                            max_width = max(max_width, len(heading))
                            width = min(element.MaxColumnWidth, max_width + 1)
                        else:
                            try:
                                width = element.ColumnWidths[i]
                            except Exception:
                                width = element.DefaultColumnWidth
                        treeview.column(
                            heading,
                            width=width * _char_width_in_pixels(font) + 10,
                            anchor=anchor,
                        )

                def add_treeview_data(node):
                    """

                    :param node:
                    :type node:

                    """
                    if node.key != "":
                        if node.icon:
                            if node.icon not in element.image_dict:
                                if type(node.icon) is bytes:
                                    photo = tk.PhotoImage(data=node.icon)
                                else:
                                    photo = tk.PhotoImage(file=node.icon)
                                element.image_dict[node.icon] = photo
                            else:
                                photo = element.image_dict.get(node.icon)

                            node.photo = photo
                            try:
                                id = treeview.insert(
                                    element.KeyToID[node.parent],
                                    "end",
                                    iid=None,
                                    text=node.text,
                                    values=node.values,
                                    open=element.ShowExpanded,
                                    image=node.photo,
                                )
                                element.IdToKey[id] = node.key
                                element.KeyToID[node.key] = id
                            except Exception as e:
                                print("Error inserting image into tree", e)
                        else:
                            id = treeview.insert(
                                element.KeyToID[node.parent],
                                "end",
                                iid=None,
                                text=node.text,
                                values=node.values,
                                open=element.ShowExpanded,
                            )
                            element.IdToKey[id] = node.key
                            element.KeyToID[node.key] = id

                    for node in node.children:
                        add_treeview_data(node)

                add_treeview_data(element.TreeData.root_node)
                treeview.column(
                    "#0",
                    width=element.Col0Width * _char_width_in_pixels(font),
                    anchor=tk.W,
                )
                treeview.heading("#0", text=element.col0_heading)

                # ----- configure colors -----
                # style_name = str(element.Key) + '.Treeview'
                style_name = _make_ttk_style_name(
                    ".Treeview", element, primary_style=True
                )
                tree_style = ttk.Style()
                _change_ttk_theme(tree_style, toplevel_form.TtkTheme)

                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    tree_style.configure(
                        style_name,
                        background=element.BackgroundColor,
                        fieldbackground=element.BackgroundColor,
                    )
                    if element.SelectedRowColors[1] is not None:
                        tree_style.map(
                            style_name,
                            background=_fixed_map(
                                tree_style,
                                style_name,
                                "background",
                                element.SelectedRowColors,
                            ),
                        )
                if (
                        element.TextColor is not None
                        and element.TextColor != COLOR_SYSTEM_DEFAULT
                ):
                    tree_style.configure(style_name, foreground=element.TextColor)
                    if element.SelectedRowColors[0] is not None:
                        tree_style.map(
                            style_name,
                            foreground=_fixed_map(
                                tree_style,
                                style_name,
                                "foreground",
                                element.SelectedRowColors,
                            ),
                        )
                if (
                        element.HeaderTextColor is not None
                        and element.HeaderTextColor != COLOR_SYSTEM_DEFAULT
                ):
                    tree_style.configure(
                        style_name + ".Heading", foreground=element.HeaderTextColor
                    )
                if (
                        element.HeaderBackgroundColor is not None
                        and element.HeaderBackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    tree_style.configure(
                        style_name + ".Heading",
                        background=element.HeaderBackgroundColor,
                    )
                if element.HeaderFont is not None:
                    tree_style.configure(
                        style_name + ".Heading", font=element.HeaderFont
                    )
                else:
                    tree_style.configure(style_name + ".Heading", font=font)
                if element.HeaderBorderWidth is not None:
                    tree_style.configure(
                        style_name + ".Heading", borderwidth=element.HeaderBorderWidth
                    )
                if element.HeaderRelief is not None:
                    tree_style.configure(
                        style_name + ".Heading", relief=element.HeaderRelief
                    )
                tree_style.configure(style_name, font=font)
                if element.RowHeight:
                    tree_style.configure(style_name, rowheight=element.RowHeight)
                else:
                    tree_style.configure(
                        style_name, rowheight=_char_height_in_pixels(font)
                    )
                if element.BorderWidth is not None:
                    tree_style.configure(style_name, borderwidth=element.BorderWidth)

                treeview.configure(
                    style=style_name
                )  # IMPORTANT! Be sure and set the style name for this widget

                if not element.HideVerticalScroll:
                    _make_ttk_scrollbar(element, "v", toplevel_form)

                    element.Widget.configure(yscrollcommand=element.vsb.set)
                    element.vsb.pack(side=tk.RIGHT, fill="y")

                # Horizontal scrollbar
                if not element.VerticalScrollOnly:
                    # element.Widget.config(wrap='none')
                    _make_ttk_scrollbar(element, "h", toplevel_form)
                    element.hsb.pack(side=tk.BOTTOM, fill="x")
                    element.Widget.configure(xscrollcommand=element.hsb.set)

                if not element.HideVerticalScroll or not element.VerticalScrollOnly:
                    # Chr0nic
                    element.Widget.bind(
                        "<Enter>", lambda event, em=element: testMouseHook(em)
                    )
                    element.Widget.bind(
                        "<Leave>", lambda event, em=element: testMouseUnhook(em)
                    )

                # Horizontal scrollbar
                # if not element.VerticalScrollOnly:
                #     element.TKText.config(wrap='none')
                #     _make_ttk_scrollbar(element, 'h')
                #     element.hsb.pack(side=tk.BOTTOM, fill='x')
                #     element.Widget.configure(xscrollcommand=element.hsb.set)

                # if not element.HideVerticalScroll or not element.VerticalScrollOnly:
                # Chr0nic
                # element.Widget.bind("<Enter>", lambda event, em=element: testMouseHook(em))
                # element.Widget.bind("<Leave>", lambda event, em=element: testMouseUnhook(em))

                # element.scrollbar = scrollbar = tk.Scrollbar(element_frame)
                # scrollbar.pack(side=tk.RIGHT, fill='y')
                # scrollbar.config(command=treeview.yview)
                # treeview.configure(yscrollcommand=scrollbar.set)

                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )
                element.TKTreeview.pack(
                    side=tk.LEFT, padx=0, pady=0, expand=expand, fill=fill
                )
                element_frame.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    expand=expand,
                    fill=fill,
                )
                if not element.visible:
                    element._pack_forget_save_settings(
                        alternate_widget=element.element_frame
                    )  # seems like it should be the frame if following other elements conventions
                    # element.TKTreeview.pack_forget()
                treeview.bind("<<TreeviewSelect>>", element._treeview_selected)
                if element.Tooltip is not None:  # tooltip
                    element.TooltipObject = ToolTip(
                        element.TKTreeview,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

                if tclversion_detailed == "8.6.9" and ENABLE_TREEVIEW_869_PATCH:
                    # print('*** tk version 8.6.9 detected.... patching ttk treeview code ***')
                    tree_style.map(
                        style_name,
                        foreground=_fixed_map(
                            tree_style,
                            style_name,
                            "foreground",
                            element.SelectedRowColors,
                        ),
                        background=_fixed_map(
                            tree_style,
                            style_name,
                            "background",
                            element.SelectedRowColors,
                        ),
                    )

            # -------------------------  Separator placement element  ------------------------- #
            elif element_type == ELEM_TYPE_SEPARATOR:
                element = element  # type: VerticalSeparator
                # style_name = str(element.Key) + "Line.TSeparator"
                style_name = _make_ttk_style_name(
                    ".Line.TSeparator", element, primary_style=True
                )
                style = ttk.Style()

                _change_ttk_theme(style, toplevel_form.TtkTheme)

                if element.color not in (None, COLOR_SYSTEM_DEFAULT):
                    style.configure(style_name, background=element.color)
                separator = element.Widget = ttk.Separator(
                    tk_row_frame,
                    orient=element.Orientation,
                )

                expand, fill, row_should_expand, row_fill_direction = _add_expansion(
                    element, row_should_expand, row_fill_direction
                )

                if element.Orientation.startswith("h"):
                    separator.pack(
                        side=tk.LEFT,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        fill=tk.X,
                        expand=True,
                    )
                else:
                    separator.pack(
                        side=tk.LEFT,
                        padx=elementpad[0],
                        pady=elementpad[1],
                        fill=tk.Y,
                        expand=False,
                    )
                element.Widget.configure(
                    style=style_name
                )  # IMPORTANT!  Apply the style
            # -------------------------  SizeGrip placement element  ------------------------- #
            elif element_type == ELEM_TYPE_SIZEGRIP:
                element = element  # type: Sizegrip
                style_name = "Sizegrip.TSizegrip"
                style = ttk.Style()

                _change_ttk_theme(style, toplevel_form.TtkTheme)

                size_grip = element.Widget = ttk.Sizegrip(tk_row_frame)
                toplevel_form.sizegrip_widget = size_grip
                # if no size is specified, then use the background color for the window
                if element.BackgroundColor != COLOR_SYSTEM_DEFAULT:
                    style.configure(style_name, background=element.BackgroundColor)
                else:
                    style.configure(style_name, background=toplevel_form.TKroot["bg"])
                size_grip.configure(style=style_name)

                size_grip.pack(
                    side=tk.BOTTOM,
                    anchor="se",
                    padx=elementpad[0],
                    pady=elementpad[1],
                    fill=tk.X,
                    expand=True,
                )
                # tricky part of sizegrip... it shouldn't cause the row to expand, but should expand and should add X axis if
                # not already filling in that direction.  Otherwise, leaves things alone!
                # row_should_expand = True
                row_fill_direction = (
                    tk.BOTH if row_fill_direction in (tk.Y, tk.BOTH) else tk.X
                )
            # -------------------------  StatusBar placement element  ------------------------- #
            elif element_type == ELEM_TYPE_STATUSBAR:
                # auto_size_text = element.AutoSizeText
                display_text = element.DisplayText  # text to display
                if not auto_size_text:
                    width, height = element_size
                else:
                    lines = display_text.split("\n")
                    max_line_len = max([len(l) for l in lines])
                    num_lines = len(lines)
                    if (
                            max_line_len > element_size[0]
                    ):  # if text exceeds element size, the will have to wrap
                        width = element_size[0]
                    else:
                        width = max_line_len
                    height = num_lines
                # ---===--- LABEL widget create and place --- #
                stringvar = tk.StringVar()
                element.TKStringVar = stringvar
                stringvar.set(display_text)
                if auto_size_text:
                    width = 0
                if element.Justification is not None:
                    justification = element.Justification
                elif toplevel_form.TextJustification is not None:
                    justification = toplevel_form.TextJustification
                else:
                    justification = DEFAULT_TEXT_JUSTIFICATION
                justify = (
                    tk.LEFT
                    if justification.startswith("l")
                    else tk.CENTER
                    if justification.startswith("c")
                    else tk.RIGHT
                )
                anchor = (
                    tk.NW
                    if justification.startswith("l")
                    else tk.N
                    if justification.startswith("c")
                    else tk.NE
                )
                # tktext_label = tk.Label(tk_row_frame, textvariable=stringvar, width=width, height=height,
                #                         justify=justify, bd=border_depth, font=font)
                tktext_label = element.Widget = tk.Label(
                    tk_row_frame,
                    textvariable=stringvar,
                    width=width,
                    height=height,
                    justify=justify,
                    bd=border_depth,
                    font=font,
                )
                # Set wrap-length for text (in PIXELS) == PAIN IN THE ASS
                wraplen = (
                        tktext_label.winfo_reqwidth() + 40
                )  # width of widget in Pixels
                if not auto_size_text and height == 1:
                    wraplen = 0
                # print("wraplen, width, height", wraplen, width, height)
                tktext_label.configure(
                    anchor=anchor, wraplen=wraplen
                )  # set wrap to width of widget
                if element.Relief is not None:
                    tktext_label.configure(relief=element.Relief)
                if (
                        element.BackgroundColor is not None
                        and element.BackgroundColor != COLOR_SYSTEM_DEFAULT
                ):
                    tktext_label.configure(background=element.BackgroundColor)
                if (
                        element.TextColor != COLOR_SYSTEM_DEFAULT
                        and element.TextColor is not None
                ):
                    tktext_label.configure(fg=element.TextColor)
                tktext_label.pack(
                    side=tk.LEFT,
                    padx=elementpad[0],
                    pady=elementpad[1],
                    fill=tk.X,
                    expand=True,
                )
                row_fill_direction = tk.X
                if not element.visible:
                    element._pack_forget_save_settings()
                    # tktext_label.pack_forget()
                element.TKText = tktext_label
                if element.ClickSubmits:
                    tktext_label.bind("<Button-1>", element._TextClickedHandler)
                if element.Tooltip is not None:
                    element.TooltipObject = ToolTip(
                        element.TKText,
                        text=element.Tooltip,
                        timeout=DEFAULT_TOOLTIP_TIME,
                    )
                _add_right_click_menu_and_grab(element)

        # ............................DONE WITH ROW pack the row of widgets ..........................#
        # done with row, pack the row of widgets
        # tk_row_frame.grid(row=row_num+2, sticky=tk.NW, padx=DEFAULT_MARGINS[0])

        anchor = "nw"

        if row_justify.lower().startswith("c"):
            anchor = "n"
            side = tk.LEFT
        elif row_justify.lower().startswith("r"):
            anchor = "ne"
            side = tk.RIGHT
        elif row_justify.lower().startswith("l"):
            anchor = "nw"
            side = tk.LEFT
        # elif toplevel_form.ElementJustification.lower().startswith('c'):
        #     anchor = 'n'
        #     side = tk.TOP
        # elif toplevel_form.ElementJustification.lower().startswith('r'):
        #     anchor = 'ne'
        #     side = tk.TOP
        # else:
        #     anchor = 'nw'
        #     side = tk.TOP

        # row_should_expand = False

        # if form.RightClickMenu:
        #     menu = form.RightClickMenu
        #     top_menu = tk.Menu(toplevel_form.TKroot, tearoff=False)
        #     AddMenuItem(top_menu, menu[1], form)
        #     tk_row_frame.bind('<Button-3>', form._RightClickMenuCallback)

        tk_row_frame.pack(
            side=tk.TOP,
            anchor=anchor,
            padx=0,
            pady=0,
            expand=row_should_expand,
            fill=row_fill_direction,
        )
        if (
                form.BackgroundColor is not None
                and form.BackgroundColor != COLOR_SYSTEM_DEFAULT
        ):
            tk_row_frame.configure(background=form.BackgroundColor)

    return


def _get_hidden_master_root():
    """
    Creates the hidden master root window.  This window is never visible and represents the overall "application"
    """

    # if one is already made, then skip making another
    if Window.hidden_master_root is None:
        Window._IncrementOpenCount()
        Window.hidden_master_root = tk.Tk()
        Window.hidden_master_root.attributes(
            "-alpha", 0
        )  # HIDE this window really really really
        # if not running_mac():
        try:
            Window.hidden_master_root.wm_overrideredirect(True)
        except Exception as e:
            if not running_mac():
                print(
                    "* Error performing wm_overrideredirect while hiding the hidden master root*",
                    e,
                )
        Window.hidden_master_root.withdraw()
    return Window.hidden_master_root


def _no_titlebar_setup(window):
    """
    Does the operations required to turn off the titlebar for the window.
    The Raspberry Pi required the settings to be make after the window's creation.
    Calling twice seems to have had better overall results so that's what's currently done.
    The MAC has been the problem with this feature.  It's been a chronic problem on the Mac.
    :param window:          window to turn off the titlebar if indicated in the settings
    :type window:           Window
    """
    try:
        if window.NoTitleBar:
            if running_linux():
                # window.TKroot.wm_attributes("-type", 'splash')
                window.TKroot.wm_attributes("-type", "dock")
            else:
                window.TKroot.wm_overrideredirect(True)
                # Special case for Mac. Need to clear flag again if not tkinter version 8.6.10+
                # Previously restricted patch to only certain tkinter versions. Now use the patch setting exclusively regardless of tk ver
                # if running_mac() and ENABLE_MAC_NOTITLEBAR_PATCH and (sum([int(i) for i in tclversion_detailed.split('.')]) < 24):
                # if running_mac() and ENABLE_MAC_NOTITLEBAR_PATCH:
                if _mac_should_apply_notitlebar_patch():
                    print("* Applying Mac no_titlebar patch *")
                    window.TKroot.wm_overrideredirect(False)
    except Exception as e:
        warnings.warn("** Problem setting no titlebar {} **".format(e), UserWarning)


def _convert_window_to_tk(window):
    """

    :type window: (Window)

    """
    master = window.TKroot
    master.title(window.Title)
    InitializeResults(window)

    PackFormIntoFrame(window, master, window)

    window.TKroot.configure(padx=window.Margins[0], pady=window.Margins[1])

    # ....................................... DONE creating and laying out window ..........................#
    if window._Size != (None, None):
        master.geometry("%sx%s" % (window._Size[0], window._Size[1]))
    screen_width = (
        master.winfo_screenwidth()
    )  # get window info to move to middle of screen
    screen_height = master.winfo_screenheight()
    if window.Location is not None:
        if window.Location != (None, None):
            x, y = window.Location
        elif DEFAULT_WINDOW_LOCATION != (None, None):
            x, y = DEFAULT_WINDOW_LOCATION
        else:
            master.update_idletasks()  # don't forget to do updates or values are bad
            win_width = master.winfo_width()
            win_height = master.winfo_height()
            x = screen_width / 2 - win_width / 2
            y = screen_height / 2 - win_height / 2
            if y + win_height > screen_height:
                y = screen_height - win_height
            if x + win_width > screen_width:
                x = screen_width - win_width

        if window.RelativeLoction != (None, None):
            x += window.RelativeLoction[0]
            y += window.RelativeLoction[1]

        move_string = "+%i+%i" % (int(x), int(y))
        master.geometry(move_string)
        window.config_last_location = (int(x), (int(y)))
        window.TKroot.x = int(x)
        window.TKroot.y = int(y)
        window.starting_window_position = (int(x), (int(y)))
        master.update_idletasks()  # don't forget
        master.geometry(move_string)
        master.update_idletasks()  # don't forget
    else:
        master.update_idletasks()
        x, y = int(master.winfo_x()), int(master.winfo_y())
        window.config_last_location = x, y
        window.TKroot.x = x
        window.TKroot.y = y
        window.starting_window_position = x, y
    _no_titlebar_setup(window)

    return


# ----====----====----====----====----==== STARTUP TK ====----====----====----====----====----#
def StartupTK(window):
    """
    NOT user callable
    Creates the window (for real) lays out all the elements, etc.  It's a HUGE set of things it does.  It's the basic
    "porting layer" that will change depending on the GUI framework PySimpleGUI is running on top of.

    :param window: you window object
    :type window:  (Window)

    """
    window = window  # type: Window
    # global _my_windows
    # ow = _my_windows.NumOpenWindows
    ow = Window.NumOpenWindows
    # print('Starting TK open Windows = {}'.format(ow))
    if ENABLE_TK_WINDOWS:
        root = tk.Tk()
    elif not ow and not window.ForceTopLevel:
        # if first window being created, make a throwaway, hidden master root.  This stops one user
        # window from becoming the child of another user window. All windows are children of this hidden window
        _get_hidden_master_root()
        root = tk.Toplevel(class_=window.Title)
    else:
        root = tk.Toplevel(class_=window.Title)
    if window.DebuggerEnabled:
        root.bind("<Cancel>", window._callback_main_debugger_window_create_keystroke)
        root.bind("<Pause>", window._callback_popout_window_create_keystroke)

    # If location is None, then there's no need to hide the window.  Let it build where it is going to end up being.
    if DEFAULT_HIDE_WINDOW_WHEN_CREATING is True and window.Location is not None:
        try:
            if (
                    not running_mac()
                    or (running_mac() and not window.NoTitleBar)
                    or (
                    running_mac()
                    and window.NoTitleBar
                    and not _mac_should_apply_notitlebar_patch()
            )
            ):
                root.attributes(
                    "-alpha", 0
                )  # hide window while building it. makes for smoother 'paint'
        except Exception as e:
            print(
                "*** Exception setting alpha channel to zero while creating window ***",
                e,
            )

    if (
            window.BackgroundColor is not None
            and window.BackgroundColor != COLOR_SYSTEM_DEFAULT
    ):
        root.configure(background=window.BackgroundColor)
    Window._IncrementOpenCount()

    window.TKroot = root

    window._create_thread_queue()

    # for the Raspberry Pi. Need to set the attributes here, prior to the building of the window
    # so going ahead and doing it for all platforms, in addition to doing it after the window is packed
    # 2023-April - this call seems to be causing problems on MacOS 13.2.1 Ventura.  Input elements become non-responsive
    # if this call is made here and at the end of building the window
    if not running_mac():
        _no_titlebar_setup(window)

    if not window.Resizable:
        root.resizable(False, False)

    if window.DisableMinimize:
        root.attributes("-toolwindow", 1)

    if window.KeepOnTop:
        root.wm_attributes("-topmost", 1)

    if window.TransparentColor is not None:
        window.SetTransparentColor(window.TransparentColor)

    if window.scaling is not None:
        root.tk.call("tk", "scaling", window.scaling)

    # root.protocol("WM_DELETE_WINDOW", MyFlexForm.DestroyedCallback())
    # root.bind('<Destroy>', MyFlexForm.DestroyedCallback())
    _convert_window_to_tk(window)

    # Make moveable window
    if window.GrabAnywhere is not False and not (
            window.NonBlocking and window.GrabAnywhere is not True
    ):
        if not (
                ENABLE_MAC_DISABLE_GRAB_ANYWHERE_WITH_TITLEBAR
                and running_mac()
                and not window.NoTitleBar
        ):
            root.bind("<ButtonPress-1>", window._StartMoveGrabAnywhere)
            root.bind("<ButtonRelease-1>", window._StopMove)
            root.bind("<B1-Motion>", window._OnMotionGrabAnywhere)
    if window.GrabAnywhereUsingControlKey is not False and not (
            window.NonBlocking and window.GrabAnywhereUsingControlKey is not True
    ):
        root.bind("<Control-Button-1>", window._StartMoveUsingControlKey)
        root.bind("<Control-ButtonRelease-1>", window._StopMove)
        root.bind("<Control-B1-Motion>", window._OnMotionUsingControlKey)
        # also enable movement using Control + Arrow key
        root.bind("<Control-Left>", window._move_callback)
        root.bind("<Control-Right>", window._move_callback)
        root.bind("<Control-Up>", window._move_callback)
        root.bind("<Control-Down>", window._move_callback)

    window.set_icon(window.WindowIcon)
    try:
        alpha_channel = 1 if window.AlphaChannel is None else window.AlphaChannel
        root.attributes("-alpha", alpha_channel)  # Make window visible again
    except Exception as e:
        print(
            "**** Error setting Alpha Channel to {} after window was created ****".format(
                alpha_channel
            ),
            e,
        )
        # pass

    if window.ReturnKeyboardEvents and not window.NonBlocking:
        root.bind("<KeyRelease>", window._KeyboardCallback)
        root.bind("<MouseWheel>", window._MouseWheelCallback)
        root.bind("<Button-4>", window._MouseWheelCallback)
        root.bind("<Button-5>", window._MouseWheelCallback)
    elif window.ReturnKeyboardEvents:
        root.bind("<Key>", window._KeyboardCallback)
        root.bind("<MouseWheel>", window._MouseWheelCallback)
        root.bind("<Button-4>", window._MouseWheelCallback)
        root.bind("<Button-5>", window._MouseWheelCallback)

    DEFAULT_WINDOW_SNAPSHOT_KEY_CODE = main_global_get_screen_snapshot_symcode()

    if DEFAULT_WINDOW_SNAPSHOT_KEY_CODE:
        # print('**** BINDING THE SNAPSHOT!', DEFAULT_WINDOW_SNAPSHOT_KEY_CODE, DEFAULT_WINDOW_SNAPSHOT_KEY)
        window.bind(
            DEFAULT_WINDOW_SNAPSHOT_KEY_CODE,
            DEFAULT_WINDOW_SNAPSHOT_KEY,
            propagate=False,
        )
        # window.bind('<Win_L><F12>', DEFAULT_WINDOW_SNAPSHOT_KEY, )

    if window.NoTitleBar:
        window.TKroot.focus_force()

    if window.AutoClose:
        # if the window is being finalized, then don't start the autoclose timer
        if not window.finalize_in_progress:
            window._start_autoclose_timer()
            # duration = DEFAULT_AUTOCLOSE_TIME if window.AutoCloseDuration is None else window.AutoCloseDuration
            # window.TKAfterID = root.after(int(duration * 1000), window._AutoCloseAlarmCallback)

    if window.Timeout is not None:
        window.TKAfterID = root.after(int(window.Timeout), window._TimeoutAlarmCallback)
    if window.NonBlocking:
        window.TKroot.protocol("WM_DESTROY_WINDOW", window._OnClosingCallback)
        window.TKroot.protocol("WM_DELETE_WINDOW", window._OnClosingCallback)

    else:  # it's a blocking form
        # print('..... CALLING MainLoop')
        window.CurrentlyRunningMainloop = True
        window.TKroot.protocol("WM_DESTROY_WINDOW", window._OnClosingCallback)
        window.TKroot.protocol("WM_DELETE_WINDOW", window._OnClosingCallback)

        if window.modal or DEFAULT_MODAL_WINDOWS_FORCED:
            window.make_modal()

        if window.enable_window_config_events:
            window.TKroot.bind("<Configure>", window._config_callback)

        # ----------------------------------- tkinter mainloop call -----------------------------------
        Window._window_running_mainloop = window
        Window._root_running_mainloop = window.TKroot
        window.TKroot.mainloop()
        window.CurrentlyRunningMainloop = False
        window.TimerCancelled = True
        # print('..... BACK from MainLoop')
        if not window.FormRemainedOpen:
            Window._DecrementOpenCount()
            # _my_windows.Decrement()
        if window.RootNeedsDestroying:
            try:
                window.TKroot.destroy()
            except Exception:
                pass
            window.RootNeedsDestroying = False
    return


def _set_icon_for_tkinter_window(root, icon=None, pngbase64=None):
    """
    At the moment, this function is only used by the get_filename or folder with the no_window option set.
    Changes the icon that is shown on the title bar and on the task bar.
    NOTE - The file type is IMPORTANT and depends on the OS!
    Can pass in:
    * filename which must be a .ICO icon file for windows, PNG file for Linux
    * bytes object
    * BASE64 encoded file held in a variable

    :param root:      The window being modified
    :type root:       (tk.Tk or tk.TopLevel)
    :param icon:      Filename or bytes object
    :type icon:       (str | bytes)
    :param pngbase64: Base64 encoded image
    :type pngbase64:  (bytes)
    """

    if type(icon) is bytes or pngbase64 is not None:
        wicon = tkinter.PhotoImage(data=icon if icon is not None else pngbase64)
        try:
            root.tk.call("wm", "iconphoto", root._w, wicon)
        except Exception:
            wicon = tkinter.PhotoImage(data=DEFAULT_BASE64_ICON)
            try:
                root.tk.call("wm", "iconphoto", root._w, wicon)
            except Exception as e:
                print("Set icon exception", e)
                pass
        return

    wicon = icon
    try:
        root.iconbitmap(icon)
    except Exception as e:
        try:
            wicon = tkinter.PhotoImage(file=icon)
            root.tk.call("wm", "iconphoto", root._w, wicon)
        except Exception as e:
            try:
                wicon = tkinter.PhotoImage(data=DEFAULT_BASE64_ICON)
                try:
                    root.tk.call("wm", "iconphoto", root._w, wicon)
                except Exception as e:
                    print("Set icon exception", e)
                    pass
            except Exception:
                print("Set icon exception", e)
                pass


# ==============================_GetNumLinesNeeded ==#
# Helper function for determining how to wrap text   #
# ===================================================#
def _GetNumLinesNeeded(text, max_line_width):
    if max_line_width == 0:
        return 1
    lines = text.split("\n")
    num_lines = len(lines)  # number of original lines of text
    max_line_len = max([len(l) for l in lines])  # longest line
    lines_used = []
    for L in lines:
        lines_used.append(
            len(L) // max_line_width + (len(L) % max_line_width > 0)
        )  # fancy math to round up
    total_lines_needed = sum(lines_used)
    return total_lines_needed


# ==============================  PROGRESS METER ========================================== #


def convert_args_to_single_string(*args):
    """

    """
    (
        max_line_total,
        width_used,
        total_lines,
    ) = 0, 0, 0
    single_line_message = ""
    # loop through args and built a SINGLE string from them
    for message in args:
        # fancy code to check if string and convert if not is not need. Just always convert to string :-)
        # if not isinstance(message, str): message = str(message)
        message = str(message)
        longest_line_len = max([len(l) for l in message.split("\n")])
        width_used = max(longest_line_len, width_used)
        max_line_total = max(max_line_total, width_used)
        lines_needed = _GetNumLinesNeeded(message, width_used)
        total_lines += lines_needed
        single_line_message += message + "\n"
    return single_line_message, width_used, total_lines


METER_REASON_CANCELLED = "cancelled"
METER_REASON_CLOSED = "closed"
METER_REASON_REACHED_MAX = "finished"
METER_OK = True
METER_STOPPED = False


class _QuickMeter(object):
    active_meters = {}
    exit_reasons = {}

    def __init__(
            self,
            title,
            current_value,
            max_value,
            key,
            *args,
            orientation="v",
            bar_color=(None, None),
            button_color=(None, None),
            size=DEFAULT_PROGRESS_BAR_SIZE,
            border_width=None,
            grab_anywhere=False,
            no_titlebar=False,
            keep_on_top=None,
            no_button=False,
    ):
        """

        :param title:         text to display in element
        :type title:          (str)
        :param current_value: current value
        :type current_value:  (int)
        :param max_value:     max value of progress meter
        :type max_value:      (int)
        :param key:           Used with window.find_element and with return values to uniquely identify this element
        :type key:            str | int | tuple | object
        :param *args:         stuff to output
        :type *args:          (Any)
        :param orientation:   'horizontal' or 'vertical' ('h' or 'v' work) (Default value = 'vertical' / 'v')
        :type orientation:    (str)
        :param bar_color:     The 2 colors that make up a progress bar. Either a tuple of 2 strings or a string. Tuple - (bar, background). A string with 1 color changes the background of the bar only. A string with 2 colors separated by "on" like "red on blue" specifies a red bar on a blue background.
        :type bar_color:      (str, str) or str
        :param button_color:  button color (foreground, background)
        :type button_color:   (str, str) | str
        :param size:          (w,h) w=characters-wide, h=rows-high (Default value = DEFAULT_PROGRESS_BAR_SIZE)
        :type size:           (int, int)
        :param border_width:  width of border around element
        :type border_width:   (int)
        :param grab_anywhere: If True: can grab anywhere to move the window (Default = False)
        :type grab_anywhere:  (bool)
        :param no_titlebar:   If True: window will be created without a titlebar
        :type no_titlebar:    (bool)
        :param keep_on_top:   If True the window will remain above all current windows
        :type keep_on_top:    (bool)
        :param no_button:     If True: window will be created without a cancel button
        :type no_button:      (bool)
        """
        self.start_time = datetime.datetime.utcnow()
        self.key = key
        self.orientation = orientation
        self.bar_color = bar_color
        self.size = size
        self.grab_anywhere = grab_anywhere
        self.button_color = button_color
        self.border_width = border_width
        self.no_titlebar = no_titlebar
        self.title = title
        self.current_value = current_value
        self.max_value = max_value
        self.close_reason = None
        self.keep_on_top = keep_on_top
        self.no_button = no_button
        self.window = self.BuildWindow(*args)

    def BuildWindow(self, *args):
        layout = []
        if self.orientation.lower().startswith("h"):
            col = []
            col += [
                [T("".join(map(lambda x: str(x) + "\n", args)), key="-OPTMSG-")]
            ]  ### convert all *args into one string that can be updated
            col += [
                [T("", size=(30, 10), key="-STATS-")],
                [
                    ProgressBar(
                        max_value=self.max_value,
                        orientation="h",
                        key="-PROG-",
                        size=self.size,
                        bar_color=self.bar_color,
                    )
                ],
            ]
            if not self.no_button:
                col += [[Cancel(button_color=self.button_color), Stretch()]]
            layout = [Column(col)]
        else:
            col = [
                [
                    ProgressBar(
                        max_value=self.max_value,
                        orientation="v",
                        key="-PROG-",
                        size=self.size,
                        bar_color=self.bar_color,
                    )
                ]
            ]
            col2 = []
            col2 += [
                [T("".join(map(lambda x: str(x) + "\n", args)), key="-OPTMSG-")]
            ]  ### convert all *args into one string that can be updated
            col2 += [[T("", size=(30, 10), key="-STATS-")]]
            if not self.no_button:
                col2 += [[Cancel(button_color=self.button_color), Stretch()]]

            layout = [Column(col), Column(col2)]
        self.window = Window(
            self.title,
            grab_anywhere=self.grab_anywhere,
            border_depth=self.border_width,
            no_titlebar=self.no_titlebar,
            disable_close=True,
            keep_on_top=self.keep_on_top,
        )
        self.window.Layout([layout]).Finalize()

        return self.window

    def UpdateMeter(
            self, current_value, max_value, *args
    ):  ### support for *args when updating
        self.current_value = current_value
        self.max_value = max_value
        self.window.Element("-PROG-").UpdateBar(self.current_value, self.max_value)
        self.window.Element("-STATS-").Update("\n".join(self.ComputeProgressStats()))
        self.window.Element("-OPTMSG-").Update(
            value="".join(map(lambda x: str(x) + "\n", args))
        )  ###  update the string with the args
        event, values = self.window.read(timeout=0)
        if event in ("Cancel", None) or current_value >= max_value:
            exit_reason = (
                METER_REASON_CANCELLED
                if event in ("Cancel", None)
                else METER_REASON_REACHED_MAX
                if current_value >= max_value
                else METER_STOPPED
            )
            self.window.close()
            del _QuickMeter.active_meters[self.key]
            _QuickMeter.exit_reasons[self.key] = exit_reason
            return _QuickMeter.exit_reasons[self.key]
        return METER_OK

    def ComputeProgressStats(self):
        utc = datetime.datetime.utcnow()
        time_delta = utc - self.start_time
        total_seconds = time_delta.total_seconds()
        if not total_seconds:
            total_seconds = 1
        try:
            time_per_item = total_seconds / self.current_value
        except Exception:
            time_per_item = 1
        seconds_remaining = (self.max_value - self.current_value) * time_per_item
        time_remaining = str(datetime.timedelta(seconds=seconds_remaining))
        time_remaining_short = time_remaining.split(".")[0]
        time_delta_short = str(time_delta).split(".")[0]
        total_time = time_delta + datetime.timedelta(seconds=seconds_remaining)
        total_time_short = str(total_time).split(".")[0]
        self.stat_messages = [
            "{} of {}".format(self.current_value, self.max_value),
            "{} %".format(100 * self.current_value // self.max_value),
            "",
            " {:6.2f} Iterations per Second".format(self.current_value / total_seconds),
            " {:6.2f} Seconds per Iteration".format(
                total_seconds / (self.current_value if self.current_value else 1)
            ),
            "",
            "{} Elapsed Time".format(time_delta_short),
            "{} Time Remaining".format(time_remaining_short),
            "{} Estimated Total Time".format(total_time_short),
        ]
        return self.stat_messages


def one_line_progress_meter(
        title,
        current_value,
        max_value,
        *args,
        key="OK for 1 meter",
        orientation="v",
        bar_color=(None, None),
        button_color=None,
        size=DEFAULT_PROGRESS_BAR_SIZE,
        border_width=None,
        grab_anywhere=False,
        no_titlebar=False,
        keep_on_top=None,
        no_button=False,
):
    """
    :param title:         text to display in titlebar of window
    :type title:          (str)
    :param current_value: current value
    :type current_value:  (int)
    :param max_value:     max value of progress meter
    :type max_value:      (int)
    :param key:           Used to differentiate between multiple meters. Used to cancel meter early. Now optional as there is a default value for single meters
    :type key:            str | int | tuple | object
    :param orientation:   'horizontal' or 'vertical' ('h' or 'v' work) (Default value = 'vertical' / 'v')
    :type orientation:    (str)
    :param bar_color:     The 2 colors that make up a progress bar. Either a tuple of 2 strings or a string. Tuple - (bar, background). A string with 1 color changes the background of the bar only. A string with 2 colors separated by "on" like "red on blue" specifies a red bar on a blue background.
    :type bar_color:      (str, str) or str
    :param button_color:  button color (foreground, background)
    :type button_color:   (str, str) | str
    :param size:          (w,h) w=characters-wide, h=rows-high (Default value = DEFAULT_PROGRESS_BAR_SIZE)
    :type size:           (int, int)
    :param border_width:  width of border around element
    :type border_width:   (int)
    :param grab_anywhere: If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:  (bool)
    :param no_titlebar:   If True: no titlebar will be shown on the window
    :type no_titlebar:    (bool)
    :param keep_on_top:   If True the window will remain above all current windows
    :type keep_on_top:    (bool)
    :param no_button:     If True: window will be created without a cancel button
    :type no_button:      (bool)
    :return:              True if updated successfully. False if user closed the meter with the X or Cancel button
    :rtype:               (bool)
    """
    if key not in _QuickMeter.active_meters:
        meter = _QuickMeter(
            title,
            current_value,
            max_value,
            key,
            *args,
            orientation=orientation,
            bar_color=bar_color,
            button_color=button_color,
            size=size,
            border_width=border_width,
            grab_anywhere=grab_anywhere,
            no_titlebar=no_titlebar,
            keep_on_top=keep_on_top,
            no_button=no_button,
        )
        _QuickMeter.active_meters[key] = meter
        _QuickMeter.exit_reasons[key] = None

    else:
        meter = _QuickMeter.active_meters[key]

    rc = meter.UpdateMeter(
        current_value, max_value, *args
    )  ### pass the *args to to UpdateMeter function
    OneLineProgressMeter.exit_reasons = getattr(
        OneLineProgressMeter, "exit_reasons", _QuickMeter.exit_reasons
    )
    exit_reason = OneLineProgressMeter.exit_reasons.get(key)
    return (
        METER_OK if exit_reason in (None, METER_REASON_REACHED_MAX) else METER_STOPPED
    )


def one_line_progress_meter_cancel(key="OK for 1 meter"):
    """
    Cancels and closes a previously created One Line Progress Meter window

    :param key: Key used when meter was created
    :type key:  (Any)
    :return:    None
    :rtype:     None
    """
    try:
        meter = _QuickMeter.active_meters[key]
        meter.window.Close()
        del _QuickMeter.active_meters[key]
        _QuickMeter.exit_reasons[key] = METER_REASON_CANCELLED
    except Exception:  # meter is already deleted
        return


def get_complimentary_hex(color):
    """
    :param color: color string, like "#RRGGBB"
    :type color:  (str)
    :return:      color string, like "#RRGGBB"
    :rtype:       (str)
    """

    # strip the # from the beginning
    color = color[1:]
    # convert the string into hex
    color = int(color, 16)
    # invert the three bytes
    # as good as substracting each of RGB component by 255(FF)
    comp_color = 0xFFFFFF ^ color
    # convert the color back to hex by prefixing a #
    comp_color = "#%06X" % comp_color
    return comp_color


# ========================  EasyPrint           =====#
# ===================================================#
class _DebugWin:
    debug_window = None

    def __init__(
            self,
            size=(None, None),
            location=(None, None),
            relative_location=(None, None),
            font=None,
            no_titlebar=False,
            no_button=False,
            grab_anywhere=False,
            keep_on_top=None,
            do_not_reroute_stdout=True,
            echo_stdout=False,
            resizable=True,
            blocking=False,
    ):
        """

        :param size:                  (w,h) w=characters-wide, h=rows-high
        :type size:                   (int, int)
        :param location:              Location of upper left corner of the window
        :type location:               (int, int)
        :param relative_location:     (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
        :type relative_location:      (int, int)
        :param font:                  specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
        :type font:                   (str or (str, int[, str]) or None)
        :param no_titlebar:           If True no titlebar will be shown
        :type no_titlebar:            (bool)
        :param no_button:             show button
        :type no_button:              (bool)
        :param grab_anywhere:         If True: can grab anywhere to move the window (Default = False)
        :type grab_anywhere:          (bool)
        :param location:              Location of upper left corner of the window
        :type location:               (int, int)
        :param do_not_reroute_stdout: bool value
        :type do_not_reroute_stdout:  (bool)
        :param echo_stdout:           If True stdout is sent to both the console and the debug window
        :type echo_stdout:            (bool)
        :param resizable:             if True, makes the window resizble
        :type resizable:              (bool)
        :param blocking:              if True, makes the window block instead of returning immediately
        :type blocking:               (bool)
        """

        # Show a form that's a running counter
        self.size = size
        self.location = location
        self.relative_location = relative_location
        self.font = font
        self.no_titlebar = no_titlebar
        self.no_button = no_button
        self.grab_anywhere = grab_anywhere
        self.keep_on_top = keep_on_top
        self.do_not_reroute_stdout = do_not_reroute_stdout
        self.echo_stdout = echo_stdout
        self.resizable = resizable
        self.blocking = blocking

        win_size = size if size != (None, None) else DEFAULT_DEBUG_WINDOW_SIZE
        self.output_element = Multiline(
            size=win_size,
            autoscroll=True,
            auto_refresh=True,
            reroute_stdout=False if do_not_reroute_stdout else True,
            echo_stdout_stderr=self.echo_stdout,
            reroute_stderr=False if do_not_reroute_stdout else True,
            expand_x=True,
            expand_y=True,
            key="-MULTILINE-",
        )
        if no_button:
            self.layout = [[self.output_element]]
        else:
            if blocking:
                self.quit_button = Button("Quit", key="Quit")
            else:
                self.quit_button = DummyButton("Quit", key="Quit")
            self.layout = [
                [self.output_element],
                [pin(self.quit_button), pin(B("Pause", key="-PAUSE-")), Stretch()],
            ]

        self.layout[-1] += [Sizegrip()]

        self.window = Window(
            "Debug Window",
            self.layout,
            no_titlebar=no_titlebar,
            auto_size_text=True,
            location=location,
            relative_location=relative_location,
            font=font or ("Courier New", 10),
            grab_anywhere=grab_anywhere,
            keep_on_top=keep_on_top,
            finalize=True,
            resizable=resizable,
        )
        return

    def reopen_window(self):
        if self.window is None or (self.window is not None and self.window.is_closed()):
            self.__init__(
                size=self.size,
                location=self.location,
                relative_location=self.relative_location,
                font=self.font,
                no_titlebar=self.no_titlebar,
                no_button=self.no_button,
                grab_anywhere=self.grab_anywhere,
                keep_on_top=self.keep_on_top,
                do_not_reroute_stdout=self.do_not_reroute_stdout,
                resizable=self.resizable,
                echo_stdout=self.echo_stdout,
            )

    def Print(
            self,
            *args,
            end=None,
            sep=None,
            text_color=None,
            background_color=None,
            erase_all=False,
            font=None,
            blocking=None,
    ):
        global SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS
        suppress = SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS
        SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS = True
        sepchar = sep if sep is not None else " "
        endchar = end if end is not None else "\n"
        self.reopen_window()  # if needed, open the window again

        timeout = 0 if not blocking else None
        if erase_all:
            self.output_element.update("")

        if self.do_not_reroute_stdout:
            end_str = str(end) if end is not None else "\n"
            sep_str = str(sep) if sep is not None else " "

            outstring = ""
            num_args = len(args)
            for i, arg in enumerate(args):
                outstring += str(arg)
                if i != num_args - 1:
                    outstring += sep_str
            outstring += end_str
            try:
                self.output_element.update(
                    outstring,
                    append=True,
                    text_color_for_value=text_color,
                    background_color_for_value=background_color,
                    font_for_value=font,
                )
            except Exception:
                self.window = None
                self.reopen_window()
                self.output_element.update(
                    outstring,
                    append=True,
                    text_color_for_value=text_color,
                    background_color_for_value=background_color,
                    font_for_value=font,
                )

        else:
            print(*args, sep=sepchar, end=endchar)
        # This is tricky....changing the button type depending on the blocking parm. If blocking, then the "Quit" button should become a normal button
        if blocking and not self.no_button:
            self.quit_button.BType = BUTTON_TYPE_READ_FORM
            try:  # The window may be closed by user at any time, so have to protect
                self.quit_button.update(text="Click to continue...")
            except Exception:
                self.window = None
        elif not self.no_button:
            self.quit_button.BType = BUTTON_TYPE_CLOSES_WIN_ONLY
            try:  # The window may be closed by user at any time, so have to protect
                self.quit_button.update(text="Quit")
            except Exception:
                self.window = None

        try:  # The window may be closed by user at any time, so have to protect
            if blocking and not self.no_button:
                self.window["-PAUSE-"].update(visible=False)
            elif not self.no_button:
                self.window["-PAUSE-"].update(visible=True)
        except Exception:
            self.window = None

        self.reopen_window()  # if needed, open the window again

        paused = None
        while True:
            event, values = self.window.read(timeout=timeout)

            if event == WIN_CLOSED:
                self.Close()
                break
            elif blocking and event == "Quit":
                break
            elif not paused and event == TIMEOUT_EVENT and not blocking:
                break
            elif event == "-PAUSE-":
                if (
                        blocking or self.no_button
                ):  # if blocking or shouldn't have been a button event, ignore the pause button entirely
                    continue
                if paused:
                    self.window["-PAUSE-"].update(text="Pause")
                    self.quit_button.update(visible=True)
                    break
                paused = True
                self.window["-PAUSE-"].update(text="Resume")
                self.quit_button.update(visible=False)
                timeout = None

        SUPPRESS_WIDGET_NOT_FINALIZED_WARNINGS = suppress

    def Close(self):
        if self.window.XFound:  # increment the number of open windows to get around a bug with debug windows
            Window._IncrementOpenCount()
        self.window.close()
        self.window = None


def easy_print(
        *args,
        size=(None, None),
        end=None,
        sep=None,
        location=(None, None),
        relative_location=(None, None),
        font=None,
        no_titlebar=False,
        no_button=False,
        grab_anywhere=False,
        keep_on_top=None,
        do_not_reroute_stdout=True,
        echo_stdout=False,
        text_color=None,
        background_color=None,
        colors=None,
        c=None,
        erase_all=False,
        resizable=True,
        blocking=None,
        wait=None,
):
    """
    Works like a "print" statement but with windowing options.  Routes output to the "Debug Window"

    In addition to the normal text and background colors, you can use a "colors" tuple/string
    The "colors" or "c" parameter defines both the text and background in a single parm.
    It can be a tuple or a single single. Both text and background colors need to be specified
    colors -(str, str) or str.  A combined text/background color definition in a single parameter
    c - (str, str) - Colors tuple has format (foreground, backgrouned)
    c - str - can also be a string of the format "foreground on background"  ("white on red")

    :param size:                  (w,h) w=characters-wide, h=rows-high
    :type size:                   (int, int)
    :param end:                   end character
    :type end:                    (str)
    :param sep:                   separator character
    :type sep:                    (str)
    :param location:              Location of upper left corner of the window
    :type location:               (int, int)
    :param relative_location:     (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:      (int, int)
    :param font:                  specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                   (str or (str, int[, str]) or None)
    :param no_titlebar:           If True no titlebar will be shown
    :type no_titlebar:            (bool)
    :param no_button:             don't show button
    :type no_button:              (bool)
    :param grab_anywhere:         If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:          (bool)
    :param background_color:      color of background
    :type background_color:       (str)
    :param text_color:            color of the text
    :type text_color:             (str)
    :param keep_on_top:           If True the window will remain above all current windows
    :type keep_on_top:            (bool)
    :param location:              Location of upper left corner of the window
    :type location:               (int, int)
    :param do_not_reroute_stdout: do not reroute stdout and stderr. If False, both stdout and stderr will reroute to here
    :type do_not_reroute_stdout:  (bool)
    :param echo_stdout:           If True stdout is sent to both the console and the debug window
    :type echo_stdout:            (bool)
    :param colors:                Either a tuple or a string that has both the text and background colors
    :type colors:                 (str) or (str, str)
    :param c:                     Either a tuple or a string that has both the text and background colors
    :type c:                      (str) or (str, str)
    :param resizable:             if True, the user can resize the debug window. Default is True
    :type resizable:              (bool)
    :param erase_all:             If True when erase the output before printing
    :type erase_all:              (bool)
    :param blocking:              if True, makes the window block instead of returning immediately. The "Quit" button changers to "More"
    :type blocking:               (bool | None)
    :param wait:                  Same as the "blocking" parm. It's an alias.  if True, makes the window block instead of returning immediately. The "Quit" button changes to "Click to Continue..."
    :type wait:                   (bool | None)
    :return:
    :rtype:
    """

    blocking = blocking or wait
    if _DebugWin.debug_window is None:
        _DebugWin.debug_window = _DebugWin(
            size=size,
            location=location,
            relative_location=relative_location,
            font=font,
            no_titlebar=no_titlebar,
            no_button=no_button,
            grab_anywhere=grab_anywhere,
            keep_on_top=keep_on_top,
            do_not_reroute_stdout=do_not_reroute_stdout,
            echo_stdout=echo_stdout,
            resizable=resizable,
            blocking=blocking,
        )
    txt_color, bg_color = _parse_colors_parm(c or colors)
    _DebugWin.debug_window.Print(
        *args,
        end=end,
        sep=sep,
        text_color=text_color or txt_color,
        background_color=background_color or bg_color,
        erase_all=erase_all,
        font=font,
        blocking=blocking,
    )


def easy_print_close():
    """
    Close a previously opened EasyPrint window

    :return:
    :rtype:
    """
    if _DebugWin.debug_window is not None:
        _DebugWin.debug_window.Close()
        _DebugWin.debug_window = None


#                            d8b          888
#                            Y8P          888
#                                         888
#   .d8888b 88888b.  888d888 888 88888b.  888888
#  d88P"    888 "88b 888P"   888 888 "88b 888
#  888      888  888 888     888 888  888 888
#  Y88b.    888 d88P 888     888 888  888 Y88b.
#   "Y8888P 88888P"  888     888 888  888  "Y888
#           888
#           888
#           888


CPRINT_DESTINATION_WINDOW = None
CPRINT_DESTINATION_MULTILINE_ELMENT_KEY = None


def cprint_set_output_destination(window, multiline_key):
    """
    Sets up the color print (cprint) output destination
    :param window:        The window that the cprint call will route the output to
    :type window:         (Window)
    :param multiline_key: Key for the Multiline Element where output will be sent
    :type multiline_key:  (Any)
    :return:              None
    :rtype:               None
    """

    global CPRINT_DESTINATION_WINDOW, CPRINT_DESTINATION_MULTILINE_ELMENT_KEY

    CPRINT_DESTINATION_WINDOW = window
    CPRINT_DESTINATION_MULTILINE_ELMENT_KEY = multiline_key


def cprint(
        *args,
        end=None,
        sep=" ",
        text_color=None,
        font=None,
        t=None,
        background_color=None,
        b=None,
        colors=None,
        c=None,
        window=None,
        key=None,
        justification=None,
        autoscroll=True,
        erase_all=False,
):
    """
    Color print to a multiline element in a window of your choice.
    Must have EITHER called cprint_set_output_destination prior to making this call so that the
    window and element key can be saved and used here to route the output, OR used the window
    and key parameters to the cprint function to specicy these items.

    args is a variable number of things you want to print.

    end - The end char to use just like print uses
    sep - The separation character like print uses
    text_color - The color of the text
            key - overrides the previously defined Multiline key
    window - overrides the previously defined window to output to
    background_color - The color of the background
    colors -(str, str) or str.  A combined text/background color definition in a single parameter

    There are also "aliases" for text_color, background_color and colors (t, b, c)
     t - An alias for color of the text (makes for shorter calls)
    b - An alias for the background_color parameter
    c - (str, str) - "shorthand" way of specifying color. (foreground, backgrouned)
    c - str - can also be a string of the format "foreground on background"  ("white on red")

    With the aliases it's possible to write the same print but in more compact ways:
    cprint('This will print white text on red background', c=('white', 'red'))
    cprint('This will print white text on red background', c='white on red')
    cprint('This will print white text on red background', text_color='white', background_color='red')
    cprint('This will print white text on red background', t='white', b='red')

    :param text_color:       Color of the text
    :type text_color:        (str)
    :param font:             specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the value being updated
    :type font:              (str or (str, int[, str]) or None)
    :param background_color: The background color of the line
    :type background_color:  (str)
    :param colors:           Either a tuple or a string that has both the text and background colors "text on background" or just the text color
    :type colors:            (str) or (str, str)
    :param t:                Color of the text
    :type t:                 (str)
    :param b:                The background color of the line
    :type b:                 (str)
    :param c:                Either a tuple or a string.  Same as the color parm
    :type c:                 (str) or (str, str)
    :param end:              end character
    :type end:               (str)
    :param sep:              separator character
    :type sep:               (str)
    :param key:              key of multiline to output to (if you want to override the one previously set)
    :type key:               (Any)
    :param window:           Window containing the multiline to output to (if you want to override the one previously set)
    :type window:            (Window)
    :param justification:    text justification. left, right, center. Can use single characters l, r, c. Sets only for this value, not entire element
    :type justification:     (str)
    :param autoscroll:       If True the contents of the element will automatically scroll as more data added to the end
    :type autoscroll:        (bool)
    :param erase_all         If True the contents of the element will be cleared before printing happens
    :type erase_all          (bool)
    """

    destination_key = CPRINT_DESTINATION_MULTILINE_ELMENT_KEY if key is None else key
    destination_window = window or CPRINT_DESTINATION_WINDOW

    if (destination_window is None and window is None) or (
            destination_key is None and key is None
    ):
        print(
            "** Warning ** Attempting to perform a cprint without a valid window & key",
            "Will instead print on Console",
            "You can specify window and key in this cprint call, or set ahead of time using cprint_set_output_destination",
        )
        print(*args)
        return

    kw_text_color = text_color or t
    kw_background_color = background_color or b
    dual_color = colors or c
    try:
        if isinstance(dual_color, tuple):
            kw_text_color = dual_color[0]
            kw_background_color = dual_color[1]
        elif isinstance(dual_color, str):
            if (
                    " on " in dual_color
            ):  # if has "on" in the string, then have both text and background
                kw_text_color = dual_color.split(" on ")[0]
                kw_background_color = dual_color.split(" on ")[1]
            else:  # if no "on" then assume the color string is just the text color
                kw_text_color = dual_color
    except Exception as e:
        print("* cprint warning * you messed up with color formatting", e)

    mline = destination_window.find_element(destination_key, silent_on_error=True)  # type: Multiline
    try:
        # mline = destination_window[destination_key]     # type: Multiline
        if erase_all:
            mline.update("")
        if end is None:
            mline.print(
                *args,
                text_color=kw_text_color,
                background_color=kw_background_color,
                end="",
                sep=sep,
                justification=justification,
                font=font,
                autoscroll=autoscroll,
            )
            mline.print("", justification=justification, autoscroll=autoscroll)
        else:
            mline.print(
                *args,
                text_color=kw_text_color,
                background_color=kw_background_color,
                end=end,
                sep=sep,
                justification=justification,
                font=font,
                autoscroll=autoscroll,
            )
    except Exception as e:
        print(
            "** cprint error trying to print to the multiline. Printing to console instead **",
            e,
        )
        print(*args, end=end, sep=sep)


# ------------------------------------------------------------------------------------------------ #
# A print-like call that can be used to output to a multiline element as if it's an Output element #
# ------------------------------------------------------------------------------------------------ #


def _print_to_element(
        multiline_element,
        *args,
        end=None,
        sep=None,
        text_color=None,
        background_color=None,
        autoscroll=None,
        justification=None,
        font=None,
):
    """
    Print like Python normally prints except route the output to a multiline element and also add colors if desired

    :param multiline_element: The multiline element to be output to
    :type multiline_element:  (Multiline)
    :param args:              The arguments to print
    :type args:               List[Any]
    :param end:               The end char to use just like print uses
    :type end:                (str)
    :param sep:               The separation character like print uses
    :type sep:                (str)
    :param text_color:        color of the text
    :type text_color:         (str)
    :param background_color:  The background color of the line
    :type background_color:   (str)
    :param autoscroll:        If True (the default), the element will scroll to bottom after updating
    :type autoscroll:         (bool)
    :param font:              specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike for the value being updated
    :type font:               str | (str, int)
    """
    end_str = str(end) if end is not None else "\n"
    sep_str = str(sep) if sep is not None else " "

    outstring = ""
    num_args = len(args)
    for i, arg in enumerate(args):
        outstring += str(arg)
        if i != num_args - 1:
            outstring += sep_str
    outstring += end_str

    multiline_element.update(
        outstring,
        append=True,
        text_color_for_value=text_color,
        background_color_for_value=background_color,
        autoscroll=autoscroll,
        justification=justification,
        font_for_value=font,
    )

    try:  # if the element is set to autorefresh, then refresh the parent window
        if multiline_element.AutoRefresh:
            multiline_element.ParentForm.refresh()
    except Exception:
        pass


def _parse_colors_parm(colors):
    """
    Parse a colors parameter into its separate colors.
    Some functions accept a dual colors string/tuple.
    This function parses the parameter into the component colors

    :param colors: Either a tuple or a string that has both the text and background colors
    :type colors:  (str) or (str, str)
    :return:       tuple with the individual text and background colors
    :rtype:        (str, str)
    """
    if colors is None:
        return None, None
    dual_color = colors
    kw_text_color = kw_background_color = None
    try:
        if isinstance(dual_color, tuple):
            kw_text_color = dual_color[0]
            kw_background_color = dual_color[1]
        elif isinstance(dual_color, str):
            if (
                    " on " in dual_color
            ):  # if has "on" in the string, then have both text and background
                kw_text_color = dual_color.split(" on ")[0]
                kw_background_color = dual_color.split(" on ")[1]
            else:  # if no "on" then assume the color string is just the text color
                kw_text_color = dual_color
    except Exception as e:
        print("* warning * you messed up with color formatting", e)

    return kw_text_color, kw_background_color


# ============================== set_global_icon ====#
# Sets the icon to be used by default                #
# ===================================================#
def set_global_icon(icon):
    """
    Sets the icon which will be used any time a window is created if an icon is not provided when the
    window is created.

    :param icon: Either a Base64 byte string or a filename
    :type icon:  bytes | str
    """

    Window._user_defined_icon = icon


# ============================== set_options ========#
# Sets the icon to be used by default                #
# ===================================================#
def set_options(
        icon=None,
        button_color=None,
        element_size=(None, None),
        button_element_size=(None, None),
        margins=(None, None),
        element_padding=(None, None),
        auto_size_text=None,
        auto_size_buttons=None,
        font=None,
        border_width=None,
        slider_border_width=None,
        slider_relief=None,
        slider_orientation=None,
        autoclose_time=None,
        message_box_line_width=None,
        progress_meter_border_depth=None,
        progress_meter_style=None,
        progress_meter_relief=None,
        progress_meter_color=None,
        progress_meter_size=None,
        text_justification=None,
        background_color=None,
        element_background_color=None,
        text_element_background_color=None,
        input_elements_background_color=None,
        input_text_color=None,
        scrollbar_color=None,
        text_color=None,
        element_text_color=None,
        debug_win_size=(None, None),
        window_location=(None, None),
        error_button_color=(None, None),
        tooltip_time=None,
        tooltip_font=None,
        use_ttk_buttons=None,
        ttk_theme=None,
        suppress_error_popups=None,
        suppress_raise_key_errors=None,
        suppress_key_guessing=None,
        warn_button_key_duplicates=False,
        enable_treeview_869_patch=None,
        enable_mac_notitlebar_patch=None,
        use_custom_titlebar=None,
        titlebar_background_color=None,
        titlebar_text_color=None,
        titlebar_font=None,
        titlebar_icon=None,
        user_settings_path=None,
        pysimplegui_settings_path=None,
        pysimplegui_settings_filename=None,
        keep_on_top=None,
        dpi_awareness=None,
        scaling=None,
        disable_modal_windows=None,
        force_modal_windows=None,
        tooltip_offset=(None, None),
        sbar_trough_color=None,
        sbar_background_color=None,
        sbar_arrow_color=None,
        sbar_width=None,
        sbar_arrow_width=None,
        sbar_frame_color=None,
        sbar_relief=None,
        alpha_channel=None,
        hide_window_when_creating=None,
        use_button_shortcuts=None,
        watermark_text=None,
):
    """
    :param icon:                            Can be either a filename or Base64 value. For Windows if filename, it MUST be ICO format. For Linux, must NOT be ICO. Most portable is to use a Base64 of a PNG file. This works universally across all OS's
    :type icon:                             bytes | str
    :param button_color:                    Color of the button (text, background)
    :type button_color:                     (str, str) | str
    :param element_size:                    element size (width, height) in characters
    :type element_size:                     (int, int)
    :param button_element_size:             Size of button
    :type button_element_size:              (int, int)
    :param margins:                         (left/right, top/bottom) tkinter margins around outsize. Amount of pixels to leave inside the window's frame around the edges before your elements are shown.
    :type margins:                          (int, int)
    :param element_padding:                 Default amount of padding to put around elements in window (left/right, top/bottom) or ((left, right), (top, bottom))
    :type element_padding:                  (int, int) or ((int, int),(int,int))
    :param auto_size_text:                  True if the Widget should be shrunk to exactly fit the number of chars to show
    :type auto_size_text:                   bool
    :param auto_size_buttons:               True if Buttons in this Window should be sized to exactly fit the text on this.
    :type auto_size_buttons:                (bool)
    :param font:                            specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                             (str or (str, int[, str]) or None)
    :param border_width:                    width of border around element
    :type border_width:                     (int)
    :param slider_border_width:             Width of the border around sliders
    :type slider_border_width:              (int)
    :param slider_relief:                   Type of relief to use for sliders
    :type slider_relief:                    (str)
    :param slider_orientation:              ???
    :type slider_orientation:               ???
    :param autoclose_time:                  ???
    :type autoclose_time:                   ???
    :param message_box_line_width:          ???
    :type message_box_line_width:           ???
    :param progress_meter_border_depth:     ???
    :type progress_meter_border_depth:      ???
    :param progress_meter_style:            You can no longer set a progress bar style. All ttk styles must be the same for the window
    :type progress_meter_style:             ???
    :param progress_meter_relief:
    :type progress_meter_relief:            ???
    :param progress_meter_color:            ???
    :type progress_meter_color:             ???
    :param progress_meter_size:             ???
    :type progress_meter_size:              ???
    :param text_justification:              Default text justification for all Text Elements in window
    :type text_justification:               'left' | 'right' | 'center'
    :param background_color:                color of background
    :type background_color:                 (str)
    :param element_background_color:        element background color
    :type element_background_color:         (str)
    :param text_element_background_color:   text element background color
    :type text_element_background_color:    (str)
    :param input_elements_background_color: Default color to use for the background of input elements
    :type input_elements_background_color:  (str)
    :param input_text_color:                Default color to use for the text for Input elements
    :type input_text_color:                 (str)
    :param scrollbar_color:                 Default color to use for the slider trough
    :type scrollbar_color:                  (str)
    :param text_color:                      color of the text
    :type text_color:                       (str)
    :param element_text_color:              Default color to use for Text elements
    :type element_text_color:               (str)
    :param debug_win_size:                  window size
    :type debug_win_size:                   (int, int)
    :param window_location:                 Default location to place windows. Not setting will center windows on the display
    :type window_location:                  (int, int) | None
    :param error_button_color:              (Default = (None))
    :type error_button_color:               ???
    :param tooltip_time:                    time in milliseconds to wait before showing a tooltip. Default is 400ms
    :type tooltip_time:                     (int)
    :param tooltip_font:                    font to use for all tooltips
    :type tooltip_font:                     str or Tuple[str, int] or Tuple[str, int, str]
    :param use_ttk_buttons:                 if True will cause all buttons to be ttk buttons
    :type use_ttk_buttons:                  (bool)
    :param ttk_theme:                       Theme to use with ttk widgets.  Choices (on Windows) include - 'default', 'winnative', 'clam', 'alt', 'classic', 'vista', 'xpnative'
    :type ttk_theme:                        (str)
    :param suppress_error_popups:           If True then error popups will not be shown if generated internally to PySimpleGUI
    :type suppress_error_popups:            (bool)
    :param suppress_raise_key_errors:       If True then key errors won't be raised (you'll still get popup error)
    :type suppress_raise_key_errors:        (bool)
    :param suppress_key_guessing:           If True then key errors won't try and find closest matches for you
    :type suppress_key_guessing:            (bool)
    :param warn_button_key_duplicates:      If True then duplicate Button Keys generate warnings (not recommended as they're expected)
    :type warn_button_key_duplicates:       (bool)
    :param enable_treeview_869_patch:       If True, then will use the treeview color patch for tk 8.6.9
    :type enable_treeview_869_patch:        (bool)
    :param enable_mac_notitlebar_patch:     If True then Windows with no titlebar use an alternative technique when tkinter version < 8.6.10
    :type enable_mac_notitlebar_patch:      (bool)
    :param use_custom_titlebar:             If True then a custom titlebar is used instead of the normal system titlebar
    :type use_custom_titlebar:              (bool)
    :param titlebar_background_color:       If custom titlebar indicated by use_custom_titlebar, then use this as background color
    :type titlebar_background_color:        str | None
    :param titlebar_text_color:             If custom titlebar indicated by use_custom_titlebar, then use this as text color
    :type titlebar_text_color:              str | None
    :param titlebar_font:                   If custom titlebar indicated by use_custom_titlebar, then use this as title font
    :type titlebar_font:                    (str or (str, int[, str]) or None) | None
    :param titlebar_icon:                   If custom titlebar indicated by use_custom_titlebar, then use this as the icon (file or base64 bytes)
    :type titlebar_icon:                    bytes | str
    :param user_settings_path:              default path for user_settings API calls. Expanded with os.path.expanduser so can contain ~ to represent user
    :type user_settings_path:               (str)
    :param pysimplegui_settings_path:       default path for the global PySimpleGUI user_settings
    :type pysimplegui_settings_path:        (str)
    :param pysimplegui_settings_filename:   default filename for the global PySimpleGUI user_settings
    :type pysimplegui_settings_filename:    (str)
    :param keep_on_top:                     If True then all windows will automatically be set to keep_on_top=True
    :type keep_on_top:                      (bool)
    :param dpi_awareness:                   If True then will turn on DPI awareness (Windows only at the moment)
    :type dpi_awareness:                    (bool)
    :param scaling:                         Sets the default scaling for all windows including popups, etc.
    :type scaling:                          (float)
    :param disable_modal_windows:           If True then all windows, including popups, will not be modal windows (unless they've been set to FORCED using another option)
    :type disable_modal_windows:            (bool)
    :param force_modal_windows:             If True then all windows will be modal (the disable option will be ignored... all windows will be forced to be modal)
    :type force_modal_windows:              (bool)
    :param tooltip_offset:                  Offset to use for tooltips as a tuple. These values will be added to the mouse location when the widget was entered.
    :type tooltip_offset:                   ((None, None) | (int, int))
    :param sbar_trough_color:               Scrollbar color of the trough
    :type sbar_trough_color:                (str)
    :param sbar_background_color:           Scrollbar color of the background of the arrow buttons at the ends AND the color of the "thumb" (the thing you grab and slide). Switches to arrow color when mouse is over
    :type sbar_background_color:            (str)
    :param sbar_arrow_color:                Scrollbar color of the arrow at the ends of the scrollbar (it looks like a button). Switches to background color when mouse is over
    :type sbar_arrow_color:                 (str)
    :param sbar_width:                      Scrollbar width in pixels
    :type sbar_width:                       (int)
    :param sbar_arrow_width:                Scrollbar width of the arrow on the scrollbar. It will potentially impact the overall width of the scrollbar
    :type sbar_arrow_width:                 (int)
    :param sbar_frame_color:                Scrollbar Color of frame around scrollbar (available only on some ttk themes)
    :type sbar_frame_color:                 (str)
    :param sbar_relief:                     Scrollbar relief that will be used for the "thumb" of the scrollbar (the thing you grab that slides). Should be a constant that is defined at starting with "RELIEF_" - RELIEF_RAISED, RELIEF_SUNKEN, RELIEF_FLAT, RELIEF_RIDGE, RELIEF_GROOVE, RELIEF_SOLID
    :type sbar_relief:                      (str)
    :param alpha_channel:                   Default alpha channel to be used on all windows
    :type alpha_channel:                    (float)
    :param hide_window_when_creating:       If True then alpha will be set to 0 while a window is made and moved to location indicated
    :type hide_window_when_creating:        (bool)
    :param use_button_shortcuts:            If True then Shortcut Char will be used with Buttons
    :type use_button_shortcuts:             (bool)
    :param watermark_text:                  Set the text that will be used if a window is watermarked
    :type watermark_text:                   (str)
    :return:                                None
    :rtype:                                 None
    """

    global DEFAULT_ELEMENT_SIZE
    global DEFAULT_BUTTON_ELEMENT_SIZE
    global DEFAULT_MARGINS  # Margins for each LEFT/RIGHT margin is first term
    global DEFAULT_ELEMENT_PADDING  # Padding between elements (row, col) in pixels
    global DEFAULT_AUTOSIZE_TEXT
    global DEFAULT_AUTOSIZE_BUTTONS
    global DEFAULT_FONT
    global DEFAULT_BORDER_WIDTH
    global DEFAULT_AUTOCLOSE_TIME
    global DEFAULT_BUTTON_COLOR
    global MESSAGE_BOX_LINE_WIDTH
    global DEFAULT_PROGRESS_BAR_BORDER_WIDTH
    global DEFAULT_PROGRESS_BAR_STYLE
    global DEFAULT_PROGRESS_BAR_RELIEF
    global DEFAULT_PROGRESS_BAR_COLOR
    global DEFAULT_PROGRESS_BAR_SIZE
    global DEFAULT_TEXT_JUSTIFICATION
    global DEFAULT_DEBUG_WINDOW_SIZE
    global DEFAULT_SLIDER_BORDER_WIDTH
    global DEFAULT_SLIDER_RELIEF
    global DEFAULT_SLIDER_ORIENTATION
    global DEFAULT_BACKGROUND_COLOR
    global DEFAULT_INPUT_ELEMENTS_COLOR
    global DEFAULT_ELEMENT_BACKGROUND_COLOR
    global DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR
    global DEFAULT_SCROLLBAR_COLOR
    global DEFAULT_TEXT_COLOR
    global DEFAULT_WINDOW_LOCATION
    global DEFAULT_ELEMENT_TEXT_COLOR
    global DEFAULT_INPUT_TEXT_COLOR
    global DEFAULT_TOOLTIP_TIME
    global DEFAULT_ERROR_BUTTON_COLOR
    global DEFAULT_TTK_THEME
    global USE_TTK_BUTTONS
    global TOOLTIP_FONT
    global SUPPRESS_ERROR_POPUPS
    global SUPPRESS_RAISE_KEY_ERRORS
    global SUPPRESS_KEY_GUESSING
    global WARN_DUPLICATE_BUTTON_KEY_ERRORS
    global ENABLE_TREEVIEW_869_PATCH
    global ENABLE_MAC_NOTITLEBAR_PATCH
    global USE_CUSTOM_TITLEBAR
    global CUSTOM_TITLEBAR_BACKGROUND_COLOR
    global CUSTOM_TITLEBAR_TEXT_COLOR
    global CUSTOM_TITLEBAR_ICON
    global CUSTOM_TITLEBAR_FONT
    global DEFAULT_USER_SETTINGS_PATH
    global DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH
    global DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME
    global DEFAULT_KEEP_ON_TOP
    global DEFAULT_SCALING
    global DEFAULT_MODAL_WINDOWS_ENABLED
    global DEFAULT_MODAL_WINDOWS_FORCED
    global DEFAULT_TOOLTIP_OFFSET
    global DEFAULT_ALPHA_CHANNEL
    global _pysimplegui_user_settings
    global ttk_part_overrides_from_options
    global DEFAULT_HIDE_WINDOW_WHEN_CREATING
    global DEFAULT_USE_BUTTON_SHORTCUTS
    # global _my_windows

    if icon:
        Window._user_defined_icon = icon
        # _my_windows._user_defined_icon = icon

    if button_color is not None:
        if button_color == COLOR_SYSTEM_DEFAULT:
            DEFAULT_BUTTON_COLOR = (COLOR_SYSTEM_DEFAULT, COLOR_SYSTEM_DEFAULT)
        else:
            DEFAULT_BUTTON_COLOR = button_color

    if element_size != (None, None):
        DEFAULT_ELEMENT_SIZE = element_size

    if button_element_size != (None, None):
        DEFAULT_BUTTON_ELEMENT_SIZE = button_element_size

    if margins != (None, None):
        DEFAULT_MARGINS = margins

    if element_padding != (None, None):
        DEFAULT_ELEMENT_PADDING = element_padding

    if auto_size_text is not None:
        DEFAULT_AUTOSIZE_TEXT = auto_size_text

    if auto_size_buttons is not None:
        DEFAULT_AUTOSIZE_BUTTONS = auto_size_buttons

    if font is not None:
        DEFAULT_FONT = font

    if border_width is not None:
        DEFAULT_BORDER_WIDTH = border_width

    if autoclose_time is not None:
        DEFAULT_AUTOCLOSE_TIME = autoclose_time

    if message_box_line_width is not None:
        MESSAGE_BOX_LINE_WIDTH = message_box_line_width

    if progress_meter_border_depth is not None:
        DEFAULT_PROGRESS_BAR_BORDER_WIDTH = progress_meter_border_depth

    if progress_meter_style is not None:
        warnings.warn(
            "You can no longer set a progress bar style. All ttk styles must be the same for the window",
            UserWarning,
        )
        # DEFAULT_PROGRESS_BAR_STYLE = progress_meter_style

    if progress_meter_relief is not None:
        DEFAULT_PROGRESS_BAR_RELIEF = progress_meter_relief

    if progress_meter_color is not None:
        DEFAULT_PROGRESS_BAR_COLOR = progress_meter_color

    if progress_meter_size is not None:
        DEFAULT_PROGRESS_BAR_SIZE = progress_meter_size

    if slider_border_width is not None:
        DEFAULT_SLIDER_BORDER_WIDTH = slider_border_width

    if slider_orientation is not None:
        DEFAULT_SLIDER_ORIENTATION = slider_orientation

    if slider_relief is not None:
        DEFAULT_SLIDER_RELIEF = slider_relief

    if text_justification is not None:
        DEFAULT_TEXT_JUSTIFICATION = text_justification

    if background_color is not None:
        DEFAULT_BACKGROUND_COLOR = background_color

    if text_element_background_color is not None:
        DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR = text_element_background_color

    if input_elements_background_color is not None:
        DEFAULT_INPUT_ELEMENTS_COLOR = input_elements_background_color

    if element_background_color is not None:
        DEFAULT_ELEMENT_BACKGROUND_COLOR = element_background_color

    if window_location != (None, None):
        DEFAULT_WINDOW_LOCATION = window_location

    if debug_win_size != (None, None):
        DEFAULT_DEBUG_WINDOW_SIZE = debug_win_size

    if text_color is not None:
        DEFAULT_TEXT_COLOR = text_color

    if scrollbar_color is not None:
        DEFAULT_SCROLLBAR_COLOR = scrollbar_color

    if element_text_color is not None:
        DEFAULT_ELEMENT_TEXT_COLOR = element_text_color

    if input_text_color is not None:
        DEFAULT_INPUT_TEXT_COLOR = input_text_color

    if tooltip_time is not None:
        DEFAULT_TOOLTIP_TIME = tooltip_time

    if error_button_color != (None, None):
        DEFAULT_ERROR_BUTTON_COLOR = error_button_color

    if ttk_theme is not None:
        DEFAULT_TTK_THEME = ttk_theme

    if use_ttk_buttons is not None:
        USE_TTK_BUTTONS = use_ttk_buttons

    if tooltip_font is not None:
        TOOLTIP_FONT = tooltip_font

    if suppress_error_popups is not None:
        SUPPRESS_ERROR_POPUPS = suppress_error_popups

    if suppress_raise_key_errors is not None:
        SUPPRESS_RAISE_KEY_ERRORS = suppress_raise_key_errors

    if suppress_key_guessing is not None:
        SUPPRESS_KEY_GUESSING = suppress_key_guessing

    if warn_button_key_duplicates is not None:
        WARN_DUPLICATE_BUTTON_KEY_ERRORS = warn_button_key_duplicates

    if enable_treeview_869_patch is not None:
        ENABLE_TREEVIEW_869_PATCH = enable_treeview_869_patch

    if enable_mac_notitlebar_patch is not None:
        ENABLE_MAC_NOTITLEBAR_PATCH = enable_mac_notitlebar_patch

    if use_custom_titlebar is not None:
        USE_CUSTOM_TITLEBAR = use_custom_titlebar

    if titlebar_background_color is not None:
        CUSTOM_TITLEBAR_BACKGROUND_COLOR = titlebar_background_color

    if titlebar_text_color is not None:
        CUSTOM_TITLEBAR_TEXT_COLOR = titlebar_text_color

    if titlebar_font is not None:
        CUSTOM_TITLEBAR_FONT = titlebar_font

    if titlebar_icon is not None:
        CUSTOM_TITLEBAR_ICON = titlebar_icon

    if user_settings_path is not None:
        DEFAULT_USER_SETTINGS_PATH = user_settings_path

    if pysimplegui_settings_path is not None:
        DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH = pysimplegui_settings_path

    if pysimplegui_settings_filename is not None:
        DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME = pysimplegui_settings_filename

    if (
            pysimplegui_settings_filename is not None
            or pysimplegui_settings_filename is not None
    ):
        _pysimplegui_user_settings = UserSettings(
            filename=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME,
            path=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH,
        )

    if keep_on_top is not None:
        DEFAULT_KEEP_ON_TOP = keep_on_top

    if dpi_awareness:
        if running_windows():
            if platform.release() == "7":
                ctypes.windll.user32.SetProcessDPIAware()
            elif platform.release() == "8" or platform.release() == "10":
                ctypes.windll.shcore.SetProcessDpiAwareness(1)

    if scaling is not None:
        DEFAULT_SCALING = scaling

    if disable_modal_windows is not None:
        DEFAULT_MODAL_WINDOWS_ENABLED = not disable_modal_windows

    if force_modal_windows is not None:
        DEFAULT_MODAL_WINDOWS_FORCED = force_modal_windows

    if tooltip_offset != (None, None):
        DEFAULT_TOOLTIP_OFFSET = tooltip_offset

    if alpha_channel is not None:
        DEFAULT_ALPHA_CHANNEL = alpha_channel

    # ---------------- ttk scrollbar section ----------------
    if sbar_background_color is not None:
        ttk_part_overrides_from_options.sbar_background_color = sbar_background_color

    if sbar_trough_color is not None:
        ttk_part_overrides_from_options.sbar_trough_color = sbar_trough_color

    if sbar_arrow_color is not None:
        ttk_part_overrides_from_options.sbar_arrow_color = sbar_arrow_color

    if sbar_frame_color is not None:
        ttk_part_overrides_from_options.sbar_frame_color = sbar_frame_color

    if sbar_relief is not None:
        ttk_part_overrides_from_options.sbar_relief = sbar_relief

    if sbar_arrow_width is not None:
        ttk_part_overrides_from_options.sbar_arrow_width = sbar_arrow_width

    if sbar_width is not None:
        ttk_part_overrides_from_options.sbar_width = sbar_width

    if hide_window_when_creating is not None:
        DEFAULT_HIDE_WINDOW_WHEN_CREATING = hide_window_when_creating

    if use_button_shortcuts is not None:
        DEFAULT_USE_BUTTON_SHORTCUTS = use_button_shortcuts

    if watermark_text is not None:
        Window._watermark_user_text = watermark_text

    return True


# ----------------------------------------------------------------- #

# .########.##.....##.########.##.....##.########..######.
# ....##....##.....##.##.......###...###.##.......##....##
# ....##....##.....##.##.......####.####.##.......##......
# ....##....#########.######...##.###.##.######....######.
# ....##....##.....##.##.......##.....##.##.............##
# ....##....##.....##.##.......##.....##.##.......##....##
# ....##....##.....##.########.##.....##.########..######.

# ----------------------------------------------------------------- #

# The official Theme code

#################### ChangeLookAndFeel #######################
# Predefined settings that will change the colors and styles #
# of the elements.                                           #
##############################################################
LOOK_AND_FEEL_TABLE = {
    "SystemDefault": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": OFFICIAL_PYSIMPLEGUI_BUTTON_COLOR,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "SystemDefaultForReal": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": COLOR_SYSTEM_DEFAULT,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "SystemDefault1": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": COLOR_SYSTEM_DEFAULT,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "Material1": {
        "BACKGROUND": "#E3F2FD",
        "TEXT": "#000000",
        "INPUT": "#86A8FF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#86A8FF",
        "BUTTON": ("#FFFFFF", "#5079D3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 0,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#FF0266",
        "ACCENT2": "#FF5C93",
        "ACCENT3": "#C5003C",
    },
    "Material2": {
        "BACKGROUND": "#FAFAFA",
        "TEXT": "#000000",
        "INPUT": "#004EA1",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#5EA7FF",
        "BUTTON": ("#FFFFFF", "#0079D3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 0,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#FF0266",
        "ACCENT2": "#FF5C93",
        "ACCENT3": "#C5003C",
    },
    "Reddit": {
        "BACKGROUND": "#ffffff",
        "TEXT": "#1a1a1b",
        "INPUT": "#dae0e6",
        "TEXT_INPUT": "#222222",
        "SCROLL": "#a5a4a4",
        "BUTTON": ("#FFFFFF", "#0079d3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#ff5414",
        "ACCENT2": "#33a8ff",
        "ACCENT3": "#dbf0ff",
    },
    "Topanga": {
        "BACKGROUND": "#282923",
        "TEXT": "#E7DB74",
        "INPUT": "#393a32",
        "TEXT_INPUT": "#E7C855",
        "SCROLL": "#E7C855",
        "BUTTON": ("#E7C855", "#284B5A"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#c15226",
        "ACCENT2": "#7a4d5f",
        "ACCENT3": "#889743",
    },
    "GreenTan": {
        "BACKGROUND": "#9FB8AD",
        "TEXT": "#000000",
        "INPUT": "#F7F3EC",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#F7F3EC",
        "BUTTON": ("#FFFFFF", "#475841"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Dark": {
        "BACKGROUND": "#404040",
        "TEXT": "#FFFFFF",
        "INPUT": "#4D4D4D",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#707070",
        "BUTTON": ("#FFFFFF", "#004F00"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightGreen": {
        "BACKGROUND": "#B7CECE",
        "TEXT": "#000000",
        "INPUT": "#FDFFF7",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#FDFFF7",
        "BUTTON": ("#FFFFFF", "#658268"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "ACCENT1": "#76506d",
        "ACCENT2": "#5148f1",
        "ACCENT3": "#0a1c84",
        "PROGRESS_DEPTH": 0,
    },
    "Dark2": {
        "BACKGROUND": "#404040",
        "TEXT": "#FFFFFF",
        "INPUT": "#FFFFFF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#707070",
        "BUTTON": ("#FFFFFF", "#004F00"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Black": {
        "BACKGROUND": "#000000",
        "TEXT": "#FFFFFF",
        "INPUT": "#4D4D4D",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#707070",
        "BUTTON": ("#000000", "#FFFFFF"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Black2": {
        "BACKGROUND": "#000000",
        "TEXT": "#FFFFFF",
        "INPUT": "#000000",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#FFFFFF",
        "BUTTON": ("#000000", "#FFFFFF"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Tan": {
        "BACKGROUND": "#fdf6e3",
        "TEXT": "#268bd1",
        "INPUT": "#eee8d5",
        "TEXT_INPUT": "#6c71c3",
        "SCROLL": "#eee8d5",
        "BUTTON": ("#FFFFFF", "#063542"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "TanBlue": {
        "BACKGROUND": "#e5dece",
        "TEXT": "#063289",
        "INPUT": "#f9f8f4",
        "TEXT_INPUT": "#242834",
        "SCROLL": "#eee8d5",
        "BUTTON": ("#FFFFFF", "#063289"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkTanBlue": {
        "BACKGROUND": "#242834",
        "TEXT": "#dfe6f8",
        "INPUT": "#97755c",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#a9afbb",
        "BUTTON": ("#FFFFFF", "#063289"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkAmber": {
        "BACKGROUND": "#2c2825",
        "TEXT": "#fdcb52",
        "INPUT": "#705e52",
        "TEXT_INPUT": "#fdcb52",
        "SCROLL": "#705e52",
        "BUTTON": ("#000000", "#fdcb52"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlue": {
        "BACKGROUND": "#1a2835",
        "TEXT": "#d1ecff",
        "INPUT": "#335267",
        "TEXT_INPUT": "#acc2d0",
        "SCROLL": "#1b6497",
        "BUTTON": ("#000000", "#fafaf8"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Reds": {
        "BACKGROUND": "#280001",
        "TEXT": "#FFFFFF",
        "INPUT": "#d8d584",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#763e00",
        "BUTTON": ("#000000", "#daad28"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Green": {
        "BACKGROUND": "#82a459",
        "TEXT": "#000000",
        "INPUT": "#d8d584",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e3ecf3",
        "BUTTON": ("#FFFFFF", "#517239"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "BluePurple": {
        "BACKGROUND": "#A5CADD",
        "TEXT": "#6E266E",
        "INPUT": "#E0F5FF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#E0F5FF",
        "BUTTON": ("#FFFFFF", "#303952"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Purple": {
        "BACKGROUND": "#B0AAC2",
        "TEXT": "#000000",
        "INPUT": "#F2EFE8",
        "SCROLL": "#F2EFE8",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#C2D4D8"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "BlueMono": {
        "BACKGROUND": "#AAB6D3",
        "TEXT": "#000000",
        "INPUT": "#F1F4FC",
        "SCROLL": "#F1F4FC",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#7186C7"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "GreenMono": {
        "BACKGROUND": "#A8C1B4",
        "TEXT": "#000000",
        "INPUT": "#DDE0DE",
        "SCROLL": "#E3E3E3",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#6D9F85"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "BrownBlue": {
        "BACKGROUND": "#64778d",
        "TEXT": "#FFFFFF",
        "INPUT": "#f0f3f7",
        "SCROLL": "#A6B2BE",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#283b5b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "BrightColors": {
        "BACKGROUND": "#b4ffb4",
        "TEXT": "#000000",
        "INPUT": "#ffff64",
        "SCROLL": "#ffb482",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#ffa0dc"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "NeutralBlue": {
        "BACKGROUND": "#92aa9d",
        "TEXT": "#000000",
        "INPUT": "#fcfff6",
        "SCROLL": "#fcfff6",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#d0dbbd"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Kayak": {
        "BACKGROUND": "#a7ad7f",
        "TEXT": "#000000",
        "INPUT": "#e6d3a8",
        "SCROLL": "#e6d3a8",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#5d907d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "SandyBeach": {
        "BACKGROUND": "#efeccb",
        "TEXT": "#012f2f",
        "INPUT": "#e6d3a8",
        "SCROLL": "#e6d3a8",
        "TEXT_INPUT": "#012f2f",
        "BUTTON": ("#FFFFFF", "#046380"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "TealMono": {
        "BACKGROUND": "#a8cfdd",
        "TEXT": "#000000",
        "INPUT": "#dfedf2",
        "SCROLL": "#dfedf2",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#183440"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "Default": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": OFFICIAL_PYSIMPLEGUI_BUTTON_COLOR,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "Default1": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": COLOR_SYSTEM_DEFAULT,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "DefaultNoMoreNagging": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": OFFICIAL_PYSIMPLEGUI_BUTTON_COLOR,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "GrayGrayGray": {
        "BACKGROUND": COLOR_SYSTEM_DEFAULT,
        "TEXT": COLOR_SYSTEM_DEFAULT,
        "INPUT": COLOR_SYSTEM_DEFAULT,
        "TEXT_INPUT": COLOR_SYSTEM_DEFAULT,
        "SCROLL": COLOR_SYSTEM_DEFAULT,
        "BUTTON": COLOR_SYSTEM_DEFAULT,
        "PROGRESS": COLOR_SYSTEM_DEFAULT,
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "LightBlue": {
        "BACKGROUND": "#E3F2FD",
        "TEXT": "#000000",
        "INPUT": "#86A8FF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#86A8FF",
        "BUTTON": ("#FFFFFF", "#5079D3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 0,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#FF0266",
        "ACCENT2": "#FF5C93",
        "ACCENT3": "#C5003C",
    },
    "LightGrey": {
        "BACKGROUND": "#FAFAFA",
        "TEXT": "#000000",
        "INPUT": "#004EA1",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#5EA7FF",
        "BUTTON": ("#FFFFFF", "#0079D3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 0,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#FF0266",
        "ACCENT2": "#FF5C93",
        "ACCENT3": "#C5003C",
    },
    "LightGrey1": {
        "BACKGROUND": "#ffffff",
        "TEXT": "#1a1a1b",
        "INPUT": "#dae0e6",
        "TEXT_INPUT": "#222222",
        "SCROLL": "#a5a4a4",
        "BUTTON": ("#FFFFFF", "#0079d3"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#ff5414",
        "ACCENT2": "#33a8ff",
        "ACCENT3": "#dbf0ff",
    },
    "DarkBrown": {
        "BACKGROUND": "#282923",
        "TEXT": "#E7DB74",
        "INPUT": "#393a32",
        "TEXT_INPUT": "#E7C855",
        "SCROLL": "#E7C855",
        "BUTTON": ("#E7C855", "#284B5A"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "ACCENT1": "#c15226",
        "ACCENT2": "#7a4d5f",
        "ACCENT3": "#889743",
    },
    "LightGreen1": {
        "BACKGROUND": "#9FB8AD",
        "TEXT": "#000000",
        "INPUT": "#F7F3EC",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#F7F3EC",
        "BUTTON": ("#FFFFFF", "#475841"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey": {
        "BACKGROUND": "#404040",
        "TEXT": "#FFFFFF",
        "INPUT": "#4D4D4D",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#707070",
        "BUTTON": ("#FFFFFF", "#004F00"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightGreen2": {
        "BACKGROUND": "#B7CECE",
        "TEXT": "#000000",
        "INPUT": "#FDFFF7",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#FDFFF7",
        "BUTTON": ("#FFFFFF", "#658268"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "ACCENT1": "#76506d",
        "ACCENT2": "#5148f1",
        "ACCENT3": "#0a1c84",
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey1": {
        "BACKGROUND": "#404040",
        "TEXT": "#FFFFFF",
        "INPUT": "#FFFFFF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#707070",
        "BUTTON": ("#FFFFFF", "#004F00"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlack": {
        "BACKGROUND": "#000000",
        "TEXT": "#FFFFFF",
        "INPUT": "#4D4D4D",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#707070",
        "BUTTON": ("#000000", "#FFFFFF"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBrown": {
        "BACKGROUND": "#fdf6e3",
        "TEXT": "#268bd1",
        "INPUT": "#eee8d5",
        "TEXT_INPUT": "#6c71c3",
        "SCROLL": "#eee8d5",
        "BUTTON": ("#FFFFFF", "#063542"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBrown1": {
        "BACKGROUND": "#e5dece",
        "TEXT": "#063289",
        "INPUT": "#f9f8f4",
        "TEXT_INPUT": "#242834",
        "SCROLL": "#eee8d5",
        "BUTTON": ("#FFFFFF", "#063289"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlue1": {
        "BACKGROUND": "#242834",
        "TEXT": "#dfe6f8",
        "INPUT": "#97755c",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#a9afbb",
        "BUTTON": ("#FFFFFF", "#063289"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBrown1": {
        "BACKGROUND": "#2c2825",
        "TEXT": "#fdcb52",
        "INPUT": "#705e52",
        "TEXT_INPUT": "#fdcb52",
        "SCROLL": "#705e52",
        "BUTTON": ("#000000", "#fdcb52"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlue2": {
        "BACKGROUND": "#1a2835",
        "TEXT": "#d1ecff",
        "INPUT": "#335267",
        "TEXT_INPUT": "#acc2d0",
        "SCROLL": "#1b6497",
        "BUTTON": ("#000000", "#fafaf8"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBrown2": {
        "BACKGROUND": "#280001",
        "TEXT": "#FFFFFF",
        "INPUT": "#d8d584",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#763e00",
        "BUTTON": ("#000000", "#daad28"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGreen": {
        "BACKGROUND": "#82a459",
        "TEXT": "#000000",
        "INPUT": "#d8d584",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e3ecf3",
        "BUTTON": ("#FFFFFF", "#517239"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBlue1": {
        "BACKGROUND": "#A5CADD",
        "TEXT": "#6E266E",
        "INPUT": "#E0F5FF",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#E0F5FF",
        "BUTTON": ("#FFFFFF", "#303952"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightPurple": {
        "BACKGROUND": "#B0AAC2",
        "TEXT": "#000000",
        "INPUT": "#F2EFE8",
        "SCROLL": "#F2EFE8",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#C2D4D8"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBlue2": {
        "BACKGROUND": "#AAB6D3",
        "TEXT": "#000000",
        "INPUT": "#F1F4FC",
        "SCROLL": "#F1F4FC",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#7186C7"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightGreen3": {
        "BACKGROUND": "#A8C1B4",
        "TEXT": "#000000",
        "INPUT": "#DDE0DE",
        "SCROLL": "#E3E3E3",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#6D9F85"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlue3": {
        "BACKGROUND": "#64778d",
        "TEXT": "#FFFFFF",
        "INPUT": "#f0f3f7",
        "SCROLL": "#A6B2BE",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#283b5b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightGreen4": {
        "BACKGROUND": "#b4ffb4",
        "TEXT": "#000000",
        "INPUT": "#ffff64",
        "SCROLL": "#ffb482",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#ffa0dc"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightGreen5": {
        "BACKGROUND": "#92aa9d",
        "TEXT": "#000000",
        "INPUT": "#fcfff6",
        "SCROLL": "#fcfff6",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#000000", "#d0dbbd"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBrown2": {
        "BACKGROUND": "#a7ad7f",
        "TEXT": "#000000",
        "INPUT": "#e6d3a8",
        "SCROLL": "#e6d3a8",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#5d907d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBrown3": {
        "BACKGROUND": "#efeccb",
        "TEXT": "#012f2f",
        "INPUT": "#e6d3a8",
        "SCROLL": "#e6d3a8",
        "TEXT_INPUT": "#012f2f",
        "BUTTON": ("#FFFFFF", "#046380"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBlue3": {
        "BACKGROUND": "#a8cfdd",
        "TEXT": "#000000",
        "INPUT": "#dfedf2",
        "SCROLL": "#dfedf2",
        "TEXT_INPUT": "#000000",
        "BUTTON": ("#FFFFFF", "#183440"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "LightBrown4": {
        "BACKGROUND": "#d7c79e",
        "TEXT": "#a35638",
        "INPUT": "#9dab86",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#a35638",
        "BUTTON": ("#FFFFFF", "#a35638"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#a35638", "#9dab86", "#e08f62", "#d7c79e"],
    },
    "DarkTeal": {
        "BACKGROUND": "#003f5c",
        "TEXT": "#fb5b5a",
        "INPUT": "#bc4873",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#bc4873",
        "BUTTON": ("#FFFFFF", "#fb5b5a"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#003f5c", "#472b62", "#bc4873", "#fb5b5a"],
    },
    "DarkPurple": {
        "BACKGROUND": "#472b62",
        "TEXT": "#fb5b5a",
        "INPUT": "#bc4873",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#bc4873",
        "BUTTON": ("#FFFFFF", "#472b62"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#003f5c", "#472b62", "#bc4873", "#fb5b5a"],
    },
    "LightGreen6": {
        "BACKGROUND": "#eafbea",
        "TEXT": "#1f6650",
        "INPUT": "#6f9a8d",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#1f6650",
        "BUTTON": ("#FFFFFF", "#1f6650"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#1f6650", "#6f9a8d", "#ea5e5e", "#eafbea"],
    },
    "DarkGrey2": {
        "BACKGROUND": "#2b2b28",
        "TEXT": "#f8f8f8",
        "INPUT": "#f1d6ab",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f1d6ab",
        "BUTTON": ("#2b2b28", "#e3b04b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#2b2b28", "#e3b04b", "#f1d6ab", "#f8f8f8"],
    },
    "LightBrown6": {
        "BACKGROUND": "#f9b282",
        "TEXT": "#8f4426",
        "INPUT": "#de6b35",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#8f4426",
        "BUTTON": ("#FFFFFF", "#8f4426"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#8f4426", "#de6b35", "#64ccda", "#f9b282"],
    },
    "DarkTeal1": {
        "BACKGROUND": "#396362",
        "TEXT": "#ffe7d1",
        "INPUT": "#f6c89f",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f6c89f",
        "BUTTON": ("#ffe7d1", "#4b8e8d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#396362", "#4b8e8d", "#f6c89f", "#ffe7d1"],
    },
    "LightBrown7": {
        "BACKGROUND": "#f6c89f",
        "TEXT": "#396362",
        "INPUT": "#4b8e8d",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#396362",
        "BUTTON": ("#FFFFFF", "#396362"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#396362", "#4b8e8d", "#f6c89f", "#ffe7d1"],
    },
    "DarkPurple1": {
        "BACKGROUND": "#0c093c",
        "TEXT": "#fad6d6",
        "INPUT": "#eea5f6",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#eea5f6",
        "BUTTON": ("#FFFFFF", "#df42d1"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#0c093c", "#df42d1", "#eea5f6", "#fad6d6"],
    },
    "DarkGrey3": {
        "BACKGROUND": "#211717",
        "TEXT": "#dfddc7",
        "INPUT": "#f58b54",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f58b54",
        "BUTTON": ("#dfddc7", "#a34a28"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#211717", "#a34a28", "#f58b54", "#dfddc7"],
    },
    "LightBrown8": {
        "BACKGROUND": "#dfddc7",
        "TEXT": "#211717",
        "INPUT": "#a34a28",
        "TEXT_INPUT": "#dfddc7",
        "SCROLL": "#211717",
        "BUTTON": ("#dfddc7", "#a34a28"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#211717", "#a34a28", "#f58b54", "#dfddc7"],
    },
    "DarkBlue4": {
        "BACKGROUND": "#494ca2",
        "TEXT": "#e3e7f1",
        "INPUT": "#c6cbef",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#c6cbef",
        "BUTTON": ("#FFFFFF", "#8186d5"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#494ca2", "#8186d5", "#c6cbef", "#e3e7f1"],
    },
    "LightBlue4": {
        "BACKGROUND": "#5c94bd",
        "TEXT": "#470938",
        "INPUT": "#1a3e59",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#470938",
        "BUTTON": ("#FFFFFF", "#470938"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#470938", "#1a3e59", "#5c94bd", "#f2d6eb"],
    },
    "DarkTeal2": {
        "BACKGROUND": "#394a6d",
        "TEXT": "#c0ffb3",
        "INPUT": "#52de97",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#52de97",
        "BUTTON": ("#c0ffb3", "#394a6d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#394a6d", "#3c9d9b", "#52de97", "#c0ffb3"],
    },
    "DarkTeal3": {
        "BACKGROUND": "#3c9d9b",
        "TEXT": "#c0ffb3",
        "INPUT": "#52de97",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#52de97",
        "BUTTON": ("#c0ffb3", "#394a6d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#394a6d", "#3c9d9b", "#52de97", "#c0ffb3"],
    },
    "DarkPurple5": {
        "BACKGROUND": "#730068",
        "TEXT": "#f6f078",
        "INPUT": "#01d28e",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#01d28e",
        "BUTTON": ("#f6f078", "#730068"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#730068", "#434982", "#01d28e", "#f6f078"],
    },
    "DarkPurple2": {
        "BACKGROUND": "#202060",
        "TEXT": "#b030b0",
        "INPUT": "#602080",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#602080",
        "BUTTON": ("#FFFFFF", "#202040"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#202040", "#202060", "#602080", "#b030b0"],
    },
    "DarkBlue5": {
        "BACKGROUND": "#000272",
        "TEXT": "#ff6363",
        "INPUT": "#a32f80",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#a32f80",
        "BUTTON": ("#FFFFFF", "#341677"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#000272", "#341677", "#a32f80", "#ff6363"],
    },
    "LightGrey2": {
        "BACKGROUND": "#f6f6f6",
        "TEXT": "#420000",
        "INPUT": "#d4d7dd",
        "TEXT_INPUT": "#420000",
        "SCROLL": "#420000",
        "BUTTON": ("#420000", "#d4d7dd"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#420000", "#d4d7dd", "#eae9e9", "#f6f6f6"],
    },
    "LightGrey3": {
        "BACKGROUND": "#eae9e9",
        "TEXT": "#420000",
        "INPUT": "#d4d7dd",
        "TEXT_INPUT": "#420000",
        "SCROLL": "#420000",
        "BUTTON": ("#420000", "#d4d7dd"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#420000", "#d4d7dd", "#eae9e9", "#f6f6f6"],
    },
    "DarkBlue6": {
        "BACKGROUND": "#01024e",
        "TEXT": "#ff6464",
        "INPUT": "#8b4367",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#8b4367",
        "BUTTON": ("#FFFFFF", "#543864"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#01024e", "#543864", "#8b4367", "#ff6464"],
    },
    "DarkBlue7": {
        "BACKGROUND": "#241663",
        "TEXT": "#eae7af",
        "INPUT": "#a72693",
        "TEXT_INPUT": "#eae7af",
        "SCROLL": "#a72693",
        "BUTTON": ("#eae7af", "#160f30"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#160f30", "#241663", "#a72693", "#eae7af"],
    },
    "LightBrown9": {
        "BACKGROUND": "#f6d365",
        "TEXT": "#3a1f5d",
        "INPUT": "#c83660",
        "TEXT_INPUT": "#f6d365",
        "SCROLL": "#3a1f5d",
        "BUTTON": ("#f6d365", "#c83660"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3a1f5d", "#c83660", "#e15249", "#f6d365"],
    },
    "DarkPurple3": {
        "BACKGROUND": "#6e2142",
        "TEXT": "#ffd692",
        "INPUT": "#e16363",
        "TEXT_INPUT": "#ffd692",
        "SCROLL": "#e16363",
        "BUTTON": ("#ffd692", "#943855"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#6e2142", "#943855", "#e16363", "#ffd692"],
    },
    "LightBrown10": {
        "BACKGROUND": "#ffd692",
        "TEXT": "#6e2142",
        "INPUT": "#943855",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#6e2142",
        "BUTTON": ("#FFFFFF", "#6e2142"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#6e2142", "#943855", "#e16363", "#ffd692"],
    },
    "DarkPurple4": {
        "BACKGROUND": "#200f21",
        "TEXT": "#f638dc",
        "INPUT": "#5a3d5c",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#5a3d5c",
        "BUTTON": ("#FFFFFF", "#382039"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#200f21", "#382039", "#5a3d5c", "#f638dc"],
    },
    "LightBlue5": {
        "BACKGROUND": "#b2fcff",
        "TEXT": "#3e64ff",
        "INPUT": "#5edfff",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#3e64ff",
        "BUTTON": ("#FFFFFF", "#3e64ff"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3e64ff", "#5edfff", "#b2fcff", "#ecfcff"],
    },
    "DarkTeal4": {
        "BACKGROUND": "#464159",
        "TEXT": "#c7f0db",
        "INPUT": "#8bbabb",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#8bbabb",
        "BUTTON": ("#FFFFFF", "#6c7b95"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#464159", "#6c7b95", "#8bbabb", "#c7f0db"],
    },
    "LightTeal": {
        "BACKGROUND": "#c7f0db",
        "TEXT": "#464159",
        "INPUT": "#6c7b95",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#464159",
        "BUTTON": ("#FFFFFF", "#464159"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#464159", "#6c7b95", "#8bbabb", "#c7f0db"],
    },
    "DarkTeal5": {
        "BACKGROUND": "#8bbabb",
        "TEXT": "#464159",
        "INPUT": "#6c7b95",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#464159",
        "BUTTON": ("#c7f0db", "#6c7b95"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#464159", "#6c7b95", "#8bbabb", "#c7f0db"],
    },
    "LightGrey4": {
        "BACKGROUND": "#faf5ef",
        "TEXT": "#672f2f",
        "INPUT": "#99b19c",
        "TEXT_INPUT": "#672f2f",
        "SCROLL": "#672f2f",
        "BUTTON": ("#672f2f", "#99b19c"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#672f2f", "#99b19c", "#d7d1c9", "#faf5ef"],
    },
    "LightGreen7": {
        "BACKGROUND": "#99b19c",
        "TEXT": "#faf5ef",
        "INPUT": "#d7d1c9",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#d7d1c9",
        "BUTTON": ("#FFFFFF", "#99b19c"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#672f2f", "#99b19c", "#d7d1c9", "#faf5ef"],
    },
    "LightGrey5": {
        "BACKGROUND": "#d7d1c9",
        "TEXT": "#672f2f",
        "INPUT": "#99b19c",
        "TEXT_INPUT": "#672f2f",
        "SCROLL": "#672f2f",
        "BUTTON": ("#FFFFFF", "#672f2f"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#672f2f", "#99b19c", "#d7d1c9", "#faf5ef"],
    },
    "DarkBrown3": {
        "BACKGROUND": "#a0855b",
        "TEXT": "#f9f6f2",
        "INPUT": "#f1d6ab",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f1d6ab",
        "BUTTON": ("#FFFFFF", "#38470b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#38470b", "#a0855b", "#f1d6ab", "#f9f6f2"],
    },
    "LightBrown11": {
        "BACKGROUND": "#f1d6ab",
        "TEXT": "#38470b",
        "INPUT": "#a0855b",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#38470b",
        "BUTTON": ("#f9f6f2", "#a0855b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#38470b", "#a0855b", "#f1d6ab", "#f9f6f2"],
    },
    "DarkRed": {
        "BACKGROUND": "#83142c",
        "TEXT": "#f9d276",
        "INPUT": "#ad1d45",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#ad1d45",
        "BUTTON": ("#f9d276", "#ad1d45"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#44000d", "#83142c", "#ad1d45", "#f9d276"],
    },
    "DarkTeal6": {
        "BACKGROUND": "#204969",
        "TEXT": "#fff7f7",
        "INPUT": "#dadada",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#dadada",
        "BUTTON": ("#000000", "#fff7f7"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#204969", "#08ffc8", "#dadada", "#fff7f7"],
    },
    "DarkBrown4": {
        "BACKGROUND": "#252525",
        "TEXT": "#ff0000",
        "INPUT": "#af0404",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#af0404",
        "BUTTON": ("#FFFFFF", "#252525"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#252525", "#414141", "#af0404", "#ff0000"],
    },
    "LightYellow": {
        "BACKGROUND": "#f4ff61",
        "TEXT": "#27aa80",
        "INPUT": "#32ff6a",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#27aa80",
        "BUTTON": ("#f4ff61", "#27aa80"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#27aa80", "#32ff6a", "#a8ff3e", "#f4ff61"],
    },
    "DarkGreen1": {
        "BACKGROUND": "#2b580c",
        "TEXT": "#fdef96",
        "INPUT": "#f7b71d",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f7b71d",
        "BUTTON": ("#fdef96", "#2b580c"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#2b580c", "#afa939", "#f7b71d", "#fdef96"],
    },
    "LightGreen8": {
        "BACKGROUND": "#c8dad3",
        "TEXT": "#63707e",
        "INPUT": "#93b5b3",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#63707e",
        "BUTTON": ("#FFFFFF", "#63707e"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#63707e", "#93b5b3", "#c8dad3", "#f2f6f5"],
    },
    "DarkTeal7": {
        "BACKGROUND": "#248ea9",
        "TEXT": "#fafdcb",
        "INPUT": "#aee7e8",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#aee7e8",
        "BUTTON": ("#000000", "#fafdcb"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#248ea9", "#28c3d4", "#aee7e8", "#fafdcb"],
    },
    "DarkBlue8": {
        "BACKGROUND": "#454d66",
        "TEXT": "#d9d872",
        "INPUT": "#58b368",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#58b368",
        "BUTTON": ("#000000", "#009975"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#009975", "#454d66", "#58b368", "#d9d872"],
    },
    "DarkBlue9": {
        "BACKGROUND": "#263859",
        "TEXT": "#ff6768",
        "INPUT": "#6b778d",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#6b778d",
        "BUTTON": ("#ff6768", "#263859"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#17223b", "#263859", "#6b778d", "#ff6768"],
    },
    "DarkBlue10": {
        "BACKGROUND": "#0028ff",
        "TEXT": "#f1f4df",
        "INPUT": "#10eaf0",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#10eaf0",
        "BUTTON": ("#f1f4df", "#24009c"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#24009c", "#0028ff", "#10eaf0", "#f1f4df"],
    },
    "DarkBlue11": {
        "BACKGROUND": "#6384b3",
        "TEXT": "#e6f0b6",
        "INPUT": "#b8e9c0",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#b8e9c0",
        "BUTTON": ("#e6f0b6", "#684949"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#684949", "#6384b3", "#b8e9c0", "#e6f0b6"],
    },
    "DarkTeal8": {
        "BACKGROUND": "#71a0a5",
        "TEXT": "#212121",
        "INPUT": "#665c84",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#212121",
        "BUTTON": ("#fab95b", "#665c84"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#212121", "#665c84", "#71a0a5", "#fab95b"],
    },
    "DarkRed1": {
        "BACKGROUND": "#c10000",
        "TEXT": "#eeeeee",
        "INPUT": "#dedede",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#dedede",
        "BUTTON": ("#c10000", "#eeeeee"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#c10000", "#ff4949", "#dedede", "#eeeeee"],
    },
    "LightBrown5": {
        "BACKGROUND": "#fff591",
        "TEXT": "#e41749",
        "INPUT": "#f5587b",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e41749",
        "BUTTON": ("#fff591", "#e41749"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#e41749", "#f5587b", "#ff8a5c", "#fff591"],
    },
    "LightGreen9": {
        "BACKGROUND": "#f1edb3",
        "TEXT": "#3b503d",
        "INPUT": "#4a746e",
        "TEXT_INPUT": "#f1edb3",
        "SCROLL": "#3b503d",
        "BUTTON": ("#f1edb3", "#3b503d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3b503d", "#4a746e", "#c8cf94", "#f1edb3"],
        "DESCRIPTION": ["Green", "Turquoise", "Yellow"],
    },
    "DarkGreen2": {
        "BACKGROUND": "#3b503d",
        "TEXT": "#f1edb3",
        "INPUT": "#c8cf94",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#c8cf94",
        "BUTTON": ("#f1edb3", "#3b503d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3b503d", "#4a746e", "#c8cf94", "#f1edb3"],
        "DESCRIPTION": ["Green", "Turquoise", "Yellow"],
    },
    "LightGray1": {
        "BACKGROUND": "#f2f2f2",
        "TEXT": "#222831",
        "INPUT": "#393e46",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#222831",
        "BUTTON": ("#f2f2f2", "#222831"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#222831", "#393e46", "#f96d00", "#f2f2f2"],
        "DESCRIPTION": ["#000000", "Grey", "Orange", "Grey", "Autumn"],
    },
    "DarkGrey4": {
        "BACKGROUND": "#52524e",
        "TEXT": "#e9e9e5",
        "INPUT": "#d4d6c8",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#d4d6c8",
        "BUTTON": ("#FFFFFF", "#9a9b94"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#52524e", "#9a9b94", "#d4d6c8", "#e9e9e5"],
        "DESCRIPTION": ["Grey", "Pastel", "Winter"],
    },
    "DarkBlue12": {
        "BACKGROUND": "#324e7b",
        "TEXT": "#f8f8f8",
        "INPUT": "#86a6df",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#86a6df",
        "BUTTON": ("#FFFFFF", "#5068a9"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#324e7b", "#5068a9", "#86a6df", "#f8f8f8"],
        "DESCRIPTION": ["Blue", "Grey", "Cold", "Winter"],
    },
    "DarkPurple6": {
        "BACKGROUND": "#070739",
        "TEXT": "#e1e099",
        "INPUT": "#c327ab",
        "TEXT_INPUT": "#e1e099",
        "SCROLL": "#c327ab",
        "BUTTON": ("#e1e099", "#521477"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#070739", "#521477", "#c327ab", "#e1e099"],
        "DESCRIPTION": ["#000000", "Purple", "Yellow", "Dark"],
    },
    "DarkPurple7": {
        "BACKGROUND": "#191930",
        "TEXT": "#B1B7C5",
        "INPUT": "#232B5C",
        "TEXT_INPUT": "#D0E3E7",
        "SCROLL": "#B1B7C5",
        "BUTTON": ("#272D38", "#B1B7C5"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBlue13": {
        "BACKGROUND": "#203562",
        "TEXT": "#e3e8f8",
        "INPUT": "#c0c5cd",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#c0c5cd",
        "BUTTON": ("#FFFFFF", "#3e588f"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#203562", "#3e588f", "#c0c5cd", "#e3e8f8"],
        "DESCRIPTION": ["Blue", "Grey", "Wedding", "Cold"],
    },
    "DarkBrown5": {
        "BACKGROUND": "#3c1b1f",
        "TEXT": "#f6e1b5",
        "INPUT": "#e2bf81",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e2bf81",
        "BUTTON": ("#3c1b1f", "#f6e1b5"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3c1b1f", "#b21e4b", "#e2bf81", "#f6e1b5"],
        "DESCRIPTION": ["Brown", "Red", "Yellow", "Warm"],
    },
    "DarkGreen3": {
        "BACKGROUND": "#062121",
        "TEXT": "#eeeeee",
        "INPUT": "#e4dcad",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e4dcad",
        "BUTTON": ("#eeeeee", "#181810"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#062121", "#181810", "#e4dcad", "#eeeeee"],
        "DESCRIPTION": ["#000000", "#000000", "Brown", "Grey"],
    },
    "DarkBlack1": {
        "BACKGROUND": "#181810",
        "TEXT": "#eeeeee",
        "INPUT": "#e4dcad",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e4dcad",
        "BUTTON": ("#FFFFFF", "#062121"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#062121", "#181810", "#e4dcad", "#eeeeee"],
        "DESCRIPTION": ["#000000", "#000000", "Brown", "Grey"],
    },
    "DarkGrey5": {
        "BACKGROUND": "#343434",
        "TEXT": "#f3f3f3",
        "INPUT": "#e9dcbe",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e9dcbe",
        "BUTTON": ("#FFFFFF", "#8e8b82"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#343434", "#8e8b82", "#e9dcbe", "#f3f3f3"],
        "DESCRIPTION": ["Grey", "Brown"],
    },
    "LightBrown12": {
        "BACKGROUND": "#8e8b82",
        "TEXT": "#f3f3f3",
        "INPUT": "#e9dcbe",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e9dcbe",
        "BUTTON": ("#f3f3f3", "#8e8b82"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#343434", "#8e8b82", "#e9dcbe", "#f3f3f3"],
        "DESCRIPTION": ["Grey", "Brown"],
    },
    "DarkTeal9": {
        "BACKGROUND": "#13445a",
        "TEXT": "#fef4e8",
        "INPUT": "#446878",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#446878",
        "BUTTON": ("#fef4e8", "#446878"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#13445a", "#970747", "#446878", "#fef4e8"],
        "DESCRIPTION": ["Red", "Grey", "Blue", "Wedding", "Retro"],
    },
    "DarkBlue14": {
        "BACKGROUND": "#21273d",
        "TEXT": "#f1f6f8",
        "INPUT": "#b9d4f1",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#b9d4f1",
        "BUTTON": ("#FFFFFF", "#6a759b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#21273d", "#6a759b", "#b9d4f1", "#f1f6f8"],
        "DESCRIPTION": ["Blue", "#000000", "Grey", "Cold", "Winter"],
    },
    "LightBlue6": {
        "BACKGROUND": "#f1f6f8",
        "TEXT": "#21273d",
        "INPUT": "#6a759b",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#21273d",
        "BUTTON": ("#f1f6f8", "#6a759b"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#21273d", "#6a759b", "#b9d4f1", "#f1f6f8"],
        "DESCRIPTION": ["Blue", "#000000", "Grey", "Cold", "Winter"],
    },
    "DarkGreen4": {
        "BACKGROUND": "#044343",
        "TEXT": "#e4e4e4",
        "INPUT": "#045757",
        "TEXT_INPUT": "#e4e4e4",
        "SCROLL": "#045757",
        "BUTTON": ("#e4e4e4", "#045757"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#222222", "#044343", "#045757", "#e4e4e4"],
        "DESCRIPTION": ["#000000", "Turquoise", "Grey", "Dark"],
    },
    "DarkGreen5": {
        "BACKGROUND": "#1b4b36",
        "TEXT": "#e0e7f1",
        "INPUT": "#aebd77",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#aebd77",
        "BUTTON": ("#FFFFFF", "#538f6a"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#1b4b36", "#538f6a", "#aebd77", "#e0e7f1"],
        "DESCRIPTION": ["Green", "Grey"],
    },
    "DarkTeal10": {
        "BACKGROUND": "#0d3446",
        "TEXT": "#d8dfe2",
        "INPUT": "#71adb5",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#71adb5",
        "BUTTON": ("#FFFFFF", "#176d81"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#0d3446", "#176d81", "#71adb5", "#d8dfe2"],
        "DESCRIPTION": ["Grey", "Turquoise", "Winter", "Cold"],
    },
    "DarkGrey6": {
        "BACKGROUND": "#3e3e3e",
        "TEXT": "#ededed",
        "INPUT": "#68868c",
        "TEXT_INPUT": "#ededed",
        "SCROLL": "#68868c",
        "BUTTON": ("#FFFFFF", "#405559"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3e3e3e", "#405559", "#68868c", "#ededed"],
        "DESCRIPTION": ["Grey", "Turquoise", "Winter"],
    },
    "DarkTeal11": {
        "BACKGROUND": "#405559",
        "TEXT": "#ededed",
        "INPUT": "#68868c",
        "TEXT_INPUT": "#ededed",
        "SCROLL": "#68868c",
        "BUTTON": ("#ededed", "#68868c"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#3e3e3e", "#405559", "#68868c", "#ededed"],
        "DESCRIPTION": ["Grey", "Turquoise", "Winter"],
    },
    "LightBlue7": {
        "BACKGROUND": "#9ed0e0",
        "TEXT": "#19483f",
        "INPUT": "#5c868e",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#19483f",
        "BUTTON": ("#FFFFFF", "#19483f"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#19483f", "#5c868e", "#ff6a38", "#9ed0e0"],
        "DESCRIPTION": ["Orange", "Blue", "Turquoise"],
    },
    "LightGreen10": {
        "BACKGROUND": "#d8ebb5",
        "TEXT": "#205d67",
        "INPUT": "#639a67",
        "TEXT_INPUT": "#FFFFFF",
        "SCROLL": "#205d67",
        "BUTTON": ("#d8ebb5", "#205d67"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#205d67", "#639a67", "#d9bf77", "#d8ebb5"],
        "DESCRIPTION": ["Blue", "Green", "Brown", "Vintage"],
    },
    "DarkBlue15": {
        "BACKGROUND": "#151680",
        "TEXT": "#f1fea4",
        "INPUT": "#375fc0",
        "TEXT_INPUT": "#f1fea4",
        "SCROLL": "#375fc0",
        "BUTTON": ("#f1fea4", "#1c44ac"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#151680", "#1c44ac", "#375fc0", "#f1fea4"],
        "DESCRIPTION": ["Blue", "Yellow", "Cold"],
    },
    "DarkBlue16": {
        "BACKGROUND": "#1c44ac",
        "TEXT": "#f1fea4",
        "INPUT": "#375fc0",
        "TEXT_INPUT": "#f1fea4",
        "SCROLL": "#375fc0",
        "BUTTON": ("#f1fea4", "#151680"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#151680", "#1c44ac", "#375fc0", "#f1fea4"],
        "DESCRIPTION": ["Blue", "Yellow", "Cold"],
    },
    "DarkTeal12": {
        "BACKGROUND": "#004a7c",
        "TEXT": "#fafafa",
        "INPUT": "#e8f1f5",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#e8f1f5",
        "BUTTON": ("#fafafa", "#005691"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#004a7c", "#005691", "#e8f1f5", "#fafafa"],
        "DESCRIPTION": ["Grey", "Blue", "Cold", "Winter"],
    },
    "LightBrown13": {
        "BACKGROUND": "#ebf5ee",
        "TEXT": "#921224",
        "INPUT": "#bdc6b8",
        "TEXT_INPUT": "#921224",
        "SCROLL": "#921224",
        "BUTTON": ("#FFFFFF", "#921224"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#921224", "#bdc6b8", "#bce0da", "#ebf5ee"],
        "DESCRIPTION": ["Red", "Blue", "Grey", "Vintage", "Wedding"],
    },
    "DarkBlue17": {
        "BACKGROUND": "#21294c",
        "TEXT": "#f9f2d7",
        "INPUT": "#f2dea8",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#f2dea8",
        "BUTTON": ("#f9f2d7", "#141829"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#141829", "#21294c", "#f2dea8", "#f9f2d7"],
        "DESCRIPTION": ["#000000", "Blue", "Yellow"],
    },
    "DarkBlue18": {
        "BACKGROUND": "#0c1825",
        "TEXT": "#d1d7dd",
        "INPUT": "#001c35",
        "TEXT_INPUT": "#d1d7dd",
        "SCROLL": "#00438e",
        "BUTTON": ("#75b7ff", "#001c35"),
        "PROGRESS": ("#0074ff", "#75b7ff"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#141829", "#21294c", "#f2dea8", "#f9f2d7"],
        "DESCRIPTION": ["#000000", "Blue", "Yellow"],
    },
    "DarkBrown6": {
        "BACKGROUND": "#785e4d",
        "TEXT": "#f2eee3",
        "INPUT": "#baaf92",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#baaf92",
        "BUTTON": ("#FFFFFF", "#785e4d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#785e4d", "#ff8426", "#baaf92", "#f2eee3"],
        "DESCRIPTION": ["Grey", "Brown", "Orange", "Autumn"],
    },
    "DarkGreen6": {
        "BACKGROUND": "#5c715e",
        "TEXT": "#f2f9f1",
        "INPUT": "#ddeedf",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#ddeedf",
        "BUTTON": ("#f2f9f1", "#5c715e"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#5c715e", "#b6cdbd", "#ddeedf", "#f2f9f1"],
        "DESCRIPTION": ["Grey", "Green", "Vintage"],
    },
    "DarkGreen7": {
        "BACKGROUND": "#0C231E",
        "TEXT": "#efbe1c",
        "INPUT": "#153C33",
        "TEXT_INPUT": "#efbe1c",
        "SCROLL": "#153C33",
        "BUTTON": ("#efbe1c", "#153C33"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey7": {
        "BACKGROUND": "#4b586e",
        "TEXT": "#dddddd",
        "INPUT": "#574e6d",
        "TEXT_INPUT": "#dddddd",
        "SCROLL": "#574e6d",
        "BUTTON": ("#dddddd", "#43405d"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#43405d", "#4b586e", "#574e6d", "#dddddd"],
        "DESCRIPTION": ["Grey", "Winter", "Cold"],
    },
    "DarkRed2": {
        "BACKGROUND": "#ab1212",
        "TEXT": "#f6e4b5",
        "INPUT": "#cd3131",
        "TEXT_INPUT": "#f6e4b5",
        "SCROLL": "#cd3131",
        "BUTTON": ("#f6e4b5", "#ab1212"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#ab1212", "#1fad9f", "#cd3131", "#f6e4b5"],
        "DESCRIPTION": ["Turquoise", "Red", "Yellow"],
    },
    "LightGrey6": {
        "BACKGROUND": "#e3e3e3",
        "TEXT": "#233142",
        "INPUT": "#455d7a",
        "TEXT_INPUT": "#e3e3e3",
        "SCROLL": "#233142",
        "BUTTON": ("#e3e3e3", "#455d7a"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
        "COLOR_LIST": ["#233142", "#455d7a", "#f95959", "#e3e3e3"],
        "DESCRIPTION": ["#000000", "Blue", "Red", "Grey"],
    },
    "HotDogStand": {
        "BACKGROUND": "red",
        "TEXT": "yellow",
        "INPUT": "yellow",
        "TEXT_INPUT": "#000000",
        "SCROLL": "yellow",
        "BUTTON": ("red", "yellow"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey8": {
        "BACKGROUND": "#19232D",
        "TEXT": "#ffffff",
        "INPUT": "#32414B",
        "TEXT_INPUT": "#ffffff",
        "SCROLL": "#505F69",
        "BUTTON": ("#ffffff", "#32414B"),
        "PROGRESS": ("#505F69", "#32414B"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey9": {
        "BACKGROUND": "#36393F",
        "TEXT": "#DCDDDE",
        "INPUT": "#40444B",
        "TEXT_INPUT": "#ffffff",
        "SCROLL": "#202225",
        "BUTTON": ("#202225", "#B9BBBE"),
        "PROGRESS": ("#202225", "#40444B"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey10": {
        "BACKGROUND": "#1c1e23",
        "TEXT": "#cccdcf",
        "INPUT": "#272a31",
        "TEXT_INPUT": "#8b9fde",
        "SCROLL": "#313641",
        "BUTTON": ("#f5f5f6", "#2e3d5a"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey11": {
        "BACKGROUND": "#1c1e23",
        "TEXT": "#cccdcf",
        "INPUT": "#313641",
        "TEXT_INPUT": "#cccdcf",
        "SCROLL": "#313641",
        "BUTTON": ("#f5f5f6", "#313641"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey12": {
        "BACKGROUND": "#1c1e23",
        "TEXT": "#8b9fde",
        "INPUT": "#313641",
        "TEXT_INPUT": "#8b9fde",
        "SCROLL": "#313641",
        "BUTTON": ("#cccdcf", "#2e3d5a"),
        "PROGRESS": DEFAULT_PROGRESS_BAR_COMPUTE,
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey13": {
        "BACKGROUND": "#1c1e23",
        "TEXT": "#cccdcf",
        "INPUT": "#272a31",
        "TEXT_INPUT": "#cccdcf",
        "SCROLL": "#313641",
        "BUTTON": ("#8b9fde", "#313641"),
        "PROGRESS": ("#cccdcf", "#272a31"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey14": {
        "BACKGROUND": "#24292e",
        "TEXT": "#fafbfc",
        "INPUT": "#1d2125",
        "TEXT_INPUT": "#fafbfc",
        "SCROLL": "#1d2125",
        "BUTTON": ("#fafbfc", "#155398"),
        "PROGRESS": ("#155398", "#1d2125"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey15": {
        "BACKGROUND": "#121212",
        "TEXT": "#dddddd",
        "INPUT": "#1e1e1e",
        "TEXT_INPUT": "#69b1ef",
        "SCROLL": "#272727",
        "BUTTON": ("#69b1ef", "#2e2e2e"),
        "PROGRESS": ("#69b1ef", "#2e2e2e"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkGrey16": {
        "BACKGROUND": "#353535",
        "TEXT": "#ffffff",
        "INPUT": "#191919",
        "TEXT_INPUT": "#ffffff",
        "SCROLL": "#454545",
        "BUTTON": ("#ffffff", "#454545"),
        "PROGRESS": ("#757575", "#454545"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "DarkBrown7": {
        "BACKGROUND": "#2c2417",
        "TEXT": "#baa379",
        "INPUT": "#baa379",
        "TEXT_INPUT": "#000000",
        "SCROLL": "#392e1c",
        "BUTTON": ("#000000", "#baa379"),
        "PROGRESS": ("#baa379", "#453923"),
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "Python": {
        "BACKGROUND": "#3d7aab",
        "TEXT": "#ffde56",
        "INPUT": "#295273",
        "TEXT_INPUT": "#ffde56",
        "SCROLL": "#295273",
        "BUTTON": ("#ffde56", "#295273"),
        "PROGRESS": ("#ffde56", "#295273"),
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "PythonPlus": {
        "BACKGROUND": "#001d3c",
        "TEXT": "#ffffff",
        "INPUT": "#015bbb",
        "TEXT_INPUT": "#fed500",
        "SCROLL": "#015bbb",
        "BUTTON": ("#fed500", "#015bbb"),
        "PROGRESS": ("#015bbb", "#fed500"),
        "BORDER": 1,
        "SLIDER_DEPTH": 1,
        "PROGRESS_DEPTH": 0,
    },
    "NeonBlue1": {
        "BACKGROUND": "#000000",
        "TEXT": "#ffffff",
        "INPUT": "#000000",
        "TEXT_INPUT": "#33ccff",
        "SCROLL": "#33ccff",
        "BUTTON": ("#33ccff", "#000000"),
        "PROGRESS": ("#33ccff", "#ffffff"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "NeonGreen1": {
        "BACKGROUND": "#000000",
        "TEXT": "#ffffff",
        "INPUT": "#000000",
        "TEXT_INPUT": "#96ff7b",
        "SCROLL": "#96ff7b",
        "BUTTON": ("#96ff7b", "#000000"),
        "PROGRESS": ("#96ff7b", "#ffffff"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
    "NeonYellow1": {
        "BACKGROUND": "#000000",
        "TEXT": "#ffffff",
        "INPUT": "#000000",
        "TEXT_INPUT": "#ffcf40",
        "SCROLL": "#ffcf40",
        "BUTTON": ("#ffcf40", "#000000"),
        "PROGRESS": ("#ffcf40", "#ffffff"),
        "BORDER": 1,
        "SLIDER_DEPTH": 0,
        "PROGRESS_DEPTH": 0,
    },
}


def list_of_look_and_feel_values():
    """
    Get a list of the valid values to pass into your call to change_look_and_feel

    :return: list of valid string values
    :rtype:  List[str]
    """

    return sorted(list(LOOK_AND_FEEL_TABLE.keys()))


def theme(new_theme=None):
    """
    Sets / Gets the current Theme.  If none is specified then returns the current theme.
    This call replaces the ChangeLookAndFeel / change_look_and_feel call which only sets the theme.

    :param new_theme: the new theme name to use
    :type new_theme:  (str)
    :return:          the currently selected theme
    :rtype:           (str)
    """
    global TRANSPARENT_BUTTON

    if new_theme is not None:
        change_look_and_feel(new_theme)
        TRANSPARENT_BUTTON = (theme_background_color(), theme_background_color())
    return CURRENT_LOOK_AND_FEEL


def theme_background_color(color=None):
    """
    Sets/Returns the background color currently in use
    Used for Windows and containers (Column, Frame, Tab) and tables

    :param color: new background color to use (optional)
    :type color:  (str)
    :return:      color string of the background color currently in use
    :rtype:       (str)
    """
    if color is not None:
        set_options(background_color=color)
    return DEFAULT_BACKGROUND_COLOR


# This "constant" is misleading but rather than remove and break programs, will try this method instead
TRANSPARENT_BUTTON = (
    theme_background_color(),
    theme_background_color(),
)  # replaces an older version that had hardcoded numbers


def theme_element_background_color(color=None):
    """
    Sets/Returns the background color currently in use for all elements except containers

    :return: (str) - color string of the element background color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(element_background_color=color)
    return DEFAULT_ELEMENT_BACKGROUND_COLOR


def theme_text_color(color=None):
    """
    Sets/Returns the text color currently in use

    :return: (str) - color string of the text color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(text_color=color)
    return DEFAULT_TEXT_COLOR


def theme_text_element_background_color(color=None):
    """
    Sets/Returns the background color for text elements

    :return: (str) - color string of the text background color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(text_element_background_color=color)
    return DEFAULT_TEXT_ELEMENT_BACKGROUND_COLOR


def theme_input_background_color(color=None):
    """
    Sets/Returns the input element background color currently in use

    :return: (str) - color string of the input element background color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(input_elements_background_color=color)
    return DEFAULT_INPUT_ELEMENTS_COLOR


def theme_input_text_color(color=None):
    """
    Sets/Returns the input element entry color (not the text but the thing that's displaying the text)

    :return: (str) - color string of the input element color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(input_text_color=color)
    return DEFAULT_INPUT_TEXT_COLOR


def theme_button_color(color=None):
    """
    Sets/Returns the button color currently in use

    :return: (str, str) - TUPLE with color strings of the button color currently in use (button text color, button background color)
    :rtype:  (str, str)
    """
    if color is not None:
        if color == COLOR_SYSTEM_DEFAULT:
            color_tuple = (COLOR_SYSTEM_DEFAULT, COLOR_SYSTEM_DEFAULT)
        else:
            color_tuple = button_color_to_tuple(color, (None, None))
        if color_tuple == (None, None):
            if not SUPPRESS_ERROR_POPUPS:
                popup_error("theme_button_color - bad color string passed in", color)
            else:
                print(
                    "** Badly formatted button color... not a tuple nor string **",
                    color,
                )
            set_options(button_color=color)  # go ahead and try with their string
        else:
            set_options(button_color=color_tuple)
    return DEFAULT_BUTTON_COLOR


def theme_button_color_background():
    """
    Returns the button color background currently in use. Note this function simple calls the theme_button_color
    function and splits apart the tuple

    :return: color string of the button color background currently in use
    :rtype:  (str)
    """
    return theme_button_color()[1]


def theme_button_color_text():
    """
    Returns the button color text currently in use.  Note this function simple calls the theme_button_color
    function and splits apart the tuple

    :return: color string of the button color text currently in use
    :rtype:  (str)
    """
    return theme_button_color()[0]


def theme_progress_bar_color(color=None):
    """
    Sets/Returns the progress bar colors by the current color theme

    :return: (str, str) - TUPLE with color strings of the ProgressBar color currently in use(button text color, button background color)
    :rtype:  (str, str)
    """
    if color is not None:
        set_options(progress_meter_color=color)
    return DEFAULT_PROGRESS_BAR_COLOR


def theme_slider_color(color=None):
    """
    Sets/Returns the slider color (used for sliders)

    :return: color string of the slider color currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(scrollbar_color=color)
    return DEFAULT_SCROLLBAR_COLOR


def theme_border_width(border_width=None):
    """
    Sets/Returns the border width currently in use
    Used by non ttk elements at the moment

    :return: border width currently in use
    :rtype:  (int)
    """
    if border_width is not None:
        set_options(border_width=border_width)
    return DEFAULT_BORDER_WIDTH


def theme_slider_border_width(border_width=None):
    """
    Sets/Returns the slider border width currently in use

    :return: border width currently in use for sliders
    :rtype:  (int)
    """
    if border_width is not None:
        set_options(slider_border_width=border_width)
    return DEFAULT_SLIDER_BORDER_WIDTH


def theme_progress_bar_border_width(border_width=None):
    """
    Sets/Returns the progress meter border width currently in use

    :return: border width currently in use for progress meters
    :rtype:  (int)
    """
    if border_width is not None:
        set_options(progress_meter_border_depth=border_width)
    return DEFAULT_PROGRESS_BAR_BORDER_WIDTH


def theme_element_text_color(color=None):
    """
    Sets/Returns the text color used by elements that have text as part of their display (Tables, Trees and Sliders)

    :return: color string currently in use
    :rtype:  (str)
    """
    if color is not None:
        set_options(element_text_color=color)
    return DEFAULT_ELEMENT_TEXT_COLOR


def theme_list():
    """
    Returns a sorted list of the currently available color themes

    :return: A sorted list of the currently available color themes
    :rtype:  List[str]
    """
    return list_of_look_and_feel_values()


def theme_add_new(new_theme_name, new_theme_dict):
    """
    Add a new theme to the dictionary of themes

    :param new_theme_name: text to display in element
    :type new_theme_name:  (str)
    :param new_theme_dict: text to display in element
    :type new_theme_dict:  (dict)
    """
    global LOOK_AND_FEEL_TABLE
    try:
        LOOK_AND_FEEL_TABLE[new_theme_name] = new_theme_dict
    except Exception as e:
        print("Exception during adding new theme {}".format(e))


def theme_use_custom_titlebar():
    """
    Returns True if a custom titlebar will be / should be used.
    The setting is in the Global Settings window and can be overwridden
    using set_options call

    :return:        True if a custom titlebar / custom menubar should be used
    :rtype:         (bool)
    """
    if USE_CUSTOM_TITLEBAR is False:
        return False

    return USE_CUSTOM_TITLEBAR or pysimplegui_user_settings.get(
        "-custom titlebar-", False
    )


def theme_global(new_theme=None):
    """
    Sets / Gets the global PySimpleGUI Theme.  If none is specified then returns the global theme from user settings.
    Note the theme must be a standard, built-in PySimpleGUI theme... not a user-created theme.

    :param new_theme: the new theme name to use
    :type new_theme:  (str)
    :return:          the currently selected theme
    :rtype:           (str)
    """
    if new_theme is not None:
        if new_theme not in theme_list():
            popup_error_with_traceback(
                "Cannot use custom themes with theme_global call",
                "Your request to use theme {} cannot be performed.".format(new_theme),
                "The PySimpleGUI Global User Settings are meant for PySimpleGUI standard items, not user config items",
                "You can use any of the many built-in themes instead or use your own UserSettings file to store your custom theme",
            )
            return pysimplegui_user_settings.get("-theme-", CURRENT_LOOK_AND_FEEL)
        pysimplegui_user_settings.set("-theme-", new_theme)
        theme(new_theme)
        return new_theme
    else:
        return pysimplegui_user_settings.get("-theme-", CURRENT_LOOK_AND_FEEL)


def theme_previewer(
        columns=12,
        scrollable=False,
        scroll_area_size=(None, None),
        search_string=None,
        location=(None, None),
):
    """
    Displays a "Quick Reference Window" showing all of the different Look and Feel settings that are available.
    They are sorted alphabetically.  The legacy color names are mixed in, but otherwise they are sorted into Dark and Light halves

    :param columns:          The number of themes to display per row
    :type columns:           int
    :param scrollable:       If True then scrollbars will be added
    :type scrollable:        bool
    :param scroll_area_size: Size of the scrollable area (The Column Element used to make scrollable)
    :type scroll_area_size:  (int, int)
    :param search_string:    If specified then only themes containing this string will be shown
    :type search_string:     str
    :param location:         Location on the screen to place the window. Defaults to the center like all windows
    :type location:          (int, int)
    """

    current_theme = theme()

    # Show a "splash" type message so the user doesn't give up waiting
    popup_quick_message(
        "Hang on for a moment, this will take a bit to create....",
        keep_on_top=True,
        background_color="red",
        text_color="#FFFFFF",
        auto_close=True,
        non_blocking=True,
    )

    web = False

    win_bg = "black"

    def sample_layout():
        return [
            [Text("Text element"), InputText("Input data here", size=(10, 1))],
            [
                Button("Ok"),
                Button("Disabled", disabled=True),
                Slider((1, 10), orientation="h", size=(5, 15)),
            ],
        ]

    names = list_of_look_and_feel_values()
    names.sort()
    if search_string not in (None, ""):
        names = [
            name
            for name in names
            if search_string.lower().replace(" ", "") in name.lower().replace(" ", "")
        ]

    if search_string not in (None, ""):
        layout = [
            [
                Text(
                    'Themes containing "{}"'.format(search_string),
                    font="Default 18",
                    background_color=win_bg,
                )
            ]
        ]
    else:
        layout = [
            [Text("List of all themes", font="Default 18", background_color=win_bg)]
        ]

    col_layout = []
    row = []
    for count, theme_name in enumerate(names):
        theme(theme_name)
        if not count % columns:
            col_layout += [row]
            row = []
        row += [
            Frame(
                theme_name,
                sample_layout() if not web else [[T(theme_name)]] + sample_layout(),
                pad=(2, 2),
            )
        ]
    if row:
        col_layout += [row]

    layout += [
        [
            Column(
                col_layout,
                scrollable=scrollable,
                size=scroll_area_size,
                pad=(0, 0),
                background_color=win_bg,
                key="-COL-",
            )
        ]
    ]
    window = Window(
        "Preview of Themes",
        layout,
        background_color=win_bg,
        resizable=True,
        location=location,
        keep_on_top=True,
        finalize=True,
        modal=True,
    )
    window["-COL-"].expand(
        True, True, True
    )  # needed so that col will expand with the window
    window.read(close=True)
    theme(current_theme)


preview_all_look_and_feel_themes = theme_previewer


def _theme_preview_window_swatches():
    # Begin the layout with a header
    layout = [
        [
            Text(
                "Themes as color swatches",
                text_color="white",
                background_color="black",
                font="Default 25",
            )
        ],
        [
            Text(
                "Tooltip and right click a color to get the value",
                text_color="white",
                background_color="black",
                font="Default 15",
            )
        ],
        [
            Text(
                "Left click a color to copy to clipboard",
                text_color="white",
                background_color="black",
                font="Default 15",
            )
        ],
    ]
    layout = [[Column(layout, element_justification="c", background_color="black")]]
    # Create the pain part, the rows of Text with color swatches
    for i, theme_name in enumerate(theme_list()):
        theme(theme_name)
        colors = [
            theme_background_color(),
            theme_text_color(),
            theme_input_background_color(),
            theme_input_text_color(),
        ]
        if theme_button_color() != COLOR_SYSTEM_DEFAULT:
            colors.append(theme_button_color()[0])
            colors.append(theme_button_color()[1])
        colors = list(set(colors))  # de-duplicate items
        row = [
            T(
                theme(),
                background_color="black",
                text_color="white",
                size=(20, 1),
                justification="r",
            )
        ]
        for color in colors:
            if color != COLOR_SYSTEM_DEFAULT:
                row.append(
                    T(
                        SYMBOL_SQUARE,
                        text_color=color,
                        background_color="black",
                        pad=(0, 0),
                        font="DEFAUlT 20",
                        right_click_menu=["Nothing", [color]],
                        tooltip=color,
                        enable_events=True,
                        key=(i, color),
                    )
                )
        layout += [row]
    # place layout inside of a Column so that it's scrollable
    layout = [
        [
            Column(
                layout,
                size=(500, 900),
                scrollable=True,
                vertical_scroll_only=True,
                background_color="black",
            )
        ]
    ]
    # finish the layout by adding an exit button
    layout += [[B("Exit")]]

    # create and return Window that uses the layout
    return Window(
        "Theme Color Swatches",
        layout,
        background_color="black",
        finalize=True,
        keep_on_top=True,
    )


def theme_previewer_swatches():
    """
    Display themes in a window as color swatches.
    Click on a color swatch to see the hex value printed on the console.
    If you hover over a color or right click it you'll also see the hext value.
    """
    current_theme = theme()
    popup_quick_message(
        "This is going to take a minute...",
        text_color="white",
        background_color="red",
        font="Default 20",
        keep_on_top=True,
    )
    window = _theme_preview_window_swatches()
    theme(OFFICIAL_PYSIMPLEGUI_THEME)
    # col_height = window.get_screen_size()[1]-200
    # if window.size[1] > 100:
    #     window.size = (window.size[0], col_height)
    # window.move(window.get_screen_size()[0] // 2 - window.size[0] // 2, 0)

    while True:  # Event Loop
        event, values = window.read()
        if event == WIN_CLOSED or event == "Exit":
            break
        if isinstance(event, tuple):  # someone clicked a swatch
            chosen_color = event[1]
        else:
            if event[0] == "#":  # someone right clicked
                chosen_color = event
            else:
                chosen_color = ""
        print("Copied to clipboard color = ", chosen_color)
        clipboard_set(chosen_color)
        # window.TKroot.clipboard_clear()
        # window.TKroot.clipboard_append(chosen_color)
    window.close()
    theme(current_theme)


def change_look_and_feel(index, force=False):
    """
    Change the "color scheme" of all future PySimpleGUI Windows.
    The scheme are string names that specify a group of colors. Background colors, text colors, button colors.
    There are 13 different color settings that are changed at one time using a single call to ChangeLookAndFeel
    The look and feel table itself has these indexes into the dictionary LOOK_AND_FEEL_TABLE.
    The original list was (prior to a major rework and renaming)... these names still work...
    In Nov 2019 a new Theme Formula was devised to make choosing a theme easier:
    The "Formula" is:
    ["Dark" or "Light"] Color Number
    Colors can be Blue Brown Grey Green Purple Red Teal Yellow Black
    The number will vary for each pair. There are more DarkGrey entries than there are LightYellow for example.
    Default = The default settings (only button color is different than system default)
    Default1 = The full system default including the button (everything's gray... how sad... don't be all gray... please....)
    :param index: the name of the index into the Look and Feel table (does not have to be exact, can be "fuzzy")
    :type index:  (str)
    :param force: no longer used
    :type force:  (bool)
    :return:      None
    :rtype:       None
    """
    global CURRENT_LOOK_AND_FEEL

    # if running_mac() and not force:
    #     print('*** Changing look and feel is not supported on Mac platform ***')
    #     return

    requested_theme_name = index
    theme_names_list = list_of_look_and_feel_values()
    # normalize available l&f values by setting all to lower case
    lf_values_lowercase = [item.lower() for item in theme_names_list]
    # option 1
    opt1 = requested_theme_name.replace(" ", "").lower()
    # option 3 is option 1 with gray replaced with grey
    opt3 = opt1.replace("gray", "grey")
    # option 2 (reverse lookup)
    optx = requested_theme_name.lower().split(" ")
    optx.reverse()
    opt2 = "".join(optx)

    # search for valid l&f name
    if requested_theme_name in theme_names_list:
        ix = theme_names_list.index(requested_theme_name)
    elif opt1 in lf_values_lowercase:
        ix = lf_values_lowercase.index(opt1)
    elif opt2 in lf_values_lowercase:
        ix = lf_values_lowercase.index(opt2)
    elif opt3 in lf_values_lowercase:
        ix = lf_values_lowercase.index(opt3)
    else:
        ix = random.randint(0, len(lf_values_lowercase) - 1)
        print(
            "** Warning - {} Theme is not a valid theme. Change your theme call. **".format(
                index
            )
        )
        print("valid values are", list_of_look_and_feel_values())
        print(
            "Instead, please enjoy a random Theme named {}".format(
                list_of_look_and_feel_values()[ix]
            )
        )

    selection = theme_names_list[ix]
    CURRENT_LOOK_AND_FEEL = selection
    try:
        colors = LOOK_AND_FEEL_TABLE[selection]

        # Color the progress bar using button background and input colors...unless they're the same
        if colors["PROGRESS"] != COLOR_SYSTEM_DEFAULT:
            if colors["PROGRESS"] == DEFAULT_PROGRESS_BAR_COMPUTE:
                if (
                        colors["BUTTON"][1] != colors["INPUT"]
                        and colors["BUTTON"][1] != colors["BACKGROUND"]
                ):
                    colors["PROGRESS"] = colors["BUTTON"][1], colors["INPUT"]
                else:  # if the same, then use text input on top of input color
                    colors["PROGRESS"] = (colors["TEXT_INPUT"], colors["INPUT"])
        else:
            colors["PROGRESS"] = DEFAULT_PROGRESS_BAR_COLOR_OFFICIAL
        # call to change all the colors
        SetOptions(
            background_color=colors["BACKGROUND"],
            text_element_background_color=colors["BACKGROUND"],
            element_background_color=colors["BACKGROUND"],
            text_color=colors["TEXT"],
            input_elements_background_color=colors["INPUT"],
            # button_color=colors['BUTTON'] if not running_mac() else None,
            button_color=colors["BUTTON"],
            progress_meter_color=colors["PROGRESS"],
            border_width=colors["BORDER"],
            slider_border_width=colors["SLIDER_DEPTH"],
            progress_meter_border_depth=colors["PROGRESS_DEPTH"],
            scrollbar_color=(colors["SCROLL"]),
            element_text_color=colors["TEXT"],
            input_text_color=colors["TEXT_INPUT"],
        )
    except Exception:  # most likely an index out of range
        print("** Warning - Theme value not valid. Change your theme call. **")
        print("valid values are", list_of_look_and_feel_values())


# ------------------------ Color processing functions ------------------------


def _hex_to_hsl(hex):
    r, g, b = _hex_to_rgb(hex)
    return _rgb_to_hsl(r, g, b)


def _hex_to_rgb(hex):
    hex = hex.lstrip("#")
    hlen = len(hex)
    return tuple(int(hex[i: i + hlen // 3], 16) for i in range(0, hlen, hlen // 3))


def _rgb_to_hsl(r, g, b):
    r = float(r)
    g = float(g)
    b = float(b)
    high = max(r, g, b)
    low = min(r, g, b)
    h, s, v = ((high + low) / 2,) * 3
    if high == low:
        h = s = 0.0
    else:
        d = high - low
        l = (high + low) / 2
        s = d / (2 - high - low) if l > 0.5 else d / (high + low)
        h = {
            r: (g - b) / d + (6 if g < b else 0),
            g: (b - r) / d + 2,
            b: (r - g) / d + 4,
        }[high]
        h /= 6
    return h, s, v


def _hsl_to_rgb(h, s, l):
    def hue_to_rgb(p, q, t):
        t += 1 if t < 0 else 0
        t -= 1 if t > 1 else 0
        if t < 1 / 6:
            return p + (q - p) * 6 * t
        if t < 1 / 2:
            return q
        if t < 2 / 3:
            p + (q - p) * (2 / 3 - t) * 6
        return p

    if s == 0:
        r, g, b = l, l, l
    else:
        q = l * (1 + s) if l < 0.5 else l + s - l * s
        p = 2 * l - q
        r = hue_to_rgb(p, q, h + 1 / 3)
        g = hue_to_rgb(p, q, h)
        b = hue_to_rgb(p, q, h - 1 / 3)

    return r, g, b


def _hsv_to_hsl(h, s, v):
    l = 0.5 * v * (2 - s)
    s = v * s / (1 - fabs(2 * l - 1))
    return h, s, l


def _hsl_to_hsv(h, s, l):
    v = (2 * l + s * (1 - fabs(2 * l - 1))) / 2
    s = 2 * (v - l) / v
    return h, s, v


# Converts an object's contents into a nice printable string.  Great for dumping debug data
def obj_to_string_single_obj(obj):
    """
    Dumps an Object's values as a formatted string.  Very nicely done. Great way to display an object's member variables in human form
    Returns only the top-most object's variables instead of drilling down to dispolay more
    :param obj: The object to display
    :type obj:  (Any)
    :return:    Formatted output of the object's values
    :rtype:     (str)
    """
    if obj is None:
        return "None"
    return (
            str(obj.__class__)
            + "\n"
            + "\n".join(
        (
            repr(item) + " = " + repr(obj.__dict__[item])
            for item in sorted(obj.__dict__)
        )
    )
    )


def obj_to_string(obj, extra="    "):
    """
    Dumps an Object's values as a formatted string.  Very nicely done. Great way to display an object's member variables in human form
    :param obj:   The object to display
    :type obj:    (Any)
    :param extra: extra stuff (Default value = '    ')
    :type extra:  (str)
    :return:      Formatted output of the object's values
    :rtype:       (str)
    """
    if obj is None:
        return "None"
    return (
            str(obj.__class__)
            + "\n"
            + "\n".join(
        (
            extra
            + (
                    str(item)
                    + " = "
                    + (
                        ObjToString(obj.__dict__[item], extra + "    ")
                        if hasattr(obj.__dict__[item], "__dict__")
                        else str(obj.__dict__[item])
                    )
            )
            for item in sorted(obj.__dict__)
        )
    )
    )


# ...######..##.......####.########..########...#######.....###....########..########.
# ..##....##.##........##..##.....##.##.....##.##.....##...##.##...##.....##.##.....##
# ..##.......##........##..##.....##.##.....##.##.....##..##...##..##.....##.##.....##
# ..##.......##........##..########..########..##.....##.##.....##.########..##.....##
# ..##.......##........##..##........##.....##.##.....##.#########.##...##...##.....##
# ..##....##.##........##..##........##.....##.##.....##.##.....##.##....##..##.....##
# ...######..########.####.##........########...#######..##.....##.##.....##.########.
# ................................................................................
# ..########.##.....##.##....##..######..########.####..#######..##....##..######.
# ..##.......##.....##.###...##.##....##....##.....##..##.....##.###...##.##....##
# ..##.......##.....##.####..##.##..........##.....##..##.....##.####..##.##......
# ..######...##.....##.##.##.##.##..........##.....##..##.....##.##.##.##..######.
# ..##.......##.....##.##..####.##..........##.....##..##.....##.##..####.......##
# ..##.......##.....##.##...###.##....##....##.....##..##.....##.##...###.##....##
# ..##........#######..##....##..######.....##....####..#######..##....##..######.


def clipboard_set(new_value):
    """
    Sets the clipboard to a specific value.
    IMPORTANT NOTE - Your PySimpleGUI application needs to remain running until you've pasted
    your clipboard. This is a tkinter limitation.  A workaround was found for Windows, but you still
    need to stay running for Linux systems.

    :param new_value: value to set the clipboard to. Will be converted to a string
    :type new_value:  (str | bytes)
    """
    root = _get_hidden_master_root()
    root.clipboard_clear()
    root.clipboard_append(str(new_value))
    root.update()


def clipboard_get():
    """
    Gets the clipboard current value.

    :return: The current value of the clipboard
    :rtype:  (str)
    """
    root = _get_hidden_master_root()

    try:
        value = root.clipboard_get()
    except Exception:
        value = ""
    root.update()
    return value


# MM"""""""`YM
# MM  mmmmm  M
# M'        .M .d8888b. 88d888b. dP    dP 88d888b. .d8888b.
# MM  MMMMMMMM 88'  `88 88'  `88 88    88 88'  `88 Y8ooooo.
# MM  MMMMMMMM 88.  .88 88.  .88 88.  .88 88.  .88       88
# MM  MMMMMMMM `88888P' 88Y888P' `88888P' 88Y888P' `88888P'
# MMMMMMMMMMMM          88                88
#                       dP                dP
# ------------------------------------------------------------------------------------------------------------------ #
# =====================================   Upper PySimpleGUI ======================================================== #
# ------------------------------------------------------------------------------------------------------------------ #
# ----------------------------------- The mighty Popup! ------------------------------------------------------------ #


def popup(
        *args,
        title: str = None,
        button_color=None,
        background_color=None,
        text_color=None,
        button_type=POPUP_BUTTONS_OK,
        auto_close=False,
        auto_close_duration=None,
        custom_text=(None, None),
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=True,
        location=(None, None),
        relative_location=(None, None),
        any_key_closes=False,
        image=None,
        modal=True,
        button_justification=None,
        drop_whitespace=True,
):
    """
    Popup - Display a popup Window with as many parms as you wish to include.  This is the GUI equivalent of the
    "print" statement.  It's also great for "pausing" your program's flow until the user can read some error messages.

    If this popup doesn't have the features you want, then you can easily make your own. Popups can be accomplished in 1 line of code:
    choice, _ = sg.Window('Continue?', [[sg.T('Do you want to continue?')], [sg.Yes(s=10), sg.No(s=10)]], disable_close=True).read(close=True)


    :param title:                 Optional title for the window. If none provided, the first arg will be used instead.
    :type title:                  (str)
    :param button_color:          Color of the buttons shown (text color, button color)
    :type button_color:           (str, str) | str
    :param background_color:      Window's background color
    :type background_color:       (str)
    :param text_color:            text color
    :type text_color:             (str)
    :param button_type:           NOT USER SET!  Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK). There are many Popup functions and they call Popup, changing this parameter to get the desired effect.
    :type button_type:            (int)
    :param auto_close:            If True the window will automatically close
    :type auto_close:             (bool)
    :param auto_close_duration:   time in seconds to keep window open before closing it automatically
    :type auto_close_duration:    (int)
    :param custom_text:           A string or pair of strings that contain the text to display on the buttons
    :type custom_text:            (str, str) | str
    :param non_blocking:          If True then will immediately return from the function without waiting for the user's input.
    :type non_blocking:           (bool)
    :param icon:                  icon to display on the window. Same format as a Window call
    :type icon:                   str | bytes
    :param line_width:            Width of lines in characters.  Defaults to MESSAGE_BOX_LINE_WIDTH
    :type line_width:             (int)
    :param font:                  specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                   str | Tuple[font_name, size, modifiers]
    :param no_titlebar:           If True will not show the frame around the window and the titlebar across the top
    :type no_titlebar:            (bool)
    :param grab_anywhere:         If True can grab anywhere to move the window. If no_titlebar is True, grab_anywhere should likely be enabled too
    :type grab_anywhere:          (bool)
    :param location:              Location on screen to display the top left corner of window. Defaults to window centered on screen
    :type location:               (int, int)
    :param relative_location:     (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:      (int, int)
    :param keep_on_top:           If True the window will remain above all current windows
    :type keep_on_top:            (bool)
    :param any_key_closes:        If True then will turn on return_keyboard_events for the window which will cause window to close as soon as any key is pressed.  Normally the return key only will close the window.  Default is false.
    :type any_key_closes:         (bool)
    :param image:                 Image to include at the top of the popup window
    :type image:                  (str) or (bytes)
    :param modal:                 If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                  bool
    :param button_justification:  Speficies if buttons should be left, right or centered. Default is left justified
    :type button_justification:   str
    :param drop_whitespace:       Controls is whitespace should be removed when wrapping text.  Parameter is passed to textwrap.fill. Default is to drop whitespace (so popup remains backward compatible)
    :type drop_whitespace:        bool
    :return:                      Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                       str | None
    """

    if not args:
        args_to_print = [""]
    else:
        args_to_print = args
    if line_width is not None:
        local_line_width = line_width
    else:
        local_line_width = MESSAGE_BOX_LINE_WIDTH
    _title = title if title is not None else args_to_print[0]

    layout = [[]]
    max_line_total, total_lines = 0, 0
    if image is not None:
        if isinstance(image, str):
            layout += [[Image(filename=image)]]
        else:
            layout += [[Image(data=image)]]

    for message in args_to_print:
        # fancy code to check if string and convert if not is not need. Just always convert to string :-)
        # if not isinstance(message, str): message = str(message)
        message = str(message)
        if message.count(
                "\n"
        ):  # if there are line breaks, then wrap each segment separately
            # message_wrapped = message         # used to just do this, but now breaking into smaller pieces
            message_wrapped = ""
            msg_list = message.split(
                "\n"
            )  # break into segments that will each be wrapped
            message_wrapped = "\n".join(
                [textwrap.fill(msg, local_line_width) for msg in msg_list]
            )
        else:
            message_wrapped = textwrap.fill(
                message, local_line_width, drop_whitespace=drop_whitespace
            )
        message_wrapped_lines = message_wrapped.count("\n") + 1
        longest_line_len = max([len(l) for l in message.split("\n")])
        width_used = min(longest_line_len, local_line_width)
        max_line_total = max(max_line_total, width_used)
        # height = _GetNumLinesNeeded(message, width_used)
        height = message_wrapped_lines
        layout += [
            [
                Text(
                    message_wrapped,
                    auto_size_text=True,
                    text_color=text_color,
                    background_color=background_color,
                )
            ]
        ]
        total_lines += height

    if non_blocking:
        PopupButton = (
            DummyButton  # important to use or else button will close other windows too!
        )
    else:
        PopupButton = Button
    # show either an OK or Yes/No depending on paramater
    if custom_text != (None, None):
        if type(custom_text) is not tuple:
            layout += [
                [
                    PopupButton(
                        custom_text,
                        size=(len(custom_text), 1),
                        button_color=button_color,
                        focus=True,
                        bind_return_key=True,
                    )
                ]
            ]
        elif custom_text[1] is None:
            layout += [
                [
                    PopupButton(
                        custom_text[0],
                        size=(len(custom_text[0]), 1),
                        button_color=button_color,
                        focus=True,
                        bind_return_key=True,
                    )
                ]
            ]
        else:
            layout += [
                [
                    PopupButton(
                        custom_text[0],
                        button_color=button_color,
                        focus=True,
                        bind_return_key=True,
                        size=(len(custom_text[0]), 1),
                    ),
                    PopupButton(
                        custom_text[1],
                        button_color=button_color,
                        size=(len(custom_text[1]), 1),
                    ),
                ]
            ]
    elif button_type == POPUP_BUTTONS_YES_NO:
        layout += [
            [
                PopupButton(
                    "Yes",
                    button_color=button_color,
                    focus=True,
                    bind_return_key=True,
                    size=(5, 1),
                ),
                PopupButton("No", button_color=button_color, size=(5, 1)),
            ]
        ]
    elif button_type == POPUP_BUTTONS_CANCELLED:
        layout += [
            [
                PopupButton(
                    "Cancelled",
                    button_color=button_color,
                    focus=True,
                    bind_return_key=True,
                )
            ]
        ]
    elif button_type == POPUP_BUTTONS_ERROR:
        layout += [
            [
                PopupButton(
                    "Error",
                    size=(6, 1),
                    button_color=button_color,
                    focus=True,
                    bind_return_key=True,
                )
            ]
        ]
    elif button_type == POPUP_BUTTONS_OK_CANCEL:
        layout += [
            [
                PopupButton(
                    "OK",
                    size=(6, 1),
                    button_color=button_color,
                    focus=True,
                    bind_return_key=True,
                ),
                PopupButton("Cancel", size=(6, 1), button_color=button_color),
            ]
        ]
    elif button_type == POPUP_BUTTONS_NO_BUTTONS:
        pass
    else:
        layout += [
            [
                PopupButton(
                    "OK",
                    size=(5, 1),
                    button_color=button_color,
                    focus=True,
                    bind_return_key=True,
                )
            ]
        ]
    if button_justification is not None:
        justification = button_justification.lower()[0]
        if justification == "r":
            layout[-1] = [Push()] + layout[-1]
        elif justification == "c":
            layout[-1] = [Push()] + layout[-1] + [Push()]

    window = Window(
        _title,
        layout,
        auto_size_text=True,
        background_color=background_color,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        icon=icon,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        return_keyboard_events=any_key_closes,
        modal=modal,
    )

    if non_blocking:
        button, values = window.read(timeout=0)
    else:
        button, values = window.read()
        window.close()
        del window

    return button


# ==============================  MsgBox============#
# Lazy function. Same as calling Popup with parms   #
# This function WILL Disappear perhaps today        #
# ==================================================#
# MsgBox is the legacy call and should not be used any longer
def MsgBox(*args):
    """
    Do not call this anymore it will raise exception.  Use Popups instead
    """
    raise DeprecationWarning(
        "MsgBox is no longer supported... change your call to Popup"
    )


# ========================  Scrolled Text Box   =====#
# ===================================================#
def popup_scrolled(
        *args,
        title=None,
        button_color=None,
        background_color=None,
        text_color=None,
        yes_no=False,
        no_buttons=False,
        button_justification="l",
        auto_close=False,
        auto_close_duration=None,
        size=(None, None),
        location=(None, None),
        relative_location=(None, None),
        non_blocking=False,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        font=None,
        image=None,
        icon=None,
        modal=True,
        no_sizegrip=False,
):
    """
    Show a scrolled Popup window containing the user's text that was supplied.  Use with as many items to print as you
    want, just like a print statement.

    :param title:                 Title to display in the window.
    :type title:                  (str)
    :param button_color:          button color (foreground, background)
    :type button_color:           (str, str) | str
    :param yes_no:                If True, displays Yes and No buttons instead of Ok
    :type yes_no:                 (bool)
    :param no_buttons:            If True, no buttons will be shown. User will have to close using the "X"
    :type no_buttons:             (bool)
    :param button_justification:  How buttons should be arranged.  l, c, r for Left, Center or Right justified
    :type button_justification:   (str)
    :param auto_close:            if True window will close itself
    :type auto_close:             (bool)
    :param auto_close_duration:   Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:    int | float
    :param size:                  (w,h) w=characters-wide, h=rows-high
    :type size:                   (int, int)
    :param location:              Location on the screen to place the upper left corner of the window
    :type location:               (int, int)
    :param relative_location:     (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:      (int, int)
    :param non_blocking:          if True the call will immediately return rather than waiting on user input
    :type non_blocking:           (bool)
    :param background_color:      color of background
    :type background_color:       (str)
    :param text_color:            color of the text
    :type text_color:             (str)
    :param no_titlebar:           If True no titlebar will be shown
    :type no_titlebar:            (bool)
    :param grab_anywhere:         If True, than can grab anywhere to move the window (Default = False)
    :type grab_anywhere:          (bool)
    :param keep_on_top:           If True the window will remain above all current windows
    :type keep_on_top:            (bool)
    :param font:                  specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                   (str or (str, int[, str]) or None)
    :param image:                 Image to include at the top of the popup window
    :type image:                  (str) or (bytes)
    :param icon:                  filename or base64 string to be used for the window's icon
    :type icon:                   bytes | str
    :param modal:                 If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                  bool
    :param no_sizegrip:           If True no Sizegrip will be shown when there is no titlebar. It's only shown if there is no titlebar
    :type no_sizegrip:            (bool)
    :return:                      Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                       str | None | TIMEOUT_KEY
    """

    if not args:
        return
    width, height = size
    width = width if width else MESSAGE_BOX_LINE_WIDTH

    layout = [[]]

    if image is not None:
        if isinstance(image, str):
            layout += [[Image(filename=image)]]
        else:
            layout += [[Image(data=image)]]
    max_line_total, max_line_width, total_lines, height_computed = 0, 0, 0, 0
    complete_output = ""
    for message in args:
        # fancy code to check if string and convert if not is not need. Just always convert to string :-)
        # if not isinstance(message, str): message = str(message)
        message = str(message)
        longest_line_len = max([len(l) for l in message.split("\n")])
        width_used = min(longest_line_len, width)
        max_line_total = max(max_line_total, width_used)
        max_line_width = width
        lines_needed = _GetNumLinesNeeded(message, width_used)
        height_computed += lines_needed + 1
        complete_output += message + "\n"
        total_lines += lines_needed
    height_computed = (
        MAX_SCROLLED_TEXT_BOX_HEIGHT
        if height_computed > MAX_SCROLLED_TEXT_BOX_HEIGHT
        else height_computed
    )
    if height:
        height_computed = height
    layout += [
        [
            Multiline(
                complete_output,
                size=(max_line_width, height_computed),
                background_color=background_color,
                text_color=text_color,
                expand_x=True,
                expand_y=True,
                k="-MLINE-",
            )
        ]
    ]
    # show either an OK or Yes/No depending on paramater
    button = DummyButton if non_blocking else Button

    if yes_no:
        buttons = [button("Yes"), button("No")]
    elif no_buttons:
        buttons = [button("OK", size=(5, 1), button_color=button_color)]
    else:
        buttons = None

    if buttons is not None:
        if button_justification.startswith("l"):
            layout += [buttons]
        elif button_justification.startswith("c"):
            layout += [[Push()] + buttons + [Push()]]
        else:
            layout += [[Push()] + buttons]

    if no_sizegrip:
        layout[-1] += [Sizegrip()]

    window = Window(
        title or args[0],
        layout,
        auto_size_text=True,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        location=location,
        relative_location=relative_location,
        resizable=True,
        font=font,
        background_color=background_color,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        modal=modal,
        icon=icon,
    )
    if non_blocking:
        button, values = window.read(timeout=0)
    else:
        button, values = window.read()
        window.close()
        del window
    return button


# ============================== sprint ======#
# Is identical to the Scrolled Text Box       #
# Provides a crude 'print' mechanism but in a #
# GUI environment                             #
# This is in addition to the Print function   #
# which routes output to a "Debug Window"     #
# ============================================#


# --------------------------- popup_no_buttons ---------------------------
def popup_no_buttons(
        *args,
        title=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """Show a Popup but without any buttons

    :param keep_on_top:
    :param title:               Title to display in the window.
    :type title:                (str)
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        If True then will immediately return from the function without waiting for the user's input. (Default = False)
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True, than can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY"""
    Popup(
        *args,
        title=title,
        background_color=background_color,
        text_color=text_color,
        button_type=POPUP_BUTTONS_NO_BUTTONS,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_non_blocking ---------------------------
def popup_non_blocking(
        *args,
        title=None,
        button_type=POPUP_BUTTONS_OK,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=True,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=False,
):
    """
    Show Popup window and immediately return (does not block)

    :param keep_on_top:
    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_type:         Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK).
    :type button_type:          (int)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = False
    :type modal:                bool
    :return:                    Reason for popup closing
    :rtype:                     str | None
    """

    return popup(
        *args,
        title=title,
        button_color=button_color,
        background_color=background_color,
        text_color=text_color,
        button_type=button_type,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_quick - a NonBlocking, Self-closing Popup  ---------------------------
def popup_quick(
        *args,
        title=None,
        button_type=POPUP_BUTTONS_OK,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=True,
        auto_close_duration=2,
        non_blocking=True,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=False,
):
    """
    Show Popup box that doesn't block and closes itself

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_type:         Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK).
    :type button_type:          (int)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = False
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """

    return popup(
        *args,
        title=title,
        button_color=button_color,
        background_color=background_color,
        text_color=text_color,
        button_type=button_type,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_quick_message - a NonBlocking, Self-closing Popup with no titlebar and no buttons ---------------------------
def popup_quick_message(
        *args,
        title=None,
        button_type=POPUP_BUTTONS_NO_BUTTONS,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=True,
        auto_close_duration=2,
        non_blocking=True,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=True,
        grab_anywhere=False,
        keep_on_top=True,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=False,
):
    """
    Show Popup window with no titlebar, doesn't block, and auto closes itself.

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_type:         Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK).
    :type button_type:          (int)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = False
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """
    return popup(
        *args,
        title=title,
        button_color=button_color,
        background_color=background_color,
        text_color=text_color,
        button_type=button_type,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- PopupNoTitlebar ---------------------------
def popup_no_titlebar(
        *args,
        title=None,
        button_type=POPUP_BUTTONS_OK,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        grab_anywhere=True,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Display a Popup without a titlebar.   Enables grab anywhere so you can move it

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_type:         Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK).
    :type button_type:          (int)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """
    return popup(
        *args,
        title=title,
        button_color=button_color,
        background_color=background_color,
        text_color=text_color,
        button_type=button_type,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=True,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- PopupAutoClose ---------------------------
def popup_auto_close(
        *args,
        title=None,
        button_type=POPUP_BUTTONS_OK,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=True,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """Popup that closes itself after some time period

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_type:         Determines which pre-defined buttons will be shown (Default value = POPUP_BUTTONS_OK).
    :type button_type:          (int)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """

    return popup(
        *args,
        title=title,
        button_color=button_color,
        background_color=background_color,
        text_color=text_color,
        button_type=button_type,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_error ---------------------------
def popup_error(
        *args,
        title=None,
        button_color=(None, None),
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Popup with colored button and 'Error' as button text

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """
    tbutton_color = (
        DEFAULT_ERROR_BUTTON_COLOR if button_color == (None, None) else button_color
    )
    return popup(
        *args,
        title=title,
        button_type=POPUP_BUTTONS_ERROR,
        background_color=background_color,
        text_color=text_color,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        button_color=tbutton_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_cancel ---------------------------
def popup_cancel(
        *args,
        title=None,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Display Popup with "cancelled" button text

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """
    return popup(
        *args,
        title=title,
        button_type=POPUP_BUTTONS_CANCELLED,
        background_color=background_color,
        text_color=text_color,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_ok ---------------------------
def popup_ok(
        *args,
        title=None,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Display Popup with OK button only

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    Returns text of the button that was pressed.  None will be returned if user closed window with X
    :rtype:                     str | None | TIMEOUT_KEY
    """
    return popup(
        *args,
        title=title,
        button_type=POPUP_BUTTONS_OK,
        background_color=background_color,
        text_color=text_color,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_ok_cancel ---------------------------
def popup_ok_cancel(
        *args,
        title=None,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=DEFAULT_WINDOW_ICON,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Display popup with OK and Cancel buttons

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    clicked button
    :rtype:                     "OK" | "Cancel" | None
    """
    return popup(
        *args,
        title=title,
        button_type=POPUP_BUTTONS_OK_CANCEL,
        background_color=background_color,
        text_color=text_color,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


# --------------------------- popup_yes_no ---------------------------
def popup_yes_no(
        *args,
        title=None,
        button_color=None,
        background_color=None,
        text_color=None,
        auto_close=False,
        auto_close_duration=None,
        non_blocking=False,
        icon=None,
        line_width=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        modal=True,
):
    """
    Display Popup with Yes and No buttons

    :param title:               Title to display in the window.
    :type title:                (str)
    :param button_color:        button color (foreground, background)
    :type button_color:         (str, str) | str
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param auto_close:          if True window will close itself
    :type auto_close:           (bool)
    :param auto_close_duration: Older versions only accept int. Time in seconds until window will close
    :type auto_close_duration:  int | float
    :param non_blocking:        if True the call will immediately return rather than waiting on user input
    :type non_blocking:         (bool)
    :param icon:                filename or base64 string to be used for the window's icon
    :type icon:                 bytes | str
    :param line_width:          Width of lines in characters
    :type line_width:           (int)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 (str or (str, int[, str]) or None)
    :param no_titlebar:         If True no titlebar will be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True the window will remain above all current windows
    :type keep_on_top:          (bool)
    :param location:            Location of upper left corner of the window
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param image:               Image to include at the top of the popup window
    :type image:                (str) or (bytes)
    :param modal:               If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                bool
    :return:                    clicked button
    :rtype:                     "Yes" | "No" | None
    """
    return popup(
        *args,
        title=title,
        button_type=POPUP_BUTTONS_YES_NO,
        background_color=background_color,
        text_color=text_color,
        non_blocking=non_blocking,
        icon=icon,
        line_width=line_width,
        button_color=button_color,
        auto_close=auto_close,
        auto_close_duration=auto_close_duration,
        font=font,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        image=image,
        modal=modal,
    )


##############################################################################
#   The popup_get_____ functions - Will return user input                     #
##############################################################################

# --------------------------- popup_get_folder ---------------------------


def popup_get_folder(
        message,
        title=None,
        default_path="",
        no_window=False,
        size=(None, None),
        button_color=None,
        background_color=None,
        text_color=None,
        icon=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        initial_folder=None,
        image=None,
        modal=True,
        history=False,
        history_setting_filename=None,
):
    """
    Display popup with text entry field and browse button so that a folder can be chosen.

    :param message:                  message displayed to user
    :type message:                   (str)
    :param title:                    Window title
    :type title:                     (str)
    :param default_path:             path to display to user as starting point (filled into the input field)
    :type default_path:              (str)
    :param no_window:                if True, no PySimpleGUI window will be shown. Instead just the tkinter dialog is shown
    :type no_window:                 (bool)
    :param size:                     (width, height) of the InputText Element
    :type size:                      (int, int)
    :param button_color:             button color (foreground, background)
    :type button_color:              (str, str) | str
    :param background_color:         color of background
    :type background_color:          (str)
    :param text_color:               color of the text
    :type text_color:                (str)
    :param icon:                     filename or base64 string to be used for the window's icon
    :type icon:                      bytes | str
    :param font:                     specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                      (str or (str, int[, str]) or None)
    :param no_titlebar:              If True no titlebar will be shown
    :type no_titlebar:               (bool)
    :param grab_anywhere:            If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:             (bool)
    :param keep_on_top:              If True the window will remain above all current windows
    :type keep_on_top:               (bool)
    :param location:                 Location of upper left corner of the window
    :type location:                  (int, int)
    :param relative_location:        (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:         (int, int)
    :param initial_folder:           location in filesystem to begin browsing
    :type initial_folder:            (str)
    :param image:                    Image to include at the top of the popup window
    :type image:                     (str) or (bytes)
    :param modal:                    If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                     bool
    :param history:                  If True then enable a "history" feature that will display previous entries used. Uses settings filename provided or default if none provided
    :type history:                   bool
    :param history_setting_filename: Filename to use for the User Settings. Will store list of previous entries in this settings file
    :type history_setting_filename:  (str)
    :return:                         string representing the path chosen, None if cancelled or window closed with X
    :rtype:                          str | None
    """

    # First setup the history settings file if history feature is enabled
    if history and history_setting_filename is not None:
        try:
            history_settings = UserSettings(history_setting_filename)
        except Exception as e:
            _error_popup_with_traceback(
                "popup_get_folder - Something is wrong with your supplied history settings filename",
                "Exception: {}".format(e),
            )
            return None
    elif history:
        history_settings_filename = os.path.basename(inspect.stack()[1].filename)
        history_settings_filename = (
                os.path.splitext(history_settings_filename)[0] + ".json"
        )
        history_settings = UserSettings(history_settings_filename)
    else:
        history_settings = None

    # global _my_windows
    if no_window:
        _get_hidden_master_root()
        root = tk.Toplevel()

        try:
            root.attributes(
                "-alpha", 0
            )  # hide window while building it. makes for smoother 'paint'
            # if not running_mac():
            try:
                root.wm_overrideredirect(True)
            except Exception as e:
                print(
                    "* Error performing wm_overrideredirect while hiding the window during creation in get folder *",
                    e,
                )
            root.withdraw()
        except Exception:
            pass
        folder_name = tk.filedialog.askdirectory(
            initialdir=initial_folder
        )  # show the 'get folder' dialog box

        root.destroy()

        return folder_name

    browse_button = FolderBrowse(initial_folder=initial_folder)

    if image is not None:
        if isinstance(image, str):
            layout = [[Image(filename=image)]]
        else:
            layout = [[Image(data=image)]]
    else:
        layout = [[]]

    layout += [
        [
            Text(
                message,
                auto_size_text=True,
                text_color=text_color,
                background_color=background_color,
            )
        ]
    ]

    if not history:
        layout += [
            [
                InputText(default_text=default_path, size=size, key="-INPUT-"),
                browse_button,
            ]
        ]
    else:
        file_list = history_settings.get("-PSG folder list-", [])
        last_entry = file_list[0] if file_list else ""
        layout += [
            [
                Combo(
                    file_list,
                    default_value=last_entry,
                    key="-INPUT-",
                    size=size if size != (None, None) else (80, 1),
                    bind_return_key=True,
                ),
                browse_button,
                Button(
                    "Clear History",
                    tooltip="Clears the list of folders shown in the combobox",
                ),
            ]
        ]

    layout += [
        [Button("Ok", size=(6, 1), bind_return_key=True), Button("Cancel", size=(6, 1))]
    ]

    window = Window(
        title=title or message,
        layout=layout,
        icon=icon,
        auto_size_text=True,
        button_color=button_color,
        font=font,
        background_color=background_color,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        modal=modal,
    )

    while True:
        event, values = window.read()
        if event in ("Cancel", WIN_CLOSED):
            break
        elif event == "Clear History":
            history_settings.set("-PSG folder list-", [])
            window["-INPUT-"].update("", [])
            popup_quick_message(
                "History of Previous Choices Cleared",
                background_color="red",
                text_color="white",
                font="_ 20",
                keep_on_top=True,
            )
        elif event in ("Ok", "-INPUT-"):
            if values["-INPUT-"] != "":
                if history_settings is not None:
                    list_of_entries = history_settings.get("-PSG folder list-", [])
                    if values["-INPUT-"] in list_of_entries:
                        list_of_entries.remove(values["-INPUT-"])
                    list_of_entries.insert(0, values["-INPUT-"])
                    history_settings.set("-PSG folder list-", list_of_entries)
            break

    window.close()
    del window
    if event in ("Cancel", WIN_CLOSED):
        return None

    return values["-INPUT-"]


# --------------------------- popup_get_file ---------------------------


def popup_get_file(
        message,
        title=None,
        default_path="",
        default_extension="",
        save_as=False,
        multiple_files=False,
        file_types=FILE_TYPES_ALL_FILES,
        no_window=False,
        size=(None, None),
        button_color=None,
        background_color=None,
        text_color=None,
        icon=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        initial_folder=None,
        image=None,
        files_delimiter=BROWSE_FILES_DELIMITER,
        modal=True,
        history=False,
        show_hidden=True,
        history_setting_filename=None,
):
    """
    Display popup window with text entry field and browse button so that a file can be chosen by user.

    :param message:                  message displayed to user
    :type message:                   (str)
    :param title:                    Window title
    :type title:                     (str)
    :param default_path:             path to display to user as starting point (filled into the input field)
    :type default_path:              (str)
    :param default_extension:        If no extension entered by user, add this to filename (only used in saveas dialogs)
    :type default_extension:         (str)
    :param save_as:                  if True, the "save as" dialog is shown which will verify before overwriting
    :type save_as:                   (bool)
    :param multiple_files:           if True, then allows multiple files to be selected that are returned with ';' between each filename
    :type multiple_files:            (bool)
    :param file_types:               List of extensions to show using wildcards. All files (the default) = (("ALL Files", "*.* *"),).
    :type file_types:                Tuple[Tuple[str,str]]
    :param no_window:                if True, no PySimpleGUI window will be shown. Instead just the tkinter dialog is shown
    :type no_window:                 (bool)
    :param size:                     (width, height) of the InputText Element or Combo element if using history feature
    :type size:                      (int, int)
    :param button_color:             Color of the button (text, background)
    :type button_color:              (str, str) | str
    :param background_color:         background color of the entire window
    :type background_color:          (str)
    :param text_color:               color of the text
    :type text_color:                (str)
    :param icon:                     filename or base64 string to be used for the window's icon
    :type icon:                      bytes | str
    :param font:                     specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                      (str or (str, int[, str]) or None)
    :param no_titlebar:              If True no titlebar will be shown
    :type no_titlebar:               (bool)
    :param grab_anywhere:            If True: can grab anywhere to move the window (Default = False)
    :type grab_anywhere:             (bool)
    :param keep_on_top:              If True the window will remain above all current windows
    :type keep_on_top:               (bool)
    :param location:                 Location of upper left corner of the window
    :type location:                  (int, int)
    :param relative_location:        (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:         (int, int)
    :param initial_folder:           location in filesystem to begin browsing
    :type initial_folder:            (str)
    :param image:                    Image to include at the top of the popup window
    :type image:                     (str) or (bytes)
    :param files_delimiter:          String to place between files when multiple files are selected. Normally a ;
    :type files_delimiter:           str
    :param modal:                    If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                     bool
    :param history:                  If True then enable a "history" feature that will display previous entries used. Uses settings filename provided or default if none provided
    :type history:                   bool
    :param show_hidden:              If True then enables the checkbox in the system dialog to select hidden files to be shown
    :type show_hidden:               bool
    :param history_setting_filename: Filename to use for the User Settings. Will store list of previous entries in this settings file
    :type history_setting_filename:  (str)
    :return:                         string representing the file(s) chosen, None if cancelled or window closed with X
    :rtype:                          str | None
    """

    # First setup the history settings file if history feature is enabled
    if history and history_setting_filename is not None:
        try:
            history_settings = UserSettings(history_setting_filename)
        except Exception as e:
            _error_popup_with_traceback(
                "popup_get_file - Something is wrong with your supplied history settings filename",
                "Exception: {}".format(e),
            )
            return None
    elif history:
        history_settings_filename = os.path.basename(inspect.stack()[1].filename)
        history_settings_filename = (
                os.path.splitext(history_settings_filename)[0] + ".json"
        )
        history_settings = UserSettings(history_settings_filename)
    else:
        history_settings = None

    if icon is None:
        icon = Window._user_defined_icon or DEFAULT_BASE64_ICON
    if no_window:
        _get_hidden_master_root()
        root = tk.Toplevel()

        try:
            root.attributes(
                "-alpha", 0
            )  # hide window while building it. makes for smoother 'paint'
            # if not running_mac():
            try:
                root.wm_overrideredirect(True)
            except Exception as e:
                print("* Error performing wm_overrideredirect in get file *", e)
            root.withdraw()
        except Exception:
            pass

        if not show_hidden:
            try:
                # call a dummy dialog with an impossible option to initialize the file
                # dialog without really getting a dialog window; this will throw a
                # TclError, so we need a try...except :
                try:
                    root.tk.call("tk_getOpenFile", "-foobarbaz")
                except tk.TclError:
                    pass
                # now set the magic variables accordingly
                root.tk.call("set", "::tk::dialog::file::showHiddenBtn", "1")
                root.tk.call("set", "::tk::dialog::file::showHiddenVar", "0")
            except Exception:
                pass

        if root and icon is not None:
            _set_icon_for_tkinter_window(root, icon=icon)
        # for Macs, setting parent=None fixes a warning problem.
        if save_as:
            if running_mac():
                is_all = [
                    (x, y) for (x, y) in file_types if all(ch in "* ." for ch in y)
                ]
                if not len(set(file_types)) > 1 and (
                        len(is_all) != 0 or file_types == FILE_TYPES_ALL_FILES
                ):
                    filename = tk.filedialog.asksaveasfilename(
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get file' dialog box
                else:
                    filename = tk.filedialog.asksaveasfilename(
                        filetypes=file_types,
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get file' dialog box
            else:
                filename = tk.filedialog.asksaveasfilename(
                    filetypes=file_types,
                    initialdir=initial_folder,
                    initialfile=default_path,
                    parent=root,
                    defaultextension=default_extension,
                )  # show the 'get file' dialog box
        elif multiple_files:
            if running_mac():
                is_all = [
                    (x, y) for (x, y) in file_types if all(ch in "* ." for ch in y)
                ]
                if not len(set(file_types)) > 1 and (
                        len(is_all) != 0 or file_types == FILE_TYPES_ALL_FILES
                ):
                    filename = tk.filedialog.askopenfilenames(
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get file' dialog box
                else:
                    filename = tk.filedialog.askopenfilenames(
                        filetypes=file_types,
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get file' dialog box
            else:
                filename = tk.filedialog.askopenfilenames(
                    filetypes=file_types,
                    initialdir=initial_folder,
                    initialfile=default_path,
                    parent=root,
                    defaultextension=default_extension,
                )  # show the 'get file' dialog box
        else:
            if running_mac():
                is_all = [
                    (x, y) for (x, y) in file_types if all(ch in "* ." for ch in y)
                ]
                if not len(set(file_types)) > 1 and (
                        len(is_all) != 0 or file_types == FILE_TYPES_ALL_FILES
                ):
                    filename = tk.filedialog.askopenfilename(
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get files' dialog box
                else:
                    filename = tk.filedialog.askopenfilename(
                        filetypes=file_types,
                        initialdir=initial_folder,
                        initialfile=default_path,
                        defaultextension=default_extension,
                    )  # show the 'get files' dialog box
            else:
                filename = tk.filedialog.askopenfilename(
                    filetypes=file_types,
                    initialdir=initial_folder,
                    initialfile=default_path,
                    parent=root,
                    defaultextension=default_extension,
                )  # show the 'get files' dialog box
        root.destroy()

        if not multiple_files and type(filename) in (tuple, list):
            if len(filename):  # only if not 0 length, otherwise will get an error
                filename = filename[0]
        if not filename:
            return None
        return filename

    if save_as:
        browse_button = SaveAs(
            file_types=file_types,
            initial_folder=initial_folder,
            default_extension=default_extension,
        )
    elif multiple_files:
        browse_button = FilesBrowse(
            file_types=file_types,
            initial_folder=initial_folder,
            files_delimiter=files_delimiter,
        )
    else:
        browse_button = FileBrowse(file_types=file_types, initial_folder=initial_folder)

    if image is not None:
        if isinstance(image, str):
            layout = [[Image(filename=image)]]
        else:
            layout = [[Image(data=image)]]
    else:
        layout = [[]]

    layout += [
        [
            Text(
                message,
                auto_size_text=True,
                text_color=text_color,
                background_color=background_color,
            )
        ]
    ]

    if not history:
        layout += [
            [
                InputText(default_text=default_path, size=size, key="-INPUT-"),
                browse_button,
            ]
        ]
    else:
        file_list = history_settings.get("-PSG file list-", [])
        last_entry = file_list[0] if file_list else ""
        layout += [
            [
                Combo(
                    file_list,
                    default_value=last_entry,
                    key="-INPUT-",
                    size=size if size != (None, None) else (80, 1),
                    bind_return_key=True,
                ),
                browse_button,
                Button(
                    "Clear History",
                    tooltip="Clears the list of files shown in the combobox",
                ),
            ]
        ]

    layout += [
        [Button("Ok", size=(6, 1), bind_return_key=True), Button("Cancel", size=(6, 1))]
    ]

    window = Window(
        title=title or message,
        layout=layout,
        icon=icon,
        auto_size_text=True,
        button_color=button_color,
        font=font,
        background_color=background_color,
        no_titlebar=no_titlebar,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        modal=modal,
        finalize=True,
    )

    if running_linux() and show_hidden is True:
        window.TKroot.tk.eval(
            "catch {tk_getOpenFile -badoption}"
        )  # dirty hack to force autoloading of Tk's file dialog code
        window.TKroot.setvar(
            "::tk::dialog::file::showHiddenBtn", 1
        )  # enable the "show hidden files" checkbox (it's necessary)
        window.TKroot.setvar(
            "::tk::dialog::file::showHiddenVar", 0
        )  # start with the hidden files... well... hidden

    while True:
        event, values = window.read()
        if event in ("Cancel", WIN_CLOSED):
            break
        elif event == "Clear History":
            history_settings.set("-PSG file list-", [])
            window["-INPUT-"].update("", [])
            popup_quick_message(
                "History of Previous Choices Cleared",
                background_color="red",
                text_color="white",
                font="_ 20",
                keep_on_top=True,
            )
        elif event in ("Ok", "-INPUT-"):
            if values["-INPUT-"] != "":
                if history_settings is not None:
                    list_of_entries = history_settings.get("-PSG file list-", [])
                    if values["-INPUT-"] in list_of_entries:
                        list_of_entries.remove(values["-INPUT-"])
                    list_of_entries.insert(0, values["-INPUT-"])
                    history_settings.set("-PSG file list-", list_of_entries)
            break

    window.close()
    del window
    if event in ("Cancel", WIN_CLOSED):
        return None

    return values["-INPUT-"]


# --------------------------- popup_get_text ---------------------------


def popup_get_text(
        message,
        title=None,
        default_text="",
        password_char="",
        size=(None, None),
        button_color=None,
        background_color=None,
        text_color=None,
        icon=None,
        font=None,
        no_titlebar=False,
        grab_anywhere=False,
        keep_on_top=None,
        location=(None, None),
        relative_location=(None, None),
        image=None,
        history=False,
        history_setting_filename=None,
        modal=True,
):
    """
    Display Popup with text entry field. Returns the text entered or None if closed / cancelled

    :param message:                  message displayed to user
    :type message:                   (str)
    :param title:                    Window title
    :type title:                     (str)
    :param default_text:             default value to put into input area
    :type default_text:              (str)
    :param password_char:            character to be shown instead of actually typed characters. WARNING - if history=True then can't hide passwords
    :type password_char:             (str)
    :param size:                     (width, height) of the InputText Element
    :type size:                      (int, int)
    :param button_color:             Color of the button (text, background)
    :type button_color:              (str, str) | str
    :param background_color:         background color of the entire window
    :type background_color:          (str)
    :param text_color:               color of the message text
    :type text_color:                (str)
    :param icon:                     filename or base64 string to be used for the window's icon
    :type icon:                      bytes | str
    :param font:                     specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                      (str or (str, int[, str]) or None)
    :param no_titlebar:              If True no titlebar will be shown
    :type no_titlebar:               (bool)
    :param grab_anywhere:            If True can click and drag anywhere in the window to move the window
    :type grab_anywhere:             (bool)
    :param keep_on_top:              If True the window will remain above all current windows
    :type keep_on_top:               (bool)
    :param location:                 (x,y) Location on screen to display the upper left corner of window
    :type location:                  (int, int)
    :param relative_location:        (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:         (int, int)
    :param image:                    Image to include at the top of the popup window
    :type image:                     (str) or (bytes)
    :param history:                  If True then enable a "history" feature that will display previous entries used. Uses settings filename provided or default if none provided
    :type history:                   bool
    :param history_setting_filename: Filename to use for the User Settings. Will store list of previous entries in this settings file
    :type history_setting_filename:  (str)
    :param modal:                    If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                     bool
    :return:                         Text entered or None if window was closed or cancel button clicked
    :rtype:                          str | None
    """

    # First setup the history settings file if history feature is enabled
    if history and history_setting_filename is not None:
        try:
            history_settings = UserSettings(history_setting_filename)
        except Exception as e:
            _error_popup_with_traceback(
                "popup_get_file - Something is wrong with your supplied history settings filename",
                "Exception: {}".format(e),
            )
            return None
    elif history:
        history_settings_filename = os.path.basename(inspect.stack()[1].filename)
        history_settings_filename = (
                os.path.splitext(history_settings_filename)[0] + ".json"
        )
        history_settings = UserSettings(history_settings_filename)
    else:
        history_settings = None

    if image is not None:
        if isinstance(image, str):
            layout = [[Image(filename=image)]]
        else:
            layout = [[Image(data=image)]]
    else:
        layout = [[]]

    layout += [
        [
            Text(
                message,
                auto_size_text=True,
                text_color=text_color,
                background_color=background_color,
            )
        ]
    ]
    if not history:
        layout += [
            [
                InputText(
                    default_text=default_text,
                    size=size,
                    key="-INPUT-",
                    password_char=password_char,
                )
            ]
        ]
    else:
        text_list = history_settings.get("-PSG text list-", [])
        last_entry = text_list[0] if text_list else default_text
        layout += [
            [
                Combo(
                    text_list,
                    default_value=last_entry,
                    key="-INPUT-",
                    size=size if size != (None, None) else (80, 1),
                    bind_return_key=True,
                ),
                Button(
                    "Clear History",
                    tooltip="Clears the list of files shown in the combobox",
                ),
            ]
        ]

    layout += [
        [Button("Ok", size=(6, 1), bind_return_key=True), Button("Cancel", size=(6, 1))]
    ]

    window = Window(
        title=title or message,
        layout=layout,
        icon=icon,
        auto_size_text=True,
        button_color=button_color,
        no_titlebar=no_titlebar,
        background_color=background_color,
        grab_anywhere=grab_anywhere,
        keep_on_top=keep_on_top,
        location=location,
        relative_location=relative_location,
        finalize=True,
        modal=modal,
        font=font,
    )

    while True:
        event, values = window.read()
        if event in ("Cancel", WIN_CLOSED):
            break
        elif event == "Clear History":
            history_settings.set("-PSG text list-", [])
            window["-INPUT-"].update("", [])
            popup_quick_message(
                "History of Previous Choices Cleared",
                background_color="red",
                text_color="white",
                font="_ 20",
                keep_on_top=True,
            )
        elif event in ("Ok", "-INPUT-"):
            if values["-INPUT-"] != "":
                if history_settings is not None:
                    list_of_entries = history_settings.get("-PSG text list-", [])
                    if values["-INPUT-"] in list_of_entries:
                        list_of_entries.remove(values["-INPUT-"])
                    list_of_entries.insert(0, values["-INPUT-"])
                    history_settings.set("-PSG text list-", list_of_entries)
            break

    window.close()
    del window
    if event in ("Cancel", WIN_CLOSED):
        return None
    else:
        text = values["-INPUT-"]
        return text


def popup_get_date(
        start_mon=None,
        start_day=None,
        start_year=None,
        begin_at_sunday_plus=0,
        no_titlebar=True,
        title="Choose Date",
        keep_on_top=True,
        location=(None, None),
        relative_location=(None, None),
        close_when_chosen=False,
        icon=None,
        locale=None,
        month_names=None,
        day_abbreviations=None,
        day_font="TkFixedFont 9",
        mon_year_font="TkFixedFont 10",
        arrow_font="TkFixedFont 7",
        modal=True,
):
    """
    Display a calendar window, get the user's choice, return as a tuple (mon, day, year)

    :param start_mon:            The starting month
    :type start_mon:             (int)
    :param start_day:            The starting day - optional. Set to None or 0 if no date to be chosen at start
    :type start_day:             int | None
    :param start_year:           The starting year
    :type start_year:            (int)
    :param begin_at_sunday_plus: Determines the left-most day in the display. 0=sunday, 1=monday, etc
    :type begin_at_sunday_plus:  (int)
    :param icon:                 Same as Window icon parameter. Can be either a filename or Base64 value. For Windows if filename, it MUST be ICO format. For Linux, must NOT be ICO
    :type icon:                  (str | bytes)
    :param location:             (x,y) location on the screen to place the top left corner of your window. Default is to center on screen
    :type location:              (int, int)
    :param relative_location:    (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:     (int, int)
    :param title:                Title that will be shown on the window
    :type title:                 (str)
    :param close_when_chosen:    If True, the window will close and function return when a day is clicked
    :type close_when_chosen:     (bool)
    :param locale:               locale used to get the day names
    :type locale:                (str)
    :param no_titlebar:          If True no titlebar will be shown
    :type no_titlebar:           (bool)
    :param keep_on_top:          If True the window will remain above all current windows
    :type keep_on_top:           (bool)
    :param month_names:          optional list of month names to use (should be 12 items)
    :type month_names:           List[str]
    :param day_abbreviations:    optional list of abbreviations to display as the day of week
    :type day_abbreviations:     List[str]
    :param day_font:             Font and size to use for the calendar
    :type day_font:              str | tuple
    :param mon_year_font:        Font and size to use for the month and year at the top
    :type mon_year_font:         str | tuple
    :param arrow_font:           Font and size to use for the arrow buttons
    :type arrow_font:            str | tuple
    :param modal:                If True then makes the popup will behave like a Modal window... all other windows are non-operational until this one is closed. Default = True
    :type modal:                 bool
    :return:                     Tuple containing (month, day, year) of chosen date or None if was cancelled
    :rtype:                      None | (int, int, int)
    """

    if month_names is not None and len(month_names) != 12:
        if not SUPPRESS_ERROR_POPUPS:
            popup_error(
                "Incorrect month names list specified. Must have 12 entries.",
                "Your list:",
                month_names,
            )

    if day_abbreviations is not None and len(day_abbreviations) != 7:
        if not SUPPRESS_ERROR_POPUPS:
            popup_error(
                "Incorrect day abbreviation list. Must have 7 entries.",
                "Your list:",
                day_abbreviations,
            )

    now = datetime.datetime.now()
    cur_month, cur_day, cur_year = now.month, now.day, now.year
    cur_month = start_mon or cur_month
    if start_mon is not None:
        cur_day = start_day
    else:
        cur_day = cur_day
    cur_year = start_year or cur_year

    def update_days(window, month, year, begin_at_sunday_plus):
        [window[(week, day)].update("") for day in range(7) for week in range(6)]
        weeks = calendar.monthcalendar(year, month)
        month_days = list(
            itertools.chain.from_iterable(
                [[0 for _ in range(8 - begin_at_sunday_plus)]] + weeks
            )
        )
        if month_days[6] == 0:
            month_days = month_days[7:]
            if month_days[6] == 0:
                month_days = month_days[7:]
        for i, day in enumerate(month_days):
            offset = i
            if offset >= 6 * 7:
                break
            window[(offset // 7, offset % 7)].update(str(day) if day else "")

    def make_days_layout():
        days_layout = []
        for week in range(6):
            row = []
            for day in range(7):
                row.append(
                    T(
                        "",
                        size=(4, 1),
                        justification="c",
                        font=day_font,
                        key=(week, day),
                        enable_events=True,
                        pad=(0, 0),
                    )
                )
            days_layout.append(row)
        return days_layout

    # Create table of month names and week day abbreviations

    if day_abbreviations is None or len(day_abbreviations) != 7:
        fwday = calendar.SUNDAY
        try:
            if locale is not None:
                _cal = calendar.LocaleTextCalendar(fwday, locale)
            else:
                _cal = calendar.TextCalendar(fwday)
            day_names = _cal.formatweekheader(3).split()
        except Exception as e:
            print("Exception building day names from locale", locale, e)
            day_names = ("Sun", "Mon", "Tue", "Wed", "Th", "Fri", "Sat")
    else:
        day_names = day_abbreviations

    mon_names = (
        month_names
        if month_names is not None and len(month_names) == 12
        else [calendar.month_name[i] for i in range(1, 13)]
    )
    days_layout = make_days_layout()

    layout = [
        [
            B(
                "◄◄",
                font=arrow_font,
                border_width=0,
                key="-YEAR-DOWN-",
                pad=((10, 2), 2),
            ),
            B("◄", font=arrow_font, border_width=0, key="-MON-DOWN-", pad=(0, 2)),
            Text(
                "{} {}".format(mon_names[cur_month - 1], cur_year),
                size=(16, 1),
                justification="c",
                font=mon_year_font,
                key="-MON-YEAR-",
                pad=(0, 2),
            ),
            B("►", font=arrow_font, border_width=0, key="-MON-UP-", pad=(0, 2)),
            B("►►", font=arrow_font, border_width=0, key="-YEAR-UP-", pad=(2, 2)),
        ]
    ]
    layout += [
        [
            Col(
                [
                    [
                        T(
                            day_names[i - (7 - begin_at_sunday_plus) % 7],
                            size=(4, 1),
                            font=day_font,
                            background_color=theme_text_color(),
                            text_color=theme_background_color(),
                            pad=(0, 0),
                        )
                        for i in range(7)
                    ]
                ],
                background_color=theme_text_color(),
                pad=(0, 0),
            )
        ]
    ]
    layout += days_layout
    if not close_when_chosen:
        layout += [
            [
                Button("Ok", border_width=0, font="TkFixedFont 8"),
                Button("Cancel", border_width=0, font="TkFixedFont 8"),
            ]
        ]

    window = Window(
        title,
        layout,
        no_titlebar=no_titlebar,
        grab_anywhere=True,
        keep_on_top=keep_on_top,
        font="TkFixedFont 12",
        use_default_focus=False,
        location=location,
        relative_location=relative_location,
        finalize=True,
        icon=icon,
    )

    update_days(window, cur_month, cur_year, begin_at_sunday_plus)

    prev_choice = chosen_mon_day_year = None

    if cur_day:
        chosen_mon_day_year = cur_month, cur_day, cur_year
        for week in range(6):
            for day in range(7):
                if window[(week, day)].DisplayText == str(cur_day):
                    window[(week, day)].update(
                        background_color=theme_text_color(),
                        text_color=theme_background_color(),
                    )
                    prev_choice = (week, day)
                    break

    if modal or DEFAULT_MODAL_WINDOWS_FORCED:
        window.make_modal()

    while True:  # Event Loop
        event, values = window.read()
        if event in (None, "Cancel"):
            chosen_mon_day_year = None
            break
        if event == "Ok":
            break
        if event in ("-MON-UP-", "-MON-DOWN-", "-YEAR-UP-", "-YEAR-DOWN-"):
            cur_month += event == "-MON-UP-"
            cur_month -= event == "-MON-DOWN-"
            cur_year += event == "-YEAR-UP-"
            cur_year -= event == "-YEAR-DOWN-"
            if cur_month > 12:
                cur_month = 1
                cur_year += 1
            elif cur_month < 1:
                cur_month = 12
                cur_year -= 1
            window["-MON-YEAR-"].update(
                "{} {}".format(mon_names[cur_month - 1], cur_year)
            )
            update_days(window, cur_month, cur_year, begin_at_sunday_plus)
            if prev_choice:
                window[prev_choice].update(
                    background_color=theme_background_color(),
                    text_color=theme_text_color(),
                )
        elif type(event) is tuple:
            if window[event].DisplayText != "":
                chosen_mon_day_year = (
                    cur_month,
                    int(window[event].DisplayText),
                    cur_year,
                )
                if prev_choice:
                    window[prev_choice].update(
                        background_color=theme_background_color(),
                        text_color=theme_text_color(),
                    )
                window[event].update(
                    background_color=theme_text_color(),
                    text_color=theme_background_color(),
                )
                prev_choice = event
                if close_when_chosen:
                    break
    window.close()
    return chosen_mon_day_year


# --------------------------- PopupAnimated ---------------------------


def popup_animated(
        image_source,
        message=None,
        background_color=None,
        text_color=None,
        font=None,
        no_titlebar=True,
        grab_anywhere=True,
        keep_on_top=True,
        location=(None, None),
        relative_location=(None, None),
        alpha_channel=None,
        time_between_frames=0,
        transparent_color=None,
        title="",
        icon=None,
        no_buffering=False,
):
    """
     Show animation one frame at a time.  This function has its own internal clocking meaning you can call it at any frequency
     and the rate the frames of video is shown remains constant.  Maybe your frames update every 30 ms but your
     event loop is running every 10 ms.  You don't have to worry about delaying, just call it every time through the
     loop.

    :param image_source:        Either a filename or a base64 string. Use None to close the window.
    :type image_source:         str | bytes | None
    :param message:             An optional message to be shown with the animation
    :type message:              (str)
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 str | tuple
    :param no_titlebar:         If True then the titlebar and window frame will not be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True then you can move the window just clicking anywhere on window, hold and drag
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True then Window will remain on top of all other windows currently shownn
    :type keep_on_top:          (bool)
    :param location:            (x,y) location on the screen to place the top left corner of your window. Default is to center on screen
    :type location:             (int, int)
    :param relative_location:   (x,y) location relative to the default location of the window, in pixels. Normally the window centers.  This location is relative to the location the window would be created. Note they can be negative.
    :type relative_location:    (int, int)
    :param alpha_channel:       Window transparency 0 = invisible 1 = completely visible. Values between are see through
    :type alpha_channel:        (float)
    :param time_between_frames: Amount of time in milliseconds between each frame
    :type time_between_frames:  (int)
    :param transparent_color:   This color will be completely see-through in your window. Can even click through
    :type transparent_color:    (str)
    :param title:               Title that will be shown on the window
    :type title:                (str)
    :param icon:                Same as Window icon parameter. Can be either a filename or Base64 byte string. For Windows if filename, it MUST be ICO format. For Linux, must NOT be ICO
    :type icon:                 str | bytes
    :param no_buffering:        If True then no buffering will be used for the GIF. May work better if you have a large animation
    :type no_buffering:         (bool)
    :return:                    True if the window updated OK. False if the window was closed
    :rtype:                     bool
    """
    if image_source is None:
        for image in Window._animated_popup_dict:
            window = Window._animated_popup_dict[image]
            window.close()
        Window._animated_popup_dict = {}
        return

    if image_source not in Window._animated_popup_dict:
        if type(image_source) is bytes or len(image_source) > 300:
            layout = [
                [
                    Image(
                        data=image_source,
                        background_color=background_color,
                        key="-IMAGE-",
                    )
                ],
            ]
        else:
            layout = [
                [
                    Image(
                        filename=image_source,
                        background_color=background_color,
                        key="-IMAGE-",
                    )
                ],
            ]
        if message:
            layout.append(
                [
                    Text(
                        message,
                        background_color=background_color,
                        text_color=text_color,
                        font=font,
                    )
                ]
            )

        window = Window(
            title,
            layout,
            no_titlebar=no_titlebar,
            grab_anywhere=grab_anywhere,
            keep_on_top=keep_on_top,
            background_color=background_color,
            location=location,
            alpha_channel=alpha_channel,
            element_padding=(0, 0),
            margins=(0, 0),
            transparent_color=transparent_color,
            finalize=True,
            element_justification="c",
            icon=icon,
            relative_location=relative_location,
        )
        Window._animated_popup_dict[image_source] = window
    else:
        window = Window._animated_popup_dict[image_source]
        if no_buffering:
            window["-IMAGE-"].update_animation_no_buffering(
                image_source, time_between_frames=time_between_frames
            )
        else:
            window["-IMAGE-"].update_animation(
                image_source, time_between_frames=time_between_frames
            )
    event, values = window.read(1)
    if event == WIN_CLOSED:
        return False
    # window.refresh() # call refresh instead of Read to save significant CPU time
    return True


# Popup Notify
def popup_notify(
        *args,
        title="",
        icon=SYSTEM_TRAY_MESSAGE_ICON_INFORMATION,
        display_duration_in_ms=SYSTEM_TRAY_MESSAGE_DISPLAY_DURATION_IN_MILLISECONDS,
        fade_in_duration=SYSTEM_TRAY_MESSAGE_FADE_IN_DURATION,
        alpha=0.9,
        location=None,
):
    """
    Displays a "notification window", usually in the bottom right corner of your display.  Has an icon, a title, and a message.  It is more like a "toaster" window than the normal popups.

    The window will slowly fade in and out if desired.  Clicking on the window will cause it to move through the end the current "phase". For example, if the window was fading in and it was clicked, then it would immediately stop fading in and instead be fully visible.  It's a way for the user to quickly dismiss the window.

    The return code specifies why the call is returning (e.g. did the user click the message to dismiss it)

    :param title:                  Text to be shown at the top of the window in a larger font
    :type title:                   (str)
    :param icon:                   A base64 encoded PNG/GIF image or PNG/GIF filename that will be displayed in the window
    :type icon:                    bytes | str
    :param display_duration_in_ms: Number of milliseconds to show the window
    :type display_duration_in_ms:  (int)
    :param fade_in_duration:       Number of milliseconds to fade window in and out
    :type fade_in_duration:        (int)
    :param alpha:                  Alpha channel. 0 - invisible 1 - fully visible
    :type alpha:                   (float)
    :param location:               Location on the screen to display the window
    :type location:                (int, int)
    :return:                       reason for returning
    :rtype:                        (int)
    """

    if not args:
        args_to_print = [""]
    else:
        args_to_print = args
    output = ""
    max_line_total, total_lines, local_line_width = (
        0,
        0,
        SYSTEM_TRAY_MESSAGE_MAX_LINE_LENGTH,
    )
    for message in args_to_print:
        # fancy code to check if string and convert if not is not need. Just always convert to string :-)
        # if not isinstance(message, str): message = str(message)
        message = str(message)
        if message.count("\n"):
            message_wrapped = message
        else:
            message_wrapped = textwrap.fill(message, local_line_width)
        message_wrapped_lines = message_wrapped.count("\n") + 1
        longest_line_len = max([len(l) for l in message.split("\n")])
        width_used = min(longest_line_len, local_line_width)
        max_line_total = max(max_line_total, width_used)
        # height = _GetNumLinesNeeded(message, width_used)
        height = message_wrapped_lines
        output += message_wrapped + "\n"
        total_lines += height

    message = output

    # def __init__(self, menu=None, filename=None, data=None, data_base64=None, tooltip=None, metadata=None):
    return SystemTray.notify(
        title=title,
        message=message,
        icon=icon,
        display_duration_in_ms=display_duration_in_ms,
        fade_in_duration=fade_in_duration,
        alpha=alpha,
        location=location,
    )


def popup_menu(window, element, menu_def, title=None, location=(None, None)):
    """
    Makes a "popup menu"
    This type of menu is what you get when a normal menu or a right click menu is torn off
    The settings for the menu are obtained from the window parameter's Window


    :param window:   The window associated with the popup menu. The theme and right click menu settings for this window will be used
    :type window:    Window
    :param element:  An element in your window to associate the menu to. It can be any element
    :type element:   Element
    :param menu_def: A menu definition. This will be the same format as used for Right Click Menus1
    :type  menu_def: List[List[ List[str] | str ]]
    :param title:    The title that will be shown on the torn off menu window. Defaults to window titlr
    :type title:     str
    :param location: The location on the screen to place the window
    :type location:  (int, int) | (None, None)
    """

    element._popup_menu_location = location
    top_menu = tk.Menu(
        window.TKroot, tearoff=True, tearoffcommand=element._tearoff_menu_callback
    )
    if window.right_click_menu_background_color not in (COLOR_SYSTEM_DEFAULT, None):
        top_menu.config(bg=window.right_click_menu_background_color)
    if window.right_click_menu_text_color not in (COLOR_SYSTEM_DEFAULT, None):
        top_menu.config(fg=window.right_click_menu_text_color)
    if window.right_click_menu_disabled_text_color not in (COLOR_SYSTEM_DEFAULT, None):
        top_menu.config(disabledforeground=window.right_click_menu_disabled_text_color)
    if window.right_click_menu_font is not None:
        top_menu.config(font=window.right_click_menu_font)
    if window.right_click_menu_selected_colors[0] != COLOR_SYSTEM_DEFAULT:
        top_menu.config(activeforeground=window.right_click_menu_selected_colors[0])
    if window.right_click_menu_selected_colors[1] != COLOR_SYSTEM_DEFAULT:
        top_menu.config(activebackground=window.right_click_menu_selected_colors[1])
    top_menu.config(title=window.Title if title is None else title)
    AddMenuItem(top_menu, menu_def[1], element, right_click_menu=True)
    # element.Widget.bind('<Button-3>', element._RightClickMenuCallback)
    top_menu.invoke(0)


def popup_error_with_traceback(title, *messages, emoji=None):
    """
    Show an error message and as many additoinal lines of messages as you want.
    Will show the same error window as PySimpleGUI uses internally.  Has a button to
    take the user to the line of code you called this popup from.
    If you include the Exception information in your messages, then it will be parsed and additional information
    will be in the window about such as the specific line the error itself occurred on.

    :param title:     The title that will be shown in the popup's titlebar and in the first line of the window
    :type title:      str
    :param emoji:     An optional BASE64 Encoded image to shows in the error window
    :type emoji:      bytes
    """

    # For now, call the function that PySimpleGUI uses internally
    _error_popup_with_traceback(str(title), *messages, emoji=emoji)


def _error_popup_with_traceback(title, *args, emoji=None):
    if SUPPRESS_ERROR_POPUPS:
        return
    trace_details = traceback.format_stack()
    error_message = ""
    file_info_pysimplegui = None
    for line in reversed(trace_details):
        if __file__ not in line:
            file_info_pysimplegui = line.split(",")[0]
            error_message = line
            break
    if file_info_pysimplegui is None:
        _error_popup_with_code(
            title, None, None, "Did not find your traceback info", *args, emoji=emoji
        )
        return

    error_parts = None
    if error_message != "":
        error_parts = error_message.split(", ")
        if len(error_parts) < 4:
            error_message = (
                    error_parts[0] + "\n" + error_parts[1] + "\n" + "".join(error_parts[2:])
            )
    if error_parts is None:
        print("*** Error popup attempted but unable to parse error details ***")
        print(trace_details)
        return
    filename = error_parts[0][error_parts[0].index("File ") + 5:]
    line_num = error_parts[1][error_parts[1].index("line ") + 5:]
    _error_popup_with_code(title, filename, line_num, error_message, *args, emoji=emoji)


def _error_popup_with_code(title, filename, line_num, *args, emoji=None):
    """
    Makes the error popup window

    :param title:     The title that will be shown in the popup's titlebar and in the first line of the window
    :type title:      str
    :param filename:  The filename to show.. may not be the filename that actually encountered the exception!
    :type filename:   str
    :param line_num:  Line number within file with the error
    :type line_num:   int | str
    :param args:      A variable number of lines of messages
    :type args:       *Any
    :param emoji:     An optional BASE64 Encoded image to shows in the error window
    :type emoji:      bytes
    """
    editor_filename = execute_get_editor()
    emoji_data = emoji if emoji is not None else _random_error_emoji()
    layout = [[Text("ERROR"), Text(title)], [Image(data=emoji_data)]]
    lines = []
    for msg in args:
        if isinstance(msg, Exception):
            lines += [
                [
                    "Additional Exception info pased in by PySimpleGUI or user: Error type is: {}".format(
                        type(msg).__name__
                    )
                ]
            ]
            lines += [
                [
                    "In file {} Line number {}".format(
                        __file__, msg.__traceback__.tb_lineno
                    )
                ]
            ]
            lines += [[str(msg)]]
        else:
            lines += [str(msg).split("\n")]
    max_line_len = 0
    for line in lines:
        max_line_len = max(max_line_len, max([len(s) for s in line]))

    layout += [
        [Text("".join(line), size=(min(max_line_len, 90), None))] for line in lines
    ]
    layout += [
        [
            Button("Close"),
            Button("Take me to error", disabled=True if not editor_filename else False),
            Button("Kill Application", button_color="white on red"),
        ]
    ]
    if not editor_filename:
        layout += [
            [
                Text(
                    'Configure editor in the Global settings to enable "Take me to error" feature'
                )
            ]
        ]
    window = Window(title, layout, keep_on_top=True)

    while True:
        event, values = window.read()
        if event in ("Close", WIN_CLOSED):
            break
        if event == "Kill Application":
            window.close()
            popup_quick_message(
                "KILLING APP!  BYE!",
                font="_ 18",
                keep_on_top=True,
                text_color="white",
                background_color="red",
                non_blocking=False,
            )
            sys.exit()
        if (
                event == "Take me to error"
                and filename is not None
                and line_num is not None
        ):
            execute_editor(filename, line_num)

    window.close()


#####################################################################
# Animated window while shell command is executed
#####################################################################


def _process_thread(*args):
    global __shell_process__

    # start running the command with arugments
    try:
        __shell_process__ = subprocess.run(args, shell=True, stdout=subprocess.PIPE)
    except Exception as e:
        print("Exception running process args = {}".format(args))
        __shell_process__ = None


def shell_with_animation(
        command,
        args=None,
        image_source=DEFAULT_BASE64_LOADING_GIF,
        message=None,
        background_color=None,
        text_color=None,
        font=None,
        no_titlebar=True,
        grab_anywhere=True,
        keep_on_top=True,
        location=(None, None),
        alpha_channel=None,
        time_between_frames=100,
        transparent_color=None,
):
    """
    Execute a "shell command" (anything capable of being launched using subprocess.run) and
    while the command is running, show an animated popup so that the user knows that a long-running
    command is being executed.  Without this mechanism, the GUI appears locked up.

    :param command:             The command to run
    :type command:              (str)
    :param args:                List of arguments
    :type args:                 List[str]
    :param image_source:        Either a filename or a base64 string.
    :type image_source:         str | bytes
    :param message:             An optional message to be shown with the animation
    :type message:              (str)
    :param background_color:    color of background
    :type background_color:     (str)
    :param text_color:          color of the text
    :type text_color:           (str)
    :param font:                specifies the  font family, size, etc. Tuple or Single string format 'name size styles'. Styles: italic * roman bold normal underline overstrike
    :type font:                 str | tuple
    :param no_titlebar:         If True then the titlebar and window frame will not be shown
    :type no_titlebar:          (bool)
    :param grab_anywhere:       If True then you can move the window just clicking anywhere on window, hold and drag
    :type grab_anywhere:        (bool)
    :param keep_on_top:         If True then Window will remain on top of all other windows currently shownn
    :type keep_on_top:          (bool)
    :param location:            (x,y) location on the screen to place the top left corner of your window. Default is to center on screen
    :type location:             (int, int)
    :param alpha_channel:       Window transparency 0 = invisible 1 = completely visible. Values between are see through
    :type alpha_channel:        (float)
    :param time_between_frames: Amount of time in milliseconds between each frame
    :type time_between_frames:  (int)
    :param transparent_color:   This color will be completely see-through in your window. Can even click through
    :type transparent_color:    (str)
    :return:                    The resulting string output from stdout
    :rtype:                     (str)
    """

    global __shell_process__

    real_args = [command]
    if args is not None:
        for arg in args:
            real_args.append(arg)
        # real_args.append(args)
    thread = threading.Thread(target=_process_thread, args=real_args, daemon=True)
    thread.start()

    # Poll to see if the thread is still running.  If so, then continue showing the animation
    while True:
        popup_animated(
            image_source=image_source,
            message=message,
            time_between_frames=time_between_frames,
            transparent_color=transparent_color,
            text_color=text_color,
            background_color=background_color,
            font=font,
            no_titlebar=no_titlebar,
            grab_anywhere=grab_anywhere,
            keep_on_top=keep_on_top,
            location=location,
            alpha_channel=alpha_channel,
        )
        thread.join(timeout=time_between_frames / 1000)
        if not thread.is_alive():
            break
    popup_animated(None)  # stop running the animation

    output = __shell_process__.__str__().replace(
        "\\r\\n", "\n"
    )  # fix up the output string
    output = output[output.index("stdout=b'") + 9: -2]
    return output


#######################################################################
#  8888888888
#  888
#  888
#  8888888    888d888 888d888 .d88b.  888d888
#  888        888P"   888P"  d88""88b 888P"
#  888        888     888    888  888 888
#  888        888     888    Y88..88P 888
#  8888888888 888     888     "Y88P"  888
#
#
#
#  888b     d888
#  8888b   d8888
#  88888b.d88888
#  888Y88888P888  .d88b.  .d8888b  .d8888b   8888b.   .d88b.   .d88b.
#  888 Y888P 888 d8P  Y8b 88K      88K          "88b d88P"88b d8P  Y8b
#  888  Y8P  888 88888888 "Y8888b. "Y8888b. .d888888 888  888 88888888
#  888   "   888 Y8b.          X88      X88 888  888 Y88b 888 Y8b.
#  888       888  "Y8888   88888P'  88888P' "Y888888  "Y88888  "Y8888
#                                                         888
#                                                    Y8b d88P
#                                                     "Y88P"
# Code to make messages to help user find errors in their code
#######################################################################


def _create_error_message():
    """
    Creates an error message containing the filename and line number of the users
    code that made the call into PySimpleGUI
    :return: Error string to display with file, line number, and line of code
    :rtype:  str
    """

    called_func = inspect.stack()[1].function
    trace_details = traceback.format_stack()
    error_message = ""
    file_info_pysimplegui = trace_details[-1].split(",")[0]
    for line in reversed(trace_details):
        if line.split(",")[0] != file_info_pysimplegui:
            error_message = line
            break
    if error_message != "":
        error_parts = error_message.split(", ")
        if len(error_parts) < 4:
            error_message = (
                    error_parts[0] + "\n" + error_parts[1] + "\n" + "".join(error_parts[2:])
            )
    return (
            "The PySimpleGUI internal reporting function is "
            + called_func
            + "\n"
            + "The error originated from:\n"
            + error_message
    )


#   .d8888b.           888    888    d8b
#  d88P  Y88b          888    888    Y8P
#  Y88b.               888    888
#   "Y888b.    .d88b.  888888 888888 888 88888b.   .d88b.  .d8888b
#      "Y88b. d8P  Y8b 888    888    888 888 "88b d88P"88b 88K
#        "888 88888888 888    888    888 888  888 888  888 "Y8888b.
#  Y88b  d88P Y8b.     Y88b.  Y88b.  888 888  888 Y88b 888      X88
#   "Y8888P"   "Y8888   "Y888  "Y888 888 888  888  "Y88888  88888P'
#                                                      888
#                                                 Y8b d88P
#                                                  "Y88P"

# Interface to saving / loading user program settings in json format
# This is a new set of APIs supplied by PySimpleGUI that enables users to easily set/save/load individual
# settings.  They are automatically saved to a JSON file. If no file/path is specified then a filename is
# created from the source file filename.


class UserSettings:
    # A reserved settings object for use by the setting functions. It's a way for users
    # to access the user settings without diarectly using the UserSettings class
    _default_for_function_interface = None  # type: UserSettings

    def __init__(
            self,
            filename=None,
            path=None,
            silent_on_error=False,
            autosave=True,
            use_config_file=None,
            convert_bools_and_none=True,
    ):
        """
        User Settings

        :param filename:               The name of the file to use. Can be a full path and filename or just filename
        :type filename:                (str or None)
        :param path:                   The folder that the settings file will be stored in. Do not include the filename.
        :type path:                    (str or None)
        :param silent_on_error:        If True errors will not be reported
        :type silent_on_error:         (bool)
        :param autosave:               If True the settings file is saved after every update
        :type autosave:                (bool)
        :param use_config_file:        If True then the file format will be a config.ini rather than json
        :type use_config_file:         (bool)
        :param convert_bools_and_none: If True then "True", "False", "None" will be converted to the Python values True, False, None when using INI files. Default is TRUE
        :type convert_bools_and_none:  (bool)
        """

        self.path = path
        self.filename = filename
        self.full_filename = None
        self.dict = {}
        self.default_value = None
        self.silent_on_error = silent_on_error
        self.autosave = autosave
        if (
                filename is not None
                and filename.endswith(".ini")
                and use_config_file is None
        ):
            warnings.warn(
                "[UserSettings] You have specified a filename with .ini extension but did not set use_config_file. Setting use_config_file for you.",
                UserWarning,
            )
            use_config_file = True
        self.use_config_file = use_config_file
        # self.retain_config_comments = retain_config_comments
        self.convert_bools = convert_bools_and_none
        if use_config_file:
            self.config = configparser.ConfigParser()
            self.config.optionxform = str
            # self.config_dict = {}
            self.section_class_dict = {}  # type: dict[_SectionDict]
        if filename is not None or path is not None:
            self.load(filename=filename, path=path)

    ########################################################################################################
    ## FIRST is the _SectionDict helper class
    ## It is typically not directly accessed, although it is possible to call delete_section, get, set
    ########################################################################################################

    class _SectionDict:
        item_count = 0

        def __init__(
                self, section_name, section_dict, config, user_settings_parent
        ):  # (str, Dict, configparser.ConfigParser)
            """
            The Section Dictionary.  It holds the values for a section.

            :param section_name:                Name of the section
            :type section_name:                 str
            :param section_dict:                Dictionary of values for the section
            :type section_dict:                 dict
            :param config:                      The configparser object
            :type config:                       configparser.ConfigParser
            :param user_settings_parent:        The parent UserSettings object that hdas this section
            :type user_settings_parent:         UserSettings
            """
            self.section_name = section_name
            self.section_dict = section_dict  # type: Dict
            self.new_section = False
            self.config = config  # type: configparser.ConfigParser
            self.user_settings_parent = user_settings_parent  # type: UserSettings
            UserSettings._SectionDict.item_count += 1

            if self.user_settings_parent.convert_bools:
                for key, value in self.section_dict.items():
                    if value == "True":
                        value = True
                        self.section_dict[key] = value
                    elif value == "False":
                        value = False
                        self.section_dict[key] = value
                    elif value == "None":
                        value = None
                        self.section_dict[key] = value
            # print(f'++++++ making a new SectionDict with name = {section_name}')

        def __repr__(self):
            """
            Converts the settings dictionary into a string for easy display

            :return: the dictionary as a string
            :rtype:  (str)
            """
            return_string = "{}:\n".format(self.section_name)
            for entry in self.section_dict.keys():
                return_string += "          {} : {}\n".format(
                    entry, self.section_dict[entry]
                )

            return return_string

        def get(self, key, default=None):
            """
            Returns the value of a specified setting.  If the setting is not found in the settings dictionary, then
            the user specified default value will be returned.  It no default is specified and nothing is found, then
            the "default value" is returned.  This default can be specified in this call, or previously defined
            by calling set_default. If nothing specified now or previously, then None is returned as default.

            :param key:     Key used to lookup the setting in the settings dictionary
            :type key:      (Any)
            :param default: Value to use should the key not be found in the dictionary
            :type default:  (Any)
            :return:        Value of specified settings
            :rtype:         (Any)
            """
            value = self.section_dict.get(key, default)
            if self.user_settings_parent.convert_bools:
                if value == "True":
                    value = True
                elif value == "False":
                    value = False
            return value

        def set(self, key, value):
            value = str(value)  # all values must be strings
            if self.new_section:
                self.config.add_section(self.section_name)
                self.new_section = False
            self.config.set(section=self.section_name, option=key, value=value)
            self.section_dict[key] = value
            if self.user_settings_parent.autosave:
                self.user_settings_parent.save()

        def delete_section(self):
            # print(f'** Section Dict deleting section = {self.section_name}')
            self.config.remove_section(section=self.section_name)
            del self.user_settings_parent.section_class_dict[self.section_name]
            if self.user_settings_parent.autosave:
                self.user_settings_parent.save()

        def __getitem__(self, item):
            # print('*** In SectionDict Get ***')
            return self.get(item)

        def __setitem__(self, item, value):
            """
            Enables setting a setting by using [ ] notation like a dictionary.
            Your code will have this kind of design pattern:
            settings = sg.UserSettings()
            settings[item] = value

            :param item:  The key for the setting to change. Needs to be a hashable type. Basically anything but a list
            :type item:   Any
            :param value: The value to set the setting to
            :type value:  Any
            """
            # print(f'*** In SectionDict SET *** item = {item} value = {value}')
            self.set(item, value)
            self.section_dict[item] = value

        def __delitem__(self, item):
            """
            Delete an individual user setting.  This is the same as calling delete_entry.  The syntax
            for deleting the item using this manner is:
                del settings['entry']
            :param item: The key for the setting to delete
            :type item:  Any
            """
            # print(f'** In SectionDict delete! section name = {self.section_name} item = {item} ')
            self.config.remove_option(section=self.section_name, option=item)
            try:
                del self.section_dict[item]
            except Exception as e:
                pass
                # print(e)
            if self.user_settings_parent.autosave:
                self.user_settings_parent.save()

    ########################################################################################################

    def __repr__(self):
        """
        Converts the settings dictionary into a string for easy display

        :return: the dictionary as a string
        :rtype:  (str)
        """
        if not self.use_config_file:
            return pprint.pformat(self.dict)
        else:
            # rvalue = '-------------------- Settings ----------------------\n'
            rvalue = ""
            for name, section in self.section_class_dict.items():
                rvalue += str(section)

            # rvalue += '\n-------------------- Settings End----------------------\n'
            rvalue += "\n"
            return rvalue

    def set_default_value(self, default):
        """
        Set the value that will be returned if a requested setting is not found

        :param default: value to be returned if a setting is not found in the settings dictionary
        :type default:  Any
        """
        self.default_value = default

    def _compute_filename(self, filename=None, path=None):
        """
        Creates the full filename given the path or the filename or both.

        :param filename: The name of the file to use. Can be a full path and filename or just filename
        :type filename:  (str or None)
        :param path:     The folder that the settings file will be stored in. Do not include the filename.
        :type path:      (str or None)
        :return:         Tuple with (full filename, path, filename)
        :rtype:          Tuple[str, str, str]
        """
        if filename is not None:
            dirname_from_filename = os.path.dirname(
                filename
            )  # see if a path was provided as part of filename
            if dirname_from_filename:
                path = dirname_from_filename
                filename = os.path.basename(filename)
        elif self.filename is not None:
            filename = self.filename
        else:
            if not self.use_config_file:
                filename = (
                        os.path.splitext(
                            os.path.basename(sys.modules["__main__"].__file__)
                        )[0]
                        + ".json"
                )
            else:
                filename = (
                        os.path.splitext(
                            os.path.basename(sys.modules["__main__"].__file__)
                        )[0]
                        + ".ini"
                )

        if path is None:
            if self.path is not None:
                # path = self.path
                path = os.path.expanduser(
                    self.path
                )  # expand user provided path in case it has user ~ in it. Don't think it'll hurt
            elif (
                    DEFAULT_USER_SETTINGS_PATH is not None
            ):  # if user set the path manually system-wide using set options
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_PATH)
            elif running_trinket():
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_TRINKET_PATH)
            elif running_replit():
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_REPLIT_PATH)
            elif running_windows():
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_WIN_PATH)
            elif running_linux():
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_LINUX_PATH)
            elif running_mac():
                path = os.path.expanduser(DEFAULT_USER_SETTINGS_MAC_PATH)
            else:
                path = "."

        full_filename = os.path.join(path, filename)
        return full_filename, path, filename

    def set_location(self, filename=None, path=None):
        """
        Sets the location of the settings file

        :param filename: The name of the file to use. Can be a full path and filename or just filename
        :type filename:  (str or None)
        :param path:     The folder that the settings file will be stored in. Do not include the filename.
        :type path:      (str or None)
        """
        cfull_filename, cpath, cfilename = self._compute_filename(
            filename=filename, path=path
        )

        self.filename = cfilename
        self.path = cpath
        self.full_filename = cfull_filename

    def get_filename(self, filename=None, path=None):
        """
        Sets the filename and path for your settings file.  Either paramter can be optional.

        If you don't choose a path, one is provided for you that is OS specific
        Windows path default = users/name/AppData/Local/PySimpleGUI/settings.

        If you don't choose a filename, your application's filename + '.json' will be used.

        Normally the filename and path are split in the user_settings calls. However for this call they
        can be combined so that the filename contains both the path and filename.

        :param filename: The name of the file to use. Can be a full path and filename or just filename
        :type filename:  (str or None)
        :param path:     The folder that the settings file will be stored in. Do not include the filename.
        :type path:      (str or None)
        :return:         The full pathname of the settings file that has both the path and filename combined.
        :rtype:          (str)
        """
        if (
                filename is not None
                or path is not None
                or (filename is None and path is None and self.full_filename is None)
        ):
            self.set_location(filename=filename, path=path)
            self.read()
        return self.full_filename

    def save(self, filename=None, path=None):
        """
        Saves the current settings dictionary.  If a filename or path is specified in the call, then it will override any
        previously specitfied filename to create a new settings file.  The settings dictionary is then saved to the newly defined file.

        :param filename: The fFilename to save to. Can specify a path or just the filename. If no filename specified, then the caller's filename will be used.
        :type filename:  (str or None)
        :param path:     The (optional) path to use to save the file.
        :type path:      (str or None)
        :return:         The full path and filename used to save the settings
        :rtype:          (str)
        """
        if filename is not None or path is not None:
            self.set_location(filename=filename, path=path)
        try:
            if not os.path.exists(self.path):
                os.makedirs(self.path)
            with open(self.full_filename, "w") as f:
                if not self.use_config_file:
                    json.dump(self.dict, f)
                else:
                    self.config.write(f)
        except Exception as e:
            if not self.silent_on_error:
                _error_popup_with_traceback(
                    "UserSettings.save error",
                    "*** UserSettings.save()  Error saving settings to file:***\n",
                    self.full_filename,
                    e,
                )

        return self.full_filename

    def load(self, filename=None, path=None):
        """
        Specifies the path and filename to use for the settings and reads the contents of the file.
        The filename can be a full filename including a path, or the path can be specified separately.
        If  no filename is specified, then the caller's filename will be used with the extension ".json"

        :param filename: Filename to load settings from (and save to in the future)
        :type filename:  (str or None)
        :param path:     Path to the file. Defaults to a specific folder depending on the operating system
        :type path:      (str or None)
        :return:         The settings dictionary (i.e. all settings)
        :rtype:          (dict)
        """
        if filename is not None or path is not None or self.full_filename is None:
            self.set_location(filename, path)
        self.read()
        return self.dict

    def delete_file(self, filename=None, path=None, report_error=False):
        """
        Deltes the filename and path for your settings file.  Either paramter can be optional.
        If you don't choose a path, one is provided for you that is OS specific
        Windows path default = users/name/AppData/Local/PySimpleGUI/settings.
        If you don't choose a filename, your application's filename + '.json' will be used
        Also sets your current dictionary to a blank one.

        :param filename:     The name of the file to use. Can be a full path and filename or just filename
        :type filename:      (str or None)
        :param path:         The folder that the settings file will be stored in. Do not include the filename.
        :type path:          (str or None)
        :param report_error: Determines if an error should be shown if a delete error happen (i.e. file isn't present)
        :type report_error:  (bool)
        """

        if (
                filename is not None
                or path is not None
                or (filename is None and path is None)
        ):
            self.set_location(filename=filename, path=path)
        try:
            os.remove(self.full_filename)
        except Exception as e:
            if report_error:
                _error_popup_with_traceback(
                    "UserSettings delete_file warning ***",
                    "Exception trying to perform os.remove",
                    e,
                )
        self.dict = {}

    def write_new_dictionary(self, settings_dict):
        """
        Writes a specified dictionary to the currently defined settings filename.

        :param settings_dict: The dictionary to be written to the currently defined settings file
        :type settings_dict:  (dict)
        """
        if self.full_filename is None:
            self.set_location()
        self.dict = settings_dict
        self.save()

    def read(self):
        """
        Reads settings file and returns the dictionary.
        If you have anything changed in an existing settings dictionary, you will lose your changes.
        :return: settings dictionary
        :rtype:  (dict)
        """
        if self.full_filename is None:
            return {}
        try:
            if os.path.exists(self.full_filename):
                with open(self.full_filename, "r") as f:
                    if not self.use_config_file:  # if using json
                        self.dict = json.load(f)
                    else:  # if using a config file
                        self.config.read_file(f)
                        # Make a dictionary of SectionDict classses. Keys are the config.sections().
                        self.section_class_dict = {}
                        for section in self.config.sections():
                            section_dict = dict(self.config[section])
                            self.section_class_dict[section] = self._SectionDict(
                                section, section_dict, self.config, self
                            )

                        self.dict = self.section_class_dict
                        self.config_sections = self.config.sections()
                        # self.config_dict = {section_name : dict(self.config[section_name]) for section_name in self.config.sections()}
                    # if self.retain_config_comments:
                    #     self.config_file_contents = f.readlines()
        except Exception as e:
            if not self.silent_on_error:
                _error_popup_with_traceback(
                    "User Settings read warning",
                    "Error reading settings from file",
                    self.full_filename,
                    e,
                )
                # print('*** UserSettings.read - Error reading settings from file: ***\n', self.full_filename, e)
                # print(_create_error_message())

        return self.dict

    def exists(self, filename=None, path=None):
        """
        Check if a particular settings file exists.  Returns True if file exists

        :param filename: The name of the file to use. Can be a full path and filename or just filename
        :type filename:  (str or None)
        :param path:     The folder that the settings file will be stored in. Do not include the filename.
        :type path:      (str or None)
        """
        cfull_filename, cpath, cfilename = self._compute_filename(
            filename=filename, path=path
        )
        if os.path.exists(cfull_filename):
            return True
        return False

    def delete_entry(self, key, section=None, silent_on_error=None):
        """
        Deletes an individual entry.  If no filename has been specified up to this point,
        then a default filename will be used.
        After value has been deleted, the settings file is written to disk.

        :param section:
        :param key: Setting to be deleted. Can be any valid dictionary key type (i.e. must be hashable)
        :type key:  (Any)
        :param silent_on_error: Determines if error should be shown. This parameter overrides the silent on error setting for the object.
        :type silent_on_error:  (bool)
        """
        if self.full_filename is None:
            self.set_location()
            self.read()
        if not self.use_config_file:  # Is using JSON file
            if key in self.dict:
                del self.dict[key]
                if self.autosave:
                    self.save()
            else:
                if silent_on_error is False or (
                        silent_on_error is not True and not self.silent_on_error
                ):
                    _error_popup_with_traceback(
                        "User Settings delete_entry Warning - key",
                        key,
                        " not found in settings",
                    )

        else:
            if section is not None:
                section_dict = self.get(section)
                # print(f'** Trying to delete an entry with a config file in use ** id of section_dict = {id(section_dict)}')
                # section_dict = self.section_class_dict[section]
                del self.get(section)[key]
                # del section_dict[key]
                # del section_dict[key]

    def delete_section(self, section):
        """
        Deletes a section with the name provided in the section parameter.  Your INI file will be saved afterwards if auto-save enabled (default is ON)
        :param section:     Name of the section to delete
        :type section:      str
        """
        if not self.use_config_file:
            return

        section_dict = self.section_class_dict.get(section, None)
        section_dict.delete_section()
        del self.section_class_dict[section]
        if self.autosave:
            self.save()

    def set(self, key, value):
        """
        Sets an individual setting to the specified value.  If no filename has been specified up to this point,
        then a default filename will be used.
        After value has been modified, the settings file is written to disk.
        Note that this call is not value for a config file normally. If it is, then the key is assumed to be the
            Section key and the value written will be the default value.
        :param key:      Setting to be saved. Can be any valid dictionary key type
        :type key:       (Any)
        :param value:    Value to save as the setting's value. Can be anything
        :type value:     (Any)
        :return:         value that key was set to
        :rtype:          (Any)
        """

        if self.full_filename is None:
            self.set_location()
        # if not autosaving, then don't read the file or else will lose changes
        if not self.use_config_file:
            if self.autosave or self.dict == {}:
                self.read()
            self.dict[key] = value
        else:
            self.section_class_dict[key].set(value, self.default_value)

        if self.autosave:
            self.save()
        return value

    def get(self, key, default=None):
        """
        Returns the value of a specified setting.  If the setting is not found in the settings dictionary, then
        the user specified default value will be returned.  It no default is specified and nothing is found, then
        the "default value" is returned.  This default can be specified in this call, or previously defined
        by calling set_default. If nothing specified now or previously, then None is returned as default.

        :param key:     Key used to lookup the setting in the settings dictionary
        :type key:      (Any)
        :param default: Value to use should the key not be found in the dictionary
        :type default:  (Any)
        :return:        Value of specified settings
        :rtype:         (Any)
        """
        if self.default_value is not None:
            default = self.default_value

        if self.full_filename is None:
            self.set_location()
            if self.autosave or self.dict == {}:
                self.read()
        if not self.use_config_file:
            value = self.dict.get(key, default)
        else:
            value = self.section_class_dict.get(key, None)
            if key not in list(self.section_class_dict.keys()):
                self.section_class_dict[key] = self._SectionDict(
                    key, {}, self.config, self
                )
                value = self.section_class_dict[key]
                value.new_section = True
        return value

    def get_dict(self):
        """
        Returns the current settings dictionary.  If you've not setup the filename for the
        settings, a default one will be used and then read.

        Note that you can display the dictionary in text format by printing the object itself.

        :return: The current settings dictionary
        :rtype:  Dict
        """
        if self.full_filename is None:
            self.set_location()
            if self.autosave or self.dict == {}:
                self.read()
                self.save()
        return self.dict

    def __setitem__(self, item, value):
        """
        Enables setting a setting by using [ ] notation like a dictionary.
        Your code will have this kind of design pattern:
        settings = sg.UserSettings()
        settings[item] = value

        :param item:  The key for the setting to change. Needs to be a hashable type. Basically anything but a list
        :type item:   Any
        :param value: The value to set the setting to
        :type value:  Any
        """
        return self.set(item, value)

    def __getitem__(self, item):
        """
        Enables accessing a setting using [ ] notation like a dictionary.
        If the entry does not exist, then the default value will be returned.  This default
        value is None unless user sets by calling UserSettings.set_default_value(default_value)

        :param item: The key for the setting to change. Needs to be a hashable type. Basically anything but a list
        :type item:  Any
        :return:     The setting value
        :rtype:      Any
        """
        return self.get(item, self.default_value)

    def __delitem__(self, item):
        """
        Delete an individual user setting.  This is the same as calling delete_entry.  The syntax
        for deleting the item using this manner is:
            del settings['entry']
        :param item: The key for the setting to delete
        :type item:  Any
        """
        if self.use_config_file:
            return self.get(item)
        else:
            self.delete_entry(key=item)


# Create a singleton for the settings information so that the settings functions can be used
if UserSettings._default_for_function_interface is None:
    UserSettings._default_for_function_interface = UserSettings()


def user_settings_filename(filename=None, path=None):
    """
    Sets the filename and path for your settings file.  Either paramter can be optional.

    If you don't choose a path, one is provided for you that is OS specific
    Windows path default = users/name/AppData/Local/PySimpleGUI/settings.

    If you don't choose a filename, your application's filename + '.json' will be used.

    Normally the filename and path are split in the user_settings calls. However for this call they
    can be combined so that the filename contains both the path and filename.

    :param filename: The name of the file to use. Can be a full path and filename or just filename
    :type filename:  (str)
    :param path:     The folder that the settings file will be stored in. Do not include the filename.
    :type path:      (str)
    :return:         The full pathname of the settings file that has both the path and filename combined.
    :rtype:          (str)
    """
    settings = UserSettings._default_for_function_interface
    return settings.get_filename(filename, path)


def user_settings_delete_filename(filename=None, path=None, report_error=False):
    """
    Deltes the filename and path for your settings file.  Either paramter can be optional.
    If you don't choose a path, one is provided for you that is OS specific
    Windows path default = users/name/AppData/Local/PySimpleGUI/settings.
    If you don't choose a filename, your application's filename + '.json' will be used
    Also sets your current dictionary to a blank one.

    :param report_error:
    :param filename: The name of the file to use. Can be a full path and filename or just filename
    :type filename:  (str)
    :param path:     The folder that the settings file will be stored in. Do not include the filename.
    :type path:      (str)
    """
    settings = UserSettings._default_for_function_interface
    settings.delete_file(filename, path, report_error=report_error)


def user_settings_set_entry(key, value):
    """
    Sets an individual setting to the specified value.  If no filename has been specified up to this point,
    then a default filename will be used.
    After value has been modified, the settings file is written to disk.

    :param key:   Setting to be saved. Can be any valid dictionary key type
    :type key:    (Any)
    :param value: Value to save as the setting's value. Can be anything
    :type value:  (Any)
    """
    settings = UserSettings._default_for_function_interface
    settings.set(key, value)


def user_settings_delete_entry(key, silent_on_error=None):
    """
    Deletes an individual entry.  If no filename has been specified up to this point,
    then a default filename will be used.
    After value has been deleted, the settings file is written to disk.

    :param key: Setting to be saved. Can be any valid dictionary key type (hashable)
    :type key:  (Any)
    :param silent_on_error: Determines if an error popup should be shown if an error occurs. Overrides the silent onf effort setting from initialization
    :type silent_on_error:  (bool)
    """
    settings = UserSettings._default_for_function_interface
    settings.delete_entry(key, silent_on_error=silent_on_error)


def user_settings_get_entry(key, default=None):
    """
    Returns the value of a specified setting.  If the setting is not found in the settings dictionary, then
    the user specified default value will be returned.  It no default is specified and nothing is found, then
    None is returned.  If the key isn't in the dictionary, then it will be added and the settings file saved.
    If no filename has been specified up to this point, then a default filename will be assigned and used.
    The settings are SAVED prior to returning.

    :param key:     Key used to lookup the setting in the settings dictionary
    :type key:      (Any)
    :param default: Value to use should the key not be found in the dictionary
    :type default:  (Any)
    :return:        Value of specified settings
    :rtype:         (Any)
    """
    settings = UserSettings._default_for_function_interface
    return settings.get(key, default)


def user_settings_save(filename=None, path=None):
    """
    Saves the current settings dictionary.  If a filename or path is specified in the call, then it will override any
    previously specitfied filename to create a new settings file.  The settings dictionary is then saved to the newly defined file.

    :param filename: The fFilename to save to. Can specify a path or just the filename. If no filename specified, then the caller's filename will be used.
    :type filename:  (str)
    :param path:     The (optional) path to use to save the file.
    :type path:      (str)
    :return:         The full path and filename used to save the settings
    :rtype:          (str)
    """
    settings = UserSettings._default_for_function_interface
    return settings.save(filename, path)


def user_settings_load(filename=None, path=None):
    """
    Specifies the path and filename to use for the settings and reads the contents of the file.
    The filename can be a full filename including a path, or the path can be specified separately.
    If  no filename is specified, then the caller's filename will be used with the extension ".json"

    :param filename: Filename to load settings from (and save to in the future)
    :type filename:  (str)
    :param path:     Path to the file. Defaults to a specific folder depending on the operating system
    :type path:      (str)
    :return:         The settings dictionary (i.e. all settings)
    :rtype:          (dict)
    """
    settings = UserSettings._default_for_function_interface
    return settings.load(filename, path)


def user_settings_file_exists(filename=None, path=None):
    """
    Determines if a settings file exists.  If so a boolean True is returned.
    If either a filename or a path is not included, then the appropriate default
    will be used.

    :param filename: Filename to check
    :type filename:  (str)
    :param path:     Path to the file. Defaults to a specific folder depending on the operating system
    :type path:      (str)
    :return:         True if the file exists
    :rtype:          (bool)
    """
    settings = UserSettings._default_for_function_interface
    return settings.exists(filename=filename, path=path)


def user_settings_write_new_dictionary(settings_dict):
    """
    Writes a specified dictionary to the currently defined settings filename.

    :param settings_dict: The dictionary to be written to the currently defined settings file
    :type settings_dict:  (dict)
    """
    settings = UserSettings._default_for_function_interface
    settings.write_new_dictionary(settings_dict)


def user_settings_silent_on_error(silent_on_error=False):
    """
    Used to control the display of error messages.  By default, error messages are displayed to stdout.

    :param silent_on_error: If True then all error messages are silenced (not displayed on the console)
    :type silent_on_error:  (bool)
    """
    settings = UserSettings._default_for_function_interface
    settings.silent_on_error = silent_on_error


def user_settings():
    """
    Returns the current settings dictionary.  If you've not setup the filename for the
    settings, a default one will be used and then read.
    :return:            The current settings dictionary as a dictionary or a nicely formatted string representing it
    :rtype:             (dict or str)
    """
    settings = UserSettings._default_for_function_interface
    return settings.get_dict()


def user_settings_object():
    """
    Returns the object that is used for the function version of this API.
    With this object you can use the object interface, print it out in a nice format, etc.

    :return:    The UserSettings obect used for the function level interface
    :rtype:     (UserSettings)
    """
    return UserSettings._default_for_function_interface


"""
'########:'##::::'##:'########::'######::'##::::'##:'########:'########:
 ##.....::. ##::'##:: ##.....::'##... ##: ##:::: ##:... ##..:: ##.....::
 ##::::::::. ##'##::: ##::::::: ##:::..:: ##:::: ##:::: ##:::: ##:::::::
 ######:::::. ###:::: ######::: ##::::::: ##:::: ##:::: ##:::: ######:::
 ##...:::::: ## ##::: ##...:::: ##::::::: ##:::: ##:::: ##:::: ##...::::
 ##:::::::: ##:. ##:: ##::::::: ##::: ##: ##:::: ##:::: ##:::: ##:::::::
 ########: ##:::. ##: ########:. ######::. #######::::: ##:::: ########:
........::..:::::..::........:::......::::.......::::::..:::::........::
:::'###::::'########::'####::'######::                                  
::'## ##::: ##.... ##:. ##::'##... ##:                                  
:'##:. ##:: ##:::: ##:: ##:: ##:::..::                                  
'##:::. ##: ########::: ##::. ######::                                  
 #########: ##.....:::: ##:::..... ##:                                  
 ##.... ##: ##::::::::: ##::'##::: ##:                                  
 ##:::: ##: ##::::::::'####:. ######::                                  
..:::::..::..:::::::::....:::......:::        



These are the functions used to implement the subprocess APIs (Exec APIs) of PySimpleGUI

"""


def execute_command_subprocess(
        command,
        *args,
        wait=False,
        cwd=None,
        pipe_output=False,
        merge_stderr_with_stdout=True,
        stdin=None,
):
    """
    Runs the specified command as a subprocess.
    By default the call is non-blocking.
    The function will immediately return without waiting for the process to complete running. You can use the returned Popen object to communicate with the subprocess and get the results.
    Returns a subprocess Popen object.

    :param command:                  The command/file to execute. What you would type at a console to run a program or shell command.
    :type command:                   (str)
    :param wait:                     If True then wait for the subprocess to finish
    :type wait:                      (bool)
    :param cwd:                      Working directory to use when executing the subprocess
    :type cwd:                       (str))
    :param pipe_output:              If True then output from the subprocess will be piped. You MUST empty the pipe by calling execute_get_results or your subprocess will block until no longer full
    :type pipe_output:               (bool)
    :param merge_stderr_with_stdout: If True then output from the subprocess stderr will be merged with stdout. The result is ALL output will be on stdout.
    :type merge_stderr_with_stdout:  (bool)
    :param stdin:                    Value passed to the Popen call. Defaults to subprocess.DEVNULL so that the pyinstaller created executable work correctly
    :type stdin:                     (bool)
    :return:                         Popen object
    :rtype:                          (subprocess.Popen)
    """
    if stdin is None:
        stdin = subprocess.DEVNULL
    try:
        if args is not None:
            expanded_args = " ".join(args)
            # print('executing subprocess command:',command, 'args:',expanded_args)
            if command[0] != '"' and " " in command:
                command = '"' + command + '"'
            # print('calling popen with:', command +' '+ expanded_args)
            # sp = subprocess.Popen(command +' '+ expanded_args, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd)
            if pipe_output:
                if merge_stderr_with_stdout:
                    sp = subprocess.Popen(
                        command + " " + expanded_args,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        cwd=cwd,
                        stdin=stdin,
                    )
                else:
                    sp = subprocess.Popen(
                        command + " " + expanded_args,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        cwd=cwd,
                        stdin=stdin,
                    )
            else:
                sp = subprocess.Popen(
                    command + " " + expanded_args,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    cwd=cwd,
                    stdin=stdin,
                )
        else:
            sp = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=cwd,
                stdin=stdin,
            )
        if wait:
            out, err = sp.communicate()
            if out:
                print(out.decode("utf-8"))
            if err:
                print(err.decode("utf-8"))
    except Exception as e:
        warnings.warn("Error in execute_command_subprocess {}".format(e), UserWarning)
        _error_popup_with_traceback(
            "Error in execute_command_subprocess",
            e,
            "command={}".format(command),
            "args={}".format(args),
            "cwd={}".format(cwd),
        )
        sp = None
    return sp


def execute_py_file(
        pyfile,
        parms=None,
        cwd=None,
        interpreter_command=None,
        wait=False,
        pipe_output=False,
        merge_stderr_with_stdout=True,
):
    """
    Executes a Python file.
    The interpreter to use is chosen based on this priority order:
        1. interpreter_command paramter
        2. global setting "-python command-"
        3. the interpreter running running PySimpleGUI
    :param pyfile:                   the file to run
    :type pyfile:                    (str)
    :param parms:                    parameters to pass on the command line
    :type parms:                     (str)
    :param cwd:                      the working directory to use
    :type cwd:                       (str)
    :param interpreter_command:      the command used to invoke the Python interpreter
    :type interpreter_command:       (str)
    :param wait:                     the working directory to use
    :type wait:                      (bool)
    :param pipe_output:              If True then output from the subprocess will be piped. You MUST empty the pipe by calling execute_get_results or your subprocess will block until no longer full
    :type pipe_output:               (bool)
    :param merge_stderr_with_stdout: If True then output from the subprocess stderr will be merged with stdout. The result is ALL output will be on stdout.
    :type merge_stderr_with_stdout:  (bool)
    :return:                         Popen object
    :rtype:                          (subprocess.Popen) | None
    """

    if cwd is None:
        # if the specific file is not found (not an absolute path) then assume it's relative to '.'
        if not os.path.exists(pyfile):
            cwd = "."

    if pyfile[0] != '"' and " " in pyfile:
        pyfile = '"' + pyfile + '"'
    if interpreter_command is not None:
        python_program = interpreter_command
    else:
        # use the version CURRENTLY RUNNING if nothing is specified. Previously used the one from the settings file
        # ^ hmmm... that's not the code is doing now... it's getting the one from the settings file first
        pysimplegui_user_settings.load()  # Refresh the settings just in case they've changed via another program
        python_program = pysimplegui_user_settings.get("-python command-", "")
        if (
                python_program == ""
        ):  # if no interpreter set in the settings, then use the current one
            python_program = sys.executable
            # python_program = 'python' if running_windows() else 'python3'
    if parms is not None and python_program:
        sp = execute_command_subprocess(
            python_program,
            pyfile,
            parms,
            wait=wait,
            cwd=cwd,
            pipe_output=pipe_output,
            merge_stderr_with_stdout=merge_stderr_with_stdout,
        )
    elif python_program:
        sp = execute_command_subprocess(
            python_program,
            pyfile,
            wait=wait,
            cwd=cwd,
            pipe_output=pipe_output,
            merge_stderr_with_stdout=merge_stderr_with_stdout,
        )
    else:
        print("execute_py_file - No interpreter has been configured")
        sp = None
    return sp


def execute_py_get_interpreter():
    """
    Returns Python Interpreter from the system settings. If none found in the settings file
    then the currently running interpreter is returned.

    :return: Full path to python interpreter (uses settings file or sys.executable)
    :rtype:  (str)
    """
    pysimplegui_user_settings.load()  # Refresh the settings just in case they've changed via another program
    interpreter = pysimplegui_user_settings.get("-python command-", "")
    if interpreter == "":
        interpreter = sys.executable
    return interpreter


def execute_py_get_running_interpreter():
    """
    Returns the command that is currently running.

    :return: Full path to python interpreter (uses sys.executable)
    :rtype:  (str)
    """
    return sys.executable


def execute_editor(file_to_edit, line_number=None):
    """
    Runs the editor that was configured in the global settings and opens the file to a specific line number.
    Two global settings keys are used.
    '-editor program-' the command line used to startup your editor. It's set
        in the global settings window or by directly manipulating the PySimpleGUI settings object
    '-editor format string-' a string containing 3 "tokens" that describes the command that is executed
            <editor> <file> <line>
    :param file_to_edit: the full path to the file to edit
    :type file_to_edit:  (str)
    :param line_number:  optional line number to place the cursor
    :type line_number:   (int)
    :return:             Popen object
    :rtype:              (subprocess.Popen) | None
    """
    if (
            file_to_edit is not None
            and len(file_to_edit) != 0
            and file_to_edit[0] not in ('"', "'")
            and " " in file_to_edit
    ):
        file_to_edit = '"' + file_to_edit + '"'
    pysimplegui_user_settings.load()  # Refresh the settings just in case they've changed via another program
    editor_program = pysimplegui_user_settings.get("-editor program-", None)
    if editor_program is not None:
        format_string = pysimplegui_user_settings.get("-editor format string-", None)
        # if no format string, then just launch the editor with the filename
        if not format_string or line_number is None:
            sp = execute_command_subprocess(editor_program, file_to_edit)
        else:
            command = _create_full_editor_command(
                file_to_edit, line_number, format_string
            )
            # print('final command line = ', command)
            sp = execute_command_subprocess(editor_program, command)
    else:
        print("No editor has been configured in the global settings")
        sp = None
    return sp


def execute_get_results(subprocess_id, timeout=None):
    """
    Get the text results of a previously executed execute call
    Returns a tuple of the strings (stdout, stderr)
    :param subprocess_id: a Popen subprocess ID returned from a previous execute call
    :type subprocess_id:  (subprocess.Popen)
    :param timeout:       Time in fractions of a second to wait. Returns '','' if timeout. Default of None means wait forever
    :type timeout:        (None | float)
    :returns:             Tuple with 2 strings (stdout, stderr)
    :rtype:               (str | None , str | None)
    """

    out_decoded = err_decoded = None
    if subprocess_id is not None:
        try:
            out, err = subprocess_id.communicate(timeout=timeout)
            if out:
                out_decoded = out.decode("utf-8")
            if err:
                err_decoded = err.decode("utf-8")
        except ValueError:
            # will get an error if stdout and stderr are combined and attempt to read stderr
            # so ignore the error that would be generated
            pass
        except subprocess.TimeoutExpired:
            # a Timeout error is not actually an error that needs to be reported
            pass
        except Exception as e:
            popup_error("Error in execute_get_results", e)
    return out_decoded, err_decoded


def execute_subprocess_still_running(subprocess_id):
    """
    Returns True is the subprocess ID provided is for a process that is still running

    :param subprocess_id: ID previously returned from Exec API calls that indicate this value is returned
    :type subprocess_id:  (subprocess.Popen)
    :return:              True if the subproces is running
    :rtype:               bool
    """
    if subprocess_id.poll() == 0:
        return False
    return True


def execute_file_explorer(folder_to_open=""):
    """
    The global settings has a setting called -   "-explorer program-"
    It defines the program to run when this function is called.
    The optional folder paramter specified which path should be opened.

    :param folder_to_open: The path to open in the explorer program
    :type folder_to_open:  str
    :return:               Popen object
    :rtype:                (subprocess.Popen) | None
    """
    pysimplegui_user_settings.load()  # Refresh the settings just in case they've changed via another program
    explorer_program = pysimplegui_user_settings.get("-explorer program-", None)
    if explorer_program is not None:
        sp = execute_command_subprocess(explorer_program, folder_to_open)
    else:
        print("No file explorer has been configured in the global settings")
        sp = None
    return sp


def execute_find_callers_filename():
    """
    Returns the first filename found in a traceback that is not the name of this file (__file__)
    Used internally with the debugger for example.

    :return: filename of the caller, assumed to be the first non PySimpleGUI file
    :rtype:  str
    """
    try:  # lots can go wrong so wrapping the entire thing
        trace_details = traceback.format_stack()
        file_info_pysimplegui, error_message = None, ""
        for line in reversed(trace_details):
            if __file__ not in line:
                file_info_pysimplegui = line.split(",")[0]
                error_message = line
                break
        if file_info_pysimplegui is None:
            return ""
        error_parts = None
        if error_message != "":
            error_parts = error_message.split(", ")
            if len(error_parts) < 4:
                error_message = (
                        error_parts[0]
                        + "\n"
                        + error_parts[1]
                        + "\n"
                        + "".join(error_parts[2:])
                )
        if error_parts is None:
            print("*** Error popup attempted but unable to parse error details ***")
            print(trace_details)
            return ""
        filename = error_parts[0][error_parts[0].index("File ") + 5:]
        return filename
    except Exception:
        return ""


def _create_full_editor_command(file_to_edit, line_number, edit_format_string):
    """
    The global settings has a setting called -   "-editor format string-"
    It uses 3 "tokens" to describe how to invoke the editor in a way that starts at a specific line #
    <editor> <file> <line>

    :param file_to_edit:
    :type file_to_edit:        str
    :param edit_format_string:
    :type edit_format_string:  str
    :return:
    :rtype:
    """

    command = edit_format_string
    command = command.replace("<editor>", "")
    command = command.replace("<file>", file_to_edit)
    command = command.replace(
        "<line>", str(line_number) if line_number is not None else ""
    )
    return command


def execute_get_editor():
    """
    Get the path to the editor based on user settings or on PySimpleGUI's global settings

    :return: Path to the editor
    :rtype:  str
    """
    try:  # in case running with old version of PySimpleGUI that doesn't have a global PSG settings path
        global_editor = pysimplegui_user_settings.get("-editor program-")
    except Exception:
        global_editor = ""

    return user_settings_get_entry("-editor program-", global_editor)


"""
'##::::'##::::'###:::::'######::::::'######::'########::'########::'######::'####:'########:'####::'######::
 ###::'###:::'## ##:::'##... ##::::'##... ##: ##.... ##: ##.....::'##... ##:. ##:: ##.....::. ##::'##... ##:
 ####'####::'##:. ##:: ##:::..::::: ##:::..:: ##:::: ##: ##::::::: ##:::..::: ##:: ##:::::::: ##:: ##:::..::
 ## ### ##:'##:::. ##: ##::::::::::. ######:: ########:: ######::: ##:::::::: ##:: ######:::: ##:: ##:::::::
 ##. #: ##: #########: ##:::::::::::..... ##: ##.....::: ##...:::: ##:::::::: ##:: ##...::::: ##:: ##:::::::
 ##:.:: ##: ##.... ##: ##::: ##::::'##::: ##: ##:::::::: ##::::::: ##::: ##:: ##:: ##:::::::: ##:: ##::: ##:
 ##:::: ##: ##:::: ##:. ######:::::. ######:: ##:::::::: ########:. ######::'####: ##:::::::'####:. ######::
..:::::..::..:::::..:::......:::::::......:::..:::::::::........:::......:::....::..::::::::....:::......:::
"""

"""
The Mac problems have been significant enough to warrant the addition of a series of settings that allow
users to turn specific patches and features on or off depending on their setup.  There is not enough information
available to make this process more atuomatic.  

"""

# Dictionary of Mac Patches.  Used to find the key in the global settings and the default value
MAC_PATCH_DICT = {
    "Enable No Titlebar Patch": ("-mac feature enable no titlebar patch-", False),
    "Disable Modal Windows": ("-mac feature disable modal windows-", True),
    "Disable Grab Anywhere with Titlebar": (
        "-mac feature disable grab anywhere with titlebar-",
        True,
    ),
    "Set Alpha Channel to 0.99 for MacOS >= 12.3": (
        "-mac feature disable Alpha 0.99",
        True,
    ),
}


def _read_mac_global_settings():
    """
    Reads the settings from the PySimpleGUI Global Settings and sets variables that
    are used at runtime to control how certain features behave
    """

    global ENABLE_MAC_MODAL_DISABLE_PATCH
    global ENABLE_MAC_NOTITLEBAR_PATCH
    global ENABLE_MAC_DISABLE_GRAB_ANYWHERE_WITH_TITLEBAR
    global ENABLE_MAC_ALPHA_99_PATCH

    ENABLE_MAC_MODAL_DISABLE_PATCH = pysimplegui_user_settings.get(
        MAC_PATCH_DICT["Disable Modal Windows"][0],
        MAC_PATCH_DICT["Disable Modal Windows"][1],
    )
    ENABLE_MAC_NOTITLEBAR_PATCH = pysimplegui_user_settings.get(
        MAC_PATCH_DICT["Enable No Titlebar Patch"][0],
        MAC_PATCH_DICT["Enable No Titlebar Patch"][1],
    )
    ENABLE_MAC_DISABLE_GRAB_ANYWHERE_WITH_TITLEBAR = pysimplegui_user_settings.get(
        MAC_PATCH_DICT["Disable Grab Anywhere with Titlebar"][0],
        MAC_PATCH_DICT["Disable Grab Anywhere with Titlebar"][1],
    )
    ENABLE_MAC_ALPHA_99_PATCH = pysimplegui_user_settings.get(
        MAC_PATCH_DICT["Set Alpha Channel to 0.99 for MacOS >= 12.3"][0],
        MAC_PATCH_DICT["Set Alpha Channel to 0.99 for MacOS >= 12.3"][1],
    )


def _mac_should_apply_notitlebar_patch():
    """
    Uses a combination of the tkinter version number and the setting from the global settings
    to determine if the notitlebar patch should be applied

    :return:    True if should apply the no titlebar patch on the Mac
    :rtype:     (bool)
    """

    if not running_mac():
        return False

    try:
        tver = [int(n) for n in framework_version.split(".")]
        if (
                tver[0] == 8
                and tver[1] == 6
                and tver[2] < 10
                and ENABLE_MAC_NOTITLEBAR_PATCH
        ):
            return True
    except Exception as e:
        warnings.warn(
            "Exception while trying to parse tkinter version {} Error = {}".format(
                framework_version, e
            ),
            UserWarning,
        )

    return False


def _mac_should_set_alpha_to_99():
    if not running_mac():
        return False

    if not ENABLE_MAC_ALPHA_99_PATCH:
        return False

    # ONLY enable this patch for tkinter version 8.6.12
    if framework_version != "8.6.12":
        return False

    # At this point, we're running a Mac and the alpha patch is enabled
    # Final check is to see if Mac OS version is 12.3 or later
    try:
        platform_mac_ver = platform.mac_ver()[0]
        mac_ver = (
            platform_mac_ver.split(".")
            if "." in platform_mac_ver
            else (platform_mac_ver, 0)
        )
        if (int(mac_ver[0]) >= 12 and int(mac_ver[1]) >= 3) or int(mac_ver[0]) >= 13:
            # print("Mac OS Version is {} and patch enabled so applying the patch".format(platform_mac_ver))
            return True
    except Exception as e:
        warnings.warn(
            "_mac_should_seet_alpha_to_99 Exception while trying check mac_ver. Error = {}".format(
                e
            ),
            UserWarning,
        )
        return False

    return False


def main_mac_feature_control():
    """
    Window to set settings that will be used across all PySimpleGUI programs that choose to use them.
    Use set_options to set the path to the folder for all PySimpleGUI settings.

    :return: True if settings were changed
    :rtype:  (bool)
    """

    current_theme = theme()
    theme("dark red")

    layout = [
        [T("Mac PySimpleGUI Feature Control", font="DEFAIULT 18")],
        [T("Use this window to enable / disable features.")],
        [
            T(
                "Unfortunately, on some releases of tkinter on the Mac, there are problems that"
            )
        ],
        [
            T(
                "create the need to enable and disable sets of features. This window facilitates the control."
            )
        ],
        [T("Feature Control / Settings", font="_ 16 bold")],
        [
            T("You are running tkinter version:", font="_ 12 bold"),
            T(framework_version, font="_ 12 bold"),
        ],
    ]

    for key, value in MAC_PATCH_DICT.items():
        layout += [
            [
                Checkbox(
                    key,
                    k=value[0],
                    default=pysimplegui_user_settings.get(value[0], value[1]),
                )
            ]
        ]
    layout += [
        [
            T(
                "Currently the no titlebar patch "
                + ("WILL" if _mac_should_apply_notitlebar_patch() else "WILL NOT")
                + " be applied"
            )
        ],
        [T("The no titlebar patch will ONLY be applied on tkinter versions < 8.6.10")],
    ]
    layout += [[Button("Ok"), Button("Cancel")]]

    window = Window("Mac Feature Control", layout, keep_on_top=True, finalize=True)
    while True:
        event, values = window.read()
        if event in ("Cancel", WIN_CLOSED):
            break
        if event == "Ok":
            for key, value in values.items():
                print("setting {} to {}".format(key, value))
                pysimplegui_user_settings.set(key, value)
            break
    window.close()
    theme(current_theme)


"""
'########::'########:'########::'##::::'##::'######::::'######:::'########:'########::
 ##.... ##: ##.....:: ##.... ##: ##:::: ##:'##... ##::'##... ##:: ##.....:: ##.... ##:
 ##:::: ##: ##::::::: ##:::: ##: ##:::: ##: ##:::..::: ##:::..::: ##::::::: ##:::: ##:
 ##:::: ##: ######::: ########:: ##:::: ##: ##::'####: ##::'####: ######::: ########::
 ##:::: ##: ##...:::: ##.... ##: ##:::: ##: ##::: ##:: ##::: ##:: ##...:::: ##.. ##:::
 ##:::: ##: ##::::::: ##:::: ##: ##:::: ##: ##::: ##:: ##::: ##:: ##::::::: ##::. ##::
 ########:: ########: ########::. #######::. ######:::. ######::: ########: ##:::. ##:
........:::........::........::::.......::::......:::::......::::........::..:::::..::                                                                    
"""

#####################################################################################################
# Debugger
#####################################################################################################


red_x = b"R0lGODlhEAAQAPeQAIsAAI0AAI4AAI8AAJIAAJUAAJQCApkAAJoAAJ4AAJkJCaAAAKYAAKcAAKcCAKcDA6cGAKgAAKsAAKsCAKwAAK0AAK8AAK4CAK8DAqUJAKULAKwLALAAALEAALIAALMAALMDALQAALUAALYAALcEALoAALsAALsCALwAAL8AALkJAL4NAL8NAKoTAKwbAbEQALMVAL0QAL0RAKsREaodHbkQELMsALg2ALk3ALs+ALE2FbgpKbA1Nbc1Nb44N8AAAMIWAMsvAMUgDMcxAKVABb9NBbVJErFYEq1iMrtoMr5kP8BKAMFLAMxKANBBANFCANJFANFEB9JKAMFcANFZANZcANpfAMJUEMZVEc5hAM5pAMluBdRsANR8AM9YOrdERMpIQs1UVMR5WNt8X8VgYMdlZcxtYtx4YNF/btp9eraNf9qXXNCCZsyLeNSLd8SSecySf82kd9qqc9uBgdyBgd+EhN6JgtSIiNuJieGHhOGLg+GKhOKamty1ste4sNO+ueenp+inp+HHrebGrefKuOPTzejWzera1O7b1vLb2/bl4vTu7fbw7ffx7vnz8f///wAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAJAALAAAAAAQABAAAAjUACEJHEiwYEEABniQKfNFgQCDkATQwAMokEU+PQgUFDAjjR09e/LUmUNnh8aBCcCgUeRmzBkzie6EeQBAoAAMXuA8ciRGCaJHfXzUMCAQgYooWN48anTokR8dQk4sELggBhQrU9Q8evSHiJQgLCIIfMDCSZUjhbYuQkLFCRAMAiOQGGLE0CNBcZYmaRIDLqQFGF60eTRoSxc5jwjhACFWIAgMLtgUocJFy5orL0IQRHAiQgsbRZYswbEhBIiCCH6EiJAhAwQMKU5DjHCi9gnZEHMTDAgAOw=="


class _Debugger:
    debugger = None
    DEBUGGER_MAIN_WINDOW_THEME = "dark grey 13"
    DEBUGGER_POPOUT_THEME = "dark grey 13"
    WIDTH_VARIABLES = 23
    WIDTH_RESULTS = 46

    WIDTH_WATCHER_VARIABLES = 20
    WIDTH_WATCHER_RESULTS = 60

    WIDTH_LOCALS = 80
    NUM_AUTO_WATCH = 9

    MAX_LINES_PER_RESULT_FLOATING = 4
    MAX_LINES_PER_RESULT_MAIN = 3

    DEBUGGER_POPOUT_WINDOW_FONT = "Sans 8"
    DEBUGGER_VARIABLE_DETAILS_FONT = "Courier 10"

    """
        #     #                    ######
        ##   ##   ##   # #    #    #     # ###### #####  #    #  ####   ####  ###### #####
        # # # #  #  #  # ##   #    #     # #      #    # #    # #    # #    # #      #    #
        #  #  # #    # # # #  #    #     # #####  #####  #    # #      #      #####  #    #
        #     # ###### # #  # #    #     # #      #    # #    # #  ### #  ### #      #####
        #     # #    # # #   ##    #     # #      #    # #    # #    # #    # #      #   #
        #     # #    # # #    #    ######  ###### #####   ####   ####   ####  ###### #    #
    """

    def __init__(self):
        self.watcher_window = None  # type: Window
        self.popout_window = None  # type: Window
        self.local_choices = {}
        self.myrc = ""
        self.custom_watch = ""
        self.locals = {}
        self.globals = {}
        self.popout_choices = {}

    # Includes the DUAL PANE (now 2 tabs)!  Don't forget REPL is there too!
    def _build_main_debugger_window(self, location=(None, None)):
        old_theme = theme()
        theme(_Debugger.DEBUGGER_MAIN_WINDOW_THEME)

        def InVar(key1):
            row1 = [
                T("    "),
                I(key=key1, size=(_Debugger.WIDTH_VARIABLES, 1)),
                T("", key=key1 + "CHANGED_", size=(_Debugger.WIDTH_RESULTS, 1)),
                B("Detail", key=key1 + "DETAIL_"),
                B("Obj", key=key1 + "OBJ_"),
            ]
            return row1

        variables_frame = [
            InVar("_VAR0_"),
            InVar("_VAR1_"),
            InVar("_VAR2_"),
        ]

        interactive_frame = [
            [
                T(">>> "),
                In(
                    size=(83, 1),
                    key="-REPL-",
                    tooltip='Type in any "expression" or "statement"\n and it will be disaplayed below.\nPress RETURN KEY instead of "Go"\nbutton for faster use',
                ),
                B("Go", bind_return_key=True, visible=True),
            ],
            [
                Multiline(
                    size=(93, 26), key="-OUTPUT-", autoscroll=True, do_not_clear=True
                )
            ],
        ]

        autowatch_frame = [
            [
                Button("Choose Variables To Auto Watch", key="-LOCALS-"),
                Button("Clear All Auto Watches"),
                Button("Show All Variables", key="-SHOW_ALL-"),
                Button("Locals", key="-ALL_LOCALS-"),
                Button("Globals", key="-GLOBALS-"),
                Button("Popout", key="-POPOUT-"),
            ]
        ]

        var_layout = []
        for i in range(_Debugger.NUM_AUTO_WATCH):
            var_layout.append(
                [
                    T(
                        "",
                        size=(_Debugger.WIDTH_WATCHER_VARIABLES, 1),
                        key="_WATCH%s_" % i,
                    ),
                    T(
                        "",
                        size=(
                            _Debugger.WIDTH_WATCHER_RESULTS,
                            _Debugger.MAX_LINES_PER_RESULT_MAIN,
                        ),
                        key="_WATCH%s_RESULT_" % i,
                    ),
                ]
            )

        col1 = [
            # [Frame('Auto Watches', autowatch_frame+variable_values, title_color='blue')]
            [
                Frame(
                    "Auto Watches",
                    autowatch_frame + var_layout,
                    title_color=theme_button_color()[0],
                )
            ]
        ]

        col2 = [
            [
                Frame(
                    "Variables or Expressions to Watch",
                    variables_frame,
                    title_color=theme_button_color()[0],
                ),
            ],
            [
                Frame(
                    "REPL-Light - Press Enter To Execute Commands",
                    interactive_frame,
                    title_color=theme_button_color()[0],
                ),
            ],
        ]

        # Tab based layout
        layout = [
            [Text("Debugging: " + self._find_users_code())],
            [TabGroup([[Tab("Variables", col1), Tab("REPL & Watches", col2)]])],
        ]

        # ------------------------------- Create main window -------------------------------
        window = Window(
            "PySimpleGUI Debugger",
            layout,
            icon=PSG_DEBUGGER_LOGO,
            margins=(0, 0),
            location=location,
            keep_on_top=True,
            right_click_menu=[
                [""],
                [
                    "Exit",
                ],
            ],
        )

        Window._read_call_from_debugger = True
        window.finalize()
        Window._read_call_from_debugger = False

        window.Element("_VAR1_").SetFocus()
        self.watcher_window = window
        theme(old_theme)
        return window

    """
        #     #                    #######                               #
        ##   ##   ##   # #    #    #       #    # ###### #    # #####    #        ####   ####  #####
        # # # #  #  #  # ##   #    #       #    # #      ##   #   #      #       #    # #    # #    #
        #  #  # #    # # # #  #    #####   #    # #####  # #  #   #      #       #    # #    # #    #
        #     # ###### # #  # #    #       #    # #      #  # #   #      #       #    # #    # #####
        #     # #    # # #   ##    #        #  #  #      #   ##   #      #       #    # #    # #
        #     # #    # # #    #    #######   ##   ###### #    #   #      #######  ####   ####  #
    """

    def _refresh_main_debugger_window(self, mylocals, myglobals):
        if not self.watcher_window:  # if there is no window setup, nothing to do
            return False
        event, values = self.watcher_window.read(timeout=1)
        if event in (None, "Exit", "_EXIT_", "-EXIT-"):  # EXIT BUTTON / X BUTTON
            try:
                self.watcher_window.close()
            except Exception:
                pass
            self.watcher_window = None
            return False
        # ------------------------------- Process events from REPL Tab -------------------------------
        cmd = values["-REPL-"]  # get the REPL entered
        # BUTTON - GO (NOTE - This button is invisible!!)
        if event == "Go":  # GO BUTTON
            self.watcher_window.Element("-REPL-").Update("")
            self.watcher_window.Element("-OUTPUT-").Update(
                ">>> {}\n".format(cmd), append=True, autoscroll=True
            )

            try:
                result = eval("{}".format(cmd), myglobals, mylocals)
            except Exception as e:
                if sys.version_info[0] < 3:
                    result = "Not available in Python 2"
                else:
                    try:
                        result = exec("{}".format(cmd), myglobals, mylocals)
                    except Exception as e:
                        result = "Exception {}\n".format(e)

            self.watcher_window.Element("-OUTPUT-").Update(
                "{}\n".format(result), append=True, autoscroll=True
            )
        # BUTTON - DETAIL
        elif event.endswith("_DETAIL_"):  # DETAIL BUTTON
            var = values["_VAR{}_".format(event[4])]
            try:
                result = str(eval(str(var), myglobals, mylocals))
            except Exception:
                result = ""
            old_theme = theme()
            theme(_Debugger.DEBUGGER_MAIN_WINDOW_THEME)
            popup_scrolled(
                str(values["_VAR{}_".format(event[4])]) + "\n" + result,
                title=var,
                non_blocking=True,
                font=_Debugger.DEBUGGER_VARIABLE_DETAILS_FONT,
            )
            theme(old_theme)
        # BUTTON - OBJ
        elif event.endswith("_OBJ_"):  # OBJECT BUTTON
            var = values["_VAR{}_".format(event[4])]
            try:
                result = ObjToStringSingleObj(mylocals[var])
            except Exception as e:
                try:
                    result = eval("{}".format(var), myglobals, mylocals)
                    result = ObjToStringSingleObj(result)
                except Exception as e:
                    result = "{}\nError showing object {}".format(e, var)
            old_theme = theme()
            theme(_Debugger.DEBUGGER_MAIN_WINDOW_THEME)
            popup_scrolled(
                str(var) + "\n" + str(result),
                title=var,
                non_blocking=True,
                font=_Debugger.DEBUGGER_VARIABLE_DETAILS_FONT,
            )
            theme(old_theme)
        # ------------------------------- Process Watch Tab -------------------------------
        # BUTTON - Choose Locals to see
        elif event == "-LOCALS-":  # Show all locals BUTTON
            self._choose_auto_watches(mylocals)
        # BUTTON - Locals (quick popup)
        elif event == "-ALL_LOCALS-":
            self._display_all_vars("All Locals", mylocals)
        # BUTTON - Globals (quick popup)
        elif event == "-GLOBALS-":
            self._display_all_vars("All Globals", myglobals)
        # BUTTON - clear all
        elif event == "Clear All Auto Watches":
            if (
                    popup_yes_no(
                        "Do you really want to clear all Auto-Watches?", "Really Clear??"
                    )
                    == "Yes"
            ):
                self.local_choices = {}
                self.custom_watch = ""
        # BUTTON - Popout
        elif event == "-POPOUT-":
            if not self.popout_window:
                self._build_floating_window()
        # BUTTON - Show All
        elif event == "-SHOW_ALL-":
            for key in self.locals:
                self.local_choices[key] = not key.startswith("_")

        # -------------------- Process the manual "watch list" ------------------
        for i in range(3):
            key = "_VAR{}_".format(i)
            out_key = "_VAR{}_CHANGED_".format(i)
            self.myrc = ""
            if self.watcher_window.Element(key):
                var = values[key]
                try:
                    result = eval(str(var), myglobals, mylocals)
                except Exception:
                    result = ""
                self.watcher_window.Element(out_key).Update(str(result))
            else:
                self.watcher_window.Element(out_key).Update("")

        # -------------------- Process the automatic "watch list" ------------------
        slot = 0
        for key in self.local_choices:
            if key == "-CUSTOM_WATCH-":
                continue
            if self.local_choices[key]:
                self.watcher_window.Element("_WATCH{}_".format(slot)).Update(key)
                try:
                    self.watcher_window.Element(
                        "_WATCH{}_RESULT_".format(slot), silent_on_error=True
                    ).Update(mylocals[key])
                except Exception:
                    self.watcher_window.Element("_WATCH{}_RESULT_".format(slot)).Update(
                        ""
                    )
                slot += 1

            if (
                    slot + int(not self.custom_watch in (None, ""))
                    >= _Debugger.NUM_AUTO_WATCH
            ):
                break
        # If a custom watch was set, display that value in the window
        if self.custom_watch:
            self.watcher_window.Element("_WATCH{}_".format(slot)).Update(
                self.custom_watch
            )
            try:
                self.myrc = eval(self.custom_watch, myglobals, mylocals)
            except Exception:
                self.myrc = ""
            self.watcher_window.Element("_WATCH{}_RESULT_".format(slot)).Update(
                self.myrc
            )
            slot += 1
        # blank out all of the slots not used (blank)
        for i in range(slot, _Debugger.NUM_AUTO_WATCH):
            self.watcher_window.Element("_WATCH{}_".format(i)).Update("")
            self.watcher_window.Element("_WATCH{}_RESULT_".format(i)).Update("")

        return True  # return indicating the window stayed open

    def _find_users_code(self):
        try:  # lots can go wrong so wrapping the entire thing
            trace_details = traceback.format_stack()
            file_info_pysimplegui, error_message = None, ""
            for line in reversed(trace_details):
                if __file__ not in line:
                    file_info_pysimplegui = line.split(",")[0]
                    error_message = line
                    break
            if file_info_pysimplegui is None:
                return ""
            error_parts = None
            if error_message != "":
                error_parts = error_message.split(", ")
                if len(error_parts) < 4:
                    error_message = (
                            error_parts[0]
                            + "\n"
                            + error_parts[1]
                            + "\n"
                            + "".join(error_parts[2:])
                    )
            if error_parts is None:
                print("*** Error popup attempted but unable to parse error details ***")
                print(trace_details)
                return ""
            filename = error_parts[0][error_parts[0].index("File ") + 5:]
            return filename
        except Exception:
            return

    """
        ######                                 #     #
        #     #  ####  #####  #    # #####     #  #  # # #    # #####   ####  #    #
        #     # #    # #    # #    # #    #    #  #  # # ##   # #    # #    # #    #
        ######  #    # #    # #    # #    #    #  #  # # # #  # #    # #    # #    #
        #       #    # #####  #    # #####     #  #  # # #  # # #    # #    # # ## #
        #       #    # #      #    # #         #  #  # # #   ## #    # #    # ##  ##
        #        ####  #       ####  #          ## ##  # #    # #####   ####  #    #
    
        ######                                    #                     #     #
        #     # #    # #    # #####   ####       # #   #      #         #     #   ##   #####   ####
        #     # #    # ##  ## #    # #          #   #  #      #         #     #  #  #  #    # #
        #     # #    # # ## # #    #  ####     #     # #      #         #     # #    # #    #  ####
        #     # #    # #    # #####       #    ####### #      #          #   #  ###### #####       #
        #     # #    # #    # #      #    #    #     # #      #           # #   #    # #   #  #    #
        ######   ####  #    # #       ####     #     # ###### ######       #    #    # #    #  ####
    """

    # displays them into a single text box

    def _display_all_vars(self, title, dict):
        num_cols = 3
        output_text = ""
        num_lines = 2
        cur_col = 0
        out_text = title + "\n"
        longest_line = max([len(key) for key in dict])
        line = []
        sorted_dict = {}
        for key in sorted(dict.keys()):
            sorted_dict[key] = dict[key]
        for key in sorted_dict:
            value = dict[key]
            # wrapped_list = textwrap.wrap(str(value), 60)
            # wrapped_text = '\n'.join(wrapped_list)
            wrapped_text = str(value)
            out_text += "{} - {}\n".format(key, wrapped_text)
            # if cur_col + 1 == num_cols:
            #     cur_col = 0
            #     num_lines += len(wrapped_list)
            # else:
            #     cur_col += 1
        old_theme = theme()
        theme(_Debugger.DEBUGGER_MAIN_WINDOW_THEME)
        popup_scrolled(
            out_text,
            title=title,
            non_blocking=True,
            font=_Debugger.DEBUGGER_VARIABLE_DETAILS_FONT,
            keep_on_top=True,
            icon=PSG_DEBUGGER_LOGO,
        )
        theme(old_theme)

    """
        #####                                        #     #
       #     # #    #  ####   ####   ####  ######    #  #  #   ##   #####  ####  #    #
       #       #    # #    # #    # #      #         #  #  #  #  #    #   #    # #    #
       #       ###### #    # #    #  ####  #####     #  #  # #    #   #   #      ######
       #       #    # #    # #    #      # #         #  #  # ######   #   #      #    #
       #     # #    # #    # #    # #    # #         #  #  # #    #   #   #    # #    #
        #####  #    #  ####   ####   ####  ######     ## ##  #    #   #    ####  #    #
    
        #     #                                                       #     #
        #     #   ##   #####  #   ##   #####  #      ######  ####     #  #  # # #    #
        #     #  #  #  #    # #  #  #  #    # #      #      #         #  #  # # ##   #
        #     # #    # #    # # #    # #####  #      #####   ####     #  #  # # # #  #
         #   #  ###### #####  # ###### #    # #      #           #    #  #  # # #  # #
          # #   #    # #   #  # #    # #    # #      #      #    #    #  #  # # #   ##
           #    #    # #    # # #    # #####  ###### ######  ####      ## ##  # #    #
    """

    def _choose_auto_watches(self, my_locals):
        old_theme = theme()
        theme(_Debugger.DEBUGGER_MAIN_WINDOW_THEME)
        num_cols = 3
        output_text = ""
        num_lines = 2
        cur_col = 0
        layout = [
            [
                Text(
                    'Choose your "Auto Watch" variables',
                    font="ANY 14",
                    text_color="red",
                )
            ]
        ]
        longest_line = max([len(key) for key in my_locals])
        line = []
        sorted_dict = {}
        for key in sorted(my_locals.keys()):
            sorted_dict[key] = my_locals[key]
        for key in sorted_dict:
            line.append(
                CB(
                    key,
                    key=key,
                    size=(longest_line, 1),
                    default=self.local_choices[key]
                    if key in self.local_choices
                    else False,
                )
            )
            if cur_col + 1 == num_cols:
                cur_col = 0
                layout.append(line)
                line = []
            else:
                cur_col += 1
        if cur_col:
            layout.append(line)

        layout += [
            [
                Text("Custom Watch (any expression)"),
                Input(
                    default_text=self.custom_watch, size=(40, 1), key="-CUSTOM_WATCH-"
                ),
            ]
        ]
        layout += [
            [
                Ok(),
                Cancel(),
                Button("Clear All"),
                Button("Select [almost] All", key="-AUTO_SELECT-"),
            ]
        ]

        window = Window(
            "Choose Watches",
            layout,
            icon=PSG_DEBUGGER_LOGO,
            finalize=True,
            keep_on_top=True,
        )

        while True:  # event loop
            event, values = window.read()
            if event in (None, "Cancel", "-EXIT-"):
                break
            elif event == "Ok":
                self.local_choices = values
                self.custom_watch = values["-CUSTOM_WATCH-"]
                break
            elif event == "Clear All":
                popup_quick_message(
                    "Cleared Auto Watches",
                    auto_close=True,
                    auto_close_duration=3,
                    non_blocking=True,
                    text_color="red",
                    font="ANY 18",
                )
                for key in sorted_dict:
                    window.Element(key).Update(False)
                window.Element("-CUSTOM_WATCH-").Update("")
            elif event == "Select All":
                for key in sorted_dict:
                    window.Element(key).Update(False)
            elif event == "-AUTO_SELECT-":
                for key in sorted_dict:
                    window.Element(key).Update(not key.startswith("_"))

        # exited event loop
        window.Close()
        theme(old_theme)

    """
        ######                            #######
        #     # #    # # #      #####     #       #       ####    ##   ##### # #    #  ####
        #     # #    # # #      #    #    #       #      #    #  #  #    #   # ##   # #    #
        ######  #    # # #      #    #    #####   #      #    # #    #   #   # # #  # #
        #     # #    # # #      #    #    #       #      #    # ######   #   # #  # # #  ###
        #     # #    # # #      #    #    #       #      #    # #    #   #   # #   ## #    #
        ######   ####  # ###### #####     #       ######  ####  #    #   #   # #    #  ####
    
        #     #
        #  #  # # #    # #####   ####  #    #
        #  #  # # ##   # #    # #    # #    #
        #  #  # # # #  # #    # #    # #    #
        #  #  # # #  # # #    # #    # # ## #
        #  #  # # #   ## #    # #    # ##  ##
         ## ##  # #    # #####   ####  #    #
    """

    def _build_floating_window(self, location=(None, None)):
        """

        :param location:
        :type location:

        """
        if self.popout_window:  # if floating window already exists, close it first
            self.popout_window.Close()
        old_theme = theme()
        theme(_Debugger.DEBUGGER_POPOUT_THEME)
        num_cols = 2
        width_var = 15
        width_value = 30
        layout = []
        line = []
        col = 0
        # self.popout_choices = self.local_choices
        self.popout_choices = {}
        if (
                self.popout_choices == {}
        ):  # if nothing chosen, then choose all non-_ variables
            for key in sorted(self.locals.keys()):
                self.popout_choices[key] = not key.startswith("_")

        width_var = max([len(key) for key in self.popout_choices])
        for key in self.popout_choices:
            if self.popout_choices[key] is True:
                value = str(self.locals.get(key))
                h = min(
                    len(value) // width_value + 1,
                    _Debugger.MAX_LINES_PER_RESULT_FLOATING,
                )
                line += [
                    Text(
                        "{}".format(key),
                        size=(width_var, 1),
                        font=_Debugger.DEBUGGER_POPOUT_WINDOW_FONT,
                    ),
                    Text(" = ", font=_Debugger.DEBUGGER_POPOUT_WINDOW_FONT),
                    Text(
                        value,
                        key=key,
                        size=(width_value, h),
                        font=_Debugger.DEBUGGER_POPOUT_WINDOW_FONT,
                    ),
                ]
                if col + 1 < num_cols:
                    line += [VerticalSeparator(), T(" ")]
                col += 1
            if col >= num_cols:
                layout.append(line)
                line = []
                col = 0
        if col != 0:
            layout.append(line)
        layout = [
            [T(SYMBOL_X, enable_events=True, key="-EXIT-", font="_ 7")],
            [Column(layout)],
        ]

        Window._read_call_from_debugger = True
        self.popout_window = Window(
            "Floating",
            layout,
            alpha_channel=0,
            no_titlebar=True,
            grab_anywhere=True,
            element_padding=(0, 0),
            margins=(0, 0),
            keep_on_top=True,
            right_click_menu=["&Right", ["Debugger::RightClick", "Exit::RightClick"]],
            location=location,
            finalize=True,
        )
        Window._read_call_from_debugger = False

        if location == (None, None):
            screen_size = self.popout_window.GetScreenDimensions()
            self.popout_window.Move(screen_size[0] - self.popout_window.Size[0], 0)
        self.popout_window.SetAlpha(1)
        theme(old_theme)
        return True

    """
        ######
        #     # ###### ###### #####  ######  ####  #    #
        #     # #      #      #    # #      #      #    #
        ######  #####  #####  #    # #####   ####  ######
        #   #   #      #      #####  #           # #    #
        #    #  #      #      #   #  #      #    # #    #
        #     # ###### #      #    # ######  ####  #    #
    
        #######
        #       #       ####    ##   ##### # #    #  ####
        #       #      #    #  #  #    #   # ##   # #    #
        #####   #      #    # #    #   #   # # #  # #
        #       #      #    # ######   #   # #  # # #  ###
        #       #      #    # #    #   #   # #   ## #    #
        #       ######  ####  #    #   #   # #    #  ####
    
        #     #
        #  #  # # #    # #####   ####  #    #
        #  #  # # ##   # #    # #    # #    #
        #  #  # # # #  # #    # #    # #    #
        #  #  # # #  # # #    # #    # # ## #
        #  #  # # #   ## #    # #    # ##  ##
         ## ##  # #    # #####   ####  #    #
    """

    def _refresh_floating_window(self):
        if not self.popout_window:
            return
        for key in self.popout_choices:
            if self.popout_choices[key] is True and key in self.locals:
                if key is not None and self.popout_window is not None:
                    self.popout_window.Element(key, silent_on_error=True).Update(
                        self.locals.get(key)
                    )
        event, values = self.popout_window.read(timeout=5)
        if event in (None, "_EXIT_", "Exit::RightClick", "-EXIT-"):
            self.popout_window.Close()
            self.popout_window = None
        elif event == "Debugger::RightClick":
            show_debugger_window()


# 888     888                                .d8888b.         d8888 888 888          888      888
# 888     888                               d88P  Y88b       d88888 888 888          888      888
# 888     888                               888    888      d88P888 888 888          888      888
# 888     888 .d8888b   .d88b.  888d888     888            d88P 888 888 888  8888b.  88888b.  888  .d88b.
# 888     888 88K      d8P  Y8b 888P"       888           d88P  888 888 888     "88b 888 "88b 888 d8P  Y8b
# 888     888 "Y8888b. 88888888 888         888    888   d88P   888 888 888 .d888888 888  888 888 88888888
# Y88b. .d88P      X88 Y8b.     888         Y88b  d88P  d8888888888 888 888 888  888 888 d88P 888 Y8b.
#  "Y88888P"   88888P'  "Y8888  888          "Y8888P"  d88P     888 888 888 "Y888888 88888P"  888  "Y8888

# 8888888888                            888    d8b
# 888                                   888    Y8P
# 888                                   888
# 8888888    888  888 88888b.   .d8888b 888888 888  .d88b.  88888b.  .d8888b
# 888        888  888 888 "88b d88P"    888    888 d88""88b 888 "88b 88K
# 888        888  888 888  888 888      888    888 888  888 888  888 "Y8888b.
# 888        Y88b 888 888  888 Y88b.    Y88b.  888 Y88..88P 888  888      X88
# 888         "Y88888 888  888  "Y8888P  "Y888 888  "Y88P"  888  888  88888P'


def show_debugger_window(location=(None, None), *args):
    """
    Shows the large main debugger window
    :param location: Locations (x,y) on the screen to place upper left corner of the window
    :type location:  (int, int)
    :return:         None
    :rtype:          None
    """
    if _Debugger.debugger is None:
        _Debugger.debugger = _Debugger()
    debugger = _Debugger.debugger
    frame = inspect.currentframe()
    prev_frame = inspect.currentframe().f_back
    # frame, *others = inspect.stack()[1]
    try:
        debugger.locals = frame.f_back.f_locals
        debugger.globals = frame.f_back.f_globals
    finally:
        del frame

    if not debugger.watcher_window:
        debugger.watcher_window = debugger._build_main_debugger_window(
            location=location
        )
    return True


def show_debugger_popout_window(location=(None, None), *args):
    """
    Shows the smaller "popout" window.  Default location is the upper right corner of your screen

    :param location: Locations (x,y) on the screen to place upper left corner of the window
    :type location:  (int, int)
    :return:         None
    :rtype:          None
    """
    if _Debugger.debugger is None:
        _Debugger.debugger = _Debugger()
    debugger = _Debugger.debugger
    frame = inspect.currentframe()
    prev_frame = inspect.currentframe().f_back
    # frame = inspect.getframeinfo(prev_frame)
    # frame, *others = inspect.stack()[1]
    try:
        debugger.locals = frame.f_back.f_locals
        debugger.globals = frame.f_back.f_globals
    finally:
        del frame
    if debugger.popout_window:
        debugger.popout_window.Close()
        debugger.popout_window = None
    debugger._build_floating_window(location=location)


def _refresh_debugger():
    """
    Refreshes the debugger windows. USERS should NOT be calling this function. Within PySimpleGUI it is called for the USER every time the Window.Read function is called.

    :return: return code False if user closed the main debugger window.
    :rtype:  (bool)
    """
    if _Debugger.debugger is None:
        _Debugger.debugger = _Debugger()
    debugger = _Debugger.debugger
    Window._read_call_from_debugger = True
    rc = None
    # frame = inspect.currentframe()
    # frame = inspect.currentframe().f_back

    frame, *others = inspect.stack()[1]
    try:
        debugger.locals = frame.f_back.f_locals
        debugger.globals = frame.f_back.f_globals
    finally:
        del frame
    if debugger.popout_window:
        rc = debugger._refresh_floating_window()
    if debugger.watcher_window:
        rc = debugger._refresh_main_debugger_window(debugger.locals, debugger.globals)
    Window._read_call_from_debugger = False
    return rc


def _debugger_window_is_open():
    """
    Determines if one of the debugger window is currently open
    :return: returns True if the popout window or the main debug window is open
    :rtype: (bool)
    """

    if _Debugger.debugger is None:
        return False
    debugger = _Debugger.debugger
    if debugger.popout_window or debugger.watcher_window:
        return True
    return False


def get_versions():
    """
    Returns a human-readable string of version numbers for:

    Python version
    Platform (Win, Mac, Linux)
    Platform version (tuple with information from the platform module)
    PySimpleGUI Port (PySimpleGUI in this case)
    tkinter version
    PySimpleGUI version
    The location of the PySimpleGUI.py file

    The format is a newline between each value and descriptive text for each line

    :return:
    :rtype:  str
    """
    if running_mac():
        platform_name, platform_ver = "Mac", platform.mac_ver()
    elif running_windows():
        platform_name, platform_ver = "Windows", platform.win32_ver()
    elif running_linux():
        platform_name, platform_ver = "Linux", platform.libc_ver()
    else:
        platform_name, platform_ver = "Unknown platorm", "Unknown platform version"

    versions = "Python Interpeter: {}\nPython version: {}.{}.{}\nPlatform: {}\nPlatform version: {}\nPort: {}\ntkinter version: {}\nPySimpleGUI version: {}\nPySimpleGUI filename: {}".format(
        sys.executable,
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro,
        platform_name,
        platform_ver,
        port,
        tclversion_detailed,
        ver,
        __file__,
    )
    return versions


def scheck_hh():
    with open(__file__, "r", encoding="utf8") as file:
        lines_in_file = file.readlines()
    combined_lines = "".join(lines_in_file[:-1])
    entire_file_bytes = bytearray(combined_lines, encoding="utf8")
    cfileh = hh(entire_file_bytes)
    return cfileh.hexdigest()


def read_last_line():
    with open(__file__, "r", encoding="utf8") as file:
        last_line = file.readlines()[-1]
    return last_line


# ==================================================#
#
# MM""""""""`M                     oo oo
# MM  mmmmmmmM
# M`      MMMM 88d8b.d8b. .d8888b. dP dP
# MM  MMMMMMMM 88'`88'`88 88'  `88 88 88
# MM  MMMMMMMM 88  88  88 88.  .88 88 88
# MM        .M dP  dP  dP `88888P' 88 dP
# MMMMMMMMMMMM                     88
#                                  dP
# M""MMM""MMM""M                   dP       dP
# M  MMM  MMM  M                   88       88
# M  MMP  MMP  M .d8888b. 88d888b. 88 .d888b88
# M  MM'  MM' .M 88'  `88 88'  `88 88 88'  `88
# M  `' . '' .MM 88.  .88 88       88 88.  .88
# M    .d  .dMMM `88888P' dP       dP `88888P8
# MMMMMMMMMMMMMM
#
# When things look bleak, show your user an emoji
# and a little hope
# ==================================================#


EMOJI_BASE64_FACEPALM = b" "
EMOJI_BASE64_FRUSTRATED = b" "
EMOJI_BASE64_NOTUNDERSTANDING = b" "
EMOJI_BASE64_PONDER = b" "
EMOJI_BASE64_SAD = b" "
EMOJI_BASE64_SKEPTICAL = b" "
EMOJI_BASE64_THINK = b" "
EMOJI_BASE64_DREAMING = b" "
EMOJI_BASE64_WEARY = b" "
EMOJI_BASE64_YIKES = b" "

EMOJI_BASE64_HAPPY_GASP = b" "

EMOJI_BASE64_HAPPY_IDEA = b" "

EMOJI_BASE64_HAPPY_JOY = b" "

EMOJI_BASE64_HAPPY_LAUGH = b" "

EMOJI_BASE64_HAPPY_STARE = b" "

ICON_BUY_ME_A_COFFEE = b" "

EMOJI_BASE64_HAPPY_RELIEF = b" "

EMOJI_BASE64_HAPPY_BIG_SMILE = b" "
EMOJI_BASE64_HAPPY_CONTENT = b" "

EMOJI_BASE64_HAPPY_HEARTS = b" "

EMOJI_BASE64_HAPPY_THUMBS_UP = b" "

EMOJI_BASE64_HAPPY_WINK = b" "

EMOJI_BASE64_CRY = b" "

EMOJI_BASE64_DEAD = b" "

EMOJI_BASE64_FINGERS_CROSSED = b" "

EMOJI_BASE64_GUESS = b" "

EMOJI_BASE64_CLAP = b" "

EMOJI_BASE64_NO_HEAR = b" "

EMOJI_BASE64_NO_SEE = b" "

EMOJI_BASE64_NO_SPEAK = b" "

EMOJI_BASE64_PRAY = b" "

EMOJI_BASE64_ZIPPED_SHUT = b" "

EMOJI_BASE64_NO_HEAR = b" "

EMOJI_BASE64_MIKE = b" "

EMOJI_BASE64_SUPERHERO = b" "

EMOJI_BASE64_BLANK_STARE = b" "

EMOJI_BASE64_COOL = b" "

EMOJI_BASE64_EYE_ROLL = b" "

EMOJI_BASE64_OK = b" "

EMOJI_BASE64_QUESTION = b" "

EMOJI_BASE64_SKEPTIC = b" "

EMOJI_BASE64_SLEEPING = b" "

EMOJI_BASE64_TEAR = b" "

EMOJI_BASE64_UPSIDE_DOWN = b" "

EMOJI_BASE64_WIZARD = b" "

EMOJI_BASE64_CRAZY = b" "

EMOJI_BASE64_GLASSES = b" "

EMOJI_BASE64_HEAD_EXPLODE = b" "

EMOJI_BASE64_LAPTOP = b" "

EMOJI_BASE64_PARTY = b" "

EMOJI_BASE64_RAINEDON = b" "

EMOJI_BASE64_READING = b" "

EMOJI_BASE64_SANTA = b" "

EMOJI_BASE64_SEARCH = b" "

EMOJI_BASE64_WAVE = b" "

EMOJI_BASE64_DEPRESSED = b" "

EMOJI_BASE64_GOLD_STAR = b" "

EMOJI_BASE64_HONEST = b" "

EMOJI_BASE64_ILL = b" "

EMOJI_BASE64_ILL2 = b" "

EMOJI_BASE64_KEY = b" "

EMOJI_BASE64_MASK = b" "

EMOJI_BASE64_SALUTE = b" "

EMOJI_BASE64_SCREAM = b" "

EMOJI_BASE64_SMIRKING = b" "

EMOJI_BASE64_WARNING2 = b" "

EMOJI_BASE64_WARNING = b" "

EMOJI_BASE64_JEDI = b" "

EMOJI_BASE64_HAPPY_LIST = [
    EMOJI_BASE64_HAPPY_STARE,
    EMOJI_BASE64_BLANK_STARE,
    EMOJI_BASE64_HAPPY_LAUGH,
    EMOJI_BASE64_HAPPY_JOY,
    EMOJI_BASE64_HAPPY_IDEA,
    EMOJI_BASE64_HAPPY_GASP,
    EMOJI_BASE64_HAPPY_RELIEF,
    EMOJI_BASE64_HAPPY_WINK,
    EMOJI_BASE64_HAPPY_THUMBS_UP,
    EMOJI_BASE64_HAPPY_HEARTS,
    EMOJI_BASE64_HAPPY_CONTENT,
    EMOJI_BASE64_HAPPY_BIG_SMILE,
    EMOJI_BASE64_PRAY,
    EMOJI_BASE64_GUESS,
    EMOJI_BASE64_FINGERS_CROSSED,
    EMOJI_BASE64_CLAP,
    EMOJI_BASE64_COOL,
    EMOJI_BASE64_UPSIDE_DOWN,
    EMOJI_BASE64_OK,
    EMOJI_BASE64_COOL,
    EMOJI_BASE64_GLASSES,
    EMOJI_BASE64_HEAD_EXPLODE,
    EMOJI_BASE64_GLASSES,
    EMOJI_BASE64_LAPTOP,
    EMOJI_BASE64_PARTY,
    EMOJI_BASE64_READING,
    EMOJI_BASE64_SANTA,
    EMOJI_BASE64_SEARCH,
    EMOJI_BASE64_WAVE,
    EMOJI_BASE64_KEY,
    EMOJI_BASE64_SALUTE,
    EMOJI_BASE64_HONEST,
    EMOJI_BASE64_WIZARD,
    EMOJI_BASE64_JEDI,
    EMOJI_BASE64_GOLD_STAR,
    EMOJI_BASE64_SMIRKING,
]

EMOJI_BASE64_SAD_LIST = [
    EMOJI_BASE64_YIKES,
    EMOJI_BASE64_WEARY,
    EMOJI_BASE64_DREAMING,
    EMOJI_BASE64_SLEEPING,
    EMOJI_BASE64_THINK,
    EMOJI_BASE64_SKEPTIC,
    EMOJI_BASE64_SKEPTICAL,
    EMOJI_BASE64_FACEPALM,
    EMOJI_BASE64_FRUSTRATED,
    EMOJI_BASE64_PONDER,
    EMOJI_BASE64_NOTUNDERSTANDING,
    EMOJI_BASE64_QUESTION,
    EMOJI_BASE64_CRY,
    EMOJI_BASE64_TEAR,
    EMOJI_BASE64_DEAD,
    EMOJI_BASE64_ZIPPED_SHUT,
    EMOJI_BASE64_NO_HEAR,
    EMOJI_BASE64_NO_SEE,
    EMOJI_BASE64_NO_SPEAK,
    EMOJI_BASE64_EYE_ROLL,
    EMOJI_BASE64_CRAZY,
    EMOJI_BASE64_RAINEDON,
    EMOJI_BASE64_DEPRESSED,
    EMOJI_BASE64_ILL,
    EMOJI_BASE64_ILL2,
    EMOJI_BASE64_MASK,
    EMOJI_BASE64_WARNING,
    EMOJI_BASE64_WARNING2,
    EMOJI_BASE64_SCREAM,
]
EMOJI_BASE64_LIST = EMOJI_BASE64_HAPPY_LIST + EMOJI_BASE64_SAD_LIST

EMOJI_BASE64_JASON = EMOJI_BASE64_WIZARD
EMOJI_BASE64_TANAY = EMOJI_BASE64_JEDI


def _random_error_emoji():
    c = random.choice(EMOJI_BASE64_SAD_LIST)
    return c


def _random_happy_emoji():
    c = random.choice(EMOJI_BASE64_HAPPY_LIST)
    return c


'''
M"""""`'"""`YM                            
M  mm.  mm.  M                            
M  MMM  MMM  M .d8888b. 88d888b. .d8888b. 
M  MMM  MMM  M 88'  `88 88'  `88 88ooood8 
M  MMM  MMM  M 88.  .88 88       88.  ... 
M  MMM  MMM  M `88888P' dP       `88888P' 
MMMMMMMMMMMMMM                            
                                          
M#"""""""'M                             .d8888P dP   dP 
##  mmmm. `M                            88'     88   88 
#'        .M .d8888b. .d8888b. .d8888b. 88baaa. 88aaa88 
M#  MMMb.'YM 88'  `88 Y8ooooo. 88ooood8 88` `88      88 
M#  MMMM'  M 88.  .88       88 88.  ... 8b. .d8      88 
M#       .;M `88888P8 `88888P' `88888P' `Y888P'      dP 
M#########M                                             
                                                        
M""M                                                
M  M                                                
M  M 88d8b.d8b. .d8888b. .d8888b. .d8888b. .d8888b. 
M  M 88'`88'`88 88'  `88 88'  `88 88ooood8 Y8ooooo. 
M  M 88  88  88 88.  .88 88.  .88 88.  ...       88 
M  M dP  dP  dP `88888P8 `8888P88 `88888P' `88888P' 
MMMM                          .88                   
                          d8888P
'''

"""

90 x 90 pixel images

These images are intentionally a little large so that you can use the image_subsample to reduce their size.

This offers more flexibility for use in a main window (larger) or perhaps a titlebar (smaller)

"""

HEART_FLAT_BASE64 = b" "

HEART_3D_BASE64 = b" "

PYTHON_COLORED_HEARTS_BASE64 = b" "

RED_X_BASE64 = b" "

GREEN_CHECK_BASE64 = b"iVBORw0KGgoAAAANSUhEUgAAAFoAAABaCAYAAAA4qEECAAAJV0lEQVR4nO2cTWwc5RnHf8/M7Dq7ttdxIIIUcqGA1BQU6Ac9VSkp0NwoJE5PJJygKki9tIIEO7ND3ICEeqJUJYcqCYdKDoS0lWgpH21KuVShH/TjUolLkIpKguO1vWvvfDw9zOxH1l8zjnc3Xs/vFEXy7uzPz/7f93nnGUNKSkpKSkpKSkpKSkpKzyFMYDKC2e0L2TjYGN2+hN5DkXoVP1s4wdjgDwB4jEw3L6u30CguAJzCCV4YUp4bUuzC94BlZaclHx9hPwb78bELp8jJQaa1yrx65OQljhSe4DguLy8uOxUdhzAuDE5HkvvlEWbVRcgSYDKnHnn5CXbhSR5fXHYqemXCSj6Nj1M4Qb88wrR6EMkUpC47Jy8yFsm2sa58kZSlUYTTUVw4hRPkjIPMBC6ySDwoioHPJrEo65M8W3qJx8hwHBdS0UujTZVcLJwkLweY0cUlN35GEQJyYlLRJ3BKP2UEk9P4qejFWTyTibGFq1V2ViwqPMXRqRcYwUgzupXmha9YOJlIMoSZ7ROQEZBgJ6DsQNKKbmZBJsvBFeOilQCPQbGo6Ens0qNRdARpRddollwsnAwXPq0mkgwug2Ixq69glx7Fjr4ZoGlFhyzM5KSVrLgMSIZZfQWndKBWyYBCuo9erhlJIrnKgJGhrKdwSgeYwGSiIRnS7V1Dci2Tp9XDuLLZWJZaJdcyOTw6DZCGZNjIFR0eEDVJNsKFL4lkIsllPVVf+BaRDBu1olfTjCzEpX/pTG5lI1Z0Q7JdOEVeDqwik0PJtUweWZjJrWws0VfbjISv4TJghJlcLB2sL3yLxEUzGyc62tiMsEwl19gYFd2OZiRGXDSzESq67c1IHHq7ojvUjMShlyu6Y81IHHqzojvcjMSh9yq6C81IHHqtorvSjMShd0R3sRmJQ29ER5ebkTjEE21j8EWE/fhr8aZrTFhvgoaZbBxgJqgiZBO8xsJMXqNKblzkStgYOAQL/n2tUB9UKfy8W81IHJbPaBsLh4DRgS8wVvgWDkHrBE5Xscni4Bk69H2GjEeY1fluNCNxWLqid2FxDo9nCp8ny/v0yQ1U/L04M2d4mQyPhxM4XSOaAio4N391Wqbf0ECHUQzixuEaNiNxWLyi7Ujy6OBtZHkPU25gTj2yxgSjAw8vNlvWUWwsjuMOjt30tWlj5k019HoChPiL+5o2I3FYeGFhXHg8PXg7A/I2yHaq6gMGJoopwpz/MOMzZ5tnyzpGdH2FwzffM52f+Y1qsAUXH4n9iMOaNyNxuFJ0TfIPB29jSN5BZDvz6iFR9SoayTZw/YdwZs52NEai68uPfu7uSt/sO4oOJ5KsTZVcLB1sx+5iKRqiJzDZj8/TQ7eQ1z9iyk3M68IP0ZAtzLGP8akz0aJUbeuVRpKH7G1fKlmz7yoMJZdsZKgEHcnkVsKMtuuT7LeS1/eXlAy12TLBVyXHBIcH9uJQbeszHJHk3OEbvzJllkPJVYLYkgO8cOELGs3I/s5JBpDGE0XDOzD9NzBl+5KSm1ECTMACZoN9HJt5vS2ZXYuLseu/XO5z30T1uqvO5A7FRTMG1JoQ/2fkje1UtIoR40MIBj7gAXnjDKMD3+Y47ppWdiQ5Yw/dVelzf5tYsi6x8HVYMoSig7Cqze9SDi6QkyxBzFY7lB2OqW4yXmds6KHlHphJxGNkcPAyo1t3ehbvqOr1CSV3rBmJQ6Oldib/ic9ufP2EPjHR2LKlIZtXGRvYy+O49cfEVkO0T87bW+9ys/PnFN0SO5MVRZlnQLJUgsYpXAcXvsVIvutYilpmmyjzwXc4OnOmfmyZhFpcjA7d7fbxFnAdbszrCKfthYJAqfNbuOVodIb78bGxeH7qI6b1XlQvRJXtxXolwcADAkyxjBMjE3YmPIBPcObdLHkTb5JMsk8WEZVJqyRPUiwdBOhWJrdypQQHDxuLF6b/w4zeh+oFsmLFjhEDAx9fTcm99u8Xz47YI1mKaCzZtWZpdPhOt4+3UN2aSHIGUzAuDTK4xytefimKLqFLmdzK4mcD9Q89eBsZOYcl2xLFSEDAgBjGvPHruz++Ze8H2z4If1FLHbHWK3n4TjfrncOQYaoxF76G5MlBb2BPyfn4zx1poBKy8uldmNl/wkwoO9paSdX45b4P79t7esfpsLJaZdclb97pZv3fIxK/rQ4IyGJIwPRgMLS75Fw435Xzlxgs/ZU+F8XI81MfUeLrBPoxfSTZjWSYVVezwYOv3vm718SRULA2/XJr3xw7f5e7Sd9GjPiSw0w2BJnMycCuknPhfG23Euv6OkycOyxXnuaJbGdO/VhNTUhY2WX9lRZLD9ZFFzFx8Hgqv5NB6y2QrVQTZrLIpZybeaDsXPxL/TqvUeLeM2zIzsu7GHJTbCnQfGp2ln+V9rEDwcHjUP8d5M0/APE7vkgyyKWcl9tTcT45f61LhiR3weuyC7eS5z1MuXE1mY2rZxgt7cUevgPLfw9hc+yFL8pk4HK+2n9f+eh/P1gPkiHpuMHVNzUeebGoBOdAbiebYIGtVzKXM17fva7z6d/Wi2RYzVzHSjcHViIgICcGnoIbdXIr0ZTJltu323X+9+F6kgyrHaBZ7HbXfIJJzXDnIkiMRkbxyYiJcDE/n9lTPnpx3cRFM6ufVGptavpkG+UEMRKHmmT4LFPJ3O8eu/Z3F0txdSNhTU2N5PmFCvfgaxDd9r86wn2yic9UxjV2ueOX/75eJcNazN5F00uCYBS3OH7OO0I54XBhK7WFT+Qz5oxvMD75j/UsGdZqyDE8NDLEEc90ho94m3yHirooVuL3UHyyYgKfUuYBjk2tq93FUqztNKmNJQ6e6WwZ9Tb5R6moF8mOR9PCl5njAXd86q+9IBnaMbYbyRZ782iQ11B2gLXiO9UkazBJ1byXdZ7JrbRjPlqww3MMoyF7+RipLXyBTlK1dvVCJrfSvkH0aILJKBaeCXIyHi2QC2XXFz4uMufvZny25yRDOx+tiP6iYVAs/YiKHiYvGcLhhMYdj3omy6e43v29Khk68WhF7SD+SOEQ/XIsWiBNlCBqRi4xL9/stUxupf0PCx2PRnyfLT3HrH+YnFgoLhlMVC9T9nb3uuTOUptgOlI4xI+HlKOFixzqvwNoejwiZW2oCS0WnuBw4Z4r/i9ljWkePUj/ZHubsbFSySkpKSkpKSkpKSkpKSkpKW3g/3+PYisYNf7zAAAAAElFTkSuQmCC"

'''
M""MMMMM""M                                           dP          
M  MMMMM  M                                           88          
M  MMMMM  M 88d888b. .d8888b. 88d888b. .d8888b. .d888b88 .d8888b. 
M  MMMMM  M 88'  `88 88'  `88 88'  `88 88'  `88 88'  `88 88ooood8 
M  `MMM'  M 88.  .88 88.  .88 88       88.  .88 88.  .88 88.  ... 
Mb       dM 88Y888P' `8888P88 dP       `88888P8 `88888P8 `88888P' 
MMMMMMMMMMM 88            .88                                     
            dP        d8888P                                      
MP""""""`MM                            oo                   
M  mmmmm..M                                                 
M.      `YM .d8888b. 88d888b. dP   .dP dP .d8888b. .d8888b. 
MMMMMMM.  M 88ooood8 88'  `88 88   d8' 88 88'  `"" 88ooood8 
M. .MMM'  M 88.  ... 88       88 .88'  88 88.  ... 88.  ... 
Mb.     .dM `88888P' dP       8888P'   dP `88888P' `88888P' 
MMMMMMMMMMM
'''

__upgrade_server_ip = "upgradeapi.PySimpleGUI.com"
__upgrade_server_port = "5353"


def __send_dict(ip, port, dict_to_send):
    """
    Send a dictionary to the upgrade server and get back a dictionary in response
    :param ip:           ip address of the upgrade server
    :type ip:            str
    :param port:         port number
    :type port:          int | str
    :param dict_to_send: dictionary of items to send
    :type dict_to_send:  dict
    :return:             dictionary that is the reply
    :rtype:              dict
    """

    # print(f'sending dictionary to ip {ip} port {port}')
    try:
        # Create a socket object
        s = socket.socket()

        s.settimeout(5.0)  # set a 5 second timeout

        # connect to the server on local computer
        s.connect((ip, int(port)))
        # send a python dictionary
        s.send(json.dumps(dict_to_send).encode())

        # receive data from the server
        reply_data = s.recv(1024).decode()
        # close the connection
        s.close()
    except Exception as e:
        # print(f'Error sending to server:', e)
        # print(f'payload:\n', dict_to_send)
        reply_data = e
    try:
        data_dict = json.loads(reply_data)
    except Exception as e:
        # print(f'UPGRADE THREAD - Error decoding reply {reply_data} as a dictionary. Error = {e}')
        data_dict = {}
    return data_dict


def __show_previous_upgrade_information():
    """
    Shows information about upgrades if upgrade information is waiting to be shown

    :return:
    """

    # if nothing to show, then just return
    if pysimplegui_user_settings.get(
            "-upgrade info seen-", True
    ) and not pysimplegui_user_settings.get("-upgrade info available-", False):
        return
    if (
            pysimplegui_user_settings.get("-upgrade show only critical-", False)
            and pysimplegui_user_settings.get("-severity level-", "") != "Critical"
    ):
        return

    message1 = pysimplegui_user_settings.get("-upgrade message 1-", "")
    message2 = pysimplegui_user_settings.get("-upgrade message 2-", "")
    recommended_version = pysimplegui_user_settings.get("-upgrade recommendation-", "")
    severity_level = pysimplegui_user_settings.get("-severity level-", "")

    if severity_level != "Critical":
        return

    layout = [
        [
            Image(EMOJI_BASE64_HAPPY_THUMBS_UP),
            T("An upgrade is available & recommended", font="_ 14"),
        ],
        [T("It is recommended you upgrade to version {}".format(recommended_version))],
        [T(message1, enable_events=True, k="-MESSAGE 1-")],
        [T(message2, enable_events=True, k="-MESSAGE 2-")],
        [
            CB(
                "Do not show this message again in the future",
                default=True,
                k="-SKIP IN FUTURE-",
            )
        ],
        [
            B("Close"),
            T("This window auto-closes in"),
            T("30", k="-CLOSE TXT-", text_color="white", background_color="red"),
            T("seconds"),
        ],
    ]

    window = Window("PySimpleGUI Intelligent Upgrade", layout, finalize=True)
    if "http" in message1:
        window["-MESSAGE 1-"].set_cursor("hand1")
    if "http" in message2:
        window["-MESSAGE 2-"].set_cursor("hand1")

    seconds_left = 30
    while True:
        event, values = window.read(timeout=1000)
        if event in ("Close", WIN_CLOSED) or seconds_left < 1:
            break
        if values["-SKIP IN FUTURE-"]:
            if not running_trinket():
                pysimplegui_user_settings["-upgrade info available-"] = False
                pysimplegui_user_settings["-upgrade info seen-"] = True
        if event == "-MESSAGE 1-" and "http" in message1 and webbrowser_available:
            webbrowser.open_new_tab(message1)
        elif event == "-MESSAGE 2-" and "http" in message2 and webbrowser_available:
            webbrowser.open_new_tab(message2)
        window["-CLOSE TXT-"].update(seconds_left)
        seconds_left -= 1

    window.close()


def __get_linux_distribution():
    line_tuple = ("Linux Distro", "Unknown", "No lines Found in //etc//os-release")
    try:
        with open("/etc/os-release") as f:
            data = f.read()
        lines = data.split("\n")
        for line in lines:
            if line.startswith("PRETTY_NAME"):
                line_split = line.split("=")[1].strip('"')
                line_tuple = tuple(line_split.split(" "))
                return line_tuple
    except Exception:
        line_tuple = (
            "Linux Distro",
            "Exception",
            "Error reading//processing //etc//os-release",
        )

    return line_tuple


# =========================================================================#
# MP""""""`MM                                                                dP                  dP
# M  mmmmm..M                                                                88                  88
# M.      `YM .d8888b. 88d888b. .d8888b. .d8888b. .d8888b. 88d888b. .d8888b. 88d888b. .d8888b. d8888P
# MMMMMMM.  M 88'  `"" 88'  `88 88ooood8 88ooood8 88ooood8 88'  `88 Y8ooooo. 88'  `88 88'  `88   88
# M. .MMM'  M 88.  ... 88       88.  ... 88.  ... 88.  ... 88    88       88 88    88 88.  .88   88
# Mb.     .dM `88888P' dP       `88888P' `88888P' `88888P' dP    dP `88888P' dP    dP `88888P'   dP
# MMMMMMMMMMM
#
# M"""""`'"""`YM                   oo
# M  mm.  mm.  M
# M  MMM  MMM  M .d8888b. .d8888b. dP .d8888b.
# M  MMM  MMM  M 88'  `88 88'  `88 88 88'  `""
# M  MMM  MMM  M 88.  .88 88.  .88 88 88.  ...
# M  MMM  MMM  M `88888P8 `8888P88 dP `88888P'
# MMMMMMMMMMMMMM               .88
#                          d8888P
# M#"""""""'M                    oo                      M""MMMMM""MM
# ##  mmmm. `M                                           M  MMMMM  MM
# #'        .M .d8888b. .d8888b. dP 88d888b. .d8888b.    M         `M .d8888b. 88d888b. .d8888b.
# M#  MMMb.'YM 88ooood8 88'  `88 88 88'  `88 Y8ooooo.    M  MMMMM  MM 88ooood8 88'  `88 88ooood8
# M#  MMMM'  M 88.  ... 88.  .88 88 88    88       88    M  MMMMM  MM 88.  ... 88       88.  ...
# M#       .;M `88888P' `8888P88 dP dP    dP `88888P'    M  MMMMM  MM `88888P' dP       `88888P'
# M#########M                .88                         MMMMMMMMMMMM
#                        d8888P
# =========================================================================#


# =========================================================================#

# MP""""""`MM   dP                       dP               .8888b
# M  mmmmm..M   88                       88               88   "
# M.      `YM d8888P .d8888b. 88d888b. d8888P    .d8888b. 88aaa
# MMMMMMM.  M   88   88'  `88 88'  `88   88      88'  `88 88
# M. .MMM'  M   88   88.  .88 88         88      88.  .88 88
# Mb.     .dM   dP   `88888P8 dP         dP      `88888P' dP
# MMMMMMMMMMM
#
# dP dP                     oo          dP dP
# dP dP                                 dP dP
#       88d8b.d8b. .d8888b. dP 88d888b.
#       88'`88'`88 88'  `88 88 88'  `88
#       88  88  88 88.  .88 88 88    88
#       dP  dP  dP `88888P8 dP dP    dP
#
#
# MM""""""""`M            dP
# MM  mmmmmmmM            88
# M`      MMMM 88d888b. d8888P 88d888b. dP    dP
# MM  MMMMMMMM 88'  `88   88   88'  `88 88    88
# MM  MMMMMMMM 88    88   88   88       88.  .88
# MM        .M dP    dP   dP   dP       `8888P88
# MMMMMMMMMMMM                               .88
#                                        d8888P
# MM"""""""`YM          oo            dP
# MM  mmmmm  M                        88
# M'        .M .d8888b. dP 88d888b. d8888P .d8888b.
# MM  MMMMMMMM 88'  `88 88 88'  `88   88   Y8ooooo.
# MM  MMMMMMMM 88.  .88 88 88    88   88         88
# MM  MMMMMMMM `88888P' dP dP    dP   dP   `88888P'
# MMMMMMMMMMMM

# ==========================================================================#


# M"""""`'"""`YM          oo
# M  mm.  mm.  M
# M  MMM  MMM  M .d8888b. dP 88d888b.
# M  MMM  MMM  M 88'  `88 88 88'  `88
# M  MMM  MMM  M 88.  .88 88 88    88
# M  MMM  MMM  M `88888P8 dP dP    dP
# MMMMMMMMMMMMMM
#
# MM"""""""`YM                     dP      MM'"""""`MM oo   dP   M""MMMMM""MM          dP
# MM  mmmmm  M                     88      M' .mmm. `M      88   M  MMMMM  MM          88
# M'        .M .d8888b. .d8888b. d8888P    M  MMMMMMMM dP d8888P M         `M dP    dP 88d888b.
# MM  MMMMMMMM 88'  `88 Y8ooooo.   88      M  MMM   `M 88   88   M  MMMMM  MM 88    88 88'  `88
# MM  MMMMMMMM 88.  .88       88   88      M. `MMM' .M 88   88   M  MMMMM  MM 88.  .88 88.  .88
# MM  MMMMMMMM `88888P' `88888P'   dP      MM.     .MM dP   dP   M  MMMMM  MM `88888P' 88Y8888'
# MMMMMMMMMMMM                             MMMMMMMMMMM           MMMMMMMMMMMM
#
# M""M
# M  M
# M  M .d8888b. .d8888b. dP    dP .d8888b.
# M  M Y8ooooo. Y8ooooo. 88    88 88ooood8
# M  M       88       88 88.  .88 88.  ...
# M  M `88888P' `88888P' `88888P' `88888P'
# MMMM


def _github_issue_post_make_markdown(
        issue_type,
        operating_system,
        os_ver,
        psg_port,
        psg_ver,
        gui_ver,
        python_ver,
        python_exp,
        prog_exp,
        used_gui,
        gui_notes,
        cb_docs,
        cb_demos,
        cb_demo_port,
        cb_readme_other,
        cb_command_line,
        cb_issues,
        cb_latest_pypi,
        cb_github,
        detailed_desc,
        code,
        project_details,
        where_found,
):
    body = """
## Type of Issue (Enhancement, Error, Bug, Question)

{}

----------------------------------------

## Environment 

#### Operating System

{}  version {}

#### PySimpleGUI Port (tkinter, Qt, Wx, Web)

{}

----------------------------------------

## Versions


#### Python version (`sg.sys.version`)

{}

#### PySimpleGUI Version (`sg.__version__`)

{}

#### GUI Version  (tkinter (`sg.tclversion_detailed`), PySide2, WxPython, Remi)

{}
""".format(
        issue_type,
        operating_system,
        os_ver,
        psg_port,
        python_ver,
        psg_ver,
        gui_ver,
        project_details,
    )

    body2 = """


---------------------

## Your Experience In Months or Years (optional)

{} Years Python programming experience
{} Years Programming experience overall
{} Have used another Python GUI Framework? (tkinter, Qt, etc) (yes/no is fine)
{}

---------------------

## Troubleshooting

These items may solve your problem. Please check those you've done by changing - [ ] to - [X]

- [{}] Searched main docs for your problem  www.PySimpleGUI.org
- [{}] Looked for Demo Programs that are similar to your goal. It is recommend you use the Demo Browser! Demos.PySimpleGUI.org
- [{}] If not tkinter - looked for Demo Programs for specific port
- [{}] For non tkinter - Looked at readme for your specific port if not PySimpleGUI (Qt, WX, Remi)
- [{}] Run your program outside of your debugger (from a command line)
- [{}] Searched through Issues (open and closed) to see if already reported Issues.PySimpleGUI.org
- [{}] Upgraded to the latest official release of PySimpleGUI on PyPI
- [{}] Tried using the PySimpleGUI.py file on GitHub. Your problem may have already been fixed but not released

## Detailed Description

{}

#### Code To Duplicate


```python
{}


```

#### Screenshot, Sketch, or Drawing



""".format(
        python_exp,
        prog_exp,
        used_gui,
        gui_notes,
        cb_docs,
        cb_demos,
        cb_demo_port,
        cb_readme_other,
        cb_command_line,
        cb_issues,
        cb_latest_pypi,
        cb_github,
        detailed_desc,
        code if len(code) > 10 else "# Paste your code here",
    )

    if project_details or where_found:
        body2 += "------------------------"

    if project_details:
        body2 += """
## Watcha Makin?
{}
""".format(str(project_details))

    if where_found:
        body2 += """
## How did you find PySimpleGUI?
{}
""".format(str(where_found))
    return body + body2


def _github_issue_post_make_github_link(title, body):
    pysimplegui_url = "https://github.com/PySimpleGUI/PySimpleGUI"
    pysimplegui_issues = "{}/issues/new?".format(pysimplegui_url)

    # Fix body cuz urllib can't do it smfh
    getVars = {"title": str(title), "body": str(body)}
    return pysimplegui_issues + urllib.parse.urlencode(getVars).replace("%5Cn", "%0D")


#########################################################################################################


def _github_issue_post_validate(values, checklist, issue_types):
    issue_type = None
    for itype in issue_types:
        if values[itype]:
            issue_type = itype
            break
    if issue_type is None:
        popup_error("Must choose issue type", keep_on_top=True)
        return False
    if values["-OS WIN-"]:
        os_ver = values["-OS WIN VER-"]
    elif values["-OS LINUX-"]:
        os_ver = values["-OS LINUX VER-"]
    elif values["-OS MAC-"]:
        os_ver = values["-OS MAC VER-"]
    elif values["-OS OTHER-"]:
        os_ver = values["-OS OTHER VER-"]
    else:
        popup_error("Must choose Operating System", keep_on_top=True)
        return False

    if os_ver == "":
        popup_error("Must fill in an OS Version", keep_on_top=True)
        return False

    checkboxes = any([values[("-CB-", i)] for i in range(len(checklist))])
    if not checkboxes:
        popup_error(
            "None of the checkboxes were checked.... you need to have tried something...anything...",
            keep_on_top=True,
        )
        return False

    title = values["-TITLE-"].strip()
    if len(title) == 0:
        popup_error("Title can't be blank", keep_on_top=True)
        return False
    elif title[1: len(title) - 1] == issue_type:
        popup_error(
            "Title can't be blank (only the type of issue isn't enough)",
            keep_on_top=True,
        )
        return False

    if len(values["-ML DETAILS-"]) < 4:
        popup_error("A little more details would be awesome", keep_on_top=True)
        return False

    return True


def _github_issue_help():
    heading_font = "_ 12 bold underline"
    text_font = "_ 10"

    def HelpText(text):
        return Text(text, size=(80, None), font=text_font)

    help_why = """ Let's start with a review of the Goals of the PySimpleGUI project
1. To have fun
2. For you to be successful

This form is as important as the documentation and the demo programs to meeting those goals.

The GitHub Issue GUI is here to help you more easily log issues on the PySimpleGUI GitHub Repo. """

    help_goals = """ The goals of using GitHub Issues for PySimpleGUI question, problems and suggestions are:
* Give you direct access to engineers with the most knowledge of PySimpleGUI
* Answer your questions in the most precise and correct way possible
* Provide the highest quality solutions possible
* Give you a checklist of things to try that may solve the problem
* A single, searchable database of known problems and their workarounds
* Provide a place for the PySimpleGUI project to directly provide support to users
* A list of requested enhancements
* An easy to use interface to post code and images
* A way to track the status and have converstaions about issues
* Enable multiple people to help users """

    help_explain = """ GitHub does not provide a "form" that normal bug-tracking-databases provide. As a result, a form was created specifically for the PySimpleGUI project.

The most obvious questions about this form are
* Why is there a form? Other projects don't have one?
* My question is an easy one, why does it still need a form?

The answer is:
I want you to get your question answered with the highest quality answer possible as quickly as possible.

The longer answer - For quite a while there was no form. It resulted the same back and forth, multiple questions comversation.  "What version are you running?"  "What OS are you using?"  These waste precious time.

If asking nicely helps... PLEASE ... please fill out the form.

I can assure you that this form is not here to punish you. It doesn't exist to make you angry and frustrated.  It's not here for any purpose than to try and get you support and make PySimpleGUI better. """

    help_experience = """ Not many Bug-tracking systems ask about you as a user. Your experience in programming, programming in Python and programming a GUI are asked to provide you with the best possible answer.  Here's why it's helpful.  You're a human being, with a past, and a some amount of experience.  Being able to taylor the reply to your issue in a way that fits you and your experience will result in a reply that's efficient and clear.  It's not something normally done but perhaps it should be. It's meant to provide you with a personal response.

If you've been programming for a month, the person answering your question can answer your question in a way that's understandable to you.  Similarly, if you've been programming for 20 years and have used multiple Python GUI frameworks, then you are unlikely to need as much explanation.  You'll also have a richer GUI vocabularly. It's meant to try and give you a peronally crafted response that's on your wavelength. Fun & success... Remember those are our shared goals"""

    help_steps = """ The steps to log an issue are:
1. Fill in the form
2. Click Post Issue """

    # layout = [  [T('Goals', font=heading_font, pad=(0,0))],
    #             [HelpText(help_goals)],
    #             [T('Why?', font=heading_font, pad=(0,0))],
    #             [HelpText(help_why)],
    #             [T('FAQ', font=heading_font, pad=(0,0))],
    #             [HelpText(help_explain)],
    #             [T('Experience (optional)', font=heading_font)],
    #             [HelpText(help_experience)],
    #             [T('Steps', font=heading_font, pad=(0,0))],
    #             [HelpText(help_steps)],
    #             [B('Close')]]

    t_goals = Tab("Goals", [[HelpText(help_goals)]])
    t_why = Tab("Why", [[HelpText(help_why)]])
    t_faq = Tab("FAQ", [[HelpText(help_explain)]])
    t_exp = Tab("Experience", [[HelpText(help_experience)]])
    t_steps = Tab("Steps", [[HelpText(help_steps)]])

    layout = [[TabGroup([[t_goals, t_why, t_faq, t_exp, t_steps]])], [B("Close")]]

    Window("GitHub Issue GUI Help", layout, keep_on_top=True).read(close=True)

    return


def main_open_github_issue():
    font_frame = "_ 14"
    issue_types = ("Question", "Bug", "Enhancement", "Error Message")
    frame_type = [
        [Radio(t, 1, size=(10, 1), enable_events=True, k=t)] for t in issue_types
    ]

    v_size = (15, 1)
    frame_versions = [
        [T("Python", size=v_size), In(sys.version, size=(20, 1), k="-VER PYTHON-")],
        [T("PySimpleGUI", size=v_size), In(ver, size=(20, 1), k="-VER PSG-")],
        [
            T("tkinter", size=v_size),
            In(tclversion_detailed, size=(20, 1), k="-VER TK-"),
        ],
    ]

    frame_platforms = [
        [T("OS                 "), T("Details")],
        [
            Radio("Windows", 2, running_windows(), size=(8, 1), k="-OS WIN-"),
            In(size=(8, 1), k="-OS WIN VER-"),
        ],
        [
            Radio("Linux", 2, running_linux(), size=(8, 1), k="-OS LINUX-"),
            In(size=(8, 1), k="-OS LINUX VER-"),
        ],
        [
            Radio("Mac", 2, running_mac(), size=(8, 1), k="-OS MAC-"),
            In(size=(8, 1), k="-OS MAC VER-"),
        ],
        [
            Radio("Other", 2, size=(8, 1), k="-OS OTHER-"),
            In(size=(8, 1), k="-OS OTHER VER-"),
        ],
    ]

    col_experience = [
        [T("Optional Experience Info")],
        [In(size=(4, 1), k="-EXP PROG-"), T("Years Programming")],
        [In(size=(4, 1), k="-EXP PYTHON-"), T("Years Writing Python")],
        [CB("Previously programmed a GUI", k="-CB PRIOR GUI-")],
        [T("Share more if you want....")],
        [In(size=(25, 1), k="-EXP NOTES-", expand_x=True)],
    ]

    checklist = (
        ("Searched main docs for your problem", "www.PySimpleGUI.org"),
        (
            "Looked for Demo Programs that are similar to your goal.\nIt is recommend you use the Demo Browser!",
            "https://Demos.PySimpleGUI.org",
        ),
        ("If not tkinter - looked for Demo Programs for specific port", ""),
        (
            "For non tkinter - Looked at readme for your specific port if not PySimpleGUI (Qt, WX, Remi)",
            "",
        ),
        ("Run your program outside of your debugger (from a command line)", ""),
        (
            "Searched through Issues (open and closed) to see if already reported",
            "https://Issues.PySimpleGUI.org",
        ),
        (
            "Upgraded to the latest official release of PySimpleGUI on PyPI",
            "https://Upgrading.PySimpleGUI.org",
        ),
        (
            "Tried using the PySimpleGUI.py file on GitHub. Your problem may have already been fixed but not released.",
            "",
        ),
    )

    checklist_col1 = Col(
        [
            [CB(c, k=("-CB-", i)), T(t, k="-T{}-".format(i), enable_events=True)]
            for i, (c, t) in enumerate(checklist[:4])
        ],
        k="-C FRAME CBs1-",
    )
    checklist_col2 = Col(
        [
            [
                CB(c, k=("-CB-", i + 4)),
                T(t, k="-T{}-".format(i + 4), enable_events=True),
            ]
            for i, (c, t) in enumerate(checklist[4:])
        ],
        pad=(0, 0),
        k="-C FRAME CBs2-",
    )
    checklist_tabgropup = TabGroup(
        [
            [
                Tab("Checklist 1 *", [[checklist_col1]], expand_x=True, expand_y=True),
                Tab("Checklist 2  *", [[checklist_col2]]),
                Tab("Experience", col_experience, k="-Tab Exp-", pad=(0, 0)),
            ]
        ],
        expand_x=True,
        expand_y=True,
    )

    frame_details = [
        [
            Multiline(
                size=(65, 10),
                font="Courier 10",
                k="-ML DETAILS-",
                expand_x=True,
                expand_y=True,
            )
        ]
    ]

    tooltip_project_details = "If you care to share a little about your project,\nthen by all means tell us what you are making!"
    frame_project_details = [
        [
            Multiline(
                size=(65, 10),
                font="Courier 10",
                k="-ML PROJECT DETAILS-",
                expand_x=True,
                expand_y=True,
                tooltip=tooltip_project_details,
            )
        ]
    ]

    tooltip_where_find_psg = "Where did you learn about PySimpleGUI?"
    frame_where_you_found_psg = [
        [
            Multiline(
                size=(65, 10),
                font="Courier 10",
                k="-ML FOUND PSG-",
                expand_x=True,
                expand_y=True,
                tooltip=tooltip_where_find_psg,
            )
        ]
    ]

    tooltip_code = "A short program that can be immediately run will considerably speed up getting you quality help."
    frame_code = [
        [
            Multiline(
                size=(80, 10),
                font="Courier 8",
                k="-ML CODE-",
                expand_x=True,
                expand_y=True,
                tooltip=tooltip_code,
            )
        ]
    ]

    frame_markdown = [
        [
            Multiline(
                size=(80, 10),
                font="Courier 8",
                k="-ML MARKDOWN-",
                expand_x=True,
                expand_y=True,
            )
        ]
    ]

    top_layout = [
        [
            Col(
                [[Text("Open A GitHub Issue (* = Required Info)", font="_ 15")]],
                expand_x=True,
            ),
            Col([[B("Help")]]),
        ],
        [
            Frame(
                "Title *",
                [[Input(k="-TITLE-", size=(50, 1), font="_ 14", focus=True)]],
                font=font_frame,
            )
        ],
        # Image(data=EMOJI_BASE64_WEARY)],
        vtop(
            [
                Frame("Platform *", frame_platforms, font=font_frame),
                Frame("Type of Issue *", frame_type, font=font_frame),
                Frame("Versions *", frame_versions, font=font_frame),
            ]
        ),
    ]

    middle_layout = [
        [
            Frame(
                "Checklist * (note that you can click the links)",
                [[checklist_tabgropup]],
                font=font_frame,
                k="-CLIST FRAME-",
                expand_x=True,
                expand_y=True,
            )
        ],
        [HorizontalSeparator()],
        [
            T(
                SYMBOL_DOWN
                + " If you need more room for details grab the dot and drag to expand",
                background_color="red",
                text_color="white",
            )
        ],
    ]

    bottom_layout = [
        [
            TabGroup(
                [
                    [
                        Tab("Details *\n", frame_details, pad=(0, 0)),
                        Tab(
                            "SHORT Program\nto duplicate problem *",
                            frame_code,
                            pad=(0, 0),
                        ),
                        Tab(
                            "Your Project Details\n(optional)",
                            frame_project_details,
                            pad=(0, 0),
                        ),
                        Tab(
                            "Where you found us?\n(optional)",
                            frame_where_you_found_psg,
                            pad=(0, 0),
                        ),
                        Tab("Markdown Output\n", frame_markdown, pad=(0, 0)),
                    ]
                ],
                k="-TABGROUP-",
                expand_x=True,
                expand_y=True,
            ),
        ]
    ]

    layout_pane = Pane(
        [Col(middle_layout), Col(bottom_layout)],
        key="-PANE-",
        expand_x=True,
        expand_y=True,
    )

    layout = [
        [
            pin(
                B(
                    SYMBOL_DOWN,
                    pad=(0, 0),
                    k="-HIDE CLIST-",
                    tooltip="Hide/show upper sections of window",
                )
            ),
            pin(Col(top_layout, k="-TOP COL-")),
        ],
        [layout_pane],
        [Col([[B("Post Issue"), B("Create Markdown Only"), B("Quit")]])],
    ]

    window = Window(
        "Open A GitHub Issue",
        layout,
        finalize=True,
        resizable=True,
        enable_close_attempted_event=True,
        margins=(0, 0),
    )

    # for i in range(len(checklist)):
    [window["-T{}-".format(i)].set_cursor("hand1") for i in range(len(checklist))]
    # window['-TABGROUP-'].expand(True, True, True)
    # window['-ML CODE-'].expand(True, True, True)
    # window['-ML DETAILS-'].expand(True, True, True)
    # window['-ML MARKDOWN-'].expand(True, True, True)
    # window['-PANE-'].expand(True, True, True)

    if running_mac():
        window["-OS MAC VER-"].update(platform.mac_ver())
    elif running_windows():
        window["-OS WIN VER-"].update(platform.win32_ver())
    elif running_linux():
        window["-OS LINUX VER-"].update(platform.libc_ver())

    window.bring_to_front()
    while True:  # Event Loop
        event, values = window.read()
        # print(event, values)
        if event in (WINDOW_CLOSE_ATTEMPTED_EVENT, "Quit"):
            if (
                    popup_yes_no(
                        "Do you really want to exit?",
                        'If you have not clicked Post Issue button and then clicked "Submit New Issue" button '
                        "then your issue will not have been submitted to GitHub.\n"
                        "If you are having trouble with PySimpleGUI opening your browser, consider generating "
                        "the markdown, copying it to a text file, and then using it later to manually paste into a new issue "
                        "\n"
                        "Are you sure you want to quit?",
                        image=EMOJI_BASE64_PONDER,
                        keep_on_top=True,
                    )
                    == "Yes"
            ):
                break
        if event == WIN_CLOSED:
            break
        if event in ["-T{}-".format(i) for i in range(len(checklist))]:
            webbrowser.open_new_tab(window[event].get())
        if event in issue_types:
            title = str(values["-TITLE-"])
            if len(title) != 0:
                if title[0] == "[" and title.find("]"):
                    title = title[title.find("]") + 1:]
                    title = title.strip()
            window["-TITLE-"].update("[{}] {}".format(event, title))
        if event == "-HIDE CLIST-":
            window["-TOP COL-"].update(visible=not window["-TOP COL-"].visible)
            window["-HIDE CLIST-"].update(
                text=SYMBOL_UP
                if window["-HIDE CLIST-"].get_text() == SYMBOL_DOWN
                else SYMBOL_DOWN
            )
        if event == "Help":
            _github_issue_help()
        elif event in ("Post Issue", "Create Markdown Only"):
            issue_type = None
            for itype in issue_types:
                if values[itype]:
                    issue_type = itype
                    break
            if issue_type is None:
                popup_error("Must choose issue type", keep_on_top=True)
                continue
            if values["-OS WIN-"]:
                operating_system = "Windows"
                os_ver = values["-OS WIN VER-"]
            elif values["-OS LINUX-"]:
                operating_system = "Linux"
                os_ver = values["-OS LINUX VER-"]
            elif values["-OS MAC-"]:
                operating_system = "Mac"
                os_ver = values["-OS MAC VER-"]
            elif values["-OS OTHER-"]:
                operating_system = "Other"
                os_ver = values["-OS OTHER VER-"]
            else:
                popup_error("Must choose Operating System", keep_on_top=True)
                continue
            checkboxes = [
                "X" if values[("-CB-", i)] else " " for i in range(len(checklist))
            ]

            if not _github_issue_post_validate(values, checklist, issue_types):
                continue

            cb_dict = {
                "cb_docs": checkboxes[0],
                "cb_demos": checkboxes[1],
                "cb_demo_port": checkboxes[2],
                "cb_readme_other": checkboxes[3],
                "cb_command_line": checkboxes[4],
                "cb_issues": checkboxes[5],
                "cb_latest_pypi": checkboxes[6],
                "cb_github": checkboxes[7],
                "detailed_desc": values["-ML DETAILS-"],
                "code": values["-ML CODE-"],
                "project_details": values["-ML PROJECT DETAILS-"].rstrip(),
                "where_found": values["-ML FOUND PSG-"],
            }

            markdown = _github_issue_post_make_markdown(
                issue_type,
                operating_system,
                os_ver,
                "tkinter",
                values["-VER PSG-"],
                values["-VER TK-"],
                values["-VER PYTHON-"],
                values["-EXP PYTHON-"],
                values["-EXP PROG-"],
                "Yes" if values["-CB PRIOR GUI-"] else "No",
                values["-EXP NOTES-"],
                **cb_dict,
            )
            window["-ML MARKDOWN-"].update(markdown)
            link = _github_issue_post_make_github_link(
                values["-TITLE-"], window["-ML MARKDOWN-"].get()
            )
            if event == "Post Issue":
                webbrowser.open_new_tab(link)
            else:
                popup("Your markdown code is in the Markdown tab", keep_on_top=True)

    window.close()


def _main_entry_point():
    # print('Restarting main as a new process...(needed in case you want to GitHub Upgrade)')
    # Relaunch using the same python interpreter that was used to run this function
    interpreter = sys.executable
    if "pythonw" in interpreter:
        interpreter = interpreter.replace("pythonw", "python")
    execute_py_file(__file__, interpreter_command=interpreter)


####################################################################################################

# M"""""`'"""`YM          oo
# M  mm.  mm.  M
# M  MMM  MMM  M .d8888b. dP 88d888b.
# M  MMM  MMM  M 88'  `88 88 88'  `88
# M  MMM  MMM  M 88.  .88 88 88    88
# M  MMM  MMM  M `88888P8 dP dP    dP
# MMMMMMMMMMMMMM
#
# MM'"""""`MM            dP
# M' .mmm. `M            88
# M  MMMMMMMM .d8888b. d8888P
# M  MMM   `M 88ooood8   88
# M. `MMM' .M 88.  ...   88
# MM.     .MM `88888P'   dP
# MMMMMMMMMMM
#
# M""""""'YMM          dP
# M  mmmm. `M          88
# M  MMMMM  M .d8888b. 88d888b. dP    dP .d8888b.
# M  MMMMM  M 88ooood8 88'  `88 88    88 88'  `88
# M  MMMM' .M 88.  ... 88.  .88 88.  .88 88.  .88
# M       .MM `88888P' 88Y8888' `88888P' `8888P88
# MMMMMMMMMMM                                 .88
#                                         d8888P
# M""""""'YMM            dP
# M  mmmm. `M            88
# M  MMMMM  M .d8888b. d8888P .d8888b.
# M  MMMMM  M 88'  `88   88   88'  `88
# M  MMMM' .M 88.  .88   88   88.  .88
# M       .MM `88888P8   dP   `88888P8
# MMMMMMMMMMM


def main_get_debug_data(suppress_popup=False):
    """
    Collect up and display the data needed to file GitHub issues.
    This function will place the information on the clipboard.
    You MUST paste the information from the clipboard prior to existing your application (except on Windows).
    :param suppress_popup: If True no popup window will be shown. The string will be only returned, not displayed
    :type suppress_popup:  (bool)
    :returns:              String containing the information to place into the GitHub Issue
    :rtype:                (str)
    """
    message = get_versions()
    clipboard_set(message)

    if not suppress_popup:
        popup_scrolled(
            "*** Version information copied to your clipboard. Paste into your GitHub Issue. ***\n",
            message,
            title="Select and copy this info to your GitHub Issue",
            keep_on_top=True,
            size=(100, 10),
        )

    return message


# ..######...##........#######..########.....###....##.........
# .##....##..##.......##.....##.##.....##...##.##...##.........
# .##........##.......##.....##.##.....##..##...##..##.........
# .##...####.##.......##.....##.########..##.....##.##.........
# .##....##..##.......##.....##.##.....##.#########.##.........
# .##....##..##.......##.....##.##.....##.##.....##.##.........
# ..######...########..#######..########..##.....##.########...
# ..######..########.########.########.########.####.##....##..######....######.
# .##....##.##..........##.......##.......##.....##..###...##.##....##..##....##
# .##.......##..........##.......##.......##.....##..####..##.##........##......
# ..######..######......##.......##.......##.....##..##.##.##.##...####..######.
# .......##.##..........##.......##.......##.....##..##..####.##....##........##
# .##....##.##..........##.......##.......##.....##..##...###.##....##..##....##
# ..######..########....##.......##.......##....####.##....##..######....######.


def _global_settings_get_ttk_scrollbar_info():
    """
    This function reads the ttk scrollbar settings from the global PySimpleGUI settings file.
    Each scrollbar setting is stored with a key that's a TUPLE, not a normal string key.
    The settings are for pieces of the scrollbar and their associated piece of the PySimpleGUI theme.

    The whole ttk scrollbar feature is based on mapping parts of the scrollbar to parts of the PySimpleGUI theme.
    That is what the ttk_part_mapping_dict does, maps between the two lists of items.
    For example, the scrollbar arrow color may map to the theme input text color.

    """
    global ttk_part_mapping_dict, DEFAULT_TTK_THEME
    for ttk_part in TTK_SCROLLBAR_PART_LIST:
        value = pysimplegui_user_settings.get(
            json.dumps(("-ttk scroll-", ttk_part)), ttk_part_mapping_dict[ttk_part]
        )
        ttk_part_mapping_dict[ttk_part] = value

    DEFAULT_TTK_THEME = pysimplegui_user_settings.get("-ttk theme-", DEFAULT_TTK_THEME)


def _global_settings_get_watermark_info():
    if (
            not pysimplegui_user_settings.get("-watermark-", False)
            and not Window._watermark_temp_forced
    ):
        Window._watermark = None
        return
    forced = Window._watermark_temp_forced
    prefix_text = pysimplegui_user_settings.get("-watermark text-", "")

    ver_text = (
        " " + version.split(" ", 1)[0]
        if pysimplegui_user_settings.get(
            "-watermark ver-", False if not forced else True
        )
           or forced
        else ""
    )
    framework_ver_text = (
        " Tk " + framework_version
        if pysimplegui_user_settings.get(
            "-watermark framework ver-", False if not forced else True
        )
           or forced
        else ""
    )
    watermark_font = pysimplegui_user_settings.get("-watermark font-", "_ 9 bold")
    # background_color = pysimplegui_user_settings.get('-watermark bg color-', 'window.BackgroundColor')
    user_text = pysimplegui_user_settings.get("-watermark text-", "")
    python_text = " Py {}.{}.{}".format(
        sys.version_info.major, sys.version_info.minor, sys.version_info.micro
    )
    if user_text:
        text = str(user_text)
    else:
        text = prefix_text + ver_text + python_text + framework_ver_text
    Window._watermark = lambda window: Text(
        text, font=watermark_font, background_color=window.BackgroundColor
    )


def main_global_get_screen_snapshot_symcode():
    pysimplegui_user_settings = UserSettings(
        filename=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME,
        path=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH,
    )

    settings = pysimplegui_user_settings.read()

    screenshot_keysym = ""
    for i in range(4):
        keysym = settings.get(json.dumps(("-snapshot keysym-", i)), "")
        if keysym:
            screenshot_keysym += "<{}>".format(keysym)

    screenshot_keysym_manual = settings.get("-snapshot keysym manual-", "")

    # print('BINDING INFO!', screenshot_keysym, screenshot_keysym_manual)
    if screenshot_keysym_manual:
        return screenshot_keysym_manual
    elif screenshot_keysym:
        return screenshot_keysym
    return ""


def main_global_pysimplegui_settings_erase():
    """
    *** WARNING ***
    Deletes the PySimpleGUI settings file without asking for verification


    """
    print(
        "********** WARNING - you are deleting your PySimpleGUI settings file **********"
    )
    print("The file being deleted is:", pysimplegui_user_settings.full_filename)


def main_global_pysimplegui_settings():
    """
    Window to set settings that will be used across all PySimpleGUI programs that choose to use them.
    Use set_options to set the path to the folder for all PySimpleGUI settings.

    :return: True if settings were changed
    :rtype:  (bool)
    """
    global DEFAULT_WINDOW_SNAPSHOT_KEY_CODE, ttk_part_mapping_dict, DEFAULT_TTK_THEME

    key_choices = tuple(sorted(tkinter_keysyms))

    settings = pysimplegui_user_settings.read()

    editor_format_dict = {
        "pycharm": "<editor> --line <line> <file>",
        "notepad++": "<editor> -n<line> <file>",
        "sublime": "<editor> <file>:<line>",
        "vim": "<editor> +<line> <file>",
        "wing": "<editor> <file>:<line>",
        "visual studio": '<editor> <file> /command "edit.goto <line>"',
        "atom": "<editor> <file>:<line>",
        "spyder": "<editor> <file>",
        "thonny": "<editor> <file>",
        "pydev": "<editor> <file>:<line>",
        "idle": "<editor> <file>",
    }

    tooltip = (
            "Format strings for some popular editors/IDEs:\n"
            + "PyCharm - <editor> --line <line> <file>\n"
            + "Notepad++ - <editor> -n<line> <file>\n"
            + "Sublime - <editor> <file>:<line>\n"
            + "vim -  <editor> +<line> <file>\n"
            + "wing - <editor> <file>:<line>\n"
            + 'Visual Studio - <editor> <file> /command "edit.goto <line>"\n'
            + "Atom - <editor> <file>:<line>\n"
            + "Spyder - <editor> <file>\n"
            + "Thonny - <editor> <file>\n"
            + "PyDev - <editor> <file>:<line>\n"
            + "IDLE - <editor> <file>\n"
    )

    tooltip_file_explorer = (
            'This is the program you normally use to "Browse" for files\n'
            + 'For Windows this is normally "explorer". On Linux "nemo" is sometimes used.'
    )

    tooltip_theme = (
            'The normal default theme for PySimpleGUI is "Dark Blue 13\n'
            + 'If you do not call theme("theme name") by your program to change the theme, then the default is used.\n'
            + "This setting allows you to set the theme that PySimpleGUI will use for ALL of your programs that\n"
            + "do not set a theme specifically."
    )

    # ------------------------- TTK Tab -------------------------
    ttk_scrollbar_tab_layout = [
        [
            T("Default TTK Theme", font="_ 16"),
            Combo(
                [],
                DEFAULT_TTK_THEME,
                readonly=True,
                size=(20, 10),
                key="-TTK THEME-",
                font="_ 16",
            ),
        ],
        [HorizontalSeparator()],
        [T("TTK Scrollbar Settings", font="_ 16")],
    ]

    t_len = max([len(l) for l in TTK_SCROLLBAR_PART_LIST])
    ttk_layout = [[]]
    for key, item in ttk_part_mapping_dict.items():
        if key in TTK_SCROLLBAR_PART_THEME_BASED_LIST:
            ttk_layout += [
                [
                    T(key, s=t_len, justification="r"),
                    Combo(
                        PSG_THEME_PART_LIST,
                        default_value=settings.get(("-ttk scroll-", key), item),
                        key=("-TTK SCROLL-", key),
                    ),
                ]
            ]
        elif key in (TTK_SCROLLBAR_PART_ARROW_WIDTH, TTK_SCROLLBAR_PART_SCROLL_WIDTH):
            ttk_layout += [
                [
                    T(key, s=t_len, justification="r"),
                    Combo(
                        list(range(100)),
                        default_value=settings.get(("-ttk scroll-", key), item),
                        key=("-TTK SCROLL-", key),
                    ),
                ]
            ]
        elif key == TTK_SCROLLBAR_PART_RELIEF:
            ttk_layout += [
                [
                    T(key, s=t_len, justification="r"),
                    Combo(
                        RELIEF_LIST,
                        default_value=settings.get(("-ttk scroll-", key), item),
                        readonly=True,
                        key=("-TTK SCROLL-", key),
                    ),
                ]
            ]

    ttk_scrollbar_tab_layout += ttk_layout
    ttk_scrollbar_tab_layout += [
        [Button("Reset Scrollbar Settings"), Button("Test Scrollbar Settings")]
    ]
    ttk_tab = Tab("TTK", ttk_scrollbar_tab_layout)

    layout = [
        [
            T(
                "Global PySimpleGUI Settings",
                text_color=theme_button_color()[0],
                background_color=theme_button_color()[1],
                font="_ 18",
                expand_x=True,
                justification="c",
            )
        ]
    ]

    # ------------------------- Interpreter Tab -------------------------

    interpreter_tab = Tab(
        "Python Interpreter",
        [
            [T("Normally leave this blank")],
            [
                T("Command to run a python program:"),
                In(
                    settings.get("-python command-", ""),
                    k="-PYTHON COMMAND-",
                    enable_events=True,
                ),
                FileBrowse(),
            ],
        ],
        font="_ 16",
        expand_x=True,
    )

    # ------------------------- Editor Tab -------------------------

    editor_tab = Tab(
        "Editor Settings",
        [
            [
                T("Command to invoke your editor:"),
                In(
                    settings.get("-editor program-", ""),
                    k="-EDITOR PROGRAM-",
                    enable_events=True,
                ),
                FileBrowse(),
            ],
            [T("String to launch your editor to edit at a particular line #.")],
            [T("Use tags <editor> <file> <line> to specify the string")],
            [T("that will be executed to edit python files using your editor")],
            [
                T("Edit Format String (hover for tooltip)", tooltip=tooltip),
                In(
                    settings.get("-editor format string-", "<editor> <file>"),
                    k="-EDITOR FORMAT-",
                    tooltip=tooltip,
                ),
            ],
        ],
        font="_ 16",
        expand_x=True,
    )

    # ------------------------- Explorer Tab -------------------------

    explorer_tab = Tab(
        "Explorer Program",
        [
            [
                In(
                    settings.get("-explorer program-", ""),
                    k="-EXPLORER PROGRAM-",
                    tooltip=tooltip_file_explorer,
                )
            ]
        ],
        font="_ 16",
        expand_x=True,
        tooltip=tooltip_file_explorer,
    )

    # ------------------------- Snapshots Tab -------------------------

    snapshots_tab = Tab(
        "Window Snapshots",
        [
            [
                Combo(
                    ("",) + key_choices,
                    default_value=settings.get(
                        json.dumps(("-snapshot keysym-", i)), ""
                    ),
                    readonly=True,
                    k=("-SNAPSHOT KEYSYM-", i),
                    s=(None, 30),
                )
                for i in range(4)
            ],
            [
                T("Manually Entered Bind String:"),
                Input(
                    settings.get("-snapshot keysym manual-", ""),
                    k="-SNAPSHOT KEYSYM MANUAL-",
                ),
            ],
            [
                T("Folder to store screenshots:"),
                Push(),
                In(settings.get("-screenshots folder-", ""), k="-SCREENSHOTS FOLDER-"),
                FolderBrowse(),
            ],
            [
                T("Screenshots Filename or Prefix:"),
                Push(),
                In(
                    settings.get("-screenshots filename-", ""),
                    k="-SCREENSHOTS FILENAME-",
                ),
                FileBrowse(),
            ],
            [Checkbox("Auto-number Images", k="-SCREENSHOTS AUTONUMBER-")],
        ],
        font="_ 16",
        expand_x=True,
    )

    # ------------------------- Theme Tab -------------------------

    theme_tab = Tab(
        "Theme",
        [
            [
                T(
                    'Leave blank for "official" PySimpleGUI default theme: {}'.format(
                        OFFICIAL_PYSIMPLEGUI_THEME
                    )
                )
            ],
            [
                T("Default Theme For All Programs:"),
                Combo(
                    [""] + theme_list(),
                    settings.get("-theme-", None),
                    readonly=True,
                    k="-THEME-",
                    tooltip=tooltip_theme,
                ),
                Checkbox(
                    "Always use custom Titlebar",
                    default=pysimplegui_user_settings.get("-custom titlebar-", False),
                    k="-CUSTOM TITLEBAR-",
                ),
            ],
            [
                Frame(
                    "Window Watermarking",
                    [
                        [
                            Checkbox(
                                "Enable Window Watermarking",
                                pysimplegui_user_settings.get("-watermark-", False),
                                k="-WATERMARK-",
                            )
                        ],
                        [
                            T("Prefix Text String:"),
                            Input(
                                pysimplegui_user_settings.get("-watermark text-", ""),
                                k="-WATERMARK TEXT-",
                            ),
                        ],
                        [
                            Checkbox(
                                "PySimpleGUI Version",
                                pysimplegui_user_settings.get("-watermark ver-", False),
                                k="-WATERMARK VER-",
                            )
                        ],
                        [
                            Checkbox(
                                "Framework Version",
                                pysimplegui_user_settings.get(
                                    "-watermark framework ver-", False
                                ),
                                k="-WATERMARK FRAMEWORK VER-",
                            )
                        ],
                        [
                            T("Font:"),
                            Input(
                                pysimplegui_user_settings.get(
                                    "-watermark font-", "_ 9 bold"
                                ),
                                k="-WATERMARK FONT-",
                            ),
                        ],
                        # [T('Background Color:'), Input(pysimplegui_user_settings.get('-watermark bg color-', 'window.BackgroundColor'), k='-WATERMARK BG COLOR-')],
                    ],
                    font="_ 16",
                    expand_x=True,
                )
            ],
        ],
    )

    settings_tab_group = TabGroup(
        [
            [
                theme_tab,
                ttk_tab,
                interpreter_tab,
                explorer_tab,
                editor_tab,
                snapshots_tab,
            ]
        ]
    )
    layout += [[settings_tab_group]]
    # [T('Buttons (Leave Unchecked To Use Default) NOT YET IMPLEMENTED!',  font='_ 16')],
    #      [Checkbox('Always use TTK buttons'), CBox('Always use TK Buttons')],
    layout += [[B("Ok", bind_return_key=True), B("Cancel"), B("Mac Patch Control")]]

    window = Window("Settings", layout, keep_on_top=True, modal=False, finalize=True)

    # fill in the theme list into the Combo element - must do this AFTER the window is created or a tkinter temp window is auto created by tkinter
    ttk_theme_list = ttk.Style().theme_names()

    window["-TTK THEME-"].update(value=DEFAULT_TTK_THEME, values=ttk_theme_list)

    while True:
        event, values = window.read()
        if event in ("Cancel", WIN_CLOSED):
            break
        if event == "Ok":
            new_theme = (
                OFFICIAL_PYSIMPLEGUI_THEME
                if values["-THEME-"] == ""
                else values["-THEME-"]
            )
            pysimplegui_user_settings.set(
                "-editor program-", values["-EDITOR PROGRAM-"]
            )
            pysimplegui_user_settings.set(
                "-explorer program-", values["-EXPLORER PROGRAM-"]
            )
            pysimplegui_user_settings.set(
                "-editor format string-", values["-EDITOR FORMAT-"]
            )
            pysimplegui_user_settings.set(
                "-python command-", values["-PYTHON COMMAND-"]
            )
            pysimplegui_user_settings.set(
                "-custom titlebar-", values["-CUSTOM TITLEBAR-"]
            )
            pysimplegui_user_settings.set("-theme-", new_theme)
            pysimplegui_user_settings.set("-watermark-", values["-WATERMARK-"])
            pysimplegui_user_settings.set(
                "-watermark text-", values["-WATERMARK TEXT-"]
            )
            pysimplegui_user_settings.set("-watermark ver-", values["-WATERMARK VER-"])
            pysimplegui_user_settings.set(
                "-watermark framework ver-", values["-WATERMARK FRAMEWORK VER-"]
            )
            pysimplegui_user_settings.set(
                "-watermark font-", values["-WATERMARK FONT-"]
            )
            # pysimplegui_user_settings.set('-watermark bg color-', values['-WATERMARK BG COLOR-'])

            # TTK SETTINGS
            pysimplegui_user_settings.set("-ttk theme-", values["-TTK THEME-"])
            DEFAULT_TTK_THEME = values["-TTK THEME-"]

            # Snapshots portion
            screenshot_keysym_manual = values["-SNAPSHOT KEYSYM MANUAL-"]
            pysimplegui_user_settings.set(
                "-snapshot keysym manual-", values["-SNAPSHOT KEYSYM MANUAL-"]
            )
            screenshot_keysym = ""
            for i in range(4):
                pysimplegui_user_settings.set(
                    json.dumps(("-snapshot keysym-", i)),
                    values[("-SNAPSHOT KEYSYM-", i)],
                )
                if values[("-SNAPSHOT KEYSYM-", i)]:
                    screenshot_keysym += "<{}>".format(values[("-SNAPSHOT KEYSYM-", i)])
            if screenshot_keysym_manual:
                DEFAULT_WINDOW_SNAPSHOT_KEY_CODE = screenshot_keysym_manual
            elif screenshot_keysym:
                DEFAULT_WINDOW_SNAPSHOT_KEY_CODE = screenshot_keysym

            pysimplegui_user_settings.set(
                "-screenshots folder-", values["-SCREENSHOTS FOLDER-"]
            )
            pysimplegui_user_settings.set(
                "-screenshots filename-", values["-SCREENSHOTS FILENAME-"]
            )

            # TTK Scrollbar portion
            for key, value in values.items():
                if isinstance(key, tuple):
                    if key[0] == "-TTK SCROLL-":
                        pysimplegui_user_settings.set(
                            json.dumps(("-ttk scroll-", key[1])), value
                        )

            # Upgrade Service Settings
            pysimplegui_user_settings.set(
                "-upgrade show only critical-", values["-UPGRADE SHOW ONLY CRITICAL-"]
            )

            theme(new_theme)

            _global_settings_get_ttk_scrollbar_info()
            _global_settings_get_watermark_info()

            window.close()
            return True
        elif event == "-EDITOR PROGRAM-":
            for key in editor_format_dict.keys():
                if key in values["-EDITOR PROGRAM-"].lower():
                    window["-EDITOR FORMAT-"].update(value=editor_format_dict[key])
        elif event == "Mac Patch Control":
            main_mac_feature_control()
            # re-read the settings in case they changed
            _read_mac_global_settings()
        elif event == "Reset Scrollbar Settings":
            ttk_part_mapping_dict = copy.copy(DEFAULT_TTK_PART_MAPPING_DICT)
            for key, item in ttk_part_mapping_dict.items():
                window[("-TTK SCROLL-", key)].update(item)
        elif event == "Test Scrollbar Settings":
            for ttk_part in TTK_SCROLLBAR_PART_LIST:
                value = values[("-TTK SCROLL-", ttk_part)]
                ttk_part_mapping_dict[ttk_part] = value
            DEFAULT_TTK_THEME = values["-TTK THEME-"]
            for i in range(100):
                Print(i, keep_on_top=True)
            Print("Close this window to continue...", keep_on_top=True)

    window.close()
    # In case some of the settings were modified and tried out, reset the ttk info to be what's in the config file
    style = ttk.Style(Window.hidden_master_root)
    _change_ttk_theme(style, DEFAULT_TTK_THEME)
    _global_settings_get_ttk_scrollbar_info()

    return False


# ..######..########..##....##....##.....##.########.##.......########.
# .##....##.##.....##.##...##.....##.....##.##.......##.......##.....##
# .##.......##.....##.##..##......##.....##.##.......##.......##.....##
# ..######..##.....##.#####.......#########.######...##.......########.
# .......##.##.....##.##..##......##.....##.##.......##.......##.......
# .##....##.##.....##.##...##.....##.....##.##.......##.......##.......
# ..######..########..##....##....##.....##.########.########.##.......


def main_sdk_help():
    """
    Display a window that will display the docstrings for each PySimpleGUI Element and the Window object

    """
    online_help_links = {
        "Button": r"https://PySimpleGUI.org/en/latest/call%20reference/#button-element",
        "ButtonMenu": r"https://PySimpleGUI.org/en/latest/call%20reference/#buttonmenu-element",
        "Canvas": r"https://PySimpleGUI.org/en/latest/call%20reference/#canvas-element",
        "Checkbox": r"https://PySimpleGUI.org/en/latest/call%20reference/#checkbox-element",
        "Column": r"https://PySimpleGUI.org/en/latest/call%20reference/#column-element",
        "Combo": r"https://PySimpleGUI.org/en/latest/call%20reference/#combo-element",
        "Frame": r"https://PySimpleGUI.org/en/latest/call%20reference/#frame-element",
        "Graph": r"https://PySimpleGUI.org/en/latest/call%20reference/#graph-element",
        "HorizontalSeparator": r"https://PySimpleGUI.org/en/latest/call%20reference/#horizontalseparator-element",
        "Image": r"https://PySimpleGUI.org/en/latest/call%20reference/#image-element",
        "Input": r"https://PySimpleGUI.org/en/latest/call%20reference/#input-element",
        "Listbox": r"https://PySimpleGUI.org/en/latest/call%20reference/#listbox-element",
        "Menu": r"https://PySimpleGUI.org/en/latest/call%20reference/#menu-element",
        "MenubarCustom": r"https://PySimpleGUI.org/en/latest/call%20reference/#menubarcustom-element",
        "Multiline": r"https://PySimpleGUI.org/en/latest/call%20reference/#multiline-element",
        "OptionMenu": r"https://PySimpleGUI.org/en/latest/call%20reference/#optionmenu-element",
        "Output": r"https://PySimpleGUI.org/en/latest/call%20reference/#output-element",
        "Pane": r"https://PySimpleGUI.org/en/latest/call%20reference/#pane-element",
        "ProgressBar": r"https://PySimpleGUI.org/en/latest/call%20reference/#progressbar-element",
        "Radio": r"https://PySimpleGUI.org/en/latest/call%20reference/#radio-element",
        "Slider": r"https://PySimpleGUI.org/en/latest/call%20reference/#slider-element",
        "Spin": r"https://PySimpleGUI.org/en/latest/call%20reference/#spin-element",
        "StatusBar": r"https://PySimpleGUI.org/en/latest/call%20reference/#statusbar-element",
        "Tab": r"https://PySimpleGUI.org/en/latest/call%20reference/#tab-element",
        "TabGroup": r"https://PySimpleGUI.org/en/latest/call%20reference/#tabgroup-element",
        "Table": r"https://PySimpleGUI.org/en/latest/call%20reference/#table-element",
        "Text": r"https://PySimpleGUI.org/en/latest/call%20reference/#text-element",
        "Titlebar": r"https://PySimpleGUI.org/en/latest/call%20reference/#titlebar-element",
        "Tree": r"https://PySimpleGUI.org/en/latest/call%20reference/#tree-element",
        "VerticalSeparator": r"https://PySimpleGUI.org/en/latest/call%20reference/#verticalseparator-element",
        "Window": r"https://PySimpleGUI.org/en/latest/call%20reference/#window",
    }

    NOT_AN_ELEMENT = "Not An Element"
    element_classes = Element.__subclasses__()
    element_names = {element.__name__: element for element in element_classes}
    element_names["Window"] = Window
    element_classes.append(Window)
    element_arg_default_dict, element_arg_default_dict_update = {}, {}
    vars3 = [m for m in inspect.getmembers(sys.modules[__name__])]

    functions = [
        m for m in inspect.getmembers(sys.modules[__name__], inspect.isfunction)
    ]
    functions_names_lower = [f for f in functions if f[0][0].islower()]
    functions_names_upper = [f for f in functions if f[0][0].isupper()]
    functions_names = sorted(functions_names_lower) + sorted(functions_names_upper)

    for element in element_classes:
        # Build info about init method
        args = inspect.getfullargspec(element.__init__).args[1:]
        defaults = inspect.getfullargspec(element.__init__).defaults
        # print('------------- {element}----------')
        # print(args)
        # print(defaults)
        if len(args) != len(defaults):
            diff = len(args) - len(defaults)
            defaults = ("NO DEFAULT",) * diff + defaults
        args_defaults = []
        for i, a in enumerate(args):
            args_defaults.append((a, defaults[i]))
        element_arg_default_dict[element.__name__] = args_defaults

        # Build info about update method
        try:
            args = inspect.getfullargspec(element.update).args[1:]
            defaults = inspect.getfullargspec(element.update).defaults
            if args is None or defaults is None:
                element_arg_default_dict_update[element.__name__] = (("", ""),)
                continue
            if len(args) != len(defaults):
                diff = len(args) - len(defaults)
                defaults = ("NO DEFAULT",) * diff + defaults
            args_defaults = []
            for i, a in enumerate(args):
                args_defaults.append((a, defaults[i]))
            element_arg_default_dict_update[element.__name__] = (
                args_defaults if len(args_defaults) else (("", ""),)
            )
        except Exception as e:
            pass

    # Add on the pseudo-elements
    element_names["MenubarCustom"] = MenubarCustom
    element_names["Titlebar"] = Titlebar

    buttons = [
        [B(e, pad=(0, 0), size=(22, 1), font="Courier 10")]
        for e in sorted(element_names.keys())
    ]
    buttons += [[B("Func Search", pad=(0, 0), size=(22, 1), font="Courier 10")]]
    button_col = Col(buttons, vertical_alignment="t")
    mline_col = Column(
        [
            [
                Multiline(
                    size=(100, 46),
                    key="-ML-",
                    write_only=True,
                    reroute_stdout=True,
                    font="Courier 10",
                    expand_x=True,
                    expand_y=True,
                )
            ],
            [
                T(
                    size=(80, 1),
                    font="Courier 10 underline",
                    k="-DOC LINK-",
                    enable_events=True,
                )
            ],
        ],
        pad=(0, 0),
        expand_x=True,
        expand_y=True,
        vertical_alignment="t",
    )
    layout = [[button_col, mline_col]]
    layout += [
        [
            CBox("Summary Only", enable_events=True, k="-SUMMARY-"),
            CBox("Display Only PEP8 Functions", default=True, k="-PEP8-"),
        ]
    ]
    # layout = [[Column(layout, scrollable=True, p=0, expand_x=True, expand_y=True, vertical_alignment='t'), Sizegrip()]]
    layout += [[Button("Exit", size=(15, 1)), Sizegrip()]]

    window = Window(
        "SDK API Call Reference",
        layout,
        resizable=True,
        use_default_focus=False,
        keep_on_top=True,
        icon=EMOJI_BASE64_THINK,
        finalize=True,
        right_click_menu=MENU_RIGHT_CLICK_EDITME_EXIT,
    )
    window["-DOC LINK-"].set_cursor("hand1")
    online_help_link = ""
    ml = window["-ML-"]
    current_element = ""
    try:
        while True:  # Event Loop
            event, values = window.read()
            if event in (WIN_CLOSED, "Exit"):
                break
            if event == "-DOC LINK-":
                if webbrowser_available and online_help_link:
                    webbrowser.open_new_tab(online_help_link)
            if event == "-SUMMARY-":
                event = current_element

            if event in element_names.keys():
                current_element = event
                window["-ML-"].update("")
                online_help_link = online_help_links.get(event, "")
                window["-DOC LINK-"].update(online_help_link)
                if not values["-SUMMARY-"]:
                    elem = element_names[event]
                    ml.print(pydoc.help(elem))
                    # print the aliases for the class
                    ml.print("\n--- Shortcut Aliases for Class ---")
                    for v in vars3:
                        if elem == v[1] and elem.__name__ != v[0]:
                            print(v[0])
                    ml.print("\n--- Init Parms ---")
                else:
                    elem = element_names[event]
                    if inspect.isfunction(elem):
                        ml.print(
                            "Not a class...It is a function",
                            background_color="red",
                            text_color="white",
                        )
                    else:
                        element_methods = [
                            m[0]
                            for m in inspect.getmembers(Element, inspect.isfunction)
                            if not m[0].startswith("_") and not m[0][0].isupper()
                        ]
                        methods = inspect.getmembers(elem, inspect.isfunction)
                        methods = [
                            m[0]
                            for m in methods
                            if not m[0].startswith("_") and not m[0][0].isupper()
                        ]

                        unique_methods = [
                            m
                            for m in methods
                            if m not in element_methods and not m[0][0].isupper()
                        ]

                        properties = inspect.getmembers(
                            elem, lambda o: isinstance(o, property)
                        )
                        properties = [
                            p[0] for p in properties if not p[0].startswith("_")
                        ]
                        ml.print(
                            "--- Methods ---",
                            background_color="red",
                            text_color="white",
                        )
                        ml.print("\n".join(methods))
                        ml.print(
                            "--- Properties ---",
                            background_color="red",
                            text_color="white",
                        )
                        ml.print("\n".join(properties))
                        if elem != NOT_AN_ELEMENT:
                            if issubclass(elem, Element):
                                ml.print(
                                    "Methods Unique to This Element",
                                    background_color="red",
                                    text_color="white",
                                )
                                ml.print("\n".join(unique_methods))
                        ml.print(
                            "========== Init Parms ==========",
                            background_color="#FFFF00",
                            text_color="black",
                        )
                        elem_text_name = event
                        for parm, default in element_arg_default_dict[elem_text_name]:
                            ml.print("{:18}".format(parm), end=" = ")
                            ml.print(default, end=",\n")
                        if elem_text_name in element_arg_default_dict_update:
                            ml.print(
                                "========== Update Parms ==========",
                                background_color="#FFFF00",
                                text_color="black",
                            )
                            for parm, default in element_arg_default_dict_update[
                                elem_text_name
                            ]:
                                ml.print("{:18}".format(parm), end=" = ")
                                ml.print(default, end=",\n")
                ml.set_vscroll_position(0)  # scroll to top of multoline
            elif event == "Func Search":
                search_string = popup_get_text(
                    "Search for this in function list:", keep_on_top=True
                )
                if search_string is not None:
                    online_help_link = ""
                    window["-DOC LINK-"].update("")
                    ml.update("")
                    for f_entry in functions_names:
                        f = f_entry[0]
                        if search_string in f.lower() and not f.startswith("_"):
                            if (values["-PEP8-"] and not f[0].isupper()) or not values[
                                "-PEP8-"
                            ]:
                                if values["-SUMMARY-"]:
                                    ml.print(f)
                                else:
                                    ml.print(
                                        "=========== " + f + "===========",
                                        background_color="#FFFF00",
                                        text_color="black",
                                    )
                                    ml.print(pydoc.help(f_entry[1]))
                ml.set_vscroll_position(0)  # scroll to top of multoline
    except Exception as e:
        _error_popup_with_traceback("Exception in SDK reference", e)
    window.close()


#                     oo
#
# 88d8b.d8b. .d8888b. dP 88d888b.
# 88'`88'`88 88'  `88 88 88'  `88
# 88  88  88 88.  .88 88 88    88
# dP  dP  dP `88888P8 dP dP    dP
#
#
# M""MMM""MMM""M oo                dP
# M  MMM  MMM  M                   88
# M  MMP  MMP  M dP 88d888b. .d888b88 .d8888b. dP  dP  dP
# M  MM'  MM' .M 88 88'  `88 88'  `88 88'  `88 88  88  88
# M  `' . '' .MM 88 88    88 88.  .88 88.  .88 88.88b.88'
# M    .d  .dMMM dP dP    dP `88888P8 `88888P' 8888P Y8P
# MMMMMMMMMMMMMM
#
# MP""""""`MM   dP                       dP               dP
# M  mmmmm..M   88                       88               88
# M.      `YM d8888P .d8888b. 88d888b. d8888P .d8888b.    88d888b. .d8888b. 88d888b. .d8888b.
# MMMMMMM.  M   88   88'  `88 88'  `88   88   Y8ooooo.    88'  `88 88ooood8 88'  `88 88ooood8
# M. .MMM'  M   88   88.  .88 88         88         88    88    88 88.  ... 88       88.  ...
# Mb.     .dM   dP   `88888P8 dP         dP   `88888P'    dP    dP `88888P' dP       `88888P'
# MMMMMMMMMMM


def _main_switch_theme():
    layout = [
        [Text("Click a look and feel color to see demo window")],
        [Listbox(values=theme_list(), size=(20, 20), key="-LIST-")],
        [Button("Choose"), Button("Cancel")],
    ]

    window = Window("Change Themes", layout)

    event, values = window.read(close=True)

    if event == "Choose":
        theme_name = values["-LIST-"][0]
        theme(theme_name)


def _create_main_window():
    """
    Creates the main test harness window.

    :return: The test window
    :rtype:  Window
    """

    # theme('dark blue 3')
    # theme('dark brown 2')
    # theme('dark')
    # theme('dark red')
    # theme('Light Green 6')
    # theme('Dark Grey 8')

    tkversion = tkinter.TkVersion
    tclversion = tkinter.TclVersion
    tclversion_detailed = tkinter.Tcl().eval("info patchlevel")

    print("Starting up PySimpleGUI Diagnostic & Help System")
    print("PySimpleGUI long version = ", version)
    print(
        "PySimpleGUI Version ",
        ver,
        "\ntcl ver = {}".format(tclversion),
        "tkinter version = {}".format(tkversion),
        "\nPython Version {}".format(sys.version),
    )
    print("tcl detailed version = {}".format(tclversion_detailed))
    print("PySimpleGUI.py location", __file__)
    # ------ Menu Definition ------ #
    menu_def = [
        ["&File", ["!&Open", "&Save::savekey", "---", "&Properties", "E&xit"]],
        [
            "&Edit",
            ["&Paste", ["Special", "Normal", "!Disabled"], "Undo"],
        ],
        ["&Debugger", ["Popout", "Launch Debugger"]],
        ["!&Disabled", ["Popout", "Launch Debugger"]],
        ["&Toolbar", ["Command &1", "Command &2", "Command &3", "Command &4"]],
        ["&Help", "&About..."],
    ]

    button_menu_def = [
        "unused",
        ["&Paste", ["Special", "Normal", "!Disabled"], "Undo", "Exit"],
    ]
    treedata = TreeData()

    treedata.Insert(
        "",
        "_A_",
        "Tree Item 1",
        [1, 2, 3],
    )
    treedata.Insert(
        "",
        "_B_",
        "B",
        [4, 5, 6],
    )
    treedata.Insert(
        "_A_",
        "_A1_",
        "Sub Item 1",
        ["can", "be", "anything"],
    )
    treedata.Insert(
        "",
        "_C_",
        "C",
        [],
    )
    treedata.Insert(
        "_C_",
        "_C1_",
        "C1",
        ["or"],
    )
    treedata.Insert("_A_", "_A2_", "Sub Item 2", [None, None])
    treedata.Insert("_A1_", "_A3_", "A30", ["getting deep"])
    treedata.Insert("_C_", "_C2_", "C2", ["nothing", "at", "all"])

    for i in range(100):
        treedata.Insert("_C_", i, i, [])

    frame1 = [
        [
            Input("Input Text", size=(25, 1)),
        ],
        [Multiline(size=(30, 5), default_text="Multiline Input")],
    ]

    frame2 = [
        # [ProgressBar(100, bar_color=('red', 'green'), orientation='h')],
        [
            Listbox(
                ["Listbox 1", "Listbox 2", "Listbox 3"],
                select_mode=SELECT_MODE_EXTENDED,
                size=(20, 5),
                no_scrollbar=True,
            ),
            Spin([1, 2, 3, "a", "b", "c"], initial_value="a", size=(4, 3), wrap=True),
        ],
        [
            Combo(
                ["Combo item %s" % i for i in range(5)],
                size=(20, 3),
                default_value="Combo item 2",
                key="-COMBO1-",
            )
        ],
        [
            Combo(
                ["Combo item %s" % i for i in range(5)],
                size=(20, 3),
                font="Courier 14",
                default_value="Combo item 2",
                key="-COMBO2-",
            )
        ],
        # [Combo(['Combo item 1', 2,3,4], size=(20, 3), readonly=False, text_color='blue', background_color='red', key='-COMBO2-')],
    ]

    frame3 = [
        [Checkbox("Checkbox1", True, k="-CB1-"), Checkbox("Checkbox2", k="-CB2-")],
        [
            Radio("Radio Button1", 1, key="-R1-"),
            Radio("Radio Button2", 1, default=True, key="-R2-", tooltip="Radio 2"),
        ],
        [T("", size=(1, 4))],
    ]

    frame4 = [
        [
            Slider(
                range=(0, 100),
                orientation="v",
                size=(7, 15),
                default_value=40,
                key="-SLIDER1-",
            ),
            Slider(
                range=(0, 100),
                orientation="h",
                size=(11, 15),
                default_value=40,
                key="-SLIDER2-",
            ),
        ],
    ]
    matrix = [[str(x * y) for x in range(1, 5)] for y in range(1, 8)]

    frame5 = [
        vtop(
            [
                Table(
                    values=matrix,
                    headings=matrix[0],
                    auto_size_columns=False,
                    display_row_numbers=True,
                    change_submits=False,
                    justification="right",
                    header_border_width=4,
                    # header_relief=RELIEF_GROOVE,
                    num_rows=10,
                    alternating_row_color="lightblue",
                    key="-TABLE-",
                    col_widths=[5, 5, 5, 5],
                ),
                Tree(
                    data=treedata,
                    headings=["col1", "col2", "col3"],
                    col_widths=[5, 5, 5, 5],
                    change_submits=True,
                    auto_size_columns=False,
                    header_border_width=4,
                    # header_relief=RELIEF_GROOVE,
                    num_rows=8,
                    col0_width=8,
                    key="-TREE-",
                    show_expanded=True,
                ),
            ]
        )
    ]
    frame7 = [
        [
            Image(EMOJI_BASE64_HAPPY_HEARTS, enable_events=True, k="-EMOJI-HEARTS-"),
            T("Do you"),
            Image(HEART_3D_BASE64, subsample=3, enable_events=True, k="-HEART-"),
            T("so far?"),
        ],
        [
            T(
                'Want to be taught PySimpleGUI?\nThen maybe the "Official PySimpleGUI Course" on Udemy is for you.'
            )
        ],
        [
            B(image_data=UDEMY_ICON, enable_events=True, k="-UDEMY-"),
            T("Check docs, announcements, easter eggs on this page for coupons."),
        ],
    ]

    pop_test_tab_layout = [
        [Image(EMOJI_BASE64_HAPPY_IDEA), T("Popup tests? Good idea!")],
        [
            B("Popup", k="P "),
            B("No Titlebar", k="P NoTitle"),
            B("Not Modal", k="P NoModal"),
            B("Non Blocking", k="P NoBlock"),
            B("Auto Close", k="P AutoClose"),
        ],
        [T('"Get" popups too!')],
        [B("Get File"), B("Get Folder"), B("Get Date"), B("Get Text")],
    ]

    GRAPH_SIZE = (500, 200)
    graph_elem = Graph(GRAPH_SIZE, (0, 0), GRAPH_SIZE, key="+GRAPH+")

    frame6 = [[VPush()], [graph_elem]]

    themes_tab_layout = [
        [
            T(
                "You can see a preview of the themes, the color swatches, or switch themes for this window"
            )
        ],
        [
            T(
                "If you want to change the default theme for PySimpleGUI, use the Global Settings"
            )
        ],
        [B("Themes"), B("Theme Swatches"), B("Switch Themes")],
    ]

    upgrade_recommendation_tab_layout = [
        [T("Latest Recommendation and Announcements For You", font="_ 14")],
        [
            T("Severity Level of Update:"),
            T(pysimplegui_user_settings.get("-severity level-", "")),
        ],
        [
            T("Recommended Version To Upgrade To:"),
            T(pysimplegui_user_settings.get("-upgrade recommendation-", "")),
        ],
        [T(pysimplegui_user_settings.get("-upgrade message 1-", ""))],
        [T(pysimplegui_user_settings.get("-upgrade message 2-", ""))],
        [
            Checkbox(
                "Show Only Critical Messages",
                default=pysimplegui_user_settings.get(
                    "-upgrade show only critical-", False
                ),
                key="-UPGRADE SHOW ONLY CRITICAL-",
                enable_events=True,
            )
        ],
        [
            Button("Show Notification Again"),
        ],
    ]
    tab_upgrade = Tab("Upgrade\n", upgrade_recommendation_tab_layout, expand_x=True)

    tab1 = Tab("Graph\n", frame6, tooltip="Graph is in here", title_color="red")
    tab2 = Tab(
        "CB, Radio\nList, Combo",
        [
            [
                Frame(
                    "Multiple Choice Group",
                    frame2,
                    title_color="#FFFFFF",
                    tooltip="Checkboxes, radio buttons, etc",
                    vertical_alignment="t",
                ),
                Frame(
                    "Binary Choice Group",
                    frame3,
                    title_color="#FFFFFF",
                    tooltip="Binary Choice",
                    vertical_alignment="t",
                ),
            ]
        ],
    )
    # tab3 = Tab('Table and Tree', [[Frame('Structured Data Group', frame5, title_color='red', element_justification='l')]], tooltip='tab 3', title_color='red', )
    tab3 = Tab(
        "Table &\nTree",
        [[Column(frame5, element_justification="l", vertical_alignment="t")]],
        tooltip="tab 3",
        title_color="red",
        k="-TAB TABLE-",
    )
    tab4 = Tab(
        "Sliders\n",
        [[Frame("Variable Choice Group", frame4, title_color="blue")]],
        tooltip="tab 4",
        title_color="red",
        k="-TAB VAR-",
    )
    tab5 = Tab(
        "Input\nMultiline",
        [[Frame("TextInput", frame1, title_color="blue")]],
        tooltip="tab 5",
        title_color="red",
        k="-TAB TEXT-",
    )
    tab6 = Tab("Course or\nSponsor", frame7, k="-TAB SPONSOR-")
    tab7 = Tab("Popups\n", pop_test_tab_layout, k="-TAB POPUP-")
    tab8 = Tab("Themes\n", themes_tab_layout, k="-TAB THEMES-")

    def VerLine(version, description, justification="r", size=(40, 1)):
        return [
            T(
                version,
                justification=justification,
                font="Any 12",
                text_color="yellow",
                size=size,
                pad=(0, 0),
            ),
            T(description, font="Any 12", pad=(0, 0)),
        ]

    layout_top = Column(
        [
            [
                Image(
                    EMOJI_BASE64_HAPPY_BIG_SMILE,
                    enable_events=True,
                    key="-LOGO-",
                    tooltip="This is PySimpleGUI logo",
                ),
                Image(
                    data=DEFAULT_BASE64_LOADING_GIF, enable_events=True, key="-IMAGE-"
                ),
                Text(
                    "PySimpleGUI Test Harness",
                    font="ANY 14",
                    tooltip="My tooltip",
                    key="-TEXT1-",
                ),
            ],
            VerLine(ver, "PySimpleGUI Version") + [Image(HEART_3D_BASE64, subsample=4)],
            # VerLine('{}/{}'.format(tkversion, tclversion), 'TK/TCL Versions'),
            VerLine(tclversion_detailed, "detailed tkinter version"),
            VerLine(
                os.path.dirname(os.path.abspath(__file__)),
                "PySimpleGUI Location",
                size=(40, None),
            ),
            VerLine(sys.executable, "Python Executable"),
            VerLine(sys.version, "Python Version", size=(40, 2))
            + [
                Image(
                    PYTHON_COLORED_HEARTS_BASE64,
                    subsample=3,
                    k="-PYTHON HEARTS-",
                    enable_events=True,
                )
            ],
        ],
        pad=0,
    )

    layout_bottom = [
        [
            B(SYMBOL_DOWN, pad=(0, 0), k="-HIDE TABS-"),
            pin(
                Col(
                    [
                        [
                            TabGroup(
                                [
                                    [
                                        tab1,
                                        tab2,
                                        tab3,
                                        tab6,
                                        tab4,
                                        tab5,
                                        tab7,
                                        tab8,
                                        tab_upgrade,
                                    ]
                                ],
                                key="-TAB_GROUP-",
                            )
                        ]
                    ],
                    k="-TAB GROUP COL-",
                )
            ),
        ],
        [
            B("Button", highlight_colors=("yellow", "red"), pad=(1, 0)),
            B(
                "ttk Button",
                use_ttk_buttons=True,
                tooltip="This is a TTK Button",
                pad=(1, 0),
            ),
            B(
                "See-through Mode",
                tooltip="Make the background transparent",
                pad=(1, 0),
            ),
            # B('Upgrade PySimpleGUI from GitHub', button_color='white on red', key='-INSTALL-',pad=(1, 0)),
            B(
                "Global Settings",
                tooltip="Settings across all PySimpleGUI programs",
                pad=(1, 0),
            ),
            B("Exit", tooltip="Exit button", pad=(1, 0)),
        ],
        # [B(image_data=ICON_BUY_ME_A_COFFEE,pad=(1, 0), key='-COFFEE-'),
        [
            B(image_data=UDEMY_ICON, pad=(1, 0), key="-UDEMY-"),
            B("SDK Reference", pad=(1, 0)),
            B("Open GitHub Issue", pad=(1, 0)),
            B("Versions for GitHub", pad=(1, 0)),
            ButtonMenu(
                "ButtonMenu",
                button_menu_def,
                pad=(1, 0),
                key="-BMENU-",
                tearoff=True,
                disabled_text_color="yellow",
            ),
        ],
    ]

    layout = [[]]

    if not theme_use_custom_titlebar():
        layout += [
            [
                Menu(
                    menu_def,
                    key="-MENU-",
                    font="Courier 15",
                    background_color="red",
                    text_color="white",
                    disabled_text_color="yellow",
                    tearoff=True,
                )
            ]
        ]
    else:
        layout += [
            [
                MenubarCustom(
                    menu_def,
                    key="-MENU-",
                    font="Courier 15",
                    bar_background_color=theme_background_color(),
                    bar_text_color=theme_text_color(),
                    background_color="red",
                    text_color="white",
                    disabled_text_color="yellow",
                )
            ]
        ]

    layout += [
        [layout_top]
        + [ProgressBar(max_value=800, size=(20, 25), orientation="v", key="+PROGRESS+")]
    ]
    layout += layout_bottom

    window = Window(
        "PySimpleGUI Main Test Harness",
        layout,
        # font=('Helvetica', 18),
        # background_color='black',
        right_click_menu=[
            "&Right",
            ["Right", "Edit Me", "!&Click", "&Menu", "E&xit", "Properties"],
        ],
        # transparent_color= '#9FB8AD',
        resizable=True,
        keep_on_top=False,
        element_justification="left",  # justify contents to the left
        metadata="My window metadata",
        finalize=True,
        # grab_anywhere=True,
        enable_close_attempted_event=True,
        modal=False,
        # ttk_theme=THEME_CLASSIC,
        # scaling=2,
        # icon=PSG_DEBUGGER_LOGO,
        # icon=PSGDebugLogo,
    )
    # window['-SPONSOR-'].set_cursor(cursor='hand2')
    window._see_through = False
    return window


# M"""""`'"""`YM          oo
# M  mm.  mm.  M
# M  MMM  MMM  M .d8888b. dP 88d888b.
# M  MMM  MMM  M 88'  `88 88 88'  `88
# M  MMM  MMM  M 88.  .88 88 88    88
# M  MMM  MMM  M `88888P8 dP dP    dP
# MMMMMMMMMMMMMM


def main():
    """
    The PySimpleGUI "Test Harness".  This is meant to be a super-quick test of the Elements.
    """
    forced_modal = DEFAULT_MODAL_WINDOWS_FORCED
    # set_options(force_modal_windows=True)
    window = _create_main_window()
    set_options(keep_on_top=True)
    graph_elem = window["+GRAPH+"]
    i = 0
    graph_figures = []
    # Don't use the debug window
    # Print('', location=(0, 0), font='Courier 10', size=(100, 20), grab_anywhere=True)
    # print(window.element_list())
    while True:  # Event Loop
        event, values = window.read(timeout=5)
        if event != TIMEOUT_KEY:
            print(event, values)
            # Print(event, text_color='white', background_color='red', end='')
            # Print(values)
        if (
                event == WIN_CLOSED
                or event == WIN_CLOSE_ATTEMPTED_EVENT
                or event == "Exit"
                or (event == "-BMENU-" and values["-BMENU-"] == "Exit")
        ):
            break
        if i < graph_elem.CanvasSize[0]:
            x = i % graph_elem.CanvasSize[0]
            fig = graph_elem.draw_line(
                (x, 0),
                (x, random.randint(0, graph_elem.CanvasSize[1])),
                width=1,
                color="#{:06x}".format(random.randint(0, 0xFFFFFF)),
            )
            graph_figures.append(fig)
        else:
            x = graph_elem.CanvasSize[0]
            graph_elem.move(-1, 0)
            fig = graph_elem.draw_line(
                (x, 0),
                (x, random.randint(0, graph_elem.CanvasSize[1])),
                width=1,
                color="#{:06x}".format(random.randint(0, 0xFFFFFF)),
            )
            graph_figures.append(fig)
            graph_elem.delete_figure(graph_figures[0])
            del graph_figures[0]
        window["+PROGRESS+"].UpdateBar(i % 800)
        window.Element("-IMAGE-").UpdateAnimation(
            DEFAULT_BASE64_LOADING_GIF, time_between_frames=50
        )
        if event == "Button":
            window.Element("-TEXT1-").SetTooltip("NEW TEXT")
            window.Element("-MENU-").Update(visible=True)
        elif event == "Popout":
            show_debugger_popout_window()
        elif event == "Launch Debugger":
            show_debugger_window()
        elif event == "About...":
            popup(
                "About this program...",
                "You are looking at the test harness for the PySimpleGUI program",
                version,
                keep_on_top=True,
                image=DEFAULT_BASE64_ICON,
            )
        elif event.startswith("See"):
            window._see_through = not window._see_through
            window.set_transparent_color(
                theme_background_color() if window._see_through else ""
            )
        elif event == "Popup":
            popup("This is your basic popup", keep_on_top=True)
        elif event == "Get File":
            popup_scrolled("Returned:", popup_get_file("Get File", keep_on_top=True))
        elif event == "Get Folder":
            popup_scrolled(
                "Returned:", popup_get_folder("Get Folder", keep_on_top=True)
            )
        elif event == "Get Date":
            popup_scrolled("Returned:", popup_get_date(keep_on_top=True))
        elif event == "Get Text":
            popup_scrolled(
                "Returned:", popup_get_text("Enter some text", keep_on_top=True)
            )

        elif event in ("-EMOJI-HEARTS-", "-HEART-", "-PYTHON HEARTS-"):
            popup_scrolled(
                "Oh look!  It's a Udemy discount coupon!",
                "522B20BF5EF123C4AB30",
                "A personal message from Mike -- thank you so very much for supporting PySimpleGUI!",
                title="Udemy Coupon",
                image=EMOJI_BASE64_MIKE,
                keep_on_top=True,
            )
        elif event == "Themes":
            search_string = popup_get_text(
                "Enter a search term or leave blank for all themes",
                "Show Available Themes",
                keep_on_top=True,
            )
            if search_string is not None:
                theme_previewer(search_string=search_string)
        elif event == "Theme Swatches":
            theme_previewer_swatches()
        elif event == "Switch Themes":
            window.close()
            _main_switch_theme()
            window = _create_main_window()
            graph_elem = window["+GRAPH+"]
        elif event == "-HIDE TABS-":
            window["-TAB GROUP COL-"].update(
                visible=window["-TAB GROUP COL-"].metadata == True
            )
            window["-TAB GROUP COL-"].metadata = not window["-TAB GROUP COL-"].metadata
            window["-HIDE TABS-"].update(
                text=SYMBOL_UP if window["-TAB GROUP COL-"].metadata else SYMBOL_DOWN
            )
        elif event == "SDK Reference":
            main_sdk_help()
        elif event == "Global Settings":
            if main_global_pysimplegui_settings():
                theme(
                    pysimplegui_user_settings.get("-theme-", OFFICIAL_PYSIMPLEGUI_THEME)
                )
                window.close()
                window = _create_main_window()
                graph_elem = window["+GRAPH+"]
            else:
                Window("", layout=[[Multiline()]], alpha_channel=0).read(
                    timeout=1, close=True
                )
        elif event.startswith("P "):
            if event == "P ":
                popup("Normal Popup - Modal", keep_on_top=True)
            elif event == "P NoTitle":
                popup_no_titlebar("No titlebar", keep_on_top=True)
            elif event == "P NoModal":
                set_options(force_modal_windows=False)
                popup(
                    "Normal Popup - Not Modal",
                    "You can interact with main window menubar ",
                    "but will have no effect immediately",
                    "button clicks will happen after you close this popup",
                    modal=False,
                    keep_on_top=True,
                )
                set_options(force_modal_windows=forced_modal)
            elif event == "P NoBlock":
                popup_non_blocking(
                    "Non-blocking",
                    "The background window should still be running",
                    keep_on_top=True,
                )
            elif event == "P AutoClose":
                popup_auto_close(
                    "Will autoclose in 3 seconds",
                    auto_close_duration=3,
                    keep_on_top=True,
                )
        elif event == "Versions for GitHub":
            main_get_debug_data()
        elif event == "Edit Me":
            execute_editor(__file__)
        elif event == "Open GitHub Issue":
            window.minimize()
            main_open_github_issue()
            window.normal()
        elif event == "Show Notification Again":
            if not running_trinket():
                pysimplegui_user_settings.set("-upgrade info seen-", False)
            __show_previous_upgrade_information()
        elif event == "-UPGRADE SHOW ONLY CRITICAL-":
            if not running_trinket():
                pysimplegui_user_settings.set(
                    "-upgrade show only critical-",
                    values["-UPGRADE SHOW ONLY CRITICAL-"],
                )

        i += 1
        # _refresh_debugger()
    print("event = ", event)
    window.close()
    set_options(force_modal_windows=forced_modal)


# ------------------------ PEP8-ify The SDK ------------------------#

ChangeLookAndFeel = change_look_and_feel
ConvertArgsToSingleString = convert_args_to_single_string
EasyPrint = easy_print
Print = easy_print
eprint = easy_print
sgprint = easy_print
PrintClose = easy_print_close
sgprint_close = easy_print_close
EasyPrintClose = easy_print_close
FillFormWithValues = fill_form_with_values
GetComplimentaryHex = get_complimentary_hex
ListOfLookAndFeelValues = list_of_look_and_feel_values
ObjToString = obj_to_string
ObjToStringSingleObj = obj_to_string_single_obj
OneLineProgressMeter = one_line_progress_meter
OneLineProgressMeterCancel = one_line_progress_meter_cancel
Popup = popup
PopupNoFrame = popup_no_titlebar
popup_no_frame = popup_no_titlebar
PopupNoBorder = popup_no_titlebar
popup_no_border = popup_no_titlebar
PopupAnnoying = popup_no_titlebar
popup_annoying = popup_no_titlebar
PopupAnimated = popup_animated
PopupAutoClose = popup_auto_close
PopupCancel = popup_cancel
PopupError = popup_error
PopupGetFile = popup_get_file
PopupGetFolder = popup_get_folder
PopupGetText = popup_get_text
PopupNoButtons = popup_no_buttons
PopupNoTitlebar = popup_no_titlebar
PopupNoWait = popup_non_blocking
popup_no_wait = popup_non_blocking
PopupNonBlocking = popup_non_blocking
PopupOK = popup_ok
PopupOKCancel = popup_ok_cancel
PopupQuick = popup_quick
PopupQuickMessage = popup_quick_message
PopupScrolled = popup_scrolled
PopupTimed = popup_auto_close
popup_timed = popup_auto_close
PopupYesNo = popup_yes_no

RGB = rgb
SetGlobalIcon = set_global_icon
SetOptions = set_options
sprint = popup_scrolled
ScrolledTextBox = popup_scrolled
TimerStart = timer_start
TimerStop = timer_stop
test = main
sdk_help = main_sdk_help

pysimplegui_user_settings = UserSettings(
    filename=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_FILENAME,
    path=DEFAULT_USER_SETTINGS_PYSIMPLEGUI_PATH,
)
# ------------------------ Set the "Official PySimpleGUI Theme Colors" ------------------------


theme(theme_global())
# ------------------------ Read the ttk scrollbar info ------------------------
_global_settings_get_ttk_scrollbar_info()

# ------------------------ Read the window watermark info ------------------------
_global_settings_get_watermark_info()

# See if running on Trinket. If Trinket, then use custom titlebars since Trinket doesn't supply any
if running_trinket():
    USE_CUSTOM_TITLEBAR = True

if tclversion_detailed.startswith("8.5"):
    warnings.warn(
        "You are running a VERY old version of tkinter {}. You cannot use PNG formatted images for example.  Please upgrade to 8.6.x".format(
            tclversion_detailed
        ),
        UserWarning,
    )

# Enables the correct application icon to be shown on the Windows taskbar
if running_windows():
    try:
        myappid = "mycompany.myproduct.subproduct.version"  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    except Exception as e:
        print("Error using the taskbar icon patch", e)

_read_mac_global_settings()

if _mac_should_set_alpha_to_99():
    # Applyting Mac OS 12.3+ Alpha Channel fix.  Sets the default Alpha Channel to 0.99
    set_options(alpha_channel=0.99)

# -------------------------------- ENTRY POINT IF RUN STANDALONE -------------------------------- #
if __name__ == "__main__":
    # To execute the upgrade from command line, type:
    # python -m PySimpleGUI.PySimpleGUI upgrade
    if len(sys.argv) > 1 and sys.argv[1] == "help":
        main_sdk_help()
        exit(0)
    main()
    exit(0)
